﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace Acop
{
  [ToolboxItem(true)]
  [ToolboxBitmap(typeof(CheckedListBox))]
  public partial class AcopBitRegister : AcopUserComponent
  {
#   pragma warning disable 1591
    private const int GAP = 5;
    private const int MAX_BIT_LENGTH = 64;

    private int xydim = 15;
    private int sBit = 0;
    private int bLength = 8;
    private int rSize = 0;
    private int maxDescSize = 0;

    private long rValue = 0L;

    private Boolean edit = false;
    private Boolean cmEdit = false;
    private Boolean sBitNum = true;
    private Boolean sDesc = false;
    private Boolean wasError = false;

    private String[] descriptionList = new String[] { };

    private List<Label> lBitNumList = new List<Label>();
    private List<DescriptionLabel> lDescList = new List<DescriptionLabel>();
    private List<RegisterBitLabel> bitList = new List<RegisterBitLabel>();

    private Color bitOnBackCol = Color.Green;
    private Color bitOffBackCol = Color.White;

    private ContextMenu cm;
    private MenuItem inputItem;

    private int lastLinkStatus = 0;

    public enum orientationEnum
    {
      Horizontal,
      Vertical
    }
    private orientationEnum ori = orientationEnum.Vertical;

    public AcopBitRegister()
    {
      InitializeComponent();
      SetDefaultUpdater(defaultUpdate);
      setControlsList(new Control[] { this }, true);
      StartLateInitialization();
    }

    private void InitializeComponent()
    {
      InitBits(sBit, bLength);
      this.Size = new Size(50, rSize);
      this.MinimumSize = new Size(50, rSize);
      AddAll();
      cm = new ContextMenu();
      inputItem = new MenuItem("Open input dialog");
      inputItem.Click += new EventHandler(this.OpenInputWindow);
      cm.MenuItems.Add(inputItem);
    }

    private void InitBits(int startBit, int bitLength)
    {
      int gm = 0;
      sBit = startBit;
      bLength = bitLength;

      for(int i = 0; i < bLength; i++)
      {
        RegisterBitLabel bit = new RegisterBitLabel();
        bit.Size = new Size(xydim, xydim);
        bit.BitValue = 0;
        bit.BitOnColor = bitOnBackCol;
        bit.BitOffColor = bitOffBackCol;

        Label lb = new Label();
        lb.Size = new Size(xydim + 5, xydim);
        lb.Font = new Font(FontFamily.GenericMonospace, 7, FontStyle.Regular);
        lb.Text = Convert.ToString(sBit + i);

        DescriptionLabel desc = new DescriptionLabel();
        desc.TextAlign = ContentAlignment.MiddleLeft;

        if (i == 8) gm = 1;
        else if (i == 16) gm = 2;
        else if (i == 24) gm = 3;
        else if (i == 32) gm = 4;
        else if (i == 40) gm = 5;
        else if (i == 48) gm = 6;
        else if (i == 56) gm = 7;

        int pos = i * xydim + (gm * GAP);

        switch (ori)
        {
          case orientationEnum.Horizontal:
            lb.Location = new Point(pos, 0);
            lb.TextAlign = ContentAlignment.BottomCenter;
            bit.Location = new Point(pos, xydim + 5);
            desc.Location = new Point(pos, 2 * xydim + 5);
            desc.RotateAngle = 90;
            break;
          case orientationEnum.Vertical:
            lb.Location = new Point(0, pos);
            lb.TextAlign = ContentAlignment.MiddleRight;
            bit.Location = new Point(xydim + 5, pos);
            desc.Location = new Point(2 * xydim + 5, pos);
            desc.RotateAngle = 0;
            break;
        }

        rSize = pos + xydim;

        bitList.Add(bit);
        lDescList.Add(desc);
        lBitNumList.Add(lb);
      }
    }

    private void UpdateBitRegister()
    {
      RemoveAll();
      InitBits(sBit, bLength);
      AddAll();
      switch (ori)
      {
        case orientationEnum.Horizontal:
          this.Size = new Size(rSize, 2 * xydim + 10 + maxDescSize);
          this.MinimumSize = new Size(rSize, 2 * xydim + 10 + maxDescSize);
          break;
        case orientationEnum.Vertical:
          this.Size = new Size(2 * xydim + 10 + maxDescSize, rSize);
          this.MinimumSize = new Size(2 * xydim + 10 + maxDescSize, rSize);
          break;
      }
    }

    private void AddAll()
    {
      foreach (RegisterBitLabel bit in bitList)
      {
        this.Controls.Add(bit);
      }
      if(sBitNum)
      {
        foreach (Label lb in lBitNumList)
        {
          this.Controls.Add(lb);
        }
      }
      if(sDesc)
      {
        int i = 0;
        foreach (DescriptionLabel lb in lDescList)
        {
          if (i < descriptionList.Length)
          {
            lb.NewText = descriptionList[i];
            if (lb.maxStringSize > maxDescSize) maxDescSize = lb.maxStringSize;
          }
          this.Controls.Add(lb);
          i++;
        }
      }
    }

    private void RemoveAll()
    {
      foreach (RegisterBitLabel bit in bitList)
      {
        this.Controls.Remove(bit);
        bit.Dispose();
      }
      foreach (Label lb in lBitNumList)
      {
        this.Controls.Remove(lb);
        lb.Dispose();
      }
      foreach (DescriptionLabel lb in lDescList)
      {
        this.Controls.Remove(lb);
        lb.Dispose();
      }
      bitList.Clear();
      lBitNumList.Clear();
      lDescList.Clear();
      rSize = 0;
      maxDescSize = 0;
    }

    private void SetBitOnColor(Color value)
    {
      this.bitOnBackCol = value;
      foreach (RegisterBitLabel bit in bitList)
      {
        bit.BitOnColor = value;
      }
    }
    private void SetBitOffColor(Color value)
    {
      this.bitOffBackCol = value;
      foreach (RegisterBitLabel bit in bitList)
      {
        bit.BitOffColor = value;
      }
    }

    private void SetRegisterValue(long value)
    {
      if (value == rValue) return;
      this.rValue = value;

      for(int i = 0; i < bLength; i++)
      {
        if((rValue & (long)Math.Pow(2.0, i+sBit)) > 0)
        {
          bitList[i].BitValue = 1;
        }
        else
        {
          bitList[i].BitValue = 0;
        }
      }
    }
    public long GetRegisterValue()
    {
      long bVal = 0;
      for(int i = 0; i < bLength; i++)
      {
        if (bitList[i].BitValue == 1) bVal += (long)Math.Pow(2.0, i + sBit);
      }
      return bVal;
    }

    private void SetIsEditable(Boolean value)
    {
      if (value == cmEdit) return;
      cmEdit = value;
      if (cmEdit) this.ContextMenu = cm;
      else this.ContextMenu = null;
    }

    private void SetStartBit(int value)
    {
      if (value < 0 || value == sBit) return;
      sBit = value;
      UpdateBitRegister();
    }
    private void SetBitLength(int value)
    {
      if (value < 1 || value == bLength || value > MAX_BIT_LENGTH) return;
      bLength = value;
      UpdateBitRegister();
    }

    private void ShowBitNum(Boolean value)
    {
      if (value == sBitNum) return;
      sBitNum = value;
      foreach (Label lb in lBitNumList)
      {
        if (sBitNum) this.Controls.Add(lb);
        else this.Controls.Remove(lb);
      }
    }
    private void ShowBitDesc(Boolean value)
    {
      if (value == sDesc) return;
      sDesc = value;
      UpdateBitRegister();
    }

    private void setBitSize(int value)
    {
      if (value == xydim || value < 10) return;
      xydim = value;
      UpdateBitRegister();
    }

    private String[] GetDescArray()
    {
      return descriptionList;
    }
    private void SetDescArray(String[] items)
    {
      descriptionList = items;
      if(descriptionList.Length == 0)
      {
        foreach(DescriptionLabel lb in lDescList)
        {
          lb.NewText = "";
        }
      }

      for (int i = 0; i < descriptionList.Length; i++)
      {
        if (i == lDescList.Count) break;
        lDescList[i].NewText = descriptionList[i];
      }
      UpdateBitRegister();
    }

    private void SetErrorState(Boolean error)
    {
      foreach(RegisterBitLabel bit in bitList)
      {
        bit.HasError = error;
      }
    }

    private void SetOrientation(orientationEnum value)
    {
      if(value == ori) return;
      ori = value;
      UpdateBitRegister();
    }

    public void ApplyErrorStatus(String sts, Boolean applyToTTT)
    {
      SetErrorState(true);
      wasError = true;
      if (applyToTTT) applyToolTipText(sts);
    }

    private void defaultUpdate(AcopControl.AcopLink lnk)
    {
      if (!lateInitializationDone) return;
      if (getAutoUpdate())
      {
        int lnkSts = lnk.GetStatus();
        if (lnkSts != 0 && ErrorValue.Length == 0)
        {
          ApplyErrorStatus(lnk.GetStatusString(), lnkSts != lastLinkStatus);
        }
        else
        {
          if(wasError)
          {
            SetErrorState(false);
            wasError = false;
          }
          if (lnkSts != lastLinkStatus) applyToolTipText(null);
          double v = getValueAsDouble(lnk);
          if(getValueIsInteger())
          {
            SetRegisterValue((long)v);
          }
        }
      }
    }

    public void ApplySendValue(object sender, long value)
    {
      int cc = ApplyRegisterSettings(sender, value);
    }

    public int ApplyRegisterSettings(object sender, long value)
    {
      Type type = getValueType(); // as long as getValueAsDouble() is called in the updater this will be something ...
      if (type == null) type = typeof(Int32);
      object inpt = AcopControl.AcopLink.GetDataObjectFromString(type, value.ToString());
      AcopControl.AcopLink lnk = AcopControl.AcopLink.GetAcopLink(LinkProtocol, LinkAddress, LinkProperty, null, inpt, 0, this);
      return applySetting(sender, inpt, lnk);
    }

    private void AcopMouseClick(object sender, MouseEventArgs e)
    {
      if (e.Button != MouseButtons.Left) return;
      RegisterBitLabel bit = sender as RegisterBitLabel;
      if (bit.BitValue == 0) bit.BitValue = 1;
      else bit.BitValue = 0;
      rValue = GetRegisterValue();
    }

    public void ApplyLockClick(object sender)
    {
      applyLock((CheckBox)sender);
    }

    /// <summary>
    /// If true, the editing of the bits is activated
    /// </summary>
    /// <param name="mode"></param>
    public void setBitEditMode(Boolean mode)
    {
      if (mode == edit) return;
      edit = mode;
      foreach (RegisterBitLabel bit in bitList)
      {
        if (edit) bit.MouseClick += new MouseEventHandler(this.AcopMouseClick);
        else bit.MouseClick -= new MouseEventHandler(this.AcopMouseClick);
      }
    }

    private void OpenInputWindow(object sender, EventArgs e)
    {
      InputDialog id = new InputDialog(this);
      id.StartPosition = FormStartPosition.CenterParent;
      id.ShowDialog();
    }

    #region Properties
    [Category("Acop.Display")]
    [Description("The width and height of the bits")]
    [DefaultValue(15)]
    public int BitSize { get { return xydim; } set { setBitSize(value); } }
    [Category("Acop.Display")]
    [Description("The color if the bit has the value 1")]
    [DefaultValue(typeof(Color), "0x008000")]
    public Color BitOnColor { get { return bitOnBackCol; } set { SetBitOnColor(value); } }
    [Category("Acop.Display")]
    [Description("The color if the bit has the value 0")]
    [DefaultValue(typeof(Color), "0xFFFFFF")]
    public Color BitOffColor { get { return bitOffBackCol; } set { SetBitOffColor(value); } }
    [Category("Acop.Display")]
    [Description("The register value")]
    [DefaultValue(0L)]
    public long RegisterValue { get { return rValue; } set { SetRegisterValue(value); } }
    [Category("Acop.Display")]
    [Description("If true allows to set the bits over the contextmenu")]
    [DefaultValue(false)]
    public Boolean IsEditable { get { return cmEdit; } set { SetIsEditable(value); } }
    [Category("Acop.Display")]
    [Description("Specifies the start bit")]
    [DefaultValue(0)]
    public int DisplayedFirstBit { get { return sBit; } set { SetStartBit(value); } }
    [Category("Acop.Display")]
    [Description("The number of bits to be displayed")]
    [DefaultValue(8)]
    public int DisplayedLength { get { return bLength; } set { SetBitLength(value); } }
    [Category("Acop.Display")]
    [Description("Set the register orientation")]
    [DefaultValue(orientationEnum.Vertical)]
    public orientationEnum Orientation { get { return ori; } set { SetOrientation(value); } }
    [Category("Acop.Display")]
    [Description("Show the bit numbers")]
    [DefaultValue(true)]
    public Boolean ShowBitNumbers { get { return sBitNum; } set { ShowBitNum(value); } }
    [Category("Acop.Display")]
    [Description("Show the description labels for each bit")]
    [DefaultValue(false)]
    public Boolean ShowBitDescription { get { return sDesc; } set { ShowBitDesc(value); } }
    [Category("Acop.Display")]
    [Description("The description of the individual bits")]
    [DesignerSerializationVisibility(DesignerSerializationVisibility.Content)]
    [Editor("System.Windows.Forms.Design.StringArrayEditor, System.Design, Version = 2.0.0.0, Culture = neutral, PublicKeyToken = b03f5f7f11d50a3a", typeof(System.Drawing.Design.UITypeEditor))]
    public String[] DescriptionList { get { return GetDescArray(); } set { SetDescArray(value); } }

    [Category("Acop.Display")]
    [Browsable(false), EditorBrowsable(EditorBrowsableState.Never)]
    public new int Precision { get; set; }
    [Category("Acop.Display")]
    [Browsable(false), EditorBrowsable(EditorBrowsableState.Never)]
    public new String Description { get; set; }
    [Category("Acop.Display")]
    [Browsable(false), EditorBrowsable(EditorBrowsableState.Never)]
    public new String Units { get; set; }
    [Category("Acop.Display")]
    [Browsable(false), EditorBrowsable(EditorBrowsableState.Never)]
    public new Boolean DisplayAsMultiChannel { get; set; }
    [Category("Acop.Transport")]
    [Browsable(false), EditorBrowsable(EditorBrowsableState.Never)]
    public new Boolean LinkGrouped { get; set; }
    #endregion
  }

  [ToolboxItem(false)]
  public partial class RegisterBitLabel : Label
  {
#   pragma warning disable 1591
    private Boolean error = false;
    private int lBitValue = 0;

    private Pen grayPen = new Pen(Color.Gray, 0.5f);
    private Pen errorPen = new Pen(Color.Red, 0.5f);

    private Color bitOnBackCol = Color.Green;
    private Color bitOffBackCol = Color.White;

    private Brush errorBrush = new SolidBrush(Color.White);

    private delegate void RefreshDelegate(int value);
    private delegate void ErrorDelegate(Boolean value);

    public RegisterBitLabel()
    {
      
    }

    private void setBitValue(int value)
    {
      if (this.InvokeRequired)
      {
        Invoke(new RefreshDelegate(setBitValue), value);
        return;
      }
      else
      {
        if (value == 0)
        {
          lBitValue = value;
          BackColor = bitOffBackCol;
        }
        if (value == 1)
        {
          lBitValue = value;
          BackColor = bitOnBackCol;
        }
      }
    }

    private void setError(Boolean value)
    {
      if (value == error) return;
      error = value;
      if(this.InvokeRequired)
      {
        Invoke(new ErrorDelegate(setError), value);
        return;
      }
      else
      {
        this.Invalidate();
      }
    }

    protected override void OnPaint(PaintEventArgs e)
    {
      base.OnPaint(e);
      Graphics g = e.Graphics;
      
      if(!error) g.DrawRectangle(grayPen, 0, 0, this.Width - 1, this.Height - 1);
      else
      {
        g.FillRectangle(errorBrush, 0, 0, this.Width - 1, this.Height - 1);
        g.DrawRectangle(errorPen, 0, 0, this.Width - 1, this.Height - 1);
        g.DrawLine(errorPen, 0, 0, this.Width - 1, this.Height - 1);
        g.DrawLine(errorPen, 0, this.Height - 1, this.Width - 1, 0);
      }
    }

    public int BitValue { get { return lBitValue; } set { setBitValue(value); } }
    public Boolean HasError { get { return error; } set { setError(value); } }
    public Color BitOnColor { get { return bitOnBackCol; } set { bitOnBackCol = value; } }
    public Color BitOffColor { get { return bitOffBackCol; } set { bitOffBackCol = value; } }
  }

  [ToolboxItem(false)]
  public partial class DescriptionLabel : Label 
  {
    //source frome: https://stackoverflow.com/questions/1371943/c-sharp-vertical-label-in-a-windows-forms
#   pragma warning disable 1591
    private int m_RotateAngle = 0;
    private string m_NewText = string.Empty;

    public int RotateAngle { get { return m_RotateAngle; } set { m_RotateAngle = value; Invalidate(); } }
    public string NewText { get { return m_NewText; } set { m_NewText = value; Invalidate(); } }
    public int maxStringSize { get { return getMaxStringSize(); } }

    private int getMaxStringSize()
    {
      int stringSize = 0;
      using (Graphics g = CreateGraphics())
      {
        SizeF size = g.MeasureString(m_NewText, this.Font);
        stringSize = (int)size.Width;
      }
      return stringSize;
    }

    protected override void OnPaint(PaintEventArgs e)
    {
      Func<double, double> DegToRad = (angle) => Math.PI * angle / 180.0;

      Brush b = new SolidBrush(this.ForeColor);
      SizeF size = e.Graphics.MeasureString(this.NewText, this.Font, this.Parent.Width);

      int normalAngle = ((RotateAngle % 360) + 360) % 360;
      double normaleRads = DegToRad(normalAngle);

      int hSinTheta = (int)Math.Ceiling((size.Height * Math.Sin(normaleRads)));
      int wCosTheta = (int)Math.Ceiling((size.Width * Math.Cos(normaleRads)));
      int wSinTheta = (int)Math.Ceiling((size.Width * Math.Sin(normaleRads)));
      int hCosTheta = (int)Math.Ceiling((size.Height * Math.Cos(normaleRads)));

      int rotatedWidth = Math.Abs(hSinTheta) + Math.Abs(wCosTheta);
      int rotatedHeight = Math.Abs(wSinTheta) + Math.Abs(hCosTheta);

      this.Width = rotatedWidth;
      this.Height = rotatedHeight;

      int numQuadrants =
          (normalAngle >= 0 && normalAngle < 90) ? 1 :
          (normalAngle >= 90 && normalAngle < 180) ? 2 :
          (normalAngle >= 180 && normalAngle < 270) ? 3 :
          (normalAngle >= 270 && normalAngle < 360) ? 4 :
          0;

      int horizShift = 0;
      int vertShift = 0;

      if (numQuadrants == 1)
      {
        horizShift = Math.Abs(hSinTheta);
      }
      else if (numQuadrants == 2)
      {
        horizShift = rotatedWidth;
        vertShift = Math.Abs(hCosTheta);
      }
      else if (numQuadrants == 3)
      {
        horizShift = Math.Abs(wCosTheta);
        vertShift = rotatedHeight;
      }
      else if (numQuadrants == 4)
      {
        vertShift = Math.Abs(wSinTheta);
      }

      e.Graphics.TranslateTransform(horizShift, vertShift);
      e.Graphics.RotateTransform(this.RotateAngle);

      e.Graphics.DrawString(this.NewText, this.Font, b, 0f, 0f);
      base.OnPaint(e);
    }
  }

  [ToolboxItem(false)]
  public partial class InputDialog : Form
  {
    private long rValue = 0;

    private AcopBitRegister inputReg;
    private AcopBitRegister mainReg;

    private CheckBox lockCheck = new CheckBox();

    private Button setBtn = new Button();
    private Button cancelBtn = new Button();

    public InputDialog(AcopBitRegister reg)
    {
      mainReg = reg;
      rValue = mainReg.RegisterValue;
      InitializeComponent();
      inputReg.RegisterValue = rValue;
    }

    private void InitializeComponent()
    {
      this.Text = "Bit input";
      this.ControlBox = false;
      this.MaximizeBox = false;
      this.MinimizeBox = false;
      this.CancelButton = cancelBtn;

      inputReg = new AcopBitRegister();
      inputReg.Location = new Point(5, 5);
      inputReg.DisplayedFirstBit = mainReg.DisplayedFirstBit;
      inputReg.DisplayedLength = mainReg.DisplayedLength;
      inputReg.DescriptionList = mainReg.DescriptionList;
      inputReg.ShowBitDescription = true;
      inputReg.setBitEditMode(true);

      lockCheck.Location = new Point(5, inputReg.Height + 10);
      lockCheck.Text = "lock access";
      lockCheck.Size = new Size(100, 20);
      lockCheck.MouseClick += new MouseEventHandler(this.ApplyLockClick);

      setBtn.Location = new Point(5, lockCheck.Location.Y + 30);
      setBtn.Size = new Size(100, 20);
      setBtn.Text = "Set";
      setBtn.Click += new EventHandler(this.ApplySetClick);

      cancelBtn.Location = new Point(5, setBtn.Location.Y + 25);
      cancelBtn.Size = new Size(100, 20);
      cancelBtn.Text = "Cancel";
      cancelBtn.Click += new EventHandler(this.ApplyCancelClick);

      this.Controls.Add(inputReg);
      //this.Controls.Add(lockCheck);
      this.Controls.Add(setBtn);
      this.Controls.Add(cancelBtn);
      this.Size = new Size(120, cancelBtn.Location.Y + 70);
    }

    private void ApplySetClick(object sender, EventArgs e)
    {
      mainReg.ApplySendValue(sender, inputReg.GetRegisterValue());
      this.Dispose();
    }

    private void ApplyCancelClick(object sender, EventArgs e)
    {
      this.Dispose();
    }

    private void ApplyLockClick(object sender, EventArgs e)
    {
      mainReg.ApplyLockClick(sender);
    }
  }
}
