﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace Acop
{
  [ToolboxItem(true)]
  [ToolboxBitmap(typeof(Button))]
  public partial class AcopButton : AcopUserComponent
  {
#   pragma warning disable 1591
    public AcopButton()
    {
      InitializeComponent();
      SetDefaultUpdater(defaultUpdate);
      setControlsList(new Control[] { Button }, true);
      Button.Text = setTexts[setValuesIndex];
      SetLateInitializerProlog(initTransportInfo);
      StartLateInitialization();
    }
    public void applyErrorStatus(String sts)
    {
      applyToolTipText(sts);
      Button.ForeColor = Color.DarkRed;
    }

    private void initTransportInfo()
    {
      if (!IsDesignMode())
      {
        String[] labels = null;
        AcopTransport.DisplaySettings dsp = null;
        if (!AutoLinkAttach && !AutoLinkUpdate)
        {
          displaySettingsAcquired = true;
          return;
        }
        dsp = acquireAndApplyDisplaySettings(LinkAddress, LinkProperty, ref labels, false);
        if (dsp == null || dsp.SettingsStale()) return;
        displaySettingsAcquired = true;
     }
    }
    private void defaultUpdate(AcopControl.AcopLink lnk)
    {
      if (!lateInitializationDone) return;
      if (getAutoUpdate())
      {
        int lnkSts = lnk.GetStatus();
        if (lnkSts != 0)
        {
          applyErrorStatus(lnk.GetStatusString());
        }
        else
        {
          applyToolTipText(null);
          double v = getValueAsDouble(lnk);
          String txt;
          if (ValueIsNumber())
          {
            txt = getLabelString(v, 6);
          }
          else
          {
            String[] txtlines = lnk.GetOutputDataAsString();
            txt = txtlines[0];
          }
          if (lnk.GetDisplayId() < 0)
          {
            if (!displaySettingsAcquired) initTransportInfo();
            if (Button.Text.Length == 0)
            {
              Button.Text = "-> " + setValues[setValuesIndex];
            }
            //if ((txt.CompareTo(setValues[setValuesIndex]) != 0))
            //{
            //  if (isToggleButton) toggleButtonState();
            //}
          }

          Button.ForeColor = (txt.Equals(setValues[setValuesIndex], StringComparison.OrdinalIgnoreCase)) ? labelOutOfSyncForeColor : labelInSyncForeColor;

        }
      }
    }

    private Boolean isToggleButton = false;
    private String[] setTexts = new String[] { "Set", "Set" };
    private String[] setValues = new String[] { "", "" };
    private int setValuesIndex = 0;
    private String getPrimaryValue(String[] tgt) { return tgt[0]; }
    private void setPrimaryValue(String[] tgt, String value)
    {
      if (value == null) value = "";
      tgt[0] = value;
    }
    private String getSecondaryValue(String[] tgt) { return tgt[1]; }
    private void setSecondaryValue(String[] tgt, String value)
    {
      if (value == null) value = "";
      tgt[1] = value;
    }
    private void toggleButtonState()
    {
      setValuesIndex = (setValuesIndex + 1) % 2;
      Button.Text = setTexts[setValuesIndex];
    }
    //private object[] setObjects = new object[] { null, null };
    private Boolean showErrorMessages = true;
    private void ButtonClicked(object sender, EventArgs e)
    {
      object inpt = setValues[setValuesIndex];
      if (((String)inpt).Length == 0) inpt = null;
      if (inpt != null)
      {
        Type type = getValueType(); // as long as getValueAsDouble() is called in the updater this will be something ...
        if (type == null) type = getValueType((String)inpt);
        inpt = type == null ? null : AcopControl.AcopLink.GetDataObjectFromString(type, setValues[setValuesIndex]);
      }
      AcopControl.AcopLink lnk = null;
      if (!hasTargetComponent())
      {
        lnk = AcopControl.AcopLink.GetAcopLink(LinkProtocol, LinkAddress, LinkProperty, null, inpt, 0, this);
      }
      if (applySetting(sender, inpt, lnk) == 0)
      {
        if (isToggleButton) toggleButtonState();
      }
      else
      {
        if (showErrorMessages)
        {
          String err = "apply setting "+setValues[setValuesIndex];
          if (setValues[setValuesIndex].Length > 0) err += " ";
          if (lnk != null)
          {
            err += "to " + LinkAddress + " [" + LinkProperty+ "] : " +lnk.GetStatusString();
          }
          else if (hasTargetComponent())
          {
            err += "to " + getTargetComponent() + " failed"; 
          }
          else
          {
            err += "failed";
          }
          MessageBox.Show(err);
        }
      }
    }

    private Color labelInSyncForeColor = Color.DarkBlue;
    private Color labelOutOfSyncForeColor = Color.DarkRed;
    private Color buttonBackgroundColor = SystemColors.Control;
    private FlatStyle btnStyle = FlatStyle.Standard;
    
    private void SetButtonFlatStyle(FlatStyle value)
    {
      Button.FlatStyle = btnStyle = value;
    }
    /// <summary>
    /// Determines whether a message box appears when a button click results in an error
    /// </summary>
    [Category("Acop.Display")]
    [Description("Determines whether a message box appears when a button click results in an error")]
    [DefaultValue(true)]
    public Boolean ShowErrorMessages { get { return showErrorMessages; } set { showErrorMessages = value; } }

    /// <summary>
    /// Determines whether this button toggles between primary and secondary settings
    /// </summary>
    [Category("Acop.Display")]
    [Description("Determines whether this button toggles between primary and secondary settings")]
    [DefaultValue(false)]
    public Boolean IsToggleButton { get { return isToggleButton; } set { isToggleButton = value; } }
    /// <summary>
    /// Primary setting to be used when button is pressed
    /// </summary>
    [Category("Acop.Display")]
    [Description("Primary setting to be used when button is pressed")]
    public String PrimarySetValue { get { return getPrimaryValue(setValues); } set { setPrimaryValue(setValues, value); } }
    /// <summary>
    /// Primary button caption displayed when the primary setting is active
    /// </summary>
    [Category("Acop.Display")]
    [Description("Primary button caption displayed when the primary setting is active")]
    [DefaultValue("Set")]
    public String PrimaryButtonText { get { return getPrimaryValue(setTexts); } set { setPrimaryValue(setTexts, value); Button.Text = value; } }
    /// <summary>
    /// Secondary setting to be used when button is pressed (toggle button only)
    /// </summary>
    [Category("Acop.Display")]
    [Description("Secondary setting to be used when button is pressed (toggle button only)")]
    public String SecondarySetValue { get { return getSecondaryValue(setValues); } set { setSecondaryValue(setValues, value); } }
    /// <summary>
    /// Secondary button caption displayed when the primary setting is active (toggle button only)
    /// </summary>
    [Category("Acop.Display")]
    [Description("Secondary button caption displayed when the primary setting is active (toggle button only)")]
    [DefaultValue("Set")]
    public String SecondaryButtonText { get { return getSecondaryValue(setTexts); } set { setSecondaryValue(setTexts, value); } }
    private void setLabelInSyncForeColor(Color value)
    {
      labelInSyncForeColor = value;
    }
    /// <summary>
    /// Button background color
    /// </summary>
    [Category("Acop.Display")]
    [Description("The background color")]
    public Color ButtonBackColor { get { return buttonBackgroundColor; } set { setButtonBackColor(value); } }
    private void setButtonBackColor(Color value)
    {
      buttonBackgroundColor = value;
      Button.BackColor = buttonBackgroundColor;
    }
    /// <summary>
    /// ForeColor to use if setting is synchronized with the current readback value
    /// </summary>
    [Category("Acop.Display")]
    [Description("ForeColor to use if setting is synchronized with the current readback value")]
    [DefaultValue(typeof(Color), "0x00008B")]
    public Color LabelInSyncForeColor { get { return labelInSyncForeColor; } set { setLabelInSyncForeColor(value); } }
    private void setLabelOutOfSyncForeColor(Color value)
    {
      labelOutOfSyncForeColor = value;
    }
    /// <summary>
    /// Gets or sets the flat style appearance of the button control
    /// </summary>
    [Category("Acop.Display")]
    [Description("Determines the appearance of the control when a user moves the mouse over the control and clicks.")]
    [DefaultValue(FlatStyle.Standard)]
    public FlatStyle ButtonFlatStyle { get { return btnStyle;  }  set { SetButtonFlatStyle(value); } }
    /// <summary>
    /// ForeColor to use if setting is not synchronized with the current readback value
    /// </summary>
    [Category("Acop.Display")]
    [Description("ForeColor to use if setting is not synchronized with the current readback value")]
    [DefaultValue(typeof(Color), "0x8B0000")]
    public Color LabelOutOfSyncForeColor { get { return labelOutOfSyncForeColor; } set { setLabelOutOfSyncForeColor(value); } }
    //[Category("Acop.Display")]
    //[Description("If not empty, should be in the form of 'ControlName.Property' : apply setting will set the control property to the slider value")]
    /// <summary>
    /// If not empty, should be in the form of 'ControlName.Property' : apply setting will set the control property to the slider value
    /// </summary>
    public String TargetComponent { get { return getTargetComponent(); } set { setTargetComponent(value); } }
    [Category("Acop.Display")]
    [Description("If not empty, attach the target control property to the slider")]
    [Editor(typeof(TargetControlEditor), typeof(System.Drawing.Design.UITypeEditor))]
    [DesignerSerializationVisibility(DesignerSerializationVisibility.Content)]
    [TypeConverter(typeof(TargetControlEditorConverter))]
    public TargetControlEditorValue Target { get { return getTargetControlEditorValue(); } set { setTargetControlEditorValue(value); } }

    [Category("Acop.Display")]
    [Browsable(false), EditorBrowsable(EditorBrowsableState.Never)]
    public new Boolean DisplayAsMultiChannel { get; set; }

    [Category("Acop.Display")]
    [Browsable(false), EditorBrowsable(EditorBrowsableState.Never)]
    public new int Precision { get; set; }

    [Category("Acop.Display")]
    [Browsable(false), EditorBrowsable(EditorBrowsableState.Never)]
    public new String Units { get; set; }

    [Category("Acop.Display")]
    [Browsable(false), EditorBrowsable(EditorBrowsableState.Never)]
    public new String ErrorValue { get; set; }
    [Category("Acop.Display")]
    [Browsable(false), EditorBrowsable(EditorBrowsableState.Never)]
    public new String Description { get; set; }

    [Category("Acop.Transport")]
    [Browsable(false), EditorBrowsable(EditorBrowsableState.Never)]
    public new Boolean LinkGrouped { get; set; }
  }
}
