﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace Acop.AcopChartUtil
{
  public partial class AcopAxis
  {
#   pragma warning disable 1591
    protected int axisIndex;       	// own axis index number
    protected int npoints;    		// display points
    protected internal Boolean axisON;			// on or off

    protected double[] Position;          // input display data
    protected Color[] penColor;
    protected int[] penWidth, penStyle;
    protected Pen[] axisStroke;
    private int strokeWidth, strokeStyle, penIndex;
    private Color lastColor;

    private AcopChart acop;

    public AcopAxis(AcopChart acop, int index)
    {
      this.acop = acop;
      npoints = 0;
      axisON = false;
      axisIndex = index;
    }

    protected internal int loadData(int size, Object PositionArray, Color[] ColorArray, int[] StyleArray, int[] WidthArray)
    {
      int i, ndata, dataChanged, penChange;

      dataChanged = 0;
      penChange = 0;

      if (npoints < size)
      {
        dataChanged++;
        npoints = size;
        Position = new double[npoints];
        penColor = new Color[npoints];
        penWidth = new int[npoints];
        penStyle = new int[npoints];
        for (i = 0; i < npoints; i++)
        {
          penColor[i] = acop.sca[0].gridLineColor;
          Position[i] = 0.0;
          penWidth[i] = 1;
          penStyle[i] = AcopConst.PS_DOT;
        }
        penChange++;
      }
      ndata = AcopHisto.getArrayLength(PositionArray);
      if (ndata > 0)
      {
        dataChanged++;
        if (npoints < ndata) ndata = npoints;
        AcopHisto.copyArray(Position, PositionArray, 0, ndata, ndata, 1);
      }
      ndata = AcopHisto.getArrayLength(ColorArray);
      if (ndata > 0)
      {
        dataChanged++;
        if (penColor[0] != ColorArray[0]) penChange++;
        if (npoints < ndata) ndata = npoints;
        for (i = 0; i < ndata; i++) penColor[i] = ColorArray[i];
      }
      ndata = AcopHisto.getArrayLength(StyleArray);
      if (ndata > 0)
      {
        dataChanged++;
        if (penStyle[0] != StyleArray[0]) penChange++;
        if (npoints < ndata) ndata = npoints;
        for (i = 0; i < ndata; i++) penStyle[i] = StyleArray[i];
      }
      ndata = AcopHisto.getArrayLength(WidthArray);
      if (ndata > 0)
      {
        dataChanged++;
        if (penWidth[0] != WidthArray[0]) penChange++;
        if (npoints < ndata) ndata = npoints;
        for (i = 0; i < ndata; i++) penWidth[i] = WidthArray[i];
      }
      if (penChange != 0) getStroke(0, true);
      return dataChanged;
    }

    private void getStroke(int i, Boolean creat)
    {
      strokeWidth = penWidth[i];
      strokeStyle = penStyle[i];
      lastColor = penColor[i];
      penIndex = i;
      if (creat) axisStroke[i] = AcopConst.createStroke(lastColor, strokeWidth, strokeStyle);
    }

    protected internal void drawAxis(Graphics a)
    {
      int xtemp, ytemp, lastWidth, lastStyle;
      double tmp;
      Color lastColor;

      if (axisON == false || npoints <= 0) return;
      lastColor = penColor[0];
      getStroke(0, false);
      for (int i = 0; i < npoints; i++)
      {
        tmp = acop.sca[axisIndex].log != 0 ? Math.Log(AcopConst.log_check(Position[i])) / Math.Log(10) : Position[i];
        if (tmp < acop.sca[axisIndex].min || tmp > acop.sca[axisIndex].max) continue;

        if (strokeWidth != penWidth[i] || strokeStyle != penStyle[i] || lastColor != penColor[i]) getStroke(i, true);
        if (axisIndex == 0)	// xaxis
        {
          xtemp = (int)((tmp - acop.sca[axisIndex].min) / acop.sca[axisIndex].dispSize * acop.aFrame.dxwidth + acop.aFrame.xleft + 0.01);
          a.DrawLine(axisStroke[penIndex], xtemp, acop.aFrame.histRect.Height, xtemp, 0);
        }
        else
        {
          ytemp = (int)(acop.aFrame.histRect.Height - ((tmp - acop.sca[axisIndex].min) / acop.sca[axisIndex].dispSize * acop.aFrame.histRect.Height));
          a.DrawLine(axisStroke[penIndex], acop.aFrame.xleft, ytemp, acop.aFrame.dxwidth, ytemp);
        }
      }
    }

  }
}
