﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace Acop.AcopChartUtil
{
  [ToolboxItem(false)]
  public partial class AcopChartContextMenu : ContextMenu
  {
    private AcopChart chart;

    private MenuItem AutoOnce;
    private MenuItem AutoAlways;
    private MenuItem DefaultScale;
    private MenuItem SingleUpdate;
    private MenuItem ChartProp;
    private MenuItem MoreOp;

    private ChartProperties cp;

    private Boolean showMoreOption = false;

    public AcopChartContextMenu(AcopChart chart)
    {
      this.chart = chart;

      AutoOnce = new MenuItem("AutoScale once");
      AutoOnce.Click += new EventHandler(this.applyAutoScaleOnce);

      AutoAlways = new MenuItem("AutoScale always");
      AutoAlways.Checked = false;
      AutoAlways.Click += new EventHandler(this.applyAutoScaleAlways);

      DefaultScale = new MenuItem("Apply default scale");
      DefaultScale.Checked = false;
      DefaultScale.Click += new EventHandler(this.applyDefaultScale);

      SingleUpdate = new MenuItem("Apply single update");
      SingleUpdate.Checked = false;
      SingleUpdate.Click += new EventHandler(this.applySingleUpdate);

      ChartProp = new MenuItem("Chart properties");
      ChartProp.Click += new EventHandler(this.applyChartProp);

      MoreOp = new MenuItem("More options");
      MoreOp.Click += new EventHandler(this.ApplyMoreOpt);

      this.MenuItems.Add(AutoOnce);
      this.MenuItems.Add(AutoAlways);
      this.MenuItems.Add(DefaultScale);
      this.MenuItems.Add(SingleUpdate);
      this.MenuItems.Add("-");
      this.MenuItems.Add(ChartProp);
    }

    private void applyAutoScaleOnce(object sender, EventArgs e)
    {
      DefaultScale.Checked = false;
      chart.AutoScaleUpdate = AcopChart.AutoScaleMode.once;
    }

    private void applyAutoScaleAlways(object sender, EventArgs e)
    {
      DefaultScale.Checked = false;
      if (AutoAlways.Checked)
      {
        AutoAlways.Checked = false;
        chart.AutoScaleUpdate = AcopChart.AutoScaleMode.off;
      }
      else
      {
        AutoAlways.Checked = true;
        chart.AutoScaleUpdate = AcopChart.AutoScaleMode.always;
      }
    }
    private void applyDefaultScale(object sender, EventArgs e)
    {
      AutoAlways.Checked = false;
      chart.AutoScaleUpdate = AcopChart.AutoScaleMode.off;
      chart.SetDefaultScale();
    }
    private void applySingleUpdate(object sender, EventArgs e)
    {
      SingleUpdate.Checked = false;
      chart.SingleUpdate();
    }
    private void applyChartProp(object sender, EventArgs e)
    {
      cp = new ChartProperties(chart);
      cp.ShowDialog();
    }

    private void ApplyMoreOpt(object sender, EventArgs e)
    {
      chart.OnMoreOptionsClick(sender, e);
    }

    private void ShowMoreOp(Boolean value)
    {
      showMoreOption = value;
      if (showMoreOption) this.MenuItems.Add(MoreOp);
      else this.MenuItems.Remove(MoreOp);
    }

    public Boolean ScaleAlways { get { return AutoAlways.Checked; } set { AutoAlways.Checked = value; } }
    public Boolean MoreOptions { get { return showMoreOption; } set { ShowMoreOp(value); } }
  }

  public class ChartProperties : Form
  {
    private AcopChart chart;

    private DataGridView dgv = new DataGridView();

    private Boolean linePropertyChanged = false;
    private String trendTime;

    public ChartProperties(AcopChart chart)
    {
      this.chart = chart;
      InitializeComponent();
      getChartProperties();
      getChartLinesProperties();
    }

    public void getChartProperties()
    {
      tb_XLab.Text = chart.XAxisLabel;
      tb_YLab.Text = chart.YAxisLabel;
      cb_XGrid.Checked = chart.XGrid;
      cb_YGrid.Checked = chart.YGrid;
      cb_XBestScale.Checked = chart.XBestScale;
      cb_YBestScale.Checked = chart.YBestScale;
      tb_YMin.Text = chart.YMin.ToString();
      tb_YMax.Text = chart.YMax.ToString();
      if (chart.AutoTrendUpdate)
      {
        tb_Trend.Enabled = true;
        tb_Trend.Text = chart.LinkTrendHours.ToString();
        trendTime = chart.LinkTrendHours.ToString();
        tb_XMin.Enabled = false;
        tb_XMax.Enabled = false;
      }
      else
      {
        tb_XMin.Text = chart.XMin.ToString();
        tb_XMax.Text = chart.XMax.ToString();
      }
      AcopChart.AutoScaleMode asm = chart.AutoScaleUpdate;
      this.c_AutoScale.SelectedIndex = (int)asm;
      AcopChart.AcopMarkerModeEnum amm = chart.CursorMarkerMode;
      this.c_Cursor.SelectedIndex = (int)amm;
    }

    private void btn_OK_Click(object sender, EventArgs e)
    {
      chart.XAxisLabel = tb_XLab.Text;
      chart.YAxisLabel = tb_YLab.Text;
      chart.XGrid = cb_XGrid.Checked;
      chart.YGrid = cb_YGrid.Checked;
      chart.XBestScale = cb_XBestScale.Checked;
      chart.YBestScale = cb_YBestScale.Checked;
      try
      {
        chart.YMin = Convert.ToDouble(tb_YMin.Text);
        chart.YMax = Convert.ToDouble(tb_YMax.Text);
        if (chart.AutoTrendUpdate)
        {
          if (!trendTime.Equals(tb_Trend.Text))
          {
            chart.LinkTrendHours = Convert.ToDouble(tb_Trend.Text);
          }
        }
        else
        {
          chart.XMin = Convert.ToDouble(tb_XMin.Text);
          chart.XMax = Convert.ToDouble(tb_XMax.Text);
        }
      }
      catch(Exception exc)
      {
        Console.WriteLine(exc.ToString());
      }
      chart.AutoScaleUpdate = (AcopChart.AutoScaleMode)c_AutoScale.SelectedIndex;
      chart.CursorMarkerMode = (AcopChart.AcopMarkerModeEnum)c_Cursor.SelectedIndex;

      setChartLinesProperties();

      this.Dispose();
    }

    private void btn_Cancel_Click(object sender, EventArgs e)
    {
      this.Dispose();
    }

    private void setChartLinesProperties()
    {
      if (!linePropertyChanged) return;
      foreach (DataGridViewRow row in dgv.Rows)
      {
        int i = Convert.ToInt32(row.Cells[0].Value.ToString());
        if (chart.histo[i].displayON == 1 && Convert.ToInt32(row.Cells[1].Value.ToString()) == 0) chart.ClearScreen(i);
        else if (chart.histo[i].displayON == 0 && Convert.ToInt32(row.Cells[1].Value.ToString()) == 1) chart.RecallScreen(i);
        chart.histo[i].setCurveColor(row.Cells[2].Style.BackColor);
        chart.histo[i].setCurveWidth(Convert.ToInt32(row.Cells[3].Value));
        chart.histo[i].applyStrokeChange();
      }
    }

    private void getChartLinesProperties()
    {
      for(int i=0; i<chart.histo.Length; i++)
      {
        if(chart.histo[i] != null)
        {
          dgv.Rows.Add(i, chart.histo[i].displayON, chart.histo[i].getCurveColor().Name, chart.histo[i].penWidth);
          dgv.Rows[i].Cells[2].Style.BackColor = chart.histo[i].getCurveColor();
        }
      }
    }

    private void dgv_CellContentClick(object sender, DataGridViewCellEventArgs e)
    {
      if (e.ColumnIndex < 0) return;
      var senderGrid = (DataGridView)sender;
      if (senderGrid.Columns[e.ColumnIndex] is DataGridViewButtonColumn && e.RowIndex >= 0)
      {
        ColorDialog cd = new ColorDialog();
        cd.ShowDialog();
        dgv.Rows[e.RowIndex].Cells[e.ColumnIndex].Style.BackColor = cd.Color;
        dgv.Rows[e.RowIndex].Cells[e.ColumnIndex].Value = cd.Color.Name;
        linePropertyChanged = true;
      }
      if(senderGrid.Columns[e.ColumnIndex] is DataGridViewCheckBoxColumn && e.RowIndex >= 0)
      {
        linePropertyChanged = true;
      }
      if(senderGrid.Columns[e.ColumnIndex] is DataGridViewTextBoxColumn && e.RowIndex >= 0)
      {
        if (e.ColumnIndex > 0) linePropertyChanged = true;
      }
    }

    private void dgv_CellSelectionChanged(object sender, EventArgs e)
    {
      this.dgv.ClearSelection();
    }

    private void InitializeComponent()
    {
      this.btn_OK = new Button();
      this.btn_Cancel = new Button();
      this.groupBox1 = new GroupBox();
      this.tb_XMin = new TextBox();
      this.label2 = new Label();
      this.cb_XBestScale = new CheckBox();
      this.cb_XGrid = new CheckBox();
      this.tb_XLab = new TextBox();
      this.label1 = new Label();
      this.tb_XMax = new TextBox();
      this.label3 = new Label();
      this.tb_YMax = new TextBox();
      this.label4 = new Label();
      this.tb_YMin = new TextBox();
      this.label5 = new Label();
      this.cb_YBestScale = new CheckBox();
      this.cb_YGrid = new CheckBox();
      this.tb_YLab = new TextBox();
      this.label6 = new Label();
      this.c_AutoScale = new ComboBox();
      this.label7 = new Label();
      this.gb_Lines = new GroupBox();
      this.label8 = new Label();
      this.c_Cursor = new ComboBox();
      this.tb_Trend = new TextBox();
      this.label9 = new Label();
      this.label10 = new Label();
      this.groupBox1.SuspendLayout();
      this.SuspendLayout();
      // 
      // btn_OK
      // 
      this.btn_OK.Location = new System.Drawing.Point(549, 12);
      this.btn_OK.Name = "btn_OK";
      this.btn_OK.Size = new System.Drawing.Size(75, 23);
      this.btn_OK.TabIndex = 0;
      this.btn_OK.Text = "OK";
      this.btn_OK.UseVisualStyleBackColor = true;
      this.btn_OK.Click += new EventHandler(this.btn_OK_Click);
      // 
      // btn_Cancel
      // 
      this.btn_Cancel.Location = new System.Drawing.Point(549, 41);
      this.btn_Cancel.Name = "btn_Cancel";
      this.btn_Cancel.Size = new System.Drawing.Size(75, 23);
      this.btn_Cancel.TabIndex = 1;
      this.btn_Cancel.Text = "Cancel";
      this.btn_Cancel.UseVisualStyleBackColor = true;
      this.btn_Cancel.Click += new EventHandler(this.btn_Cancel_Click);
      // 
      // tb_XMin
      // 
      this.tb_XMin.Location = new System.Drawing.Point(54, 62);
      this.tb_XMin.Name = "tb_XMin";
      this.tb_XMin.Size = new System.Drawing.Size(100, 20);
      this.tb_XMin.TabIndex = 5;
      // 
      // label2
      // 
      this.label2.AutoSize = true;
      this.label2.Location = new System.Drawing.Point(6, 65);
      this.label2.Name = "label2";
      this.label2.Size = new System.Drawing.Size(37, 13);
      this.label2.TabIndex = 4;
      this.label2.Text = "X Min:";
      // 
      // cb_XBestScale
      // 
      this.cb_XBestScale.AutoSize = true;
      this.cb_XBestScale.Location = new System.Drawing.Point(74, 39);
      this.cb_XBestScale.Name = "cb_XBestScale";
      this.cb_XBestScale.Size = new System.Drawing.Size(82, 17);
      this.cb_XBestScale.TabIndex = 3;
      this.cb_XBestScale.Text = "X Bestscale";
      this.cb_XBestScale.UseVisualStyleBackColor = true;
      // 
      // cb_XGrid
      // 
      this.cb_XGrid.AutoSize = true;
      this.cb_XGrid.Location = new System.Drawing.Point(9, 39);
      this.cb_XGrid.Name = "cb_XGrid";
      this.cb_XGrid.Size = new System.Drawing.Size(55, 17);
      this.cb_XGrid.TabIndex = 2;
      this.cb_XGrid.Text = "X Grid";
      this.cb_XGrid.UseVisualStyleBackColor = true;
      // 
      // tb_XLab
      // 
      this.tb_XLab.Location = new System.Drawing.Point(54, 13);
      this.tb_XLab.Name = "tb_XLab";
      this.tb_XLab.Size = new System.Drawing.Size(100, 20);
      this.tb_XLab.TabIndex = 1;
      // 
      // label1
      // 
      this.label1.AutoSize = true;
      this.label1.Location = new System.Drawing.Point(6, 16);
      this.label1.Name = "label1";
      this.label1.Size = new System.Drawing.Size(42, 13);
      this.label1.TabIndex = 0;
      this.label1.Text = "X label:";
      // 
      // tb_XMax
      // 
      this.tb_XMax.Location = new System.Drawing.Point(54, 88);
      this.tb_XMax.Name = "tb_XMax";
      this.tb_XMax.Size = new System.Drawing.Size(100, 20);
      this.tb_XMax.TabIndex = 7;
      // 
      // label3
      // 
      this.label3.AutoSize = true;
      this.label3.Location = new System.Drawing.Point(6, 91);
      this.label3.Name = "label3";
      this.label3.Size = new System.Drawing.Size(40, 13);
      this.label3.TabIndex = 6;
      this.label3.Text = "X Max:";
      // 
      // tb_YMax
      // 
      this.tb_YMax.Location = new System.Drawing.Point(220, 88);
      this.tb_YMax.Name = "tb_YMax";
      this.tb_YMax.Size = new System.Drawing.Size(100, 20);
      this.tb_YMax.TabIndex = 15;
      // 
      // label4
      // 
      this.label4.AutoSize = true;
      this.label4.Location = new System.Drawing.Point(173, 91);
      this.label4.Name = "label4";
      this.label4.Size = new System.Drawing.Size(40, 13);
      this.label4.TabIndex = 14;
      this.label4.Text = "Y Max:";
      // 
      // tb_YMin
      // 
      this.tb_YMin.Location = new System.Drawing.Point(220, 62);
      this.tb_YMin.Name = "tb_YMin";
      this.tb_YMin.Size = new System.Drawing.Size(100, 20);
      this.tb_YMin.TabIndex = 13;
      // 
      // label5
      // 
      this.label5.AutoSize = true;
      this.label5.Location = new System.Drawing.Point(172, 65);
      this.label5.Name = "label5";
      this.label5.Size = new System.Drawing.Size(37, 13);
      this.label5.TabIndex = 12;
      this.label5.Text = "Y Min:";
      // 
      // cb_YBestScale
      // 
      this.cb_YBestScale.AutoSize = true;
      this.cb_YBestScale.Location = new System.Drawing.Point(240, 39);
      this.cb_YBestScale.Name = "cb_YBestScale";
      this.cb_YBestScale.Size = new System.Drawing.Size(82, 17);
      this.cb_YBestScale.TabIndex = 11;
      this.cb_YBestScale.Text = "Y Bestscale";
      this.cb_YBestScale.UseVisualStyleBackColor = true;
      // 
      // cb_YGrid
      // 
      this.cb_YGrid.AutoSize = true;
      this.cb_YGrid.Location = new System.Drawing.Point(175, 39);
      this.cb_YGrid.Name = "cb_YGrid";
      this.cb_YGrid.Size = new System.Drawing.Size(55, 17);
      this.cb_YGrid.TabIndex = 10;
      this.cb_YGrid.Text = "Y Grid";
      this.cb_YGrid.UseVisualStyleBackColor = true;
      // 
      // tb_YLab
      // 
      this.tb_YLab.Location = new System.Drawing.Point(220, 13);
      this.tb_YLab.Name = "tb_YLab";
      this.tb_YLab.Size = new System.Drawing.Size(100, 20);
      this.tb_YLab.TabIndex = 9;
      // 
      // label6
      // 
      this.label6.AutoSize = true;
      this.label6.Location = new System.Drawing.Point(172, 16);
      this.label6.Name = "label6";
      this.label6.Size = new System.Drawing.Size(42, 13);
      this.label6.TabIndex = 8;
      this.label6.Text = "Y label:";
      // 
      // c_AutoScale
      // 
      this.c_AutoScale.FormattingEnabled = true;
      this.c_AutoScale.Location = new System.Drawing.Point(387, 31);
      this.c_AutoScale.Name = "c_AutoScale";
      this.c_AutoScale.Size = new System.Drawing.Size(121, 21);
      this.c_AutoScale.TabIndex = 16;
      String[] items = new String[] { "off", "once", "always" };
      foreach (String s in items) c_AutoScale.Items.Add(s);
      // 
      // label7
      // 
      this.label7.AutoSize = true;
      this.label7.Location = new System.Drawing.Point(384, 15);
      this.label7.Name = "label7";
      this.label7.Size = new System.Drawing.Size(57, 13);
      this.label7.TabIndex = 17;
      this.label7.Text = "Auto scale";
      // 
      // label8
      // 
      this.label8.AutoSize = true;
      this.label8.Location = new System.Drawing.Point(384, 58);
      this.label8.Name = "label8";
      this.label8.Size = new System.Drawing.Size(101, 13);
      this.label8.TabIndex = 19;
      this.label8.Text = "Cursor marker mode";
      // 
      // c_Cursor
      // 
      this.c_Cursor.FormattingEnabled = true;
      this.c_Cursor.Location = new System.Drawing.Point(387, 74);
      this.c_Cursor.Name = "c_Cursor";
      this.c_Cursor.Size = new System.Drawing.Size(121, 21);
      this.c_Cursor.TabIndex = 18;
      String[] cursorItems = new String[] { "None", "Horizontal", "Vertical", "CrossHair" };
      foreach (String s in cursorItems) c_Cursor.Items.Add(s);
      // 
      // groupBox1
      //
      this.groupBox1.Controls.Add(this.label10);
      this.groupBox1.Controls.Add(this.label9);
      this.groupBox1.Controls.Add(this.tb_Trend);
      this.groupBox1.Controls.Add(this.label8);
      this.groupBox1.Controls.Add(this.c_Cursor);
      this.groupBox1.Controls.Add(this.label7);
      this.groupBox1.Controls.Add(this.c_AutoScale);
      this.groupBox1.Controls.Add(this.tb_YMax);
      this.groupBox1.Controls.Add(this.label4);
      this.groupBox1.Controls.Add(this.tb_YMin);
      this.groupBox1.Controls.Add(this.label5);
      this.groupBox1.Controls.Add(this.cb_YBestScale);
      this.groupBox1.Controls.Add(this.cb_YGrid);
      this.groupBox1.Controls.Add(this.tb_YLab);
      this.groupBox1.Controls.Add(this.label6);
      this.groupBox1.Controls.Add(this.tb_XMax);
      this.groupBox1.Controls.Add(this.label3);
      this.groupBox1.Controls.Add(this.tb_XMin);
      this.groupBox1.Controls.Add(this.label2);
      this.groupBox1.Controls.Add(this.cb_XBestScale);
      this.groupBox1.Controls.Add(this.cb_XGrid);
      this.groupBox1.Controls.Add(this.tb_XLab);
      this.groupBox1.Controls.Add(this.label1);
      this.groupBox1.Location = new System.Drawing.Point(12, 12);
      this.groupBox1.Name = "groupBox1";
      this.groupBox1.Size = new System.Drawing.Size(531, 143);
      this.groupBox1.TabIndex = 2;
      this.groupBox1.TabStop = false;
      this.groupBox1.Text = "X/Y Axis";


      DataGridViewTextBoxColumn idColum = new DataGridViewTextBoxColumn()
      {
        Name = "ID", ReadOnly = true, Visible = true,
        SortMode = DataGridViewColumnSortMode.NotSortable
      };

      DataGridViewCheckBoxColumn visibleColumn = new DataGridViewCheckBoxColumn()
      {
        Name = "Visible", FalseValue = 0, TrueValue = 1, Visible = true,
        SortMode = DataGridViewColumnSortMode.NotSortable
      };
      DataGridViewButtonColumn colorColumn = new DataGridViewButtonColumn()
      {
        Name = "Color", FlatStyle = FlatStyle.Flat, Visible = true,
        SortMode = DataGridViewColumnSortMode.NotSortable
      };
      DataGridViewTextBoxColumn widthColum = new DataGridViewTextBoxColumn()
      {
        Name = "Width", Visible = true,
        SortMode = DataGridViewColumnSortMode.NotSortable
      };

      this.dgv.Dock = DockStyle.Fill;
      this.dgv.Columns.Insert(0, idColum);
      this.dgv.Columns.Insert(1, visibleColumn);
      this.dgv.Columns.Insert(2, colorColumn);
      this.dgv.Columns.Insert(3, widthColum);
      this.dgv.MultiSelect = false;
      this.dgv.AutoGenerateColumns = false;
      this.dgv.AllowUserToAddRows = false;
      this.dgv.CellClick += new DataGridViewCellEventHandler(this.dgv_CellContentClick);
      this.dgv.SelectionChanged += new EventHandler(this.dgv_CellSelectionChanged); // added to disable selection color
      // 
      // gb_Lines
      // 
      this.gb_Lines.Location = new System.Drawing.Point(12, 161);
      this.gb_Lines.Name = "gb_Lines";
      this.gb_Lines.Size = new System.Drawing.Size(612, 168);
      this.gb_Lines.TabIndex = 3;
      this.gb_Lines.TabStop = false;
      this.gb_Lines.Text = "Lines";
      // 
      // tb_Trend
      // 
      this.tb_Trend.Location = new System.Drawing.Point(54, 114);
      this.tb_Trend.Name = "tb_Trend";
      this.tb_Trend.Size = new System.Drawing.Size(41, 20);
      this.tb_Trend.TabIndex = 20;
      this.tb_Trend.Enabled = false;
      // 
      // label9
      // 
      this.label9.AutoSize = true;
      this.label9.Location = new System.Drawing.Point(6, 117);
      this.label9.Name = "label9";
      this.label9.Size = new System.Drawing.Size(38, 13);
      this.label9.TabIndex = 21;
      this.label9.Text = "Trend:";
      // 
      // label10
      // 
      this.label10.AutoSize = true;
      this.label10.Location = new System.Drawing.Point(101, 117);
      this.label10.Name = "label10";
      this.label10.Size = new System.Drawing.Size(16, 13);
      this.label10.TabIndex = 22;
      this.label10.Text = "hr";

      this.gb_Lines.Controls.Add(dgv);
      // 
      // Form2
      // 
      this.AutoScaleDimensions = new System.Drawing.SizeF(6F, 13F);
      this.AutoScaleMode = AutoScaleMode.Font;
      this.ClientSize = new System.Drawing.Size(636, 343);
      this.ControlBox = false;
      this.Controls.Add(this.gb_Lines);
      this.Controls.Add(this.groupBox1);
      this.Controls.Add(this.btn_Cancel);
      this.Controls.Add(this.btn_OK);
      this.MaximizeBox = false;
      this.MaximumSize = new System.Drawing.Size(652, 381);
      this.MinimizeBox = false;
      this.MinimumSize = new System.Drawing.Size(652, 381);
      this.Name = "Form2";
      this.ShowIcon = false;
      this.ShowInTaskbar = false;
      this.StartPosition = FormStartPosition.CenterParent;
      this.Text = "Chart Properties";
      this.groupBox1.ResumeLayout(false);
      this.groupBox1.PerformLayout();
      this.ResumeLayout(false);
    }

    private Button btn_OK;
    private Button btn_Cancel;
    private GroupBox groupBox1;
    private CheckBox cb_XGrid;
    private TextBox tb_XLab;
    private Label label1;
    private CheckBox cb_XBestScale;
    private TextBox tb_XMin;
    private Label label2;
    private TextBox tb_XMax;
    private Label label3;
    private TextBox tb_YMax;
    private Label label4;
    private TextBox tb_YMin;
    private Label label5;
    private CheckBox cb_YBestScale;
    private CheckBox cb_YGrid;
    private TextBox tb_YLab;
    private Label label6;
    private Label label7;
    private ComboBox c_AutoScale;
    private GroupBox gb_Lines;
    private Label label8;
    private ComboBox c_Cursor;
    private TextBox tb_Trend;
    private Label label10;
    private Label label9;
  }
}
