﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace Acop.AcopChartUtil
{
  /// <summary>
  /// An ACOP Event Object
  /// </summary>
  public partial class AcopEvent
  {
#   pragma warning disable 1591
    protected int event_type;             // event type
    protected int link_id, link_error;    // receive event
    protected int histo_index;            // histo index
    protected Point mouse_point;          // mouse position for mouse event
    protected double mouseUserX, mouseUserY, mouseUserSecondY; // converted user mouse position for mouse event
    protected double zoomX1, zoomY1, zoomX2, zoomY2;        // converted zoom rectangle
    protected double secondaryYMin, secondaryYMax;
    protected MouseEventArgs mouseEvent;
    protected int outsideHistoRect;       // zero means inside histo area

    // total AcopEventType
    public readonly static int ACOP_MOUSE_MOVE_EVENT = 1;
    public readonly static int ACOP_MOUSE_CLICK_EVENT = 2;
    public readonly static int ACOP_MOUSE_ZOOM_EVENT = 3;
    public readonly static int ACOP_MOUSE_ENTER_EVENT = 4;
    public readonly static int ACOP_MOUSE_EXIT_EVENT = 5;
    public readonly static int ACOP_MOUSE_DCLICK_EVENT = 6;
    public readonly static int ACOP_MOUSE_PRESSED_EVENT = 7;
    public readonly static int ACOP_MOUSE_DRAG_EVENT = 8;
    public readonly static int ACOP_RECEIVE_EVENT = 10;
    public readonly static int ACOP_ANNOATATION_EVENT = 11;

    // the constructor
    public AcopEvent(Object source, int eventId, int linkId, int error)
    {
      // pass the source object to the superclass
      //    base(source);

      // save data
      event_type = eventId;
      link_id = linkId;
      link_error = error;
    }

    // the constructor
    public AcopEvent(Object source, int eventId, int histIndex, int x, int y, double userx, double usery, double userSecondy, MouseEventArgs mEvent, int mArea)
    {
      // pass the source object to the superclass
      //    super(source);

      event_type = eventId;

      // save data
      histo_index = histIndex;
      mouse_point = new Point(x, y);
      mouseUserX = userx;
      mouseUserY = usery;
      mouseUserSecondY = userSecondy;
      mouseEvent = mEvent;
      outsideHistoRect = mArea;
    }

    // the constructor
    public AcopEvent(Object source, int eventId, double xMin, double yMin, double xMax, double yMax, double secondaryYMin, double secondaryYMax)
    {
      // pass the source object to the superclass
      //    super(source);

      event_type = eventId;

      // save data
      zoomX1 = xMin;
      zoomY1 = yMin;
      zoomX2 = xMax;
      zoomY2 = yMax;
      this.secondaryYMax = secondaryYMax;
      this.secondaryYMin = secondaryYMin;
    }

    public int GetEventType()
    {
      return event_type;
    }

    public int GetLinkId()
    {
      return link_id;
    }

    public int GetLinkError()
    {
      return link_error;
    }
    /// <summary>
    /// Returns the array index of the primary plotted graph at the mouse position.
    /// </summary>
    /// <returns>The array index at the mouse position</returns>
    public int GetArrayIndex()
    {
      return histo_index;
    }
    /// <summary>
    /// Returns the Point Object giving the current mouse position
    /// </summary>
    /// <returns>the Point Object giving the current mouse position</returns>
    public Point GetMousePoint()
    {
      return mouse_point;
    }
    /// <summary>
    /// Returns the current mouse horizontal (X) position
    /// </summary>
    /// <returns>the current mouse horizontal (X) position</returns>
    public double GetMousePositionX()
    {
      return mouseUserX;
    }
    /// <summary>
    /// Returns the current mouse vertical (Y) position
    /// </summary>
    /// <returns>the current mouse vertical (Y) position</returns>
    public double GetMousePositionY()
    {
      return mouseUserY;
    }
    /// <summary>
    /// Returns the current mouse second vertical (Y) position
    /// </summary>
    /// <returns>the current mouse second vertical (Y) position</returns>
    public double GetMousePositionSecondY()
    {
      return mouseUserSecondY;
    }
    /// <summary>
    /// Returns the minimum horizontal (X) coordinate of the current zoomed rectangle
    /// </summary>
    /// <returns>the minimum horizontal (X) coordinate of the current zoomed rectangle</returns>
    public double GetZoomXMin()
    {
      return zoomX1;
    }
    /// <summary>
    /// Returns the minimum vertical (Y) coordinate of the current zoomed rectangle
    /// </summary>
    /// <returns>the minimum vertical (Y) coordinate of the current zoomed rectangle</returns>
    public double GetZoomYMin()
    {
      return zoomY1;
    }

    /// <summary>
    /// Returns the maximum horizontal (X) coordinate of the current zoomed rectangle
    /// </summary>
    /// <returns>the maximum horizontal (X) coordinate of the current zoomed rectangle</returns>
    public double GetZoomXMax()
    {
      return zoomX2;
    }

    /// <summary>
    /// Returns the maximum vertical (Y) coordinate of the current zoomed rectangle
    /// </summary>
    /// <returns>the maximum vertical (Y) coordinate of the current zoomed rectangle</returns>
    public double GetZoomYMax()
    {
      return zoomY2;
    }

    /// <summary>
    /// Returns the 'secondary' maximum vertical (Y) coordinate of the current zoomed rectangle
    /// </summary>
    /// <returns>the 'secondary' maximum vertical (Y) coordinate of the current zoomed rectangle</returns>
    public double GetSecondaryYMax()
    {
      return secondaryYMax;
    }

    /// <summary>
    /// Returns the 'secondary' minimum vertical (Y) coordinate of the current zoomed rectangle
    /// </summary>
    /// <returns>the 'secondary' minimum vertical (Y) coordinate of the current zoomed rectangle</returns>
    public double GetSecondaryYMin()
    {
      return secondaryYMin;
    }

    /// <summary>
    /// Returns the mouseEvente
    /// </summary>
    /// <returns>the mouseEvent</returns>
    public MouseEventArgs GetmouseEvent()
    {
      return mouseEvent;
    }

    /// <summary>
    /// Returns the mouseEvente
    /// </summary>
    /// <returns>the mouseEvent</returns>
    public int GetframeClick()
    {
      return outsideHistoRect;
    }
  }
}
