﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Globalization;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace Acop.AcopChartUtil
{
  public partial class AcopHisto
  {
#   pragma warning disable 1591
    //    static FontFamily histFont = new FontFamily("Century");
    static FontFamily histFont = new FontFamily("Times New Roman");
    static Font ANNOTATION_FONT = new Font(histFont, AcopConst.AnnotationSize - 2, FontStyle.Bold | FontStyle.Italic);
    //static Font ANNOTATION_FONT = new Font(histFont, 15, FontStyle.Bold);
    static String ANNOTATION_TEXT = "i";
    static Color ANNOTATION_COLOR = Color.FromArgb(200, 200, 200);
    static SolidBrush ANNOTATION_BRUSH = new SolidBrush(ANNOTATION_COLOR);
    //    static SolidBrush ANNOTATION_BRUSH = new SolidBrush(Color.FromArgb(200, 200, 200));
    protected SolidBrush ANNOTATION_STRING_BRUSH;
    protected int hisIndex; // own histo index number
    protected internal int npoints; // display points
    protected internal int displayON, // on or off
        number, // the display order
        markerFlag, marker_ary_cpy, xextent, yextent;
    protected internal int lockFlag, dismode;
    protected internal int displayModeSetting;
    protected internal int disBarHisto, disDots;
    protected internal int xlabON, disFlag, xlabLength;
    protected internal int wrapindex, wraphigh;
    protected double originxmin, originxmax; // x range
    // CPen pen;
    // LOGBRUSH lb, lbEdge, lbMarker;
    protected internal int penStyle, penWidth, halfWidth;
    protected Pen stroke, tmpPen;
    protected internal Color penColor, tmpColor;
    protected int pixelWidth;
    protected internal int[] ix, iy, imy; // converted int array, marker array
    protected internal double[] x, y; // input display data
    protected String[] str;
    protected double[] markery;
    // char *buf; // diaplay buffer (npoints*4*3] convert integer x and y arrays
    protected int[] dispOn, dispxOn, dispyOn; // diaplay flag for each point(on:
    // within the area)
    protected internal char[] disable;
    protected int[] dispindex; // diaplay index starting point
    protected int[] dispnumber; // number poinst to diaplay
    protected internal Color[] m_tagColorArray; // keep tagged color
    protected internal String[] xlab; // xlabel array
    protected internal double xext, yext;
    protected internal Boolean colorfilled;
    protected internal int barlineColorFilled;
    protected internal int xaryFlag;
    // int aryinput,byref;
    protected int format;
    protected Boolean saveInvert;
    protected int DrawDataSize;
    protected double saveYShift, saveYScale, saveXShift, saveXScale;
    protected int saveBitDisplayMode;
    protected double saveGanttBitDisplayBase, saveGanttBitDisplayHeight;
    protected int saveBitDisplayMask, mapedBase;
    protected internal short tagHist, initTagBuffer; // flag for tagging display
    protected internal Color tagColor, vorColor; // tagging and histo background
    protected internal int tagWidth;
    // color for this histogram!
    protected internal Boolean yScaleEnable, errWindowEnable;
    protected internal Color errorColor;
    protected internal double err_xmin, err_xmax, err_ymin, err_ymax; // error window
    // display rectangle
    protected int arrpoints, nextPos; // for appendscreen
    protected int gaps;
    protected int moving, fFT;
    protected internal int edgeIndex, edgeMode;
    // CPen edgePen;
    protected Pen edgeStroke;
    protected int edgeStyle, edgeWidth;
    protected Color edgeColor;
    protected double edgePercentX, edgePercentY;
    protected double edgeExtentX, edgeExtentY;
    protected int edgePixel;
    // CPen markerPen;
    protected Pen mStroke;
    protected int markerStyle, markerWidth;
    protected Color markerColor;
    protected int markerMode;
    protected double markerExtentX, markerExtentY;
    protected double markerPercentX, markerPercentY;
    protected int polycounter; // number of polypoints inside display window
    protected internal Boolean yReferenceEnable;
    protected internal Boolean bindSecondaryY;
    private AcopChart acop;
    private AcopFFT acopfft;
    private int[] dotsScale = { 200, 200, 150, 140, 130, 120, 110, 105 };
    protected internal double[] curveFitResults;
    private NumberFormatInfo nfi = new NumberFormatInfo();
    private Boolean showInDataGrid = true;

    public AcopHisto(AcopChart acop, int index)
    {
      this.acop = acop;
      npoints = 0;
      y = null;
      hisIndex = index;
      curveFitResults = new double[100];
      nfi.NumberDecimalSeparator = ".";
      // {{INIT_CONTROLS
      // }}
    }

    // {{DECLARE_CONTROLS
    protected int calcDetermination(double[] matrix, int dimens, double[] pOut)
    {
      int row, col, tmpCol, revCol;
      double tmp = 0, tmpNatural = 0, tmpReverse = 0;

      if (dimens < 2) return -1;
      if (dimens == 2) tmp = matrix[0] * matrix[3] - matrix[1] * matrix[2];
      else for (col = 0; col < dimens; col++)
        {
          tmpCol = col;
          revCol = dimens - col - 1;
          tmpNatural = matrix[tmpCol];
          tmpReverse = matrix[revCol];
          for (row = 1; row < dimens; row++)
          {
            if (++tmpCol >= dimens) tmpCol = 0;
            if (--revCol < 0) revCol = dimens - 1;
            tmpNatural *= matrix[row * dimens + tmpCol];
            tmpReverse *= matrix[row * dimens + revCol];
          }
          tmp += (tmpNatural - tmpReverse);
        }
      pOut[0] = tmp;
      return 0;
    }
    protected int matrixCalculation(double[] matDiv, double[] matTmp, double[] vector, double[] result, int n)
    {
      int ret = 0, col, row;
      double[] divider, FenZi;

      divider = new double[1];
      FenZi = new double[1];
      ret = calcDetermination(matDiv, n, divider);
      if (ret == 0) for (col = 0; col < n; col++)
        {
          Array.Copy(matDiv, 0, matTmp, 0, n * n);
          for (row = 0; row < n; row++) matTmp[row * n + col] = vector[row];
          ret = calcDetermination(matTmp, n, FenZi);
          if (ret != 0) break;
          result[col] = FenZi[0] / divider[0];
        }
      return ret;
    }

    protected int polyfit(double[] xin, double[] yin, int length, int dimension, double[] fitParameter)
    {
      double[] matDiv, matTmp, vector, result, divider, FenZi;
      double sigma, tmp, div;
      int n, row, col, i, ret;

      if (dimension < 1) return -1;
      if (length <= dimension) return -2;
      if (xin.Length < length || yin.Length < length) return -3;

      n = dimension + 1;
      matDiv = new double[n * n];
      vector = new double[n];
      matTmp = new double[n * n];
      result = new double[n];
      divider = new double[1];
      FenZi = new double[1];
      sigma = 0;    // standard diviation

      for (row = 0; row < n; row++)
      {
        vector[row] = 0;
        for (i = 0; i < length; i++) vector[row] += yin[i] * Math.Pow(xin[i], n - row - 1);
        for (col = 0; col < n; col++)
          if (row <= col) for (i = 0; i < length; i++) matDiv[row * n + col] += Math.Pow(xin[i], n - col - 1) * Math.Pow(xin[i], n - row - 1);
          else matDiv[row * n + col] = matDiv[col * n + row];
      }
      ret = matrixCalculation(matDiv, matTmp, vector, result, n);
      Array.Copy(result, 0, fitParameter, 0, n);
      for (div = 0, i = 0; i < length; i++)
      {
        sigma = yin[i] - fitParameter[n - 1];    // residual standard deviation
        tmp = 1;
        for (col = n - 2; col >= 0; col--)
        {
          tmp *= xin[i];
          sigma -= fitParameter[col] * tmp;
        }
        div += sigma * sigma;
      }
      fitParameter[n] = Math.Sqrt(div / (length - n));
      return ret;
    }
    protected int gaussianNonLinearFit(double[] xin, double[] yin, int length, int flag, double[] fitParameter)
    {
      double[] matF, matTmp, vectB, lamda;
      double tmp1, delta, x2square, ampl, mu, sigma;
      int i, j, k, n, ret = 0;

      n = flag == 1 ? 4 : 3;  // gaussian = 1, exponential = 0
      matF = new double[n * n];
      matTmp = new double[n * n];
      vectB = new double[n];
      lamda = new double[n];

      ampl = fitParameter[0];
      mu = fitParameter[1];
      sigma = fitParameter[2];
      x2square = 0;
      for (i = 0; i < n; i++)
      {
        vectB[i] = 0;
        for (k = 0; k < length; k++)
        {
          if (flag == 1)  // gaussian fit: A * exp( -(x-mu)**2/(2*sigma**2) ) + offset
          {
            tmp1 = Math.Exp(-0.5 * Math.Pow(xin[k] - mu, 2) / Math.Pow(sigma, 2));
            delta = yin[k] - ampl * tmp1 - fitParameter[3];
            lamda[0] = tmp1;
            lamda[1] = ampl * (xin[k] - mu) / Math.Pow(sigma, 2) * tmp1;
            lamda[2] = ampl * Math.Pow(xin[k] - mu, 2) / Math.Pow(sigma, 3) * tmp1;
            lamda[3] = 1;
          }
          else  // exponential fit: A * exp( Bx ) + C
          {
            tmp1 = Math.Exp(mu * xin[k]);
            delta = yin[k] - ampl * tmp1 - sigma;
            lamda[0] = tmp1;
            lamda[1] = ampl * tmp1 * xin[k];
            lamda[2] = 1;
          }
          vectB[i] += lamda[i] * delta;
          for (j = 0; j < n; j++)
            if (i <= j) matF[i * n + j] += lamda[i] * lamda[j];
            else matF[i * n + j] = matF[j * n + i];
          if (i == 0) x2square += delta * delta;
        }
      }
      delta = Math.Sqrt(x2square / (length - n));
      ret = matrixCalculation(matF, matTmp, vectB, fitParameter, n);
      fitParameter[n] = delta;
      return ret;
    }
    protected int curveFit(int npoints, double[] x, double[] y, int flag, double[] fitResults)
    {
      double[] logy;
      int i;
      if (flag == AcopConst.FFT_LINE_FIT)
        return polyfit(x, y, npoints, 1, fitResults);
      else if (flag == AcopConst.FFT_EXPONENTIAL_FIT || flag == AcopConst.FFT_GAUSSIAN_FIT)
      {
        logy = new double[npoints];
        for (i = 0; i < npoints; i++) logy[i] = Math.Log(y[i]);
        return polyfit(x, logy, npoints, flag == AcopConst.FFT_EXPONENTIAL_FIT ? 1 : 2, fitResults);
      }
      if (flag == AcopConst.FFT_POLYNORM_QUADRTIC_FIT) i = 2;
      else if (flag == AcopConst.FFT_POLYNORM_CUBIC_FIT) i = 3;
      else i = 4;
      return polyfit(x, y, npoints, i, fitResults);
    }
    protected int drawChanges(int flag)
    {
      double slop, offset, ampl, sumTmp, cut;
      double[] tmp, tmp2, tmpSave;
      int ret = 0, i, j, k, increase, n_interval, itmp = 0, nInterction = 300;
      int[] convergent;

      if (flag >= AcopConst.FFT_POLYNORM_QUARTIC_FIT || flag <= AcopConst.FFT_RI_INVERSE) return -1;
      if (flag >= AcopConst.FFT_LINE_FIT && flag <= AcopConst.FFT_POLYNORM_QUARTIC_FIT)
      {
        curveFitResults[0] = flag;
        tmp = new double[20];
        tmp2 = new double[20];
        tmpSave = new double[20];
        convergent = new int[20];
        for (i = 0; i < 20; i++) tmp[i] = 0;
        ret = curveFit(npoints, x, y, flag, tmp);
        Array.Copy(tmp, 0, curveFitResults, 1, 20);
        slop = tmp[0];
        offset = tmp[1];
        if (ret != 0) return ret;
        if (flag == AcopConst.FFT_LINE_FIT)
          for (i = 0; i < npoints; i++) y[i] = slop * x[i] + offset;
        else if (flag == AcopConst.FFT_EXPONENTIAL_FIT)
        {
          tmp[0] = Math.Exp(offset);
          tmp[1] = slop;
          tmp[2] = 0;
          cut = 1.0e-10;
          for (i = 0; i < nInterction; i++)
          {
            Array.Copy(tmp, 0, tmp2, 0, 20);
            gaussianNonLinearFit(x, y, npoints, 0, tmp);
            for (itmp = 0, j = 0; j < 3; j++) if (Math.Abs(tmp[j]) < cut) itmp++;
            for (j = 0; j < 3; j++)
            {
              if (i == 0) convergent[j] = 0;
              else
              {
                if (Math.Abs(tmp[j]) > Math.Abs(tmpSave[j])) convergent[j] = 0;
                else convergent[j]++;
              }
              tmpSave[j] = tmp[j];
            }
            for (j = 0; j < 3; j++) tmp[j] = tmp2[j] + tmp[j];
            if (itmp == 3) break;
          }
          curveFitResults[5] = i;  // number of iteration
          for (i = 0; i < npoints; i++) y[i] = tmp[0] * Math.Exp(tmp[1] * x[i]) + tmp[2];
          Array.Copy(tmp, 0, curveFitResults, 1, 4);
          for (j = 0; j < 3; j++) curveFitResults[6 + j] = convergent[j];
        }
        else if (flag == AcopConst.FFT_GAUSSIAN_FIT)
        {
          if (tmp[0] >= 0) // curve fit can not find a reasonable imput value
          {
            ampl = 0;
            slop = 0;
            offset = 0;
          }
          else
          {
            slop = Math.Sqrt(-1 / (2 * tmp[0]));  // sigma
            offset = tmp[1] * slop * slop;        // mu
            ampl = Math.Exp(tmp[2] + offset * offset / (2 * slop * slop)); // amplitute
          }
          tmp[0] = ampl;
          tmp[1] = offset;
          tmp[2] = slop;
          tmp[3] = 1;     // try offset
          cut = 1.0e-3;
          double tmpStore = ampl;
          double tmpSigma = slop;
          double tmpMu = tmp[1];
          double tmpXmin = x[0], tmpXmax = x[npoints - 1];
          tmpStore = 0;//tmp[0];
          tmpSigma = tmp[2];
          for (i = 0; i < npoints; i++) if (y[i] > tmpStore) tmpStore = y[i];  // find maximum
          for (i = 0, j = 0, sumTmp = 0; i < npoints; i++)     // find offset
            if (y[i] < tmpStore * 0.1)
            {
              sumTmp += y[i];
              j++;
            }
          if (j > 0) tmp[3] = sumTmp / j;
          if (tmpStore > 2 * ampl || tmpStore < ampl)  // if the input ampl is too small, adjust amplitue and sigma
          {
            tmp[0] = tmpStore * 3 / 4;
            tmpStore = tmp[0];
            for (j = 0; j < npoints; j++)         // find Xmin
            {
              if (y[j] > tmpStore)
              {
                tmpXmin = x[j];
                break;
              }
            }
            for (j = npoints - 1; j >= 0; j--)    // find Xmax
            {
              if (y[j] > tmpStore)
              {
                tmpXmax = x[j];
                break;
              }
            }
            tmp[2] = (tmpXmax - tmpXmin) / 2;
            tmpSigma = tmp[2];
            tmp[1] = tmpXmin + (tmpXmax - tmpXmin) / 2;
            tmpMu = tmp[1];
          }
          for (i = 0; i < nInterction; i++)
          {
            //          System.out.println("Interactio: " + i );
            //          for (k=0; k<5; k++) System.out.println(tmp[k]);          
            Array.Copy(tmp, 0, tmp2, 0, 20);
            gaussianNonLinearFit(x, y, npoints, 1, tmp);
            //          System.out.println("Afterwards: " + i );
            //          for (k=0; k<5; k++) System.out.println(tmp[k]);
            for (itmp = 0, j = 0; j < 4; j++) if (Math.Abs(tmp[j]) < cut) itmp++;
            for (j = 0; j < 4; j++)
            {
              if (i == 0) convergent[j] = 0;
              else
              {
                if (Math.Abs(tmp[j]) > Math.Abs(tmpSave[j])) convergent[j] = 0;
                else convergent[j]++;
              }
              tmpSave[j] = tmp[j];
            }
            for (j = 0; j < 4; j++) tmp[j] = tmp2[j] + tmp[j];          // load new value
            if (itmp == 4)
            {
              if ((tmp[0] < 0 || tmp[2] < 0 || tmp[3] < 0) && i < (nInterction - 1))
              {
                if (tmp[0] < 0) tmp[0] = tmpStore;
                if (tmp[2] < 0) tmp[2] = -tmp[2];
                if (tmp[3] < 0) tmp[3] = -tmp[3];
                continue;
              }
              break;
            }
            if (tmp[0] < tmpStore * 0.2) tmp[0] = tmpStore;  // check negative ampl, sigma and offset
            if (tmp[2] < 0) tmp[2] = tmpSigma;
            if (tmp[3] < 0) tmp[3] = -tmp[3];
            if (tmp[1] < tmpXmin || tmp[1] > tmpXmax) tmp[1] = tmpMu;
          }
          ampl = tmp[0];
          offset = tmp[1];
          slop = tmp[2];
          curveFitResults[6] = i;  // number of iteration
          for (i = 0; i < npoints; i++) y[i] = ampl * Math.Exp(-(x[i] - offset) * (x[i] - offset) / (2 * slop * slop)) + tmp[3];
          Array.Copy(tmp, 0, curveFitResults, 1, 5);
          for (j = 0; j < 4; j++) curveFitResults[7 + j] = convergent[j];
          //        curveFitResults[5] = tmp[4];  // residual standard diviation
        }
        else if (flag == AcopConst.FFT_POLYNORM_QUADRTIC_FIT)
          for (i = 0; i < npoints; i++) y[i] = slop * x[i] * x[i] + offset * x[i] + tmp[2];
        else if (flag == AcopConst.FFT_POLYNORM_CUBIC_FIT)
          for (i = 0; i < npoints; i++) y[i] = slop * Math.Pow(x[i], 3) + offset * Math.Pow(x[i], 2) + tmp[2] * x[i] + tmp[3];
        else
          for (i = 0; i < npoints; i++) y[i] = slop * Math.Pow(x[i], 4) + offset * Math.Pow(x[i], 3) + tmp[2] * Math.Pow(x[i], 2) + tmp[3] * x[i] + tmp[4];
        for (i = 0; i < npoints; i++)
        {
          if (Math.Abs(saveYShift) > AcopConst.zero_check) y[i] -= saveYShift;
          if (Math.Abs(saveXShift) > AcopConst.zero_check) x[i] -= saveXShift;
        }
      }
      else if (flag == AcopConst.FFT_DIFFER_STATISTIC || flag == AcopConst.FFT_INTEGRAL_STATISTIC)
      {
        if (npoints < 2) return -1;
        increase = x[1] > x[0] ? 1 : 0;
        ret = 0;
        for (i = 1; i < npoints; i++)
        {
          if (increase != 0 && (x[i] - x[i - 1]) < AcopConst.zero_check) break;
          else if (increase == 0 && (x[i - 1] - x[i]) < AcopConst.zero_check) break;
        }
        n_interval = i;
        tmp = new double[n_interval];
        tmp[n_interval - 1] = 0;
        for (i = 1; i < n_interval; i++)
        {
          tmp[i - 1] = 0;
          for (j = 0; j < npoints; j++)
          {
            if (increase != 0)
            {
              if (y[j] >= x[i - 1] && y[j] < x[i]) tmp[i - 1] += 1;//disp[n]->y[j];
            }
            else
            {
              if (y[j] < x[i - 1] && y[j] >= x[i]) tmp[i - 1] += 1;//disp[n]->y[j];
            }
          }
        }
        npoints = n_interval;
        for (i = 0; i < npoints; i++)
        {
          //      disp[n]->x[i] = (disp[n]->x[i+1] + disp[n]->x[i]) / 2;
          y[i] = tmp[i];
          if (flag == AcopConst.FFT_INTEGRAL_STATISTIC && i != 0) y[i] += y[i - 1];
        }
      }
      return ret;
    }
    protected int disHisto(int start, int stop)
    {
      int i, j;
      for (j = start; j < stop; j++)
      {
        i = j < npoints ? j : j - npoints;
        if (disable[i] != 0) continue;
        disFlag++;
        disable[i] = (char)1; // disable
      }
      return 0;
    }
    protected static internal int getArrayLength(Object inputArray)
    {
      int length;
      if (inputArray == null)
        length = 0;
      else if (inputArray is double[])
        length = ((double[])inputArray).Length;
      else if (inputArray is float[])
        length = ((float[])inputArray).Length;
      else if (inputArray is int[])
        length = ((int[])inputArray).Length;
      else if (inputArray is short[])
        length = ((short[])inputArray).Length;
      else if (inputArray is byte[])
        length = ((byte[])inputArray).Length;
      else if (inputArray is char[])
        length = ((char[])inputArray).Length;
      else if (inputArray is Color[])
        length = ((Color[])inputArray).Length;
      else
        length = 0;
      return length;
    }
    protected static internal int copyArray(double[] doubleArray, Object inputArray, int start, int length,
        int aryMaxNumber, int loadToDoubleArray)
    {
      int i, j, k;
      if (inputArray is double[])
        for (k = start; k < length; k++)
        {
          i = k - start;
          j = k < aryMaxNumber ? k : (k - aryMaxNumber);
          if (loadToDoubleArray == 1)
            doubleArray[j] = (double)((double[])inputArray)[i];
          else
            ((double[])inputArray)[i] = (double)doubleArray[j];
        }
      else if (inputArray is float[])
        for (k = start; k < length; k++)
        {
          i = k - start;
          j = k < aryMaxNumber ? k : (k - aryMaxNumber);
          if (loadToDoubleArray == 1)
            doubleArray[j] = (double)((float[])inputArray)[i];
          else
            ((float[])inputArray)[i] = (float)doubleArray[j];
        }
      else if (inputArray is int[])
        for (k = start; k < length; k++)
        {
          i = k - start;
          j = k < aryMaxNumber ? k : (k - aryMaxNumber);
          if (loadToDoubleArray == 1)
            doubleArray[j] = (double)((int[])inputArray)[i];
          else
            ((int[])inputArray)[i] = (int)doubleArray[j];
        }
      else if (inputArray is short[])
        for (k = start; k < length; k++)
        {
          i = k - start;
          j = k < aryMaxNumber ? k : (k - aryMaxNumber);
          if (loadToDoubleArray == 1)
            doubleArray[j] = (double)((short[])inputArray)[i];
          else
            ((short[])inputArray)[i] = (short)doubleArray[j];
        }
      else if (inputArray is byte[])
        for (k = start; k < length; k++)
        {
          i = k - start;
          j = k < aryMaxNumber ? k : (k - aryMaxNumber);
          if (loadToDoubleArray == 1)
            doubleArray[j] = (double)((byte[])inputArray)[i];
          else
            ((byte[])inputArray)[i] = (byte)doubleArray[j];
        }
      else if (inputArray is char[])
        for (k = start; k < length; k++)
        {
          i = k - start;
          j = k < aryMaxNumber ? k : (k - aryMaxNumber);
          if (loadToDoubleArray == 1)
            doubleArray[j] = (double)((char[])inputArray)[i];
          else
            ((char[])inputArray)[i] = (char)doubleArray[j];
        }
      else
        length = 0;
      return length;
    }
    protected static int copyCharArray(char[] charArray, Object inputArray, int length)
    {
      int i;
      if (inputArray is double[])
        for (i = 0; i < length; i++)
          charArray[i] = (char)((double[])inputArray)[i];
      else if (inputArray is float[])
        for (i = 0; i < length; i++)
          charArray[i] = (char)((float[])inputArray)[i];
      else if (inputArray is int[])
        for (i = 0; i < length; i++)
          charArray[i] = (char)((int[])inputArray)[i];
      else if (inputArray is short[])
        for (i = 0; i < length; i++)
          charArray[i] = (char)((short[])inputArray)[i];
      else if (inputArray is byte[])
        for (i = 0; i < length; i++)
          charArray[i] = (char)((byte[])inputArray)[i];
      else if (inputArray is char[])
        for (i = 0; i < length; i++)
          charArray[i] = (char)((char[])inputArray)[i];
      else
        length = 0;
      return length;
    }
    protected internal void keephisto()
    {
      Boolean change;
      change = (penWidth != acop.keepWidth || penStyle != acop.keepStyle || penColor != acop.keepColor) ? true : false;
      penWidth = acop.keepWidth;
      halfWidth = penWidth / 2;
      penStyle = acop.keepStyle;
      penColor = acop.keepColor;
      if (change == false) return;
      stroke = AcopConst.createStroke(penColor, penWidth, penStyle);
      ANNOTATION_STRING_BRUSH = new SolidBrush(penColor);
    }
    private Boolean needsPruning = false;
    protected internal Boolean getNeedsPruning() { return needsPruning;  }
    protected internal void setNeedsPruning(Boolean value) { needsPruning = value;  }
    protected internal int getNumberDrawnData()
    {
      int n = 0;
      if (disFlag != 0)
      {
        for (n = 0; n < npoints; n++) if (disable[n] != 0) break;
      }
      else
      {
        for (int i = 0; i < npoints; i++) if (x[i] != 0 && y[i] != 0) n++;
      }
      return n;
    }
    protected internal int getNumberDrawnPoints()
    {
      int n = 0;
      for (n = 0; n < npoints; n++) if (disable[n] != 0) break;
      return n;
    }
    protected internal int getDrawnData(Object Yarray, Object Xarray, int Npoints, int StartIndex, int hDisplay)
    {
      int length, num, start, i, j, k;
      num = Npoints; // number of points, at least 1 point
      if (num < 0)
        num = npoints;
      else if (num == 0) return -7;
      start = StartIndex; // start point: 0 to (npoints-1)
      if (start < 0)
        start = 0;
      else if (start >= npoints) return -8;
      if (num + start > npoints) num = npoints - start;
      num += start;
      if (Yarray == null && Xarray == null) return -2;
      if (disFlag != 0)
      {
        for (i = start; i < num; i++)
          if (disable[i] != 0) break;
        if (i != num) num = i;
        if (i == start) return -9; // no valid data
      }
      if (Yarray != null)
      {
        length = getArrayLength(Yarray);
        if (length < num - start)
          length += start;
        else
          length = num;
        copyArray(y, Yarray, start, length, npoints, 0);
      }
      if (Xarray != null)
      {
        length = getArrayLength(Xarray);
        if (length < num - start)
          length += start;
        else
          length = num;

        if (Xarray is String[])
          for (k = start; k < num; k++)
          {
            i = k - start;
            j = k < npoints ? k : (k - npoints);
            if (xlab != null)
              ((String[])Xarray)[i] = xlab[j];
            else
              ((String[])Xarray)[i] = j.ToString();
          }
        else
          copyArray(x, Xarray, start, num, npoints, 0);
      }
      return num - start; // return number of valid data
    }
    private double[] getPlotableArray(Object data)
    {
      if (!(data is Array))
      {
        if (data is double) return new double[] { (double)data };
        if (data is float) return new double[] { (float)data };
        if (data is byte) return new double[] { (byte)data };
        if (data is short) return new double[] { (short)data };
        if (data is int) return new double[] { (int)data };
        if (data is long) return new double[] { (double)((long)data) };
      }
      if (data is double[]) return (double[])data;
      if (data is float[])
      {
        int siz = ((float[])data).Length;
        double[] x = new double[siz];
        for (int i = 0; i < siz; i++) x[i] = ((float[])data)[i];
        return x;
      }
      if (data is byte[])
      {
        int siz = ((byte[])data).Length;
        double[] x = new double[siz];
        for (int i = 0; i < siz; i++) x[i] = ((byte[])data)[i];
        return x;
      }
      if (data is short[])
      {
        int siz = ((short[])data).Length;
        double[] x = new double[siz];
        for (int i = 0; i < siz; i++) x[i] = ((short[])data)[i];
        return x;
      }
      if (data is int[])
      {
        int siz = ((int[])data).Length;
        double[] x = new double[siz];
        for (int i = 0; i < siz; i++) x[i] = ((int[])data)[i];
        return x;
      }
      if (data is long[])
      {
        int siz = ((long[])data).Length;
        double[] x = new double[siz];
        for (int i = 0; i < siz; i++) x[i] = ((long[])data)[i];
        return x;
      }
      return null;
    }
    protected internal int updateScreen(Boolean AppendFlag, Object Yarray, int Npoints, int StartIndex, Object Xarray)
    {
      int format, i, num, start, j, ndata, fsize, k, newX;
      /*
       * SAFEARRAY FAR* sa; HRESULT hr; char *data; double *df;
       */
      double maximum, xpostmp;
      int aryType, saveNext;
      int refBegin, refEnd; // refresh screen left and right histoindexes
      int enaUpdate, changeStart;
      double lowTmp = 0, highTmp = 0;
      //      int autoTimeScaling, DataSize;
      double autoTimeScaling;
      int DataSize;
      Boolean tmpboolean;
      double[] yaxis;
      double[] xaxis;
      int scaleIndex = bindSecondaryY ? 2 : 1;
      //    if (acop.getAcopTransport() != null)
      //    {
      //      yaxis = acop.getAcopTransport().getYAxis(Yarray);
      //      xaxis = acop.getAcopTransport().getXAxis(Xarray);
      //    }
      //    else
      {
        yaxis = getPlotableArray(Yarray);
        xaxis = getPlotableArray(Xarray);
      }
      changeStart = 0;
      autoTimeScaling = 0;
      newX = 0;
      if (Npoints < 1)
        Npoints = AppendFlag ? 1 : npoints;
      else if (Npoints > npoints) return -7;
      num = Npoints; // number of points, at least 1 point
      start = StartIndex; // start point: 0 to (npoints-1)
      if (start >= npoints) return -8;
      if (start < 0)
        start = AppendFlag ? nextPos : 0;
      else if (start != nextPos) changeStart = 1;
      saveNext = nextPos;
      if (AppendFlag)
      {
        if (num + gaps > npoints) num = npoints - gaps;
        if (moving == 0)
        {
          // case 1
          if (start >= npoints || (start + num) > npoints)
          { // make a gap
            if (gaps != 0) disHisto(npoints - gaps, npoints);
            // make copy, what about x??
            for (j = 0, i = start - (npoints - gaps - num); i < start; i++, j++)
            {
              y[j] = y[i];
              if (markerFlag != 0) markery[j] = markery[i];
              if (acop.sca[0].xtime == 1) x[j] = x[i]; // time plot
            }
            // change start point
            start = npoints - gaps - num;
            refBegin = 0; // refresh whole screen
            refEnd = npoints;
          }
          else
          // case 2
          {
            // get data from start to start + num, refresh part screen start to
            // start + num, only
            refBegin = start; // refresh only this part
            refEnd = start + num;
          }
          nextPos = start + num;
        }
        else
        {
          fsize = 0;
          if (start >= npoints)
          {
            start = 0;
            fsize = 1;
          }
          if (gaps != 0)
          {
            if (changeStart != 0 || num > gaps)
              disHisto(start + num, start + num + gaps);
            else
              disHisto(start + gaps, start + num + gaps);
          }
          nextPos = start + num;
          // if ((start + num) >= disp[hDisplay]->npoints) disp[hDisplay]->nextPos
          // -= disp[hDisplay]->npoints;
          refBegin = start; // refresh this part + gaps
          refEnd = start + num + gaps;
          if (edgeMode != 0 && fsize != 0) refBegin = npoints - 1;
        }
        num += start;
      }
      else
      // case 0
      {
        if (num + start > npoints) num = npoints - start;
        num += start;
        refBegin = start;
        refEnd = num;
      }
      DataSize = 0;
      fsize = getArrayLength(yaxis);
      if (fsize <= 0)
      {
        if (AppendFlag) nextPos = saveNext;
        return fsize - 1;
      }
      if (fsize < num - start) num = start + fsize;
      enaUpdate = (AppendFlag || (num - start) < npoints) ? 1 : 0;
      // RedrawTextHist = 1;
      // df = disp[hDisplay]->y ; // load new y array
      // SetArrayFromVariant(df, data, start, num, format, aryType, DataSize,
      // disp[hDisplay]->npoints, 1);
      copyArray(y, yaxis, start, num, npoints, 1);
      acop.averagingScaling(y, start, num, hisIndex);
      for (k = start; k < num; k++)
      {
        j = k < npoints ? k : (k - npoints);
        // if ( fabs(disp[hDisplay]->shift) > 1.0e-14 ) df[j] +=
        // (disp[hDisplay]->number * disp[hDisplay]->shift);
        if (Math.Abs(saveYShift) > AcopConst.zero_check) y[j] += saveYShift;// shift
        // first
        y[j] *= saveYScale;
        // enable refreshed or updated points
        if (disFlag != 0 && enaUpdate != 0 && disable[j] != 0)
        {
          disFlag--;
          disable[j] = (char)0; // enable for update
          if (dismode == AcopConst.mode_polyline)
          {
            dispxOn[j] = 0;
            dispyOn[j] = 0;
          }
        }
      }
      if (saveInvert) for (k = start; k < num; k++)
        {
          i = k < npoints ? k : (k - npoints);
          if (Math.Abs(y[i]) > AcopConst.zero_check) y[i] = 1 / y[i];
        }
      if (saveBitDisplayMode != 0) for (k = start; k < num; k++)
        {
          i = k < npoints ? k : (k - npoints);
          if (saveBitDisplayMode == 1)
          {
            if ((int)y[i] != 0) y[i] = saveGanttBitDisplayBase + saveGanttBitDisplayHeight;
            else y[i] = saveGanttBitDisplayBase;
          }
          else
          {
            if ((saveBitDisplayMask == 0 && (int)y[i] != 0) || ((int)y[i] & saveBitDisplayMask) != 0) y[i] = saveGanttBitDisplayBase + saveGanttBitDisplayHeight;
            else y[i] = saveGanttBitDisplayBase;
          }
        }
      // fft => overwrite y array !!! it works only when input whole array!!!
      if (fFT > 0 && fFT <= AcopConst.FFT_RI_INVERSE)
      {
        double[] maxPerCut = new double[3];
        maxPerCut[1] = 0.5;
        maxPerCut[2] = 12;
        acopfft.fFT(npoints, y, y, maxPerCut, fFT);
      }
      if (AppendFlag && edgeMode != 0)
      {
        j = num - 1;
        edgeIndex = j < npoints ? j : (j - npoints);
      }
      // check Xarray
      DataSize = 0;
      fsize = getArrayLength(xaxis);
      // if ( fsize == 0 && (aryType || format == disp[hDisplay]->format) )
      if (fsize > 0)
      {
        if (fsize < num - start) num = start + fsize;
        copyArray(x, xaxis, start, num, npoints, 1);
        j = k;
        for (k = start; k < num; k++)
        {
          j = k < npoints ? k : (k - npoints);
          if (Math.Abs(saveXShift) > AcopConst.zero_check) x[j] += saveXShift;// shift
          // first
          x[j] *= saveXScale;
        }
        // should check direction?????
        if (acop.sca[0].xtime == 1 && x[j] > acop.sca[0].max && x[j] > acop.sca[0].usermax)
          if (acop.isZoomedFlag == false || acop.zoomXmaxFlag == 1) // 2013-01-10 not zoomed or only in y direction
                                                                    //            if (acop.isZoomedFlag == false || Math.Abs(acop.sca[0].userSize - acop.sca[0].dispSize) < 1.0e-7) // 2010-10-25 not zoomed or only in y direction
                                                                    //            autoTimeScaling = (int)(x[j] + gaps * AcopConst.timeScale[acop.sca[0].time_scale]);
            autoTimeScaling = (x[j] + gaps * AcopConst.timeScale[acop.sca[0].time_scale]);
        newX = 1;
      }
      if (fFT > AcopConst.FFT_RI_INVERSE) drawChanges(fFT);
      if (acop.aFrame.drawFlag == false && acop.m_lock == 0) acop.m_histstart = hisIndex;
      displayON = 1;
      acop.aFrame.drawFlag = true;
      if (dismode == AcopConst.mode_histogram || dismode == AcopConst.mode_histogram_rastoring
          || dismode == AcopConst.mode_dots || dismode == AcopConst.mode_rectangle
          || dismode == AcopConst.mode_circle || dismode == AcopConst.mode_annotation)
      { // bring to top histogram!
        tagHist = 1;
        acop.lastTagHisto = hisIndex;
      }
      acop.lastDisp = hisIndex;
      // refresh new input data
      if (autoTimeScaling > 0.001 || (refEnd - refBegin) >= npoints || dismode >= AcopConst.mode_textbox)
      { // redraw complete curver
        if (autoTimeScaling > 0.001) // one should not use xbestsclae flag!
        {
          maximum = autoTimeScaling - acop.sca[0].usermax;
          acop.sca[0].setDispscale(acop.sca[0].min + maximum, acop.sca[0].max + maximum);
          acop.sca[0].usermin += maximum;
          acop.sca[0].usermax = autoTimeScaling;
          if (acop.sca[scaleIndex].log != 0)
            acop.sca[scaleIndex].setDispscale(Math.Pow(10.0, acop.sca[scaleIndex].min), Math.Pow(10.0, acop.sca[scaleIndex].max));
          tmpboolean = acop.sca[0].bestScale;
          acop.sca[0].bestScale = false;
          acop.sca[0].scaleToGrid();
          if (acop.m_marker == 2) acop.setmarkerflag(2); // recalculate marker
          // array
          acop.rescaleFlag = 1;
          acop.rescaleTextFlag = AcopConst.NEW_ZOOM_WINDOW;
          acop.prepareDraw();
          // de.desy.acop.chart.repaint(de.desy.acop.chart.sca[0].ticksRect.x, -de.desy.acop.chart.sca[0].ticksRect.y
          // ,de.desy.acop.chart.sca[0].ticksRect.width, de.desy.acop.chart.sca[0].ticksRect.height);
          // prepareDraw(TRUE, -1, FALSE, nScaleText); // rescale and not set max
          // limit
          acop.sca[0].bestScale = tmpboolean;
        }
        else
        {
          if (acop.wrapAround && acop.sca[0].log == 0) setwrapindex(); // find
          // wrapindex
          if (acop.m_marker == 2) acop.setmarkerflag(2); // recalculate marker
          // array
          acop.histIndex = hisIndex;
          acop.prepareDraw();
          // prepareDraw(FALSE,hDisplay,FALSE, 0);
        }
      }
      else
      { // fast update
        int[] repLow = new int[2];
        int[] repHigh = new int[2];
        int repeat, lowFlag;
        int xpos1 = 0, xpos2, ittpp, xonLast = 0, yonLast = 0;
        // int *iy, *ix, n;
        double xtemp = 0, ytemp, ttpp, xlast = 0, ylast = 0;
        double lowT = 0, highT = 0, ttmmpp = 0;
        int xxPP = 0; // temperary
        double overflow;
        // LPPOINT ipoint;
        Boolean xlow, xhigh, ylow, yhigh;
        repLow[0] = refBegin;
        if (refEnd > npoints || refEnd < refBegin) // Appendscreen case 5
        {
          repeat = 2;
          repHigh[0] = npoints;
          repLow[1] = 0;
          repHigh[1] = refEnd > npoints ? refEnd - npoints : refEnd;
        }
        else
        {
          repeat = 1;
          repHigh[0] = refEnd;
        }
        for (k = 0; k < repeat; k++)
        { // preset for finding refresh window
          xpostmp = xaryFlag == 1 ? -1.7e+308 : 1.7e+308;
          lowFlag = 0;
          lowTmp = 1.7e+308; // for case of ( disp[n]->xaryFlag > 1 )
          highTmp = -1.7e+308;
          for (i = (repLow[k] != 0 ? repLow[k] - 1 : repLow[k]); i < (repHigh[k] == npoints ? repHigh[k]
              : repHigh[k] + 1); i++)
          {
            ytemp = acop.sca[scaleIndex].log != 0 ? Math.Log(AcopConst.log_check(y[i])) / Math.Log(10) : y[i];
            xtemp = correctWrapPosition(i);
            ix[i] = (int)((xtemp - acop.sca[0].min) / acop.sca[0].dispSize * acop.aFrame.dxwidth
                + acop.aFrame.xleft + 0.01);
            if (ix[i] <= -16777216) ix[i] = -16777216;  // ovoid overflow
            else if (ix[i] >= 16777216) ix[i] = 16777216;
            iy[i] = (int)(acop.aFrame.histRect.Height - ((ytemp - acop.sca[scaleIndex].min) / acop.sca[scaleIndex].dispSize * acop.aFrame.histRect.Height));
            if (iy[i] <= -16777216 || iy[i] >= 16777216) iy[i] = 16777216;  // ovoid overflow
            if (i < repLow[k]) // get left point x position
            {
              if (disable[i] == 0 && (dismode == AcopConst.mode_barline ? dispxOn[i] : dispOn[i]) != 0)
              {
                xpostmp = xtemp;
                lowFlag = 1;
              }
            }
            else if (i < repHigh[k]) // refresh points
            {
              xlow = xtemp < acop.sca[0].min ? true : false;
              xhigh = xtemp > acop.sca[0].max ? true : false;
              ylow = ytemp < acop.sca[scaleIndex].min ? true : false;
              yhigh = ytemp > acop.sca[scaleIndex].max ? true : false;
              if (dismode != AcopConst.mode_polyline)
              {
                if (dismode == AcopConst.mode_barline)
                {
                  dispxOn[i] = (xlow || xhigh) ? 0 : 1;
                  dispyOn[i] = (ylow || yhigh) ? 0 : 1;
                  if (i != (repLow[k] != 0 ? repLow[k] - 1 : repLow[k])) // avoid
                  // first
                  // index
                  {
                    if ((dispxOn[i] == 0 && xonLast == 0
                        && ((xlow && xlast > acop.sca[0].max) || (xlast < acop.sca[0].min && xhigh)) && yonLast != 0)
                        || (dispyOn[i] == 0 && yonLast == 0
                            && ((ylow && ylast > acop.sca[scaleIndex].max) || (ylast < acop.sca[scaleIndex].min && yhigh)) && dispxOn[i] != 0))
                    {
                      dispxOn[i] = 1;
                      dispyOn[i] = 1;
                    }
                  }
                  // store last value from the first index of the loop
                  xonLast = dispxOn[i];
                  yonLast = dispyOn[i];
                  xlast = xtemp;
                  ylast = ytemp;
                  dispOn[i] = dispxOn[i] * dispyOn[i];
                }
                else if (dismode == AcopConst.mode_dots || dismode == AcopConst.mode_rectangle
                    || dismode == AcopConst.mode_circle || dismode == AcopConst.mode_annotation)
                {
                  dispOn[i] = (xlow || xhigh || ylow || yhigh) ? 0 : 1;
                }
                else
                  dispOn[i] = (xlow || xhigh) ? 0 : 1;
                if (disable[i] != 0) dispOn[i] = 0; // disable this point
                if (dismode == AcopConst.mode_histogram || dismode == AcopConst.mode_histogram_rastoring)
                { // avoid 100% cpu time to display a huge number!!
                  if (iy[i] < 0)
                    iy[i] = -20;
                  else if (iy[i] > acop.aFrame.histRect.Height) iy[i] = acop.aFrame.histRect.Height + 20;
                }
              }
              else
              // fastest drawing mode is to use polyline()
              {
                dispOn[i] = disable[i] != 0 ? 0 : 1;
                if (dispOn[i] != 0)
                {
                  dispxOn[i] = 0;
                  if (xlow) dispxOn[i] |= 0x1; // mask 4 cases
                  if (xhigh) dispxOn[i] |= 0x2;
                  if (ylow) dispxOn[i] |= 0x4;
                  if (yhigh) dispxOn[i] |= 0x8;
                  if (dispxOn[i] == 0)
                  {
                    if (dispyOn[i] == 0) // not marked yet
                    {
                      dispyOn[i] = 1; // point inside
                      polycounter++;
                    }
                  }
                  else
                  {
                    if (dispyOn[i] != 0) // was marked
                    {
                      dispyOn[i] = 0; // point outside
                      polycounter--;
                    }
                  }
                }
              }
              if (marker_ary_cpy != 0 && markerFlag != 0) loadMarker(i);
              if (newX != 0 && xaryFlag <= 1 && lowFlag == 1) // check x direction
              {
                if (xaryFlag == 1)
                {
                  if (xpostmp < xtemp) xaryFlag = 2;
                }
                else if (xpostmp > xtemp) xaryFlag = 2;
              }
              if (i == repLow[k] && xaryFlag <= 1) // calculate xpos1 for
              // freshscreen area
              {
                if (lowFlag == 1)
                {
                  if (xaryFlag == 1)
                  {
                    ttpp = xpostmp > xtemp ? xpostmp : xtemp;
                    if (xpostmp > xtemp) xpostmp = xtemp; // select small one
                  }
                  else
                  {
                    ttpp = xpostmp < xtemp ? xpostmp : xtemp;
                    if (xpostmp < xtemp) xpostmp = xtemp; // select large one
                  }
                }
                else
                {
                  xpostmp = xtemp; // set the left one as xpos1
                  lowFlag = 1;
                  ttpp = xtemp;
                }
                xpos1 = (int)((ttpp - acop.sca[0].min) / acop.sca[0].dispSize * acop.aFrame.dxwidth
                    + acop.aFrame.xleft + 0.01);
              }
              // compare every point for xpos2
              if ((dismode == AcopConst.mode_barline ? dispxOn[i] : dispOn[i]) != 0)
              {
                if (xaryFlag == 1)
                {
                  if (xpostmp > xtemp) xpostmp = xtemp; // select small one
                }
                else
                {
                  if (xpostmp < xtemp) xpostmp = xtemp; // select large one
                }
              }
            }
            else
            { // get right point x position
              if (disable[i] == 0 && (dismode == AcopConst.mode_barline ? dispxOn[i] : dispOn[i]) != 0)
              {
                if (xaryFlag <= 1)
                {
                  if (xaryFlag == 1 ? (xpostmp < xtemp) : (xpostmp > xtemp)) xtemp = xpostmp;
                }
                if (newX != 0 && xaryFlag <= 1 && lowFlag == 1) // check x
                // direction
                {
                  if (xaryFlag == 1)
                  {
                    if (xpostmp < xtemp) xaryFlag = 2;
                  }
                  else if (xpostmp > xtemp) xaryFlag = 2;
                }
              }
              // else if (repeat == 1)xtemp = xpostmp; // load valid xtemp
            }
            // if ( disp[n]->xaryFlag > 1 ) // mixed up and down x
            if (disable[i] == 0 && (dismode == AcopConst.mode_barline ? dispxOn[i] : dispOn[i]) != 0)
            {
              if (xtemp < lowTmp) lowTmp = xtemp;
              if (xtemp > highTmp) highTmp = xtemp;
            }
          }
          if (repeat == 2 && k == 0) // save first part
          {
            lowT = lowTmp;
            highT = highTmp;
            ttmmpp = xtemp;
            xxPP = xpos1; // temperary
          }
        }
        if (dismode == AcopConst.mode_polyline) makePolyArray();
        if (marker_ary_cpy != 0 && markerFlag != 0) marker_ary_cpy = 0;
        for (k = repeat - 1; k >= 0; k--)
        {
          if (repeat == 2 && k == 0)
          {
            lowTmp = lowT;
            highTmp = highT;
            xtemp = ttmmpp;
            xpos1 = xxPP; // restore
          }
          if (xaryFlag <= 1)
          {
            xpos2 = (int)((xtemp - acop.sca[0].min) / acop.sca[0].dispSize * acop.aFrame.dxwidth
                + acop.aFrame.xleft + 0.01);
          }
          else
          {
            xpos1 = (int)((lowTmp - acop.sca[0].min) / acop.sca[0].dispSize * acop.aFrame.dxwidth
                + acop.aFrame.xleft + 0.01);
            xpos2 = (int)((highTmp - acop.sca[0].min) / acop.sca[0].dispSize * acop.aFrame.dxwidth
                + acop.aFrame.xleft + 0.01);
          }
          // if ( (xpos1 >= de.desy.acop.chart.aFrame.histRect.x && xpos1 <=
          // de.desy.acop.chart.aFrame.histRect.x + de.desy.acop.chart.aFrame.histRect.width) ||
          // (xpos2 >= de.desy.acop.chart.aFrame.histRect.x && xpos2 <= de.desy.acop.chart.aFrame.histRect.x +
          // de.desy.acop.chart.aFrame.histRect.width) )
          if ((xpos1 >= 0 && xpos1 <= acop.aFrame.histRect.Width)
              || (xpos2 >= 0 && xpos2 <= acop.aFrame.histRect.Width))
          {
            if (xaryFlag == 1)
            {
              ittpp = xpos1;
              xpos1 = xpos2;
              xpos2 = ittpp;
            }
            fsize = (edgeMode != 0 && edgePixel > 2) ? (edgePixel + 1) * acop.aFrame.histRect.Width : 0;
            if (xpos1 <= xpos2)
            {
              if (xpos1 < 0) xpos1 = 0;
              if (xpos2 > acop.aFrame.histRect.Width)
                xpos2 = acop.aFrame.histRect.Width;
              else
                xpos2 += penWidth; // disp[tagHist]
              acop.tagRect.X = xpos1 - fsize + acop.aFrame.histRect.X; // +de.desy.acop.chart.tagRect.x;
              acop.tagRect.Width = xpos2 + fsize - xpos1 + fsize;
              acop.repaintTagging();
            }
            else
            {
              if (xpos1 < acop.aFrame.histRect.Width)
              {
                acop.tagRect.X = xpos1 - fsize + acop.aFrame.histRect.X;
                acop.tagRect.Width = acop.aFrame.histRect.Width - (xpos1 - fsize);
                acop.repaintTagging();
              }
              if (xpos2 > 0)
              {
                if (xpos2 < acop.aFrame.histRect.Width) xpos2 += penWidth; // disp[tagHist]
                acop.tagRect.X = acop.aFrame.histRect.X;
                acop.tagRect.Width = xpos2;
                acop.repaintTagging();
              }
            }
          }
        }
      }
      return num - 1; // last updated point, before was hDisplay;
    }
    protected internal int setHisto(Object ydata, Object xdata, Object dis, String[] xlabel, int arraySize,
        int maxNumber, Color c)
    {
      int inputNumber, ndata, i, xlength, dislength, tmpPoints;
      int itmp = 0, fftret = 0;
      if (ydata is String[])
      {
        ndata = ((String[])ydata).Length;
        str = new String[ndata];
        for (i = 0; i < ndata; i++)
          str[i] = ((String[])ydata)[i];
        npoints = ndata;
        return 0;
      }
      if (ydata is String)
      {
        str = ((String)ydata).Split('\n');
        npoints = str.Length;
        return 0;
      }
      if (c == null) c = Color.Black;
      str = null;
      double[] yaxis; // = acop.getAcopTransport().getYAxis(ydata);
      double[] xaxis; // = acop.getAcopTransport().getXAxis(xdata);
      //    if (acop.getAcopTransport() != null && !Beans.isDesignTime())
      //    {
      //      yaxis = acop.getAcopTransport().getYAxis(ydata);
      //      xaxis = acop.getAcopTransport().getXAxis(xdata);
      //    }
      //    else
      {
        yaxis = getPlotableArray(ydata);
        xaxis = getPlotableArray(xdata);
      }
      if (ydata == null) npoints = 0;
      ndata = getArrayLength(yaxis); // ydata.length;
      if (arraySize > 0 && arraySize < ndata) ndata = arraySize; // user given
      // input size
      xlength = getArrayLength(xaxis);
      dislength = getArrayLength(dis);
      inputNumber = ndata; // at least y array size or less if user give a
      // smaller number!
      if (maxNumber > ndata) ndata = maxNumber; // user request more memory to be
      // allocated!
      if (ndata <= (acop.globalDisplayMode < AcopConst.mode_histogram ? 1 : 0)) return -1;
      for (i = 0; i < 2; i++)
        if (acop.sca[i].usermax <= acop.sca[i].usermin && acop.sca[i].bestScale == false
            && acop.sca[i].log == 0) return -1;
      if (ndata > AcopConst.max_points) ndata = AcopConst.max_points;
      if (y == null || y.Length < ndata)
      {
        displayON = 0;
        markerFlag = 0;
        marker_ary_cpy = 0;
        lockFlag = 0;
        if (acop.pTagColorArray == m_tagColorArray) acop.pTagColorArray = null;
        iy = new int[ndata];
        ix = new int[ndata];
        imy = new int[ndata];
        y = new double[ndata];
        x = new double[ndata];
        markery = new double[ndata];
        dispOn = new int[ndata];
        dispxOn = new int[ndata];
        dispyOn = new int[ndata];
        disable = new char[ndata];
        dispindex = new int[ndata];
        dispnumber = new int[ndata];
        m_tagColorArray = new Color[ndata];
        //      xlab = new java.lang.String[ndata];
        wrapindex = 0; // drawing dtart index for wraparound
        wraphigh = 0; // right boundary greater than s_xmax
        disFlag = 0;
        originxmin = -1.7e+308;
        originxmax = 1.7e+308;
        initTagBuffer = 0;
        errWindowEnable = acop.errWindowOn;
        yScaleEnable = acop.ScaleEnable;
        yReferenceEnable = acop.ReferenceEnable;
        err_xmin = acop.e_xmin;
        err_xmax = acop.e_xmax;
        err_ymin = acop.e_ymin;
        err_ymax = acop.e_ymax;
        edgeIndex = 0;
      }
      bindSecondaryY = false;
      arrpoints = inputNumber; // input data length, user given or y size
      npoints = ndata; // total displayable data
      if (xlength > ndata) xlength = ndata;
      if (dislength > ndata) dislength = ndata;
      errorColor = acop.m_errorColor;
      polycounter = 0;
      if (npoints > acop.drawWidthPointsLimit && acop.drawWidth > 1 && !acop.forceWidth) acop.drawWidth = 1; //2008-11 hwu
      if (stroke == null || penWidth != acop.drawWidth || penStyle != acop.drawStyle || penColor != c)
      {
        penWidth = acop.drawWidth;
        penStyle = acop.drawStyle;
        penColor = c; // getForeground();
        ANNOTATION_STRING_BRUSH = new SolidBrush(penColor);
        stroke = AcopConst.createStroke(c, penWidth, penStyle);
      }
      halfWidth = acop.drawWidth / 2;
      xaryFlag = 0;
      // fill in the LeadingEdge parameters here:
      gaps = acop._LeadingEdgeGapSize;
      if (gaps >= npoints) gaps = 0;
      moving = acop._LeadingEdgeMotion;
      edgeMode = acop._LeadingEdgeMode;
      edgeExtentX = acop._LeadingEdgeXExtent;
      edgeExtentY = acop._LeadingEdgeYExtent;
      edgePercentX = -1;
      edgePercentY = -1;
      edgePixel = acop._LeadingEdgeWidth; // de.desy.acop.chart.aFrame.dxwidth;
      if (edgeStroke == null || edgeWidth != acop._LeadingEdgeWidth || edgeStyle != acop._LeadingEdgeStyle || edgeColor != acop._LeadingEdgeColor)
        edgeStroke = AcopConst.createStroke(acop._LeadingEdgeColor, acop._LeadingEdgeWidth, acop._LeadingEdgeStyle);
      edgeWidth = acop._LeadingEdgeWidth;
      edgeStyle = acop._LeadingEdgeStyle;
      edgeColor = acop._LeadingEdgeColor;
      // fill in the marker parameters here:
      if (mStroke == null || markerWidth != acop.m_markerWidth || markerStyle != acop.m_markerStyle || markerColor != acop.m_markerColor)
        mStroke = AcopConst.createStroke(acop.m_markerColor, acop.m_markerWidth, acop.m_markerStyle);
      markerMode = acop.m_markerMode;
      markerStyle = acop.m_markerStyle;
      markerWidth = acop.m_markerWidth;
      markerColor = acop.m_markerColor;
      markerExtentX = acop.m_markerXExtent;
      markerExtentY = acop.m_markerYExtent;
      markerPercentX = -1;
      markerPercentY = -1;
      if (acop.globalDisplayMode == AcopConst.mode_histogram
          || acop.globalDisplayMode == AcopConst.mode_histogram_rastoring || acop.globalDisplayMode == AcopConst.mode_dots
          || acop.globalDisplayMode == AcopConst.mode_rectangle || acop.globalDisplayMode == AcopConst.mode_circle || acop.globalDisplayMode == AcopConst.mode_annotation)
      { // bring to top histogram!
        if (acop.pTagColorArray == null) // init tagColorArray only for first
        // time, Dont care data size changed!!
        {
          acop.nElement = ndata;
          for (i = 0; i < ndata; i++)
            m_tagColorArray[i] = penColor;
          acop.pTagColorArray = m_tagColorArray;
          acop.TagBackColor = penColor;
        }
        else if (initTagBuffer == 0
            || (dismode != AcopConst.mode_histogram && dismode != AcopConst.mode_histogram_rastoring
                && dismode != AcopConst.mode_dots && dismode != AcopConst.mode_rectangle && dismode != AcopConst.mode_circle && dismode != AcopConst.mode_annotation))
        { // new buffer or change mode ==> sign new backgroud color (tagging
          // inheritance!!)
          for (i = 0; i < ndata; i++)
          {
            if (i < acop.nElement)
              m_tagColorArray[i] = acop.TagBackColor == acop.pTagColorArray[i] ? penColor
                  : acop.pTagColorArray[i];
            else
              m_tagColorArray[i] = penColor;
          }
          acop.pTagColorArray = m_tagColorArray;
          acop.TagBackColor = penColor;
          acop.nElement = ndata;
        }
        else if (vorColor != penColor)
        { // only change vorcolor
          for (i = 0; i < ndata; i++)
            if (m_tagColorArray[i] == vorColor) m_tagColorArray[i] = penColor;
          acop.pTagColorArray = m_tagColorArray;
          acop.TagBackColor = penColor;
          acop.nElement = ndata;
        }
        initTagBuffer = 1;
        tagHist = 1;
        tagColor = acop.m_tagColor; // store pen color and tagging color, sign top
        tagWidth = acop.m_tagWidth;
        // hist number
        vorColor = penColor; // TagBackColor;
        acop.lastTagHisto = hisIndex;
      }
      else
        tagHist = 0;
      if (xlabel == null)
        xlabON = 0;
      else
      {
        xlabON = 1;
        xlabLength = xlabel.Length;
        if (xlab == null || xlab.Length != ndata) xlab = new String[ndata];
        for (i = 0; i < xlabel.Length; i++)
        {
          if (i >= ndata) break;
          xlab[i] = xlabel[i];
        }
      }
      tmpPoints = arrpoints < npoints ? arrpoints : npoints;
      copyArray(y, yaxis, 0, tmpPoints, tmpPoints, 1);
      acop.averagingScaling(y, 0, npoints, hisIndex);
      if (xlength != 0) copyArray(x, xaxis, 0, xlength, xlength, 1);
      for (i = xlength; i < tmpPoints; i++)
      {
        if (xaxis == null)
          x[i] = i;
        else
          x[i] = x[xlength - 1];
      }
      disFlag = 0;
      if (dislength != 0) copyCharArray(disable, dis, dislength);
      for (i = 0; i < tmpPoints; i++)
      {
        if (dis == null)
          disable[i] = (char)0;
        else if (i < dislength)
        {
          if (disable[i] != 0) disFlag++;
        }
        else
        {
          disable[i] = (char)1;
          disFlag++;
        }
      }
      if (arrpoints < npoints) for (i = arrpoints; i < ndata; i++)
        {
          y[i] = y[arrpoints - 1];
          x[i] = x[arrpoints - 1];
          disable[i] = (char)1;
          disFlag++;
        }
      if (acop.displayMode == AcopConst.mode_barline_bit) saveBitDisplayMode = 2;
      else if (acop.displayMode == AcopConst.mode_barline_ganttplot) saveBitDisplayMode = 1;
      else saveBitDisplayMode = 0;
      for (i = 0; i < ndata; i++)
      {
        if (i >= npoints) break;
        if (Math.Abs(acop._XShift) > AcopConst.zero_check) x[i] += acop._XShift; // shift
        // first
        if (Math.Abs(acop._XScale) > AcopConst.zero_check) x[i] *= acop._XScale;
        if (Math.Abs(acop._YShift) > AcopConst.zero_check) y[i] += acop._YShift; // shift
        // first
        if (Math.Abs(acop._YScale) > AcopConst.zero_check) y[i] *= acop._YScale;
        if (acop.invert && Math.Abs(y[i]) > AcopConst.zero_check) y[i] = 1 / y[i];
        if (acop.displayMode == AcopConst.mode_barline_bit || acop.displayMode == AcopConst.mode_barline_ganttplot)
          if (saveBitDisplayMode != 0)
          {
            if (saveBitDisplayMode == 1)
            {
              if ((int)y[i] != 0) y[i] = acop.ganttBitDisplayBase + acop.ganttBitDisplayHeight;
              else y[i] = acop.ganttBitDisplayBase;
            }
            else
            {
              if ((acop.bitDisplayMask == 0 && (int)y[i] != 0) || ((int)y[i] & acop.bitDisplayMask) != 0) y[i] = acop.ganttBitDisplayBase + acop.ganttBitDisplayHeight;
              else y[i] = acop.ganttBitDisplayBase;
            }
          }
        if (xaxis == null) continue; // dont check x array
        if (i > 0)
        {
          if (itmp == 0) // check down
          {
            if (x[i] < x[i - 1])
            {
              fftret++;
              if (i != 1 && fftret == 1) fftret++; // catch first up then down
              itmp = 1;
            }
          }
          else
          { // check up
            if (x[i] > x[i - 1])
            {
              fftret++;
              itmp = 0;
            }
          }
        }
      }
      saveXShift = acop._XShift;
      saveXScale = acop._XScale;
      saveYShift = acop._YShift;
      saveYScale = acop._YScale;
      saveInvert = acop.invert;
      saveGanttBitDisplayBase = acop.ganttBitDisplayBase;
      saveGanttBitDisplayHeight = acop.ganttBitDisplayHeight;
      saveBitDisplayMask = acop.bitDisplayMask;
      if (xaxis == null)
        xaryFlag = acop._XShift < 0 ? 1 : 0;
      else
        xaryFlag = fftret;
      fFT = acop.FFT;
      if (fFT > 0 && fFT <= AcopConst.FFT_RI_INVERSE) // fft => overwrite y array
      {
        if (acopfft == null) acopfft = new AcopFFT();
        double[] maxPerCut = new double[3];
        maxPerCut[1] = 0.5;
        maxPerCut[2] = 12;
        acopfft.fFT(npoints, y, y, maxPerCut, fFT);
      }
      else if (fFT > AcopConst.FFT_RI_INVERSE) drawChanges(fFT);
      nextPos = disFlag < npoints ? npoints : 0;
      // set nextPos for Appendscreen if available data points are less then
      // requested maxpoints
      if (arrpoints < npoints && disFlag == npoints - arrpoints) nextPos = arrpoints;
      return 0;
    }
    private Color selTagErrPen(Graphics a, int i, Color lastColor)
    {
      Color selColor;
      // error color take precedence
      if (errWindowEnable && (x[i] < err_xmin || x[i] > err_xmax || y[i] < err_ymin || y[i] > err_ymax))
        selColor = errorColor;
      else if (tagHist != 0 && acop.m_tagging)
        selColor = m_tagColorArray[i];
      else
        selColor = penColor;
      if (selColor == null) selColor = Color.Black; //2012-11
      //    if (selColor != lastColor) a.setColor(selColor);
      return selColor;
    }
    protected internal void setExtent()
    {
      int scaleIndex = bindSecondaryY ? 2 : 1;
      if (dismode != AcopConst.mode_rectangle && dismode != AcopConst.mode_circle && dismode != AcopConst.mode_dots && disDots == 0) return;
      if (dismode != AcopConst.mode_dots && disDots == 0 && dismode != AcopConst.mode_annotation)
      {
        xextent = (short)(acop.aFrame.dxwidth * Math.Abs(xext / 2) / (acop.sca[0].log != 0 ? (Math.Pow(10.0,
          acop.sca[0].max) - Math.Pow(10.0, acop.sca[0].min)) : acop.sca[0].dispSize));
        yextent = (short)(acop.aFrame.histRect.Height * Math.Abs(yext / 2) / (acop.sca[scaleIndex].log != 0 ? (Math.Pow(
          10.0, acop.sca[scaleIndex].max) - Math.Pow(10.0, acop.sca[scaleIndex].min)) : acop.sca[scaleIndex].dispSize));
      }
      else if (disDots != 0)
      {
        if (penWidth > 7)
        {
          xextent = penWidth * 105 / 100;
          yextent = penWidth * 105 / 100;
        }
        else
        {
          xextent = penWidth * dotsScale[penWidth] / 100;
          yextent = penWidth * dotsScale[penWidth] / 100;
        }
      }
      else
      {
        xextent = 1;
        yextent = 1;
      }
      if (xextent == 0) xextent = penWidth;
      if (yextent == 0) yextent = penWidth;
    }
    private void loadMarker(int i)
    {
      double ytemp;
      int scaleIndex = bindSecondaryY ? 2 : 1;
      if (marker_ary_cpy == 1) // store marker array
      {
        markery[i] = y[i];
        imy[i] = iy[i];
      }
      else
      // get stored marker array
      {
        ytemp = acop.sca[scaleIndex].log != 0 ? Math.Log(AcopConst.log_check(markery[i])) / Math.Log(10) : markery[i];
        imy[i] = (int)(acop.aFrame.histRect.Height - ((ytemp - acop.sca[scaleIndex].min) / acop.sca[scaleIndex].dispSize * acop.aFrame.histRect.Height));
        // avoid 100% cpu time to display a huge number!!
        if (imy[i] < 0)
          imy[i] = -20;
        else if (imy[i] > acop.aFrame.histRect.Height) imy[i] = acop.aFrame.histRect.Height + 20;
      }
    }
    private double correctWrapPosition(int i)
    {
      double xtemp;
      xtemp = acop.sca[0].log != 0 ? Math.Log(AcopConst.log_check(x[i])) / Math.Log(10) : x[i];
      if (acop.wrapAround == false || wraphigh == 0 || acop.sca[0].log != 0) return xtemp;
      if (wraphigh < 0)
      {
        if (i >= -(wraphigh + 1))
        {
          if (xaryFlag == 1)
            xtemp += (acop.wrapxmax - acop.wrapxmin);
          else
            xtemp -= (acop.wrapxmax - acop.wrapxmin);
        }
      }
      else
      {
        if (i < wraphigh)
        {
          if (xaryFlag == 1)
            xtemp -= (acop.wrapxmax - acop.wrapxmin);
          else
            xtemp += (acop.wrapxmax - acop.wrapxmin);
        }
      }
      return xtemp;
    }
    protected internal void mapScreen(int index)
    {
      double xtemp, ytemp;
      int kk, i, start, stop, scaleIndex;
      Boolean xlow, xhigh, ylow, yhigh;
      int xonLast = 0, yonLast = 0;
      double xlast = 0, ylast = 0;
      String stmp;
      if (index < 0)
      {
        start = 0;
        stop = npoints;
        polycounter = 0;
      }
      else
      {
        start = index;
        stop = index + 1;
      }
      String allStmp = "";
      scaleIndex = bindSecondaryY ? 2 : 1;
      for (kk = start, i = wrapindex; kk < stop; kk++, i++)
      {
        if (i >= npoints) i = 0;
        if (dismode == AcopConst.mode_textbox)
        {
          if (str != null)
          {
            if (i >= str.Length) return;
            if (acop.appendLineIndexInTextMode)
              stmp = "(" + hisIndex.ToString() + "," + i.ToString() + ")  " + str[i];
            else
              stmp = str[i];
          }
          else
          {
            if (acop.appendLineIndexInTextMode)
              stmp = "(" + hisIndex.ToString() + "," + i.ToString() + ")  " + x[i].ToString()
              //              + "    " + AcopConst.textFormat.format(y[i]);
              + "    " + String.Format("{0:0.###}", y[i]);
            else
              stmp = x[i].ToString() + "    " + String.Format("{0:0.###}", y[i]);
          }
          if (disable != null && i < disable.Length && disable[i] != 0) stmp += " [disabled]";
          //        acop.listTextBox.add(stmp); // addItem(stmp);
          allStmp = allStmp + stmp + "\n";
          continue;
        }
        ytemp = acop.sca[scaleIndex].log != 0 ? Math.Log(AcopConst.log_check(y[i])) / Math.Log(10) : y[i];
        xtemp = correctWrapPosition(i);
        ix[i] = (int)((xtemp - acop.sca[0].min) / acop.sca[0].dispSize * acop.aFrame.dxwidth
            + acop.aFrame.xleft + 0.01);
        iy[i] = (int)(acop.aFrame.histRect.Height - ((ytemp - acop.sca[scaleIndex].min) / acop.sca[scaleIndex].dispSize * acop.aFrame.histRect.Height));
        if (iy[i] <= -16777216 || iy[i] >= 16777216) iy[i] = 16777216;  // ovoid overflow
        xlow = xtemp < acop.sca[0].min ? true : false;
        xhigh = xtemp > acop.sca[0].max ? true : false;
        ylow = ytemp < acop.sca[scaleIndex].min ? true : false;
        yhigh = ytemp > acop.sca[scaleIndex].max ? true : false;
        if (ix[i] < -acop.aFrame.histRect.Width * 4)
          ix[i] = -acop.aFrame.histRect.Width * 4;
        else if (ix[i] > acop.aFrame.histRect.Width * 4) ix[i] = acop.aFrame.histRect.Width * 4;
        if (dismode != AcopConst.mode_polyline) // map x,y to display area
        {
          if (dismode == AcopConst.mode_barline)
          {
            dispxOn[i] = (xlow || xhigh) ? 0 : 1;
            dispyOn[i] = (ylow || yhigh) ? 0 : 1;
            if ((wrapindex != npoints && i != wrapindex) || (wrapindex == npoints && i != 0)) // avoid
            // first
            // point
            {
              if ((dispxOn[i] == 0 && xonLast == 0
                  && ((xlow && xlast > acop.sca[0].max) || (xlast < acop.sca[0].min && xhigh)) && yonLast != 0)
                  || (dispyOn[i] == 0 && yonLast == 0
                      && ((ylow && ylast > acop.sca[scaleIndex].max) || (ylast < acop.sca[scaleIndex].min && yhigh)) && dispxOn[i] != 0))
              {
                dispxOn[i] = 1;
                dispyOn[i] = 1;
              }
            }
            // store last value
            xonLast = dispxOn[i];
            yonLast = dispyOn[i];
            xlast = xtemp;
            ylast = ytemp;
            dispOn[i] = dispxOn[i] * dispyOn[i];
          }
          else if (dismode == AcopConst.mode_dots || dismode == AcopConst.mode_rectangle
              || dismode == AcopConst.mode_circle || dismode == AcopConst.mode_annotation)
          {
            dispOn[i] = (xlow || xhigh || ylow || yhigh) ? 0 : 1;
          }
          else
            dispOn[i] = (xlow || xhigh) ? 0 : 1;
          if (disable[i] != 0) dispOn[i] = 0; // disable this point
          if (dismode == AcopConst.mode_histogram || dismode == AcopConst.mode_histogram_rastoring)
          { // avoid 100% cpu time to display a huge number!!
            if (iy[i] < 0)
              iy[i] = -20;
            else if (iy[i] > acop.aFrame.histRect.Height) iy[i] = acop.aFrame.histRect.Height + 20;
          }
        }
        else
        // fastest drawing mode is to use polyline()
        {
          dispOn[i] = disable[i] != 0 ? 0 : 1;
          if (dispOn[i] != 0)
          {
            dispxOn[i] = 0;
            dispyOn[i] = 0;
            if (xlow) dispxOn[i] |= 0x1; // mask 4 cases
            if (xhigh) dispxOn[i] |= 0x2;
            if (ylow) dispxOn[i] |= 0x4;
            if (yhigh) dispxOn[i] |= 0x8;
            if (dispxOn[i] == 0)
            {
              if (dispyOn[i] == 0 || index < 0) // not marked yet
              {
                dispyOn[i] = 1; // point inside
                polycounter++;
              }
            }
            else if (index >= 0 && dispyOn[i] != 0) // was marked
            {
              dispyOn[i] = 0; // point outside
              polycounter--;
            }
          }
        }
        if (marker_ary_cpy != 0 && markerFlag != 0) loadMarker(i);
      }
      if (dismode == AcopConst.mode_textbox)
      {
        acop.addItemsToListBox(allStmp.Split('\n'));
      }
      if (dismode == AcopConst.mode_polyline) makePolyArray();
      if (marker_ary_cpy != 0 && markerFlag != 0) marker_ary_cpy = 0;
      if (saveBitDisplayMode > 0)
        mapedBase = (int)(acop.aFrame.histRect.Height - ((saveGanttBitDisplayBase - acop.sca[scaleIndex].min) / acop.sca[scaleIndex].dispSize * acop.aFrame.histRect.Height));
    }
    protected int getIndex(int x)
    {
      int i, j, tmpflg = 0, XarrayElement, xtmp, lastIndex;
      if (ix == null) return 0;
      if (dismode == AcopConst.mode_textbox) return 0;
      lastIndex = wrapindex;
      xtmp = 0;
      for (i = 0, j = wrapindex; i < npoints; i++, j++)
      {
        if (j >= npoints) j = 0;
        if (dispOn[j] == 0) // this point is not displayed, check the previous posint
        {
          if (xaryFlag == 1)
          {
            if (x >= ix[lastIndex] - xtmp) break;
          }
          else
          {
            if (x < ix[lastIndex] + xtmp) break;
          }
        }
        if (j == wrapindex)
        {
          if (xaryFlag == 1)
          {
            if (x >= ix[j])
            {
              tmpflg = 1;
              break;
            }
          }
          else
          {
            if (x < ix[j])
            {
              tmpflg = 1;
              break;
            }
          }
          xtmp = penWidth;
        }
        else
        {
          xtmp = j == 0 ? npoints - 1 : j - 1;
          // TODO: there's (sometimes) an array index out of bound exception here when there's a
          // mouse move over a text-mode display. 
          // 1). why ?
          // 2).  text mode should not even check for this event
          xtmp = (ix[j] - ix[xtmp]) / 2;
        }
        lastIndex = j;
        if (xaryFlag == 1)
        {
          if (x >= ix[j] + xtmp) break;
        }
        else
        {
          if (x < ix[j] - xtmp) break;
        }
      }
      if (tmpflg == 1)
        XarrayElement = j;
      else
      {
        if (xaryFlag == 1)
        {
          XarrayElement = j == npoints ? j - 1 : j;
        }
        else
        {
          if (wrapindex != 0 && j == 0) j = npoints;
          XarrayElement = j == 0 ? 0 : j - 1;
        }
      }
      return XarrayElement;
    }
    protected internal int getHistoIndex(double x, double y)
    {
      int devX;
      double tmpX;
      tmpX = acop.sca[0].log != 0 ? Math.Log(AcopConst.log_check(x)) / Math.Log(10) : x;
      devX = (int)((tmpX - acop.sca[0].min) * acop.aFrame.dxwidth / acop.sca[0].dispSize + acop.aFrame.xleft);
      return getIndex(devX);
    }
    protected internal int getMousePosition(Point point, double[] userPoint)
    {
      int i, j, tmpflg = 0, XarrayElement, xtmp;
      int scaleIndex = bindSecondaryY ? 2 : 1;
      int lastIndex = scaleIndex == 1 ? 2 : 1;
      point.X -= acop.aFrame.histRect.X;
      point.Y -= acop.aFrame.histRect.Y;
      userPoint[0] = acop.sca[0].dispSize * (point.X - acop.aFrame.xleft) / acop.aFrame.dxwidth
          + acop.sca[0].min;
      userPoint[1] = acop.sca[scaleIndex].dispSize * (acop.aFrame.histRect.Height - point.Y)
          / acop.aFrame.histRect.Height + acop.sca[scaleIndex].min;
      userPoint[2] = acop.sca[lastIndex].dispSize * (acop.aFrame.histRect.Height - point.Y)
          / acop.aFrame.histRect.Height + acop.sca[lastIndex].min;
      if (acop.sca[0].log != 0) userPoint[0] = Math.Pow(10.0, userPoint[0]);
      if (acop.sca[scaleIndex].log != 0) userPoint[1] = Math.Pow(10.0, userPoint[1]);
      if (acop.sca[lastIndex].log != 0) userPoint[2] = Math.Pow(10.0, userPoint[2]);
      return getIndex(point.X);
    }
    protected internal double maxOf(double[] arr)
    {
      int i;
      double max = -1.7E308;
      if (disFlag == 0)
      {
        for (i = 0; i < npoints; i++)
          if (arr[i] > max) max = arr[i];
      }
      else
      {
        for (i = 0; i < npoints; i++)
        {
          if (disable[i] != 0) continue;
          if (arr[i] > max) max = arr[i];
        }
      }
      return max;
    }
    protected internal double minOf(double[] arr)
    {
      int i;
      double min = 1.7E308;
      if (disFlag == 0)
      {
        for (i = 0; i < npoints; i++)
          if (arr[i] < min) min = arr[i];
      }
      else
      {
        for (i = 0; i < npoints; i++)
        {
          if (disable[i] != 0) continue;
          if (arr[i] < min) min = arr[i];
        }
      }
      return min;
    }
    protected internal void getMaxWindow()
    {
      double off, tmin, tmax;
      int setFirstTime = 0;
      if (npoints < 2)
      {
        tmax = x[0] + 1;
        tmin = x[0] - 1;
      }
      else
      {
        tmax = maxOf(x);
        tmin = minOf(x);
        // all points disabled!
        if (Math.Abs(tmax + 1.7e+308) < AcopConst.zero_check
            && Math.Abs(originxmin + 1.7e+308) < AcopConst.zero_check
            && Math.Abs(originxmax - 1.7e+308) < AcopConst.zero_check)
        {
          setFirstTime = 1;
          tmax = acop.sca[0].dispSize / 2 + 1;
          tmin = tmax - 2;
        }
        off = (tmax - tmin) / (2 * (npoints - 1));
        if (off < AcopConst.zero_check) off = 1;
        tmax += off;
        tmin -= off;
      }
      // if different x size, then
      if (Math.Abs(tmax - originxmax) > AcopConst.zero_check
          || Math.Abs(tmin - originxmin) > AcopConst.zero_check || setFirstTime != 0)
      {
        originxmax = tmax;
        originxmin = tmin;
        if (hisIndex == acop.m_histstart)
        {
          acop.wrapxmin = originxmin;
          acop.wrapxmax = originxmax;
        }
      }
      if (acop.sca[0].log != 0) setwrapindex();
    }
    protected internal void setwrapindex()
    {
      int i, j, highfound;
      double xtemp;
      if (acop.sca[0].min < acop.wrapxmin && acop.sca[0].max < acop.wrapxmax)
        j = xaryFlag == 1 ? 2 : 1; // left
      else if (acop.sca[0].max > acop.wrapxmax && acop.sca[0].min > acop.wrapxmin)
        j = xaryFlag == 1 ? 1 : 2;// right
      else
        j = 0;
      wrapindex = wraphigh = 0;
      if (j > 0)
      {
        for (i = 0, highfound = -1; i < npoints; i++)
        {
          xtemp = x[i];
          if (j == 1)
          {
            if (xaryFlag == 1)
            {
              if (xtemp <= acop.wrapxmin + (acop.sca[0].max - acop.wrapxmax) // ???
                                                                             // change
                                                                             // to
                                                                             // <=
                  && xtemp <= acop.sca[0].min) break;
            }
            else
            {
              if (xtemp >= acop.wrapxmax - (acop.wrapxmin - acop.sca[0].min) // ???
                                                                             // change
                                                                             // to
                                                                             // >=
                  && xtemp >= acop.sca[0].max) break;
            }
          }
          else
          {
            if (xaryFlag == 1)
            {
              if (xtemp >= acop.wrapxmax - (acop.wrapxmin - acop.sca[0].min) && xtemp > acop.sca[0].max)
                highfound = i;
              if (xtemp <= acop.sca[0].max) break;
            }
            else
            {
              if (xtemp <= acop.wrapxmin + (acop.sca[0].max - acop.wrapxmax) && xtemp < acop.sca[0].min)
                highfound = i;
              if (xtemp >= acop.sca[0].min) break;
            }
          }
        }
        wrapindex = i == npoints ? 0 : i;
        if (j == 1)
        {
          if (i != npoints) wraphigh = -(wrapindex + 1);
        }
        else if (highfound != -1)
        {
          wraphigh = highfound;
          if (highfound < npoints - 1 && i != npoints)
            wraphigh = i == (highfound + 1) ? highfound : highfound + 1;
          wraphigh++;
        }
        // disp[n]->wraphigh = (j == 1) ? 0 : (k+1);
      }
    }
    private void makePolyArray()
    {
      int i, j, k, kk;
      if (disFlag == 0 && (acop.wrapAround == false || acop.sca[0].log != 0))
      {
        dispindex[0] = 0;
        dispnumber[0] = npoints; // # of points
        dispindex[1] = -1;
      }
      else
      {
        j = 0;
        k = 0;
        dispindex[k] = -1;
        for (kk = 0, i = wrapindex; kk < npoints && k < npoints; kk++, i++)
        {
          if (i >= npoints) i = 0;
          if (dispOn[i] != 0)
          {
            if (wrapindex != 0 && wraphigh > 0 && i >= wraphigh && i < wrapindex) break;
            if (dispindex[k] == -1) dispindex[k] = i; // load first index
            j++;
            if (wrapindex != 0 && i == (npoints - 1))
            {
              dispnumber[k] = j; // # of points
              j = 0;
              dispindex[++k] = -1;
            }
          }
          else
          {
            if (dispindex[k] != -1)
            {
              dispnumber[k] = j; // # of points
              j = 0;
              dispindex[++k] = -1;
            }
          }
        }
        if (dispindex[k] != -1)
        {
          dispnumber[k] = j;
          dispindex[++k] = -1;
        }
      }
    }
    private void drawPoly(Graphics a, Pen pen, int pstart, int pend)
    {
      Point[] pdata = new Point[pend - pstart];
      for (int i = 0, j = pstart; i < pend - pstart; i++, j++)
      {
        pdata[i].X = ix[j];
        pdata[i].Y = iy[j];
      }
      a.DrawLines(pen, pdata);
      //    if (pstart == 0)
      //    {
      //      a.drawPolyline(ix,iy,pend);
      //      return;
      //    }
      //    int[] tix = new int[pend-pstart], tiy = new int[pend-pstart];
      //    Array.Copy(ix, pstart, tix, 0, pend-pstart);
      //    Array.Copy(iy, pstart, tiy, 0, pend-pstart);
      //    a.drawPolyline(tix,tiy,pend-pstart);
    }

    protected internal void drawhist(Graphics a)
    {
      //by jbobnar: Do not draw the histo if acop is not visible
      if (!acop.Visible) return;
      int i, j, k, OndisFlag, wx, wy, mx, tmpWidth;
      Color selColor, lastColor;
      Boolean colorCheck;
      Pen lastPen;
      SolidBrush lastBrush;
      if (displayON == 0) return;
      lastColor = penColor; // init tag drawing
      lastPen = stroke;
      lastBrush = new SolidBrush(lastColor);
      if (dismode == AcopConst.mode_textbox)
      {
        acop.listTextBox.Visible = true;
        return;
      }
      //    a.setStroke(stroke);
      lock (stroke)
      {
        acop.listTextBox.Visible = false;
        if (dismode == AcopConst.mode_polyline)
        {
          double tantheta, atantheta, abssin, abscos, delta;
          int xstart, ystart, xstart2, ystart2, delx, dely;
          /*
           * for (i=wrapindex, j=0; j<npoints -1; j++,i++) { if ( i >= npoints ) i =
           * 0; // i: current point k = i+1; if ( k >= npoints ) k = 0; // i:
           * current point if ( dispOn[i] == 0 || dispOn[k] == 0 ) continue; //
           * aLine.DrawAcopLine(ix[i],iy[i],ix[k],iy[k]);
           * a.drawLine(ix[i],iy[i],ix[k],iy[k]); } if ( i >= -10 ) return;
           */
          for (i = 0; i < npoints; i++)
          {
            if ((wx = dispindex[i]) == -1) break;
            wy = wx; // load first index
            for (j = 1; j < dispnumber[i]; j++)
            {
              // both outside, draw last part if there is one, then set index = -1
              if ((dispxOn[wx + j] & dispxOn[wx + j - 1]) != 0)
              {
                if (wy != -1 && (wx + j - 1) > wy)
                {
                  //              a.drawPolyline(tix,tiy,wx+j-wy);
                  drawPoly(a, stroke, wy, wx + j);
                  if (disDots != 0)
                  {
                    for (k = wy; k < wx + j - 1; k++)
                    {
                      //                a.drawLine(ix[k], iy[k], ix[k + 1], iy[k + 1]);
                      if (dispyOn[k] != 0)
                      {
                        if (disDots == 2)
                        {
                          a.DrawLine(stroke, ix[k] - xextent, iy[k] - yextent, ix[k] + xextent, iy[k] + yextent);
                          a.DrawLine(stroke, ix[k] - xextent, iy[k] + yextent, ix[k] + xextent, iy[k] - yextent);
                        }
                        else if (disDots != 0) a.FillRectangle(ANNOTATION_STRING_BRUSH, ix[k] - xextent, iy[k] - yextent, xextent * 2, yextent * 2);

                      }
                    }
                    if (dispyOn[k + 1] != 0)
                    {
                      if (disDots == 2)
                      {
                        a.DrawLine(stroke, ix[k + 1] - xextent, iy[k + 1] - yextent, ix[k + 1] + xextent, iy[k + 1] + yextent);
                        a.DrawLine(stroke, ix[k + 1] - xextent, iy[k + 1] + yextent, ix[k + 1] + xextent, iy[k + 1] - yextent);
                      }
                      else if (disDots != 0) a.FillRectangle(ANNOTATION_STRING_BRUSH, ix[k + 1] - xextent, iy[k + 1] - yextent, xextent * 2, yextent * 2);
                    }
                  }
                }
                wy = -1;
                continue;
              }
              // otherwise: if nindex not yet set, load new index number
              if (wy == -1)
              {
                // pdc->MoveTo(ipoint[wx+j-1].x, ipoint[wx+j-1].y); // move to
                // lastpoint
                wy = wx + j - 1;
              }
              // pdc->LineTo(ipoint[wx+j].x, ipoint[wx+j].y);
            }
            // draw last part if there is one
            if (wy != -1 && (wx + j - 1) > wy)
            {
              // drawPolyline(ix[wy],iy[wy],wx+j-wy);
              drawPoly(a, stroke, wy, wx + j);
              if (disDots != 0)
              {
                for (k = wy; k < wx + j - 1; k++)
                {
                  //            a.drawLine(ix[k], iy[k], ix[k + 1], iy[k + 1]);
                  if (dispyOn[k] != 0)
                  {
                    if (disDots == 2)
                    {
                      a.DrawLine(stroke, ix[k] - xextent, iy[k] - yextent, ix[k] + xextent, iy[k] + yextent);
                      a.DrawLine(stroke, ix[k] - xextent, iy[k] + yextent, ix[k] + xextent, iy[k] - yextent);
                    }
                    else if (disDots != 0) a.FillRectangle(ANNOTATION_STRING_BRUSH, ix[k] - xextent, iy[k] - yextent, xextent * 2, yextent * 2);
                  }
                }
                if (dispyOn[k] != 0)
                {
                  if (disDots == 2)
                  {
                    a.DrawLine(stroke, ix[k] - xextent, iy[k] - yextent, ix[k] + xextent, iy[k] + yextent);
                    a.DrawLine(stroke, ix[k] - xextent, iy[k] + yextent, ix[k] + xextent, iy[k] - yextent);
                  }
                  else if (disDots != 0) a.FillRectangle(ANNOTATION_STRING_BRUSH, ix[k] - xextent, iy[k] - yextent, xextent * 2, yextent * 2);
                }
              }
            }
          }
          if (acop.wrapAround && wrapindex != 0 && acop.sca[0].log == 0) // connect
          // end
          // with
          // begining
          {
            i = npoints - 1;
            if (dispOn[0] != 0 && dispOn[i] != 0 && (dispxOn[0] & dispxOn[i]) == 0)
            {
              OndisFlag = 0;
              if (xaryFlag == 1)
              {
                if (ix[0] <= ix[i]) OndisFlag = 1;
              }
              else if (ix[0] >= ix[i]) OndisFlag = 1;
              if (OndisFlag != 0) a.DrawLine(stroke, ix[0], iy[0], ix[i], iy[i]);
            }
          }
          // if ( i >= -10 ) return;
          /*
           * if ( penWidth < 2 ) a.drawPolyline(ix, iy, npoints); else { //
           * java.awt.Polygon poly = new java.awt.Polygon(ix, iy, npoints); // Line
           * penLine = new Line(); for (i = 0; i < npoints-1; i++) { if ( ix[i] ==
           * ix[i+1] || iy[i] == iy[i+1] ) {
           * a.drawLine(ix[i],iy[i],ix[i+1],iy[i+1]); continue; } tantheta =
           * (iy[i+1]-iy[i])/(ix[i+1]-ix[i]); atantheta = Math.atan(tantheta);
           * abssin = Math.abs(Math.sin(atantheta)); abscos =
           * Math.abs(Math.cos(atantheta)); // abssin = Math.abs(tantheta /
           * Math.sqrt(1+tantheta*tantheta)); // abscos = Math.abs(abssin/tantheta);
           * if ( tantheta > 1 || tantheta < -1 ) { delta = 1/tantheta; if (tantheta<0.)
           * abscos = -abscos; } else { delta = tantheta; if (tantheta<0.) abssin =
           * -abssin; } xstart = (int)(ix[i] - penWidth / 2 * abssin); xstart2 =
           * (int)(ix[i+1] - penWidth / 2 * abssin); ystart = (int)(iy[i] + penWidth /
           * 2 * abscos); ystart2 = (int)(iy[i+1] + penWidth / 2 * abscos); // for
           * (int j = 0; j < penWidth; j++) // { // delx = (int) (j * abssin); //
           * dely = (int) (j * abscos); // AcopLine.a.drawLine(xstart+delx,
           * ystart-dely,xstart2+delx, ystart2-dely); // } if ( tantheta > 1 ||
           * tantheta < -1 ) { delx = (int) (penWidth * abssin); for (j=0; j < delx;
           * j++) { dely = (int) (delta * j) ; a.drawLine(xstart+j, ystart
           * -dely,xstart2+j, ystart2-dely); } } else { dely = (int) (penWidth *
           * abscos); for (j=0; j < dely; j++) { delx = (int) (delta * j) ;
           * a.drawLine(xstart+delx, ystart-j ,xstart2+delx, ystart2-j); } } } //
           * for (int j = 0; j < penWidth; j++) // for (int i = 0; i < npoints-1;
           * i++) AcopLine.a.drawLine(ix[i],iy[i]-penWidth / 2+ j
           * ,ix[i+1],iy[i+1]-penWidth / 2 + j); }
           */
        }
        else if (dismode == AcopConst.mode_barline)
        {
          for (i = wrapindex, j = 0; j < npoints - 1; j++, i++)
          {
            if (i >= npoints) i = 0; // i: current point
            k = i + 1;
            if (k >= npoints) k = 0; // i: current point
            if (disable[k] != 0 || disable[i] != 0) continue; // current disable

            if (disBarHisto != 0)
            {
              mx = ix[i] + (ix[k] - ix[i]) / 2;
              if (barlineColorFilled == 0)
              {
                a.DrawLine(stroke, ix[i], iy[i], mx, iy[i]);
                a.DrawLine(stroke, mx, iy[i], mx, iy[k]);
                a.DrawLine(stroke, mx, iy[k], ix[k], iy[k]);
              }
              else if (barlineColorFilled < 0)
              {
                a.FillRectangle(ANNOTATION_STRING_BRUSH, ix[i], iy[i], mx - ix[i], acop.aFrame.histRect.Height - iy[i]);
                a.FillRectangle(ANNOTATION_STRING_BRUSH, mx, iy[k], ix[k] - mx, acop.aFrame.histRect.Height - iy[k]);
              }
              else
              {
                a.FillRectangle(ANNOTATION_STRING_BRUSH, ix[i], 0, mx - ix[i], iy[i]);
                a.FillRectangle(ANNOTATION_STRING_BRUSH, mx, 0, ix[k] - mx, iy[k]);
              }
            }
            else if (saveBitDisplayMode == 2)
            {
              a.DrawLine(stroke, ix[i], iy[i], ix[k], iy[i]);
              a.DrawLine(stroke, ix[k], iy[i], ix[k], iy[k]);
            }
            else if (saveBitDisplayMode == 1)
            {
              if (y[i] > saveGanttBitDisplayBase + 0.001)
                a.FillRectangle(ANNOTATION_STRING_BRUSH, ix[i], iy[i], ix[k] - ix[i], mapedBase - iy[i]);
            }
            else
            {
              if (barlineColorFilled == 0)
              {
                a.DrawLine(stroke, ix[i], iy[i], ix[k], iy[i]);
                a.DrawLine(stroke, ix[k], iy[i], ix[k], iy[k]);
              }
              else if (barlineColorFilled < 0)
                a.FillRectangle(ANNOTATION_STRING_BRUSH, ix[i], iy[i], ix[k] - ix[i], acop.aFrame.histRect.Height - iy[i]);
              else
                a.FillRectangle(ANNOTATION_STRING_BRUSH, ix[i], 0, ix[k] - ix[i], iy[i]);
            }
            if (disDots != 0)
            {
              // draw first dot (2018-1)
              if ( j == 0 && disable[i] == 0 ) a.FillRectangle(ANNOTATION_STRING_BRUSH, ix[i] - xextent, iy[i] - yextent, xextent * 2, yextent * 2);
              a.FillRectangle(ANNOTATION_STRING_BRUSH, ix[k] - xextent, iy[k] - yextent, xextent * 2, yextent * 2);
            }
          }
        }
        else if (dismode == AcopConst.mode_histogram || dismode == AcopConst.mode_histogram_rastoring)
        {
          colorCheck = (errWindowEnable || (tagHist != 0 && acop.m_tagging)) ? true : false;
          for (i = 0; i < npoints; i++)
          {
            if (dispOn[i] == 0) continue;
            // a.drawLine(ix[i],de.desy.acop.chart.aFrame.histRect.height,ix[i],iy[i]);
            if (colorCheck && (selColor = selTagErrPen(a, i, lastColor)) != lastColor)
            {
              if (selColor != penColor && selColor != tmpColor)
              {
                if (tagHist != 0 && acop.m_tagging && selColor != penColor) tmpWidth = tagWidth;
                else tmpWidth = penWidth;
                tmpPen = AcopConst.createStroke(selColor, tmpWidth, penStyle);
                tmpColor = selColor;
              }
              lastColor = selColor;
              lastPen = lastColor == penColor ? stroke : tmpPen;
              lastBrush = new SolidBrush(lastColor);
            }
            if (iy[i] != acop.yzero || acop.drawWidth > 2)
            {
              a.DrawLine(lastPen, ix[i], acop.yzero, ix[i], iy[i]);
              if (iy[i] == acop.yzero) a.DrawLine(lastPen, ix[i], acop.yzero, ix[i], acop.yzero + 1);
              else a.DrawLine(lastPen, ix[i], acop.yzero, ix[i], iy[i]);
              if (disDots != 0) a.FillRectangle(lastBrush, ix[i] - xextent, iy[i] - yextent, xextent * 2, yextent * 2);
            }
            else
              a.DrawRectangle(lastPen, ix[i] - 1, iy[i] - 1, 2, 2);
          }
        }
        else if (dismode == AcopConst.mode_dots)
        {
          colorCheck = (errWindowEnable || (tagHist != 0 && acop.m_tagging)) ? true : false;
          for (i = 0; i < npoints; i++)
          {
            if (dispOn[i] == 0) continue;
            if (colorCheck && (selColor = selTagErrPen(a, i, lastColor)) != lastColor)
            {
              if (selColor != penColor && selColor != tmpColor)
              {
                tmpPen = AcopConst.createStroke(selColor, penWidth, penStyle);
                tmpColor = selColor;
              }
              lastColor = selColor;
              lastPen = lastColor == penColor ? stroke : tmpPen;
              lastBrush = new SolidBrush(lastColor);
            }
            //        a.fillRect(ix[i] - 1, iy[i] - 1, 3, 3);
            a.FillRectangle(lastBrush, ix[i] - xextent, iy[i] - yextent, xextent * 2 + 1, yextent * 2 + 1);
          }
        }
        else if (dismode == AcopConst.mode_annotation)
        {
          for (i = 0; i < npoints; i++)
          {
            //          if (dispOn[i] == 0) continue;
            //          if (colorCheck && (selColor = selTagErrPen(a, i, lastColor)) != lastColor) lastColor = selColor;
            //          a.fillRect(ix[i] - 1, iy[i] - 1, 3, 3);
            //          a.setColor(ANNOTATION_COLOR);
            //          a.fillRect(ix[i]-4,iy[i]-12,15,15);
            a.FillRectangle(ANNOTATION_BRUSH, ix[i] - 4, 1, AcopConst.AnnotationSize, AcopConst.AnnotationSize);
            //          a.setColor(penColor);
            //            a.DrawString(ANNOTATION_TEXT, ANNOTATION_FONT, ANNOTATION_STRING_BRUSH, ix[i] - 3, -5);
            a.DrawString(ANNOTATION_TEXT, ANNOTATION_FONT, ANNOTATION_STRING_BRUSH, ix[i] - 2, -1);
          }
          //        a.setFont(f);
        }
        else if (dismode == AcopConst.mode_status)
        {
          colorCheck = (errWindowEnable || (tagHist != 0 && acop.m_tagging)) ? true : false;
          //int width = a.getFontMetrics().stringWidth("X")/2;
          int width = (int)a.MeasureString("X", ANNOTATION_FONT).Width / 2;
          for (i = 0; i < npoints; i++)
          {
            //        	if (dispOn[i] == 0) continue;
            if (colorCheck && (selColor = selTagErrPen(a, i, lastColor)) != lastColor) lastColor = selColor;
            a.DrawString("X", ANNOTATION_FONT, ANNOTATION_STRING_BRUSH, ix[i] - width, acop.histRoundRect.Height);
          }
        }
        else if (dismode == AcopConst.mode_rectangle)
        {
          colorCheck = (errWindowEnable || (tagHist != 0 && acop.m_tagging)) ? true : false;
          for (i = 0; i < npoints; i++)
          {
            if (dispOn[i] == 0) continue;
            if (colorCheck && (selColor = selTagErrPen(a, i, lastColor)) != lastColor)
            {
              if (selColor != penColor && selColor != tmpColor)
              {
                tmpPen = AcopConst.createStroke(selColor, penWidth, penStyle);
                tmpColor = selColor;
              }
              lastColor = selColor;
              lastPen = lastColor == penColor ? stroke : tmpPen;
              lastBrush = new SolidBrush(lastColor);
            }
            if (colorfilled)
            {
              a.FillRectangle(lastBrush, ix[i] - xextent, iy[i] - yextent, xextent * 2, yextent * 2);
            }
            else
            {
              a.DrawRectangle(lastPen, ix[i] - xextent, iy[i] - yextent, xextent * 2, yextent * 2);
            }
          }
        }
        else if (dismode == AcopConst.mode_circle)
        {
          colorCheck = (errWindowEnable || (tagHist != 0 && acop.m_tagging)) ? true : false;
          for (i = 0; i < npoints; i++)
          {
            if (dispOn[i] == 0) continue;
            if (colorCheck && (selColor = selTagErrPen(a, i, lastColor)) != lastColor)
            {
              if (selColor != penColor && selColor != tmpColor)
              {
                tmpPen = AcopConst.createStroke(selColor, penWidth, penStyle);
                tmpColor = selColor;
              }
              lastColor = selColor;
              lastPen = lastColor == penColor ? stroke : tmpPen;
              lastBrush = new SolidBrush(lastColor);
            }
            a.DrawEllipse(lastPen, ix[i] - xextent, iy[i] - yextent, xextent * 2, yextent * 2);
          }
        }
        else
        {
          acop.listTextBox.Visible = true;
        }
      }
      if (edgeMode != 0) drawLeadingEdgeMarker(a);
      drawMarker(a);
    }
    private void drawLeadingEdgeMarker(Graphics a)
    {
      int mx;
      int scaleIndex = bindSecondaryY ? 2 : 1;
      //    a.setColor(edgeColor);
      //    a.setStroke(edgeStroke);
      if ((edgeMode & 0x1) != 0)
      {
        if (edgePercentX < 0)
        {
          if (edgeExtentX < 0)
            edgePercentX = 1;
          else
            edgePercentX = (acop.sca[0].log != 0 ? Math.Log(edgeExtentX) / Math.Log(10) : edgeExtentX)
                / (acop.sca[0].xtime != 0 ? acop.sca[0].dispSize : (acop.sca[0].tick_max - acop.sca[0].tick_min)) / 2;
          edgePixel = (int)(acop.aFrame.dxwidth * edgePercentX * 2);
          if (edgePixel < edgeWidth) edgePixel = edgeWidth;
        }
        mx = (int)(acop.aFrame.dxwidth * edgePercentX);
        if (mx < 1) mx = 1; // at least 1 pixel
        a.DrawLine(edgeStroke, ix[edgeIndex] - mx, iy[edgeIndex], ix[edgeIndex] + mx, iy[edgeIndex]);
      }
      if ((edgeMode & 0x2) != 0)
      {
        if (edgePercentY < 0)
        {
          if (edgeExtentY < 0)
            edgePercentY = 1;
          else
            edgePercentY = (acop.sca[scaleIndex].log != 0 ? Math.Log(edgeExtentY) / Math.Log(10) : edgeExtentY)
                / acop.sca[scaleIndex].dispSize / 2;
        }
        mx = (int)(acop.aFrame.histRect.Height * edgePercentY);
        a.DrawLine(edgeStroke, ix[edgeIndex], iy[edgeIndex] - mx, ix[edgeIndex], iy[edgeIndex] + mx);
      }
    }
    private void drawMarker(Graphics a)
    {
      int wy, mx, tickflag;
      int scaleIndex = bindSecondaryY ? 2 : 1;
      Pen tmpPen;
      if (acop.m_marker == 0 || markerFlag == 0) return;
      if (acop.m_markerMode == 0)
      {
        if (acop.saveMarkerWidth != penWidth)
        {
          acop.markerStroke = AcopConst.createStroke(acop.m_markerColor, penWidth, AcopConst.PS_SOLID);
          acop.saveMarkerWidth = penWidth;
        }
        //      a.setColor(acop.m_markerColor);
        //      a.setStroke(acop.markerStroke);
        tmpPen = acop.markerStroke;
      }
      else
      {
        //      a.setColor(markerColor);
        //      a.setStroke(mStroke);
        tmpPen = mStroke;
      }
      for (int i = 0; i < npoints; i++)
      {
        if (dispOn[i] == 0) continue;
        tickflag = acop.m_marker == 2 ? imy[i] : iy[i];
        mx = ix[i];
        if (acop.m_markerMode == 0)
        {
          if (tickflag > 0 && tickflag < acop.aFrame.histRect.Height)
            a.DrawLine(tmpPen, mx - acop.m_markerWidth, tickflag, mx + acop.m_markerWidth, tickflag);
        }
        else
        // new one
        {
          if ((markerMode & 0x1) != 0)
          {
            if (markerPercentX < 0) // new one
            {
              if (markerExtentX < 0)
                markerPercentX = 1;
              else
                markerPercentX = (acop.sca[0].log != 0 ? Math.Log(markerExtentX) / Math.Log(10) : markerExtentX)
                    / (acop.sca[0].xtime != 0 ? acop.sca[0].dispSize : (acop.sca[0].tick_max - acop.sca[0].tick_min)) / 2;
            }
            wy = (int)(acop.aFrame.histRect.Width * markerPercentX);
            if (wy < 1) wy = 1; // at least 1 pixel
            a.DrawLine(tmpPen, mx - wy, tickflag, mx + wy, tickflag);
          }
          if ((markerMode & 0x2) != 0)
          {
            if (markerPercentY < 0)
            {
              if (markerExtentY < 0)
                markerPercentY = 1;
              else
                markerPercentY = (acop.sca[scaleIndex].log != 0 ? Math.Log(markerExtentY) / Math.Log(10) : markerExtentY)
                    / acop.sca[scaleIndex].dispSize / 2;
            }
            wy = (int)(acop.aFrame.histRect.Height * markerPercentY);
            a.DrawLine(tmpPen, mx, tickflag - wy, mx, tickflag + wy);
          }
        }
      }
    }
    public int[] getAnnotationPoints()
    {
      if (dismode == AcopConst.mode_annotation)
      {
        return ix;
      }
      else
      {
        return null;
      }
    }

    public int[] getStatusPoints()
    {
      if (dismode == AcopConst.mode_status)
      {
        return ix;
      }
      else
      {
        return null;
      }
    }
    public double getYDataMax(Boolean yFlag)
    {
      int i;
      double dTemp = yFlag ? y[0] : x[0];
      if (yFlag)
      {
        for (i = 0; i < npoints; i++) if (y[i] > dTemp) dTemp = y[i];
      }
      else
      {
        for (i = 0; i < npoints; i++) if (x[i] > dTemp) dTemp = x[i];
      }
      return dTemp;
    }
    public double getYDataMin(Boolean yFlag)
    {
      int i;
      double dTemp = yFlag ? y[0] : x[0];
      if (yFlag)
      {
        for (i = 0; i < npoints; i++) if (y[i] < dTemp) dTemp = y[i];
      }
      else
      {
        for (i = 0; i < npoints; i++) if (x[i] < dTemp) dTemp = x[i];
      }
      return dTemp;
    }
    private double getYDataSum(Boolean yFlag)
    {
      int i;
      double sum = 0;
      if (yFlag)
      {
        for (i = 0; i < npoints; i++) sum += y[i];
      }
      else
      {
        for (i = 0; i < npoints; i++) sum += x[i];
      }
      return sum;
    }
    public double getYDataMean(Boolean yFlag)
    {
      double dTemp = 0, sum = getYDataSum(!yFlag);
      if (yFlag)
      {
        dTemp = y[npoints - 1];
        for (int i = 1; i < npoints; i++)
        {
          sum = y[i - 1];
          if (acop.weightStatistics) sum *= (x[i] - x[i - 1]);
          dTemp += sum;
        }
        sum = acop.weightStatistics ? x[npoints - 1] - x[0] + 1 : npoints;
        return dTemp / sum;
      }
      for (int i = 0; i < npoints; i++) dTemp += x[i] * y[i];
      return dTemp / sum;
    }
    public double getYDataRms(Boolean yFlag)
    {
      double dTemp = 0, sum = getYDataSum(!yFlag);
      if (yFlag)
      {
        //      for (int i=0; i<npoints; i++) dTemp += x[i] * y[i] * y[i];
        dTemp = y[npoints - 1] * y[npoints - 1];
        for (int i = 1; i < npoints; i++)
        {
          sum = y[i - 1] * y[i - 1];
          if (acop.weightStatistics) sum *= (x[i] - x[i - 1]);
          dTemp += sum;
        }
        sum = acop.weightStatistics ? x[npoints - 1] - x[0] + 1 : npoints;
        return Math.Sqrt(dTemp / sum);
      }
      else
      {
        for (int i = 0; i < npoints; i++) dTemp += y[i] * x[i] * x[i];
      }
      return Math.Sqrt(dTemp / sum);
    }
    public double getYDataStdDiv(Boolean yFlag)
    {
      double dTemp = 0, sum = getYDataSum(!yFlag), mean = getYDataMean(yFlag);
      if (yFlag)
      {
        //      for (int i=0; i<npoints; i++) dTemp +=  x[i] * (y[i] - mean ) * (y[i] - mean );
        dTemp = getYDataRms(yFlag);
        mean = getYDataMean(yFlag);
        return Math.Sqrt(dTemp * dTemp - mean * mean);
      }
      else
      {
        for (int i = 0; i < npoints; i++) dTemp += y[i] * (x[i] - mean) * (x[i] - mean);
      }
      return Math.Sqrt(dTemp / sum);
    }
    // }}
    public Color getCurveColor()
    {
      return penColor;
    }
    public void setCurveColor(Color c)
    {
      penColor = c;
      ANNOTATION_STRING_BRUSH = new SolidBrush(penColor);
    }
    public int getCurveWidth()
    {
      return penWidth;
    }
    public void setCurveWidth(int w)
    {
      penWidth = w;
    }
    public void applyStrokeChange()
    {
      stroke = AcopConst.createStroke(penColor, penWidth, penStyle);
    }
  }
}
