﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace Acop.AcopChartUtil
{
  public partial class AcopPesist
  {
#   pragma warning disable 1591
    protected int size; // persistent depth
    protected int[] ids; // id array of size
    protected internal int hisIndex; // primary id
    protected int position, numCurves;
    protected Boolean copyMemory;
    protected Boolean endlessFlag;
    protected Color vorColor, pesisColor;
    private AcopChart acop;

    public AcopPesist(AcopChart acop, int index)
    {
      this.acop = acop;
      resetPesist(index);
    }
    private void resetPesist(int index)
    {
      size = 0;
      hisIndex = index;
      position = -1;
      numCurves = 0;
      copyMemory = false;
      endlessFlag = false;
    }
    private void clearPesistScreen()
    {
      if (size > 1) for (int i = 1; i < size; i++)
        {
          if (ids[i] < 0) break;
          acop.ClearScreen(ids[i]);
        }
    }
    public int initPesist(int depth, Boolean memCopy, Color c)
    {
      return initPesist(depth, memCopy, true, c);
    }
    public int initPesist(int depth, Boolean memCopy, Boolean endless, Color c)
    {
      int ret = 0;

      clearPesistScreen();
      pesisColor = c;
      size = depth;
      copyMemory = memCopy;
      endlessFlag = endless;
      if (size < 1) size = 1;
      ids = new int[size];
      ids[0] = hisIndex;
      for (int i = 1; i < size; i++) ids[i] = -1;
      vorColor = acop.histo[hisIndex].penColor;
      position = 0;
      numCurves = 1;
      return ret;
    }

    public int refreshPesist(Object Yarray)
    {
      int pos;

      // refresh last curve with persistent color
      if (numCurves < size || (size == 1 && endlessFlag) || (size > 1 && copyMemory == false))
      {
        acop.histo[ids[position]].penColor = pesisColor;
        acop.repainHisto = true;
        acop.histo[ids[position]].updateScreen(false, acop.histo[ids[position]].y, -1, -1, null);
        acop.repainHisto = false;
      }
      // draw new curve
      if (numCurves < size)
      {
        Color tmpForColor = acop.GetForeground();
        // draw current curve
        acop.SetForeground(vorColor);
        ids[position + 1] = acop.Draw(Yarray, acop.histo[hisIndex].x);
        acop.SetForeground(tmpForColor);
        numCurves++;
        position++;
        if (ids[position] >= 0) return position;
        // memory full, adjust the size
        size = position;
        numCurves = size;
        position--;
      }
      // copy curves and refreshScreen, or move the position
      if (copyMemory) for (int i = 1; i < size; i++)
        {
          Array.Copy(acop.histo[ids[i]].y, 0, acop.histo[ids[i - 1]].y, 0, acop.histo[ids[i]].npoints);
          acop.repainHisto = true;
          acop.histo[ids[i - 1]].updateScreen(false, acop.histo[ids[i - 1]].y, -1, -1, null);
          acop.repainHisto = false;
        }
      else if (size > 1) if (++position >= size) position = 0;

      // refresh new curve
      if (endlessFlag) acop.repainHisto = true;
      acop.histo[ids[position]].penColor = vorColor;
      acop.histo[ids[position]].updateScreen(false, Yarray, -1, -1, null);
      if (endlessFlag) acop.repainHisto = false;

      return position;
    }

    public void clearPesist()
    {
      if (size <= 0 || hisIndex < 0) return;
      clearPesistScreen();
      if (ids[0] >= 0) acop.histo[ids[0]].penColor = vorColor;
      resetPesist(-1);
    }
  }
}
