﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Globalization;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace Acop.AcopChartUtil
{
  public partial class AcopScale
  {
#   pragma warning disable 1591
    private static int longTick = 4, shortTick = 2;
    protected internal int gridLineMode = AcopConst.PS_DOT;
    protected internal int gridLineWidth = 1;
    protected internal Color gridLineColor = Color.Black;//Color.Cyan;
    protected internal Pen gridStroke;
    private double Rfirst, Rlast, Rinc;
    private short Nteile, Ntick;
    private double Rtick;
    private double Rmin, Rmax;
    // flags set by property
    protected internal Boolean bestScale, grid, ticktext;
    protected internal int ticks, fineTicks, xtime, log;
    protected internal String AxisLabel = "";//"(your label here)";
    protected internal Color tickLabelForeColor;
    protected internal SolidBrush tickLabelBrush;
    private Boolean xAxis; // x or y axis
    private Boolean secondYAxis; // second  y axis
    protected internal int time_scale;
    protected internal Boolean leftBottom; // where to put ticks, or right_top?
    // for x-axis: Bottom ot Top side of Histogram
    // for y-axis: Left ot Right side of Histogram
    protected internal AcopChart.AcopAxisLabelAnchor labelAnchor = AcopChart.AcopAxisLabelAnchor.bottom;
    protected internal Boolean leftBottLabel; // where to put Label
    protected internal Rectangle labelRect; // where to put label
    protected internal Rectangle ticksRect; // tick drawing rectangle
    protected internal double tick_min, tick_max; // ticks area
    protected internal double min, max, dispSize; // data (user) display area after
    // scaling
    protected internal double usermin, usermax, userSize; // user input value(property min
    // and max)
    private NumberFormatInfo nfi = new NumberFormatInfo();
    //private DecimalFormat tickFormat = new DecimalFormat("#####.###");
    //  private FontMetrics fMetrix;
    //    private int sDigits = 3;
    //    private String stickFormat = "{0:###.#}";
    private int sDigits = 4;
    private String stickFormat = "{0:###.#####}";
    private AcopChart acop;
    private AcopFrame aFrame;
    protected internal Boolean OnFlag;
    protected internal int labelOffsetMode;
    protected internal double labelOffset;
    protected static String dtFormat = "ddd MMM dd HH:mm:ss.fff yyyy";
    public AcopScale(AcopChart acop, Boolean xax, Boolean sYAxis)
    {
      this.acop = acop;
      this.aFrame = acop.aFrame;
      xAxis = xax;
      secondYAxis = sYAxis;
      nfi.NumberDecimalSeparator = ".";
      gridStroke = AcopConst.createStroke(gridLineColor, gridLineWidth, gridLineMode);
    }
    static readonly DateTime UnixEpoch = new DateTime(1970, 1, 1, 1, 0, 0, 0, DateTimeKind.Utc);
    static readonly double MaxUnixSeconds = (DateTime.MaxValue - UnixEpoch).TotalSeconds;
    static DateTime DateTimeFromUnixTimestampMillis(long millis)
    {
      return UnixEpoch.AddMilliseconds(millis);
    }
    protected static DateTime UnixTimeStampToDateTime(double unixTimeStamp)
    {
      return unixTimeStamp > MaxUnixSeconds
         ? UnixEpoch.AddMilliseconds(unixTimeStamp)
         : UnixEpoch.AddSeconds(unixTimeStamp);
    }
    protected static DateTime JavaLongToCSharpLong(long javaLong)
    {
      //    return DateTimeFromUnixTimestampMillis(javaLong);
      //    return UnixTimeStampToDateTime((double)javaLong/1000);
      //      TimeSpan ss = TimeSpan.FromMilliseconds(javaLong * 1000);
      TimeSpan ss = TimeSpan.FromMilliseconds(javaLong);
      DateTime Jan1st1970 =
          new DateTime(1970, 1, 1, 0, 0, 0, 0, DateTimeKind.Utc);
      DateTime ddd = Jan1st1970.Add(ss);
      //DateTime final = ddd.ToUniversalTime();
      // fix sommer time problem 2015-4-21
      DateTime final = ddd.ToLocalTime();

      return final;
    }
    protected internal int getScreenPosition(double userPosition)
    {
      double tmp;
      if (xAxis)
      {
        if (xtime != 0)
          tmp = (userPosition - min) / dispSize;
        else
          tmp = ((log != 0 ? (Math.Log(userPosition) / Math.Log(10) + 0.0001) : userPosition) - tick_min)
              / (tick_max - tick_min);
        // add left to fix markercursor x poition jump, 2009
        tmp = tmp * aFrame.dxwidth + aFrame.histRect.X + aFrame.xleft;
      }
      else
        tmp = aFrame.histRect.Height
            * (1 - ((log != 0 ? (Math.Log(userPosition) / Math.Log(10) + 0.0001) : userPosition) - min)
                / dispSize) + aFrame.histRect.Y;
      return (int)tmp;
    }
    protected internal double getUserPosition(int screenPosition)
    {
      double tmp;
      if (xAxis)
      {
        tmp = (screenPosition - aFrame.xleft) / aFrame.dxwidth;
        if (xtime != 0)
          tmp = tmp * dispSize + min;
        else
          tmp = tmp * (tick_max - tick_min) + tick_min;
      }
      else
        tmp = (screenPosition - aFrame.histRect.Y) / aFrame.histRect.Width * dispSize + min;
      if (log != 0) tmp = Math.Pow(10.0, tmp);
      return tmp;
    }
    protected internal double getMin()
    {
      return log != 0 ? Math.Pow(10.0, min) : min;
    }
    protected internal double getMax()
    {
      return log != 0 ? Math.Pow(10.0, max) : max;
    }
    protected internal void setDispscale(double low, double high) // set scaled size
    {
      min = low;
      max = high;
      dispSize = max - min;
    }
    protected internal void setscale()
    {
      setDispscale(usermin, usermax);
    }
    protected internal void setUserscale(double low, double high) // set scaled size
    {
      usermin = low;
      usermax = high;
      userSize = usermax - usermin;
    }
    protected internal Boolean isOriginMaxMin()
    {
      if (Math.Abs(min - usermin) < AcopConst.zero_check && Math.Abs(max - usermax) < AcopConst.zero_check)
        return true;
      else
        return false;
    }
    private void logRound()
    {
      int i1, i2;
      int zero_flag = 0;
      double tmpmin, tmpmax;
      tmpmin = Rmin;
      tmpmax = Rmax;
      if (Rmin < AcopConst.zero_check)
      { /* check zero */
        Rmin = 0.1; // 10 to -1
        zero_flag = 1;
      }
      if (Rmax < AcopConst.zero_check)
      {
        Rmax = 10.0; // 10 to 1
        zero_flag += 2;
      }
      if (zero_flag == 1)
      { /* make new boundary */
        if (Rmax < Rmin * 5.0) if (Rmax * 0.01 > tmpmin) Rmin = Rmax * 0.01;
      }
      else if ((zero_flag == 0 && Rmax < Rmin * 5.0) || zero_flag == 2)
        if (Rmin * 100.0 < tmpmax) Rmax = Rmin * 100.0;
      if (Rmin < tmpmin && (zero_flag & 0x1) == 0) Rmin = tmpmin; /* check limit */
      if (Rmax > tmpmax && (zero_flag & 0x2) == 0) Rmax = tmpmax;
      i1 = Rmin >= 1.0 ? (int)(Math.Log(Rmin) / Math.Log(10) + 0.01)
          : (int)(Math.Log(Rmin) / Math.Log(10) - 0.99);
      i2 = Rmax > 1.0 ? (int)(Math.Log(Rmax) / Math.Log(10) + 0.99)
          : (int)(Math.Log(Rmax) / Math.Log(10) - 0.01);
      if (i1 == i2) i1--;
      Rfirst = i1;
      Rlast = i2;
      Rinc = 1;
      Nteile = (short)(i2 - i1 + 1);
      Rtick = 2; // take log10
      Ntick = 5;
    }
    private void lineRound()
    {
      double del, dl, f, dlm, fdl;
      int ideca;
      del = Rmax - Rmin;
      if (Math.Abs(del) < AcopConst.zero_check)
      { /* ??????? */
        Rfirst = Rmin;
        Rinc = 1.0;
        Rlast = Rmax;
        Nteile = 2;
        Ntick = 1;
        Rtick = 0.5;
      }
      else
      {
        if (del < 0.0)
        { /* revers it */
          dlm = Rmin;
          Rmin = Rmax;
          Rmax = dlm;
          del = Math.Abs(del);
        }
        dl = Math.Log(del) / Math.Log(10);
        dlm = Math.Abs(Rmin) > Math.Abs(Rmax) ? Math.Log(Math.Abs(Rmin)) / Math.Log(10) : Math.Log(Math
            .Abs(Rmax))
            / Math.Log(10);
        if (Rmin < 0.0 && Rmax > 0.0) dl = dlm;
        dlm = (double)((int)(dlm));
        fdl = (double)((int)(dl));
        f = Math.Abs(dl - fdl);
        if (fdl > 0.0)
        {
          if (fdl > dlm + 2.0)
          {
            fdl = dlm + 1.0;
            f = 0.0;
          }
        }
        else
        {
          if (fdl < dlm - 2.0)
          {
            fdl = dlm - 3.0;
            f = 0.0;
          }
        }
        Rinc = 1.0;
        if (del > 1.0)
        {
          Rinc = Math.Pow(10.0, fdl);
          if (fdl < 10) Rinc = (double)Math.Round(Rinc);
          // sca[index].Rinc = (double) round ( pow (10.0, fdl) );
          if (f <= 0.31)
            Rinc /= 5.0;
          else if (f <= 0.71) Rinc /= 2.0;
        }
        else
        {
          fdl -= 1.0;
          Rinc = Math.Pow(10.0, fdl);
          if (f > 0.69)
            Rinc /= 5.0;
          else if (f > 0.29) Rinc /= 2.0;
        }
        Rfirst = Rmin >= 0.0 ? ((int)(Rmin / Rinc)) * Rinc : ((int)(Rmin / Rinc - 0.99)) * Rinc;
        Rlast = Rmax >= 0.0 ? ((int)(Rmax / Rinc + 0.99)) * Rinc : ((int)(Rmax / Rinc)) * Rinc;
        Nteile = (short)(Math.Round((Rlast - Rfirst) / Rinc + 1.0));
        ideca = (int)(Math.Log(Math.Abs(Rinc)) / Math.Log(10));
        if (ideca < 0) ideca--;
        Rtick = Math.Pow(10.0, (double)ideca);
        Ntick = (short)(Math.Round(Rinc / Rtick) - 1);
        if (Ntick < 1) Ntick = 1; // for loop to wotk!!!???
      }
      ideca = ticks; // user defines ticks!!
      if (ideca < 0)
        return;
      else if (ideca == 0) ideca = 1;
      del = bestScale ? (Rlast - Rfirst) : (Rmax - Rmin);
      Nteile = (short)(ideca + 1);
      Rinc = del / ideca;
      ideca = fineTicks;
      Ntick = (short)(ideca <= 1 ? 1 : ideca);
      Rtick = Rinc / Ntick;
    }
    private void setLebelOffset()
    {
      if (log != 0) return;
      if (labelOffsetMode == 0) labelOffset = 0;
      else if (labelOffsetMode == 1) labelOffset = tick_min;
      else if (labelOffsetMode == 2)
      {
        long diff = (long)(tick_max - tick_min + 0.5);
        int digits = diff.ToString().Length;
        long factor = (long)Math.Pow(10, digits);
        labelOffset = (((long)(tick_min / factor)) * factor);
      }
    }
    protected internal void scaleToGrid()
    {
      int xstep, i;
      if (xtime != 0) // convert to time tick
      {
        if (min < 0) min = 0; // time cannt be < 0!
        if (max - min < 0.001) max = min + 0.001; // one second is minimal => change to 1ms
                                                  //        if (max - min < 1) max = min + 1; // one second is minimal
        dispSize = max - min;
        xstep = (int)dispSize;
        for (i = 0; i < AcopConst.N_TIME; i++)
        {
          if (xstep > AcopConst.timeMax[i] * AcopConst.timeScale[i]) continue;
          time_scale = i;
          break;
        }
        if (i == AcopConst.N_TIME) i = AcopConst.N_TIME - 1;  // 2007 hwu
        if (aFrame.leftTimeLabel == true)
        {
          tick_min = 0;
          tick_max = dispSize / AcopConst.timeScale[i];
        }
        else
        {
          tick_min = -dispSize / AcopConst.timeScale[i];
          tick_max = 0;
        }
      }
      else
      {
        tick_min = min;
        tick_max = max;
      }
      setLebelOffset();
    }
    protected internal void xyscale(Boolean setYaxis)
    {
      if (OnFlag == false) return;
      if (xAxis == false)
      {
        tick_min = min;
        tick_max = max;
      }
      Rmin = tick_min;
      Rmax = tick_max;
      if (log != 0)
        logRound();
      else
        lineRound();
      if (bestScale || log != 0)
      {
        tick_min = Rfirst;
        tick_max = Rlast;
        if (xtime != 0)
        {
          if (aFrame.leftTimeLabel)
          {
            min = min + (log != 0 ? Math.Pow(10.0, tick_min) : tick_min) * AcopConst.timeScale[time_scale]; // back
            // to
            // normal
            // units
            max = ((log != 0 ? Math.Pow(10.0, tick_max) : tick_max) - (log != 0 ? Math.Pow(10.0, tick_min)
                : tick_min))
                * AcopConst.timeScale[time_scale] + min;
          }
          else
          {
            max = max + (log != 0 ? Math.Pow(10.0, tick_max) : tick_max) * AcopConst.timeScale[time_scale]; // back
            // to
            // normal
            // units
            min = max
                - ((log != 0 ? Math.Pow(10.0, tick_max) : tick_max) - (log != 0 ? Math.Pow(10.0, tick_min)
                    : tick_min)) * AcopConst.timeScale[time_scale];
          }
          if (log != 0)
          {
            min = Math.Log(AcopConst.log_check(min)) / Math.Log(10);
            max = Math.Log(AcopConst.log_check(max)) / Math.Log(10);
          }
        }
        else
        {
          min = tick_min;
          max = tick_max;
        }
        dispSize = max - min;
        if (xAxis == false && setYaxis && log == 0)
        {
          usermin = min;
          usermax = max;
          userSize = usermax - usermin;
        }
      }
      setLebelOffset();
    }
    protected internal void drawGrid(Graphics a)
    {
      Boolean tickflag;
      int wx, wy, k, m;
      double tmp;
      AcopHisto hist;
      if (!grid) return;
      tmp = tick_max - tick_min;
      tickflag = (Nteile <= 2 && log != 0) ? true : false;
      wy = xAxis ? aFrame.dxwidth : aFrame.histRect.Height;
      //    for (int i = (tickflag ? 1 : 0); i < (tickflag ? Ntick : Nteile); i++)
      for (int i = 0; i < (tickflag ? (Ntick + 1) : Nteile); i++)
      {
        if (i == 0 && tickflag) wx = xAxis ? 0 : wy;
        else if (xAxis)
          wx = (int)(wy * (tickflag ? Math.Log(Rtick * i) / Math.Log(10) : (Rinc * i)) / tmp + 0.01);
        else
          wx = (int)(wy * (1 - (tickflag ? Math.Log(Rtick * i) / Math.Log(10) : (Rinc * i)) / tmp) + 0.01);
        if (wx > wy || wx < 0) continue;
        // fix y grid diplay problem, 2009
        if (xAxis) wx += aFrame.xleft;
        // if ( wx >= wy || wx == 0 ) continue;
        if (xAxis)
        {
          hist = acop.histo[acop.m_histstart];
          if (hist == null) return;
          if (aFrame.drawFlag == false || hist.xlabON == 0 || log != 0 || xtime != 0)
            a.DrawLine(gridStroke, wx, 0, wx, aFrame.histRect.Height);
          else
          {
            for (k = 0, m = hist.wrapindex; k < hist.npoints; k++, m++)
            {
              if (m >= hist.npoints) m = 0;
              if (wx <= hist.ix[m]) break;
            }
            if (k == hist.npoints) break;
            if (hist.ix[m] <= aFrame.xright && hist.ix[m] > 1) // within
            // histogram
            {
              wx = hist.ix[m];
              a.DrawLine(gridStroke, wx, 0, wx, aFrame.histRect.Height);
            }
          }
        }
        else
          a.DrawLine(gridStroke, 0, wx, aFrame.histRect.Width, wx);
      }
    }
    private void convert(String str, double w, int[] format)
    {
      /*
       * char s_format [10] ; int i, j, k ; s_format [0] ='%' ; if ( format[2] )
       * sprintf ( &s_format[1], ".%de", format[1]) ; else sprintf (&s_format[1],
       * ".%df", format[1]) ; sprintf ( string, s_format, w) ; if ( format[2] ) //
       * delete unneccesary + & 0 { for ( i = 0; string [i++] != 'e'; ) ; k = i-1 ;
       * j = string [i++] == '+' ? i : i+1 ; // store copy position while ( string
       * [i] == '0' && string [i] != 0 ) i++ ; if ( string [i] == 0 ) { string
       * [j=k] = 0 ; // no exp format[2] = 0 ; } else { // copy rest significant
       * digits do { string[j++] = string[i++] ; } while ( string[i-1] != 0 ) ;
       * j-- ; } } else for (j = 0 ; string[j] ; j++ ); format[3] = j; // length
       * of the string
       */
    }
    private void g_stellen(double wert, int[] format)
    {
      double z;
      int stellen = 5; // default effective # of digits
      format[1] = 1; // hinter der komma
      format[2] = 0; // exp format
      format[4] = 0; // exp format digits after e;
      z = Math.Abs(wert) > AcopConst.zero_check ? // make z = etwas * 1.0e3
          wert / Math.Pow(10.0, (double)(int)(Math.Log(Math.Abs(wert)) / Math.Log(10.0) * 1.001)) * 1.0e3 : 0.0;
      z = Math.Abs(z);
      wert = Math.Abs(wert);
      z = wert < 1.0 ? z * 10.0 + 0.05 : z + 0.05; // make etwas=x.xxx
      while (z - (int)z < 0.1 && stellen > 1) // get effective digits
      {
        stellen--;
        z /= 10.0;
      }
      if (wert >= 100000.0 || (wert > AcopConst.zero_check && wert < 0.001))
      {
        format[2] = 2; // { mit Exponent }
        if (wert >= 1.0E10)
          format[2] = 3;
        else if (wert <= 1.0E-10) format[2] = 4;
        z = wert * 1.0001; // avoid 10e+-xx problems 2005-Feb
        format[4] = (int)(Math.Log(z) / Math.Log(10.0));
        // format[4] = (int)(Math.log(wert)/Math.log(10.));
        if (format[4] < 0) format[4]--; // digits after exponential
      }
      format[0] = 1; // vor der komma
      if (wert >= 0.001 && wert < .999)
      {
        format[0] = 0;
        if (wert < 0.00999)
        {
          if (stellen > 3)
            stellen = 3;
          else
            stellen += 2;
        }
        else if (wert < 0.0999)
        {
          if (stellen > 4)
            stellen = 2;
          else
            stellen++;
        }
      }
      if (wert >= 10.0 && wert < 100.0)
        format[0] = 2;
      else if (wert >= 100.0 && wert < 1000.0)
        format[0] = 3;
      else if (wert >= 1000.0 && wert < 10000.0)
        format[0] = 4;
      else if (wert >= 10000.0 && wert < 100000.0) format[0] = 5;
      format[1] = stellen - format[0];
      if (format[1] < 0) format[1] = 0;
    }
    protected void drawLabel(Graphics a)
    {
      int wz, hei;
      if (AxisLabel == null || AxisLabel.Length == 0) return;
      RectangleF Cliprect = a.ClipBounds;
      if (Cliprect != null && Cliprect.IntersectsWith(labelRect) == false) return;
      SizeF labelStringSize;
      labelStringSize = a.MeasureString(AxisLabel, aFrame.frameFont);

      a.SetClip(labelRect);   
      // property XTicksBottom -> leftBottom
      // property XLabelLowerLeft -> leftBottLabel
      if (xAxis)
      {
        a.TranslateTransform(labelRect.X, labelRect.Y);
        hei = (int)labelStringSize.Height;
        //wz = labelAnchor == AcopChart.AcopAxisLabelAnchor.bottom ?
        //  labelRect.Height - 4 - hei : 2 * hei + 4;
        if (labelAnchor == AcopChart.AcopAxisLabelAnchor.bottom)
          wz = leftBottom ? labelRect.Height - 4 - hei : labelRect.Height - 4 - 2 * hei;
        else
          wz = leftBottom ? hei + 4 : hei + 4;
        a.DrawString(AxisLabel, aFrame.frameFont, tickLabelBrush, (labelRect.Width - labelStringSize.Width) / 2, wz);
        a.TranslateTransform(-labelRect.X, -labelRect.Y);
      }
      else
      {
        hei = (int)labelStringSize.Height;   
        if (leftBottLabel == true)
          a.TranslateTransform(labelRect.X + hei / 4, labelRect.Y + (labelRect.Height + labelStringSize.Width) / 2);
        else
          a.TranslateTransform(labelRect.X + labelRect.Width - hei, labelRect.Y + (labelRect.Height + labelStringSize.Width) / 2);
        a.RotateTransform(-90);
        a.DrawString(AxisLabel, aFrame.frameFont, tickLabelBrush, 0, 0);
        a.RotateTransform(90);
        if (leftBottLabel == true)
          a.TranslateTransform(-labelRect.X - hei / 4, -labelRect.Y - (labelRect.Height + labelStringSize.Width) / 2);
        else
          a.TranslateTransform(-labelRect.X - labelRect.Width + hei, -labelRect.Y - (labelRect.Height + labelStringSize.Width) / 2);
      }
      a.SetClip(Cliprect);
    }
    protected internal void drawTicks(Graphics a, AcopHisto pHisto)
    {
      int wx, wy, wz, lastright, k, m, xpos = 0, xtp, ytp;
      double tmp, rtmp;
      int[] strformat = { 0, 0, 0, 0, 0 }; // vor, hinter, texp, lang, exp-digits;
      String str = "erty";
      int siz, hei, tlen, strmaxlen, itmp;
      Rectangle txtRect = new Rectangle();
      //    Date date = new Date();
      DateTime dt = DateTime.Now;
      SizeF textStringSize;
      if ((xAxis ? ticksRect.Height : ticksRect.Width) < 1) return;
      RectangleF Cliprect = a.ClipBounds;
      if (Cliprect != null && Cliprect.IntersectsWith(ticksRect) == false) return;
      //    a.setFont(aFrame.frameFont);
      //    a.setColor(aFrame.frameForeColor);
      //    fMetrix = a.getFontMetrics();
      //    SizeF txtSize = a.MeasureString("12345", aFrame.frameFont);
      strmaxlen = 0;
      if (xAxis)
      {
        a.SetClip(ticksRect); //clipRect(ticksRect.X, ticksRect.Y - 1, ticksRect.Width + 1, ticksRect.Height + 4);
        a.TranslateTransform(aFrame.histRect.X + aFrame.xleft, ticksRect.Y - 1);
        //      hei = fMetrix.getAscent();
        hei = (int)aFrame.frameFont.Size * aFrame.frameFont.FontFamily.GetCellAscent(aFrame.frameFont.Style) / aFrame.frameFont.FontFamily.GetEmHeight(aFrame.frameFont.Style);
        txtRect.Y = leftBottom ? 4 : 0;
        txtRect.Height = ticksRect.Height - 4;
        wy = aFrame.dxwidth;
        lastright = -100000;
      }
      else
      {
        a.SetClip(ticksRect);  //.clipRect(ticksRect.X - 2, ticksRect.Y, ticksRect.Width + 3, ticksRect.Height);
        a.TranslateTransform(ticksRect.X, aFrame.histRect.Y);
        txtRect.X = 0;
        txtRect.Width = ticksRect.Width - 4;
        //      hei = (fMetrix.getDescent() + fMetrix.getAscent()) / 2;
        //      hei = aFrame.frameFont.FontFamily.GetEmHeight(aFrame.frameFont.Style) / 2;
        hei = (int)aFrame.frameFont.Size;
        wy = aFrame.histRect.Height;
        lastright = 100000;
      }
      if (ticks != 0 && (xtime == 0 || aFrame.absTimeDisplay == false) && Math.Abs(Rinc) < 0.001)  // more than 3 digits after comma
      {
        rtmp = Math.Abs(Rinc);
        itmp = (int)-Math.Log10(rtmp);
        if (itmp != sDigits)
        {
          stickFormat = "{0:###.#";  //"{0:0.###}"
          for (int it = 0; it < itmp && it < 10; it++) stickFormat += '#';
          stickFormat += '}';
          sDigits = itmp;
        }
      }
      if (ticks != 0)
        for (int i = 0; i < Nteile; i++)
          for (int j = 0; j < Ntick; j++)
          {
            tmp = Rtick * j; // linear scale
            if (j > 0 && log != 0) tmp = Math.Log(tmp) / Math.Log(10.0); // log
            // scale
            tmp += Rinc * i;
            if (xAxis)
            {
              wx = (int)(wy * tmp / (tick_max - tick_min));
              if (wx < 0) wx = 16777216;  // ovoid overflow
              if (wx > wy) continue;
            }
            else
            {
              wx = (int)(wy * (1 - tmp / (tick_max - tick_min)));
              if (wx < 0) continue;
            }
            if (j == 0) // convert to string
            {
              tmp += tick_min;
              if (log == 0 && labelOffsetMode > 0) tmp -= labelOffset;
              if (xtime != 0 && aFrame.absTimeDisplay)
              {
                dt = JavaLongToCSharpLong((long)(((aFrame.leftTimeLabel ? getMin() : getMax()) + tmp * AcopConst.timeScale[time_scale]) * 1000));
                //                dt = new DateTime((long)(((aFrame.leftTimeLabel ? getMin() : getMax()) + tmp * AcopConst.timeScale[time_scale]) * 1000));
                //               date.setTime((long)(((aFrame.leftTimeLabel ? getMin() : getMax()) + tmp * AcopConst.timeScale[time_scale]) * 1000));
                if (time_scale < AcopConst.tScale_second)
                {
                  /*                  String tstr = "." + dt.ToUniversalTime().Ticks; // date.getTime();
                                    int len = tstr.Length;
                                    if (len > 2)*/
                  if (dt.Millisecond >= 1)  // 0.001
                  {
                    /*                    str = dt.ToLongTimeString();
                                        str = str.Substring(str.Length - 5, 3) + "." + tstr.Substring(4, 3);*/
                    str = dt.ToString("ss.fff");
                    //                    str = dt.ToString().Substring(17, 19) + "." + tstr.Substring(len - 3, len);
                  }
                  else
                  {
                    /*                    str = dt.ToLongTimeString();
                                        str = str.Substring(str.Length - 5, 3);*/
                    str = dt.ToString("ss");
                    //                    str = dt.ToString().Substring(17, 19);
                  }
                  if (getMax() - getMin() < 0.0051)
                  {
                    long uSec = (long)(((aFrame.leftTimeLabel ? getMin() : getMax()) + tmp * AcopConst.timeScale[time_scale]) * 1000000) % 1000;
                    uSec += 1000;
                    String tmpStr = uSec.ToString();
                    str = str + tmpStr.Substring(1);
                  }
                }
                else if (time_scale < AcopConst.tScale_minute)
                {
                  if (Nteile / (tick_max - tick_min) >= 2)
                  {
                    /*                    String tstr = "." + dt.ToUniversalTime().Ticks; // date.getTime();
                                        int len = tstr.Length;
                                        str = dt.ToLongTimeString();
                                        str = str.Substring(str.Length - 8, 5) + "." + tstr.Substring(4, 3);*/
                    //                    str = dt.ToString().Substring(14, 19) + "." + tstr.Substring(len - 3, len);
                    str = dt.ToString("mm:ss.fff");
                  }
                  else
                  {
                    //str = dt.ToLongTimeString();
                    //str = str.Substring(str.Length - 8, 5);
                    str = dt.ToString("HH:mm:ss");
                    //                  str = dt.ToLongTimeString().Substring(3,5);
                    //                  str = dt.ToString().Substring(14, 19);
                  }
                }
                else if (time_scale < AcopConst.tScale_hour)
                {
                  if (Nteile / (tick_max - tick_min) >= 2)
                    str = dt.ToString("HH:mm:ss");//str = dt.ToString().Substring(11, 19);
                  else
                    str = dt.ToString("HH:mm"); //str = dt.ToString().Substring(11, 16);
                }
                else if (time_scale < AcopConst.tScale_day)
                {
                  if ((tick_max - tick_min) > 24)
                    str = dt.ToString("MMM dd") + "/" + dt.ToString("HH") + "h";//str = dt.ToString().Substring(4, 10) + "/" + dt.ToString().Substring(11, 13) + "h";
                  else
                    str = dt.ToString("HH:mm"); //str = dt.ToString().Substring(11, 16);
                }
                else if (time_scale < AcopConst.tScale_month)
                {
                  if (Nteile / (tick_max - tick_min) >= 2)
                    str = dt.ToString("MMM dd") + "/" + dt.ToString("HH") + "h";//str = dt.ToString().Substring(4, 10) + "/" + dt.ToString().Substring(11, 13) + "h";
                  else
                    str = dt.ToString("MMM dd"); //str = dt.ToString().Substring(4, 10);
                }
                else if (time_scale < AcopConst.tScale_year)
                {
                  if ((tick_max - tick_min) > 12)
                  {
                    //str = "-" + (dt.Year + 1900);
                    //str = dt.ToString().Substring(4, 11) + str.Substring(3, 5) + "'";
                    str = dt.ToString("MMM dd yy") + "'";
                  }
                  else
                    str = dt.ToString("MMM dd"); //str = dt.ToString().Substring(4, 10);
                }
                else
                {
                  str = dt.ToString("MMM yy") + "'";
                  //str = "-" + (dt.Year + 1900);
                  //str = dt.ToString().Substring(4, 8) + str.Substring(3, 5) + "'";
                }
              }
              else
              {
                if (log != 0)
                  tmp = Math.Pow(10.0, tmp);
                else if (Math.Abs(tmp) < (Rlast - Rfirst) * 1.0e-7) tmp = 0.0; // avoid
                // 0.1e-9
                //str = tickFormat.format(tmp);
                // str = tmp.ToString("####0.000", nfi);
                str = String.Format("{0:0.###}", tmp);
                g_stellen(tmp, strformat);
                //            str = strformat[4] == 0 ? tickFormat.format(tmp) : tickFormat.format(tmp
                //                / Math.Pow(10.0, (double) strformat[4]))
                //                + "e" + strformat[4];
                if (strformat[4] == 0) str = String.Format("{0:0.###}", tmp);
                else str = String.Format("{0:0.###}", tmp / Math.Pow(10.0, (double)strformat[4])) + "e" + strformat[4];
                // convert (str, tmp, strformat) ;
                // texp = g_stellen (tmp, &vor, &hinter);
                // lang = convert (str, tmp, hinter, &texp ) ;                      
                if (strformat[4] == 0 && Math.Abs(Rinc) < 0.001) str = String.Format(stickFormat, tmp);
              }
            }
            if (xAxis) // output tick and text for x
            {
              if (aFrame.drawFlag == false || pHisto.xlabON == 0 || log != 0 || xtime != 0)
              {
//                wz = leftBottom ? 0 : ticksRect.Height - 1;
                wz = leftBottom ? 0 : ticksRect.Height + 1 - (j == 0 ? longTick : shortTick);
                a.DrawLine(acop.stroke, wx, wz, wx, wz + (j == 0 ? longTick : shortTick));
                if (ticktext && j == 0)
                {
                  textStringSize = a.MeasureString(str, aFrame.frameFont);
                  //siz = fMetrix.stringWidth(str);
                  siz = (int)textStringSize.Width;
                  if ((txtRect.X = wx - siz / 2) >= lastright)
                  {
                    //                  wz = leftBottom ? (4 + hei) : (txtRect.Height);
//                   wz = leftBottom ? 2 : (txtRect.Height - hei - 2 - 8);
                    wz = leftBottom ? (hei - 2) : (txtRect.Height - hei * 2 + 2);
                    txtRect.Width = siz;
                    //                  if (tickLabelForeColor != aFrame.frameForeColor) a.setColor(tickLabelForeColor);
                    a.DrawString(str, aFrame.frameFont, tickLabelBrush, txtRect.X, wz);
                    lastright = wx + siz;
                    //                  if (tickLabelForeColor != aFrame.frameForeColor) a.setColor(aFrame.frameForeColor);
                  }
                }
              }
              // else if ( j == 0 )
              else
              {
                for (k = 0, m = pHisto.wrapindex; k < pHisto.npoints; k++, m++)
                {
                  if (m >= pHisto.npoints) m = 0;
                  xpos = pHisto.ix[m] - aFrame.xleft;
                  if (wx <= xpos) break;
                }
                if (xpos <= aFrame.xright) // within histogram
                {
                  if (m == pHisto.npoints) m--;
//                  wz = leftBottom ? 0 : ticksRect.Height - 1;
                  wz = leftBottom ? 0 : ticksRect.Height - (j == 0 ? longTick : shortTick);
                  a.DrawLine(acop.stroke, xpos, wz, xpos, wz + (j == 0 ? longTick : shortTick));
                  //                if (tickLabelForeColor != aFrame.frameForeColor) a.setColor(tickLabelForeColor);
                  if (ticktext && pHisto.xlabON != 0 && m < pHisto.xlabLength)
                  {
                    if (pHisto.xlab[m] == null) pHisto.xlab[m] = "#" + m;
                    textStringSize = a.MeasureString(pHisto.xlab[m], aFrame.frameFont);
                    //                  siz = fMetrix.stringWidth(pHisto.xlab[m]);
                    siz = (int)textStringSize.Width;   //fMetrix.stringWidth(pHisto.xlab[m]);
                    if (acop.xlabelRotation == 0)
                    {
                      if (j == 0 && (txtRect.X = xpos - siz / 2) >= lastright)
                      {
                        wz = leftBottom ? (4 + hei) : (txtRect.Height);
                        txtRect.Width = siz;
                        a.DrawString(pHisto.xlab[m], aFrame.frameFont, tickLabelBrush, txtRect.X, wz);
                        lastright = xpos + siz;
                      }
                    }
                    else
                    {
                      //                    AffineTransform oldTransform = a.getTransform();
                      xtp = xpos;
                      ytp = 4;
                      if (acop.xlabelAxis == 1) // 90 dwegree
                      {
                        xtp += hei / 2;
                        ytp += siz;
                      }
                      else if (acop.xlabelAxis == 2) // 180 dwegree
                      {
                        xtp += siz / 2;
                        ytp += hei / 2;
                      }
                      else if (acop.xlabelAxis == 3) // -90 dwegree
                      {
                        // ytp += 4 ;
                      }
                      else if (acop.xlabelQudrant == 4)
                      {
                        xtp += (int)(hei / 2 * acop.xlabelsin);
                        ytp += (int)(hei * acop.xlabelcos);
                      }
                      else if (acop.xlabelQudrant == 2)
                      {
                        xtp -= (int)(siz * acop.xlabelcos);
                        ytp += (int)(siz * acop.xlabelsin);
                      }
                      else if (acop.xlabelQudrant == 1)
                      {
                        xtp -= (int)(siz * acop.xlabelcos - hei / 2 * acop.xlabelsin);
                        ytp += (int)(siz * acop.xlabelsin + hei * acop.xlabelcos);
                        // xtp -= (int)(hei * de.desy.acop.chart.xlabelsin - siz *
                        // de.desy.acop.chart.xlabelcos) ;
                        // ytp += (int)(hei * de.desy.acop.chart.xlabelcos + siz *
                        // de.desy.acop.chart.xlabelsin) ;
                      }
                      a.TranslateTransform(xtp, ytp);
                      a.RotateTransform((float)-acop.xlabelRotateDegree);
                      a.DrawString(pHisto.xlab[m], aFrame.frameFont, tickLabelBrush, 0, 0);
                      a.RotateTransform((float)acop.xlabelRotateDegree);
                      a.TranslateTransform(-xtp, -ytp);
                      //                    a.setTransform(oldTransform);
                    }
                  }
                  //                if (tickLabelForeColor != aFrame.frameForeColor) a.setColor(aFrame.frameForeColor);
                }
              }
            }
            else if (ticksRect.Width > 3) // y axis
            {
              wz = leftBottom ? ticksRect.Width + 1 : 2;
              //              a.DrawLine(acop.stroke, wz, wx, wz - (j == 0 ? longTick : shortTick), wx);
              itmp = (int)tmp;
              if (acop.ylabON == 0 || secondYAxis == true || (j == 0 && Math.Abs(tmp - itmp) < 0.001))
                a.DrawLine(acop.stroke, wz, wx, wz - (j == 0 ? longTick : shortTick), wx);
              if (ticktext && j == 0 && ticksRect.Width > 4)
              {
                textStringSize = a.MeasureString(str, aFrame.frameFont);
                siz = (int)textStringSize.Width;   //fMetrix.stringWidth(str);
                if (siz > strmaxlen) strmaxlen = siz;
                if ((txtRect.Y = wx + hei) <= lastright)
                {
                  wz = leftBottom ? ticksRect.Width - 2 - siz : 5;
                  wz -= 2;
                  txtRect.Height = hei;
                  if (acop.ylabON == 0 || secondYAxis == true)
                    a.DrawString(str, aFrame.frameFont, tickLabelBrush, wz, txtRect.Y - hei * 5 / 3);
                  else
                  {
                    ytp = (itmp - acop.yLabelStart) % acop.yLabelInterval;
                    xtp = (itmp - acop.yLabelStart) / acop.yLabelInterval;
                    if (j == 0 && xtp >= 0 && xtp < acop.ylabLength && acop.ylab[xtp] != null && ytp == 0 && Math.Abs(tmp - itmp) < 0.001)
                      a.DrawString(acop.ylab[xtp], aFrame.frameFont, tickLabelBrush, wz - 25 - acop.yLabelDisplayOffset, txtRect.Y);
                  }
                  //                if (tickLabelForeColor != aFrame.frameForeColor) a.setColor(tickLabelForeColor);
                  //                  a.DrawString(str, aFrame.frameFont, tickLabelBrush, wz, txtRect.Y - hei * 5 / 3);
                  lastright = txtRect.Y - hei * 5 / 3;
                  //                if (tickLabelForeColor != aFrame.frameForeColor) a.setColor(aFrame.frameForeColor);
                }
              }
            }
          }
      //    if (tickLabelForeColor != aFrame.frameForeColor) a.setColor(tickLabelForeColor);
      if (xAxis) // x label and time label
      {
        wz = leftBottom ? (5 + hei * 2) : (ticksRect.Height - hei * 4);
        // if ( AxisLabel.length() != 0 && leftBottom == leftBottLabel )
        // a.drawString(AxisLabel, (aFrame.dxwidth -
        // fMetrix.stringWidth(AxisLabel))/2, wz );
        if (ticktext && xtime == 0 && ticks != 0)
        {
          siz = (int)a.MeasureString(acop.XLeftLabel, aFrame.frameFont).Width;
          tlen = (int)a.MeasureString(acop.XRightLabel, aFrame.frameFont).Width;
          //          siz = fMetrix.stringWidth(acop.XLeftLabel);
          //          tlen = fMetrix.stringWidth(acop.XRightLabel);
          if (siz > 0)
          {
            if (aFrame.absxleft < siz / 3)
            {
              a.DrawString(acop.XLeftLabel, aFrame.frameFont, tickLabelBrush, -aFrame.absxleft + 5, wz);
            }
            else
            {
              a.DrawString(acop.XLeftLabel, aFrame.frameFont, tickLabelBrush, -siz / 3, wz);
            }
          }
          if (tlen > 0)
          {
            if (acop.histRoundRect.X + acop.histRoundRect.Width + tlen / 4 > acop.Width)
            {
              a.DrawString(acop.XRightLabel, aFrame.frameFont, tickLabelBrush, acop.Width - acop.histRoundRect.X - tlen - 5, wz);
            }
            else
            {
              a.DrawString(acop.XRightLabel, aFrame.frameFont, tickLabelBrush, aFrame.dxwidth - tlen * 3 / 4, wz); // 2009-12 it was 1/2
            }
          }
        }
        if (ticktext && xtime != 0 && ticks != 0)
        {
          if (aFrame.leftTimeLabel)
            dt = JavaLongToCSharpLong((long)(getMin() * 1000));
          //            dt = new DateTime((long)getMin() * 1000);
          //          date.setTime((long) getMin() * 1000);
          else
            dt = JavaLongToCSharpLong((long)(getMax() * 1000));
          //          dt = new DateTime((long)getMax() * 1000);
          //          date.setTime((long) getMax() * 1000);
          String dateString = dt.ToString(dtFormat);
          textStringSize = a.MeasureString(dateString, aFrame.frameFont);
          siz = (int)textStringSize.Width; //fMetrix.stringWidth(date.toString());
          if (getMax() - getMin() < 2)
          {
            if (dt.Millisecond >= 1)  // 0.001
            //  if (tstr.length() > 2)
            {
              dateString = dt.ToString("ddd MMM dd HH:mm:ss.fff yyyy");
            }
            else
            {
              dateString = dt.ToString("ddd MMM dd HH:mm:ss yyyy");
            }
            if (getMax() - getMin() < 0.0051)
            {
              long uSec = (long)(((aFrame.leftTimeLabel ? getMin() : getMax())) * 1000000) % 1000;
              uSec += 1000;
              String tmpStr = uSec.ToString();
              dateString = dt.ToString("ddd MMM dd HH:mm.fff") + tmpStr.Substring(1) + " " + dt.ToString("yyyy");
            }

          }
          if (aFrame.absTimeDisplay)
          {
            str = (long)((getMax() - getMin()) / AcopConst.timeScale[time_scale] + 0.45) + " " + AcopConst.timeStr[time_scale];
            textStringSize = a.MeasureString(str, aFrame.frameFont);
            tlen = (int)textStringSize.Width;  //fMetrix.stringWidth(str);
          }
          else
          {
            textStringSize = a.MeasureString(AcopConst.timeStr[time_scale], aFrame.frameFont);
            tlen = (int)textStringSize.Width;  //fMetrix.stringWidth(AcopConst.timeStr[time_scale]);
          }
          if (aFrame.leftTimeLabel)
          {
            // a.drawString(date.toString(), -siz/4, 5+hei*2);
            a.DrawString(dateString, aFrame.frameFont, tickLabelBrush, -siz / 4, wz);
            if (aFrame.absTimeDisplay)
              a.DrawString(str, aFrame.frameFont, tickLabelBrush, aFrame.dxwidth - tlen * 8 / 15, wz);  // 2009-12 it was 1/2
            else
              a.DrawString(AcopConst.timeStr[time_scale], aFrame.frameFont, tickLabelBrush, aFrame.dxwidth - tlen * 8 / 15, wz); // 2009-12 it was 1/2
          }
          else
          {
            a.DrawString(dateString, aFrame.frameFont, tickLabelBrush, aFrame.dxwidth - siz * 4 / 5, wz);  // 2009-9 it was 3/4
            if (aFrame.absTimeDisplay)
              a.DrawString(str, aFrame.frameFont, tickLabelBrush, -tlen / 2, wz);
            else
              a.DrawString(AcopConst.timeStr[time_scale], aFrame.frameFont, tickLabelBrush, -tlen / 2, wz);
          }
        }
      }
      // else if ( AxisLabel.length() != 0 && leftBottom == leftBottLabel ) // y
      // label
      else if (false) // y label
      {
        //      AffineTransform oldTransform = a.getTransform();
        if (leftBottLabel == true)
        {
          if (ticksRect.Width >= hei + 4)
          {
            // a.translate(hei+5,
            // ticksRect.height/2+fMetrix.stringWidth(AxisLabel)/2);
            a.TranslateTransform(hei + 5, ticksRect.Height / 2);
            a.RotateTransform((float)-90);
            a.DrawString(AxisLabel, aFrame.frameFont, tickLabelBrush, 0, 0);
            a.TranslateTransform(-hei - 5, -ticksRect.Height / 2);
            a.RotateTransform((float)90);
          }
        }
        else if (acop.yAxisRect.Width != 0)
        {
          a.TranslateTransform(acop.yAxisRect.Width - hei, acop.yAxisRect.Height / 2);
          // a.translate(de.desy.acop.chart.yAxisRect.x+de.desy.acop.chart.yAxisRect.width/2,
          // de.desy.acop.chart.yAxisRect.height/2);
          a.RotateTransform((float)-90);
          a.DrawString(AxisLabel, aFrame.frameFont, tickLabelBrush, 0, 0);
          a.RotateTransform((float)90);
          a.TranslateTransform(-acop.yAxisRect.Width + hei, -acop.yAxisRect.Height / 2);
        }
        //      a.transform(oldTransform);
        /*
         * pdc->SelectObject(&y_font); pdc->SetTextAlign( TA_CENTER | TA_TOP); if (
         * !m_leftYLabel ) { if ( fRight > 0 ) { pdc->SetWindowExt (fRight, -dy);
         * pdc->SetViewportExt (yAxisRect.right - yAxisRect.left, dy);
         * pdc->SetViewportOrg (yAxisRect.left, histRect.bottom-1); //
         * pdc->GetTextMetrics(&tm); // pdc->ExtTextOut((100-tm.tmHeight)/2,
         * 0,ETO_CLIPPED, &txtRect, m_yAxisLabel,NULL); pdc->TextOut(2, -dy/2,
         * m_yAxisLabel); } } else if ( fLeft >= strmaxlen + 4 + tm.tmHeight) { //
         * pdc->GetTextMetrics(&tm); // pdc->ExtTextOut((100-tm.tmHeight)/2,
         * 0,ETO_CLIPPED, &txtRect, m_yAxisLabel,NULL); strmaxlen =
         * fLeft-strmaxlen-4-tm.tmHeight-2; pdc->TextOut( strmaxlen < 0 ? 0 :
         * strmaxlen, -dy/2, m_yAxisLabel); }
         */
      }
      if (xAxis)
        a.TranslateTransform(-aFrame.histRect.X - aFrame.xleft, -ticksRect.Y + 1);
      else
        a.TranslateTransform(-ticksRect.X, -aFrame.histRect.Y);
      a.SetClip(Cliprect);
      drawLabel(a);
      //    if (tickLabelForeColor != aFrame.frameForeColor) a.setColor(aFrame.frameForeColor);
    }
  }
}
