﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace Acop.AcopChartUtil
{
  public partial class AcopText
  {
#   pragma warning disable 1591
    protected internal Boolean textON,			// on or off
            outside,
            scalable;					// go with coordinates
    protected int xcenter,					// x input: left(0), center(1) or right(2)
            ycenter,					// y input: bottom(0), center(1) or top(2)
      moveable,
      rotationDegree,   // 0 - 365degree
            prePrint,					// textprint prio ondraw!
            width,
            height,
            length;						// char width and string length
    protected internal double text_x, text_y,		// user units??
              conv_x, conv_y,             // starting print position
        lastTextx, lastTexty;
    protected double xsize, ysize;		// user display rectangle
    protected double scaleFactor;
    protected int saveHeight, saveWidth, lfHeight, lfWidth, lfAscent, lfDescent;
    protected SizeF textStringSize;
    protected int dev_x, dev_y;		// device units??
    protected int rect_x, rect_y;
    protected Color textColor;
    protected SolidBrush textBrush;
    protected String[] textstr;
    protected Font textFont, inputFont;
    protected internal Rectangle textRect;
    protected int rescaleFlag;

    protected double rotateGrad;
    protected int offsetX, offsetY;   // device units??
    protected double sin, cos;
    protected double xOffset, yOffset;      // for text area computation
    protected double heightSin, heightCos, lengthSin, lengthCos;
    protected Pen textLineStroke;

    private AcopChart acop;

    public AcopText(AcopChart acop)
    {
      this.acop = acop;

      textON = false;
      text_x = 0;
      text_y = 0;
      length = 0;
      prePrint = 0;
      outside = true;
      scalable = false;
      moveable = 0;
      xcenter = 0;
      ycenter = 0;
      rescaleFlag = 1;

      textRect = new Rectangle();
      textLineStroke = new Pen(Color.DarkGray, 1.0f);
      //{{INIT_CONTROLS
      //}}
    }
    //{{DECLARE_CONTROLS
    protected internal String getPrintedText(double[] leftBottom, double[] size, double[] position)
    {
      if (position != null)
      {
        position[0] = text_x;
        if (position.Length > 1) position[1] = text_y;
      }
      if (size != null)
      {
        size[0] = textRect.Width * acop.sca[0].dispSize / acop.aFrame.dxwidth;
        if (size.Length > 1) size[1] = textRect.Height * acop.sca[1].dispSize / acop.aFrame.histRect.Height;
      }
      if (leftBottom != null)
      {
        leftBottom[0] = (textRect.X - acop.aFrame.histRect.X) * acop.sca[0].dispSize / acop.aFrame.dxwidth + acop.sca[0].min;
        if (leftBottom.Length > 1) leftBottom[1] = (textRect.Height - textRect.Y + acop.aFrame.histRect.Y) * acop.sca[1].dispSize / acop.aFrame.histRect.Height + acop.sca[1].min;
      }
      String sb = "";
      for (int i = 0; i < textstr.Length; i++) sb += textstr[i] + "\n";
      return sb;
    }

    protected internal void printUserText(Graphics a)
    {
      Boolean nullFlag;
      //by jbobnar: Do not print the text if acop is not visible
      if (!acop.Visible) return; // isShowing()
      if (acop.textFlag == false || textON == false || outside == true) return;
      if (rescaleFlag != 0) rescaleText(rescaleFlag);
      if (a == null)
      {
        a = (Graphics)acop.CreateGraphics();
        //                a.clipRect(acop.histRoundRect.X, acop.histRoundRect.Y, acop.histRoundRect.Width, acop.histRoundRect.Height);
        a.SetClip(acop.histRoundRect);
        a.TranslateTransform(acop.aFrame.histRect.X, acop.aFrame.histRect.Y);
        nullFlag = true;
      }
      else nullFlag = false;

      //    a.clipRect(textRect.x, textRect.y, textRect.width, textRect.height);

      //            a.setFont(textFont);
      //            a.setColor(textColor);
      if (rotationDegree != 0)
      {
        int yText = offsetY - lfDescent;
        //        int yText = offsetY;
        //                AffineTransform oldTransform = a.getTransform();
        a.TranslateTransform(dev_x + offsetX, dev_y + yText);
        a.RotateTransform((float)-rotationDegree); //.rotate(-rotateGrad);
        int textHeight = textFont.FontFamily.GetEmHeight(textFont.Style) + 1;
        for (int i = 0; i < textstr.Length; i++)
        {
          a.DrawString(textstr[i], textFont, textBrush, 0, i * textHeight);
        }
        //                a.DrawString(textstr, textFont, textBrush, offsetX, offsetY - lfDescent);
        a.RotateTransform((float)rotationDegree); //.rotate(-rotateGrad);
        a.TranslateTransform(-dev_x - offsetX, -dev_y - yText);
        //                a.setTransform(oldTransform);
      }
      else
      {
        int offset = 0;
        if (acop.printTextLine)
        {
          a.DrawLine(textLineStroke, dev_x + offsetX, 0, dev_x + offsetX, acop.histRoundRect.Height);
          offset = 2;
        }
        int yText = dev_y + offsetY - lfDescent;
        //        int yText = dev_y + offsetY;
        int textHeight = textFont.FontFamily.GetEmHeight(textFont.Style) + 1;
        for (int i = 0; i < textstr.Length; i++)
        {
          a.DrawString(textstr[i], textFont, textBrush, dev_x + offsetX + offset, yText + i * textHeight);
        }
        //              a.DrawString(textstr[0], textFont, textBrush, dev_x + offsetX, dev_y + offsetY - lfDescent);
      }

      if (nullFlag)
      {
        a.TranslateTransform(-acop.aFrame.histRect.X, -acop.aFrame.histRect.Y);
        a.Dispose();
      }
    }

    protected internal int printText(String Text, double Xposition, double Yposition, Color c, Font font)
    {
      int resca, lineFit;
      int i, move, localxcenter, localycenter, localRotation;
      Boolean localscalable;

      lineFit = Text.IndexOf((char)13);

      move = acop.printTextStickyAttribute & AcopConst.TEXT_MOVE;
      if (acop.printTextStickyAttribute == 4) move = AcopConst.TEXT_MOVE;
      localxcenter = acop.printTextLocation % 3;
      if (acop.printTextLocation < 3) localycenter = 0;
      else if (acop.printTextLocation < 6) localycenter = 1;
      else localycenter = 2;
      localRotation = acop.printTextRotation;

      localscalable = move == AcopConst.TEXT_MOVE ? true : false;
      if (acop.printTextStickyAttribute == 4) localscalable = false;

      moveable = 0;
      scalable = false;

      if (textON == true)		// get rid of old text
      {
        textON = false;
        if (outside == false) acop.Invalidate(textRect);// acop.repaint(textRect.X, textRect.Y, textRect.Width, textRect.Height);
      }

      // check if location or ration have been changed?
      resca = (localxcenter != xcenter || localycenter != ycenter) ? 1 : 0;
      if (Math.Abs(Xposition - text_x) > AcopConst.zero_check) resca++;
      text_x = Xposition;
      if (Math.Abs(Yposition - text_y) > AcopConst.zero_check) resca++;
      text_y = Yposition;
      if (localRotation != rotationDegree) resca++;

      String[] newText = Text.Split('\n');
      if (lineFit > 0)
      {		// if text + linefeed, then remove linefeed
        Text = Text.Substring(0, lineFit - 1);
        lineFit = 1;
      }
      else lineFit = 0;

      rotationDegree = localRotation;
      rotateGrad = rotationDegree * Math.PI / 180;
      if (rotationDegree == 0)
      {
        cos = 1;
        sin = 0;
      }
      else
      {
        cos = Math.Cos(rotateGrad);
        sin = Math.Sin(rotateGrad);
      }

      textColor = c;
      textBrush = new SolidBrush(c);
      textFont = font;
      inputFont = font;
      textstr = newText;
      xcenter = localxcenter;
      ycenter = localycenter;
      if (xcenter == 0) xOffset = 0;  // set rotation calculation parameters
      else if (xcenter == 1) xOffset = 0.5;
      else xOffset = 1;
      if (ycenter == 0) yOffset = 1;
      else if (ycenter == 1) yOffset = 0.5;
      else yOffset = 0;

      textON = true;
      if (resca != 0 || rescaleFlag != 0)
      {
        length = Text.Length;
        rescaleText(AcopConst.NEW_PRINT_WINDOW);
      }
      else if (length != Text.Length)
      {
        length = Text.Length;
        setTextAttribute(true, conv_x, 0);
      }
      lastTextx = text_x;       //store last text coordinates
      if (lineFit != 0)
      {
        lastTexty = acop.sca[1].dispSize * lfHeight / acop.aFrame.histRect.Height;
        if (acop.sca[1].log != 0) lastTexty = Math.Pow(10.0, lastTexty);
        lastTexty = text_y - lastTexty;
      }
      else lastTexty = text_y;

      acop.textFlag = true;
      moveable = move;
      scalable = localscalable;

      if (scalable)	// store origin display area
      {
        xsize = acop.sca[0].dispSize;
        ysize = acop.sca[1].dispSize;
        scaleFactor = 1;
        saveHeight = lfHeight;
        saveWidth = lfWidth;
      }

      if (acop.drawTicks)
      {
        acop.Invalidate();// .repaint();
        acop.drawTicks = false;
      }
      else printUserText(null);

      //      if ( de.desy.acop.chart.disp[0] == null && scalecall < 2 )
      //      {
      //    	InvalidateControl(NULL);		// draw ticks and grid...if no histo at all
      //    	scalecall = 2;
      //      }
      return (0);
    }

    protected internal void refreshText(String Text)
    {
      int scaled = 0;

      String[] newText = Text != null ? Text.Split('\n') : null;
      if (textON)    // get rid of old text
      {
        if (Text.CompareTo(textstr) == 0 && rescaleFlag == 0) return;  // already on
        if (outside == false)
        {
          textON = false;
          acop.Invalidate(textRect);  // repaint(textRect.X, textRect.Y, textRect.Width, textRect.Height);
        }
      }
      else
      {
        if (rescaleFlag != 0 && (moveable != 0 || scalable))
        {
          textON = true;
          length = Text.Length;
          rescaleText(0);
          scaled = 1;
        }
        acop.textFlag = true;
      }
      textON = true;
      textstr = newText;
      if (scaled == 0 && length != Text.Length)
      {
        length = Text.Length;
        setTextAttribute(true, conv_x, 0);
      }
      printUserText(null);
    }

    protected void setTextRotationParameters()
    {
      Graphics a = (Graphics)acop.CreateGraphics();
      if (a == null) return; // currently not displayable!! 2010-9-29
      /*
                  Font tempFont = a.getFont();
                  a.setFont(textFont);
                  FontMetrics fMetrix = a.getFontMetrics();
                  a.setFont(tempFont);
                  a.Dispose();

                  lfWidth = fMetrix.stringWidth(textstr);
                  lfHeight = fMetrix.getHeight(); // fMetrix.getAscent() + fMetrix.getDescent();
                  lfAscent = fMetrix.getAscent();
                  lfDescent = fMetrix.getDescent();
      */
      SizeF strSize;
      float len, hei;
      float length = 0;
      float height = 0;
      for (int i = 0; i < textstr.Length; i++)
      {
        strSize = a.MeasureString(textstr[i], textFont);
        len = strSize.Width;    // fMetrix.stringWidth(caption);
        hei = strSize.Height;  // (fMetrix.getDescent() + fMetrix.getAscent()) / 2;
        length = Math.Max(len, length);
        height = Math.Max(hei, height);
      }
      lfWidth = (int)length;
      if (acop.printTextLine)
      {
        lfWidth += 2;
      }
      lfAscent = textFont.FontFamily.GetCellAscent(textFont.Style);
      lfDescent = textFont.FontFamily.GetCellDescent(textFont.Style);
      lfHeight = textFont.FontFamily.GetEmHeight(textFont.Style) + lfAscent + lfDescent;

      lfAscent = (int)textFont.Size * lfAscent / textFont.FontFamily.GetEmHeight(FontStyle.Regular);
      lfDescent = (int)textFont.Size * lfDescent / textFont.FontFamily.GetEmHeight(FontStyle.Regular);
      lfHeight = (int)textFont.Size * lfHeight / textFont.FontFamily.GetEmHeight(FontStyle.Regular);

      lfHeight = (int)height;
      lfHeight *= textstr.Length;
      //     textStringSize = a.MeasureString(textstr, textFont);
      //     lfWidth = (int)textStringSize.Width;
      //     lfHeight = (int)textStringSize.Width;

      heightSin = lfHeight * sin;
      heightCos = lfHeight * cos;
      lengthSin = lfWidth * sin;
      lengthCos = lfWidth * cos;

      if (xcenter == 0) offsetX = 0;
      else if (xcenter == 1) offsetX = -lfWidth / 2;
      else offsetX = -lfWidth;

      if (ycenter == 0) offsetY = 0;
      else if (ycenter == 1) offsetY = (int)textFont.Size / 2;//lfHeight / 2;
      else offsetY = (int)textFont.Size;//lfHeight;
    }

    protected void setTextAttribute(Boolean newText, double xInput, int rescale)
    {
      double xtemp = xInput;     // can be conv_x or input after wrapText()
      double ytemp = conv_y;
      //      int x, y;
      int tmp;
      int topLeft_cx, topLeft_cy, bottomLeft_cx, bottomLeft_cy;
      int topRight_cx, topRight_cy, bottomRight_cx, bottomRight_cy;
      int mov_x, mov_y;

      if (rescale != 0)
      {
        mov_x = 1;
        mov_y = 1;
      }
      else
      {
        mov_x = moveable & 0x1;
        mov_y = moveable & 0x2;
      }

      if (newText) setTextRotationParameters();

      // rescale to device units
      if (mov_x != 0)
        dev_x = (int)((xtemp - acop.sca[0].min) / acop.sca[0].dispSize * acop.aFrame.dxwidth + acop.aFrame.xleft + 0.01);
      if (mov_y != 0)
        dev_y = (int)((acop.aFrame.histRect.Height - ((ytemp - acop.sca[1].min) / acop.sca[1].dispSize * acop.aFrame.histRect.Height)) - textFont.Size);

      if (mov_y != 0)
        rect_y = (int)(dev_y - yOffset * heightCos + xOffset * lengthSin) + acop.aFrame.histRect.Y + lfHeight; // lfDescent;
      if (mov_x != 0)
        rect_x = (int)(dev_x - yOffset * heightSin - xOffset * lengthCos) + acop.aFrame.histRect.X;

      topLeft_cx = rect_x;
      topLeft_cy = rect_y;

      bottomLeft_cx = (int)(rect_x + heightSin);
      bottomLeft_cy = (int)(rect_y + heightCos);

      topRight_cx = (int)(rect_x + lengthCos);
      topRight_cy = (int)(rect_y - lengthSin);

      bottomRight_cx = (int)(rect_x + heightSin + lengthCos);
      bottomRight_cy = (int)(rect_y + heightCos - lengthSin);

      tmp = rect_x;
      if (tmp > bottomLeft_cx) tmp = bottomLeft_cx;
      if (tmp > topRight_cx) tmp = topRight_cx;
      if (tmp > bottomRight_cx) tmp = bottomRight_cx;
      textRect.X = tmp;

      tmp = rect_x;
      if (tmp < bottomLeft_cx) tmp = bottomLeft_cx;
      if (tmp < topRight_cx) tmp = topRight_cx;
      if (tmp < bottomRight_cx) tmp = bottomRight_cx;
      textRect.Width = tmp - textRect.X;

      tmp = rect_y;
      if (tmp > bottomLeft_cy) tmp = bottomLeft_cy;
      if (tmp > topRight_cy) tmp = topRight_cy;
      if (tmp > bottomRight_cy) tmp = bottomRight_cy;
      textRect.Y = tmp;

      tmp = rect_y;
      if (tmp < bottomLeft_cy) tmp = bottomLeft_cy;
      if (tmp < topRight_cy) tmp = topRight_cy;
      if (tmp < bottomRight_cy) tmp = bottomRight_cy;
      textRect.Height = tmp - textRect.Y;

      outside = textRect.IntersectsWith(acop.histRoundRect) ? false : true;
    }

    protected void scaleFont()
    {
      double x, y;

      // relative to original size
      x = acop.sca[0].dispSize / xsize;
      y = acop.sca[1].dispSize / ysize;
      if (Math.Abs(scaleFactor - y) <= AcopConst.zero_check ||
        Math.Abs(scaleFactor - x) <= AcopConst.zero_check) return;
      if (x > 1 && y > 1)
      {
        if (x < y) y = x;
      }
      else if (x > y) y = x;

      scaleFactor = y;    // store new factor for next comparison
      //           textFont = new Font(inputFont.getName(), inputFont.getStyle(), (int)(inputFont.getSize() / y));
      textFont = new Font(inputFont.Name, (int)(inputFont.Size / y), inputFont.Style);
      lfHeight = (int)(saveHeight / y);
      lfWidth = (int)(saveWidth / x);
    }

    protected void adjustXY(int flag)
    {
      int xflag = flag & 0x1;
      int yflag = flag & 0x2;

      if (xflag != 0)
        conv_x = acop.sca[0].log != 0 ? Math.Log(AcopConst.log_check(text_x)) / Math.Log(10) : text_x;
      if (yflag != 0)
        conv_y = acop.sca[1].log != 0 ? Math.Log(AcopConst.log_check(text_y)) / Math.Log(10) : text_y;

      if (xflag != 0 || yflag != 0) setTextRotationParameters();
    }

    protected double wrapText()
    {
      int i;
      double x = conv_x;
      AcopHisto h = acop.histo[acop.m_histstart];

      i = h.wraphigh < 0 ? -(h.wraphigh + 1) : h.wraphigh;
      if (h.wraphigh < 0)
      {
        if (x >= h.x[i])
        {
          if (h.xaryFlag == 1) x += (acop.wrapxmax - acop.wrapxmin);
          else x -= (acop.wrapxmax - acop.wrapxmin);
        }
      }
      else if (x < h.x[i])
      {
        if (h.xaryFlag == 1) x -= (acop.wrapxmax - acop.wrapxmin);
        else x += (acop.wrapxmax - acop.wrapxmin);
      }
      return x;
    }

    protected internal void rescaleText(int rescaleTextFlag)
    {
      double x;

      // resize, framewidth and printtext(3); Zoom (2); move(1)
      if (textON == false)
      {
        rescaleFlag = rescaleTextFlag;  // store flag
        return;
      }

      // load flag
      if (rescaleTextFlag == 0)
      {
        if (rescaleFlag == 0) return;
        rescaleTextFlag = rescaleFlag;
        return;
      }

      // new font
      if (rescaleTextFlag == AcopConst.NEW_ZOOM_WINDOW && scalable) scaleFont();
      // new device window
      if (rescaleTextFlag == AcopConst.NEW_PRINT_WINDOW) adjustXY(AcopConst.TEXT_MOVE);
      else if (rescaleTextFlag != AcopConst.NEW_MOVE_WINDOW && moveable != 0) adjustXY(moveable);
      // wrap
      if ((rescaleTextFlag == AcopConst.NEW_PRINT_WINDOW || (moveable & 0x1) != 0) &&
                  acop.wrapAround && acop.histo[acop.m_histstart] != null && acop.sca[0].log == 0 && acop.histo[acop.m_histstart].wraphigh != 0) x = wrapText();
      else x = conv_x;
      // map
      if (rescaleTextFlag == AcopConst.NEW_PRINT_WINDOW || (moveable & AcopConst.TEXT_MOVE) != 0) setTextAttribute(false, x, rescaleTextFlag == AcopConst.NEW_PRINT_WINDOW ? 1 : 0);

      rescaleFlag = 0;     // reset flag
    }
    //}}
  }
}
