﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace Acop
{
  [ToolboxItem(true)]
  [ToolboxBitmap(typeof(CheckBox))]
  public partial class AcopCheckBox : AcopUserComponent
  {
#   pragma warning disable 1591
    private CheckBox checkBox;

    private String cbText = "Acop CheckBox";
    private double selectedVal = 1;
    private double deSelectedVal = 0;
    private Int32 bitValue = 0;
    private int selectedBit = -1;
    private Color tmpColor;

    private int lastLinkStatus = 0;
    private Boolean hasLnkError = false;

    public AcopCheckBox()
    {
      InitializeComponent();
      SetDefaultUpdater(defaultUpdate);
      setControlsList(new Control[] { checkBox }, true);
      SetLateInitializerProlog(initTransportInfo);
      StartLateInitialization();
    }

    private void InitializeComponent()
    {
      this.checkBox = new System.Windows.Forms.CheckBox();
      this.SuspendLayout();
      // 
      // checkBox
      // 
      this.checkBox.Anchor = ((System.Windows.Forms.AnchorStyles)((((System.Windows.Forms.AnchorStyles.Top | System.Windows.Forms.AnchorStyles.Bottom) 
            | System.Windows.Forms.AnchorStyles.Left) 
            | System.Windows.Forms.AnchorStyles.Right)));
      this.checkBox.AutoSize = true;
      this.checkBox.Location = new System.Drawing.Point(3, 3);
      this.checkBox.Name = "checkBox";
      this.checkBox.Size = new System.Drawing.Size(103, 17);
      this.checkBox.TabIndex = 0;
      this.checkBox.Text = "Acop CheckBox";
      this.checkBox.UseVisualStyleBackColor = true;
      this.checkBox.Click += new System.EventHandler(this.checkBox_Click);
      // 
      // AcopCheckBox
      // 
      this.Controls.Add(this.checkBox);
      this.Name = "AcopCheckBox";
      this.Size = new System.Drawing.Size(110, 21);
      this.ResumeLayout(false);
      this.PerformLayout();

    }

    private void defaultUpdate(AcopControl.AcopLink lnk)
    {
      if (getAutoUpdate())
      {
        int lnkSts = lnk.GetStatus();
        if (lnkSts != 0)
        {
          applyErrorStatus(lnk.GetStatusString(), lnkSts != lastLinkStatus);
          hasLnkError = true;
        }
        else
        {
          if (lnkSts != lastLinkStatus) applyToolTipText(null);
          if (hasLnkError)
          {
            checkBox.ForeColor = tmpColor;
            hasLnkError = false;
          }
          double v = getValueAsDouble(lnk);
          if (lnk.GetDisplayId() < 0)
          {
            if (!displaySettingsAcquired) initTransportInfo();
          }
          if (selectedBit == -1)
          {
            if (v == selectedVal) checkBox.Checked = true;
            if (v == deSelectedVal) checkBox.Checked = false;
          }
          else
          {
            bitValue = (Int32)v;
            if ((bitValue & (long)Math.Pow(2.0, selectedBit)) > 0)
            {
              checkBox.Checked = true;
            }
            else
            {
              checkBox.Checked = false;
            }
          }
        }
      }
    }

    private void initTransportInfo()
    {
      if (!IsDesignMode())
      {
        String[] labels = null;
        AcopTransport.DisplaySettings dsp = null;
        if (!AutoLinkAttach && !AutoLinkUpdate)
        {
          displaySettingsAcquired = true;
          return;
        }
        dsp = acquireAndApplyDisplaySettings(LinkAddress, LinkProperty, ref labels, false);
        if (dsp == null || dsp.SettingsStale()) return;
        displaySettingsAcquired = true;
      }
    }

    private int applyCheckState(object sender, bool selected)
    {
      object inpt;
      if (selectedBit == -1)
      {
        double value;
        if (selected) value = selectedVal;
        else value = deSelectedVal;
        Type type = typeof(double);
        inpt = AcopControl.AcopLink.GetDataObjectFromString(type, value.ToString());
      }
      else
      {
        Int32 iValue;
        if (selected) iValue = bitValue | (1 << selectedBit);
        else iValue = bitValue & ~(1 << selectedBit);
        Type type = typeof(Int32);
        inpt = AcopControl.AcopLink.GetDataObjectFromString(type, iValue.ToString());
      }
      AcopControl.AcopLink lnk = AcopControl.AcopLink.GetAcopLink(LinkProtocol, LinkAddress, LinkProperty, null, inpt, 0, this);
      return applySetting(sender, inpt, lnk);
    }
    public void applyErrorStatus(String sts, Boolean applyToTTT)
    {
      tmpColor = checkBox.ForeColor;
      checkBox.ForeColor = Color.Red;
      if (applyToTTT) applyToolTipText(sts);
    }

    private void checkBox_Click(object sender, EventArgs e)
    {
      int cc = applyCheckState(sender, checkBox.Checked);
      checkBox.Checked = false;
    }

    private void setCheckBoxText(String value)
    {
      cbText = value;
      checkBox.Text = cbText;
    }

    private void setSelectedValue(double value)
    {
      selectedVal = value;
    }
    private void setDeSelectedValue(double value)
    {
      deSelectedVal = value;
    }

    private void setSelectedBit(int value)
    {
      selectedBit = value;
    }

    #region Properties
    [Category("Acop.Display")]
    [Description("The description text")]
    [DefaultValue("Acop CheckBox")]
    public String CheckBoxText { get { return cbText; } set { setCheckBoxText(value); } }
    [Category("Acop.Display")]
    [Description("The set value when selected")]
    [DefaultValue(1.0)]
    public double SelectedValue { get { return selectedVal; } set { setSelectedValue(value); } }
    [Category("Acop.Display")]
    [Description("The set value when deselected")]
    [DefaultValue(0.0)]
    public double DeSelectedValue { get { return deSelectedVal; } set { setDeSelectedValue(value); } }
    [Category("Acop.Display")]
    [Description("Select a bit that will be toggle [1;0] - selected and deselected value will be ignored (-1 no bit selected)")]
    [DefaultValue(-1)]
    public int ToggleBit { get { return selectedBit; } set { setSelectedBit(value); } }
    [Category("Acop.Display")]
    [Browsable(false), EditorBrowsable(EditorBrowsableState.Never)]
    public new Boolean DisplayAsMultiChannel { get; set; }
    [Category("Acop.Display")]
    [Browsable(false), EditorBrowsable(EditorBrowsableState.Never)]
    public new int Precision { get; set; }

    [Category("Acop.Display")]
    [Browsable(false), EditorBrowsable(EditorBrowsableState.Never)]
    public new String Units { get; set; }

    [Category("Acop.Display")]
    [Browsable(false), EditorBrowsable(EditorBrowsableState.Never)]
    public new String ErrorValue { get; set; }
    [Category("Acop.Display")]
    [Browsable(false), EditorBrowsable(EditorBrowsableState.Never)]
    public new String Description { get; set; }

    [Category("Acop.Transport")]
    [Browsable(false), EditorBrowsable(EditorBrowsableState.Never)]
    public new Boolean LinkGrouped { get; set; }
    #endregion
  }
}
