﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using System.Windows.Forms.Design;
using System.Drawing.Design;
using System.Text.RegularExpressions;

namespace Acop
{
  [ToolboxItem(true)]
  [ToolboxBitmap(typeof(ComboBox))]
  public partial class AcopChooser : UserControl
  {
#   pragma warning disable 1591
    private static AcopComponent myAcop = new AcopComponent();
    private System.Windows.Forms.Timer lateInitializer;
    private static AcopComponent GetComponentInstance()
    {
      return myAcop;
    }
    private static AcopChooser instance = null;
    private static AcopChooser GetInstance()
    {
      if (instance == null) instance = new AcopChooser();
      return instance;
    }
    private String getDefaultDescription()
    {
      if (linkCtx == null || linkSrv == null || linkDev == null || linkProp == null)
        return null;
      StringBuilder sb = new StringBuilder();
      sb.Append("/");
      sb.Append(ctxBrowsing ? "<context>" : linkCtx);
      sb.Append("/");
      sb.Append(srvBrowsing ? "<server>" : linkSrv);
      sb.Append("/");
      sb.Append(devBrowsing ? "<device>" : linkDev);
      sb.Append("[");
      sb.Append(prpBrowsing ? "<property>" : linkProp);
      sb.Append("]");
      return sb.ToString();
    }
    private AcopChooserContextMenu cm;
    public AcopChooser()
    {
      InitializeComponent();
      instance = this;
      isDesignMode = (LicenseManager.UsageMode == LicenseUsageMode.Designtime);
      editorValue = new ApplyToEditorValue(targetList, this);
      cm = new AcopChooserContextMenu(this);
      this.ContextMenu = cm;
      if (isDesignMode)
      {
        if (linkCtx == null) linkCtx = AcopUserComponent.GetLastContextSetting();
        if (linkSrv == null) linkSrv = AcopUserComponent.GetLastServerSetting();
        if (linkDev == null) linkDev = AcopUserComponent.GetLastDeviceSetting();
        if (linkProp == null) linkProp = AcopUserComponent.GetLastPropertySetting();
      }
      else
      {
        lateInitializer = new System.Windows.Forms.Timer(components);
        lateInitializer.Tick += new System.EventHandler(lateInitializerUpdate);
        lateInitializer.Interval = 50;
      }
    }
    private ComboBoxStyle cbStyle = ComboBoxStyle.Simple;
    private void setDropdownStyle(ComboBoxStyle value)
    {
      if (value == cbStyle) return;
      this.SuspendLayout();
      ContextCombo.DropDownStyle = value;
      ServerCombo.DropDownStyle = value;
      PropertyCombo.DropDownStyle = value;
      DeviceCombo.DropDownStyle = value;
      cbStyle = value;
      this.ResumeLayout();
    }
    public delegate void addressUpdate(object sender, AcopChooser chooser);
    private Boolean hasPropertyQueryPrecedence = false;
    private Boolean hasDeviceQueryPrecedence = false;
    private Boolean isDesignMode = false;
    private Boolean initialPassDone = false;
    private Boolean ctxStale = true;
    private Boolean srvStale = true;
    private Boolean devStale = true;
    private Boolean prpStale = false;
    private Boolean subStale = false;

    private Boolean ctxBrowsing = true;
    private Boolean srvBrowsing = true;
    private Boolean devBrowsing = true;
    private Boolean prpBrowsing = true;

    private Boolean ctxApply = true;
    private Boolean srvApply = true;
    private Boolean devApply = true;
    private Boolean prpApply = true;

    public enum ApplyWhenEnum
    {
      ApplyButtonOnly = 0,
      SelectionChanges = 1,
      ContextMenuOnly = 2
    }
    ApplyWhenEnum applyWhen = ApplyWhenEnum.ApplyButtonOnly;
    private void setApplyWhen(ApplyWhenEnum value)
    {
      applyWhen = value;
      if (value != ApplyWhenEnum.ApplyButtonOnly) ApplyButton.Visible = false;
    }
    private ApplyWhenEnum getApplyWhen() { return applyWhen; }
    private int getNumChooserCombos()
    {
      int ncombos = 0;
      if (ctxBrowsing) ncombos++;
      if (srvBrowsing) ncombos++;
      if (devBrowsing) ncombos++;
      if (prpBrowsing) ncombos++;
      return ncombos;
    }
    private Boolean showHeader = true;
    private void setShowHeader(Boolean value)
    {
      showHeader = value;
      DescriptionLabel.Visible = value;
      if (applyWhen == ApplyWhenEnum.ApplyButtonOnly) ApplyButton.Visible = value;
      resizeChooser();
    }
    private void resizeChooser()
    {
      //return;
      this.SuspendLayout();
      int twidth = this.Width;
      int ypos = showHeader ? 29 : 3;
      int gwidth = 5, x = gwidth;
      int ncombos = getNumChooserCombos();
      if (ncombos > 0)
      {
        twidth -= (ncombos + 1) * gwidth;
        int cwidth = twidth / ncombos;
        if (ctxBrowsing)
        {
          this.ContextCombo.Location = new System.Drawing.Point(x, ypos);
          this.ContextCombo.Size = new System.Drawing.Size(cwidth, 21);
          x += (cwidth + gwidth);
        }
        if (srvBrowsing)
        {
          this.ServerCombo.Location = new System.Drawing.Point(x, ypos);
          this.ServerCombo.Size = new System.Drawing.Size(cwidth, 21);
          x += (cwidth + gwidth);
        }
        if (devBrowsing)
        {
          this.DeviceCombo.Location = new System.Drawing.Point(x, ypos);
          this.DeviceCombo.Size = new System.Drawing.Size(cwidth, 21);
          x += (cwidth + gwidth);
        }
        if (prpBrowsing)
        {
          this.PropertyCombo.Location = new System.Drawing.Point(x, ypos);
          this.PropertyCombo.Size = new System.Drawing.Size(cwidth, 21);
        }
      }
      this.ResumeLayout();
    }

    private void setCtxBrowsing(Boolean value)
    {
      ctxBrowsing = value;
      ContextCombo.Visible = value;
      resizeChooser();
    }
    private void setSrvBrowsing(Boolean value)
    {
      srvBrowsing = value;
      ServerCombo.Visible = value;
      resizeChooser();
    }
    private void setDevBrowsing(Boolean value)
    {
      devBrowsing = value;
      DeviceCombo.Visible = value;
      resizeChooser();
    }
    private void setPrpBrowsing(Boolean value)
    {
      prpBrowsing = value;
      PropertyCombo.Visible = value;
      resizeChooser();
    }

    private const String UNAVAILABLE = "<unavailable>";
    private String linkProt = AcopControl.AcopLink.GetDefaultProtocol();
    private String linkCtx = null;
    private String lastLinkCtx = null;
    private String linkSubsystem = null;
    private static String lastLinkSub = null;
    private String linkSrv = null;
    private String lastLinkSrv = null;
    private String linkDev = null;
    private String lastLinkDev = null;
    private String linkProp = null;
    private String lastLinkPrp = null;
    private String linkAddr = null;

    private String[] ctxLst = null;
    private String[] srvLst = null;
    private String[] devLst = null;
    private String[] prpLst = null;
    private String[] subLst = null;

    private Dictionary<String, AcopUserComponent> acopLst = new Dictionary<string, AcopUserComponent>();
    public AcopUserComponent[] getAcops()
    {
      Control parent = this.Parent;
      acopLst.Clear();
      foreach (var ac in parent.Controls.OfType<AcopUserComponent>())
      {
        acopLst.Add(ac.Name, ac);
      }
      return acopLst.Values.ToArray();
    }
    public String[] getAcopsByName()
    { // called at design time by chooserEditor : use (acop).Site.Name
      Control parent = this.Parent;
      acopLst.Clear();
      foreach (var ac in parent.Controls.OfType<AcopUserComponent>())
      {
        acopLst.Add(ac.Site.Name, (AcopUserComponent)ac);
      }
      return acopLst.Keys.ToArray();
    }
    private String targetList = "";
    public void setTargetList(String lst)
    {
      targetList = lst;
    }
    public String getTargetList()
    {
      return targetList == null ? "" : targetList;
    }
    public String[] getTargets()
    {
      return targetList.Split(new char[] { ';' }, StringSplitOptions.RemoveEmptyEntries);
    }
    private void applyToAcops()
    { // does not happen at design time : acop.Name is safe
      if (targetList == null || targetList.Length == 0) return;
      Control parent = this.Parent;
      if (parent == null) return;
      foreach (var acop in parent.Controls.OfType<AcopUserComponent>())
      {
        if (!acopLst.ContainsValue(acop)) acopLst.Add(acop.Name, acop);
      }
      AcopUserComponent ac;
      String[] tgtLst = getTargets();
      String prp, ctx, srv, dev;
      foreach (String tgt in tgtLst)
      {
        if (acopLst.TryGetValue(tgt, out ac))
        {
          ctx = ctxApply ? linkCtx : ac.LinkContext;
          srv = srvApply ? linkSrv : ac.LinkServer;
          dev = devApply ? linkDev : ac.LinkDevice;
          prp = prpApply ? linkProp : ac.LinkProperty;
          linkAddr = "/" + ctx + "/" + srv + "/" + dev;
          ac.Reassign(linkAddr, prp);
        }
      }
    }
    private bool IsDesignMode()
    {
      return isDesignMode;
    }
    private Boolean isDesignTimeSetServer()
    {
      if (IsDesignMode())
      {
        if (linkDev.CompareTo(UNAVAILABLE) == 0 ||
            linkProp.CompareTo(UNAVAILABLE) == 0)
          return true;
      }
      return false;
    }
    private void populateContexts()
    {
      if (IsDesignMode()) return;
      if (ctxLst != null) return;
      ctxLst = AcopControl.AcopLink.GetContextList(linkProt);
      if (ctxLst != null && ctxLst[0] != null)
      {
        ContextCombo.Items.Clear();
        foreach (String c in ctxLst) ContextCombo.Items.Add(c);
        if (cbHighlightFilter.Length > 0)
          setHighlightFilter(ContextCombo, cbHighlightFilter);
        if (targetedContext != null) ContextCombo.Text = targetedContext;
      }
    }
    private void populateServers()
    {
      if (IsDesignMode()) return;
      if (srvLst != null) return;
      srvLst = AcopControl.AcopLink.GetServerList(linkProt, linkCtx, linkSubsystem);
      if (srvLst != null && srvLst[0] != null)
      {
        ServerCombo.Items.Clear();
        foreach (String c in srvLst) ServerCombo.Items.Add(c);
        if (cbHighlightFilter.Length > 0)
          setHighlightFilter(ServerCombo, cbHighlightFilter);
        if (targetedServer != null) ServerCombo.Text = targetedServer;
      }
    }
    private Boolean devLstPopulated = false;
    private void populateDevices()
    {
      if (IsDesignMode()) return;
      if (devLst != null && devLstPopulated) return;
      if (linkSrv == null) return;
      devLst = AcopControl.AcopLink.GetDeviceList(linkProt, linkCtx, linkSubsystem, linkSrv, linkProp);
      devLstPopulated = false;
      if (devLst != null && devLst[0] != null)
      {
        DeviceCombo.Items.Clear();
        if (devLst.Length == 1 && devLst[0].Equals(UNAVAILABLE))
        {
          devLstPopulated = false;
          lateInitializer.Enabled = true;
          return;
        }
        foreach (String c in devLst) DeviceCombo.Items.Add(c);
        devLstPopulated = true;
        if (cbHighlightFilter.Length > 0)
          setHighlightFilter(DeviceCombo, cbHighlightFilter);
        if (targetedDevice == null) targetedDevice = DeviceCombo.Items[0].ToString();
        if (targetedDevice != null) DeviceCombo.Text = targetedDevice;
      }
      hasPropertyQueryPrecedence = AcopControl.AcopLink.HasPropertyQueryPrecedence(linkProt, linkCtx, linkSubsystem, linkSrv, linkProp);
    }
    private Boolean prpLstPopulated = false;
    private void populateProperties()
    {
      if (IsDesignMode()) return;
      if (prpLst != null && prpLstPopulated) return;
      if (linkSrv == null) return;
      prpLst = AcopControl.AcopLink.GetPropertyList(linkProt, linkCtx, linkSubsystem, linkSrv, linkDev);
      prpLstPopulated = false;
      if (prpLst != null && prpLst[0] != null)
      {
        PropertyCombo.Items.Clear();
        if (prpLst.Length == 1 && prpLst[0].Equals(UNAVAILABLE))
        {
          prpLstPopulated = false;
          lateInitializer.Enabled = true;
          return;
        }
        foreach (String c in prpLst) PropertyCombo.Items.Add(c);
        prpLstPopulated = true;
        if (cbHighlightFilter.Length > 0)
          setHighlightFilter(PropertyCombo, cbHighlightFilter);
        if (targetedProperty == null) targetedProperty = PropertyCombo.Items[0].ToString();
        if (targetedProperty != null) PropertyCombo.Text = targetedProperty;
      }
      hasDeviceQueryPrecedence = AcopControl.AcopLink.HasDeviceQueryPrecedence(linkProt, linkCtx, linkSubsystem, linkSrv);
    }
    private String getLinkProt()
    {
      instance = this;
      return linkProt == null ? AcopControl.AcopLink.GetDefaultProtocol() : linkProt;
    }
    private void setLinkProt(String protocol)
    {
      instance = this;
      linkProt = protocol;
      AcopControl.AcopLink lnk = AcopControl.AcopLink.GetAcopLink(linkProt, linkAddr, linkProp, GetComponentInstance());
      if (lnk != null) lnk.SetProtocol(linkProt);
      ctxLst = null; srvLst = null; devLst = null; prpLst = null; subLst = null;
    }
    private String getLinkAddr()
    {
      instance = this;
      return linkAddr == null ? "" : linkAddr;
    }
    private void setLinkAddr(String adr)
    {
      instance = this;
      linkAddr = adr;
    }
    String targetedServer = null;
    private void setLinkServer(String server)
    {
      instance = this;
      linkSrv = lastLinkSrv = server;
      ServerCombo.Text = server;
      targetedServer = server;
      srvStale = false;
      if (initialPassDone || isDesignTimeSetServer())
      {
        linkDev = null; lastLinkDev = null;
        linkProp = lastLinkPrp = null;
        devStale = true;
        if (IsDesignMode())
        {
          hasPropertyQueryPrecedence = false;
          hasDeviceQueryPrecedence = false;
          setLinkDevice(getFirstDevice());
        }
      }
      if (!IsDesignMode())
      {
        if (!initialPassDone)
        {
          if (linkDev == null) linkDev = lastLinkDev;
          if (linkProp == null) linkProp = lastLinkPrp;
          ToolTip.SetToolTip(DescriptionLabel, getDefaultDescription());
        }
        else
        {
          targetedProperty = null;
          targetedDevice = null;
        }
        if (devLst == null) populateDevices();
        if (prpLst == null) populateProperties();
      }
      if (linkCtx != null && linkDev != null)
      {
        linkAddr = AcopControl.AcopLink.GetConnectionAddress(linkProt, linkCtx, linkSubsystem, linkSrv, linkDev);
        setLinkAddr(linkAddr);
      }
      initialPassDone = true;
      populateServers();
      prpLst = null; devLst = null;
    }
    private String getFirstDevice()
    {
      if (!IsDesignMode()) return null;
      if (linkDev != null && linkDev.CompareTo(UNAVAILABLE) == 0)
      {
        linkDev = lastLinkDev = null;
      }
      if (linkProp != null && linkProp.CompareTo(UNAVAILABLE) == 0)
      {
        linkProp = lastLinkPrp = null;
      }
      if (!devStale && linkDev != null) return linkDev;
      if (lastLinkDev != null) return lastLinkDev;
      if (linkCtx != null && linkSrv != null)
      {
        if (linkProp == null && !hasPropertyQueryPrecedence)
        {
          hasPropertyQueryPrecedence = AcopControl.AcopLink.HasPropertyQueryPrecedence(linkProt, linkCtx, linkSubsystem, linkSrv, linkProp);
        }
        if (hasPropertyQueryPrecedence)
        { // first query
          prpStale = true;
          if (linkProp == null) return UNAVAILABLE;
        }
        devLst = AcopControl.AcopLink.GetDeviceList(linkProt, linkCtx, linkSubsystem, linkSrv, linkProp);
        if (devLst != null && devLst[0] != null)
        {
          DeviceCombo.Items.Clear();
          foreach (String c in devLst) DeviceCombo.Items.Add(c);
          return devLst[0];
        }
      }
      return "";
    }
    private String getLinkDevice()
    {
      instance = this;
      String dev = linkDev;
      if (dev == null)
      {
        linkDev = dev = getFirstDevice();
        if (linkCtx != null && linkSrv != null)
        {
          linkAddr = AcopControl.AcopLink.GetConnectionAddress(linkProt, linkCtx, linkSubsystem, linkSrv, dev);
          setLinkAddr(linkAddr);
        }
      }
      return dev == null ? UNAVAILABLE : dev;
    }
    private String targetedDevice = null;
    private void setLinkDevice(String device)
    {
      instance = this;
      linkDev = lastLinkDev = device;
      DeviceCombo.Text = device;
      targetedDevice = device;
      if (linkCtx != null && linkSrv != null)
      {
        linkAddr = AcopControl.AcopLink.GetConnectionAddress(linkProt, linkCtx, linkSubsystem, linkSrv, linkDev);
        setLinkAddr(linkAddr);
      }
      devStale = false;
      if (!hasPropertyQueryPrecedence)
      {
        prpStale = true;
        if (IsDesignMode())
        {
          linkProp = lastLinkPrp = null;
          setLinkProp(getFirstProperty());
        }
      }
      populateDevices();
      if (hasDeviceQueryPrecedence) prpLst = null;
    }
    private String getLinkProp()
    {
      instance = this;
      String prp = linkProp;
      if (prp == null)
      {
        linkProp = prp = getFirstProperty();
      }
      return prp == null ? UNAVAILABLE : prp;
    }
    private String targetedProperty = null;
    private void setLinkProp(String prp)
    {
      instance = this;
      linkProp = lastLinkPrp = prp;
      PropertyCombo.Text = prp;
      targetedProperty = prp;
      prpStale = false;
      if (hasPropertyQueryPrecedence)
      {
        devStale = true;
        if (IsDesignMode())
        {
          linkDev = lastLinkDev = null;
          setLinkDevice(getFirstDevice());
        }
        else
        {
          devLst = null;
        }
      }
      populateProperties();
    }

    private String getFirstProperty()
    {
      if (!IsDesignMode()) return null;
      if (linkProp != null && linkProp.CompareTo(UNAVAILABLE) == 0)
      {
        linkProp = lastLinkPrp = null;
      }
      if (linkDev != null && linkDev.CompareTo(UNAVAILABLE) == 0)
      {
        linkDev = lastLinkDev = null;
      }
      if (!prpStale && linkProp != null) return linkProp;
      if (linkCtx == null) linkCtx = getFirstContext();
      if (linkSrv == null) linkSrv = getFirstServer();
      if (linkDev == null) linkDev = getFirstDevice();
      if (lastLinkPrp != null) return lastLinkPrp;
      if (!hasDeviceQueryPrecedence)
      {
        hasDeviceQueryPrecedence = AcopControl.AcopLink.HasDeviceQueryPrecedence(linkProt, linkCtx, linkSubsystem, linkSrv);
      }
      prpLst = AcopControl.AcopLink.GetPropertyList(linkProt, linkCtx, linkSubsystem, linkSrv, linkDev);
      if (prpLst != null && prpLst[0] != null)
      {
        PropertyCombo.Items.Clear();
        foreach (String c in prpLst) PropertyCombo.Items.Add(c);
        return prpLst[0];
      }
      return "";
    }
    private String getFirstContext()
    {
      if (!IsDesignMode()) return null;
      if (lastLinkCtx != null) return lastLinkCtx;
      ctxLst = AcopControl.AcopLink.GetContextList(linkProt);
      if (ctxLst != null && ctxLst[0] != null)
      {
        ContextCombo.Items.Clear();
        foreach (String c in ctxLst) ContextCombo.Items.Add(c);
        return ctxLst[0];
      }
      return "";
    }
    private String getLinkContext()
    {
      instance = this;
      return linkCtx == null ? getFirstContext() : linkCtx;
    }
    private String targetedContext = null;
    private void setLinkContext(String context)
    {
      instance = this;
      linkCtx = lastLinkCtx = context;
      ContextCombo.Text = context;
      targetedContext = context;
      linkSrv = null;
      srvStale = true; ctxStale = false;
      srvLst = null; prpLst = null; devLst = null; subLst = null;
      populateContexts();
    }
    private String getFirstSubsystem()
    {
      if (!IsDesignMode()) return null;
      if (!ctxStale) return linkSubsystem;
      if (linkCtx == null) linkCtx = getFirstContext();
      if (lastLinkSub != null) return lastLinkSub;
      subLst = AcopControl.AcopLink.GetSubSystemList(linkProt, linkCtx);
      if (subLst != null && subLst[0] != null) return subLst[0];
      return "";
    }
    private String getLinkSubSystem()
    {
      instance = this;
      return linkSubsystem == null ? getFirstSubsystem() : linkSubsystem;
    }
    private void setLinkSubSystem(String subsystem)
    {
      instance = this;
      linkSubsystem = lastLinkSub = subsystem;
      srvStale = true; subStale = false;
      if (IsDesignMode()) return;
      srvLst = null; prpLst = null; devLst = null; subLst = null;
      if (subLst != null) return;
      subLst = AcopControl.AcopLink.GetSubSystemList(linkProt, linkCtx);
    }
    private String getFirstServer()
    {
      if (!IsDesignMode()) return null;
      if (!ctxStale && !srvStale && linkSrv != null) return linkSrv;
      if (linkCtx == null) linkCtx = getFirstContext();
      if (lastLinkSrv != null) return lastLinkSrv;
      srvLst = AcopControl.AcopLink.GetServerList(linkProt, linkCtx, linkSubsystem);
      if (srvLst != null && srvLst[0] != null)
      {
        ServerCombo.Items.Clear();
        foreach (String c in srvLst) ServerCombo.Items.Add(c);
        return srvLst[0];
      }
      return "";
    }
    private String getLinkServer()
    {
      instance = this;
      return linkSrv == null ? getFirstServer() : linkSrv;
    }

    #region Converter
    public class LinkProtocolConverter : StringConverter
    {
      public override bool GetStandardValuesSupported(ITypeDescriptorContext context)
      { //true means show a combobox
        return true;
      }
      public override bool GetStandardValuesExclusive(ITypeDescriptorContext context)
      { //true will limit to list. false will show the list, but allow free-form entry
        return false;
      }
      public override
      System.ComponentModel.TypeConverter.StandardValuesCollection GetStandardValues(ITypeDescriptorContext context)
      {
        return new StandardValuesCollection(AcopControl.AcopLink.GetProtocolList());
      }
    }

    public class LinkContextConverter : StringConverter
    {
      public override bool GetStandardValuesSupported(ITypeDescriptorContext context) { return true; }
      public override bool GetStandardValuesExclusive(ITypeDescriptorContext context) { return true; }
      public override
      System.ComponentModel.TypeConverter.StandardValuesCollection GetStandardValues(ITypeDescriptorContext context)
      {
        AcopChooser chs = GetInstance();
        chs.ctxLst = AcopControl.AcopLink.GetContextList(chs.linkProt);
        return new StandardValuesCollection(chs.ctxLst);
      }
    }

    public class LinkSubSystemConverter : StringConverter
    {
      public override bool GetStandardValuesSupported(ITypeDescriptorContext context) { return true; }
      public override bool GetStandardValuesExclusive(ITypeDescriptorContext context) { return true; }
      public override
      System.ComponentModel.TypeConverter.StandardValuesCollection GetStandardValues(ITypeDescriptorContext context)
      {
        AcopChooser chs = GetInstance();
        chs.subLst = AcopControl.AcopLink.GetSubSystemList(chs.linkProt, chs.linkCtx);
        return new StandardValuesCollection(chs.subLst);
      }
    }

    public class LinkServerConverter : StringConverter
    {
      public override bool GetStandardValuesSupported(ITypeDescriptorContext context) { return true; }
      public override bool GetStandardValuesExclusive(ITypeDescriptorContext context) { return true; }
      public override
      System.ComponentModel.TypeConverter.StandardValuesCollection GetStandardValues(ITypeDescriptorContext context)
      {
        AcopChooser chs = GetInstance();
        chs.srvLst = AcopControl.AcopLink.GetServerList(chs.linkProt, chs.linkCtx, chs.linkSubsystem);
        return new StandardValuesCollection(chs.srvLst);
      }
    }

    public class LinkDeviceConverter : StringConverter
    {
      public override bool GetStandardValuesSupported(ITypeDescriptorContext context) { return true; }
      public override bool GetStandardValuesExclusive(ITypeDescriptorContext context) { return false; }
      public override
      System.ComponentModel.TypeConverter.StandardValuesCollection GetStandardValues(ITypeDescriptorContext context)
      {
        AcopChooser chs = GetInstance();
        chs.hasPropertyQueryPrecedence = AcopControl.AcopLink.HasPropertyQueryPrecedence(chs.linkProt, chs.linkCtx, chs.linkSubsystem, chs.linkSrv);
        chs.devLst = AcopControl.AcopLink.GetDeviceList(chs.linkProt, chs.LinkContext, chs.linkSubsystem, chs.linkSrv, chs.linkProp);
        return new StandardValuesCollection(chs.devLst);
      }
    }

    public class LinkPropertyConverter : StringConverter
    {
      public override bool GetStandardValuesSupported(ITypeDescriptorContext context) { return true; }
      public override bool GetStandardValuesExclusive(ITypeDescriptorContext context) { return false; }
      public override
      System.ComponentModel.TypeConverter.StandardValuesCollection GetStandardValues(ITypeDescriptorContext context)
      {
        AcopChooser chs = GetInstance();
        chs.hasDeviceQueryPrecedence = AcopControl.AcopLink.HasDeviceQueryPrecedence(chs.linkProt, chs.linkCtx, chs.linkSubsystem, chs.linkSrv);
        chs.prpLst = AcopControl.AcopLink.GetPropertyList(chs.linkProt, chs.linkCtx, chs.linkSubsystem, chs.linkSrv, chs.linkDev);
        return new StandardValuesCollection(chs.prpLst);
      }
    }
    #endregion

    #region Properties
    /// <summary>
    /// The design-time browseable/editable link property (suppliments the link address)"
    /// </summary>
    [Category("Acop.Transport")]
    [Description("The design-time browseable/editable link property (suppliments the link address)")]
    [Browsable(true)]
    [DefaultValue("NSERVERS")]
    [TypeConverter(typeof(LinkPropertyConverter))]
    public String LinkProperty { get { return getLinkProp(); } set { setLinkProp(value); } }

    /// <summary>
    /// The design-time browseable/editable link device (part of the link address)
    /// </summary>
    [Category("Acop.Transport")]
    [Description("The design-time browseable/editable link device (part of the link address)")]
    [Browsable(true)]
    [DefaultValue("DEFAULT")]
    [TypeConverter(typeof(LinkDeviceConverter))]
    public String LinkDevice { get { return getLinkDevice(); } set { setLinkDevice(value); } }

    /// <summary>
    /// Identifies the transport link target. The target end point is given by the LinkAddress and the LinkProperty.
    /// </summary>
    [Category("Acop.Transport")]
    [Description("Identifies the transport link target. The target end point is given by the LinkAddress and the LinkProperty.")]
    public String LinkAddress { get { return getLinkAddr(); } set { setLinkAddr(value); } }

    /// <summary>
    /// The design-time browseable link server (part of the link address)
    /// </summary>
    [Category("Acop.Transport")]
    [Description("The design-time browseable link server (part of the link address)")]
    [Browsable(true)]
    [DefaultValue("ENS")]
    [TypeConverter(typeof(LinkServerConverter))]
    public String LinkServer { get { return getLinkServer(); } set { setLinkServer(value); } }

    /// <summary>
    /// The design-time browseable link subsystem
    /// </summary>
    [Category("Acop.Transport")]
    [Description("The design-time browseable link subsystem")]
    [Browsable(true)]
    [DefaultValue("ALL")]
    [TypeConverter(typeof(LinkSubSystemConverter))]
    public String LinkSubSystem { get { return getLinkSubSystem(); } set { setLinkSubSystem(value); } }

    /// <summary>
    /// The design-time browseable link context (part of the link address)
    /// </summary>
    [Category("Acop.Transport")]
    [Description("The design-time browseable link context (part of the link address)")]
    [Browsable(true)]
    [DefaultValue("SITE")]
    [TypeConverter(typeof(LinkContextConverter))]
    public String LinkContext { get { return getLinkContext(); } set { setLinkContext(value); } }

    /// <summary>
    /// The link control system protocol
    /// </summary>
    [Category("Acop.Transport")]
    [Description("The link control system protocol")]
    [Browsable(true)]
    [DefaultValue("SIMULATE")]
    [TypeConverter(typeof(LinkProtocolConverter))]
    public String LinkProtocol { get { return getLinkProt(); } set { setLinkProt(value); } }

    /// <summary>
    /// Optional label for the address chooser
    /// </summary>
    [Category("Acop.Display")]
    [Description("Optional label for the address chooser")]
    public String Description { get { return DescriptionLabel.Text; } set { DescriptionLabel.Text = value; } }

    /// <summary>
    /// Show description and apply button or not
    /// </summary>
    [Category("Acop.Display")]
    [Description("Show description and apply button or not")]
    public Boolean ShowHeader { get { return showHeader; } set { setShowHeader(value); } }
    /// <summary>
    /// true: apply the chosen context to targets
    /// </summary>
    [Category("Acop.Display")]
    [Description("true: apply the chosen context to targets")]
    public Boolean ApplyContext { get { return ctxApply; } set { ctxApply = value; } }
    /// <summary>
    /// true: apply the chosen server to targets
    /// </summary>
    [Category("Acop.Display")]
    [Description("true: apply the chosen server to targets")]
    public Boolean ApplyServer { get { return srvApply; } set { srvApply = value; } }
    /// <summary>
    /// true: apply the chosen device to targets
    /// </summary>
    [Category("Acop.Display")]
    [Description("true: apply the chosen device to targets")]
    public Boolean ApplyDevice { get { return devApply; } set { devApply = value; } }
    /// <summary>
    /// true: apply the chosen property to targets
    /// </summary>
    [Category("Acop.Display")]
    [Description("true: apply the chosen property to targets")]
    public Boolean ApplyProperty { get { return prpApply; } set { prpApply = value; } }
    /// <summary>
    /// display and allow context browsing in the address chooser
    /// </summary>
    [Category("Acop.Display")]
    [Description("display and allow context browsing in the address chooser")]
    public Boolean ChooseContexts { get { return ctxBrowsing; } set { setCtxBrowsing(value); } }
    /// <summary>
    /// display and allow server browsing in the address chooser
    /// </summary>
    [Category("Acop.Display")]
    [Description("display and allow server browsing in the address chooser")]
    public Boolean ChooseServer { get { return srvBrowsing; } set { setSrvBrowsing(value); } }
    /// <summary>
    /// display and allow device browsing in the address chooser
    /// </summary>
    [Category("Acop.Display")]
    [Description("display and allow device browsing in the address chooser")]
    public Boolean ChooseDevices { get { return devBrowsing; } set { setDevBrowsing(value); } }
    /// <summary>
    /// display and allow property browsing in the address chooser
    /// </summary>
    [Category("Acop.Display")]
    [Description("display and allow property browsing in the address chooser")]
    public Boolean ChooseProperties { get { return prpBrowsing; } set { setPrpBrowsing(value); } }
    /// <summary>
    /// Forecolor of items in a chooser combo box
    /// </summary>
    [Category("Acop.Display")]
    [Description("Forecolor of items in a chooser combo box")]
    public Color ItemForeColor { get { return cbForeColor; } set { setForeColor(value); } }
    /// <summary>
    /// Backcolor of items in a chooser combo box
    /// </summary>
    [Category("Acop.Display")]
    [Description("Backcolor of items in a chooser combo box")]
    public Color ItemBackColor { get { return cbBackColor; } set { setBackColor(value); } }
    /// <summary>
    /// Forecolor of items in a chooser combo box
    /// </summary>
    [Category("Acop.Display")]
    [Description("Highlighted Forecolor of items in a chooser combo box")]
    public Color HighlightForeColor { get { return cbHighlightForeColor; } set { setHighlightForeColor(value); } }
    /// <summary>
    /// Backcolor of items in a chooser combo box
    /// </summary>
    [Category("Acop.Display")]
    [Description("Highlighted Backcolor of items in a chooser combo box")]
    public Color HighlightBackColor { get { return cbHighlightBackColor; } set { setHighlightBackColor(value); } }
    /// <summary>
    /// Provides a filter string (regular expression) to use for highlighting combo box items
    /// </summary>
    [Category("Acop.Display")]
    [Description("Provides a filter string (regular expression) to use for highlighting combo box items")]
    public String HighlightFilter {  get { return cbHighlightFilter; } set { setHighlightFilter(value); } }
    /// <summary>
    /// Provides the combo box dropdown style
    /// </summary>
    [Category("Acop.Display")]
    [Description("Provides the combo box dropdown style")]
    public ComboBoxStyle ChooserStyle {  get { return cbStyle; } set { setDropdownStyle(value); } }
    [Category("Acop.Transport")]
    [Description("Occurs when the chosen address is to be applied")]
    public event addressUpdate acopAddressUpdate;

    private ApplyToEditorValue editorValue = null;
    private ApplyToEditorValue getEditorValue()
    {
      return editorValue;
    }
    private void setEditorValue(ApplyToEditorValue value)
    {
      if (value == null) value = new ApplyToEditorValue("", this);
      if (value != null) targetList = value.TargetAcops;
      editorValue = value;
    }
    [Category("Acop.Display")]
    [Description("Acop Chooser Editor")]
    [Editor(typeof(ApplyToEditor), typeof(System.Drawing.Design.UITypeEditor))]
    [DesignerSerializationVisibility(DesignerSerializationVisibility.Content)]
    [TypeConverter(typeof(ApplyToEditorConverter))]
    public ApplyToEditorValue ApplyTo { get { return getEditorValue(); } set { setEditorValue(value); } }

    /// <summary>
    /// Besides the Apply button, a selection can be applied if a single address combo box is visible, or via context menu
    /// </summary>
    [Category("Acop.Display")]
    [Description("Besides the Apply button, a selection can be applied if a single address combo box is visible, or via context menu")]
    public ApplyWhenEnum ApplyWhen { get { return getApplyWhen(); } set { setApplyWhen(value); } }

    #endregion

    private Boolean applyOnSelectionChange()
    {
      if (applyWhen != ApplyWhenEnum.SelectionChanges) return false;
      if (getNumChooserCombos() != 1) return false;
      return true;
    }
    private void DeviceChanged(object sender, EventArgs e)
    {
      if (IsDesignMode()) return;
      targetedDevice = linkDev = DeviceCombo.Text;
      if (applyOnSelectionChange())
      {
        ApplyButtonClicked(sender, e);
      }
      if (!prpBrowsing) return;
      if (!hasDeviceQueryPrecedence && prpLst != null) return;
      prpLst = null;
      linkProp = null;
      targetedProperty = null;
      populateProperties();
      if (prpLst != null) PropertyCombo.Text = prpLst[0];
      linkProp = PropertyCombo.Text;
      this.ToolTip.SetToolTip(this.DeviceCombo, linkDev);
    }

    private void PropertyChanged(object sender, EventArgs e)
    {
      if (IsDesignMode()) return;
      targetedProperty = linkProp = PropertyCombo.Text;
      if (!prpBrowsing && hasDeviceQueryPrecedence)
      { // special doocs case (get rid of the _SVR nonsense ...) 
        devLst = null;
        populateDevices();
      }
      if (applyOnSelectionChange())
      {
        ApplyButtonClicked(sender, e);
      }
      if (hasDeviceQueryPrecedence || !hasPropertyQueryPrecedence)
      { // only return if I've already determined a device list ...
        if (DeviceCombo.Items.Count > 0) return;
      }
      devLst = null;
      linkDev = null;
      targetedDevice = null;
      populateDevices();
      if (devLst != null) DeviceCombo.Text = devLst[0];
      linkDev = DeviceCombo.Text;
      this.ToolTip.SetToolTip(this.PropertyCombo, linkProp);
    }

    private void ServerChanged(object sender, EventArgs e)
    {
      if (IsDesignMode()) return;
      targetedServer = linkSrv = ServerCombo.Text;
      if (applyOnSelectionChange())
      {
        ApplyButtonClicked(sender, e);
      }
      if (devBrowsing && prpBrowsing)
      {
        hasPropertyQueryPrecedence = false;
        hasDeviceQueryPrecedence = false;
      }
      if (devBrowsing)
      {
        devLst = null;
        linkDev = null;
        populateDevices();
        if (devLst != null) DeviceCombo.Text = devLst[0];
      }
      if (prpBrowsing)
      {
        prpLst = null;
        linkProp = null;
      }
      this.ToolTip.SetToolTip(this.ServerCombo, linkSrv);
    }
    private Color cbForeColor = Color.Black;
    private Color cbBackColor = Color.White;
    private void setBackColor(Color value)
    {
      cbBackColor = value;
      DeviceCombo.BackColor = value;
      PropertyCombo.BackColor = value;
      ServerCombo.BackColor = value;
      ContextCombo.BackColor = value;
      Refresh();
    }
    private void setForeColor(Color value)
    {
      cbForeColor = value;
      DeviceCombo.ForeColor = value;
      PropertyCombo.ForeColor = value;
      ServerCombo.ForeColor = value;
      ContextCombo.ForeColor = value;
      Refresh();
    }
    private Color cbHighlightForeColor = Color.DarkBlue;
    private Color cbHighlightBackColor = Color.Azure;
    private void setHighlightForeColor(Color value)
    {
      cbHighlightForeColor = value;
    }
    private void setHighlightBackColor(Color value)
    {
      cbHighlightBackColor = value;
    }
    private String cbHighlightFilter = "";
    private Boolean TextMatches(String pattern, String txt)
    {
      if (txt == null || txt.Length == 0) return false;
      if (pattern == null || pattern.Length == 0) return true;
      if (pattern.StartsWith("-"))
      { // cut people a break with regular expressions ...
        // pattern = "^(?!.*XYZ).*$"; -> everything except XYZ matches
        pattern = "^(?!.*" + pattern.Substring(1) + ").*$";
      }
      return Regex.Match(txt, pattern).Success;
    }
    private void setHighlightFilter(ComboBox cb, String value)
    {
      if (value == null) value = "";
      cbHighlightFilter = value;
      if (value.Length == 0) return;
      if (cb.Items.Count > 0)
      {
        Boolean[] vals = new Boolean[cb.Items.Count];
        Boolean dictItemNeeded = false;
        for (int i=0; i <cb.Items.Count; i++)
        {
          vals[i] = TextMatches(value, cb.Items[i].ToString());
          if (vals[i]) dictItemNeeded = true;
        }
        if (dictItemNeeded)
        {
          if (cbIsHighlighted.ContainsKey(cb))
          {
            cbIsHighlighted.Remove(cb);
          }
          cbIsHighlighted.Add(cb, vals);
        }
      }
    }
    private void setHighlightFilter(String value)
    {
      if (value == null) value = "";
      cbHighlightFilter = value;
      if (value.Length == 0) return;
      setHighlightFilter(ContextCombo, value);
      setHighlightFilter(ServerCombo, value);
      setHighlightFilter(PropertyCombo, value);
      setHighlightFilter(DeviceCombo, value);
    }
    private Dictionary<ComboBox, Boolean[]> cbIsHighlighted = new Dictionary<ComboBox, bool[]>();
    /// <summary>
    /// Supply highlighted information via an array of boolean values
    /// </summary>
    /// <param name="values">An array containing whether a device item is to be highlighted or not</param>
    public void SetDeviceHighlighted(Boolean[] values)
    {
      if (cbIsHighlighted.ContainsKey(DeviceCombo))
      {
        cbIsHighlighted.Remove(DeviceCombo);
      }
      cbIsHighlighted.Add(DeviceCombo, values);
    }
    /// <summary>
    /// Supply highlighted information via an array of boolean values
    /// </summary>
    /// <param name="values">An array containing whether a server item is to be highlighted or not</param>
    public void SetServerHighlighted(Boolean[] values)
    {
      if (cbIsHighlighted.ContainsKey(ServerCombo))
      {
        cbIsHighlighted.Remove(ServerCombo);
      }
      cbIsHighlighted.Add(ServerCombo, values);
    }
    /// <summary>
    /// Supply highlighted information via an array of boolean values
    /// </summary>
    /// <param name="values">An array containing whether a property item is to be highlighted or not</param>
    public void SetPropertyHighlighted(Boolean[] values)
    {
      if (cbIsHighlighted.ContainsKey(PropertyCombo))
      {
        cbIsHighlighted.Remove(PropertyCombo);
      }
      cbIsHighlighted.Add(PropertyCombo, values);
    }
    /// <summary>
    /// Supply highlighted information via an array of boolean values
    /// </summary>
    /// <param name="values">An array containing whether a context item is to be highlighted or not</param>
    public void SetContextHighlighted(Boolean[] values)
    {
      if (cbIsHighlighted.ContainsKey(ContextCombo))
      {
        cbIsHighlighted.Remove(ContextCombo);
      }
      cbIsHighlighted.Add(ContextCombo, values);
    }
    private Color getItemForeColor(ComboBox cb, int index)
    {
      if (index < 0 || index >= cb.Items.Count) return cbForeColor;
      Boolean[] vals;
      if (cbIsHighlighted.TryGetValue(cb,out vals))
      {
        if (index < vals.Length)
        {
          return vals[index] ? cbHighlightForeColor : cbForeColor;
        }
      }
      return cbForeColor;
    }
    private Color getItemBackColor(ComboBox cb, int index)
    {
      if (index < 0 || index >= cb.Items.Count) return cbBackColor;
      Boolean[] vals;
      if (cbIsHighlighted.TryGetValue(cb, out vals))
      {
        if (index < vals.Length)
        {
          return vals[index] ? cbHighlightBackColor : cbBackColor;
        }
      }
      return cbBackColor;
    }
    private void ContextChanged(object sender, EventArgs e)
    {
      if (IsDesignMode()) return;
      targetedContext = linkCtx = ContextCombo.Text;
      if (applyOnSelectionChange())
      {
        ApplyButtonClicked(sender, e);
      }
      if (srvBrowsing)
      {
        srvLst = null;
        linkSrv = null;
        populateServers();
        if (srvLst != null) ServerCombo.Text = srvLst[0];
        linkSrv = ServerCombo.Text;
      }
      this.ToolTip.SetToolTip(this.ContextCombo, linkCtx);
    }
    private Boolean comboOpenedUpwards(ComboBox cb)
    {
      Point p = this.PointToScreen(cb.Location);
      int screenHeight = Screen.GetBounds(new Point(0, 0)).Bottom; // lowest point
      int ddHeight = cb.DropDownHeight * 4;
      return ((screenHeight - p.Y) < ddHeight);
    }
    private void showComboItem(ComboBox cb, DrawItemEventArgs e)
    {
      if (IsDesignMode()) return;
      if (e.Index < 0) return; // added this line thanks to Andrew's comment
      string text = cb.GetItemText(cb.Items[e.Index]);
      int bottom = 0; //  comboOpenedUpwards(cb) ? 0 : e.Bounds.Bottom;
      e.DrawBackground();
      Color fc = getItemForeColor(cb, e.Index);
      Color bc = getItemBackColor(cb, e.Index);
      SolidBrush fbr = new SolidBrush(fc);
      SolidBrush bbr = new SolidBrush(bc);
      e.Graphics.FillRectangle(bbr, e.Bounds);
      e.Graphics.DrawString(text, e.Font, fbr, e.Bounds);
      fbr.Dispose();
      bbr.Dispose();
      if (!cb.DroppedDown) return;
      if ((e.State & DrawItemState.Selected) == DrawItemState.Selected)
      {
        ToolTip.Show(text, cb, e.Bounds.Right, bottom);
      }
      e.DrawFocusRectangle();
    }
    private void ComboDrawItem(object sender, DrawItemEventArgs e)
    {
      if (sender is ComboBox)
      {
        ComboBox combo = (ComboBox)sender;
        showComboItem(combo, e);
      }
    }
    private void ComboDropDownClosed(object sender, EventArgs e)
    {
      if (sender is ComboBox)
      {
        ComboBox combo = (ComboBox)sender;
        ToolTip.Hide(combo);
      }
    }
    public void ApplySelection(object sender)
    {
      if (acopAddressUpdate != null)
      {
        acopAddressUpdate(sender, this);
      }
      else
      {
        applyToAcops();
      }
    }
    private void ApplyButtonClicked(object sender, EventArgs e)
    {
      ApplySelection(sender);
    }

    private void ChooserSizeChanged(object sender, EventArgs e)
    {
      resizeChooser();
    }

    private void handleMouseDown(MouseEventArgs e)
    {
      if (applyOnSelectionChange()) return;
      if (e.Button == MouseButtons.Right)
      {
        Point pos = this.PointToClient(Cursor.Position);
        cm.Show(this, pos, LeftRightAlignment.Right);
      }
    }
    private void ChooserMouseDown(object sender, MouseEventArgs e)
    {
      handleMouseDown(e);
    }

    private void DeviceMouseDown(object sender, MouseEventArgs e)
    {
      handleMouseDown(e);
    }

    private void PropertyMouseDown(object sender, MouseEventArgs e)
    {
      handleMouseDown(e);
    }
    private void ServerMouseDown(object sender, MouseEventArgs e)
    {
      handleMouseDown(e);
    }
    private void ContextMouseDown(object sender, MouseEventArgs e)
    {
      handleMouseDown(e);
    }
    private void lateInitializerUpdate(object sender, EventArgs e)
    {
      if (!devLstPopulated && devBrowsing)
      {
        lateInitializer.Interval = 2000; // keep trying ...
        populateDevices();
        if (!devLstPopulated) return;
        if (targetedDevice != null &&
          DeviceCombo.Items.Contains(targetedDevice))
        {
          DeviceCombo.Text = targetedDevice;
        }
      }
      if (!prpLstPopulated && prpBrowsing)
      {
        lateInitializer.Interval = 2000; // keep trying ...
        populateProperties();
        if (!prpLstPopulated) return;
        if (targetedProperty != null &&
            PropertyCombo.Items.Contains(targetedProperty))
        {
          PropertyCombo.Text = targetedProperty;
        }
      }
      lateInitializer.Enabled = false;
    }
  }
  [ToolboxItem(false)]
  public class AcopChooserContextMenu : ContextMenu
  {
#   pragma warning disable 1591
    private AcopChooser chooser;
    private MenuItem apply;

    public AcopChooserContextMenu(AcopChooser chooser)
    {
      this.chooser = chooser;
      apply = new MenuItem("Apply");
      apply.Click += new EventHandler(this.applyChanges);

      this.MenuItems.Add(apply);
    }
    private void applyChanges(object sender, EventArgs e)
    {
      chooser.ApplySelection(sender);
    }
  }
  class ChooserEditor : Form
  {
#   pragma warning disable 1591
    private AcopChooser chooser = null;
    private ApplyToEditor editor = null;
    public void SetEditor(ApplyToEditor editor)
    {
      this.editor = editor;
    }
    private String[] chosenAcops = null;
    public String[] GetChosenAcops()
    {
      return chosenAcops;
    }
    public void SetChooser(AcopChooser chooserControl)
    {
      chooser = chooserControl;
    }
    public ChooserEditor(AcopChooser chooserControl)
    {
      InitializeChooserEditorComponent();
      chooser = chooserControl;
      String[] acops = chooser.getAcopsByName();
      if (acops != null && acops.Length > 0)
      {
        foreach (String s in acops)
        {
          if (!AcopComboBox.Items.Contains(s)) AcopComboBox.Items.Add(s);
        }
        AcopComboBox.Text = acops[0];
      }
      chosenAcops = chooser.getTargets();
      if (chosenAcops != null && chosenAcops.Length > 0)
      {
        foreach (String s in chosenAcops) AcopList.Items.Add(s);
      }
    }
    /// <summary>
    /// Required designer variable.
    /// </summary>
    private System.ComponentModel.IContainer components = null;

    /// <summary>
    /// Clean up any resources being used.
    /// </summary>
    /// <param name="disposing">true if managed resources should be disposed; otherwise, false.</param>
    protected override void Dispose(bool disposing)
    {
      if (disposing && (components != null))
      {
        components.Dispose();
      }
      base.Dispose(disposing);
    }

    #region Windows Form Designer generated code

    /// <summary>
    /// Required method for Designer support - do not modify
    /// the contents of this method with the code editor.
    /// </summary>
    private void InitializeChooserEditorComponent()
    {
      this.OKButton = new System.Windows.Forms.Button();
      this.CancelButton = new System.Windows.Forms.Button();
      this.AcopList = new System.Windows.Forms.ListBox();
      this.AcopComboBox = new System.Windows.Forms.ComboBox();
      this.AddButton = new System.Windows.Forms.Button();
      this.SuspendLayout();
      // 
      // OKButton
      // 
      this.OKButton.Location = new System.Drawing.Point(144, 16);
      this.OKButton.Name = "OKButton";
      this.OKButton.Size = new System.Drawing.Size(59, 25);
      this.OKButton.TabIndex = 0;
      this.OKButton.Text = "OK";
      this.OKButton.UseVisualStyleBackColor = true;
      this.OKButton.Click += new System.EventHandler(this.OKClicked);
      // 
      // CancelButton
      // 
      this.CancelButton.Location = new System.Drawing.Point(144, 47);
      this.CancelButton.Name = "CancelButton";
      this.CancelButton.Size = new System.Drawing.Size(59, 25);
      this.CancelButton.TabIndex = 1;
      this.CancelButton.Text = "Cancel";
      this.CancelButton.UseVisualStyleBackColor = true;
      this.CancelButton.Click += new System.EventHandler(this.CancelClicked);
      // 
      // AcopList
      // 
      this.AcopList.FormattingEnabled = true;
      this.AcopList.Location = new System.Drawing.Point(9, 77);
      this.AcopList.Name = "AcopList";
      this.AcopList.Size = new System.Drawing.Size(194, 108);
      this.AcopList.TabIndex = 2;
      this.AcopList.KeyDown += new System.Windows.Forms.KeyEventHandler(this.AcopListKeyDown);
      // 
      // AcopComboBox
      // 
      this.AcopComboBox.FormattingEnabled = true;
      this.AcopComboBox.Location = new System.Drawing.Point(9, 19);
      this.AcopComboBox.Name = "AcopComboBox";
      this.AcopComboBox.Size = new System.Drawing.Size(123, 21);
      this.AcopComboBox.Sorted = true;
      this.AcopComboBox.TabIndex = 3;
      // 
      // AddButton
      // 
      this.AddButton.Location = new System.Drawing.Point(9, 46);
      this.AddButton.Name = "AddButton";
      this.AddButton.Size = new System.Drawing.Size(59, 25);
      this.AddButton.TabIndex = 4;
      this.AddButton.Text = "Add";
      this.AddButton.UseVisualStyleBackColor = true;
      this.AddButton.Click += new System.EventHandler(this.AddClicked);
      // 
      // ChooserEditor
      // 
      this.AutoScaleDimensions = new System.Drawing.SizeF(6F, 13F);
      this.AutoScaleMode = System.Windows.Forms.AutoScaleMode.Font;
      this.ClientSize = new System.Drawing.Size(212, 195);
      this.ControlBox = false;
      this.Controls.Add(this.AddButton);
      this.Controls.Add(this.AcopComboBox);
      this.Controls.Add(this.AcopList);
      this.Controls.Add(this.CancelButton);
      this.Controls.Add(this.OKButton);
      this.MaximizeBox = false;
      this.MinimizeBox = false;
      this.Name = "ChooserEditor";
      this.Text = "Choose Acop User Components";
      this.ResumeLayout(false);

    }
    #endregion
    private void CancelClicked(object sender, EventArgs e)
    {
      editor.SetValuesStale(false);
      this.Dispose();
    }
    private void OKClicked(object sender, EventArgs e)
    {
      StringBuilder sb = new StringBuilder();

      foreach (String item in AcopList.Items) sb.Append(item+";");
      chooser.setTargetList(sb.ToString());
      editor.SetValuesStale(true);
      this.Dispose();
    }
    private void AddClicked(object sender, EventArgs e)
    {
      String txt = AcopComboBox.Text;
      if (!AcopList.Items.Contains(txt)) AcopList.Items.Add(txt);
    }
    private void AcopListKeyDown(object sender, KeyEventArgs e)
    {
      if (AcopList.SelectedItems.Count == 0) return;
      if (e.KeyCode == Keys.Delete)
      {
        for (int i = AcopList.SelectedItems.Count - 1; i >= 0; i--)
          AcopList.Items.Remove(AcopList.SelectedItems[i]);
      }
    }
    private System.Windows.Forms.Button OKButton;
    private new System.Windows.Forms.Button CancelButton;
    private System.Windows.Forms.ListBox AcopList;
    private System.Windows.Forms.ComboBox AcopComboBox;
    private System.Windows.Forms.Button AddButton;
  }
  [TypeConverter(typeof(ApplyToEditorConverter))]
  public class ApplyToEditorValue
  {
#   pragma warning disable 1591
    private AcopChooser chooser = null;
    private String valuesString = null;
    private void setValuesString(String value)
    {
      valuesString = value;
      chooser.setTargetList(value);
    }
    public String TargetAcops {  get { return valuesString; } set { setValuesString(value); } }
    public ApplyToEditorValue(String value, AcopChooser chsr)
    {
      if (value == null) value = "";
      valuesString = value;
      chooser = chsr;
    }
    public override string ToString()
    {
      return valuesString;
    }
  }
  public class ApplyToEditorConverter : TypeConverter
  {
#   pragma warning disable 1591
    public override object ConvertTo(
        ITypeDescriptorContext context,
        System.Globalization.CultureInfo culture,
        object value, Type destinationType)
    {
      if (destinationType == typeof(string)) return value.ToString();
      return base.ConvertTo(context, culture, value, destinationType);
    }
    public override object ConvertFrom(
        ITypeDescriptorContext context,
        System.Globalization.CultureInfo culture,
        object value)
    {
      if (value.GetType() == typeof(string))
      {
        return new ApplyToEditorValue((string)(value),(AcopChooser)context);
      }
      else
      {
        return base.ConvertFrom(context, culture, value);
      }
    }
    public override bool CanConvertTo(ITypeDescriptorContext context, Type destinationType)
    {
      if (destinationType == typeof(string)) return true;
      return base.CanConvertTo(context, destinationType);
    }
    public override bool CanConvertFrom(ITypeDescriptorContext context, Type sourceType)
    {
      if (sourceType == typeof(string)) return true;
      return base.CanConvertFrom(context, sourceType);
    }
  }
  public class ApplyToEditor : UITypeEditor
  {
#   pragma warning disable 1591
    private Boolean valuesStale = false;
    private String valuesString = null;
    
    public ApplyToEditor()
    {
    }
    public ApplyToEditor(String value)
    {
      valuesString = value;
    }
    public override string ToString()
    {
      return valuesString;
    }
    public void SetValuesStale(Boolean value) { valuesStale = value; }
    public override UITypeEditorEditStyle GetEditStyle(System.ComponentModel.ITypeDescriptorContext context)
    {
      return UITypeEditorEditStyle.Modal;
    }
    public override object EditValue(ITypeDescriptorContext context, IServiceProvider provider, object value)
    {
      AcopChooser chs = (AcopChooser)context.Instance;
      IWindowsFormsEditorService frmsvr = (IWindowsFormsEditorService)provider.GetService(typeof(IWindowsFormsEditorService));
      if (frmsvr == null) return null;
      valuesString = chs.getTargetList(); //  chs.ApplyTo;
      ChooserEditor page = new ChooserEditor(chs);
      page.SetEditor(this);
      frmsvr.ShowDialog(page);
      if (valuesStale)
      {
        valuesString = chs.getTargetList();
        value = new ApplyToEditorValue(valuesString, chs);
      }
      return base.EditValue(context, provider, value);
    }
  }
}
