﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace Acop
{
  [ToolboxItem(true)]
  [ToolboxBitmap(typeof(ComboBox))]
  public partial class AcopComboBox : AcopUserComponent
  {
#   pragma warning disable 1591
    private ComboBox comboBox;

    private int lastLinkStatus = 0;
    private int oldIndex = -1;
    private Boolean hasLnkError = false;
    private Color tmpColor;

    private string[] cbItems = new String[] { };
    private ArrayList CBItemsList = new ArrayList();

    public AcopComboBox()
    {
      InitializeComponent();
      SetDefaultUpdater(defaultUpdate);
      setControlsList(new Control[] { comboBox }, true);
      SetLateInitializerProlog(initTransportInfo);
      StartLateInitialization();
    }

    private void InitializeComponent()
    {
      this.comboBox = new System.Windows.Forms.ComboBox();
      this.SuspendLayout();
      // 
      // comboBox
      // 
      this.comboBox.Anchor = ((System.Windows.Forms.AnchorStyles)((((System.Windows.Forms.AnchorStyles.Top | System.Windows.Forms.AnchorStyles.Bottom) 
            | System.Windows.Forms.AnchorStyles.Left) 
            | System.Windows.Forms.AnchorStyles.Right)));
      this.comboBox.FormattingEnabled = true;
      this.comboBox.Location = new System.Drawing.Point(0, 2);
      this.comboBox.Name = "comboBox";
      this.comboBox.Size = new System.Drawing.Size(121, 21);
      this.comboBox.TabIndex = 0;
      this.comboBox.SelectionChangeCommitted += new System.EventHandler(this.comboBox_SelectionChangeCommitted);
      // 
      // AcopComboBox
      // 
      this.Controls.Add(this.comboBox);
      this.Name = "AcopComboBox";
      this.Size = new System.Drawing.Size(122, 26);
      this.ResumeLayout(false);

    }

    private void comboBox_SelectionChangeCommitted(object sender, EventArgs e)
    {
      if (CBItemsList.Count < 1) return;
      applyCheckState(sender, comboBox.SelectedIndex);
    }

    private void defaultUpdate(AcopControl.AcopLink lnk)
    {
      if (getAutoUpdate())
      {
        int lnkSts = lnk.GetStatus();
        if (lnkSts != 0 && ErrorValue.Length == 0)
        {
          applyErrorStatus(lnk.GetStatusString(), lnkSts != lastLinkStatus);
          hasLnkError = true;
        }
        else
        {
          if (lnkSts != lastLinkStatus) applyToolTipText(null);
          if (hasLnkError)
          {
            comboBox.ForeColor = tmpColor;
            hasLnkError = false;
          }
          double v = getValueAsDouble(lnk);
          if (lnk.GetDisplayId() < 0)
          {
            if (!displaySettingsAcquired) initTransportInfo();
          }
          setBoxItem(v);
        }
      }
    }

    private void initTransportInfo()
    {
      if (!IsDesignMode())
      {
        String[] labels = null;
        AcopTransport.DisplaySettings dsp = null;
        if (!AutoLinkAttach && !AutoLinkUpdate)
        {
          displaySettingsAcquired = true;
          return;
        }
        dsp = acquireAndApplyDisplaySettings(LinkAddress, LinkProperty, ref labels, false);
        if (dsp == null || dsp.SettingsStale()) return;
        displaySettingsAcquired = true;
      }
    }

    private void setBoxItem(double value)
    {
      if (comboBox.DroppedDown) return;
      for (int i=0; i<CBItemsList.Count; i++)
      {
        double itemValue = ((CBItem)CBItemsList[i]).ValueAsDouble;
        if(value == itemValue)
        {
          comboBox.SelectedIndex = i;
          oldIndex = i;
          return;
        }
      }
    }

    private int applyCheckState(object sender, int index)
    {
      String value = ((CBItem)CBItemsList[index]).Value;
      Type type = getValueType();
      if (type == null) type = getValueType(value);
      object inpt = AcopControl.AcopLink.GetDataObjectFromString(type, value);
      AcopControl.AcopLink lnk = AcopControl.AcopLink.GetAcopLink(LinkProtocol, LinkAddress, LinkProperty, null, inpt, 0, this);
      return applySetting(sender, inpt, lnk);
    }

    public void applyErrorStatus(String sts, Boolean applyToTTT)
    {
      tmpColor = comboBox.ForeColor;
      comboBox.ForeColor = Color.Red;
      if (applyToTTT) applyToolTipText(sts);
    }

    private void setCBItems(string[] value)
    {
      if (value == null) return;
      
      CBItemsList.Clear();
      comboBox.Items.Clear();
      string[] seperator = new string[] { "|" };

      try
      {
        foreach (string s in value)
        {
          string[] sr = s.Split(seperator, StringSplitOptions.RemoveEmptyEntries);
          CBItemsList.Add(new CBItem(sr[0], sr[1]));
          comboBox.Items.Add(sr[0]);
        }
        cbItems = value;
      }
      catch (Exception ex)
      {
        string errorCaption = "Error: Wrong Items format";
        string errorMessage = "Wrong Items format! The format must be: Text|Value e.g. Off|0" + Environment.NewLine + ex.ToString();

        MessageBox.Show(this, errorMessage, errorCaption, MessageBoxButtons.OK, MessageBoxIcon.Error);
      }
    } 

    #region Properties
    [Category("Acop.Display")]
    [Description("The items in the combo box. (e.g. text1|value1;text2|value2)")]
    [DesignerSerializationVisibility(DesignerSerializationVisibility.Content)]
    [Editor("System.Windows.Forms.Design.StringArrayEditor, System.Design, Version = 2.0.0.0, Culture = neutral, PublicKeyToken = b03f5f7f11d50a3a", typeof(System.Drawing.Design.UITypeEditor))]
    public string[] Items { get { return cbItems; } set { setCBItems(value); } } 

    [Category("Acop.Display")]
    [Browsable(false), EditorBrowsable(EditorBrowsableState.Never)]
    public new Boolean DisplayAsMultiChannel { get; set; }
    [Category("Acop.Display")]
    [Browsable(false), EditorBrowsable(EditorBrowsableState.Never)]
    public new int Precision { get; set; }

    [Category("Acop.Display")]
    [Browsable(false), EditorBrowsable(EditorBrowsableState.Never)]
    public new String Units { get; set; }

    [Category("Acop.Display")]
    [Browsable(false), EditorBrowsable(EditorBrowsableState.Never)]
    public new String ErrorValue { get; set; }
    [Category("Acop.Display")]
    [Browsable(false), EditorBrowsable(EditorBrowsableState.Never)]
    public new String Description { get; set; }

    [Category("Acop.Transport")]
    [Browsable(false), EditorBrowsable(EditorBrowsableState.Never)]
    public new Boolean LinkGrouped { get; set; }
    #endregion
  }

  public class CBItem
  {
    private string displayName;
    private string displayValue;

    public CBItem(string dName, string dValue)
    {
      this.displayName = dName;
      this.displayValue = dValue;
    }

    public string Name { get { return displayName; } }
    public string Value { get { return displayValue; } }
    public double ValueAsDouble { get { return Convert.ToDouble(displayValue); } }
  }

}
