﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace Acop
{
  [ToolboxItem(false)]
  public class AcopComponent : Component, INotifyPropertyChanged
  {
#   pragma warning disable 1591
    static AcopComponent instance = null;
    Component component = null;
    private IDictionary<string, int> flagsDict = null;
    private AcopTransport.DisplaySettings linkDspSettings = null;
    private Type valueType = null;
    static DateTime utcThen = new DateTime(1970, 1, 1);
    private object notifyMutex = new object();

    private object linkDataOut = null;
    private object linkDataIn = null;
    private int linkFlags = 0;
    private String linkAddr = null;
    private int defaultInterval = 1000;

    private String linkProt = AcopControl.AcopLink.GetDefaultProtocol();
    private String linkFlagsString = "NONE";

    private Boolean isDesignMode = false;
    private Boolean initialPassDone = false;
    private Boolean displaySettingsAcquired = false;

    private Boolean ctxStale = true;
    private Boolean srvStale = true;
    private Boolean devStale = true;
    private Boolean prpStale = false;
    private Boolean subStale = true;

    private Boolean hasPropertyQueryPrecedence = false;
    private Boolean hasDeviceQueryPrecedence = false;
    private Boolean hasAutoAttached = false;

    private Boolean isGrouped = false;
    private Boolean autoUpdate = false;
    private Boolean autoAttach = false;

    private Boolean valueIsInteger = false;
    private Boolean valueIsNumber = false;

    private Boolean insideNotify = false;

    private String linkCtx = null;
    private String lastLinkCtx = null;
    private String linkSubsystem = null;
    private static String lastLinkSub = null;
    private String linkSrv = null;
    private String lastLinkSrv = null;
    private String linkDev = null;
    private String lastLinkDev = null;
    private String linkProp = null;
    private String lastLinkPrp = null;
    private String units = null;

    private String[] descriptionLabels = null;
    private String description = null;

    private int autoLinkCapacity = 0;
    private int precision = 0;
    protected int numAcopLinksPending = 0;
    protected int numAcopLinksInGroup = 0;

    private double resetValue = 0;
    private double readbackValue = 0;
    private String errorValue = "";
    private const String UNAVAILABLE = "<unavailable>";

    public event AcopControl.AcopLink.acopTransportUpdate update = null;
    public event PropertyChangedEventHandler PropertyChanged;

    public delegate void defaultUpdaterType(AcopControl.AcopLink lnk);
    private delegate void notifyLinkUpdateDelegate(object sender, AcopControl.AcopLink lnk);
    public delegate void acopUpdate(object sender, Link lnk);
    public delegate void acopApplySetting(object sender, EventArgs e);

    public defaultUpdaterType updater = null;

    private static String lastContextSetting = null;
    private static String lastServerSetting = null;
    private static String lastSubsystemSetting = null;
    private static String lastDeviceSetting = null;
    private static String lastPropertySetting = null;

    /// <summary>
    /// An invoker helper for cases where InvokeRequired is needed
    /// </summary>
    /// <param name="control">the parent control of this AcopLink</param>
    /// <param name="updater">is the delegate to invoke</param>
    /// <param name="sender">is the sender object</param>
    /// <param name="lnk">is the Link object</param>
    public void LinkUpdateInvoker(Control control, acopUpdate updater, object sender, Link lnk)
    {
      try
      {
        control.Invoke(new acopUpdate(updater), sender, lnk);
      }
      catch (Exception) { }
    }
    protected static AcopComponent GetInstance()
    {
      return instance;
    }
    public AcopComponent()
    {
      isDesignMode = (LicenseManager.UsageMode == LicenseUsageMode.Designtime);
      loadTransport(this);
      setNotifier(notifyLinkUpdate);
      if (IsDesignMode())
      {
        linkCtx = AcopControl.AcopLink.GetDefaultContext(LinkProtocol);
        linkSrv = AcopControl.AcopLink.GetDefaultServer(LinkProtocol);
        linkDev = AcopControl.AcopLink.GetDefaultDevice(LinkProtocol);
        linkProp = AcopControl.AcopLink.GetDefaultProperty(LinkProtocol);
        if (lastContextSetting != null) linkCtx = lastContextSetting;
        if (lastServerSetting != null) linkSrv = lastServerSetting;
        if (lastDeviceSetting != null) linkDev = lastDeviceSetting;
        if (lastPropertySetting != null) linkProp = lastPropertySetting;
      }
    }
    ~AcopComponent()
    {
      AcopControl.AcopLink.ClearAcopLinkTable(component);
    }

    protected bool IsDesignMode()
    {
      return isDesignMode;
    }

    protected void setNotifier(AcopControl.AcopLink.acopTransportUpdate notifier)
    {
      update = notifier;
    }

    protected Component GetComponentInstance()
    {
      return component;
    }
    protected void SetComponentInstance(Component control)
    {
      component = control;
    }

    protected void loadTransport(Component acopControl)
    {
      component = acopControl;
      instance = this;
    }
    void initTransport()
    {
      lastLinkSrv = null;
      lastLinkCtx = null;
      lastLinkDev = null;
      lastLinkPrp = null;
      lastLinkSub = null;
    }

    int getDefaultInterval() { return defaultInterval; }
    void setDefaultInterval(int value)
    {
      if (defaultInterval < 0) defaultInterval = 0;
      if (defaultInterval == 0)
        defaultInterval = AcopControl.AcopLink.GetDefaultInterval();
      defaultInterval = value;
    }
    object GetLinkDataOut() { return linkDataOut; }
    object GetLinkDataIn() { return linkDataIn; }
    int GetLinkFlags() { return linkFlags; }
    private String getLinkAddr()
    {
      instance = this;
      return linkAddr == null ? "" : linkAddr;
    }
    private void setLinkAddr(String adr)
    {
      instance = this;
      if (!adr.Equals(linkAddr, StringComparison.OrdinalIgnoreCase))
      {
        displaySettingsAcquired = false;
      }
      linkAddr = adr;
      checkAutoAttach();
    }
    /// <summary>
    /// Gets the integer flag representation of the input flags string
    /// </summary>
    /// <param name="flags"></param>
    /// <returns>the integer flag representation of the input flags string</returns>
    public int GetFlags(String flags)
    {
      if (flagsDict == null) flagsDict = AcopControl.AcopLink.GetFlagsList(linkProt);
      if (flagsDict == null || flags == null) return 0;
      if (!flagsDict.ContainsKey(flags)) return 0;
      int v = 0;
      flagsDict.TryGetValue(flags, out v);
      return v;
    }
    private String GetLinkFlagsString() { return linkFlagsString; }
    private void SetLinkFlagsString(String value) { linkFlagsString = value; }
    protected String getEnumLinkFlags()
    {
      instance = this;
      return GetLinkFlagsString();
    }
    protected void setEnumLinkFlags(String flags)
    {
      instance = this;
      SetLinkFlagsString(flags);
    }

    private String getLinkProt()
    {
      instance = this;
      return linkProt == null ? AcopControl.AcopLink.GetDefaultProtocol() : linkProt;
    }
    private void setLinkProt(String protocol)
    {
      instance = this;
      linkProt = protocol;
      AcopControl.AcopLink lnk = AcopControl.AcopLink.GetAcopLink(linkProt, linkAddr, linkProp, component);
      if (lnk != null) lnk.SetProtocol(linkProt);
    }

    private String getLinkCtxSetting()
    {
      instance = this;
      return this.linkCtx;
    }
    private String getFirstContext()
    {
      if (!IsDesignMode()) return null;
      if (lastLinkCtx != null) return lastLinkCtx;
      String[] lst = AcopControl.AcopLink.GetContextList(linkProt);
      if (lst != null && lst[0] != null) return lst[0];
      return "";
    }
    private String getLinkContext()
    {
      instance = this;
      return linkCtx == null ? getFirstContext() : linkCtx;
    }
    private void setLinkContext(String context)
    {
      instance = this;
      linkCtx = lastLinkCtx = context;
      linkSrv = null;
      srvStale = true; ctxStale = false;
      if (IsDesignMode())
      {
        lastContextSetting = context;
        if (initialPassDone)
        {
          lastLinkSrv = null;
          if (IsDesignMode()) setLinkServer(getFirstServer());
          linkDev = lastLinkDev = null;
          linkProp = lastLinkPrp = null;
        }
      }
    }

    private String getFirstSubsystem()
    {
      if (!IsDesignMode()) return null;
      if (!ctxStale) return linkSubsystem;
      if (linkCtx == null) linkCtx = getFirstContext();
      if (lastLinkSub != null) return lastLinkSub;
      String[] lst = AcopControl.AcopLink.GetSubSystemList(linkProt, linkCtx);
      if (lst != null && lst[0] != null) return lst[0];
      return "";
    }
    private String getLinkSubSystem()
    {
      instance = this;
      return linkSubsystem == null ? getFirstSubsystem() : linkSubsystem;
    }
    private void setLinkSubSystem(String subsystem)
    {
      instance = this;
      linkSubsystem = lastLinkSub = subsystem;
      srvStale = true; subStale = false;
    }

    private String getLinkSrvSetting()
    {
      instance = this;
      return this.linkSrv;
    }

    private String getFirstServer()
    {
      if (!IsDesignMode()) return null;
      if (!ctxStale && !srvStale && linkSrv != null) return linkSrv;
      if (linkCtx == null) linkCtx = getFirstContext();
      if (lastLinkSrv != null) return lastLinkSrv;
      String[] lst = AcopControl.AcopLink.GetServerList(linkProt, linkCtx, linkSubsystem);
      if (lst != null && lst[0] != null) return lst[0];
      return "";
    }
    private String getLinkServer()
    {
      instance = this;
      return linkSrv == null ? getFirstServer() : linkSrv;
    }
    private Boolean isDesignTimeSetServer()
    {
      if (IsDesignMode())
      {
        if (linkDev.CompareTo(UNAVAILABLE) == 0 ||
            linkProp.CompareTo(UNAVAILABLE) == 0)
          return true;
      }
      return false;
    }
    private void setLinkServer(String server)
    {
      instance = this;
      linkSrv = lastLinkSrv = server;
      srvStale = false;
      if (IsDesignMode())
      {
        if (server != null && server.Length > 0)
          lastServerSetting = server;
      }
      if (initialPassDone || isDesignTimeSetServer())
      {
        linkDev = null; lastLinkDev = null;
        linkProp = lastLinkPrp = null;
        devStale = true;
        if (IsDesignMode())
        {
          hasPropertyQueryPrecedence = false;
          hasDeviceQueryPrecedence = false;
          setLinkDevice(getFirstDevice());
          if (hasPropertyQueryPrecedence)
          {
            linkDev = lastLinkDev = null;
            linkAddr = null;
            setLinkProp(getFirstProperty());
          }
        }
      }
      if (linkCtx != null && linkDev != null)
      {
        linkAddr = AcopControl.AcopLink.GetConnectionAddress(linkProt, linkCtx, linkSubsystem, linkSrv, linkDev);
        setLinkAddr(linkAddr);
      }
      checkAutoAttach(); // in initializeComponent this is alphabetically the last thing to happen
      initialPassDone = true;
    }
    private String getFirstDevice()
    {
      if (!IsDesignMode()) return null;
      if (linkDev != null && linkDev.CompareTo(UNAVAILABLE) == 0)
      {
        linkDev = lastLinkDev = null;
      }
      if (linkProp != null && linkProp.CompareTo(UNAVAILABLE) == 0)
      {
        linkProp = lastLinkPrp = null;
      }
      if (!devStale && linkDev != null) return linkDev;
      if (lastLinkDev != null) return lastLinkDev;
      if (linkCtx != null && linkSrv != null)
      {
        if (linkProp == null && !hasPropertyQueryPrecedence)
        {
          hasPropertyQueryPrecedence = AcopControl.AcopLink.HasPropertyQueryPrecedence(linkProt, linkCtx, linkSubsystem, linkSrv, linkProp);
        }
        if (hasPropertyQueryPrecedence)
        { // first query
          prpStale = true;
          if (linkProp == null) return UNAVAILABLE;
        }
        String[] lst = AcopControl.AcopLink.GetDeviceList(linkProt, linkCtx, linkSubsystem, linkSrv, linkProp);
        if (lst != null && lst[0] != null) return lst[0];
      }
      return "";
    }
    private String getLinkDevice()
    {
      instance = this;
      String dev = linkDev;
      if (dev == null)
      {
        linkCtx = this.getLinkCtxSetting();
        linkSrv = this.getLinkSrvSetting();
        linkDev = dev = getFirstDevice();
        if (linkCtx != null && linkSrv != null)
        {
          linkAddr = AcopControl.AcopLink.GetConnectionAddress(linkProt, linkCtx, linkSubsystem, linkSrv, dev);
          setLinkAddr(linkAddr);
        }
      }
      return dev == null ? UNAVAILABLE : dev;
    }
    private void setLinkDevice(String device)
    {
      instance = this;
      linkDev = lastLinkDev = device;
      linkCtx = this.getLinkCtxSetting();
      linkSrv = this.getLinkSrvSetting();
      //MessageBox.Show("setLinkDevice: /" + linkCtx + "/" + linkSrv);
      if (linkCtx != null && linkSrv != null)
      {
        linkAddr = AcopControl.AcopLink.GetConnectionAddress(linkProt, linkCtx, linkSubsystem, linkSrv, linkDev);
        setLinkAddr(linkAddr);
        if (IsDesignMode())
        {
          lastDeviceSetting = device;
        }
      }
      devStale = false;
      if (hasDeviceQueryPrecedence && !device.Contains('*'))
      {
        prpStale = true;
        if (IsDesignMode())
        {
          linkProp = lastLinkPrp = null;
          setLinkProp(getFirstProperty());
        }
      }
    }

    private String getLinkProp()
    {
      instance = this;
      String prp = linkProp;
      if (prp == null)
      {
        linkProp = prp = getFirstProperty();
      }
      return prp == null ? UNAVAILABLE : prp;
    }
    private void setLinkProp(String prp)
    {
      instance = this;
      if (!prp.Equals(linkProp, StringComparison.OrdinalIgnoreCase))
      {
        displaySettingsAcquired = false;
      }
      linkProp = lastLinkPrp = prp;
      if (IsDesignMode())
      {
        lastPropertySetting = prp;
      }
      checkAutoAttach();
      prpStale = false;
      if (hasPropertyQueryPrecedence)
      {
        devStale = true;
        if (IsDesignMode())
        {
          linkDev = lastLinkDev = null;
          setLinkDevice(getFirstDevice());
        }
      }
    }

    private String getFirstProperty()
    {
      if (!IsDesignMode()) return null;
      if (linkProp != null && linkProp.CompareTo(UNAVAILABLE) == 0)
      {
        linkProp = lastLinkPrp = null;
      }
      if (linkDev != null && linkDev.CompareTo(UNAVAILABLE) == 0)
      {
        linkDev = lastLinkDev = null;
      }
      if (!prpStale && linkProp != null) return linkProp;
      if (linkCtx == null) linkCtx = getFirstContext();
      if (linkSrv == null) linkSrv = getFirstServer();
      if (linkDev == null) linkDev = getFirstDevice();
      if (lastLinkPrp != null) return lastLinkPrp;
      if (!hasDeviceQueryPrecedence)
      {
        hasDeviceQueryPrecedence = AcopControl.AcopLink.HasDeviceQueryPrecedence(linkProt, linkCtx, linkSubsystem, linkSrv);
      }
      String[] lst = AcopControl.AcopLink.GetPropertyList(linkProt, linkCtx, linkSubsystem, linkSrv, linkDev);
      if (lst != null && lst[0] != null) return lst[0];
      return "";
    }
    /// <summary>
    /// Returns the output data object bound to the default transport link
    /// </summary>
    /// <returns>the output data object bound to the default transport link</returns>
    public object GetLinkOutput()
    {
      AcopControl.AcopLink lnk = AcopControl.AcopLink.GetAcopLink(linkProt, linkAddr, linkProp, component);
      if (lnk == null) return null;
      return lnk.GetOutputDataObject();
    }
    /// <summary>
    /// Returns the appropriate X-Axis for the data bound to the default transport link (for use in the chart)
    /// </summary>
    /// <returns>the appropriate X-Axis for the data bound to the default transport link</returns>
    public double[] GetLinkOutputXAxis()
    {
      linkProt = getLinkProt().ToString();
      AcopControl.AcopLink lnk = AcopControl.AcopLink.GetAcopLink(linkProt, linkAddr, linkProp, component);
      if (lnk == null) return null;
      return lnk.GetOutputXAxis();
    }
    /// <summary>
    /// Returns the appropriate Y-Axis for the data bound to the default transport link (for use in the chart)
    /// </summary>
    /// <returns>the appropriate Y-Axis for the data bound to the default transport link</returns>
    public double[] GetLinkOutputYAxis()
    {
      AcopControl.AcopLink lnk = AcopControl.AcopLink.GetAcopLink(linkProt, linkAddr, linkProp, component);
      if (lnk == null) return null;
      return lnk.GetOutputYAxis();
    }
    /// <summary>
    /// Returns a list of discovered channel names for the data bound to the default transport link
    /// (In case the transport link is bound to a multi-channel array).
    /// </summary>
    /// <returns>a list of discovered channel names for the data bound to the default transport link</returns>
    public String[] GetLinkOutputChannelNames()
    {
      AcopControl.AcopLink lnk = AcopControl.AcopLink.GetAcopLink(linkProt, linkAddr, linkProp, component);
      if (lnk == null) return null;
      return lnk.GetOutputChannelNames();
    }

    /// <summary>
    /// Attaches a transport link to the link target contained in the 
    /// LinkAddress and LinkProperty fields.  Will attempt to discover 
    /// and use the default output data object
    /// </summary>
    /// <returns>A positive link ID or -1 if the link cannot be established.</returns>
    public int Attach()
    {
      int flgs = GetFlags(linkFlagsString);
      return Attach(AcopControl.AcopLink.GetOutputDataObject(linkAddr, linkProp, null, autoLinkCapacity), null, flgs, AcopControl.AcopLink.GetDefaultInterval(), null);
    }
    /// <summary>
    /// Attaches a transport link to the link target contained in the 
    /// LinkAddress and LinkProperty fields. 
    /// </summary>
    /// <param name="DataOut">A reference to the output data object which is to be filled</param>
    /// <param name="DataIn">A reference to an input data object to be sent to the link address</param>
    /// <param name="Flags">Any optional link flags</param>
    /// <param name="interval">An update interval (or timeout) in milliseconds</param>
    /// <param name="reference">Any reference which should be returned in the link's update event</param>
    /// <returns>A positive link ID or -1 if the link cannot be established.</returns>
    public int Attach(object DataOut, object DataIn, int Flags, int interval, object reference)
    {
      AcopControl.AcopLink lnk = AcopControl.AcopLink.GetAcopLink(linkProt, linkAddr, linkProp, DataOut, DataIn, Flags, component);
      if (lnk.IsAttached()) return lnk.GetLinkId();
      lnk.SetTransportUpdatedEvent(update);
      lnk.SetAcopControl(component);
      linkDataOut = DataOut; linkDataIn = DataIn; linkFlags = Flags;
      return lnk.Attach(DataOut, DataIn, Flags, interval, reference);
    }
    /// <summary>
    /// Attaches a transport link to the link target contained in the 
    /// LinkAddress and LinkProperty fields. 
    /// </summary>
    /// <param name="DataOut">A reference to the output data object which is to be filled</param>
    /// <param name="Offset">A specific offset into the output data object (array) reference. Output data will begin filling at this offset.</param>
    /// <param name="Length">The length of the output data object (array) which is to be filled in.</param>
    /// <param name="DataIn">A reference to an input data object to be sent to the link address</param>
    /// <param name="Flags">Any optional link flags</param>
    /// <param name="interval">An update interval (or timeout) in milliseconds</param>
    /// <param name="reference">Any reference which should be returned in the link's update event</param>
    /// <returns>A positive link ID or -1 if the link cannot be established.</returns>
    public int Attach(object DataOut, int Offset, int Length, object DataIn, int Flags, int interval, object reference)
    {
      AcopControl.AcopLink lnk = AcopControl.AcopLink.GetAcopLink(linkProt, linkAddr, linkProp, DataOut, DataIn, Flags, component);
      if (lnk.IsAttached()) return lnk.GetLinkId();
      lnk.SetTransportUpdatedEvent(update);
      lnk.SetAcopControl(component);
      linkDataOut = DataOut; linkDataIn = DataIn; linkFlags = Flags;
      return lnk.Attach(DataOut, Offset, Length, DataIn, Flags, interval, reference);
    }
    /// <summary>
    /// Detaches the link specifed by the link parameters from the acop control.
    /// </summary>
    /// <param name="Protocol">Control system protocol</param>
    /// <param name="Address">The link address</param>
    /// <param name="Property">The link property</param>
    /// <param name="DataOut">The output data object</param>
    /// <param name="DataIn">The input data object</param>
    /// <returns>0 upon success or -1 on error</returns>
    public int Detach(String Protocol, String Address, String Property, object DataOut, object DataIn)
    {
      AcopControl.AcopLink lnk = AcopControl.AcopLink.GetAcopLink(Protocol, Address, Property, DataOut, DataIn, linkFlags, component);
      if (lnk != null)
      {
        int rc = lnk.Detach();
        AcopControl.AcopLink.RmvAcopLink(lnk.GetConnectionKey());
        if (isGrouped && numAcopLinksPending > 0) numAcopLinksPending--;
        return rc;
      }
      return -1;
    }
    /// <summary>
    /// Detaches the link specifed by the link parameters from the acop control.
    /// </summary>
    /// <param name="Address">The link address</param>
    /// <param name="Property">The link property</param>
    /// <param name="DataOut">The output data object</param>
    /// <param name="DataIn">The input data object</param>
    /// <returns>0 upon success or -1 on error</returns>
    public int Detach(String Address, String Property, object DataOut, object DataIn)
    {
      return Detach(linkProt, Address, Property, DataOut, DataIn);
    }
    /// <summary>
    /// Detaches the link specifed by the link parameters from the acop control.
    /// </summary>
    /// <param name="Protocol">Control system protocol</param>
    /// <param name="Address">The link address</param>
    /// <param name="Property">The link property</param>
    /// <returns>0 upon success or -1 on error</returns>
    public int Detach(String Protocol, String Address, String Property)
    {
      return Detach(Protocol, Address, Property, linkDataOut, linkDataIn);
    }
    /// <summary>
    /// Detaches the link specifed by the link parameters from the acop control.
    /// </summary>
    /// <param name="Address">The link address</param>
    /// <param name="Property">The link property</param>
    /// <returns>0 upon success or -1 on error</returns>
    public int Detach(String Address, String Property)
    {
      return Detach(linkProt.ToString(), Address, Property);
    }
    /// <summary>
    /// Detaches all links attached to the acop control.
    /// </summary>
    /// <returns>0 upon success or -1 on error</returns>
    public int Detach()
    {
      AcopControl.AcopLink.ClearAcopLinkTable(component);
      return 0;
    }
    /// <summary>
    /// Executes a single synchronous transport call to the link target contained in the 
    /// LinkAddress and LinkProperty fields.  Will attempt to discover 
    /// and use the default output data object
    /// </summary>
    /// <returns>0 upon success or -1 if there is a link error.</returns>
    public int Execute()
    {
      int flgs = GetFlags(linkFlagsString);
      return Execute(AcopControl.AcopLink.GetOutputDataObject(linkAddr, linkProp), null, flgs, AcopControl.AcopLink.GetDefaultInterval());
    }
    /// <summary>
    /// Executes a single synchronous transport call to the link target contained in the 
    /// LinkAddress and LinkProperty fields.  
    /// </summary>
    /// <param name="DataOut">A reference to the output data object which is to be filled</param>
    /// <param name="DataIn">A reference to an input data object to be sent to the link address</param>
    /// <param name="Flags">Any optional link flags</param>
    /// <param name="timeout">timeout in milliseconds</param>
    /// <returns>0 upon success or -1 if there is a link error.</returns>
    public int Execute(object DataOut, object DataIn, int Flags, int timeout)
    {
      AcopControl.AcopLink lnk = AcopControl.AcopLink.GetAcopLink(linkProt, linkAddr, linkProp, DataOut, DataIn, Flags, component);
      linkDataOut = DataOut; linkDataIn = DataIn; linkFlags = Flags;
      return lnk.Execute(DataOut, DataIn, Flags, timeout);
    }
    /// <summary>
    /// Executes a single synchronous transport call to the link target contained in the 
    /// LinkAddress and LinkProperty fields.  
    /// </summary>
    /// <param name="DataOut">A reference to the output data object which is to be filled</param>
    /// <param name="Offset">A specific offset into the output data object (array) reference. Output data will begin filling at this offset.</param>
    /// <param name="Length">The length of the output data object (array) which is to be filled in.</param>
    /// <param name="DataIn">A reference to an input data object to be sent to the link address</param>
    /// <param name="Flags">Any optional link flags</param>
    /// <param name="timeout">timeout in milliseconds</param>
    /// <returns>0 upon success or -1 if there is a link error.</returns>
    public int Execute(object DataOut, int Offset, int Length, object DataIn, int Flags, int timeout)
    {
      AcopControl.AcopLink lnk = AcopControl.AcopLink.GetAcopLink(linkProt, linkAddr, linkProp, DataOut, DataIn, Flags, component);
      linkDataOut = DataOut; linkDataIn = DataIn; linkFlags = Flags;
      return lnk.Execute(DataOut, Offset, Length, DataIn, Flags, timeout);
    }
    protected Boolean needsAutoAttach()
    {
      return autoAttach && !hasAutoAttached;
    }
    private void checkAutoAttach()
    {
      if (!IsDesignMode())
      {
        if (autoAttach && !hasAutoAttached)
        {
          if (linkAddr != null && linkProp != null)
          {
            AcopTransport.DisplaySettings dsp =
              acquireAndApplyDisplaySettings(linkAddr, linkProp, ref descriptionLabels, true);
            if (dsp.SettingsStale()) return; // didn't get the settings ... 
            Attach();
            hasAutoAttached = true;
          }
        }
      }
    }

    protected AcopTransport.DisplaySettings acquireAndApplyDisplaySettings(String address, String property, ref String[] labels, Boolean acquireLabels)
    {
      if (!isGrouped && displaySettingsAcquired && (linkDspSettings != null)) return linkDspSettings;
      AcopTransport.DisplaySettings dsp = AcopControl.AcopLink.GetDisplaySettings(address, property);
      if (dsp != null)
      {
        if (units == null || units.Length == 0)
        {
          units = dsp.GetYUnits();
        }
        if (description == null || description.Length == 0)
        {
          description = dsp.GetDescription();
        }
      }
      String unitsString = ((units == null || units.Length == 0) ? "" : " (" + units + ") ");
      if (unitsString.Length > 0 && !description.Contains(unitsString))
      {
        description += unitsString;
      }
      if (!dsp.SettingsStale()) displaySettingsAcquired = true;
      linkDspSettings = dsp;
      return dsp;
    }
    /// <summary>
    /// Gets the current transport link status as a string
    /// </summary>
    /// <returns>The current transport link status as a string</returns>
    public String GetStatusString()
    {
      AcopControl.AcopLink lnk = AcopControl.AcopLink.GetAcopLink(linkProt, linkAddr, linkProp, linkDataOut, linkDataIn, linkFlags, component);
      if (lnk != null) return lnk.GetStatusString();
      return "no link exists for this acop";
    }
    /// <summary>
    /// Gets the last delivered link output completion length.
    /// </summary>
    /// <returns>The last delivered link output completion length.</returns>
    public int GetOutputCompletionLength()
    {
      AcopControl.AcopLink lnk = AcopControl.AcopLink.GetAcopLink(linkProt, linkAddr, linkProp, linkDataOut, linkDataIn, linkFlags, component);
      if (lnk != null) return lnk.GetOutputCompletionLength();
      return 0;
    }
    /// <summary>
    /// Get the last link timestamp (as a UTC double).
    /// </summary>
    /// <returns>The last link timestamp (as a UTC double).</returns>
    public double GetTimeStamp()
    {
      AcopControl.AcopLink lnk = AcopControl.AcopLink.GetAcopLink(linkProt, linkAddr, linkProp, linkDataOut, linkDataIn, linkFlags, component);
      if (lnk != null) return lnk.GetTimeStamp();
      return 0;
    }
    /// <summary>
    /// Gets the last data tag (if any) associated with the output data.
    /// </summary>
    /// <returns>The last data tag (if any) associated with the output data.</returns>
    public int GetDataTag()
    {
      AcopControl.AcopLink lnk = AcopControl.AcopLink.GetAcopLink(linkProt, linkAddr, linkProp, linkDataOut, linkDataIn, linkFlags, component);
      if (lnk != null) return lnk.GetDataTag();
      return 0;
    }
    /// <summary>
    /// Gets the current link status as an integer value
    /// </summary>
    /// <returns>The current link status as an integer value</returns>
    public int GetStatus()
    {
      AcopControl.AcopLink lnk = AcopControl.AcopLink.GetAcopLink(linkProt, linkAddr, linkProp, linkDataOut, linkDataIn, linkFlags, component);
      if (lnk != null) return lnk.GetStatus();
      return -1;
    }

    protected void setAutoLinkCapacity(int value) { autoLinkCapacity = value; }
    protected int getAutoLinkCapacity() { return autoLinkCapacity; }

    protected void setLinkGrouped(bool value) { isGrouped = value; }
    protected bool getLinkGrouped() { return isGrouped; }

    protected void setAutoUpdate(bool value) { autoUpdate = value; }
    protected bool getAutoUpdate() { return autoUpdate; }

    protected void setAutoAttach(bool value) { autoAttach = value; }
    protected bool getAutoAttach() { return autoAttach; }

    protected String getUnits() { return units; }
    protected void setUnits(String value)
    {
      if (value != null) units = value;
    }

    protected int getPrecision() { return precision; }
    protected void setPrecision(int value)
    {
      if (value >= 0) precision = value;
    }

    private double GetReadbackValue()
    {
      return readbackValue;
    }
    private void SetReadbackValue(double value)
    {
      readbackValue = value;
    }
    protected double GetResetValue()
    {
      return resetValue;
    }
    protected Boolean getValueIsInteger()
    {
      return valueIsInteger;
    }
    protected Boolean getValueIsNumber()
    {
      return valueIsNumber;
    }
    protected void SetResetValue(double value)
    {
      resetValue = value;
    }
    private String getErrorValue() { return errorValue; }
    private void setErrorValue(String value)
    {
      if (value == null) value = "";
      errorValue = value;
    }
    protected void applyErrorValue(AcopControl.AcopLink lnk)
    {
      if (errorValue.Length == 0) return;
      if (isGrouped)
      {
        AcopControl.AcopLink[] lst = AcopControl.AcopLink.GetAcopLinkTable(component);
        int gsize = lst.GetLength(0);
        foreach (AcopControl.AcopLink alnk in lst)
        {
          if (alnk.GetStatus() == 0) continue;
          alnk.FillDataObjectFromString(errorValue);
        }
      }
      else
      {
        if (lnk.GetStatus() == 0) return;
        lnk.FillDataObjectFromString(errorValue);
      }
    }
    protected Type getValueType()
    {
      return valueType;
    }
    protected String getLabelString(double v, int width)
    {
      if (valueIsInteger) return ((int)v).ToString("D");
      return v.ToString(getLabelFormat(v, width));
    }
    private String getLabelFormat(double v, int width)
    {
      if (valueIsInteger)
      {
        return "D";
      }
      if (precision == 0)
      {
        return "G" + width;
      }
      return "F" + precision;
    }

    protected Boolean ValueIsNumber()
    {
      switch (Type.GetTypeCode(valueType))
      {
        case TypeCode.Double: return true;
        case TypeCode.Int16: return true;
        case TypeCode.Int32: return true;
        case TypeCode.Int64: return true;
        case TypeCode.Single: return true;
        case TypeCode.UInt16: return true;
        case TypeCode.UInt32: return true;
        case TypeCode.UInt64: return true;
      }
      return false;
    }
    protected double getValueAsDouble(AcopControl.AcopLink lnk)
    {
      double v = 0.0;
      try
      {
        Object o = lnk.GetOutputDataObject();
        valueType = (o is Array) ? ((Array)o).GetType().GetElementType() : o.GetType();
        if (o is Boolean[])
        {
          return ((Boolean[])o)[0] ? 1.0 : 0.0;
        }
        if (o is double[]) v = ((double[])o)[0];
        else if (o is float[]) v = (double)((float[])o)[0];
        else if (o is Int16[])
        {
          v = (double)((Int16[])o)[0];
          valueIsInteger = true;
        }
        else if (o is Int32[])
        {
          v = (double)((Int32[])o)[0];
          valueIsInteger = true;
        }
        else if (o is Int64[])
        {
          v = (double)((Int64[])o)[0];
          valueIsInteger = true;
        }
        else if (o is Double) v = (Double)o;
        else if (o is Single) v = (double)((Single)o);
        else if (o is Int16)
        {
          v = (double)((Int16)o);
          valueIsInteger = true;
        }
        else if (o is Int32)
        {
          v = (double)((Int32)o);
          valueIsInteger = true;
        }
        else if (o is Int64)
        {
          v = (double)((Int64)o);
          valueIsInteger = true;
        }
        valueIsNumber = ValueIsNumber();
      }
      catch (Exception x)
      {
        valueIsNumber = false;
      }
      return v;
    }

    private String getDescription() { return description == null ? "" : description; }
    private void setDescription(String description) { this.description = description; }

    static double getUtcNow()
    {
      return (DateTime.Now - utcThen).TotalSeconds;
    }

    private Boolean AreNotificationsPending()
    {
      if (isGrouped)
      {
        AcopControl.AcopLink[] lst = AcopControl.AcopLink.GetAcopLinkTable(component);
        int gsize = lst.GetLength(0);
        if (numAcopLinksPending == 0)
        { // 1st callback: set the notification parameters ...
          numAcopLinksPending = numAcopLinksInGroup = gsize;
        }
        if (gsize != numAcopLinksInGroup)
        { // something has changed ...
          numAcopLinksPending -= (numAcopLinksInGroup - gsize);
          numAcopLinksInGroup = gsize;
        }
        if (--numAcopLinksPending > 0) return true;
        numAcopLinksPending = 0;
      }
      return false;
    }
    protected void SetDefaultUpdater(defaultUpdaterType updater)
    {
      this.updater = updater;
    }
    protected void notifyLinkUpdate(object sender, AcopControl.AcopLink lnk)
    {
      //foreach (Control c in getControlsList())
      //  if (c.InvokeRequired)
      //  {
      //    Invoke(new notifyLinkUpdateDelegate(notifyLinkUpdate), sender, lnk);
      //    return;
      //  }
      lock (notifyMutex)
      {
        if (insideNotify) return;
        insideNotify = true;
        applyErrorValue(lnk);
        if (AreNotificationsPending())
        {
          insideNotify = false;
          return;
        }
        if (acopLinkUpdate != null)
        { // there is an event delegate : fire it
          acopLinkUpdate(this, Link.GetLink(lnk));
        }
        else
        { // check for automatic display updates:
          if (updater != null) updater(lnk);
        }
        insideNotify = false;
      }
    }

    #region Converter
    public class LinkProtocolConverter : StringConverter
    {
      public override bool GetStandardValuesSupported(ITypeDescriptorContext context)
      { //true means show a combobox
        return true;
      }
      public override bool GetStandardValuesExclusive(ITypeDescriptorContext context)
      { //true will limit to list. false will show the list, but allow free-form entry
        return false;
      }
      public override
      System.ComponentModel.TypeConverter.StandardValuesCollection GetStandardValues(ITypeDescriptorContext context)
      {
        return new StandardValuesCollection(AcopControl.AcopLink.GetProtocolList());
      }
    }

    public class LinkContextConverter : StringConverter
    {
      private AcopComponent myAcop = GetInstance();
      public override bool GetStandardValuesSupported(ITypeDescriptorContext context) { return true; }
      public override bool GetStandardValuesExclusive(ITypeDescriptorContext context) { return true; }
      public override
      System.ComponentModel.TypeConverter.StandardValuesCollection GetStandardValues(ITypeDescriptorContext context)
      {
        return new StandardValuesCollection(AcopControl.AcopLink.GetContextList(myAcop.linkProt));
      }
    }

    public class LinkSubSystemConverter : StringConverter
    {
      private AcopComponent myAcop = null;
      public override bool GetStandardValuesSupported(ITypeDescriptorContext context) { return true; }
      public override bool GetStandardValuesExclusive(ITypeDescriptorContext context) { return true; }
      public override
      System.ComponentModel.TypeConverter.StandardValuesCollection GetStandardValues(ITypeDescriptorContext context)
      {
        myAcop = AcopComponent.GetInstance();
        return new StandardValuesCollection(AcopControl.AcopLink.GetSubSystemList(myAcop.linkProt, myAcop.getLinkContext()));
      }
    }

    public class LinkServerConverter : StringConverter
    {
      private AcopComponent myAcop = null;
      public override bool GetStandardValuesSupported(ITypeDescriptorContext context) { return true; }
      public override bool GetStandardValuesExclusive(ITypeDescriptorContext context) { return true; }
      public override
      System.ComponentModel.TypeConverter.StandardValuesCollection GetStandardValues(ITypeDescriptorContext context)
      {
        myAcop = GetInstance();
        return new StandardValuesCollection(AcopControl.AcopLink.GetServerList(myAcop.linkProt, myAcop.getLinkContext(), myAcop.getLinkSubSystem()));
      }
    }

    public class LinkDeviceConverter : StringConverter
    {
      private AcopComponent myAcop = null;
      public override bool GetStandardValuesSupported(ITypeDescriptorContext context) { return true; }
      public override bool GetStandardValuesExclusive(ITypeDescriptorContext context) { return false; }
      public override
      System.ComponentModel.TypeConverter.StandardValuesCollection GetStandardValues(ITypeDescriptorContext context)
      {
        myAcop = GetInstance();
        myAcop.hasPropertyQueryPrecedence = AcopControl.AcopLink.HasPropertyQueryPrecedence(myAcop.linkProt, myAcop.getLinkContext(), myAcop.getLinkSubSystem(), myAcop.getLinkServer());
        String prp = myAcop.hasPropertyQueryPrecedence ? myAcop.linkProp : null;
        return new StandardValuesCollection(AcopControl.AcopLink.GetDeviceList(myAcop.linkProt, myAcop.getLinkContext(), myAcop.getLinkSubSystem(), myAcop.getLinkServer(), prp));
      }
    }

    public class LinkPropertyConverter : StringConverter
    {
      private AcopComponent myAcop = null;
      public override bool GetStandardValuesSupported(ITypeDescriptorContext context) { return true; }
      public override bool GetStandardValuesExclusive(ITypeDescriptorContext context) { return false; }
      public override
      System.ComponentModel.TypeConverter.StandardValuesCollection GetStandardValues(ITypeDescriptorContext context)
      {
        myAcop = GetInstance();
        myAcop.hasDeviceQueryPrecedence = AcopControl.AcopLink.HasDeviceQueryPrecedence(myAcop.linkProt, myAcop.getLinkContext(), myAcop.getLinkSubSystem(), myAcop.getLinkServer());
        return new StandardValuesCollection(AcopControl.AcopLink.GetPropertyList(myAcop.linkProt, myAcop.getLinkContext(), myAcop.getLinkSubSystem(), myAcop.getLinkServer(), myAcop.getLinkDevice()));
      }
    }
    public class LinkFlagsConverter : StringConverter
    {
      private AcopComponent myAcop = null;
      public override bool GetStandardValuesSupported(ITypeDescriptorContext context) { return true; }
      public override bool GetStandardValuesExclusive(ITypeDescriptorContext context) { return false; }
      public override
      System.ComponentModel.TypeConverter.StandardValuesCollection GetStandardValues(ITypeDescriptorContext context)
      {
        myAcop = GetInstance();
        myAcop.flagsDict = AcopControl.AcopLink.GetFlagsList(myAcop.linkProt);
        String[] sv = myAcop.flagsDict.Keys.ToArray();
        return new StandardValuesCollection(sv);
      }
    }

    #endregion

    #region Properties
    [Category("Acop.Transport")]
    public event acopUpdate acopLinkUpdate;

    /// <summary>
    /// if set to true, instructs ACOP to automatically monitor the data provided by LinkAddress and LinkProperty
    /// </summary>
    [Category("Acop.Transport")]
    [Description("if set to true, instructs ACOP to automatically monitor the data provided by LinkAddress and LinkProperty")]
    public bool AutoLinkAttach { get { return getAutoAttach(); } set { setAutoAttach(value); } }
    /// <summary>
    /// if set to true, then multiple assigned links will fire a single receive event when all links have completed their update cycle
    /// </summary>
    [Category("Acop.Transport")]
    [Description("if set to true, then multiple assigned links will fire a single receive event when all links have completed their update cycle")]
    public bool LinkGrouped { get { return getLinkGrouped(); } set { setLinkGrouped(value); } }
    /// <summary>
    /// if non-zero, provides the requested data size to use when AutoLinkUpdate is set to true
    /// </summary>
    [Category("Acop.Transport")]
    [Description("if non-zero, provides the requested data size to use when AutoLinkUpdate is set to true")]
    public int AutoLinkCapacity { get { return autoLinkCapacity; } set { setAutoLinkCapacity(value); } }
    /// <summary>
    /// the design-time browseable/editable link property (suppliments the link address)
    /// </summary>
    [Category("Acop.Transport")]
    [Description("the design-time browseable/editable link property (suppliments the link address)")]
    [Browsable(true)]
    [DefaultValue("NSERVERS")]
    [TypeConverter(typeof(LinkPropertyConverter))]
    public String LinkProperty { get { return getLinkProp(); } set { setLinkProp(value); } }
    /// <summary>
    /// the design-time browseable/editable link device (part of the link address)
    /// </summary>
    [Category("Acop.Transport")]
    [Description("the design-time browseable/editable link device (part of the link address)")]
    [Browsable(true)]
    [DefaultValue("DEFAULT")]
    [TypeConverter(typeof(LinkDeviceConverter))]
    public String LinkDevice { get { return getLinkDevice(); } set { setLinkDevice(value); } }
    /// <summary>
    /// identifies the transport link target. The target end point is given by the LinkAddress and the LinkProperty.
    /// </summary>
    [Category("Acop.Transport")]
    [Description("Identifies the transport link target. The target end point is given by the LinkAddress and the LinkProperty.")]
    public String LinkAddress { get { return getLinkAddr(); } set { setLinkAddr(value); } }
    /// <summary>
    /// the design-time browseable link server (part of the link address)
    /// </summary>
    [Category("Acop.Transport")]
    [Description("the design-time browseable link server (part of the link address)")]
    [Browsable(true)]
    [DefaultValue("ENS")]
    [TypeConverter(typeof(LinkServerConverter))]
    public String LinkServer { get { return getLinkServer(); } set { setLinkServer(value); } }
    /// <summary>
    /// the design-time browseable link subsystem
    /// </summary>
    [Category("Acop.Transport")]
    [Description("the design-time browseable link subsystem")]
    [Browsable(true)]
    [DefaultValue("ALL")]
    [TypeConverter(typeof(LinkSubSystemConverter))]
    public String LinkSubSystem { get { return getLinkSubSystem(); } set { setLinkSubSystem(value); } }
    /// <summary>
    /// the design-time browseable link context (part of the link address)
    /// </summary>
    [Category("Acop.Transport")]
    [Description("the design-time browseable link context (part of the link address)")]
    [Browsable(true)]
    [DefaultValue("SITE")]
    [TypeConverter(typeof(LinkContextConverter))]
    public String LinkContext { get { return getLinkContext(); } set { setLinkContext(value); } }
    /// <summary>
    /// the link control system protocol
    /// </summary>
    [Category("Acop.Transport")]
    [Description("the link control system protocol")]
    [Browsable(true)]
    [DefaultValue("SIMULATE")]
    [TypeConverter(typeof(LinkProtocolConverter))]
    public String LinkProtocol { get { return getLinkProt(); } set { setLinkProt(value); } }
    /// <summary>
    /// an optional design-time browseable flags set which can be applied to the underlying link
    /// </summary>
    [Category("Acop.Transport")]
    [Description("an optional design-time browseable flags set which can be applied to the underlying link")]
    [Browsable(true)]
    [DefaultValue("NONE")]
    [TypeConverter(typeof(LinkFlagsConverter))]
    public String LinkFlags { get { return getEnumLinkFlags(); } set { setEnumLinkFlags(value); } }
    /// <summary>
    /// the default monitoring interval in msec
    /// </summary>
    [Category("Acop.Transport")]
    [DefaultValue(1000)]
    [Description("the default monitoring interval in msec")]
    public int DefaultMonitorInterval { get { return getDefaultInterval(); } set { setDefaultInterval(value); } }
    #endregion
  }
}
