﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace Acop
{
  [ToolboxItem(true)]
  [ToolboxBitmap(typeof(ProgressBar))]
  public partial class AcopGauge : AcopUserComponent
  {
#   pragma warning disable 1591
    private Label descLabel = new Label();
    private Label valueLabel = new Label();
    private Label unitLabel = new Label();

    private Color barBack_Color = Color.White;
    private Color barFore_Color = Color.FromArgb(100, 100, 100);
    private Color needle_Color = Color.DarkBlue;

    private Color warnBack_Color = Color.Yellow;
    private Color errorBack_Color = Color.Red;

    private Pen scalePen;
    private SolidBrush barBackBrush;
    private SolidBrush barForeBrush;
    private SolidBrush needleBrush;

    private double minValue = 0;
    private double maxValue = 100;
    private double thresholdHigh = MAX_THRESHOLD;
    private double thresholdLow = -MAX_THRESHOLD;
    private double refValue = -MAX_THRESHOLD;
    private double actValue = 0;
    private double oldValue = 0;

    private int bWidth;
    private int bHeight;

    private int xStart = 20;
    private int barTicksValue = -1;
    private int oldTicks = 10;
    private int lastLinkStatus = 0;

    private RegionOfInterest roi1;
    private RegionOfInterest roi2;

    public delegate void regionChangedEventHandler(object sender, RegionChangedEventArgs e);
    public enum regionChangedEnum
    {
      Warn1_IN_MIN,
      Warn1_OUT_MIN,
      Warn1_IN_MAX,
      Warn1_OUT_MAX,
      Warn2_IN_MIN,
      Warn2_OUT_MIN,
      Warn2_IN_MAX,
      Warn2_OUT_MAX
    }

    public enum orientationEnum
    {
      Horizontal,
      Vertical
    }
    private orientationEnum ori = orientationEnum.Horizontal;

    public AcopGauge()
    {
      InitializeComponent();
      roi1 = new RegionOfInterest(this);
      roi2 = new RegionOfInterest(this);
      SetDefaultUpdater(defaultUpdate);
      setControlsList(new Control[] { this, valueLabel, descLabel, unitLabel }, true);
      applyToolTipText(valueLabel, "the current value");
      SetLateInitializerProlog(initTransportInfo);
      StartLateInitialization();
    }

    private void InitializeComponent()
    {
      this.Size = new Size(275, 95);
      this.MinimumSize = new Size(100, 90);
      this.DoubleBuffered = true;

      descLabel.Location = new Point(1, 5);
      descLabel.Font = new Font(FontFamily.GenericSansSerif, 8.25f, FontStyle.Bold);
      descLabel.Text = "description";
      descLabel.AutoSize = true;
      descLabel.Anchor = (AnchorStyles.Left | AnchorStyles.Top | AnchorStyles.Right);

      valueLabel.Text = "0.00";
      valueLabel.Location = new Point(0, this.Height - 30);
      valueLabel.AutoSize = false;
      valueLabel.TextAlign = ContentAlignment.MiddleRight;
      valueLabel.Size = new Size(this.Width / 2, 30);
      valueLabel.Anchor = AnchorStyles.Bottom;

      unitLabel.Text = "unit";
      unitLabel.Location = new Point(this.Width / 2, this.Height - 30);
      unitLabel.AutoSize = false;
      unitLabel.TextAlign = ContentAlignment.MiddleLeft;
      unitLabel.Size = new Size(this.Width / 2, 30);
      unitLabel.Anchor = AnchorStyles.Bottom;

      barBackBrush = new SolidBrush(barBack_Color);
      barForeBrush = new SolidBrush(barFore_Color);
      needleBrush = new SolidBrush(needle_Color);

      this.Controls.Add(valueLabel);
      this.Controls.Add(unitLabel);
      this.Controls.Add(descLabel);
    }

    private void setBarBackColor(Color value)
    {
      barBack_Color = value;
      barBackBrush.Color = barBack_Color;
      Refresh();
    }

    private void setBarForeColor(Color value)
    {
      barFore_Color = value;
      barForeBrush.Color = barFore_Color;
      Refresh();
    }

    private void setNeedleColor(Color value)
    {
      needle_Color = value;
      needleBrush.Color = needle_Color;
      Refresh();
    }

    private void setBarTicks(int value)
    {
      barTicksValue = value;
      Refresh();
    }

    private void setThresholdHigh(double value)
    {
      if (value != MAX_THRESHOLD)
      {
        thresholdHigh = value;
        maxValue = value;
        Refresh();
      }
    }
    private void setThresholdLow(double value)
    {
      if (value != -MAX_THRESHOLD)
      {
        thresholdLow = value;
        minValue = value;
        Refresh();
      }
    }

    private void setRefValue(double value)
    {
      if (value != -MAX_THRESHOLD)
      {
        refValue = value;
        Refresh();
      }
    }

    private void setOrientation(orientationEnum value)
    {
      ori = value;
      switch (ori)
      {
        case orientationEnum.Horizontal:
          this.Size = new Size(275, 95);
          this.MinimumSize = new Size(100, 90);
          break;
        case orientationEnum.Vertical:
          this.Size = new Size(95, 275);
          this.MinimumSize = new Size(70, 150);
          break;
      }
      Refresh();
    }

    private void setBarValue(double value)
    {
      actValue = value;
      valueLabel.Text = value.ToString();
      Refresh();
    }

    public void applyErrorStatus(String sts, Boolean applyToTTT)
    {
      valueLabel.Text = sts;
      valueLabel.ForeColor = Color.Red;
      if (applyToTTT) applyToolTipText(sts);
    }

    private void defaultUpdate(AcopControl.AcopLink lnk)
    {
      if (getAutoUpdate())
      {
        int lnkSts = lnk.GetStatus();
        if (lnkSts != 0 && ErrorValue.Length == 0)
        {
          applyErrorStatus(lnk.GetStatusString(), lnkSts != lastLinkStatus);
        }
        else
        {
          if (lnkSts != lastLinkStatus) applyToolTipText(null);
          double v = getValueAsDouble(lnk);
          String txt;
          if (getValueAvailableAsNumber())
          {
            txt = getLabelString(v, 6);
          }
          else
          {
            String[] txtlines = lnk.GetOutputDataAsString();
            txt = txtlines[0];
          }
          if (lnk.GetDisplayId() < 0)
          {
            if (!displaySettingsAcquired) initTransportInfo();
            lnk.SetDisplayId(1);
            SetResetValue(v);
            applyToolTipText(descLabel, Description);
            valueLabel.Text = getLabelString(v, 6);
          }
          valueLabel.Text = txt;
          try
          {
            actValue = Convert.ToDouble(txt);
          }
          catch(FormatException)
          {

          }         
          Refresh();
        }
      }
    }

    private void initTransportInfo()
    {
      if (!IsDesignMode())
      {
        String[] labels = null;
        AcopTransport.DisplaySettings dsp = null;
        if (!AutoLinkAttach && !AutoLinkUpdate)
        {
          descLabel.Text = Description;
          unitLabel.Text = Units;
          displaySettingsAcquired = true;
          return;
        }
        dsp = acquireAndApplyDisplaySettings(LinkAddress, LinkProperty, ref labels, false);
        if (dsp == null || dsp.SettingsStale()) return;
        displaySettingsAcquired = true;
        unitLabel.Text = dsp.GetYUnits();
        if (thresholdHigh == MAX_THRESHOLD) maxValue = dsp.GetYMaximum();
        if (thresholdLow == -MAX_THRESHOLD) minValue = dsp.GetYMinimum();
        if (maxValue == minValue) maxValue = minValue + 100.0;
        descLabel.Text = (getIsMCA() && getDisplayAsMCA()) ? getMCADescription() : getDescription();
      }
    }

    protected override void OnPaint(PaintEventArgs e)
    {
      base.OnPaint(e);
      
      Bitmap gaugeBar;
      if (ori == orientationEnum.Horizontal)
      {
        if (this.Width < 20 || this.Height < 70) return;
        gaugeBar = new Bitmap(this.Width - 10, this.Height - 60, e.Graphics);
      }
      else
      {
        if (this.Width < 70 || this.Height < 150) return;
        gaugeBar = new Bitmap(this.Height - 60, this.Width - 10, e.Graphics);
      }
      bWidth = gaugeBar.Width;
      bHeight = gaugeBar.Height;
      Graphics gi = Graphics.FromImage(gaugeBar);
      gi.SmoothingMode = SmoothingMode.HighQuality;

      Rectangle barRect = new Rectangle(0, 0, bWidth - 1, bHeight - 1);
      scalePen = new Pen(barFore_Color, 1);
      gi.FillRectangle(barBackBrush, barRect);
      gi.DrawRectangle(scalePen, barRect);
      drawRegionsOfInterest(gi, minValue, maxValue);
      drawBarScale(gi, minValue, maxValue, barTicksValue);
      drawBarNeedle(gi, actValue, minValue, maxValue);
      if (ori == orientationEnum.Vertical) gaugeBar.RotateFlip(RotateFlipType.Rotate270FlipNone);
      e.Graphics.DrawImageUnscaled(gaugeBar, 5, 25);
    }

    private void drawBarScale(Graphics g, double minValue, double maxValue, int ticks)
    {
      if (ticks == 0) return;
      float x = xStart;
      float xOld = xStart;
      int newTicks = oldTicks;
      SizeF stringSize = new SizeF();
      StringFormat sFormat = new StringFormat();
      if (ticks == -1)
      {
        int check = (int)(bWidth - (2 * x)) / 50;
        if (check % 2 == 0)
        {
          newTicks = check;
          oldTicks = newTicks;
        }
        else if (check == 1)
        {
          newTicks = 1;
          oldTicks = newTicks;
        }
        else
        {
          newTicks = oldTicks;
        }
      }
      else
      {
        newTicks = ticks;
      }

      float gap = (bWidth - (2 * x)) / newTicks;
      for (int i = 0; i < newTicks + 1; i++)
      {
        x = xStart + (i * gap);
        double number = i * Math.Abs(maxValue-minValue) / newTicks + minValue;
        stringSize = g.MeasureString(getLabelString(number, 6), Font);
        if (ori == orientationEnum.Vertical)
        {
          sFormat.FormatFlags = StringFormatFlags.DirectionVertical;
          g.DrawString(getLabelString(number, 6), Font, barForeBrush, x - stringSize.Height/2, bHeight/2 - stringSize.Width, sFormat);
        }
        else
        {
          g.DrawString(getLabelString(number, 6), Font, barForeBrush, x - stringSize.Width / 2, bHeight / 2 - Font.Height);
        }        
        g.DrawLine(scalePen, x, bHeight / 2, x, bHeight - 10);
        drawBarSubTicks(g, xOld, x);
        xOld = x;
      }
      g.DrawLine(scalePen, 20, bHeight - 10, x, bHeight - 10);
    }

    private void drawBarSubTicks(Graphics g, float p1, float p2)
    {
      if(p2 - p1 < 15)
      {
        return;
      }
      float gap = (p2 - p1) / 5;
      for(int i = 0; i < 5; i++)
      {
        float x = p1 + (i * gap);
        g.DrawLine(scalePen, x, bHeight / 2 + (bHeight / 8), x, bHeight - 10);
      }
    }

    private void drawBarNeedle(Graphics g, double value, double minValue, double maxValue)
    {
      FillMode fm = FillMode.Winding;
      PointF p1, p2, p3;

      double x = calcPosFromValue(value, minValue, maxValue);
      if (x > bWidth) x = bWidth-1;
      if (x < 0) x = 0;
      p1 = new PointF((float)x, bHeight / 2);
      p2 = new PointF((float)x + 4, bHeight);
      p3 = new PointF((float)x - 4, bHeight);
      PointF[] points = { p1, p2, p3 };

      g.FillPolygon(needleBrush, points, fm);
      if(refValue != -MAX_THRESHOLD)
      {
        PointF p1ref, p2ref, p3ref;

        double refX = calcPosFromValue(refValue, minValue, maxValue);
        if (refX > bWidth) refX = bWidth-1;
        if (refX < 0) refX = 0;
        p1ref = new PointF((float)refX, bHeight / 2 - 10);
        p2ref = new PointF((float)refX + 4, 0);
        p3ref = new PointF((float)refX - 4, 0);
        PointF[] refPoints = { p1ref, p2ref, p3ref };
        g.FillPolygon(barForeBrush, refPoints, fm);
      }
    }

    private void drawRegionsOfInterest(Graphics g, double minValue, double maxValue)
    {
      if (!roi1.Visible && !roi2.Visible) return;
      if (roi1.Visible)
      {
        Brush regBrush = new SolidBrush(roi1.RegionBackColor);
        double xMin = 0;
        double xMax = bWidth - 2;
        if (roi1.RegionMinValue > minValue)
        {
          xMin = calcPosFromValue(roi1.RegionMinValue, minValue, maxValue);
        }
        if (roi1.RegionMaxValue < maxValue)
        {
          xMax = calcPosFromValue(roi1.RegionMaxValue, minValue, maxValue);
        }
        RectangleF regRect = new RectangleF((float)xMin, 1, (float)xMax - (float)xMin, bHeight - 3);
        g.FillRectangle(regBrush, regRect);
      }
      if (roi2.Visible)
      {
        Brush regBrush = new SolidBrush(roi2.RegionBackColor);
        double xMin = 0;
        double xMax = bWidth - 1;
        if (roi2.RegionMinValue > minValue)
        {
          xMin = calcPosFromValue(roi2.RegionMinValue, minValue, maxValue);
        }
        if (roi2.RegionMaxValue < maxValue)
        {
          xMax = calcPosFromValue(roi2.RegionMaxValue, minValue, maxValue);
        }
        RectangleF regRect = new RectangleF((float)xMin, 1, (float)xMax - (float)xMin, bHeight - 3);
        g.FillRectangle(regBrush, regRect);
      }
      checkRegionChanged(actValue, oldValue);
      oldValue = actValue;
    }

    private double calcPosFromValue(double value, double min, double max)
    {
      double pos = Math.Sqrt(Math.Pow(value - min, 2));
      if (value < min) pos = pos * -1;
      double x = xStart + ((bWidth - (2 * xStart)) * pos / Math.Abs(max - min));
      return x;
    }

    private void checkRegionChanged(double act, double old)
    {
      if (act == old) return;
      if (roi1.Visible)
      {
        if (act > old)
        {
          if (act > roi1.RegionMinValue && old <= roi1.RegionMinValue) OnRegionChanged(new RegionChangedEventArgs(regionChangedEnum.Warn1_IN_MIN));
          if (act > roi1.RegionMaxValue && old <= roi1.RegionMaxValue) OnRegionChanged(new RegionChangedEventArgs(regionChangedEnum.Warn1_OUT_MAX));
        }
        if (act < old)
        {
          if (act < roi1.RegionMinValue && old >= roi1.RegionMinValue) OnRegionChanged(new RegionChangedEventArgs(regionChangedEnum.Warn1_OUT_MIN));
          if (act < roi1.RegionMaxValue && old >= roi1.RegionMaxValue) OnRegionChanged(new RegionChangedEventArgs(regionChangedEnum.Warn1_IN_MAX));
        }
      }
      if (roi2.Visible)
      {
        if (act > old)
        {

          if (act > roi2.RegionMinValue && old <= roi2.RegionMinValue) OnRegionChanged(new RegionChangedEventArgs(regionChangedEnum.Warn2_IN_MIN));
          if (act > roi2.RegionMaxValue && old <= roi2.RegionMaxValue) OnRegionChanged(new RegionChangedEventArgs(regionChangedEnum.Warn2_OUT_MAX));
        }
        if (act < old)
        {

          if (act < roi2.RegionMinValue && old >= roi2.RegionMinValue) OnRegionChanged(new RegionChangedEventArgs(regionChangedEnum.Warn2_OUT_MIN));
          if (act < roi2.RegionMaxValue && old >= roi2.RegionMaxValue) OnRegionChanged(new RegionChangedEventArgs(regionChangedEnum.Warn2_IN_MAX));
        }
      }
    }

    protected virtual void OnRegionChanged(RegionChangedEventArgs e)
    {
      WarnRegionChanged?.Invoke(this, e);
    }

    #region Properties
    [Category("Acop.Display")]
    [Description("Fires when the value entered or leaved a warn region")]
    public event regionChangedEventHandler WarnRegionChanged;

    [Category("Acop.Display")]
    [Description("The bar background color")]
    [DefaultValue(typeof(Color), "0xFFFFFF")]
    public Color BarBackColor { get { return barBack_Color; } set { setBarBackColor(value); } }
    [Category("Acop.Display")]
    [Description("The bar foreground color")]
    [DefaultValue(typeof(Color), "0x646464")]
    public Color BarForeColor { get { return barFore_Color; } set { setBarForeColor(value); } }
    [Category("Acop.Display")]
    [Description("The needle color")]
    [DefaultValue(typeof(Color), "0x00008B")]
    public Color NeedleColor { get { return needle_Color; } set { setNeedleColor(value); } }
    [Category("Acop.Display")]
    [Description("Set the gauger orientation")]
    [DefaultValue(orientationEnum.Horizontal)]
    public orientationEnum Orientation { get { return ori; } set { setOrientation(value); } }
    [Category("Acop.Display")]
    [Description("The maximum setting value")]
    [DefaultValue(MAX_THRESHOLD)]
    public double MaximumValue { get { return thresholdHigh; } set { setThresholdHigh(value); } }
    [Category("Acop.Display")]
    [Description("The minimum setting value")]
    [DefaultValue(-MAX_THRESHOLD)]
    public double MinimumValue { get { return thresholdLow; } set { setThresholdLow(value); } }
    [Category("Acop.Display")]
    [Description("Set a reference value to be shown in the bar")]
    [DefaultValue(-MAX_THRESHOLD)]
    public double ReferenceValue { get { return refValue; } set { setRefValue(value); } }
    [Category("Acop.Display")]
    [Description("Set the tick marks (-1 : auto; 0 : no marks and scale)")]
    [DefaultValue(-1)]
    public int BarTicks { get { return barTicksValue; } set { setBarTicks(value); } }
    [Category("Acop.Display")]
    [Description("The bar value")]
    [DefaultValue(0)]
    public double BarValue { get { return actValue; } set { setBarValue(value); } }
    [Category("Acop.Display")]
    [Description("define a different backcolored area on the bar")]
    [DesignerSerializationVisibility(DesignerSerializationVisibility.Content)]
    public RegionOfInterest Region1Warn { get { return roi1; } set { if (value != null) roi1 = value; } }
    [Category("Acop.Display")]
    [Description("Define a different backcolored area on the bar")]
    [DesignerSerializationVisibility(DesignerSerializationVisibility.Content)]
    public RegionOfInterest Region2Warn { get { return roi2; } set { if (value != null) roi2 = value; } }

    [Category("Acop.Transport")]
    [Browsable(false), EditorBrowsable(EditorBrowsableState.Never)]
    public new Boolean LinkGrouped { get; set; }

    [Category("Acop.Transport")]
    [Browsable(false), EditorBrowsable(EditorBrowsableState.Never)]
    public new event acopApplySetting acopApply;

    [Category("Acop.Transport")]
    [Browsable(false), EditorBrowsable(EditorBrowsableState.Never)]
    public new ApplyEventSettingType ApplyEventOrder { get; set; }

    public override Font Font { get { return base.Font; } set { base.Font = value; } }
    #endregion
  }

  [TypeConverter(typeof(RegionOfInterestConverter))]
  public class RegionOfInterest
  {
#   pragma warning disable 1591
    private Boolean isRegionVisible;
    private Color regionColor;
    private double regionMin;
    private double regionMax;

    private AcopGauge gauger;

    public RegionOfInterest(AcopGauge ag)
    {
      isRegionVisible = false;
      regionColor = Color.Yellow;
      regionMin = 0;
      regionMax = 100;

      gauger = ag;
    }

    private void setRegionVisible(Boolean value)
    {
      isRegionVisible = value;
      gauger.Refresh();
    }
    private void setRegionColor(Color value)
    {
      regionColor = value;
      gauger.Refresh();
    }
    private void setMinValue(double value)
    {
      regionMin = value;
      gauger.Refresh();
    }
    private void setMaxValue(double value)
    {
      regionMax = value;
      gauger.Refresh();
    }

    [Category("Acop.Display")]
    [Description("Show region on bar")]
    [DefaultValue(false)]
    public Boolean Visible { get { return isRegionVisible; } set { setRegionVisible(value); } }
    [Category("Acop.Display")]
    [Description("Set the region backcolor")]
    [DefaultValue(typeof(Color), "0xFFFF00")]
    public Color RegionBackColor { get { return regionColor; } set { setRegionColor(value); } }
    [Category("Acop.Display")]
    [Description("Set the minimum region value")]
    [DefaultValue(0.0)]
    public double RegionMinValue { get { return regionMin; } set { setMinValue(value); } }
    [Category("Acop.Display")]
    [Description("Set the maximum region value")]
    [DefaultValue(100.0)]
    public double RegionMaxValue { get { return regionMax; } set { setMaxValue(value); } }
  }

  public class RegionOfInterestConverter : TypeConverter
  {
#   pragma warning disable 1591
    public override bool GetPropertiesSupported(ITypeDescriptorContext context)
    {
      return true;
    }
    public override PropertyDescriptorCollection GetProperties(ITypeDescriptorContext context, object value, Attribute[] attributes)
    {
      return TypeDescriptor.GetProperties(typeof(RegionOfInterest));
    }
  }

  public class RegionChangedEventArgs : EventArgs
  {
#   pragma warning disable 1591
    public AcopGauge.regionChangedEnum param { get; set; }

    public RegionChangedEventArgs(AcopGauge.regionChangedEnum param)
    {
      this.param = param;
    }
  }
}
