﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace Acop
{
  [ToolboxItem(true)]
  [ToolboxBitmap(typeof(TextBox))]
  public partial class AcopInputBox : AcopUserComponent
  {
#   pragma warning disable 1591
    public AcopInputBox()
    {
      InitializeComponent();
      allowDragAndDrop(new Control[] { this, ReadbackLabel });
      SetDefaultUpdater(defaultUpdate);
      SetDefaultApplier(defaultApplier);
      setControlsList(new Control[] { Label, ReadbackLabel, BarLine, DescLabel }, true);
      applyToolTipText(LockCheck, "if checked: apply an exclusive (only this application is allowed) access lock to the set value");
      applyToolTipText(label1, "the current readback value");
      SetLateInitializerProlog(initTransportInfo);
      StartLateInitialization();
    }
    private Color labelInSyncForeColor = Color.DarkBlue;
    private Color labelOutOfSyncForeColor = Color.DarkRed;
    private Boolean showDescription = true;
    private void setShowDescription(Boolean value)
    {
      this.SuspendLayout();
      DescLabel.Visible = value;
      BarLine.Visible = value;
      this.ResumeLayout();
      showDescription = value;
    }
    double getValueFromLabel()
    {
      double v = 0;
      try
      {
        v = getValueIsInteger() ?
          (double)(Int32.Parse(Label.Text)) :
          Double.Parse(Label.Text);
      }
      catch (Exception)
      {
        v = 0;
      }
      return v;
    }
    public void applyErrorStatus(String sts, Boolean applyToTTT)
    {
      ReadbackLabel.Text = sts;
      if (applyToTTT) applyToolTipText(sts);
      Label.ForeColor = Color.DarkRed;
    }
    private void initTransportInfo()
    {
      if (!IsDesignMode())
      {
        String[] labels = null;
        AcopTransport.DisplaySettings dsp = null;
        if (!AutoLinkAttach && !AutoLinkUpdate)
        {
          DescLabel.Text = Description;
          displaySettingsAcquired = true;
          return;
        }
        dsp = acquireAndApplyDisplaySettings(LinkAddress, LinkProperty, ref labels, false);
        if (dsp == null || dsp.SettingsStale()) return;
        displaySettingsAcquired = true;
        DescLabel.Text = Description;
        if (Description.Length > BarLine.Text.Length / 5)
        {
          StringBuilder sb = new StringBuilder();
          for (int i = 0; i < (Description.Length * 5) - BarLine.Text.Length; i++)
            sb.Append("--");
          BarLine.Text += sb.ToString();
        }
        applyToolTipText(DescLabel, Description);
      }
    }
    private int lastLinkStatus = 0;
    private void defaultUpdate(AcopControl.AcopLink lnk)
    {
      if (!lateInitializationDone) return;
      if (getAutoUpdate())
      {
        int lnkSts = lnk.GetStatus();
        if (lnkSts != 0 && ErrorValue.Length == 0)
        {
          applyErrorStatus(lnk.GetStatusString(), lnkSts != lastLinkStatus);
        }
        else
        {
          if (lnkSts != lastLinkStatus) applyToolTipText(null);
          double v = getValueAsDouble(lnk);
          String txt;
          if (getValueIsNumber())
          {
            txt = getLabelString(v, 6);
          }
          else
          {
            String[] txtlines = lnk.GetOutputDataAsString();
            txt = txtlines[0];
          }
          if (lnk.GetDisplayId() < 0)
          {
            if (!displaySettingsAcquired) initTransportInfo();
            lnk.SetDisplayId(1);
            SetResetValue(v);
            Label.Text = txt;
          }
          SetReadbackValue(v);
          ReadbackLabel.Text = txt;
          Label.ForeColor = (ReadbackLabel.Text.CompareTo(Label.Text) != 0) ? labelOutOfSyncForeColor : labelInSyncForeColor;
        }
      }
    }
    private int defaultApplier(AcopControl.AcopLink lnk)
    {
      int cc = lnk.GetStatus();
      if (getValueIsNumber() && cc == 0)
      {
        double v = getValueFromLabel();
        Label.Text = getLabelString(v, 6);
      }
      if (cc != 0) Label.Text = lnk.GetStatusString();
      else ReadbackLabel.Text = Label.Text;
      return cc;
    }
    private void applySettingFromLabelEvent(object sender)
    {
      Type type = getValueType(); // as long as getValueAsDouble() is called in the updater this will be something ...
      if (type == null) type = getValueType(Label.Text);
      object inpt = AcopControl.AcopLink.GetDataObjectFromString(type, Label.Text);
      AcopControl.AcopLink lnk = AcopControl.AcopLink.GetAcopLink(LinkProtocol, LinkAddress, LinkProperty, null, inpt, 0, this);
      applySetting(sender, inpt, lnk);
    }
    private void LabelKeyDown(object sender, KeyEventArgs e)
    {
      if (e.KeyCode == Keys.Enter)
      {
        applySettingFromLabelEvent(sender);
      }
    }
    private void LabelDoubleClicked(object sender, EventArgs e)
    {
      applySettingFromLabelEvent(sender);
    }
    private void ApplyLockClick(object sender, EventArgs e)
    {
      applyLock(LockCheck);
    }
    private void InputBoxDrop(object sender, DragEventArgs e)
    {
      try
      {
        Label.Text = (String)e.Data.GetData(DataFormats.StringFormat);
      }
      catch (Exception) { } // ignore ...
    }
    private void InputBoxDragEnter(object sender, DragEventArgs e)
    {
      e.Effect = DragDropEffects.All;
    }
    private Font inputBoxFont = null;
    private Font getInputBoxFont()
    {
      if (inputBoxFont == null) inputBoxFont = this.Label.Font;
      return inputBoxFont;
    }
    private void setInputBoxFont(Font font)
    {
      inputBoxFont = font;
      this.Label.Font = font;
      this.DescLabel.Font = font;
      this.label1.Font = new System.Drawing.Font(label1.Font.Name, font.Size,label1.Font.Style);
      this.ReadbackLabel.Font = font;
      this.LockCheck.Font = font;
    }
    [Category("Acop.Display")]
    [Description("Font to use int the input box")]
    public Font DisplayFont { get { return getInputBoxFont(); } set { setInputBoxFont(value); } }

    private void setLabelInSyncForeColor(Color value)
    {
      labelInSyncForeColor = value;
    }
    [Category("Acop.Display")]
    [Description("ForeColor to use if setting is synchronized with the current readback value")]
    public Color LabelInSyncForeColor { get { return labelInSyncForeColor; } set { setLabelInSyncForeColor(value); } }
    private void setLabelOutOfSyncForeColor(Color value)
    {
      labelOutOfSyncForeColor = value;
    }
    [Category("Acop.Display")]
    [Description("ForeColor to use if setting is not synchronized with the current readback value")]
    public Color LabelOutOfSyncForeColor { get { return labelOutOfSyncForeColor; } set { setLabelOutOfSyncForeColor(value); } }

    private Color readbackForeColor = Color.Black;
    private Color referenceForeColor = Color.Black;
    private void setReadbackForeColor(Color value)
    {
      readbackForeColor = value;
      ReadbackLabel.ForeColor = value;
    }
    private void setReferenceForeColor(Color value)
    {
      referenceForeColor = value;
    }
    [Category("Acop.Display")]
    [Description("Default ForeColor to use for the Readback label")]
    public Color ReadbackForeColor { get { return readbackForeColor; } set { setReadbackForeColor(value); } }

    [Category("Acop.Display")]
    [Description("Show the description as a control header")]
    public Boolean ShowDescription { get { return showDescription; } set { setShowDescription(value); } }

    [Category("Acop.Display")]
    [Browsable(false), EditorBrowsable(EditorBrowsableState.Never)]
    public new Boolean DisplayAsMultiChannel { get; set; }

    [Category("Acop.Transport")]
    [Browsable(false), EditorBrowsable(EditorBrowsableState.Never)]
    public new Boolean LinkInputValue { get; set; }

    [Category("Acop.Transport")]
    [Browsable(false), EditorBrowsable(EditorBrowsableState.Never)]
    public new Boolean LinkGrouped { get; set; }

  }
}
