﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace Acop
{
  [ToolboxItem(true)]
  [ToolboxBitmap(typeof(Label))]
  public partial class AcopLabel : AcopUserComponent
  {
#   pragma warning disable 1591
    private Boolean showSplitter = false;
    private Boolean showDescription = true;

    private int captionWidth = 0;
    private int lastLinkStatus = 0;
    private const int AT_CHANNEL = 0x10;

    private double thresholdHigh = MAX_THRESHOLD;
    private double thresholdLow = -MAX_THRESHOLD;

    private String thresholdMatch = "";

    private Color captionForeColor = Color.Black;
    private Color captionBackColor = DefaultBackColor;
    private Color labelForeColor = Color.Black;
    private Color labelBackColor = DefaultBackColor;
    private Color labelLowWarnForeColor = Color.Empty;
    private Color labelHiWarnForeColor = Color.Empty;
    private Color labelLowWarnBackColor = Color.Empty;
    private Color labelHiWarnBackColor = Color.Empty;

    private BorderStyle labelBorderStyle = BorderStyle.None;

    private StringBuilder sbCap = new StringBuilder();
    private StringBuilder sbLab = new StringBuilder();

    ContentAlignment labelAlignment = ContentAlignment.MiddleCenter;
    ContentAlignment captionAlignment = ContentAlignment.MiddleCenter;

    public AcopLabel()
    {
      InitializeComponent();
      allowDragAndDrop(new Control[] { this, Label, Caption });
      SetDefaultUpdater(defaultUpdate);
      setControlsList(new Control[] { Label, Caption }, true);
      SetLateInitializerProlog(initTransportInfo);
      Label.BackColor = labelBackColor;
      Label.ForeColor = labelForeColor;
      Caption.ForeColor = captionForeColor;
      Caption.BackColor = captionBackColor;
      StartLateInitialization();
      blink = blinkColors;
    }
    private String displayValue = "";
    private void setDisplayValue(String value)
    {
      displayValue = value;
      Label.Text = value;
      Refresh();
    }

    private Boolean isWarningCondition = false;
    private void setBlinkColors(Color fill, Color text)
    {
      Label.BackColor = fill;
      Label.ForeColor = text;
      //BackColor = fill;
    }
    private void resetBlinkColors()
    {
      resetColors();
    }
    private Boolean isBlinkCycle = false;
    void blinkColors(Boolean blink)
    {
      if (blinkMode == BlinkMode.off)
      {
        resetBlinkColors();
        return;
      }
      if (blinkMode == BlinkMode.onWarning && !isWarningCondition)
      {
        resetBlinkColors();
        return;
      }
      isBlinkCycle = blink;
      if (blink)
      {
        Color fc =
          blinkerIntensity > 0 ?
          ControlPaint.Light(currentBackColor, blinkerIntensity) :
          ControlPaint.Dark(currentBackColor, -blinkerIntensity);
        Color tc =
          blinkerIntensity > 0 ?
          ControlPaint.Dark(currentForeColor, blinkerIntensity) :
          ControlPaint.Light(currentForeColor, -blinkerIntensity);
        setBlinkColors(fc, tc);
      }
      else
      {
        setBlinkColors(currentBackColor, currentForeColor);
      }
    }
    private String GetLabelString(AcopControl.AcopLink lnk, Boolean ismca)
    {
      if (displayValue != null && displayValue.Length > 0) return displayValue;
      if (lnk == null || lnk.GetLinkId() < 0) return "";
      String[] txt = null;
      double v = getValueAsDouble(lnk);
      if (getValueIsNumber() && !ismca)
      {
        String s = getLabelString(v, 6);
        txt = new String[] { s };
      }
      else
      {
        txt = lnk.GetOutputDataAsString();
      }
      StringBuilder sb = new StringBuilder();
      if (txt != null)
      {
        for (int i = 0; i < txt.Length; i++)
        {
          if (!checkThresholds(lnk)) checkStringMatch(txt[i]);
          sb.AppendLine(txt[i]);
        }
        return sb.ToString();
      }
      return "";
    }
    private void initTransportInfo()
    {
      if (!IsDesignMode())
      {
        String[] labels = null;
        AcopTransport.DisplaySettings dsp = null;
        if (!AutoLinkAttach && !AutoLinkUpdate)
        {
          Caption.Text = Description;
          displaySettingsAcquired = true;
          return;
        }
        if (!getLinkGrouped())
        {
          dsp = acquireAndApplyDisplaySettings(LinkAddress, LinkProperty, ref labels, false);
          if (dsp == null || dsp.SettingsStale()) return;
          displaySettingsAcquired = true;
          if (thresholdHigh == MAX_THRESHOLD) thresholdHigh = dsp.GetYMaximum();
          if (thresholdLow == -MAX_THRESHOLD) thresholdLow = dsp.GetYMinimum();
          Caption.Text = (getIsMCA() && getDisplayAsMCA()) ? getMCADescription() : getDescription();
        }
        else
        {
          StringBuilder sb = new StringBuilder();
          AcopControl.AcopLink[] lst = AcopControl.AcopLink.GetAcopLinkTable(this);
          for (int i = 0; i < lst.Length; i++)
          {
            dsp = acquireAndApplyDisplaySettings(lst[i].GetAddress(), lst[i].GetProperty(), ref labels, false);
            if (dsp == null) continue;
            displaySettingsAcquired = true;
            sb.AppendLine(dsp.GetDescription());
          }
          Caption.Text = sb.ToString();
        }
      }
    }
    private void checkLabelCaption(String[] names)
    {
      if (names != null)
      {
        StringBuilder sb = new StringBuilder();
        for (int i = 0; i < names.Length; i++) sb.AppendLine(names[i] + " ");
        String txt = sb.ToString();
        String dsc = getMCADescription();
        if (dsc == null || !dsc.Equals(txt, StringComparison.CurrentCultureIgnoreCase))
        {
          Caption.Text = txt;
        }
      }
    }
    private void defaultUpdate(AcopControl.AcopLink lnk)
    {
      if (!lateInitializationDone) return;
      if (getAutoUpdate())
      {
        int lnkSts = lnk.GetStatus();
        if (lnkSts != 0 && ErrorValue.Length == 0)
        {
          applyErrorStatus(lnk.GetStatusString(), lnkSts != lastLinkStatus);
        }
        else
        {
          if (lnkSts != lastLinkStatus) applyToolTipText(null);
          Boolean ismca = getIsMCA() && getDisplayAsMCA();
          if (lnk.GetDisplayId() < 0)
          {
            if (!displaySettingsAcquired) initTransportInfo();
            if (ismca)
            { // last minute discovery ?
              checkLabelCaption(lnk.GetOutputChannelNames());
            }
            lnk.SetDisplayId(1);
          }
          String labStr = "";
          if (!getLinkGrouped())
          {
            labStr = GetLabelString(lnk, ismca);
          }
          else
          {
            AcopControl.AcopLink[] lst = AcopControl.AcopLink.GetAcopLinkTable(this);
            for (int i = 0; i < lst.Length; i++)
            {
              labStr += GetLabelString(lst[i], false);
            }
          }
          Label.Text = labStr;
        }
      }
    }
    public Boolean needResetColors = false;
    public void applyErrorStatus(String sts, Boolean applyToTTT)
    {
      Label.Text = sts;
      if (applyToTTT) applyToolTipText(sts);
      Label.ForeColor = Color.White;
      Label.BackColor = Color.DarkRed;
      needResetColors = true;
    }
    private Color currentBackColor = Color.Empty;
    private Color currentForeColor = Color.Empty;
    private void applyLowWarning()
    {
      if (isBlinkCycle) return; // only apply in an off cycle
      if (labelLowWarnBackColor != Color.Empty)
      {
        Label.BackColor = labelLowWarnBackColor;
      }
      else
      {
        Label.BackColor = BackColor;
      }
      if (labelLowWarnForeColor != Color.Empty)
      {
        Label.ForeColor = labelLowWarnForeColor;
      }
      else
      {
        Label.ForeColor = ForeColor;
      }
      currentBackColor = Label.BackColor;
      currentForeColor = Label.ForeColor;
      isWarningCondition = true;
    }
    private void applyHiWarning()
    {
      if (isBlinkCycle) return; // only apply in an off cycle
      if (labelHiWarnBackColor != Color.Empty)
      {
        Label.BackColor = labelHiWarnBackColor;
      }
      else
      {
        Label.BackColor = BackColor;
      }
      if (labelHiWarnForeColor != Color.Empty)
      {
        Label.ForeColor = labelHiWarnForeColor;
      }
      else
      {
        Label.ForeColor = ForeColor;
      }
      currentBackColor = Label.BackColor;
      currentForeColor = Label.ForeColor;
      isWarningCondition = true;
    }
    private void checkStringMatch(String txt)
    {
      if (txt == null) return;
      if (thresholdMatch.Length == 0 && !needResetColors) return;
      if (!TextMatches(thresholdMatch, txt))
      {
        applyHiWarning();
      }
      else
      {
        resetColors();
      }
    }

    private void setLabelForeColor(Color value)
    {
      labelForeColor = value;
      Label.ForeColor = value;
      currentForeColor = Label.ForeColor;
    }
    private void setLabelBackColor(Color value)
    {
      labelBackColor = value;
      Label.BackColor = value;
      currentBackColor = Label.BackColor;
    }
    private void resetColors()
    {
      Label.BackColor = labelBackColor;
      Label.ForeColor = labelForeColor;
      currentBackColor = Label.BackColor;
      currentForeColor = Label.ForeColor;
      isWarningCondition = false;
      needResetColors = false;
    }
    private Boolean checkThresholds(AcopControl.AcopLink lnk)
    {
      try
      {
        double v;
        Object o = lnk.GetOutputDataObject();
        if (o is Boolean[])
        {
          checkStringMatch(((Boolean[])o)[0].ToString());
          return true;
        }
        if (o is double[]) v = ((double[])o)[0];
        else if (o is float[]) v = (double)((float[])o)[0];
        else if (o is Int16[]) v = (double)((Int16[])o)[0];
        else if (o is Int32[]) v = (double)((Int32[])o)[0];
        else if (o is Int64[]) v = (double)((Int64[])o)[0];
        else if (o is Double) v = (Double)o;
        else if (o is Single) v = (double)((Single)o);
        else if (o is Int16) v = (double)((Int16)o);
        else if (o is Int32) v = (double)((Int32)o);
        else if (o is Int64) v = (double)((Int64)o);
        else return false;
        if (v < thresholdLow)
        {
          applyLowWarning();
        }
        else if (v > thresholdHigh)
        {
          applyHiWarning();
        }
        else
        {
          resetColors();
        }
        return true;
      }
      catch (Exception)
      {
      }
      return false;
    }
    private void setThresholdHigh(double value)
    {
      if (value != MAX_THRESHOLD) thresholdHigh = value;
    }
    private void setThresholdLow(double value)
    {
      if (value != -MAX_THRESHOLD) thresholdLow = value;
    }
    private void setLabelBorderStyle(BorderStyle value)
    {
      labelBorderStyle = value;
      Label.BorderStyle = value;
    }
    private void setCaptionWidth(int value)
    {
      if (value < 2) value = 0;
      if (value == captionWidth) return;
      captionWidth = value;
      resizeLabel();
      Refresh();
    }
    private void setCaptionAligment(ContentAlignment value)
    {
      Caption.TextAlign = captionAlignment = value;
      Refresh();
    }
    private void setLabelAligment(ContentAlignment value)
    {
      Label.TextAlign = labelAlignment = value;
      Refresh();
    }
    private void setShowSplitter(Boolean value)
    {
      showSplitter = value;
      resizeLabel();
      Refresh();
    }

    private void setShowDescription(Boolean value)
    {
      if (!value)
      {
        ShowSplitter = false;
        Caption.Visible = false;
      }
      else
      {
        Caption.Visible = true;
      }
      showDescription = value;
      resizeLabel();
      Refresh();
    }

    private void setLabelLowWarnForeColor(Color value)
    {
      labelLowWarnForeColor = value;
    }

    private void setLabelHiWarnForeColor(Color value)
    {
      labelHiWarnForeColor = value;
    }

    private void setLabelLowWarnBackColor(Color value)
    {
      labelLowWarnBackColor = value;
    }

    private void setLabelHiWarnBackColor(Color value)
    {
      labelHiWarnBackColor = value;
    }

    private void setCaptionColor(Color value)
    {
      Caption.ForeColor = value;
      captionForeColor = value;
    }
    private void setCaptionBackColor(Color value)
    {
      captionBackColor = value;
      Caption.BackColor = value;
    }

    [Category("Acop.Transport")]
    [ToolboxItem(false)]
    [Browsable(false), EditorBrowsable(EditorBrowsableState.Never)]
    public new event acopApplySetting acopApply;

    [Category("Acop.Transport")]
    [Browsable(false), EditorBrowsable(EditorBrowsableState.Never)]
    public new ApplyEventSettingType ApplyEventOrder { get; set; }

    #region Properties
    /// <summary>
    /// if non-empty and the link data is non-numerical, then the string provided is used for comparisons with the monitored data
    /// </summary>
    [Category("Acop.Display")]
    [Description("if non-empty and the link data is non-numerical, then the string provided is used for comparisons with the monitored data")]
    public String ThresholdMatch { get { return thresholdMatch; } set { thresholdMatch = value; } }
    /// <summary>
    /// if the link data is numerical, then the threshold provided is used for comparisons with the monitored data
    /// </summary>
    [Category("Acop.Display")]
    [Description("if the link data is numerical, then the threshold provided is used for comparisons with the monitored data")]
    public double ThresholdHighWarn { get { return thresholdHigh; } set { setThresholdHigh(value); } }
    /// <summary>
    /// if the link data is numerical, then the threshold provided is used for comparisons with the monitored data
    /// </summary>
    [Category("Acop.Display")]
    [Description("if the link data is numerical, then the threshold provided is used for comparisons with the monitored data")]
    public double ThresholdLowWarn { get { return thresholdLow; } set { setThresholdLow(value); } }
    /// <summary>
    /// The border style of the embedded label text
    /// </summary>
    [Category("Acop.Display")]
    [Description("The border style of the embedded label text")]
    public BorderStyle LabelBorderStyle { get { return labelBorderStyle; } set { setLabelBorderStyle(value); } }
    /// <summary>
    /// if non-zero, provides the desired caption width (if ShowDescription is true).  Else the width of the description text is used
    /// </summary>
    [Category("Acop.Display")]
    [Description("if non-zero, provides the desired caption width (if ShowDescription is true).  Else the width of the description text is used")]
    public int CaptionWidth { get { return captionWidth; } set { setCaptionWidth(value); } }
    /// <summary>
    /// provides the Caption text aligment
    /// </summary>
    [Category("Acop.Display")]
    [Description("provides the Caption text aligment")]
    public ContentAlignment CaptionAlignment { get { return captionAlignment; } set { setCaptionAligment(value); } }
    /// <summary>
    /// provides the Lable text aligment
    /// </summary>
    [Category("Acop.Display")]
    [Description("provides the Lable text aligment")]
    public ContentAlignment LabelAlignment { get { return labelAlignment; } set { setLabelAligment(value); } }
    /// <summary>
    /// if true, then the splitter between Caption and Label will be visible (if ShowDescription is true)
    /// </summary>
    [Category("Acop.Display")]
    [Description("if true, then the splitter between Caption and Label will be visible (if ShowDescription is true)")]
    public Boolean ShowSplitter { get { return showSplitter; } set { setShowSplitter(value); } }
    /// <summary>
    /// if non-zero, provides the desired splitter width (if ShowDescription and ShowSplitter are both true)
    /// </summary>
    [Category("Acop.Display")]
    [Description("if non-zero, provides the desired splitter width (if ShowDescription and ShowSplitter are both true)")]
    public int SplitterWidth { get { return splitterWidth; } set { setSplitterWidth(value); } }
    /// <summary>
    /// Color to use for the splitter (when visible)
    /// </summary>
    [Category("Acop.Display")]
    [Description("Color to use for the splitter (when visible)")]
    public Color SplitterColor { get { return splitterColor; } set { setSplitterColor(value); } }
    /// <summary>
    /// if true, then both the Label and its Caption will be visible
    /// </summary>
    [Category("Acop.Display")]
    [Description("if true, then both the Label and its Caption will be visible")]
    public Boolean ShowDescription { get { return showDescription; } set { setShowDescription(value); } }
    /// <summary>
    /// ForeColor to use if number value is less than ThresholdLowWarn or if string value equals ThresholdMatch
    /// </summary>
    [Category("Acop.Display")]
    [Description("ForeColor to use if number value is less than ThresholdLowWarn or if string value equals ThresholdMatch")]
    public Color LabelLowWarnForeColor { get { return labelLowWarnForeColor; } set { setLabelLowWarnForeColor(value); } }
    /// <summary>
    /// ForeColor to use if number value is greater than ThresholdHiWarn or if string value does not equal ThresholdMatch
    /// </summary>
    [Category("Acop.Display")]
    [Description("ForeColor to use if number value is greater than ThresholdHiWarn or if string value does not equal ThresholdMatch")]
    public Color LabelHiWarnForeColor { get { return labelHiWarnForeColor; } set { setLabelHiWarnForeColor(value); } }
    /// <summary>
    /// BackColor to use if number value is less than ThresholdLowWarn or if string value equals ThresholdMatch
    /// </summary>
    [Category("Acop.Display")]
    [Description("BackColor to use if number value is less than ThresholdLowWarn or if string value equals ThresholdMatch")]
    public Color LabelLowWarnBackColor { get { return labelLowWarnBackColor; } set { setLabelLowWarnBackColor(value); } }
    /// <summary>
    /// BackColor to use if number value is greater than ThresholdHiWarn or if string value does not equal ThresholdMatch
    /// </summary>
    [Category("Acop.Display")]
    [Description("BackColor to use if number value is greater than ThresholdHiWarn or if string value does not equal ThresholdMatch")]
    public Color LabelHiWarnBackColor { get { return labelHiWarnBackColor; } set { setLabelHiWarnBackColor(value); } }
    /// <summary>
    /// Default ForeColor to use for the Caption
    /// </summary>
    [Category("Acop.Display")]
    [Description("Default ForeColor to use for the Caption")]
    public Color CaptionForeColor { get { return captionForeColor; } set { setCaptionColor(value); } }
    /// <summary>
    /// Default BackColor to use for the Caption
    /// </summary>
    [Category("Acop.Display")]
    [Description("Default BackColor to use for the Caption")]
    public Color CaptionBackColor { get { return captionBackColor; } set { setCaptionBackColor(value); } }
    /// <summary>
    /// Default ForeColor to use for the value label
    /// </summary>
    [Category("Acop.Display")]
    [Description("Default ForeColor to use for the value label")]
    public Color LabelForeColor { get { return labelForeColor; } set { setLabelForeColor(value); } }
    /// <summary>
    /// Default BackColor to use for the value label
    /// </summary>
    [Category("Acop.Display")]
    [Description("Default BackColor to use for the value label")]
    public Color LabelBackColor { get { return labelBackColor; } set { setLabelBackColor(value); } }
    /// <summary>
    /// if non-empty then the displayed value upon update will be the value given
    /// </summary>
    [Category("Acop.Display")]
    [Description("if non-empty then the displayed value upon update will be the value given")]
    public String DisplayValue { get { return displayValue; } set { setDisplayValue(value); } }
    /// <summary>
    /// set the blink mode for the label
    /// </summary>
    [Category("Acop.Display")]
    [Description("set the blink mode for the label")]
    public BlinkMode Blink { get { return blinkMode; } set { setBlinkMode(value); } }
    /// <summary>
    /// set the blink intensity for the label (up to 100% lighter or -100% darker)
    /// </summary>
    [Category("Acop.Display")]
    [Description("set the blink intensity for the label (up to 100% lighter or -100% darker)")]
    public Int32 BlinkIntensity { get { return blinkerIntensity; } set { setBlinkerIntensity(value); } }
    /// <summary>
    /// set the blink interval (msec) for the label
    /// </summary>
    [Category("Acop.Display")]
    [Description("set the blink interval (msec) for the label")]
    public Int32 BlinkInterval { get { return blinkerInterval; } set { setBlinkerInterval(value); } }
    /// <summary>
    /// set the blink slot for the label (0 up to the number of slots)
    /// </summary>
    [Category("Acop.Display")]
    [Description("set the blink slot for the label (0 up to the number of slots)")]
    public Int32 BlinkSlot { get { return blinkerSlot; } set { setBlinkerSlot(value); } }
    /// <summary>
    /// set the number of blink slots for the label (at least 2)
    /// </summary>
    [Category("Acop.Display")]
    [Description("set the number of blink slots for the label (at least 2)")]
    public Int32 NumberBlinkSlots { get { return numberBlinkSlots; } set { setNumberBlinkerSlots(value); } }

    #endregion
    private void resizeLabel()
    {
      SuspendLayout();
      int w = this.Size.Width;
      int sw = showSplitter ? splitterWidth : 0;
      if (Caption.Visible)
      {
        w -= sw;
        w /= 2;
        if (captionWidth > 0) w = captionWidth;
        Caption.Width = w;
        Label.Location = new Point(w+sw,0);
        w = Size.Width - w - sw;
      }
      else
      {
        Label.Location = new Point(0, 0);
      }
      Label.Width = w;
      ResumeLayout();
    }
    private void AcopLabel_Resize(object sender, EventArgs e)
    {
      resizeLabel();
    }
    private int splitterWidth = 2;
    private void setSplitterWidth(int value)
    {
      if (value < 2) value = 2;
      if (value == splitterWidth) return;
      splitterWidth = value;
      resizeLabel();
      Refresh();
    }
    private Color splitterColor = Color.Black;
    private Brush splitterBrush = Brushes.Black;
    private void setSplitterColor(Color value)
    {
      if (value == splitterColor) return;
      splitterColor = value;
      splitterBrush = new SolidBrush(value);
      Refresh();
    }
    private void AcopLabel_Paint(object sender, PaintEventArgs e)
    {
      if (!showDescription || !showSplitter) return;
      Graphics g = e.Graphics;
      int w = this.Size.Width;
      int x = Caption.Width;
      Rectangle r = new Rectangle(x, 0, splitterWidth, this.Size.Height);
      g.FillRectangle(splitterBrush, r);
    }
  }
}
