﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using System.Drawing.Drawing2D;

namespace Acop
{
  [ToolboxItem(false)]
  //[ToolboxBitmap(typeof(Label))]
  public partial class AcopShape : AcopUserComponent
  {
#   pragma warning disable 1591
    private double thresholdHigh = MAX_THRESHOLD;
    private double thresholdLow = -MAX_THRESHOLD;

    private String thresholdMatch = "";

    private Color lowWarnTextColor = Color.Empty;
    private Color hiWarnTextColor = Color.Empty;
    private Color lowWarnFillColor = Color.Empty;
    private Color hiWarnFillColor = Color.Empty;

    private Color borderColor = Color.Black;
    private Color fillColor = Color.LightGray;
    private Color fillColorSetting = Color.LightGray;
    private Color textColorSetting = DefaultBackColor;
    private Color borderColorSetting = Color.Black;

    private int borderWidth = 2;
    private int cornerRadius = 0;

    private String displayValue = "";
    private String rectValue = "";
    private int lastLinkStatus = 0;
    private Brush rectBrush = new SolidBrush(Color.LightGray);  //Brushes.AliceBlue;
    private Pen rectPen = new Pen(Brushes.Black, 2);
    private Brush rectLowWarnBrush = null;
    private Brush rectHiWarnBrush = null;
    private Brush rectLowTextBrush = null;
    private Brush rectHiTextBrush = null;
    private Brush errorBrush = new SolidBrush(Color.DarkRed);
    private Brush errorTextBrush = new SolidBrush(Color.White);
    private Pen rectLowWarnPen = null;
    private Pen rectHiWarnPen = null;
    private Pen errorPen = new Pen(Brushes.White);
    private Pen rectPenSetting = new Pen(Brushes.Black, 2);
    private Brush rectBrushSetting = new SolidBrush(Color.LightGray);  //Brushes.AliceBlue;
    private void setDisplayValue(String value)
    {
      displayValue = value;
      Boolean valueEmpty = value == null || value.Length == 0;
      if (IsDesignMode() || valueEmpty) rectValue = value;
      this.Refresh();
    }
    private String getDisplayValue()
    {
      return displayValue == null ? "" : displayValue;
    }
    private void setFillColor(Color value)
    {
      if (fillColor == value) return;
      fillColorSetting = fillColor = value;
      rectBrushSetting = rectBrush = new SolidBrush(fillColor);
      this.Refresh();
    }

    private void setLowWarnTextColor(Color value)
    {
      lowWarnTextColor = value;
      Refresh();
    }
    private void setHiWarnTextColor(Color value)
    {
      hiWarnTextColor = value;
      Refresh();
    }
    private void setLowWarnFillColor(Color value)
    {
      lowWarnFillColor = value;
      Refresh();
    }
    private void setHiWarnFillColor(Color value)
    {
      hiWarnFillColor = value;
      Refresh();
    }

    private void setBorderColor(Color value)
    {
      if (borderColor == value) return;
      borderColorSetting = borderColor = value;
      rectPenSetting = rectPen = new Pen(borderColor, borderWidth);
      this.Refresh();
    }
    private void setBorderWidth(int value)
    {
      if (value < 1) value = 1;
      if (borderWidth == value) return;
      borderWidth = value;
      border = (borderWidth + 1) / 2;
      rectPenSetting = rectPen = new Pen(borderColor, borderWidth);
      establishDrawnRect();
      this.Refresh();
    }
    private void setCornerRadius(int value)
    {
      if (cornerRadius < 0) cornerRadius = 0;
      if (cornerRadius == value) return;
      cornerRadius = value;
      this.Refresh();
    }
    private Rectangle rect;
    private Rectangle rectUnrotated;
    private int border = 2;
    private Color textColor = Color.DarkGray;
    private Font textFont = new Font("Arial", 10);
    private Brush textBrush = new SolidBrush(Color.DarkGray);
    private Brush textBrushSetting = new SolidBrush(Color.DarkGray);
    private void setTextColor(Color value)
    {
      if (textColor == value) return;
      textColorSetting = textColor = value;
      textBrush = new SolidBrush(textColor);
      textBrushSetting = textBrush;
      this.Refresh();
    }
    private void setTextFont(Font value)
    {
      if (textFont == value) return;
      textFont = value;
      this.Refresh();
    }
    private StringFormat textFormat = new StringFormat();
    public enum GradientEnum
    {
      Lighter = 0,
      Darker = 1
    }
    private GradientEnum gradient = GradientEnum.Lighter;
    private void setGradient(GradientEnum value)
    {
      if (gradient == value) return;
      gradient = value;
      Refresh();
    }
    public enum GradientStyleEnum
    {
      None = 0,
      Horizontal = 1,
      Vertical = 2,
      Forward_Diagonal = 3,
      Backward_Diagonal = 4,
      Central = 5,
    }
    private GradientStyleEnum gradientStyle = GradientStyleEnum.None;
    private void setGradientStyle(GradientStyleEnum value)
    {
      if (gradientStyle == value) return;
      gradientStyle = value;
      Refresh();
    }
    private int rotation = 0;
    private void setRotation(int value)
    {
      if (shapeType == ShapeTypes.line) return;
      rotation = value;
      Rectangle r = rectUnrotated;
      Rectangle bb = AcopShapeUtils.GetBoundingBox(r, rotation);
      double scaleW = (double)r.Width / (double)bb.Width;
      double scaleH = (double)r.Height / (double)bb.Height;
      double scale = Math.Min(scaleW, scaleH);
      int rW = (int)((double)(r.Width) * scale);
      int rH = (int)((double)(r.Height) * scale);
      int rX = border + (r.Width - rW) / 2;
      int rY = border + (r.Height - rH) / 2;
      rect = new Rectangle(rX, rY, rW, rH);
      Refresh();
    }
    private Boolean rotateText = false;
    private void setRotateText(Boolean value)
    {
      if (rotateText == value) return;
      rotateText = value;
      Refresh();
    }
    public enum ShapeTypes
    {
      rectangle = 0,
      oval = 1,
      line = 2,
      triangle = 3,
      arc = 4,
    }
    private ShapeTypes shapeType = ShapeTypes.rectangle;
    public void setShapeType(ShapeTypes value)
    {
      shapeType = value;
    }
    protected Boolean botToTop = true;
    protected void SetBottomToTop(Boolean value)
    {
      if (botToTop == value) return;
      botToTop = value;
      Refresh();
    }
    private Int32 straightLineAspect = 10;
    protected void setStraightLineAspect(Int32 value)
    {
      if (value < 1) value = 1;
      if (value > 1000) value = 1000;
      straightLineAspect = value;
    }
    protected Int32 getStraightLineAspect()
    {
      return straightLineAspect;
    }
    protected Int32 lineArrowSize = 3;
    protected void setLineArrowSize(Int32 value)
    {
      if (value < 1) value = 1;
      lineArrowSize = value;
      Refresh();
    }
    protected Boolean lineLeftArrow = false;
    protected void setLineLeftArrow(Boolean value)
    {
      lineLeftArrow = value;
      Refresh();
    }
    protected Boolean lineRightArrow = false;
    protected void setLineRightArrow(Boolean value)
    {
      lineRightArrow = value;
      Refresh();
    }
    protected TriangleType triangleType = TriangleType.isosceles;
    protected void setTriangleType(TriangleType value)
    {
      triangleType = value;
    }
    private float getTriangleTextLocation(float width)
    {
      if (triangleType == TriangleType.wedgeLeft) return (float)border;
      if (triangleType == TriangleType.wedgeRight) return this.Width - width - (float)border;
      return this.Width / 2 - width / 2;
    }
    protected float arcStartAngle = 0;
    protected float arcSweepAngle = 30;
    private void RectanglePaint(object sender, PaintEventArgs e)
    {
      e.Graphics.SmoothingMode = SmoothingMode.HighQuality;
      switch (shapeType)
      {
        case ShapeTypes.rectangle:
          AcopShapeUtils.DrawRoundedRectangle(e.Graphics, rectPen, rect, cornerRadius, rotation);
          AcopShapeUtils.FillRoundedRectangle(e.Graphics, rectBrush, rect, cornerRadius, (int)gradientStyle, (int)gradient, rotation);
          break;
        case ShapeTypes.oval:
          AcopShapeUtils.DrawEllipse(e.Graphics, rectPen, rect, rotation);
          AcopShapeUtils.FillEllipse(e.Graphics, rectBrush, rect, (int)gradientStyle, (int)gradient, rotation);
          break;
        case ShapeTypes.line:
          if (lineLeftArrow)
            rectPen.CustomStartCap = new AdjustableArrowCap(lineArrowSize, lineArrowSize);
          if (lineRightArrow)
            rectPen.CustomEndCap = new AdjustableArrowCap(lineArrowSize, lineArrowSize);
          AcopShapeUtils.DrawLine(e.Graphics, rectPen, rect, botToTop, extraWidth, extraHeight, straightLineAspect);
          break;
        case ShapeTypes.triangle:
          AcopShapeUtils.DrawRoundedPolygon(e.Graphics, rectPen, rect, 3, cornerRadius, rotation, (int)triangleType);
          AcopShapeUtils.FillRoundedPolygon(e.Graphics, rectBrush, rect, 3, cornerRadius, (int)gradientStyle, (int)gradient, rotation, (int)triangleType);
          break;
        case ShapeTypes.arc:
          AcopShapeUtils.DrawArc(e.Graphics, rectPen, rect, arcStartAngle, arcSweepAngle);
          AcopShapeUtils.FillArc(e.Graphics, rectBrush, rect, arcStartAngle, arcSweepAngle, (int)gradientStyle, (int)gradient, rotation);
          break;
      }
      if (!displayText) return;
      if (rectValue == null || rectValue.Length == 0) return;
      if (rotateText && extraHeight == 0)
      {
        float transX = this.Width / 2;
        float transY = this.Height / 2;
        if (shapeType == ShapeTypes.line)
        {
          double rads = Math.Atan2(rect.Height, rect.Width);
          rotation = (int)(57.295779513082320876798154814105 * rads);
          if (rect.Height / rect.Width >= straightLineAspect) rotation = 90;
          if (botToTop) rotation *= -1;
          if (extraWidth != 0) transX = (this.Width - extraWidth) / 2;
          if (extraHeight != 0) transY = (this.Height - extraHeight) / 2;
        }
        e.Graphics.TranslateTransform(transX, transY);
        e.Graphics.RotateTransform(rotation);
        SizeF textSize = e.Graphics.MeasureString(rectValue, textFont);
        float xpos = -textSize.Width / 2;
        float ypos = -textSize.Height / 2;
        if (shapeType == ShapeTypes.line)
        {
          ypos = -3 * this.Height / 8 - textSize.Height / 4;
          if (botToTop)
          {
            xpos = -this.Width / 8 - textSize.Width / 4;
            ypos = -this.Height / 4 - textSize.Height / 4;
          }
          else
          {
            xpos = this.Width / 8 - textSize.Width / 4;
            ypos = -this.Height / 4 - textSize.Height / 4;
          }
          xpos -= extraWidth / 2;
          if (extraWidth < 0) xpos += extraWidth / 2; // trial and error ...
        }
        e.Graphics.DrawString(rectValue, textFont, textBrush, xpos, ypos);
        e.Graphics.ResetTransform();
      }
      else
      {
        if (shapeType == ShapeTypes.line)
        {
          SizeF textSize = e.Graphics.MeasureString(rectValue, textFont);
          float m = botToTop ? 1 : 3;
          float xpos = m * this.Width / 4 - textSize.Width / 2;
          float ypos = this.Height / 4 - textSize.Height / 2;
          if (extraHeight != 0 || extraWidth != 0)
          {
            xpos = this.Width / 2 - textSize.Width / 2;
            ypos = this.Height / 2 - textSize.Height / 2;
          }
          e.Graphics.DrawString(rectValue, textFont, textBrush, xpos, ypos);
        }
        else if (shapeType == ShapeTypes.triangle)
        {
          float hf = 0.67f;
          int ra = rotation % 360;
          int rf = rotation <= 180 ? Math.Abs(ra) : Math.Abs(360 - ra);
          if (rf > 135) hf = 0.33f;
          else if (rf > 45) hf = 0.5f;
          SizeF textSize = e.Graphics.MeasureString(rectValue, textFont);
          float xpos = getTriangleTextLocation(textSize.Width);
          float ypos = hf * this.Height - textSize.Height / 2;
          e.Graphics.DrawString(rectValue, textFont, textBrush, xpos, ypos);
        }
        else
        {
          e.Graphics.DrawString(rectValue, textFont, textBrush, rect, textFormat);
        }
      }
    }
    private Boolean displayText = true;
    private void setDisplayText(Boolean value)
    {
      if (displayText == value) return;
      displayText = value;
      Refresh();
    }
    protected int extraHeight = 0;
    protected int extraWidth = 0;
    protected void setExtraHeight(int value)
    {
      if (value == extraHeight) return;
      extraHeight = value;
      establishDrawnRect();
      Refresh();
    }
    protected void setExtraWidth(int value)
    {
      if (value == extraWidth) return;
      extraWidth = value;
      establishDrawnRect();
      Refresh();
    }
    private void establishDrawnRect()
    {
      if (this.Width <= 2 * border) return;
      if (this.Height <= 2 * border) return;
      int xH = 0, xW = 0;
      if (shapeType == ShapeTypes.line)
      {
        xH = Math.Abs(extraHeight);
        xW = Math.Abs(extraWidth);
      }
      if (this.Height - xH <= 2 * border) return;
      if (this.Width - xW <= 2 * border) return;
      rect = new Rectangle(border, border, this.Width - xW - 2 * border, this.Height - xH - 2 * border);
      rectUnrotated = rect;
      setRotation(rotation);
    }
    private void AcopRectagleResized(object sender, EventArgs e)
    {
      establishDrawnRect();
    }
    private void setBlinkColors(Color fill, Color text)
    {
      rectBrush = new SolidBrush(fill);
      textBrush = new SolidBrush(text);
      if (shapeType == ShapeTypes.line) rectPen = new Pen(fill, BorderWidth); ;
      this.Refresh();
    }
    private void resetBlinkColors()
    {
      setFillColor(fillColor);
      setTextColor(textColor);
      if (shapeType == ShapeTypes.line) rectPen = rectPenSetting;
    }
    private Boolean isBlinkCycle = false;
    void blinkColors(Boolean blink)
    {
      if (blinkMode == BlinkMode.off)
      {
        resetBlinkColors();
        return;
      }
      if (blinkMode == BlinkMode.onWarning && !isWarningCondition)
      {
        resetBlinkColors();
        return;
      }
      isBlinkCycle = blink;
      if (blink)
      {
        Color fc =
          blinkerIntensity > 0 ? 
          ControlPaint.Light(fillColor, blinkerIntensity) : 
          ControlPaint.Dark(fillColor, -blinkerIntensity);
        Color tc =
          blinkerIntensity > 0 ? 
          ControlPaint.Dark(textColor, blinkerIntensity) :
          ControlPaint.Light(textColor, -blinkerIntensity);
        setBlinkColors(fc, tc);
      }
      else
      {
        setBlinkColors(fillColor, textColor);
      }
    }
    public AcopShape()
    {
      InitializeComponent();
      SetDefaultUpdater(defaultUpdate);
      setControlsList(new Control[] { this }, true);
      SetLateInitializerProlog(initTransportInfo);
      textFormat.Alignment = StringAlignment.Center;
      textFormat.LineAlignment = StringAlignment.Center;
      rect = new Rectangle(border, border, this.Width - 2 * border, this.Height - 2 * border);
      rectUnrotated = rect;
      StartLateInitialization();
      DoubleBuffered = true;
      SetStyle(ControlStyles.SupportsTransparentBackColor, true);
      this.BackColor = Color.Transparent;
      blink = blinkColors;
    }
    protected override void OnPaint(PaintEventArgs e)
    {
      shapeBackgroundPaint(e);
      base.OnPaint(e);
    }
    private void shapeBackgroundPaint(PaintEventArgs e)
    {
      if (Parent != null)
      {
        Bitmap behind = new Bitmap(Parent.Width, Parent.Height);
        int z = Parent.Controls.GetChildIndex(this);
        foreach (Control c in Parent.Controls)
        {
          if (Parent.Controls.GetChildIndex(c) < z) continue;
          if (c != this && c.Bounds.IntersectsWith(this.Bounds))
            c.DrawToBitmap(behind, c.Bounds);
        }
        e.Graphics.DrawImage(behind, -Left, -Top);
        behind.Dispose();
      }
    }
    public void applyErrorStatus(String sts, Boolean applyToTTT)
    {
      rectValue = sts;
      if (applyToTTT) applyToolTipText(sts);
      textColor = Color.White;
      fillColor = Color.DarkRed;
      if (shapeType == ShapeTypes.line) rectPen = errorPen;
      rectBrush = errorBrush;
      textBrush = errorTextBrush;
    }
    private void applyLowWarning()
    {
      if (isBlinkCycle) return; // only apply in an off cycle
      textBrush = textBrushSetting;
      if (lowWarnFillColor != Color.Empty)
      {
        if (shapeType == ShapeTypes.line)
        {
          if (rectLowWarnPen == null) rectLowWarnPen = new Pen(lowWarnFillColor, BorderWidth);
          rectPen = rectLowWarnPen;
        }
        else
        {
          if (rectLowWarnBrush == null) rectLowWarnBrush = new SolidBrush(lowWarnFillColor);
          fillColor = lowWarnFillColor;
          rectBrush = rectLowWarnBrush;
        }
      }
      else
      {
        fillColor = fillColorSetting;
        rectBrush = rectBrushSetting;
        if (shapeType == ShapeTypes.line)
        {
          rectPen = rectPenSetting;
        }
      }
      if (lowWarnTextColor != Color.Empty)
      {
        if (rectLowTextBrush == null) rectLowTextBrush = new SolidBrush(lowWarnTextColor);
        textBrush = rectLowTextBrush;
        textColor = lowWarnTextColor;
      }
      else
      {
        textColor = ForeColor;
      }
      isWarningCondition = true;
    }
    private Boolean isWarningCondition = false;
    private void applyHiWarning()
    {
      if (isBlinkCycle) return; // only apply in an off cycle
      textBrush = textBrushSetting;
      if (hiWarnFillColor != Color.Empty)
      {
        if (shapeType == ShapeTypes.line)
        {
          if (rectHiWarnPen == null) rectHiWarnPen = new Pen(hiWarnFillColor, BorderWidth);
          rectPen = rectHiWarnPen;
        }
        else
        {
          if (rectHiWarnBrush == null) rectHiWarnBrush = new SolidBrush(hiWarnFillColor);
          fillColor = hiWarnFillColor;
          rectBrush = rectHiWarnBrush;
        }
      }
      else
      {
        fillColor = fillColorSetting;
        rectBrush = rectBrushSetting;
        if (shapeType == ShapeTypes.line)
        {
          rectPen = rectPenSetting;
        }
      }
      if (hiWarnTextColor != Color.Empty)
      {
        if (rectHiTextBrush == null) rectHiTextBrush = new SolidBrush(hiWarnTextColor);
        textBrush = rectHiTextBrush;
        textColor = hiWarnTextColor;
      }
      else
      {
        textColor = textColorSetting;
      }
      isWarningCondition = true;
    }
    private void checkStringMatch(String txt)
    {
      if (txt == null || thresholdMatch.Length == 0) return;
      if (!TextMatches(thresholdMatch, txt))
      {
        applyHiWarning();
      }
      else
      {
        resetColors();
      }
    }

    private void resetColors()
    {
      fillColor = fillColorSetting;
      textColor = textColorSetting;
      rectBrush = rectBrushSetting;
      rectPen = rectPenSetting;
      textBrush = textBrushSetting;
      borderColor = borderColorSetting;
      isWarningCondition = false;
    }
    private Boolean checkThresholds(AcopControl.AcopLink lnk)
    {
      try
      {
        double v;
        Object o = lnk.GetOutputDataObject();
        if (o is Boolean[])
        {
          checkStringMatch(((Boolean[])o)[0].ToString());
          return true;
        }
        if (o is double[]) v = ((double[])o)[0];
        else if (o is float[]) v = (double)((float[])o)[0];
        else if (o is Int16[]) v = (double)((Int16[])o)[0];
        else if (o is Int32[]) v = (double)((Int32[])o)[0];
        else if (o is Int64[]) v = (double)((Int64[])o)[0];
        else if (o is Double) v = (Double)o;
        else if (o is Single) v = (double)((Single)o);
        else if (o is Int16) v = (double)((Int16)o);
        else if (o is Int32) v = (double)((Int32)o);
        else if (o is Int64) v = (double)((Int64)o);
        else return false;
        if (v < thresholdLow)
        {
          applyLowWarning();
        }
        else if (v > thresholdHigh)
        {
          applyHiWarning();
        }
        else
        {
          resetColors();
        }
        return true;
      }
      catch (Exception)
      {
      }
      return false;
    }
    private void setThresholdHigh(double value)
    {
      if (value != MAX_THRESHOLD) thresholdHigh = value;
    }
    private void setThresholdLow(double value)
    {
      if (value != -MAX_THRESHOLD) thresholdLow = value;
    }
    protected void initTransportInfo()
    {
      if (!IsDesignMode())
      {
        String[] labels = null;
        AcopTransport.DisplaySettings dsp = null;
        if (!AutoLinkAttach && !AutoLinkUpdate)
        {
          displaySettingsAcquired = true;
          return;
        }
        dsp = acquireAndApplyDisplaySettings(LinkAddress, LinkProperty, ref labels, false);
        if (dsp == null || dsp.SettingsStale()) return;
        displaySettingsAcquired = true;
        if (thresholdHigh == MAX_THRESHOLD) thresholdHigh = dsp.GetYMaximum();
        if (thresholdLow == -MAX_THRESHOLD) thresholdLow = dsp.GetYMinimum();
      }
    }
    private String GetRectangleString(AcopControl.AcopLink lnk)
    {
      if (lnk == null || lnk.GetLinkId() < 0) return DisplayValue;
      String[] txt = lnk.GetOutputDataAsString();
      StringBuilder sb = new StringBuilder();
      if (txt != null)
      {
        for (int i = 0; i < txt.Length; i++)
        {
          if (!checkThresholds(lnk)) checkStringMatch(txt[i]);
          sb.AppendLine(txt[i]);
        }
        if (displayValue != null && displayValue.Length > 0) return displayValue;
        return sb.ToString();
      }
      return "";
    }
    protected void defaultUpdate(AcopControl.AcopLink lnk)
    {
      if (!lateInitializationDone) return;
      if (getAutoUpdate())
      {
        int lnkSts = lnk.GetStatus();
        if (lnkSts != 0 && ErrorValue.Length == 0)
        {
          applyErrorStatus(lnk.GetStatusString(), lnkSts != lastLinkStatus);
        }
        else
        {
          if (lnkSts != lastLinkStatus) applyToolTipText(null);
          if (lnk.GetDisplayId() < 0)
          {
            if (!displaySettingsAcquired) initTransportInfo();
            lnk.SetDisplayId(1);
          }
          String rv = GetRectangleString(lnk);
          rectValue = rv;
        }
        Refresh();
      }
    }
    [Category("Acop.Transport")]
    [ToolboxItem(false)]
    [Browsable(false), EditorBrowsable(EditorBrowsableState.Never)]
    public new event acopApplySetting acopApply;

    [Category("Acop.Transport")]
    [Browsable(false), EditorBrowsable(EditorBrowsableState.Never)]
    public new ApplyEventSettingType ApplyEventOrder { get; set; }

    [Category("Acop.Display")]
    [ToolboxItem(false)]
    [Browsable(false), EditorBrowsable(EditorBrowsableState.Never)]
    public new Boolean DisplayAsMultiChannel { get; set; }

    #region Properties
    /// <summary>
    /// if non-empty and the link data is non-numerical, then the string provided is used for comparisons with the monitored data
    /// </summary>
    [Category("Acop.Display")]
    [Description("if non-empty and the link data is non-numerical, then the string provided is used for comparisons with the monitored data")]
    public String ThresholdMatch { get { return thresholdMatch; } set { thresholdMatch = value; } }
    /// <summary>
    /// if the link data is numerical, then the threshold provided is used for comparisons with the monitored data
    /// </summary>
    [Category("Acop.Display")]
    [Description("if the link data is numerical, then the threshold provided is used for comparisons with the monitored data")]
    public double ThresholdHighWarn { get { return thresholdHigh; } set { setThresholdHigh(value); } }
    /// <summary>
    /// if the link data is numerical, then the threshold provided is used for comparisons with the monitored data
    /// </summary>
    [Category("Acop.Display")]
    [Description("if the link data is numerical, then the threshold provided is used for comparisons with the monitored data")]
    public double ThresholdLowWarn { get { return thresholdLow; } set { setThresholdLow(value); } }
    /// <summary>
    /// if non-empty then the displayed value upon update will be the value given
    /// </summary>
    [Category("Acop.Display")]
    [Description("if non-empty then the displayed value upon update will be the value given")]
    public String DisplayValue { get { return getDisplayValue(); } set { setDisplayValue(value); } }
    /// <summary>
    /// true: display input text/readback value on shape
    /// </summary>
    [Category("Acop.Display")]
    [Description("true: display input text/readback value on shape")]
    public Boolean DisplayText { get { return displayText; } set { setDisplayText(value); } }
    /// <summary>
    /// background color of rectangle region
    /// </summary>
    [Category("Acop.Display")]
    [Description("background color of rectangle region")]
    public Color FillColor { get { return fillColor; } set { setFillColor(value); } }
    /// <summary>
    /// border color of rectangle region
    /// </summary>
    [Category("Acop.Display")]
    [Description("border color of rectangle region")]
    public Color BorderColor { get { return borderColor; } set { setBorderColor(value); } }
    /// <summary>
    /// color of display text in rectangle region
    /// </summary>
    [Category("Acop.Display")]
    [Description("color of display text in rectangle region")]
    public Color TextColor { get { return textColor; } set { setTextColor(value); } }
    /// <summary>
    /// color of display text when the readback value exceeds (negative direction) the low threshold")
    /// </summary>
    [Category("Acop.Display")]
    [Description("color of display text when the readback value exceeds (negative direction) the low threshold")]
    public Color LowWarnTextColor { get { return lowWarnTextColor; } set { setLowWarnTextColor(value); } }
    /// <summary>
    /// color of display text when the readback value exceeds the high threshold
    /// </summary>
    [Category("Acop.Display")]
    [Description("color of display text when the readback value exceeds the high threshold")]
    public Color HiWarnTextColor { get { return hiWarnTextColor; } set { setHiWarnTextColor(value); } }
    /// <summary>
    /// color of display rectangle region when the readback value exceeds (negative direction) the low threshold
    /// </summary>
    [Category("Acop.Display")]
    [Description("color of display rectangle region when the readback value exceeds (negative direction) the low threshold")]
    public Color LowWarnFillColor { get { return lowWarnFillColor; } set { setLowWarnFillColor(value); } }
    /// <summary>
    /// color of display rectangle region when the readback value exceeds the high threshold
    /// </summary>
    [Category("Acop.Display")]
    [Description("color of display rectangle region when the readback value exceeds the high threshold")]
    public Color HiWarnFillColor { get { return hiWarnFillColor; } set { setHiWarnFillColor(value); } }
    /// <summary>
    /// border width of rectangle region
    /// </summary>
    [Category("Acop.Display")]
    [Description("border width of rectangle region")]
    public Int32 BorderWidth { get { return borderWidth; } set { setBorderWidth(value); } }
    /// <summary>
    /// font of display text in rectangle region
    /// </summary>
    [Category("Acop.Display")]
    [Description("font of display text in rectangle region")]
    public Font TextFont { get { return textFont; } set { setTextFont(value); } }
    /// <summary>
    /// corner radius of rectangle region
    /// </summary>
    [Category("Acop.Display")]
    [Description("corner radius of rectangle region")]
    public Int32 CornerRadius { get { return cornerRadius; } set { setCornerRadius(value); } }
    /// <summary>
    /// rotation angle of rectangle region
    /// </summary>
    [Category("Acop.Display")]
    [Description("rotation angle of rectangle region")]
    public Int32 RotationAngle { get { return rotation; } set { setRotation(value); } }
    /// <summary>
    /// true: apply rotation angle to displayed text
    /// </summary>
    [Category("Acop.Display")]
    [Description("true: apply rotation angle to displayed text")]
    public Boolean RotateText { get { return rotateText; } set { setRotateText(value); } }
    /// <summary>
    /// color gradient style for filled region
    /// </summary>
    [Category("Acop.Display")]
    [Description("color gradient style for filled region")]
    public GradientStyleEnum GradientStyle { get { return gradientStyle; } set { setGradientStyle(value); } }
    /// <summary>
    /// color gradient for filled region
    /// </summary>
    [Category("Acop.Display")]
    [Description("color gradient for filled region")]
    public GradientEnum Gradient { get { return gradient; } set { setGradient(value); } }
    /// <summary>
    /// set the blink mode for the shape component
    /// </summary>
    [Category("Acop.Display")]
    [Description("set the blink mode for the shape component")]
    public BlinkMode Blink { get { return blinkMode; } set { setBlinkMode(value); } }
    /// <summary>
    /// set the blink intensity for the shape component (up to 100% lighter or -100% darker)
    /// </summary>
    [Category("Acop.Display")]
    [Description("set the blink intensity for the shape component (up to 100% lighter or -100% darker)")]
    public Int32 BlinkIntensity {  get { return blinkerIntensity; } set { setBlinkerIntensity(value); } }
    /// <summary>
    /// set the blink interval (msec) for the shape component
    /// </summary>
    [Category("Acop.Display")]
    [Description("set the blink interval (msec) for the shape component")]
    public Int32 BlinkInterval { get { return blinkerInterval; } set { setBlinkerInterval(value); } }
    /// <summary>
    /// set the blink slot for the shape component (0 up to the number of slots)
    /// </summary>
    [Category("Acop.Display")]
    [Description("set the blink slot for the shape component (0 up to the number of slots)")]
    public Int32 BlinkSlot { get { return blinkerSlot; } set { setBlinkerSlot(value); } }
    /// <summary>
    /// set the number of blink slots for the shape component (at least 2)
    /// </summary>
    [Category("Acop.Display")]
    [Description("set the number of blink slots for the shape component (at least 2)")]
    public Int32 NumberBlinkSlots { get { return numberBlinkSlots; } set { setNumberBlinkerSlots(value); } }

    #endregion
  }
  public enum TriangleType
  {
    isosceles = 0,
    wedgeLeft = 1,
    wedgeRight = 2
  }
  public static class AcopShapeUtils
  {
    private static LinearGradientBrush linearGradientBrush = null;
    private static PathGradientBrush pathGradientBrush = null;

    private static GraphicsPath RoundedPolygon(Rectangle bounds, int sides, int radius)
    {
      return RoundedPolygon(bounds, sides, radius, 0);
    }
    private static GraphicsPath RoundedPolygon(Rectangle bounds, int sides, int radius, int flags)
    {
      if (sides < 3) sides = 3; // minimum
      if (radius > bounds.Height / 4) radius = bounds.Height / 4;
      if (radius > bounds.Width / 4) radius = bounds.Width / 4;
      int diameter = radius * 2;
      Size size = new Size(diameter, diameter);
      GraphicsPath path = new GraphicsPath();
      Point[] points = new Point[sides];
      switch (sides)
      {
        case 3:
          if (flags == (int)TriangleType.wedgeLeft)
          {
            points[0] = new Point(bounds.X, bounds.Y + bounds.Height);
            points[1] = new Point(bounds.X + bounds.Width, bounds.Y + bounds.Height);
            points[2] = new Point(bounds.X, bounds.Y);
          }
          else if (flags == (int)TriangleType.wedgeRight)
          {
            points[0] = new Point(bounds.X, bounds.Y + bounds.Height);
            points[1] = new Point(bounds.X + bounds.Width, bounds.Y + bounds.Height);
            points[2] = new Point(bounds.X + bounds.Width, bounds.Y);
          }
          else
          {
            points[0] = new Point(bounds.X, bounds.Y + bounds.Height);
            points[1] = new Point(bounds.X + bounds.Width, bounds.Y + bounds.Height);
            points[2] = new Point(bounds.X + bounds.Width / 2, bounds.Y);
          }
          break;
      }
      if (radius == 0)
      {
        path.AddPolygon(points);
      }
      else
      {
        List<Line> lines = new List<Line>();
        for (int i=0; i< sides+1; i++)
        {
          lines.Add(shortenLine(new Line(points[i%3], points[(i+1)%3]), radius));
        }
        for (int i = 0; i < lines.Count - 1; i++)
        {
          path.AddLine(lines[i].P1, lines[i].P2);
          path.AddCurve(new Point[] { lines[i].P2,
                                  supPoint(lines[i].P2, points[(i + 1) % 3], lines[i+1].P1),
                                  lines[i+1].P1 });
        }
      }
      return path;
    }
    public static void DrawRoundedPolygon(this Graphics graphics, Pen pen, Rectangle bounds, int sides, int cornerRadius, int rotation)
    {
      DrawRoundedPolygon(graphics, pen, bounds, sides, cornerRadius, rotation, 0);
    }
    public static void DrawRoundedPolygon(this Graphics graphics, Pen pen, Rectangle bounds, int sides, int cornerRadius, int rotation, int flags)
    {
      if (graphics == null)
        throw new ArgumentNullException("graphics");
      if (pen == null)
        throw new ArgumentNullException("pen");
      if (bounds.Height < 10 || bounds.Width < 10) return;
      if (Math.Abs(bounds.Height - bounds.Width) > 100) return;
      using (Matrix m = new Matrix())
      {
        m.RotateAt(rotation, new PointF(bounds.Left + (bounds.Width / 2),
                                  bounds.Top + (bounds.Height / 2)));
        graphics.Transform = m;
        using (GraphicsPath path = RoundedPolygon(bounds, sides, cornerRadius, flags))
        {
          graphics.DrawPath(pen, path);
        }
        graphics.ResetTransform();
      }
    }
    public static void FillRoundedPolygon(this Graphics graphics, Brush brush, Rectangle bounds, int sides, int cornerRadius, int gradientStyle, int gradient, int rotation)
    {
      FillRoundedPolygon(graphics, brush, bounds, sides, cornerRadius, gradientStyle, gradient, rotation, 0);
    }
    public static void FillRoundedPolygon(this Graphics graphics, Brush brush, Rectangle bounds, int sides, int cornerRadius, int gradientStyle, int gradient, int rotation, int flags)
    {
      if (graphics == null)
        throw new ArgumentNullException("graphics");
      if (brush == null)
        throw new ArgumentNullException("brush");

      using (Matrix m = new Matrix())
      {
        m.RotateAt(rotation, new PointF(bounds.Left + (bounds.Width / 2),
                                  bounds.Top + (bounds.Height / 2)));
        graphics.Transform = m;
        using (GraphicsPath path = RoundedPolygon(bounds, sides, cornerRadius, flags))
        {
          Color c1 = (brush as SolidBrush).Color;
          Color c2 = gradient == 0 ? ControlPaint.Light(c1, 50f) : ControlPaint.Dark(c1, 50f);
          switch (gradientStyle)
          {
            case 0:
              graphics.FillPath(brush, path);
              break;
            case 5:
              GraphicsPath gp = new GraphicsPath();
              gp.AddRectangle(bounds);
              pathGradientBrush = new PathGradientBrush(gp);
              pathGradientBrush.CenterPoint = new PointF(bounds.Width / 2, bounds.Height / 2);
              pathGradientBrush.CenterColor = c1;
              pathGradientBrush.SurroundColors = new Color[] { c2 };
              graphics.FillPath(pathGradientBrush, path);
              break;
            default:
              linearGradientBrush = new LinearGradientBrush(bounds, c1, c2, gradientAngles[gradientStyle]);
              graphics.FillPath(linearGradientBrush, path);
              break;
          }
        }
        graphics.ResetTransform();
      }
    }
    public static GraphicsPath RoundedRect(Rectangle bounds, int radius)
    {
      if (radius > bounds.Height / 4) radius = bounds.Height / 4;
      if (radius > bounds.Width / 4) radius = bounds.Width / 4;
      int diameter = radius * 2;
      Size size = new Size(diameter, diameter);
      Rectangle arc = new Rectangle(bounds.Location, size);
      GraphicsPath path = new GraphicsPath();
      if (radius == 0)
      {
        path.AddRectangle(bounds);
        return path;
      }
      // top left arc  
      path.AddArc(arc, 180, 90);
      // top right arc  
      arc.X = bounds.Right - diameter;
      path.AddArc(arc, 270, 90);
      // bottom right arc  
      arc.Y = bounds.Bottom - diameter;
      path.AddArc(arc, 0, 90);
      // bottom left arc 
      arc.X = bounds.Left;
      path.AddArc(arc, 90, 90);
      path.CloseFigure();
      return path;
    }
    public static Rectangle GetBoundingBox(Rectangle r, float angle)
    {
      if (angle == 0) return r;
      Rectangle bbox = r;
      angle = angle % 360;
      if (angle > 180) angle = 360 - angle;
      angle *= 0.01745329251994329576923690768489F; // convert to radians
      double ct = Math.Cos(angle);
      double st = Math.Sin(angle);
      double x_min = r.X, x_max = r.X + r.Width;
      double y_min = r.Y, y_max = r.Y + r.Height;
      if (angle > 0)
      {
        if (angle < 1.5707963267948966192313216916398) //  90 degrees
        {
          y_min = r.Y;
          y_max = r.Y + r.Height * ct + r.Width * st;
          x_min = r.X - r.Height * st;
          x_max = r.X + r.Width * ct;
        }
        else
        {
          y_min = r.Y + r.Height * ct;
          y_max = r.Y + r.Width * st;
          x_min = r.X - r.Height * st + r.Width * ct;
          x_max = r.X;
        }
      }
      else
      {
        if (angle > -1.5707963267948966192313216916398) //  -90 degrees
        {
          y_min = r.Y + r.Width * st;
          y_max = r.Y + r.Height * ct;
          x_min = r.X;
          x_max = r.X + r.Width * ct - r.Height * st;
        }
        else
        {
          y_min = r.Y + r.Width * st + r.Height * ct;
          y_max = r.Y;
          x_min = r.X + r.Width * ct;
          x_max = r.X - r.Height * st;
        }
      }
      double w = x_max - x_min;
      double h = y_max - y_min;
      if (w < 5 || h < 5) return r;
      if (w / h > 10 || w / h < 0.1) return r;
      bbox = new Rectangle((int)x_min, (int)y_min, (int)w, (int)h);
      return bbox;
    }
    public static void DrawRoundedRectangle(this Graphics graphics, Pen pen, Rectangle bounds, int cornerRadius, int rotation)
    {
      if (graphics == null)
        throw new ArgumentNullException("graphics");
      if (pen == null)
        throw new ArgumentNullException("pen");
      using (Matrix m = new Matrix())
      {
        m.RotateAt(rotation, new PointF(bounds.Left + (bounds.Width / 2),
                                  bounds.Top + (bounds.Height / 2)));
        graphics.Transform = m;
        using (GraphicsPath path = RoundedRect(bounds, cornerRadius))
        {
          graphics.DrawPath(pen, path);
        }
        graphics.ResetTransform();
      }
    }
    public static void DrawEllipse(this Graphics graphics, Pen pen, Rectangle bounds, int rotation)
    {
      if (graphics == null)
        throw new ArgumentNullException("graphics");
      if (pen == null)
        throw new ArgumentNullException("pen");

      using (Matrix m = new Matrix())
      {
        m.RotateAt(rotation, new PointF(bounds.Left + (bounds.Width / 2),
                                  bounds.Top + (bounds.Height / 2)));
        graphics.Transform = m;
        graphics.DrawEllipse(pen, bounds.X, bounds.Y, bounds.Width, bounds.Height);
        graphics.ResetTransform();
      }
    }
    public static void DrawLine(this Graphics graphics, Pen pen, Rectangle bounds, Boolean bottomToTop, int xoffset, int yoffset, Int32 straight_line_aspect)
    {
      int x1 = bounds.X;
      int y1 = bounds.Y;
      int x2 = bounds.Width;
      int y2 = bounds.Height;
      if (bottomToTop)
      {
        y1 = bounds.Height;
        y2 = bounds.Y;
      }
      if (yoffset > 0)
      {
        y1 += yoffset;
        y2 += yoffset;
      }
      if (xoffset < 0)
      {
        x1 -= xoffset;
        x2 -= xoffset;
      }
      float delta_y = (float)(y2 - y1);
      float delta_x = (float)(x2 - x1);
      if (Math.Abs(delta_y / delta_x) >= straight_line_aspect)
      { // force into vertical line
        x2 = x1 = Math.Min(x1, x2); 
      }
      if (Math.Abs(delta_x / delta_y) >= straight_line_aspect)
      { // force into horizontal line
        y2 = y1 = Math.Min(y1, y2); 
      }
      graphics.DrawLine(pen, x1, y1, x2, y2);
    }
    private struct TaggedLine
    {
      String tag;
      public Pen pen;
      public GraphicsPath path;
      public TaggedLine(String tag, Pen pen, GraphicsPath path)
      {
        this.tag = tag;
        this.pen = pen;
        this.path = path;
      }
    }
    private static Dictionary<String, TaggedLine> tlines = new Dictionary<string, TaggedLine>();
    public static Boolean IsPointOnPath(String tag, int x, int y)
    {
      TaggedLine tl;

      if (tlines.TryGetValue(tag,out tl))
      {
        if (tl.path.IsOutlineVisible(x, y, tl.pen)) return true;
      }

      return false;
    }
    /// <summary>
    /// Draws a line on a Form
    /// </summary>
    /// <param name="form">the form on which to draw</param>
    /// <param name="color">the prefered line color</param>
    /// <param name="width">the prefered pen width</param>
    /// <param name="x1">the x coordinate of the starting point</param>
    /// <param name="y1">the y coordinate of the starting point</param>
    /// <param name="x2">the x coordinate of the ending point</param>
    /// <param name="y2">the y coordinate of the ending point</param>
    /// <param name="tag">a tag name to identify this line</param>
    public static void DrawLine(Form form, Color color, int width, int x1, int y1, int x2, int y2, String tag)
    {
      if (form == null) return;
      Graphics g = form.CreateGraphics();
      g.SmoothingMode = SmoothingMode.HighQuality;
      Pen p = new Pen(color, width);
      g.DrawLine(p,x1, y1, x2, y2);
      if (tag != null)
      {
        if (tlines.ContainsKey(tag)) tlines.Remove(tag);
        GraphicsPath path = new GraphicsPath();
        path.AddLine(x1, y1, x2, y2);
        tlines.Add(tag, new TaggedLine(tag, p, path));
      }
      g.Dispose();
    }
    /// <summary>
    /// Draws a line on a Form
    /// </summary>
    /// <param name="form">the form on which to draw</param>
    /// <param name="color">the prefered line color</param>
    /// <param name="width">the prefered pen width</param>
    /// <param name="p1">the starting point</param>
    /// <param name="p2">the ending point</param>
    /// <param name="tag">a tag name to identify this line</param>
    public static void DrawLine(Form form, Color color, int width, Point p1, Point p2, String tag)
    {
      if (form == null) return;
      Graphics g = form.CreateGraphics();
      g.SmoothingMode = SmoothingMode.HighQuality;
      Pen p = new Pen(color, width);
      g.DrawLine(p, p1, p2);
      if (tag != null)
      {
        if (tlines.ContainsKey(tag)) tlines.Remove(tag);
        GraphicsPath path = new GraphicsPath();
        path.AddLine(p1, p2);
        tlines.Add(tag, new TaggedLine(tag, p, path));
      }
      g.Dispose();
    }
    /// <summary>
    /// Draws an arc on a Form
    /// </summary>
    /// <param name="form">the form on which to draw</param>
    /// <param name="color">the prefered line color</param>
    /// <param name="width">the prefered pen width</param>
    /// <param name="rect">the rectangle specifying the boundaries of the ellipse containing the arc </param>
    /// <param name="startAngle">the starting angle (degrees) from the x-axis to start of the arc </param>
    /// <param name="sweepAngle">the sweep angle (degrees) of the arc</param>
    /// <param name="tag">a tag name to identify this arc</param>
    public static void DrawArc(Form form, Color color, int width, Rectangle rect, float startAngle, float sweepAngle, String tag)
    {
      if (form == null) return;
      Graphics g = form.CreateGraphics();
      g.SmoothingMode = SmoothingMode.HighQuality;
      Pen p = new Pen(color, width);
      g.DrawArc(p, rect, startAngle, sweepAngle);
      if (tag != null)
      {
        if (tlines.ContainsKey(tag)) tlines.Remove(tag);
        GraphicsPath path = new GraphicsPath();
        path.AddArc(rect, startAngle, sweepAngle);
        tlines.Add(tag, new TaggedLine(tag, p, path));
      }
      g.Dispose();
    }
    public static void DrawArc(this Graphics graphics, Pen pen, Rectangle bounds, float startAngle, float sweepAngle)
    {
      Rectangle arc = new Rectangle(bounds.Location, bounds.Size);
      GraphicsPath path = new GraphicsPath();
      path.AddArc(arc, startAngle, sweepAngle);
      arc.X -= bounds.Width / 2;
      arc.Y -= bounds.Height / 2;
      //path.AddLine()
      path.AddArc(arc, startAngle, sweepAngle);

      path.CloseFigure();
      graphics.DrawPath(pen, path);

      //graphics.DrawArc(pen, bounds, startAngle, sweepAngle);
    }
    public static void FillArc(this Graphics graphics, Brush brush, Rectangle bounds, float startAngle, float sweepAngle, int gradientStyle, int gradient, int rotation)
    {
      Color c1 = (brush as SolidBrush).Color;
      Color c2 = gradient == 0 ? ControlPaint.Light(c1, 50f) : ControlPaint.Dark(c1, 50f);
      using (Matrix m = new Matrix())
      {
        m.RotateAt(rotation, new PointF(bounds.Left + (bounds.Width / 2),
                                  bounds.Top + (bounds.Height / 2)));
        graphics.Transform = m;
        switch (gradientStyle)
        {
          case 0:
            graphics.FillPie(brush, bounds, startAngle, sweepAngle);
            break;
          case 5:
            GraphicsPath gp = new GraphicsPath();
            gp.AddRectangle(bounds);
            pathGradientBrush = new PathGradientBrush(gp);
            pathGradientBrush.CenterPoint = new PointF(bounds.Width / 2, bounds.Height / 2);
            pathGradientBrush.CenterColor = c1;
            pathGradientBrush.SurroundColors = new Color[] { c2 };
            graphics.FillPie(pathGradientBrush, bounds, startAngle, sweepAngle);
            break;
          default:
            linearGradientBrush = new LinearGradientBrush(bounds, c1, c2, gradientAngles[gradientStyle]);
            graphics.FillPie(linearGradientBrush, bounds, startAngle, sweepAngle);
            break;
        }
        graphics.ResetTransform();
      }
    }

    /// <summary>
    /// Draws a line on a Form
    /// </summary>
    /// <param name="form">the form on which to draw</param>
    /// <param name="color">the prefered line color</param>
    /// <param name="width">the prefered pen width</param>
    /// <param name="control1">the control whose mid point gives the line start</param>
    /// <param name="control2">the control whose mid point gives the line end</param>
    /// <param name="tag">a tag name to identify this line</param>
    public static void DrawLine(Form form, Color color, int width, Control control1, Control control2, String tag)
    {
      if (form == null) return;
      Graphics g = form.CreateGraphics();
      g.SmoothingMode = SmoothingMode.HighQuality;
      Pen p = new Pen(color, width);
      int x1 = control1.Location.X + control1.Size.Width / 2;
      int y1 = control1.Location.Y + control1.Size.Height / 2;
      int x2 = control2.Location.X + control2.Size.Width / 2;
      int y2 = control2.Location.Y + control2.Size.Height / 2;
      g.DrawLine(p, x1, y1, x2, y2);
      if (tag != null)
      {
        if (tlines.ContainsKey(tag)) tlines.Remove(tag);
        GraphicsPath path = new GraphicsPath();
        path.AddLine(x1, y1, x2, y2);
        tlines.Add(tag, new TaggedLine(tag, p, path));
      }
      g.Dispose();
    }
    /// <summary>
    /// Draws a closed polygon on the form
    /// </summary>
    /// <param name="form">the form on which to draw</param>
    /// <param name="color">the prefered line color</param>
    /// <param name="width">the prefered pen width</param>
    /// <param name="points">the points defining the polygon vertices</param>
    /// <param name="tag">a tag name to identify this polygon shape</param>
    public static void DrawPolygon(Form form, Color color, int width, Point[] points, String tag)
    {
      if (form == null) return;
      if (points == null || points.Length < 3) return;
      Graphics g = form.CreateGraphics();
      g.SmoothingMode = SmoothingMode.HighQuality;
      Pen p = new Pen(color, width);
      g.DrawPolygon(p, points);
      if (tag != null)
      {
        if (tlines.ContainsKey(tag)) tlines.Remove(tag);
        GraphicsPath path = new GraphicsPath();
        path.AddPolygon(points);
        tlines.Add(tag, new TaggedLine(tag, p, path));
      }
      g.Dispose();
    }
    public static void FillEllipse(this Graphics graphics, Brush brush, Rectangle bounds, int gradientStyle, int gradient, int rotation)
    {
      Color c1 = (brush as SolidBrush).Color;
      Color c2 = gradient == 0 ? ControlPaint.Light(c1, 50f) : ControlPaint.Dark(c1, 50f);
      using (Matrix m = new Matrix())
      {
        m.RotateAt(rotation, new PointF(bounds.Left + (bounds.Width / 2),
                                  bounds.Top + (bounds.Height / 2)));
        graphics.Transform = m;
        switch (gradientStyle)
        {
          case 0:
            graphics.FillEllipse(brush, bounds);
            break;
          case 5:
            GraphicsPath gp = new GraphicsPath();
            gp.AddRectangle(bounds);
            pathGradientBrush = new PathGradientBrush(gp);
            pathGradientBrush.CenterPoint = new PointF(bounds.Width / 2, bounds.Height / 2);
            pathGradientBrush.CenterColor = c1;
            pathGradientBrush.SurroundColors = new Color[] { c2 };
            graphics.FillEllipse(pathGradientBrush, bounds);
            break;
          default:
            linearGradientBrush = new LinearGradientBrush(bounds, c1, c2, gradientAngles[gradientStyle]);
            graphics.FillEllipse(linearGradientBrush, bounds);
            break;
        }
        graphics.ResetTransform();
      }
    }
    private static int[] gradientAngles = new int[] { -1, 0, 90, 45, 135, 1000 };
    public static void FillRoundedRectangle(this Graphics graphics, Brush brush, Rectangle bounds, int cornerRadius, int gradientStyle, int gradient, int rotation)
    {
      if (graphics == null)
        throw new ArgumentNullException("graphics");
      if (brush == null)
        throw new ArgumentNullException("brush");

      using (Matrix m = new Matrix())
      {
        m.RotateAt(rotation, new PointF(bounds.Left + (bounds.Width / 2),
                                  bounds.Top + (bounds.Height / 2)));
        graphics.Transform = m;
        using (GraphicsPath path = RoundedRect(bounds, cornerRadius))
        {
          Color c1 = (brush as SolidBrush).Color;
          Color c2 = gradient == 0 ? ControlPaint.Light(c1, 50f) : ControlPaint.Dark(c1, 50f);
          switch (gradientStyle)
          {
            case 0:
              graphics.FillPath(brush, path);
              break;
            case 5:
              GraphicsPath gp = new GraphicsPath();
              gp.AddRectangle(bounds);
              pathGradientBrush = new PathGradientBrush(gp);
              pathGradientBrush.CenterPoint = new PointF(bounds.Width / 2, bounds.Height / 2);
              pathGradientBrush.CenterColor = c1;
              pathGradientBrush.SurroundColors = new Color[] { c2 };
              graphics.FillPath(pathGradientBrush, path);
              break;
            default:
              linearGradientBrush = new LinearGradientBrush(bounds, c1, c2, gradientAngles[gradientStyle]);
              graphics.FillPath(linearGradientBrush, path);
              break;
          }
        }
        graphics.ResetTransform();
      }
    }

    // a simple structure
    private struct Line
    {
      public Point P1; public Point P2;
      public Line(Point p1, Point p2) { P1 = p1; P2 = p2; }
    }
    // routine to make a line shorter on both ends
    private static Line shortenLine(Line line, int byPixels)
    {
      float len = (float)Math.Sqrt(Math.Pow(line.P1.X - line.P2.X, 2)
                                 + Math.Pow(line.P1.Y - line.P2.Y, 2));
      float prop = (len - byPixels) / len;
      Point p2 = pointBetween(line.P1, line.P2, prop);
      Point p1 = pointBetween(line.P1, line.P2, 1 - prop);
      return new Line(p1, p2);
    }
    // with a proportion of 0.5 the point sits in the middle
    private static Point pointBetween(Point p1, Point p2, float prop)
    {
      return new Point((int)(p1.X + (p2.X - p1.X) * prop),
                       (int)(p1.Y + (p2.Y - p1.Y) * prop));
    }
    private static Point supPoint(Point p1, Point p2, Point p0)
    {
      Point p12 = pointBetween(p1, p2, 0.5f);
      Point p20 = pointBetween(p2, p0, 0.5f);
      return pointBetween(p12, p20, 0.5f);       // **
    }
  }
}
