﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using System.Drawing.Drawing2D;

namespace Acop
{
  [ToolboxItem(true)]
  [ToolboxBitmap(typeof(TrackBar))]
  public partial class AcopSlider : AcopUserComponent
  {
#   pragma warning disable 1591
    private Boolean showHeader = true;
    private AcopSliderContextMenu cm;
    private void setShowHeader(Boolean value)
    {
      showHeader = value;
      resizeSlider(value);
    }
    public AcopSlider()
    {
      InitializeComponent();
      allowDragAndDrop(new Control[] { this, ReadbackLabel, DescLabel });
      SetDefaultUpdater(defaultUpdate);
      SetDefaultApplier(defaultApplier);
      setControlsList(new Control[] { Label, ReadbackLabel, DescLabel, LastLabel, Slider, panel1 }, true);
      defaultDescriptionLabel = DescLabel;
      Slider.Maximum = fullSliderRange;
      Slider.TickFrequency = fullSliderRange / 10;
      applyToolTipText(ContinuousCheck, "if checked: send setting with each value change (independent of 'set' button)");
      applyToolTipText(LockCheck, "if checked: apply an exclusive (only this application is allowed) access lock to the set value");
      applyToolTipText(ApplyButton, "send the set value to the target");
      applyToolTipText(SaveReferenceButton, "set the reference value to the current slider value");
      applyToolTipText(ResetButton, "send the reference value to the target");
      applyToolTipText(label1, "the current readback value");
      applyToolTipText(label2, "this slider's reference value");
      SetLateInitializerProlog(initTransportInfo);
      StartLateInitialization();
      cm = new AcopSliderContextMenu(this);
      this.ContextMenu = cm;
    }
    private void resizeSlider(Boolean showHeaderComponents)
    {
      this.SuspendLayout();
      int twidth = this.Width; // 315
      int hheight = 0, xpos, lwidth;

      DescLabel.Visible = ReadbackLabel.Visible = ResetButton.Visible = showHeaderComponents;
      ApplyButton.Visible = SaveReferenceButton.Visible = LastLabel.Visible = showHeaderComponents;
      label1.Visible = label2.Visible = label3.Visible = showHeaderComponents;

      // stretch ReadbackLabel, Lastlabel AND Label
      if (showHeaderComponents)
      {
        this.DescLabel.Location = new System.Drawing.Point(3, 6);
        this.label3.Location = new System.Drawing.Point(3, 22);

        this.label1.Location = new System.Drawing.Point(2, 28);
        this.ReadbackLabel.Location = new System.Drawing.Point(32, 28);

        xpos = twidth - ResetButton.Width - 4; // 268 
        this.ResetButton.Location = new System.Drawing.Point(xpos, 22);
        xpos -= SaveReferenceButton.Width + 2; // 226
        this.SaveReferenceButton.Location = new System.Drawing.Point(xpos, 22);
        xpos -= ApplyButton.Width + 3;   // 183
        this.ApplyButton.Location = new System.Drawing.Point(xpos, 22);

        lwidth = (ApplyButton.Location.X - label1.Location.X - label1.Width - label2.Width) / 2;
        xpos = ReadbackLabel.Location.X + lwidth - 4;
        this.label2.Location = new System.Drawing.Point(xpos, 28);
        lwidth -= 2;  // 46 ?
        this.ReadbackLabel.Size = new System.Drawing.Size(lwidth, 14);

        xpos = label2.Location.X + label2.Width + 1;
        this.LastLabel.Location = new System.Drawing.Point(xpos, 28);
        this.LastLabel.Size = new System.Drawing.Size(lwidth, 14);

        hheight = 50;
      }
      else
      {
        hheight = 0;
      }
      this.IncrMM.Location = new System.Drawing.Point(1, hheight + 1);
      this.IncrM.Location = new System.Drawing.Point(33, hheight + 1);
      this.Label.Location = new System.Drawing.Point(55, hheight + 3);

      xpos = twidth - LockCheck.Width - 1; // 230
      this.LockCheck.Location = new System.Drawing.Point(xpos, hheight + 5);
      xpos -= ContinuousCheck.Width - 1; // 179
      this.ContinuousCheck.Location = new System.Drawing.Point(xpos, hheight + 5);
      xpos -= IncrPP.Width + 11; // 138
      this.IncrPP.Location = new System.Drawing.Point(xpos, hheight + 1);
      xpos -= IncrP.Width + 2; // 118
      this.IncrP.Location = new System.Drawing.Point(xpos, hheight + 1);
      lwidth = xpos - 57;
      this.Label.Size = new System.Drawing.Size(lwidth, 20);

      this.MinLabel.Location = new System.Drawing.Point(9, hheight + 48);
      xpos = (twidth - MidLabel.Width) / 2;
      this.MidLabel.Location = new System.Drawing.Point(xpos, hheight + 48);
      xpos = twidth - MaxLabel.Width - 8;
      this.MaxLabel.Location = new System.Drawing.Point(xpos, hheight + 48);

      this.Slider.Location = new System.Drawing.Point(1, hheight + 26);
      this.Slider.Size = new System.Drawing.Size(twidth - 2, 45);
      this.panel1.Location = new System.Drawing.Point(1, hheight + 53);
      this.panel1.Size = new System.Drawing.Size(twidth - 4, 12);

      this.MinimumSize = new System.Drawing.Size(315, hheight + 65);

      this.ResumeLayout();
    }
    private Color labelInSyncForeColor = Color.DarkBlue;
    private Color labelOutOfSyncForeColor = Color.DarkRed;
    private void setLabelInSyncForeColor(Color value)
    {
      labelInSyncForeColor = value;
    }
    private void setLabelOutOfSyncForeColor(Color value)
    {
      labelOutOfSyncForeColor = value;
    }
    private Color readbackForeColor = Color.Black;
    private Color referenceForeColor = Color.Black;
    private void setReadbackForeColor(Color value)
    {
      readbackForeColor = value;
      ReadbackLabel.ForeColor = value;
    }
    private void setReferenceForeColor(Color value)
    {
      referenceForeColor = value;
      LastLabel.ForeColor = value;
    }
    private void applyErrorStatus(String sts, Boolean applyToTTT)
    {
      ReadbackLabel.Text = sts;
      if (applyToTTT) applyToolTipText(sts);
      Label.ForeColor = Color.DarkRed;
    }
    private int lastLinkStatus = 0;
    private void initTransportInfo()
    {
      if (!IsDesignMode())
      {
        if (hasTargetComponent())
        {
          try
          { // casting errors ?
            double v = (double)getSettingFromControlProperty(getTargetComponent());
            SetSliderValue(v);
            if (MaximumValue == MAX_THRESHOLD)
            {
              MaximumValue = v > 0 ? v * 10 : v / 10;
              MinimumValue = v > 0 ? v / 10 : v * 10;
            }
          }
          catch (Exception) { }
          displaySettingsAcquired = true;
          return;
        }
        String[] labels = null;
        AcopTransport.DisplaySettings dsp = null;
        if (!AutoLinkAttach && !AutoLinkUpdate)
        {
          DescLabel.Text = Description;
          displaySettingsAcquired = true;
          return;
        }
        dsp = acquireAndApplyDisplaySettings(LinkAddress, LinkProperty, ref labels, false);
        if (dsp == null || dsp.SettingsStale()) return;
        displaySettingsAcquired = true;
        if (thresholdHigh == MAX_THRESHOLD) thresholdHigh = dsp.GetYMaximum();
        if (thresholdLow == -MAX_THRESHOLD) thresholdLow = dsp.GetYMinimum();
        applyScaleText(dsp);
        DescLabel.Text = (getIsMCA() && getDisplayAsMCA()) ? getMCADescription() : getDescription();
      }
    }
    private void defaultUpdate(AcopControl.AcopLink lnk)
    {
      if (!lateInitializationDone) return;
      if (getAutoUpdate())
      {
        int lnkSts = lnk.GetStatus();
        if (lnkSts != 0 && ErrorValue.Length == 0)
        {
          applyErrorStatus(lnk.GetStatusString(), lnkSts != lastLinkStatus);
        }
        else
        {
          if (lnkSts != lastLinkStatus) applyToolTipText(null);
          double v = getValueAsDouble(lnk);
          String txt;
          if (getValueAvailableAsNumber())
          {
            txt = getLabelString(v, 6);
          }
          else
          {
            String[] txtlines = lnk.GetOutputDataAsString();
            txt = txtlines[0];
          }
          if (lnk.GetDisplayId() < 0)
          {
            if (!displaySettingsAcquired) initTransportInfo();
            lnk.SetDisplayId(1);
            Slider.Value = getSliderValue(v);
            SetResetValue(v);
            panel1.setRefValue = v;
            Label.Text = txt;
            LastLabel.Text = getLabelString(v, 6);
          }
          SetReadbackValue(v);
          panel1.setRbvValue = v;
          ReadbackLabel.Text = txt;
          Label.ForeColor = (ReadbackLabel.Text.CompareTo(Label.Text) != 0) ? labelOutOfSyncForeColor : labelInSyncForeColor;
        }
      }
    }
    double getValueFromLabel()
    {
      double v = 0;
      try
      {
        v = getValueIsInteger() ?
          (double)(Int32.Parse(Label.Text)) :
          Double.Parse(Label.Text);
      }
      catch (Exception)
      {
        v = 0;
      }
      return v;
    }
    private int defaultApplier(AcopControl.AcopLink lnk)
    {
      int cc = lnk.GetStatus();
      if (getValueAvailableAsNumber() && cc == 0)
      {
        double v = getValueFromLabel();
        Label.Text = getLabelString(v, 6);
      }
      if (cc != 0) Label.Text = lnk.GetStatusString();
      else ReadbackLabel.Text = Label.Text;
      return cc;
    }
    Boolean applyContinuous = false;
    private void setApplyContinuous(Boolean value)
    {
      ContinuousCheck.Checked = value;
      applyContinuous = value;
    }
    private double thresholdHigh = MAX_THRESHOLD;
    private double thresholdLow = -MAX_THRESHOLD;
    private static int fullSliderRange = 1000000;
    private void applyScaleText(AcopTransport.DisplaySettings dsp)
    {
      String sms, lgs;
      if (thresholdHigh == MAX_THRESHOLD && dsp != null) thresholdHigh = dsp.GetYMaximum();
      if (thresholdLow == -MAX_THRESHOLD && dsp != null) thresholdLow = dsp.GetYMinimum();
      if (thresholdHigh == MAX_THRESHOLD || thresholdLow == -MAX_THRESHOLD) return;
      if (thresholdHigh == thresholdLow) thresholdHigh = thresholdLow + 100.0;
#     if   EARLY_LEARNING
      if (getValueIsInteger())
      {
        Slider.Maximum = (int)thresholdHigh;
        Slider.Minimum = (int)thresholdLow;
        panel1.setMaxValue = (int)thresholdHigh;
        panel1.setMinValue = (int)thresholdLow;
        if (smallIncrement == 0) smallIncrement = (thresholdHigh - thresholdLow) / 100;
        if (largeIncrement == 0) largeIncrement = (thresholdHigh - thresholdLow) / 10;
        Slider.SmallChange = (int)smallIncrement;
        Slider.LargeChange = (int)largeIncrement;
        sms = getLabelString(smallIncrement, 3);
        lgs = getLabelString(largeIncrement, 3);
        applyToolTipText(IncrP, "adjust set value by +"+ sms);
        applyToolTipText(IncrPP, "adjust set value by +" + lgs);
        applyToolTipText(IncrM, "adjust set value by -" + sms);
        applyToolTipText(IncrMM, "adjust set value by -" + lgs);
        sliderFactor = 1.0;
        return;
      }
#     endif
      sliderFactor = (thresholdHigh - thresholdLow) / (double)fullSliderRange;
      MinLabel.Text = getLabelString(thresholdLow, 4, true);
      MaxLabel.Text = getLabelString(thresholdHigh, 4, true);
      panel1.setMaxValue = thresholdHigh;
      panel1.setMinValue = thresholdLow;
      double mv = thresholdLow + ((thresholdHigh - thresholdLow) / 2.0);
      if (mv > 100.0) mv = (double)((int)mv);
      MidLabel.Text = getLabelString(mv, 4, true);
      if (smallIncrement == 0)
      {
        mv = (thresholdHigh - thresholdLow) / 100.0;
        smallIncrement = mv > 10 ? Math.Round(mv) : mv > 1 ? Math.Round(mv, 1) : mv;
      }
      //IncrLabel.Text = getLabelString(smallIncrement, 3);
      if (largeIncrement == 0)
      {
        mv = (thresholdHigh - thresholdLow) / 10.0;
        largeIncrement = mv > 10 ? Math.Round(mv) : mv > 1 ? Math.Round(mv, 1) : mv;
      }
      int smallChange = (int)Math.Round(smallIncrement / sliderFactor);
      int largeChange = (int)Math.Round(largeIncrement / sliderFactor);
      Slider.SmallChange = smallChange;
      Slider.LargeChange = largeChange;
      //IncrLLabel.Text = getLabelString(largeIncrement, 3);
      sms = getLabelString(smallIncrement, 3);
      lgs = getLabelString(largeIncrement, 3);
      applyToolTipText(IncrP, "adjust set value by +" + sms);
      applyToolTipText(IncrPP, "adjust set value by +" + lgs);
      applyToolTipText(IncrM, "adjust set value by -" + sms);
      applyToolTipText(IncrMM, "adjust set value by -" + lgs);
    }
    private void setThresholdHigh(double value)
    {
      if (value != MAX_THRESHOLD) thresholdHigh = value;
      if (!IsDesignMode()) applyScaleText(null);
    }
    private void setThresholdLow(double value)
    {
      if (value != -MAX_THRESHOLD) thresholdLow = value;
      if (!IsDesignMode()) applyScaleText(null);
    }
    private double sliderFactor = 1.0;
    private double smallIncrement = 0;
    private double largeIncrement = 0;
    private void setSmallIncrement(double value)
    {
      if (value > 0)
      {
        if (getValueIsInteger()) value = Math.Round(value);
        smallIncrement = value;
      }
    }
    private void setLargeIncrement(double value)
    {
      if (value > 0)
      {
        if (getValueIsInteger()) value = Math.Round(value);
        largeIncrement = value;
      }
    }
    private int getSliderValue(double v)
    {
#     if EARLY_LEARNING
      if (getValueIsInteger())
      {
        int iv = (int)v;
        if (iv > Slider.Maximum) iv = Slider.Maximum;
        if (iv < Slider.Minimum) iv = Slider.Minimum;
        return iv;
      }
#     endif
      int sv = (int)Math.Round((v - thresholdLow) / sliderFactor);
      if (sv < 0) sv = 0;
      if (sv > fullSliderRange) sv = fullSliderRange;
      return sv;
    }
    private double getValueFromSlider()
    {
#     if EARLY_LEARNING
      if (getValueIsInteger())
      {
        return (double)Slider.Value;
      }
#     endif
      return thresholdLow + ((double)(Slider.Value)) * sliderFactor;
    }
    /// <summary>
    /// Sets the slider value to the value given
    /// </summary>
    /// <param name="v">the desired slider value</param>
    public void SetSliderValue(double v)
    {
      Slider.Value = getSliderValue(v);
    }
    /// <summary>
    /// Gives the current slider value
    /// </summary>
    /// <returns>the current slider value</returns>
    public double GetSliderValue()
    {
      return getValueFromSlider();
    }
    /// <summary>
    /// interal use: applies the slider value to the assigned transport link
    /// </summary>
    /// <param name="sender">the sender control</param>
    /// <returns>0 if the value was successfully applied to the target</returns>
    public int applySliderSetting(object sender)
    {
      object inpt = null;
      AcopControl.AcopLink lnk = null;
      if (hasTargetComponent())
      {
        inpt = GetSliderValue();
      }
      else
      {
        Type type = getValueType(); // as long as getValueAsDouble() is called in the updater this will be something ...
        if (type == null) type = getValueType(Label.Text);
        inpt = AcopControl.AcopLink.GetDataObjectFromString(type, Label.Text);
        lnk = AcopControl.AcopLink.GetAcopLink(LinkProtocol, LinkAddress, LinkProperty, null, inpt, 0, this);
      }
      return applySetting(sender, inpt, lnk);
    }
    private delegate void SliderValueChangedDelegate(object sender, EventArgs e);
    private void SliderValueChanged(object sender, EventArgs e)
    {
      if (Label.InvokeRequired)
      {
        Invoke(new SliderValueChangedDelegate(SliderValueChanged), sender, e);
        return;
      }
      Label.Text = getLabelString(getValueFromSlider(), 6);
      if (applyContinuous)
      {
        int cc = applySliderSetting(sender);
      }
    }
    private void ApplyContinousClick(object sender, EventArgs e)
    {
      applyContinuous = ContinuousCheck.Checked;
    }
    private void ApplyClick(object sender, EventArgs e)
    {
      int cc = applySliderSetting(sender);
    }
    private void ApplyLockClick(object sender, EventArgs e)
    {
      applyLock(LockCheck);
    }
    private void LabelDoubleCLick(object sender, MouseEventArgs e)
    {
      int cc = applySliderSetting(sender);
    }
    private void LargeNegativeClick(object sender, EventArgs e)
    {
      double dv = getValueFromSlider();
      dv -= largeIncrement;
      if (dv < thresholdLow) dv = thresholdLow;
      Slider.Value = getSliderValue(dv);
      if (applyContinuous)
      {
        int cc = applySliderSetting(sender);
      }
    }
    private void LargePositiveClick(object sender, EventArgs e)
    {
      double dv = getValueFromSlider();
      dv += largeIncrement;
      if (dv > thresholdHigh) dv = thresholdHigh;
      Slider.Value = getSliderValue(dv);
      if (applyContinuous)
      {
        int cc = applySliderSetting(sender);
      }
    }
    /// <summary>
    /// interal use: applies the current slider reference value to the assigned transport link
    /// </summary>
    /// <param name="sender">the sender control</param>
    public void Reset(object sender)
    {
      Slider.Value = getSliderValue(GetResetValue());
      if (ReadbackLabel.Text.CompareTo(LastLabel.Text) == 0) return;
      int cc = applySliderSetting(sender);
    }
    private void ResetClick(object sender, EventArgs e)
    {
      Reset(sender);
    }
    /// <summary>
    /// Assigns the current slider value to the reference value.
    /// </summary>
    public void SaveReference()
    {
      double resetValue = getValueFromSlider();
      SetResetValue(resetValue);
      LastLabel.Text = getLabelString(resetValue, 6);
      panel1.setRefValue = resetValue;
    }
    private void SaveReferenceClick(object sender, EventArgs e)
    {
      SaveReference();
    }
    private void SmallPositiveClick(object sender, EventArgs e)
    {
      double dv = getValueFromSlider();
      dv += smallIncrement;
      if (dv > thresholdHigh) dv = thresholdHigh;
      Slider.Value = getSliderValue(dv);
      if (applyContinuous)
      {
        int cc = applySliderSetting(sender);
      }
    }
    private void SmallNegativeClick(object sender, EventArgs e)
    {
      double dv = getValueFromSlider();
      dv -= smallIncrement;
      if (dv < thresholdLow) dv = thresholdLow;
      Slider.Value = getSliderValue(dv);
      if (applyContinuous)
      {
        int cc = applySliderSetting(sender);
      }
    }
    private void handleMouseDown(MouseEventArgs e)
    {
      if (showHeader) return;
      if (e.Button == MouseButtons.Right)
      {
        Point pos = this.PointToClient(Cursor.Position);
        cm.Show(this, pos, LeftRightAlignment.Right);
      }
    }
    private Font sliderFont = null;
    private Font getDisplayFont()
    {
      if (sliderFont == null) sliderFont = this.Label.Font;
      return sliderFont;
    }
    private void setDisplayFont(Font font)
    {
      sliderFont = font;
      foreach(Control c in this.Controls)
      {
        if (c is TrackBar) continue;
        if (c.Name.Equals("label3")) continue; // bar line 
        if (c.Name.Equals("label1") || c.Name.Equals("label2"))
          c.Font = new System.Drawing.Font(c.Font.Name, font.Size, c.Font.Style);
        else
          c.Font = font;
      }
    }
    /// <summary>
    /// font to use int the acop control
    /// </summary>
    [Category("Acop.Display")]
    [Description("Font to use int the acop control")]
    public Font DisplayFont { get { return getDisplayFont(); } set { setDisplayFont(value); } }
    /// <summary>
    /// the maximum setting value
    /// </summary>
    [Category("Acop.Display")]
    [Description("the maximum setting value")]
    public double MaximumValue { get { return thresholdHigh; } set { setThresholdHigh(value); } }
    /// <summary>
    /// the minimum setting value
    /// </summary>
    [Category("Acop.Display")]
    [Description("the minimum setting value")]
    public double MinimumValue { get { return thresholdLow; } set { setThresholdLow(value); } }
    /// <summary>
    /// the small increment setting value
    /// </summary>
    [Category("Acop.Display")]
    [Description("the small increment setting value")]
    public double SmallIncrement { get { return smallIncrement; } set { setSmallIncrement(value); } }
    /// <summary>
    /// Value at position of slider
    /// </summary>
    [Category("Acop.Display")]
    [Description("Value at position of slider")]
    public double SliderValue { get { return GetSliderValue(); } set { SetSliderValue(value); } }
    /// <summary>
    /// the large increment setting value
    /// </summary>
    [Category("Acop.Display")]
    [Description("the large increment setting value")]
    public double LargeIncrement { get { return largeIncrement; } set { setLargeIncrement(value); } }
    /// <summary>
    /// ForeColor to use if setting is synchronized with the current readback value
    /// </summary>
    [Category("Acop.Display")]
    [Description("ForeColor to use if setting is synchronized with the current readback value")]
    public Color LabelInSyncForeColor { get { return labelInSyncForeColor; } set { setLabelInSyncForeColor(value); } }
    /// <summary>
    /// ForeColor to use if setting is not synchronized with the current readback value
    /// </summary>
    [Category("Acop.Display")]
    [Description("ForeColor to use if setting is not synchronized with the current readback value")]
    public Color LabelOutOfSyncForeColor { get { return labelOutOfSyncForeColor; } set { setLabelOutOfSyncForeColor(value); } }
    /// <summary>
    /// Default ForeColor to use for the Readback label
    /// </summary>
    [Category("Acop.Display")]
    [Description("Default ForeColor to use for the Readback label")]
    public Color ReadbackForeColor { get { return readbackForeColor; } set { setReadbackForeColor(value); } }
    /// <summary>
    /// Default ForeColor to use for the Reference label
    /// </summary>
    [Category("Acop.Display")]
    [Description("Default ForeColor to use for the Reference label")]
    public Color ReferenceForeColor { get { return referenceForeColor; } set { setReferenceForeColor(value); } }
    /// <summary>
    /// Show the readback needle in the slider
    /// </summary>
    [Category("Acop.Display")]
    [Description("Show the readback needle in the slider")]
    [DefaultValue(false)]
    public Boolean ShowReadBack { get { return panel1.showRbv; } set { panel1.showRbv = value; } }
    /// <summary>
    /// Show the reference needle in the slider
    /// </summary>
    [Category("Acop.Display")]
    [Description("Show the reference needle in the slider")]
    [DefaultValue(false)]
    public Boolean ShowReference { get { return panel1.showRef; } set { panel1.showRef = value; } }
    /// <summary>
    /// true: displays the slider header information and buttons or false: not
    /// </summary>
    [Category("Acop.Display")]
    [Description("true: displays the slider header information and buttons or false: not")]
    public Boolean ShowHeader { get { return showHeader; } set { setShowHeader(value); } }
    /// <summary>
    /// true: any change in slider position applies a new set value
    /// </summary>
    [Category("Acop.Display")]
    [Description("true: any change in slider position applies a new set value")]
    public Boolean ApplyContinuous { get { return applyContinuous; } set { setApplyContinuous(value); } }

    [Category("Acop.Display")]
    [Browsable(false), EditorBrowsable(EditorBrowsableState.Never)]
    public new Boolean DisplayAsMultiChannel { get; set; }

    [Category("Acop.Transport")]
    [Browsable(false), EditorBrowsable(EditorBrowsableState.Never)]
    public new Boolean LinkInputValue { get; set; }

    [Category("Acop.Transport")]
    [Browsable(false), EditorBrowsable(EditorBrowsableState.Never)]
    public new Boolean LinkGrouped { get; set; }

    //[Category("Acop.Display")]
    //[Description("If not empty, should be in the form of 'ControlName.Property' : apply setting will set the control property to the slider value")]
    //public String TargetComponent { get { return getTargetComponent(); } set { setTargetComponent(value); } }

    [Category("Acop.Display")]
    [Description("If not empty, attach the target control property to the slider")]
    [Editor(typeof(TargetControlEditor), typeof(System.Drawing.Design.UITypeEditor))]
    [DesignerSerializationVisibility(DesignerSerializationVisibility.Content)]
    [TypeConverter(typeof(TargetControlEditorConverter))]
    public TargetControlEditorValue Target { get { return getTargetControlEditorValue(); } set { setTargetControlEditorValue(value); } }

    private void SliderResized(object sender, EventArgs e)
    {
      resizeSlider(showHeader);
    }

    private void SliderMouseDown(object sender, MouseEventArgs e)
    {
      handleMouseDown(e);
    }
  }
  [ToolboxItem(false)]
  public class AcopSliderContextMenu : ContextMenu
  {
#pragma warning disable 1591
    private AcopSlider slider;
    private MenuItem apply;
    private MenuItem reset;
    private MenuItem save;

    public AcopSliderContextMenu(AcopSlider slider)
    {
      this.slider = slider;
      apply = new MenuItem("Apply");
      apply.Click += new EventHandler(this.applyChanges);
      save = new MenuItem("Set Reference");
      save.Click += new EventHandler(this.applySave);
      reset = new MenuItem("Restore Reference");
      reset.Click += new EventHandler(this.applyReset);

      this.MenuItems.Add(apply);
      this.MenuItems.Add(save);
      this.MenuItems.Add(reset);
    }
    private void applyChanges(object sender, EventArgs e)
    {
      slider.applySliderSetting(sender);
    }
    private void applySave(object sender, EventArgs e)
    {
      slider.SaveReference();
    }
    private void applyReset(object sender, EventArgs e)
    {
      slider.Reset(sender);
    }
  }

  [ToolboxItem(false)]
  public partial class needlePanel : Panel
  {
#pragma warning disable 1591
    private double rbvValue = 0;
    private double refValue = 0;
    private double sliderMin = 0;
    private double sliderMax = 1000;

    private int xStart = 12;

    private int bWidth;
    private int bHeight;

    private Boolean showRbvNeedle = false;
    private Boolean showRefNeedle = false;

    private Color colorRbvNeedle = Color.DarkBlue;
    private Color colorRefNeedle = Color.DarkGray;

    public needlePanel()
    {
      this.BackColor = Color.Transparent;
      this.DoubleBuffered = true;
    }

    private void setSliderMin(double value)
    {
      sliderMin = value;
    }
    private void setSliderMax(double value)
    {
      sliderMax = value;
    }

    private void setRbv(double value)
    {
      rbvValue = value;
      Refresh();
    }
    private void setRef(double value)
    {
      refValue = value;
      Refresh();
    }
    private void setShowRbv(Boolean value)
    {
      showRbvNeedle = value;
      Refresh();
    }
    private void setShowRef(Boolean value)
    {
      showRefNeedle = value;
      Refresh();
    }

    public double setMinValue { get { return this.sliderMin; } set { setSliderMin(value); } }
    public double setMaxValue { get { return this.sliderMax; } set { setSliderMax(value); } }
    public double setRbvValue { get { return this.rbvValue; } set { setRbv(value); } }
    public double setRefValue { get { return this.refValue; } set { setRef(value); } }
    public Boolean showRbv { get { return this.showRbvNeedle; } set { setShowRbv(value); } }
    public Boolean showRef { get { return this.showRefNeedle; } set { setShowRef(value); } }

    protected override void OnPaint(PaintEventArgs e)
    {
      base.OnPaint(e);
      Bitmap needleBar = new Bitmap(this.Width, this.Height, e.Graphics);
      bWidth = needleBar.Width;
      bHeight = needleBar.Height;
      Graphics gi = Graphics.FromImage(needleBar);
      gi.SmoothingMode = SmoothingMode.HighQuality;

      drawNeedles(gi);

      e.Graphics.DrawImageUnscaled(needleBar, 0, 0);
    }

    private void drawNeedles(Graphics g)
    {
      FillMode fm = FillMode.Winding;
      PointF p1, p2, p3;

      if (showRefNeedle)
      {
        Brush refBrush = new SolidBrush(colorRefNeedle);
        double dValue = Math.Sqrt(Math.Pow(refValue - sliderMin, 2));
        double x = xStart + ((bWidth - (2 * xStart)) * dValue / Math.Abs(sliderMax - sliderMin));
        p1 = new PointF((float)x, 0);
        p2 = new PointF((float)x + 4, bHeight);
        p3 = new PointF((float)x - 4, bHeight);
        PointF[] points = { p1, p2, p3 };

        g.FillPolygon(refBrush, points, fm);
      }
      if (showRbvNeedle)
      {
        Brush rbvBrush = new SolidBrush(colorRbvNeedle);
        double dValue = Math.Sqrt(Math.Pow(rbvValue - sliderMin, 2));
        double x = xStart + ((bWidth - (2 * xStart)) * dValue / Math.Abs(sliderMax - sliderMin));
        p1 = new PointF((float)x, 0);
        p2 = new PointF((float)x + 4, bHeight);
        p3 = new PointF((float)x - 4, bHeight);
        PointF[] points = { p1, p2, p3 };

        g.FillPolygon(rbvBrush, points, fm);
      }
    }
  }
}
