﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using Acop.AcopSpiderUtil;
using System.Reflection;

namespace Acop
{
  [ToolboxItem(true)]
  [ToolboxBitmap(typeof(AcopSpider), "Spider")]
  public partial class AcopSpider : UserControl
  {
#   pragma warning disable 1591
    DebugInfo info = null;
    static Assembly asm = null;
    static AcopSpider instance = null;
    AcopTransport.AcopLink alnk = null;

    String protocol = AcopControl.AcopLink.GetDefaultProtocol();

    private string ttt = "click to see links and messages; double-click to launch debug session";

    private enum SpiderColor
    {
      green = 0,
      gray = 1,
      red = 2
    }

    private SpiderColor currentSpiderColor = SpiderColor.green;

    public static AcopSpider GetInstance()
    {
      if (instance == null) instance = new AcopSpider();
      return instance;
    }

    public AcopSpider()
    {
      InitializeComponent();
      loadTransport();
    }

    ~AcopSpider()
    {
      bkgTimer.Enabled = false;
    }

    public void loadTransport()
    {
      try
      {
        if (asm == null)
        {
          asm = Assembly.Load("AcopTransport");
        }
      }
      catch (Exception x)
      {
        //throw new Exception("cannot load transport assembly", x); 
        return;
      }
      try
      {
        if (alnk == null) alnk = (AcopTransport.AcopLink)asm.CreateInstance("AcopTransport.AcopLink");
        instance = this;
      }
      catch (Exception x)
      {
        //throw new Exception("cannot create Acop Link", x); 
        return;
      }
    }

    public String[] GetConnections(Boolean activeOnly)
    {
      if (alnk == null) return new String[] { "none" };
      return alnk.GetConnectionList(protocol, activeOnly);
    }
    public String[] GetMessages()
    {
      if (alnk == null) return new String[] { "none" };
      return alnk.GetTransportMessages(protocol);
    }
    public int GetConnectionErrors()
    {
      if (alnk == null) return 0;
      return alnk.GetTransportStatus(protocol);
    }

    public string AcopSpiderHint
    {
      get
      {
        return SpiderToolTip.GetToolTip(GetInstance());
      }
      set
      {
        SpiderToolTip.SetToolTip(GetInstance(), value);
      }
    }

    private void StartDebuggingSession()
    {
      if (alnk != null) alnk.StartDebuggingSession(protocol);
    }

    public void removeDebugWindow()
    {
      info = null;
    }

    private void AcopSpider_Load(object sender, EventArgs e)
    {
      AcopSpiderHint = ttt;
    }

    private void SpiderClick(object sender, EventArgs e)
    {
      if (info == null) info = new DebugInfo();
      info.updateInfo();
      info.Show();
      info.Focus();
    }
    private void SpiderDblClick(object sender, EventArgs e)
    {
      StartDebuggingSession();
    }

    private void BkgTimerTick(object sender, EventArgs e)
    {
      int nerrs = this.GetConnectionErrors();
      if (nerrs > 0)
      {
        if (currentSpiderColor != SpiderColor.green && currentSpiderColor != SpiderColor.red)
        {
          currentSpiderColor = SpiderColor.red;
          this.BackgroundImage = Properties.Resources.spider_red;
        }
        else
        {
          currentSpiderColor = SpiderColor.gray;
          this.BackgroundImage = Properties.Resources.spider_gray;
        }
      }
      else
      {
        if (currentSpiderColor != SpiderColor.green)
        {
          this.BackgroundImage = Properties.Resources.spider_green;
          currentSpiderColor = SpiderColor.green;
        }
      }
      if (info != null) info.refreshActiveLinksInList();

    }
  }
}
