﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace Acop
{
  [ToolboxItem(true)]
  [ToolboxBitmap(typeof(TableLayoutPanel))]

  public partial class AcopTable : AcopUserComponent
  {
#   pragma warning disable 1591
    private static AcopTable instance = null;
    private static AcopTable getAcopTableInstance()
    {
      return instance;
    }
    private AcopTableContextMenu cm;
    private AcopTableInputMenu im;
    private Boolean showHeader = true;
    private Boolean showDescriptionColumn = true;
    private Boolean showInputColumn = false;
    private Boolean descriptionColumnVisible = false;
    private int numTableRows = 1;
    private int maxTableRows = 30;
    private int initialRowCount = 0;
    private void setMaxTableRows(int value)
    {
      if (value < numTableRows) return;
      maxTableRows = value;
    }
    private void setNumTableRows(int value)
    {
      if (value == numTableRows) return; // no reason to do anything
      if (value < 1) return;
      if (value > maxTableRows) value = maxTableRows;
      lock (tblLst)
      {
        if (value < numTableRows)
        {
          removeTableRowsAfter(value);
        }
        if (value != numTableRows)
        {
          Table.RowCount = value; 
        }
        numTableRows = value;
        int ncols = hdrRow == null ? 1 : hdrRow.cells.Length;
        String[] dscs = getDescArray();
        String[] vals;
        String rowstr, dscstr;
        for (int i = 0; i < numTableRows; i++) 
        {
          dscstr = dscs != null && i < dscs.Length ? dscs[i] : "description";
          rowstr = IsDesignMode() ? "Value " + i : "";
          vals = Enumerable.Repeat(rowstr, ncols).ToArray();
          addTableEntry(dscstr, vals, "Input", i, true);
        }
      }
      Boolean vsv = Table.VerticalScroll.Visible;
      if (vsv != scrollVisible)
      {
        int offset = vsv ? SystemInformation.VerticalScrollBarWidth/4 : 0;
        setHeaderColumnSize(offset);
        scrollVisible = vsv;
      }
    }
    private void setShowHeader(Boolean value)
    {
      if (value == showHeader) return;
      Header.Visible = value;
      int hsize = value ? hdrHeight : 0;
      int w = Table.Size.Width;
      int h = this.Size.Height;
      Table.SuspendLayout();
      Table.Location = new System.Drawing.Point(0, hsize);
      Table.Size = new System.Drawing.Size(w, h - hsize);
      this.Refresh();
      Table.ResumeLayout();
      showHeader = value;
    }
    private Boolean scrollVisible = false;
    private void setShowInputColumn(Boolean value)
    {
      if (numInputColumns == 0) numInputColumns = 1;
      setShowInputColumn(0, value);
    }
    private void setShowInputColumn(int valueColumn, Boolean value)
    {
      showInputColumn = value;
      Boolean showInpt = false;
      TableRow tr = null;
      Table.SuspendLayout();
      Header.SuspendLayout();
      int colcount = 0;
      if (descriptionColumnVisible) colcount++;
      int ncols = hdrRow == null ? 1 : hdrRow.cells.Length;
      for (int i = 0; i < ncols; i++)
      { // remove and add at the right spot :
        Header.Controls.Remove(hdrRow.cells[i].valLab);
        Header.Controls.Add(hdrRow.cells[i].valLab, colcount++, 0);
        showInpt = value && (i < numInputColumns);
        Header.Controls.Remove(hdrRow.cells[i].inptLab);
        hdrRow.cells[i].showInput = showInpt;
        if (hdrRow.cells[i].showInput)
        {
          Header.Controls.Add(hdrRow.cells[i].inptLab, colcount++, 0);
        }
      }
      for (int row = 0; row < tblLst.Count; row++)
      {
        tr = tblLst.ElementAt(row);
        colcount = 0;
        if (descriptionColumnVisible) colcount++;
        for (int i = 0; i < tr.cells.Length; i++)
        { // remove and add at the right spot :
          Table.Controls.Remove(tr.cells[i].valLab);
          Table.Controls.Add(tr.cells[i].valLab, colcount++, row);
          showInpt = value && (i < numInputColumns);
          Table.Controls.Remove(tr.cells[i].inptTxt);
          tr.cells[i].showInput = showInpt;
          if (tr.cells[i].showInput)
          {
            Table.Controls.Add(tr.cells[i].inptTxt, colcount++, row);
          }
        }
      }
      // colcount should always end up at the same number ...
      Table.ColumnCount = colcount;
      Header.ColumnCount = colcount;

      Header.ResumeLayout();
      Table.ResumeLayout();
      setColumnDescriptionSize(descColSize);
    }
    private void setShowDescriptionColumn(Boolean value)
    {
      lock (initMutex)
      {
        showDescriptionColumn = value;
        TableRow tr = null;
        int colcount = 0;
        Table.SuspendLayout();
        Header.SuspendLayout();

        if (descriptionColumnVisible) Header.Controls.Remove(hdrRow.dscLab);
        if (showDescriptionColumn) Header.Controls.Add(hdrRow.dscLab, colcount++, 0);
        int ncols = hdrRow == null ? 1 : hdrRow.cells.Length;
        for (int i = 0; i < ncols; i++)
        { // remove and add at the right spot :
          Header.Controls.Remove(hdrRow.cells[i].valLab);
          Header.Controls.Add(hdrRow.cells[i].valLab, colcount++, 0);
          if (hdrRow.cells[i].showInput)
          {
            Header.Controls.Remove(hdrRow.cells[i].inptLab);
            Header.Controls.Add(hdrRow.cells[i].inptLab, colcount++, 0);
          }
        }
        for (int row = 0; row < tblLst.Count; row++)
        {
          tr = tblLst.ElementAt(row);
          colcount = 0;
          if (descriptionColumnVisible) Table.Controls.Remove(tr.dscLab);
          if (showDescriptionColumn) Table.Controls.Add(tr.dscLab, colcount++, row);
          for (int i = 0; i < tr.cells.Length; i++)
          { // remove and add at the right spot :
            Table.Controls.Remove(tr.cells[i].valLab);
            Table.Controls.Add(tr.cells[i].valLab, colcount++, row);
            if (tr.cells[i].showInput)
            {
              Table.Controls.Remove(tr.cells[i].inptTxt);
              Table.Controls.Add(tr.cells[i].inptTxt, colcount++, row);
            }
          }
        }
        Header.ColumnCount = colcount;
        Table.ColumnCount = colcount;
        descriptionColumnVisible = showDescriptionColumn;

        Table.ResumeLayout();
        Header.ResumeLayout();
      }
    }
    private int numValueColumns = 1;
    private int numInputColumns = 0;
    private object initMutex = new object();
    private delegate void setNumInptCols(int value);
    private void setNumberInputColumns(int value)
    {
      if (value < 1) return; // not allowed
      if (value > numValueColumns) value = numValueColumns;
      if (value == numInputColumns) return; // nothing to do
      foreach (Control c in getControlsList())
        if (c.InvokeRequired)
        {
          try
          {
            Invoke(new setNumInptCols(setNumberInputColumns),value);
          }
          catch (Exception) { }
          return;
        }
      lock (initMutex)
      {
        Table.SuspendLayout();
        Header.SuspendLayout();
        clearCells();
        numInputColumns = value;
        for (int i = 0; i < hdrRow.cells.Length; i++)
        {
          hdrRow.cells[i].showInput = i < numInputColumns;
        }
        TableRow tr = null;
        for (int row = 0; row < tblLst.Count; row++)
        {
          tr = tblLst.ElementAt(row);
          for (int i = 0; i < tr.cells.Length; i++)
          {
            tr.cells[i].showInput = i < numInputColumns;
          }
        }
        fillCells();
        Table.ResumeLayout();
        Header.ResumeLayout();

        setColumnDescriptionSize(descColSize);
      }
    }
    private delegate void clearCellsDelegate();
    private void clearCells()
    {
      foreach (Control c in getControlsList())
        if (c.InvokeRequired)
        {
          try
          {
            Invoke(new clearCellsDelegate(clearCells));
          }
          catch (Exception) { }
          return;
        }
      for (int i = 0; i < hdrRow.cells.Length; i++)
      { // clean it out
        Header.Controls.Remove(hdrRow.cells[i].valLab);
        if (hdrRow.cells[i].showInput)
        {
          Header.Controls.Remove(hdrRow.cells[i].inptLab);
        }
      }
      TableRow tr = null;
      for (int row = 0; row < tblLst.Count; row++)
      {
        tr = tblLst.ElementAt(row);
        for (int i = 0; i < tr.cells.Length; i++)
        { // clean it out 
          Table.Controls.Remove(tr.cells[i].valLab);
          if (tr.cells[i].showInput)
          {
            Table.Controls.Remove(tr.cells[i].inptTxt);
          }
        }
      }
    }
    private void fillCells()
    {
      TableRow tr = null;
      int colcount = 0;
      String[] strvalues = new String[numValueColumns];
      for (int i = 0; i < strvalues.Length; i++)
      {
        if (i < hdrRow.cells.Length)
          strvalues[i] = hdrRow.cells[i].valLab.Text;
        else
          strvalues[i] = "Readback";
      }
      lock (hdrRow)
      {
        hdrRow.setValueColumns(strvalues);
        if (descriptionColumnVisible) colcount++;
        for (int i = 0; i < hdrRow.cells.Length; i++)
        { 
          Header.Controls.Add(hdrRow.cells[i].valLab, colcount++, 0);
          if (hdrRow.cells[i].showInput)
          {
            Header.Controls.Add(hdrRow.cells[i].inptLab, colcount++, 0);
          }
        }
        for (int row = 0; row < tblLst.Count; row++)
        {
          tr = tblLst.ElementAt(row);
          for (int i = 0; i < strvalues.Length; i++)
          {
            strvalues[i] = IsDesignMode() ? "Value " + row.ToString() : "";
          }
          tr.setValueColumns(strvalues);
          colcount = 0;
          if (descriptionColumnVisible) colcount++;
          for (int i = 0; i < tr.cells.Length; i++)
          { // fill each row in again
            Table.Controls.Add(tr.cells[i].valLab, colcount++, 0);
            if (tr.cells[i].showInput)
            {
              Table.Controls.Add(tr.cells[i].inptTxt, colcount++, 0);
            }
          }
        }
        resetColors();
      }
      Header.ColumnCount = colcount;
      Table.ColumnCount = colcount;
    }
    private void setNumberValueColumns(int value)
    {
      if (value < 1) return; // not allowed
      lock (initMutex)
      {
        Table.SuspendLayout();
        Header.SuspendLayout();
        clearCells();
        numValueColumns = value;
        fillCells();
        Table.ResumeLayout();
        Header.ResumeLayout();
        setColumnDescriptionSize(descColSize);
      }
    }

    private BorderStyle labelBorderStyle = BorderStyle.FixedSingle;
    private void setLabelBorderStyle(BorderStyle value)
    {
      if (value == labelBorderStyle) return;
      labelBorderStyle = value;
      TableRow tr;
      for (int row = 0; row < tblLst.Count; row++)
      {
        tr = tblLst.ElementAt(row);
        foreach (TableCell c in tr.cells)
          c.valLab.BorderStyle = value;
      }
    }
    protected class TableCell
    {
      public Label valLab = null;
      public Label inptLab = null;
      public TextBox inptTxt = null;
      public Boolean showInput;
      public Boolean inptValuesSet = false;
      public Type valueType = null;
      public AcopControl.AcopLink lnk = null;
      public Boolean isSelected = false;
      public String linkAddr = null;
      public String linkProperty = null;
      public int col = 0;
      public TableRow tr = null;
      public Color currentBackColor = Color.Empty;
      public Color currentForeColor = Color.Empty;
      public Boolean isWarningCondition = false;
      public TableCell(TableRow tblRow, int column, String value, Boolean showInput)
      {
        tr = tblRow;
        col = column;
        valLab = new Label() { Text = value };
        if (col > 0 && tr.cells != null && tr.cells[0].valLab != null)
        { // must be a new added cell
          valLab.Font = tr.cells[0].valLab.Font;
          valLab.BorderStyle = tr.cells[0].valLab.BorderStyle;
          valLab.TextAlign = tr.cells[0].valLab.TextAlign;
          valLab.AutoEllipsis = tr.cells[0].valLab.AutoEllipsis;
          valLab.BackColor = tr.isHeader ? tr.tbl.BackColor : tr.cells[0].valLab.BackColor;
          valLab.Anchor = tr.cells[0].valLab.Anchor;
          valLab.AutoSize = tr.cells[0].valLab.AutoSize;
        }
        else
        {
          valLab.BorderStyle = BorderStyle.FixedSingle;
          valLab.AutoEllipsis = true;
        }
        if (col > 0 && tr.cells != null && tr.cells[0].inptLab != null)
        { // must be a new header cell 
          inptLab = new Label() { Text = "Input" };
          inptLab.Font = tr.cells[0].inptLab.Font;
          inptLab.BorderStyle = tr.cells[0].inptLab.BorderStyle;
          inptLab.TextAlign = tr.cells[0].inptLab.TextAlign;
          inptLab.MouseDown += new System.Windows.Forms.MouseEventHandler(tr.tbl.inptLabelMouseDown);
          valLab.MouseDown += new System.Windows.Forms.MouseEventHandler(tr.tbl.hdrValueLabelMouseDown);

        }
        inptTxt = new TextBox() { Text = value };
        inptTxt.TextChanged += new EventHandler(tr.applyForeColor);
        this.showInput = showInput;
      }
    }
    private Dictionary<Label, TableRow> lblLookUp = new Dictionary<Label, TableRow>();
    private Dictionary<Label, TableCell> vlblLookUp = new Dictionary<Label, TableCell>();
    protected class TableRow
    {
      public Label dscLab = null;
      public TableCell[] cells = null;
      public int row = 0;
      public Boolean needResetColors = false;
      public Boolean isSelected = false;
      public String linkAddr = null;
      public String linkProperty = null;
      public Color valForeColor = Color.Empty;
      public Color valBackColor = Color.Empty;
      public Type valueType = null;
      public AcopControl.AcopLink lnk = null;
      public AcopTable tbl = null;
      public Boolean isHeader = false;
      public TableRow(AcopTable table, String desc, String[] values, int rowNum) 
       : this(table, desc, values, rowNum, false)
      {  
      }
      public TableRow(AcopTable table, String desc, String[] values, int rowNum, Boolean isHeader)
      {
        this.tbl = table;
        this.isHeader = isHeader;
        TableCell[] cols = new TableCell[values.Length];
        Boolean showInpt = false;
        for (int i = 0; i < cols.Length; i++)
        {
          showInpt = tbl.ShowInputColumn && (i < tbl.numInputColumns);
          cols[i] = new TableCell(this, i, values[i], showInpt);
        }
        dscLab = new Label() { Text = desc };
        dscLab.AutoEllipsis = true;
        cells = cols;
        row = rowNum;
      }
      public void setValueColumns(String[] values)
      {
        if (values == null || values.Length < 1) return;
        int clen = cells.Length;
        if (clen != values.Length)
        {
          Array.Resize<TableCell>(ref cells, values.Length);
        }
        Boolean showInpt = false;
        for (int i=0; i<cells.Length; i++)
        {
          if (i >= clen)
          {
            showInpt = tbl.ShowInputColumn && (i < tbl.numInputColumns);
            cells[i] = new TableCell(this, i, values[i], showInpt);
          }
          cells[i].valLab.Text = values[i];
          cells[i].inptTxt.Text = values[i];
          if (cells[i].inptLab != null)
            cells[i].inptLab.Text = "Input";
        }
      }
      public void applyForeColor(object sender, EventArgs e)
      {
        if (tbl == null) return;
        Boolean insync;
        foreach (TableCell c in cells)
        {
          if (c == null) continue; // not yet instantiated ?
          insync = c.valLab.Text.Equals(c.inptTxt.Text, StringComparison.CurrentCultureIgnoreCase);
          c.inptTxt.ForeColor = insync ? tbl.inputSyncForeColor : tbl.inputOutofSyncColor;
        }
      }
    }
    public class AcopTableCell
    {
      private String dscLab = null;
      private String valLab = null;
      private int row = 0;
      private String linkAddr = null;
      private String linkPrp = null;
      public AcopTableCell(String description, String value, int rowNumber)
      {
        dscLab = description;
        valLab = value;
        row = rowNumber;
      }
      public AcopTableCell(String address, String property, String description, String value, int rowNumber)
      {
        dscLab = description;
        valLab = value;
        row = rowNumber;
        linkAddr = address;
        linkPrp = property;
      }
      public String GetDescription() { return dscLab; }
      public String GetValue() { return valLab; }
      public int GetTableRowNumber() { return row; }
      public String GetLinkAddress() { return linkAddr; }
      public String GetLinkProperty() { return linkPrp; }
      public void SetLinkAddress(String address) { linkAddr = address; }
      public void SetLinkProperty(String property) { linkPrp = property; }
    }
    public delegate void acopTableClickDelegate(object sender, AcopTableCell cell);
    [Category("Acop.Display")]
    [Description("Occurs a table value has been clicked")]
    public event acopTableClickDelegate acopTableClick;

    public delegate void acopTableApplySelectionDelegate(object sender, AcopTable table);
    [Category("Acop.Display")]
    [Description("Occurs when the context menu item 'Apply Selection' has been clicked")]
    public event acopTableApplySelectionDelegate acopTableApplySelection;

    public void handleApplySelection(object sender)
    {
      if (!allowSelections) return;
      if (acopTableApplySelection != null)
      {
        acopTableApplySelection(sender, this);
      }
      ClearSelection();
    }
    private void resetSyncColors()
    {
      TableRow tr;
      Boolean insync = false;
      for (int i = 0; i < tblLst.Count; i++)
      {
        tr = tblLst.ElementAt(i);
        foreach (TableCell c in tr.cells)
        {
          insync = c.valLab.Text.Equals(c.inptTxt.Text, StringComparison.CurrentCultureIgnoreCase);
          c.inptTxt.ForeColor = insync ? inputSyncForeColor : inputOutofSyncColor;
        }
      }
      needsResetSyncColors = false;
    }
    private int getBitValue(String v)
    {
      if (v == null) return 0;
      v = v.ToLower();
      if (v.Equals("on", StringComparison.CurrentCultureIgnoreCase)) return 1;
      if (v.Equals("true", StringComparison.CurrentCultureIgnoreCase)) return 1;
      if (v.Equals("yes", StringComparison.CurrentCultureIgnoreCase)) return 1;
      return 0;
    }
    private Boolean needsResetSyncColors = false;
    public void handleApplyInput(object sender)
    {
      if (!showInputColumn) return;
      Boolean isgrp = getLinkGrouped();
      Boolean ismca = !isgrp && getIsMCA() && getDisplayAsMCA();
      Boolean isarray = IsOutputArrayObject();
      Boolean isstruct = getIsStructured();
      Boolean isbits = displayBits && !isarray;
      Boolean collect = !isgrp && !ismca;
      Boolean valuesChanged = false;
      Boolean valueChanged = false;
      String[] inptArray = collect ? new String[tblLst.Count] : null;
      String inptValue = null;
      String lnkAddr;
      String lnkPrp;
      object inpt = null;
      AcopControl.AcopLink lnk = null;
      int cc = 0;
      // run thru table 
      TableRow tr;
      for (int i = 0; i < tblLst.Count; i++)
      {
        tr = tblLst.ElementAt(i);
        if (selectedColumn < 0 || selectedColumn >= tr.cells.Length) continue;
        inptValue = tr.cells[selectedColumn].inptTxt.Text;
        valueChanged = !tr.cells[selectedColumn].valLab.Text.Equals(inptValue, StringComparison.CurrentCultureIgnoreCase);
        if (collect)
        { // array, bit display, or structure ... or a single value
          if (valueChanged) valuesChanged = true;
          inptArray[i] = inptValue;
        }
        else
        { // one at a time (a group or an MCA table)
          if (!valueChanged) continue;
          inpt = AcopControl.AcopLink.GetDataObjectFromString(tr.lnk, tr.valueType, inptValue);
          if (tr.cells[selectedColumn].linkAddr == null)
            tr.cells[selectedColumn].linkAddr = tr.linkAddr;
          lnkAddr = tr.cells[selectedColumn].linkAddr;
          lnkPrp = tr.cells[selectedColumn].linkProperty;
          lnk = AcopControl.AcopLink.GetAcopLink(LinkProtocol, lnkAddr, lnkPrp, null, inpt, 0, this);
          cc = applySetting(sender, inpt, lnk);
          if (cc == 0)
          {
            tr.cells[selectedColumn].inptTxt.ForeColor = inputSyncForeColor;
          }
          else
          {
            tr.cells[selectedColumn].inptTxt.Text = lnk.GetStatusString();
          }
        }
      }
      if (collect && valuesChanged)
      {
        tr = tblLst.ElementAt(0);
        if (isbits)
        { // form value from bit information
          Int32 v = 0;
          for (int i = 0; i < inptArray.Length; i++)
          {
            v |= (getBitValue(inptArray[i]) << i);
          }
          inpt = new Int32[] { v };
        }
        else
        {
          inpt = AcopControl.AcopLink.GetDataObjectFromStringArray(tr.lnk, tr.valueType, inptArray);
        }
        lnk = AcopControl.AcopLink.GetAcopLink(LinkProtocol, LinkAddress, LinkProperty, null, inpt, 0, this);
        cc = applySetting(sender, inpt, lnk);
        for (int i = 0; i < tblLst.Count; i++)
        {
          tr = tblLst.ElementAt(i);
          if (selectedColumn < 0 || selectedColumn >= tr.cells.Length) continue;
          if (cc == 0)
          {
            tr.cells[selectedColumn].inptTxt.ForeColor = inputSyncForeColor;
          }
          else
          {
            tr.cells[selectedColumn].inptTxt.Text = lnk.GetStatusString();
          }
        }
      }
      //needsResetSyncColors = true;
    }
    /// <summary>
    /// returns an object containing the input value specified in the 'Input' column
    /// </summary>
    /// <returns>an object containing the input value specified in the 'Input' column</returns>
    public object GetTableInputValues()
    {
      return GetTableInputValues(0);
    }
    /// <summary>
    /// returns an object containing the input value specified in the 'Input' column
    /// </summary>
    /// <param name="valueColumn">the table value column containing the desired input values</param>
    /// <returns>an object containing the input value specified in the 'Input' column</returns>
    public object GetTableInputValues(int valueColumn)
    {
      if (!showInputColumn) return null;
      Boolean isgrp = getLinkGrouped();
      Boolean ismca = !isgrp && getIsMCA() && getDisplayAsMCA();
      Boolean isarray = IsOutputArrayObject();
      Boolean isstruct = getIsStructured();
      Boolean isbits = displayBits && !isarray;
      Boolean collect = !isgrp && !ismca;
      String[] inptArray = collect ? new String[tblLst.Count] : null;
      String inptValue = null;
      object inpt = null;
      object[] inptObjectArray = null;
      if (collect) inptObjectArray = new object[tblLst.Count];
      // run thru table 
      TableRow tr;
      if (valueColumn < 0) valueColumn = 0;
      for (int i = 0; i < tblLst.Count; i++)
      {
        tr = tblLst.ElementAt(i);
        if (valueColumn > tr.cells.Length) continue;
        inptValue = tr.cells[valueColumn].inptTxt.Text;
        if (collect)
        { // array, bit display, or structure ... or a single value
          inptArray[i] = inptValue;
        }
        else
        { // one at a time (a group or an MCA table)
          inptObjectArray[i] = AcopControl.AcopLink.GetDataObjectFromString(tr.lnk, tr.valueType, inptValue);
        }
      }
      if (collect)
      { // single input object (which could be an array of something)
        tr = tblLst.ElementAt(0);
        if (isbits)
        { // form value from bit information
          Int32 v = 0;
          for (int i = 0; i < inptArray.Length; i++)
          {
            v |= (getBitValue(inptArray[i]) << i);
          }
          inpt = new Int32[] { v };
        }
        else
        {
          inpt = AcopControl.AcopLink.GetDataObjectFromStringArray(tr.lnk, tr.valueType, inptArray);
        }
      }
      else
      { // individual input objects for each row
        inpt = inptObjectArray;
      }
      return inpt;
    }
    /// <summary>
    /// resets the input values to the current readback values (if ShowInputColumn is true)
    /// </summary>
    public void resetInput()
    {
      TableRow tr;
      for (int i = 0; i < tblLst.Count; i++)
      {
        tr = tblLst.ElementAt(i);
        if (selectedColumn >= tr.cells.Length) continue;
        if (selectedColumn < 0)
        {
          foreach (TableCell c in tr.cells)
          {
            c.inptTxt.Text = c.valLab.Text;
            c.inptTxt.ForeColor = inputSyncForeColor;
          }
        }
        else
        {
          tr.cells[selectedColumn].inptTxt.Text = tr.cells[selectedColumn].valLab.Text;
          tr.cells[selectedColumn].inptTxt.ForeColor = inputSyncForeColor;
        }
      }
      if (selectedColumn < 0) selectedColumn = 0;
      needResetInput = false;
    }
    /// <summary>
    /// sets the table description to the string value indicated.
    /// </summary>
    /// <param name="row">the targeted table row</param>
    /// <param name="value">the desired table description</param>
    public void SetTableDescription(int row, String value)
    {
      if (!showDescriptionColumn) return;
      if (row < 0 || row >= tblLst.Count) return;
      TableRow tr = tblLst.ElementAt(row);
      tr.dscLab.Text = value;
    }
    /// <summary>
    /// returns the current table description for the row specified
    /// </summary>
    /// <param name="row">the targeted row</param>
    /// <returns>the current table description for the row specified</returns>
    public String GetTableDescription(int row)
    {
      if (!showDescriptionColumn) return "";
      if (row < 0 || row >= tblLst.Count) return "";
      TableRow tr = tblLst.ElementAt(row);
      return tr.dscLab.Text;
    }
    /// <summary>
    /// sets the table value to the string value indicated.
    /// (if the table row is in AutoUpdate mode then the value will likely 
    /// be overwritten).
    /// </summary>
    /// <param name="row">the targeted table row</param>
    /// <param name="value">the desired table value</param>
    public void SetTableValue(int row, String value)
    {
      if (row < 0 || row >= tblLst.Count) return;
      TableRow tr = tblLst.ElementAt(row);
      //tr.valLab.Text = value;
      tr.cells[0].valLab.Text = value;
    }
    public void SetTableValue(int row, int valueColumn, String value)
    {
      if (row < 0 || row >= tblLst.Count) return;
      TableRow tr = tblLst.ElementAt(row);
      if (valueColumn < 0 || valueColumn >= tr.cells.Length) return;
      tr.cells[valueColumn].valLab.Text = value;
    }
    /// <summary>
    /// returns the current table value (as string) for the row specified
    /// </summary>
    /// <param name="row">the targeted row</param>
    /// <returns>the current table value (as string) for the row specified</returns>
    public String GetTableValue(int row)
    {
      if (row < 0 || row >= tblLst.Count) return "";
      TableRow tr = tblLst.ElementAt(row);
      return tr.cells[0].valLab.Text;
    }
    public String GetTableValue(int row,int valueColumn)
    {
      if (row < 0 || row >= tblLst.Count) return "";
      TableRow tr = tblLst.ElementAt(row);
      if (valueColumn < 0 || valueColumn >= tr.cells.Length) return "";
      return tr.cells[valueColumn].valLab.Text;
    }
    /// <summary>
    /// sets the input to the value given at the given table row (if ShowInputColumn is true)
    /// </summary>
    /// <param name="row">the desired table row</param>
    /// <param name="value">the desired input value</param>
    public void SetTableInput(int row, String value)
    {
      SetTableInput(row, 0, value);
    }
    /// <summary>
    /// sets the input to the value given at the given table row (if ShowInputColumn is true)
    /// </summary>
    /// <param name="row">the desired table row</param>
    /// <param name="valueColumn">the desired table value column</param>
    /// <param name="value">the desired input value</param>
    public void SetTableInput(int row, int valueColumn, String value)
    {
      if (!showInputColumn) return;
      if (row < 0 || row >= tblLst.Count) return;
      if (valueColumn < 0) valueColumn = 0;
      TableRow tr = tblLst.ElementAt(row);
      if (valueColumn >= tr.cells.Length) return;
      tr.cells[valueColumn].inptTxt.Text = value;
    }
    /// <summary>
    /// returns the current table input value (as string) for the row specified
    /// </summary>
    /// <param name="row">the row containing the desired input value</param>
    /// <returns>the current table input value for the row specified</returns>
    public String GetTableInput(int row)
    {
      return GetTableInput(row, 0);
    }
    public String GetTableInput(int row, int valueColumn)
    {
      if (!showInputColumn) return "";
      if (row < 0 || row >= tblLst.Count) return "";
      if (valueColumn < 0) valueColumn = 0;
      TableRow tr = tblLst.ElementAt(row);
      if (valueColumn >= tr.cells.Length) return "";
      return tr.cells[valueColumn].inptTxt.Text;
    }
    private Color getValueForeColor(TableRow tr)
    {
      return getValueForeColor(tr, 0);
    }
    private Color getValueForeColor(TableRow tr, int valueColumn)
    {
      Color c = tr.valForeColor;
      int index = tr.row;
      if (applyForeColorListToColumns && numValueColumns > 1) index = valueColumn;
      if (c == null || c == Color.Empty)
      {
        Color[] ca = GetColorList(AcopUserComponent.PRIMARY_COLOR_LIST);
        if (ca != null && index >= 0) c = ca[index % ca.Length]; 
      }
      return (c != Color.Empty && c != null) ? c : LabelForeColor;
    }
    private Boolean applyForeColorListToColumns = false;
    private Boolean applyBackColorListToColumns = false;
    private void setApplyForeColorListToColumns(Boolean value)
    {
      //if (numValueColumns < 2) return; < beware of alphabetical initialization !
      applyForeColorListToColumns = value;
      resetColors();
    }
    private void setApplyBackColorListToColumns(Boolean value)
    {
      //if (numValueColumns < 2) return;
      applyBackColorListToColumns = value;
      resetColors();
    }
    private Color getValueBackColor(TableRow tr)
    {
      return getValueBackColor(tr, 0);
    }
    private const String BACKCOLOR_LIST = "Backcolor";
    private Color getValueBackColor(TableRow tr, int valueColumn)
    {
      Color color = tr.valBackColor;
      int index = tr.row;
      if (applyBackColorListToColumns && numValueColumns > 1) index = valueColumn;
      if (color == null || color == Color.Empty)
      {
        Color[] ca = GetColorList(BACKCOLOR_LIST);
        if (ca != null && index >= 0) color = ca[index % ca.Length];
      }

      return (color != Color.Empty && color != null) ? color : LabelBackColor;
    }
    /// <summary>
    /// Sets the value cell at the given row to the desired ForeColor
    /// </summary>
    /// <param name="value">the desired ForeColor for the value cell entry</param>
    /// <param name="row"><the table row containing the cell/param>
    public void SetValueForeColor(Color value, int row)
    {
      if (row < 0 || row >= tblLst.Count) return;
      TableRow tr = tblLst.ElementAt(row);
      tr.valForeColor = value;
      foreach (TableCell c in tr.cells)
        c.valLab.ForeColor = getValueForeColor(tr);
    }
    /// <summary>
    /// Sets the value cell at the given row to the desired Back Color
    /// </summary>
    /// <param name="value">the desired Back Color for the value cell entry</param>
    /// <param name="row"><the table row containing the cell/param>
    public void SetValueBackColor(Color value, int row)
    {
      if (row < 0 || row >= tblLst.Count) return;
      TableRow tr = tblLst.ElementAt(row);
      tr.valBackColor = value;
      foreach (TableCell c in tr.cells)
        c.valLab.BackColor = getValueBackColor(tr, c.col);
    }
    private void selectCell(int row, Boolean isSelected)
    {
      selectCell(row, 0, isSelected);
    }
    private void selectCell(int row, int valueColumn, Boolean isSelected)
    {
      if (!allowSelections) return;
      if (row < 0 || row >= tblLst.Count) return;
      TableRow tr = tblLst.ElementAt(row);
      if (valueColumn < 0 || valueColumn >= tr.cells.Length) return;
      tr.isSelected = isSelected;
      tr.cells[valueColumn].valLab.BackColor = isSelected ? selectionBackColor : getValueBackColor(tr, valueColumn);
      tr.cells[valueColumn].valLab.ForeColor = isSelected ? selectionForeColor : getValueForeColor(tr, valueColumn);
    }
    private void selectRangeOfCells(int row)
    {
      if (!allowSelections) return;
      int i;
      for (i=0; i < tblLst.Count; i++)
      { // first the first one currently selected
        if (tblLst.ElementAt(i).isSelected) break;
      }
      if (i < tblLst.Count)
      {
        if (row > i) for (int k = i; k <= row; k++) selectCell(k,true);
        else for (int k = row; k <= i; k++) selectCell(k,true);
      }
    }
    private void selectRangeOfCells(int row, int valueColumn)
    {
      if (!allowSelections) return;
      if (valueColumn < 0) valueColumn = 0;
      TableRow tr;
      int i;
      for (i = 0; i < tblLst.Count; i++)
      { // first the first one currently selected
        tr = tblLst.ElementAt(i);
        if (valueColumn < tr.cells.Length)
        {
          if (tr.cells[valueColumn].isSelected) break;
        }
      }
      if (i < tblLst.Count)
      {
        if (row > i) for (int k = i; k <= row; k++) selectCell(k, valueColumn, true);
        else for (int k = row; k <= i; k++) selectCell(k, valueColumn, true);
      }
    }
    private void clearSelections(int except)
    {
      clearSelections(except, 0);
    }
    private void clearSelections(int except, int valueColumn)
    {
      if (!allowSelections) return;
      int i;
      for (i = 0; i < tblLst.Count; i++)
      { // first the first one currently selected
        if (i == except) continue;
        TableRow tr = tblLst.ElementAt(i);
        if (tr.isSelected)
        {
          tr.isSelected = false;
          for (int c = 0; c < tr.cells.Length; c++)
          {
            if (valueColumn < 0 || c == valueColumn)
            {
              tr.cells[c].valLab.BackColor = getValueBackColor(tr, i);
              tr.cells[c].valLab.ForeColor = getValueForeColor(tr, i);
              if (valueColumn >= 0) break;
            }
          }
        }
      }
    }
    /// <summary>
    /// returns the selected status of the given row
    /// </summary>
    /// <param name="row"></param>
    /// <returns>true if the row input is in a selected state</returns>
    public Boolean RowIsSelected(int row)
    {
      if (!allowSelections) return false;
      if (row < 0 || row >= tblLst.Count) return false;
      return tblLst.ElementAt(row).isSelected;
    }
    /// <summary>
    /// returns an array of all table cells (minus header)
    /// </summary>
    /// <returns>an array of all table cells</returns>
    public AcopTableCell[] GetCells()
    {
      return GetCells(0);
    }
    /// <summary>
    /// returns an array of all table cells (minus header)
    /// </summary>
    /// <param name="valueColumn">the specific column for which the cells are desired</param>
    /// <returns>an array of all table cells</returns>
    public AcopTableCell[] GetCells(int valueColumn)
    {
      String dsc, val, prp;
      List<AcopTableCell> lst = new List<AcopTableCell>();
      TableRow tr;
      if (valueColumn < 0) valueColumn = 0;
      for (int i = 0; i < tblLst.Count; i++)
      { 
        tr = tblLst.ElementAt(i);
        if (valueColumn >= tr.cells.Length) valueColumn = tr.cells.Length - 1;
        dsc = tr.dscLab != null ? tr.dscLab.Text : "";
        val = tr.cells[valueColumn].valLab != null ? tr.cells[valueColumn].valLab.Text : "";
        prp = tr.cells[valueColumn].linkProperty;
        if (prp == null) prp = tr.linkProperty;
        lst.Add(new AcopTableCell(tr.linkAddr, prp, dsc, val, tr.row));
      }
      return lst.ToArray();
    }
    /// <summary>
    /// returns an array of those cells which are currently selected
    /// </summary>
    /// <returns>an array of those cells which are currently selected</returns>
    public AcopTableCell[] GetSelectedCells()
    {
      return GetSelectedCells(0);
    }
    /// <summary>
    /// returns an array of those cells which are currently selected
    /// </summary>
    /// <param name="valueColumn">the specific column for which the cells are desired</param>
    /// <returns>an array of those cells which are currently selected</returns>
    public AcopTableCell[] GetSelectedCells(int valueColumn)
    {
      if (!allowSelections) return null;
      List<AcopTableCell> lst = getSelectedCellList(valueColumn);
      return lst.ToArray();
    }
    private List<AcopTableCell> getSelectedCellList(int valueColumn)
    {
      String dsc, val, prp;
      List<AcopTableCell> lst = new List<AcopTableCell>();
      TableRow tr;
      if (valueColumn < 0) valueColumn = 0;
      for (int i = 0; i < tblLst.Count; i++)
      { // first the first one currently selected
        tr = tblLst.ElementAt(i);
        if (!tr.isSelected) continue;
        dsc = tr.dscLab != null ? tr.dscLab.Text : "";
        val = tr.cells[valueColumn].valLab != null ? tr.cells[valueColumn].valLab.Text : "";
        prp = tr.cells[valueColumn].linkProperty;
        lst.Add(new AcopTableCell(tr.linkAddr, prp, dsc, val, tr.row));
      }
      return lst;
    }
    /// <summary>
    /// de-selects all cells.
    /// </summary>
    public void ClearSelection()
    {
      if (!allowSelections) return;
      clearSelections(-1);
    }
    /// <summary>
    /// gets a trend of history points according to the depthInSeconds
    /// </summary>
    /// <param name="row">table row containing the desired link</param>
    /// <param name="depthInSeconds">start time in UTC (seconds since Jan 1 1970)</param>
    /// <returns></returns>
    public double[][] GetTrendPoints(int row, int depthInSeconds)
    {
      if (row < 0 || row >= tblLst.Count) return null;
      double stopt = getUtcNow();
      int len = depthInSeconds;
      double[][] hdata = new double[2][];
      hdata[0] = new double[len];
      hdata[1] = new double[len];
      double startt = stopt - depthInSeconds;
      TableRow tr = tblLst.ElementAt(row);
      AcopControl.AcopLink lnk = tr.lnk;
      if (lnk == null) return null;
      Boolean isstruct = getIsStructured();
      Boolean ismca = getIsMCA() && getDisplayAsMCA();
      Boolean isarray = !ismca && IsOutputArrayObject();
      int flags = 0;
      if (isstruct || isarray) flags = row;
      if (lnk.GetHistoryPoints(hdata, ref len, startt, stopt, flags, 500) == 0)
      {
        if (len < depthInSeconds && lnk.IsAttached())
        {
          hdata[0][len] = lnk.GetOutputYAxis()[0];
          hdata[1][len] = stopt;
          len++;
        }
      }
      Array.Resize(ref hdata[0], len);
      Array.Resize(ref hdata[1], len);
      return hdata;
    }
    private int selectedColumn = 0;
    private void inptLabelMouseDown(object sender, MouseEventArgs e)
    {
      if (sender is Label)
      {
        if (e.Button == MouseButtons.Right)
        {
          selectedColumn = 0;
          for (int i=0; i<hdrRow.cells.Length; i++)
          {
            if (hdrRow.cells[i].inptLab.Equals(sender as Label))
            {
              selectedColumn = i;
              break;
            }
          }
          Point pos = this.PointToClient(Cursor.Position);
          im.Show(this, pos, LeftRightAlignment.Right);
          return;
        }
      }
    }
    private void hdrValueLabelMouseDown(object sender, MouseEventArgs e)
    {
      if (e.Button == MouseButtons.Right)
      {
        if (acopTableApplySelection != null && allowSelections)
        {
          Point pos = this.PointToClient(Cursor.Position);
          cm.Show(this, pos, LeftRightAlignment.Right);
          ClearSelection();
        }
        return;
      }
    }
    private void valueLabelMouseDown(object sender, MouseEventArgs e)
    {
      if (sender is Label)
      {
        TableRow tr;
        TableCell tc;
        AcopTableCell cell;
        if (!vlblLookUp.TryGetValue((Label)sender,out tc))
        {
          tr = tblLst.ElementAt(0);
          tc = tr.cells[0];
        }
        else
        {
          if (e.Button != MouseButtons.Left) return; // only accept normal left click
          tr = tc.tr;
          if (Control.ModifierKeys == Keys.Shift)
          { // check selection range
            selectRangeOfCells(tr.row, tc.col);
          }
          else
          {
            if (Control.ModifierKeys != Keys.Control) clearSelections(tr.row,tc.col);
            selectCell(tr.row, tc.col, !tr.isSelected);
          }
        }
        if (acopTableClick == null) return; // no delegate
        String dsc = tr.dscLab != null ? tr.dscLab.Text : "";
        String val = tc.valLab != null ? tc.valLab.Text : "";
        cell = new AcopTableCell(tr.linkAddr, tc.linkProperty, dsc, val, tr.row);
        acopTableClick(sender, cell);
      }
    }

    private int descColSize = 50; // percent
    private void setHeaderColumnSize(int value)
    {
      int wd = Table.Width;
      float cwd = (float)(descColSize - value);
      if (cwd < 1) return;
      Header.SuspendLayout();
      if (descriptionColumnVisible) Header.Controls.Remove(hdrRow.dscLab);
      foreach (TableCell c in hdrRow.cells)
      {
        Header.Controls.Remove(c.valLab);
        if (c.showInput) Header.Controls.Remove(c.inptLab);
      }
      Header.ColumnStyles.Clear();
      Header.RowStyles.Clear();
      hdrRow.dscLab.Width = wd * 2;
      Header.RowStyles.Add(new System.Windows.Forms.RowStyle(SizeType.AutoSize));
      Header.ColumnStyles.Add(new ColumnStyle(SizeType.Percent, cwd));

      int columncnt = 0;
      if (descriptionColumnVisible) Header.Controls.Add(hdrRow.dscLab, columncnt++, 0);
      foreach (TableCell c in hdrRow.cells)
      {
        c.valLab.Width = wd * 2;
        Header.ColumnStyles.Add(new ColumnStyle(SizeType.Percent, 100.0F - cwd));
        Header.Controls.Add(c.valLab, columncnt++, 0);
        if (c.showInput)
        {
          Header.ColumnStyles.Add(new ColumnStyle(SizeType.Percent, 100.0F - cwd));
          Header.Controls.Add(c.inptLab, columncnt++, 0);
        }
      }
      Header.ResumeLayout();
    }
    private void setColumnDescriptionSize(int value)
    {
      if (value < 10) value = 10;
      if (value > 90) value = 90;
      //if (value == descColSize) return;
      descColSize = value;
      if (!descriptionColumnVisible) return;
      Table.SuspendLayout();
      Header.SuspendLayout();
      TableRow tr;
      // description column is there !
      Header.Controls.Remove(hdrRow.dscLab);
      foreach (TableCell c in hdrRow.cells)
      {
        Header.Controls.Remove(c.valLab);
        if (c.showInput) Header.Controls.Remove(c.inptLab);
      }
      for (int row = 0; row < tblLst.Count; row++)
      {
        tr = tblLst.ElementAt(row);
        Table.Controls.Remove(tr.dscLab);
        foreach (TableCell c in tr.cells)
        {
          Table.Controls.Remove(c.valLab);
          if (c.showInput) Table.Controls.Remove(c.inptTxt);
        }
      }
      int wd = Table.Width;
      float cwd = (float)value;
      // get the label size 'about right' 
      // the 'make it really long' strategy makes the table 'flicker' when 
      // it readjusts the size !
      int nvcols = hdrRow.cells.Length;
      int twd = wd - (value * wd) / (100 * nvcols);
      int hcwd = twd / nvcols;

      Table.ColumnStyles.Clear();
      Table.RowStyles.Clear();
      Header.ColumnStyles.Clear();
      Header.RowStyles.Clear();
      hdrRow.dscLab.Width = wd * 2;
      int columncnt = 0;
      Header.RowStyles.Add(new System.Windows.Forms.RowStyle(SizeType.AutoSize));
      Header.ColumnStyles.Add(new ColumnStyle(SizeType.Percent, cwd));
      Header.Controls.Add(hdrRow.dscLab, columncnt++, 0);
      foreach (TableCell c in hdrRow.cells)
      {
        Header.ColumnStyles.Add(new ColumnStyle(SizeType.Percent, 100.0F - cwd));
        c.valLab.Width = wd * 2;
        Header.Controls.Add(c.valLab, columncnt++, 0);
        Header.ColumnStyles.Add(new ColumnStyle(SizeType.Percent, 100.0F - cwd));
        if (c.showInput) Header.Controls.Add(c.inptLab, columncnt++, 0);
      }
      for (int row = 0; row < tblLst.Count; row++)
      {
        tr = tblLst.ElementAt(row);
        tr.dscLab.Width = wd * 2;
        columncnt = 0;
        Table.RowStyles.Add(new System.Windows.Forms.RowStyle(SizeType.AutoSize));
        Table.ColumnStyles.Add(new ColumnStyle(SizeType.Percent, cwd));
        Table.Controls.Add(tr.dscLab, columncnt++, row);
        foreach (TableCell c in tr.cells)
        {
          c.valLab.Width = hcwd; // wd * 2;
          Table.ColumnStyles.Add(new ColumnStyle(SizeType.Percent, 100.0F - cwd));
          Table.Controls.Add(c.valLab, columncnt++, row);
          Table.ColumnStyles.Add(new ColumnStyle(SizeType.Percent, 100.0F - cwd));
          if (c.showInput) Table.Controls.Add(c.inptTxt, columncnt++, row);
        }
      }
      Header.ResumeLayout();
      Table.ResumeLayout();
    }

    private LinkedList<TableRow> tblLst = new LinkedList<TableRow>();
    private void addTableEntry(String desc, String value)
    {
      addTableEntry(desc, value, -1);
    }
    private void addTableEntry(String desc, String value, String inpt)
    {
      int ncols = hdrRow == null ? 1 : hdrRow.cells.Length;
      String[] vals = Enumerable.Repeat(value, ncols).ToArray();
      addTableEntry(desc, vals, inpt, -1, false);
    }
    private TableRow addTableEntry(String desc, String value, int row)
    {
      return addTableEntry(desc, value, row, false);
    }
    private TableRow addTableEntry(String desc, String[] values, int row)
    {
      return addTableEntry(desc, values, "Input", row, false);
    }
    private TableRow addTableEntry(String desc, String value, int row, Boolean startup)
    {
      int ncols = hdrRow == null ? 1 : hdrRow.cells.Length;
      String[] vals = Enumerable.Repeat(value, ncols).ToArray();
      return addTableEntry(desc, vals, "Input", row, startup);
    }
    private TableRow hdrRow = null;
    // values gives all value-column values for a row ...
    private TableRow addTableEntry(String desc, String[] values, String inpt, int row, Boolean startup)
    {
      if (row < 0) row = tblLst.Count;
      while (row > tblLst.Count)
      {
        addTableEntry("", "");
      }
      int columncnt = 0;
      int sz = this.Width;
      int ht = this.Height / (numTableRows + 1);
      ht -= 4;
      int wd = sz * 9 / 50; // 45 % of table width
      TableRow tr = null;
      if (hdrRow == null)
      {
        hdrRow = new TableRow(this, desc, values, 0, true);
        Font font = new System.Drawing.Font(hdrRow.dscLab.Font.Name, hdrRow.dscLab.Font.Size, FontStyle.Bold | FontStyle.Italic);
        hdrRow.dscLab.Font = font;
        hdrRow.dscLab.AutoSize = false;
        hdrRow.dscLab.Width = Width * 2;
        hdrRow.dscLab.TextAlign = descLabelAlignment;
        foreach (TableCell c in hdrRow.cells)
        {
          c.inptLab = new Label() { Text = inpt };
          c.valLab.Font = font;
          c.inptLab.Font = font;
          c.inptLab.TextAlign = ContentAlignment.MiddleCenter;
          c.valLab.AutoSize = false;
          c.valLab.Width = Width * 2;
          c.valLab.BorderStyle = BorderStyle.None;
          c.valLab.TextAlign = valueLabelAlignment;
          c.inptLab.MouseDown += new System.Windows.Forms.MouseEventHandler(inptLabelMouseDown);
          c.valLab.MouseDown += new System.Windows.Forms.MouseEventHandler(hdrValueLabelMouseDown);
        }
        Header.ColumnStyles.Add(new ColumnStyle(SizeType.Absolute, wd));
        Header.RowStyles.Add(new System.Windows.Forms.RowStyle(SizeType.AutoSize));
        columncnt = 0;
        if (showDescriptionColumn) Header.Controls.Add(hdrRow.dscLab, columncnt++, 0);
        foreach (TableCell c in hdrRow.cells)
        {
          Header.Controls.Add(c.valLab, columncnt++, 0);
          if (c.showInput)
          {
            c.valLab.ContextMenu = cm;
            c.inptLab.ContextMenu = cm;
            c.inptLab.TextAlign = valueLabelAlignment;
            c.inptLab.Anchor = AnchorStyles.Top;// | AnchorStyles.Bottom;
            Header.Controls.Add(c.inptLab, columncnt++, 0);
          }
        }
        if (row == 0)
        {
          desc = "description";
          for (int i=0; i<values.Length; i++)
            values[i] = "value 0";
        }
      }
      if (row == tblLst.Count)
      {
        tr = new TableRow(this, desc, values, row);
        tblLst.AddLast(tr);
        tr.dscLab.Anchor = AnchorStyles.Top;
        tr.dscLab.Font = tableFont;
        tr.dscLab.AutoSize = false;
        tr.dscLab.Width = sz * 2;
        tr.dscLab.TextAlign = descLabelAlignment;
        columncnt = 0;
        if (showDescriptionColumn) Table.Controls.Add(tr.dscLab, columncnt++, row);
        foreach (TableCell c in tr.cells)
        {
          vlblLookUp.Add(c.valLab, c);
          c.valLab.BorderStyle = labelBorderStyle;
          c.valLab.Anchor = AnchorStyles.Top | AnchorStyles.Left | AnchorStyles.Right;
          c.valLab.AutoSize = false;
          c.valLab.Width = sz * 2;
          c.valLab.ForeColor = getValueForeColor(tr, c.col);
          c.valLab.BackColor = getValueBackColor(tr, c.col);
          c.valLab.Font = tableFont;
          c.valLab.MouseDown += new System.Windows.Forms.MouseEventHandler(valueLabelMouseDown);
          c.valLab.TextAlign = valueLabelAlignment;
          Table.Controls.Add(c.valLab, columncnt++, row);
          if (c.showInput)
          {
            c.inptTxt.Anchor = AnchorStyles.Top;
            c.inptTxt.Font = tableFont;
            c.inptTxt.ForeColor = inputSyncForeColor;
            c.inptTxt.BackColor = inputBackColor;
            Table.Controls.Add(c.inptTxt, columncnt++, row);
          }
        }
        Table.ColumnStyles.Add(new ColumnStyle(SizeType.Absolute, wd));
        Table.RowStyles.Add(new System.Windows.Forms.RowStyle(SizeType.AutoSize));
      }
      else if (row >= 0 && row < tblLst.Count)
      {
        tr = tblLst.ElementAt(row);
        if (!startup)
        {
          tr.dscLab.Text = desc;
          for (int i = 0; i < values.Length && i < tr.cells.Length; i++)
          {
            tr.cells[i].valLab.Text = values[i];
            if (!tr.cells[i].inptValuesSet)
              tr.cells[i].inptTxt.Text = values[i];
          }
        }
        else
        {
          foreach (TableCell c in tr.cells)
          {
            c.valLab.ForeColor = getValueForeColor(tr, c.col);
            c.valLab.BackColor = getValueBackColor(tr, c.col);
          }
        }
      }
      tableCapacityStale = true;
      return tr;
    }
    private TableRow updateTableEntry(String desc, String value, String inpt, int row, int valueColumn)
    {
      TableRow tr = null;
      if (valueColumn < 0) valueColumn = 0;
      if (row >= 0 && row < tblLst.Count)
      {
        tr = tblLst.ElementAt(row);
        if (valueColumn >= tr.cells.Length) return null;  
        tr.dscLab.Text = desc;
        tr.cells[valueColumn].valLab.Text = value;
        if (!tr.cells[valueColumn].inptValuesSet)
          tr.cells[valueColumn].inptTxt.Text = value;
      }
      return tr;
    }
    private void removeTableRowsAfter(int row)
    {
      if (row < 0) return;
      if (row > tblLst.Count) return;
      TableRow tr = null;
      Table.SuspendLayout();
      for (int i = 0; i < tblLst.Count; i++)
      {
        tr = tblLst.ElementAt(i);
        if (tr.row >= row)
        {
          Table.Controls.Remove(tr.dscLab);
          foreach (TableCell c in tr.cells)
          {
            Table.Controls.Remove(c.valLab);
            if (c.showInput) Table.Controls.Remove(c.inptTxt);
          }
          tblLst.Remove(tr);
          Table.RowCount--;
          i--; // back up
        }
      }
      Table.ResumeLayout();
    }
    private void removeTableRow(int row)
    {
      Boolean done = false;
      if (row < 0) return;
      if (row > tblLst.Count) return;
      TableRow tr = null;
      Table.SuspendLayout();
      for (int i=0; i<tblLst.Count; i++)
      {
        tr = tblLst.ElementAt(i);
        if (!done && tr.row == row)
        {
          Table.Controls.Remove(tr.dscLab);
          foreach (TableCell c in tr.cells)
          {
            Table.Controls.Remove(c.valLab);
            if (c.showInput) Table.Controls.Remove(c.inptTxt);
          }
          tblLst.Remove(tr);
          Table.RowCount--;
          i--; // back up
          done = true;
        }
        if (tr.row > row) tr.row--;
      }
      Table.ResumeLayout();
    }
    private void removeHScroll()
    {
      Table.SuspendLayout();
      if (Table.HorizontalScroll.Visible)
      {
        int sz = Table.Width - (2 * SystemInformation.VerticalScrollBarWidth);
        int wd = sz * 9 / 50; // 45 % of table width
        //Table.Padding = new Padding(0, 0, wd, 0);
        foreach (Control ctl in Table.Controls)
        {
          ctl.Width = sz;
        }
      }
      Table.ResumeLayout();
    }
    private Control[] getTableControls()
    {
      List<Control> ca = new List<Control>();
      foreach (TableRow tr in tblLst)
      {
        if (showDescriptionColumn) ca.Add(tr.dscLab);
        foreach (TableCell c in tr.cells)
        {
          ca.Add(c.valLab);
          if (c.showInput) ca.Add(c.inptTxt);
        }
      }
      return ca.ToArray();
    }
    private int lastLinkStatus = 0;
    private const int AT_CHANNEL = 0x10;

    private double thresholdHigh = MAX_THRESHOLD;
    private double thresholdLow = -MAX_THRESHOLD;

    private String thresholdMatch = "";

    private String thresholdList = null;
    protected class ValueThreshold
    {
      public double hi = MAX_THRESHOLD;
      public double low = -MAX_THRESHOLD;
      public String match = "";
    }
    private ValueThreshold[] thresholds = null;
    public String GetThresholdListString()
    {
      return getThresholdList();
    }
    private String getThresholdList()
    {
      return thresholdList == null ? "" : thresholdList;
    }
    public void SetThresholdListString(String value)
    {
      setThresholdList(value);
    }
    private void setThresholdList(String value)
    {
      thresholdList = value;
      if (thresholdList != null && thresholdList.Length > 0)
      {
        String[] tla = thresholdList.Split(new char[] { ';' });
        thresholds = new ValueThreshold[tla.Length];
        String[] parts = null;
        char[] mmtok = new char[] { ':' };
        for (int i=0; i<tla.Length; i++)
        {
          thresholds[i] = new ValueThreshold();
          parts = tla[i].Split(mmtok, StringSplitOptions.RemoveEmptyEntries);
          if (parts.Length > 1)
          {
            if (!double.TryParse(parts[0],out thresholds[i].low) ||
                !double.TryParse(parts[1], out thresholds[i].hi))
            {
              thresholds[i].hi = MAX_THRESHOLD;
              thresholds[i].low = -MAX_THRESHOLD;
            }
          }
          else
          {
            thresholds[i].match = tla[i];
          }
        }
      }
      else
      {
        thresholds = null;
      }
    }
    private Color labelLowWarnForeColor = Color.Empty;
    private Color labelHiWarnForeColor = Color.Empty;
    private Color labelLowWarnBackColor = Color.Empty;
    private Color labelHiWarnBackColor = Color.Empty;
    private Color labelForeColor = DefaultForeColor;
    private Color labelBackColor = DefaultBackColor;
    private Color selectionBackColor = Color.LightSteelBlue;
    private Color selectionForeColor = DefaultForeColor;
    private Color inputSyncForeColor = DefaultForeColor;
    private Color inputBackColor = Color.White;
    private Color inputOutofSyncColor = Color.DarkRed;

    private void setInputSyncForeColor(Color value)
    {
      inputSyncForeColor = value;
      TableRow tr;
      for (int i = 0; i < numTableRows && i < numTableEntries; i++)
      {
        tr = tblLst.ElementAt(i);
        foreach (TableCell c in tr.cells)
        {
          c.inptTxt.ForeColor = value;
        }
      }
    }
    private void setInputBackColor(Color value)
    {
      inputBackColor = value;
      TableRow tr;
      for (int i = 0; i < numTableRows && i < numTableEntries; i++)
      {
        tr = tblLst.ElementAt(i);
        foreach (TableCell c in tr.cells)
        {
          c.inptTxt.BackColor = value;
        }
      }
    }
    private Font tableFont = DefaultFont;
    private Font descFont = DefaultFont;
    private Boolean needResetColors = false;
    private Boolean allowSelections = true;
    private void setAllowSelections(Boolean value)
    {
      allowSelections = value;
    }
    public void applyErrorStatus(String sts, Boolean applyToTTT)
    {
      if (applyToTTT) applyToolTipText(sts);
      TableRow tr;
      for (int i = 0; i < numTableRows && i < numTableEntries; i++)
      {
        tr = tblLst.ElementAt(i);
        foreach (TableCell c in tr.cells)
        {
          c.valLab.ForeColor = Color.White;
          c.valLab.BackColor = Color.DarkRed;
          c.valLab.Text = sts;
        }
        tr.needResetColors = true;
      }
      needResetColors = true;
    }
    private Boolean tableAutoScroll = true;
    private void setTableAutoScroll(Boolean value)
    {
      Table.SuspendLayout();
      this.Table.AutoScroll = value;
      tableAutoScroll = value;
      int pad = SystemInformation.VerticalScrollBarWidth;
      if (value)
      {
        Padding padding = new Padding(1, 1, pad / 4, 1);
        Table.Padding = padding;
        Header.Padding = padding;
      }
      Table.ResumeLayout();
    }
    private void applyTableLabels(String[] labels)
    {
      applyTableLabels(labels, 0);
    }
    private void applyTableLabels(String[] labels,int valueColumn)
    {
      if (labels == acopTablelabels) return;
      if (labels == null || labels.Length == 0) return;
      if (acopTablelabels != null && acopTablelabels.Length == labels.Length)
      {
        Boolean labelsEqual = true;
        for (int i=0; labelsEqual && i<labels.Length; i++)
        {
          labelsEqual = labels[i].Equals(acopTablelabels[i],StringComparison.CurrentCultureIgnoreCase);
        }
        if (labelsEqual) return;
      }
      Table.SuspendLayout();
      TableRow tr = null;
      String dsc = "";
      String[] vals = new String[hdrRow.cells.Length];
      for (int i = 0; i < vals.Length; i++) vals[i] = "";
      for (int i = 0; i < numTableRows; i++)
      {
        dsc = (i < labels.Length) ? labels[i] : "";
        tr = tblLst.ElementAt(i);
        if (i < labels.Length)
        {
          tr = addTableEntry(labels[i], vals, i); 
        }
        else
        {
          removeTableRow(i);
        }
      }
      numTableRows = labels.Length;
      Table.RowCount = numTableRows;
      tableCapacityStale = true;
      Table.ResumeLayout();
    }
    private Boolean tableCapacityStale = true;
    private String[] acopTablelabels = null;
    private void setMCAHeaderRow(int valueColumn)
    {
      if (valueColumn < 0 || valueColumn >= hdrRow.cells.Length) return;
      if (valueColumn < valHdrList.Length && valHdrList[valueColumn].Length > 0)
      {
        return;
      }
      lock (initMutex)
      {
        AcopTransport.DisplaySettings dsp = null;
        TableCell c = hdrRow.cells[valueColumn];
        c.valLab.Text = c.linkProperty;
        if (numValueColumns > 1) setUnits("");
        dsp = acquireAndApplyDisplaySettings(c.linkAddr, c.linkProperty, ref acopTablelabels, false);
        String unitsStr = getUnits();
        if (unitsStr.Length > 0) c.valLab.Text += " (" + unitsStr + ")";
      }
    }
    private void setStructHeaderRow(int valueColumn, AcopControl.AcopLink lnk)
    {
      if (valueColumn < 0 || valueColumn >= hdrRow.cells.Length) return;
      if (valueColumn < valHdrList.Length && valHdrList[valueColumn].Length > 0)
      {
        return;
      }
      lock (initMutex)
      {
        if (descHdrText == null || descHdrText.Length == 0) hdrRow.dscLab.Text = "Fields";
        TableCell c = hdrRow.cells[valueColumn];
        String[] parts = lnk.GetAddress().Split(new char[] { '/' }, StringSplitOptions.RemoveEmptyEntries);
        c.valLab.Text = parts.Length > 2 ? parts[2] : "device "+c.col;
      }

    }
    private Boolean setMCATableEntries(AcopControl.AcopLink lnk, AcopTransport.DisplaySettings dsp, int valueColumn)
    {
      String unitsStr = "";
      String dscStr = null;
      TableRow tr = null;
      lock (hdrRow)
      {
        int ncols = hdrRow == null ? 1 : hdrRow.cells.Length;
        String[] dscArray = getDescArray();
        String[] vals = Enumerable.Repeat(String.Empty, ncols).ToArray();
        String ttt = lnk.GetAddress() + "[" + lnk.GetProperty() + "]";
        Boolean hasDescArray = (dscArray != null && dscArray.Length > 0);
        if (acopTablelabels == null) return false;
        if (descHdrText == null || descHdrText.Length == 0) hdrRow.dscLab.Text = "Device";
        Boolean applyHeaderText = true;
        if (valHdrList.Length > valueColumn && valHdrList[valueColumn].Length > 0)
        {
          applyHeaderText = false;
        }
        if (applyHeaderText)
        {
          if (valueColumn >= hdrRow.cells.Length) return true;
          hdrRow.cells[valueColumn].valLab.Text = lnk.GetProperty();
          unitsStr = getUnits();
          if (unitsStr.Length == 0) unitsStr = dsp.GetYUnits();
          if (unitsStr.Length > 0) hdrRow.cells[valueColumn].valLab.Text += " (" + unitsStr + ")";
        }
        hdrRow.cells[valueColumn].linkProperty = lnk.GetProperty();
        hdrRow.cells[valueColumn].linkAddr = lnk.GetAddress();
        if (tableCapacityStale && displaySettingsAcquired && isInsideUpdate)
        {
          displaySettingsAcquired = false;
          return true;
        }
        int nlabels = acopTablelabels.Length;
        int cap = getAutoLinkCapacity();
        if (cap > 1 && nlabels > cap) nlabels = cap;
        if (numTableRows < nlabels)
        {
          numTableRows = Math.Min(nlabels, maxTableRows);
          Table.RowCount = numTableRows;
        }
        for (int i = 0; i < nlabels && i < numTableRows; i++)
        {
          dscStr = acopTablelabels[i];
          if (hasDescArray && i < dscArray.Length)
          {
            dscStr = dscArray[i];
          }
          tr = addTableEntry(dscStr, vals, i);
          applyToolTipText(tr.dscLab, ttt);
          foreach (TableCell c in tr.cells)
            applyToolTipText(c.valLab, ttt);
        }
      }
      return false;
    }
    private void initTransportInfo()
    {
      if (!IsDesignMode())
      {
        lock (initMutex)
        {
          Table.SuspendLayout();
          AcopTransport.DisplaySettings dsp = null;
          Boolean ismca = (getIsMCA() && getDisplayAsMCA());
          TableRow tr = null;
          String unitsStr = "";
          String[] dscArray = getDescArray();
          String dscStr = null;
          int ncols = hdrRow == null ? 1 : hdrRow.cells.Length;
          String[] vals = Enumerable.Repeat(String.Empty, ncols).ToArray();
          Boolean hasDescArray = (dscArray != null && dscArray.Length > 0);
          if (!AutoLinkAttach && !AutoLinkUpdate)
          {
            displaySettingsAcquired = true;
            Table.ResumeLayout();
            return;
          }
          if (!getLinkGrouped())
          { // not grouped !
            dsp = acquireAndApplyDisplaySettings(LinkAddress, LinkProperty, ref acopTablelabels, false);
            if (dsp == null || dsp.SettingsStale())
            {
              Table.ResumeLayout();
              return;
            }
            displaySettingsAcquired = true;
            if (thresholdHigh == MAX_THRESHOLD) thresholdHigh = dsp.GetYMaximum();
            if (thresholdLow == -MAX_THRESHOLD) thresholdLow = dsp.GetYMinimum();
            String ttt = LinkAddress + "[" + LinkProperty + "]";
            AcopControl.AcopLink lnk = GetAttachedLink();
            object dout = lnk.GetOutputDataObject();
            if (getIsMCA() && getDisplayAsMCA() && acopTablelabels != null)
            {
              setMCATableEntries(lnk, dsp, 0);
            }
            else if (dout is Array)
            {
              tr = addTableEntry(getDescription(), "", 0);
              int len = ((Array)dout).Length;
              if (len == 1 && displayBits) len = numTableRows;
              if (numTableRows < len)
              {
                numTableRows = Math.Min(len, maxTableRows);
                Table.RowCount = numTableRows;
              }
              for (int i = 0; i < len && i < numTableRows; i++)
              {
                if (hasDescArray && i < dscArray.Length)
                {
                  dscStr = dscArray[i];
                }
                else
                {
                  dscStr = getDescription() + " Element " + (i + 1);
                }
                tr = addTableEntry(dscStr, vals, i);
                applyToolTipText(tr.dscLab, ttt);
                foreach (TableCell c in tr.cells)
                  applyToolTipText(c.valLab, ttt);
              }
            }
            else
            {
              tr = addTableEntry(getDescription(), vals, 0);
              applyToolTipText(tr.dscLab, ttt);
              foreach (TableCell c in tr.cells)
                applyToolTipText(c.valLab, ttt);
            }
            int idx = 0;
            assignTableValue(lnk, ref idx, ismca);
            if (acopTablelabels != null)
              applyTableSpecificLinkAddresses(LinkAddress, acopTablelabels);
          }
          else lock (tblLst)
          { // it's grouped! 
            AcopControl.AcopLink[] lst = AcopControl.AcopLink.GetAcopLinkTable(this);
            if (numTableRows < lst.Length)
            {
              numTableRows = Math.Min(lst.Length, maxTableRows);
              Table.RowCount = numTableRows;
            }
            int idx = 0;
            Boolean delay = false;
            Boolean complete = lst.Length > 0; // optimistic ...
            for (int i = 0; i < lst.Length && i < numTableRows; i++)
            {
              if (!displaySettingsList.ContainsKey(lst[i])) displaySettingsList.Add(lst[i], false);
              if (displaySettingsList[lst[i]])
              {
                idx++;
                continue;
              }
              dsp = acquireAndApplyDisplaySettings(lst[i].GetAddress(), lst[i].GetProperty(), ref acopTablelabels, ismca);
              if (dsp == null || dsp.SettingsStale())
              {
                complete = false;
                tr = addTableEntry("unavailable", "", i);
                idx++;
                continue;
              }
              displaySettingsList[lst[i]] = true;
              if (hasDescArray && i < dscArray.Length)
              {
                dscStr = dscArray[i];
              }
              else
              {
                dscStr = dsp.GetDescription();
                unitsStr = dsp.GetYUnits();
                if (unitsStr.Length > 0) dscStr += " (" + unitsStr + ")";
              }
              if (ismca)
              {
                delay = setMCATableEntries(lst[i], dsp, i);
                assignTableValue(lst[i], ref idx, i, ismca);
                idx = 0;
                continue;
              }
              tr = addTableEntry(dscStr, vals, i);
              assignTableValue(lst[i], ref idx, ismca);
              String ttt = lst[i].GetAddress() + "[" + lst[i].GetProperty() + "]";
              applyToolTipText(tr.dscLab, ttt);
              foreach (TableCell c in tr.cells)
                applyToolTipText(c.valLab, ttt);
            }
            if (!delay)
            {
              displaySettingsAcquired = complete;
              lastGrpLstLength = lst.Length;
            }
          }
          Table.ResumeLayout();
          if (!tableCapacityStale) tableCapacityStale = (Table.RowCount > initialRowCount);
          initialRowCount = Table.RowCount;
        }
      }
    }
    void initComponentInfo()
    {
      setControlsList(getTableControls(), true);
      tableCapacityStale = false;
      //removeHScroll();
    }
    private String descHdrText = null;
    private String valHdrText = null;
    private String[] valHdrList = new String[] { "" };
    private String getDescriptionHeader()
    {
      return descHdrText == null ? "" : descHdrText;
    }
    private void setDescriptionHeader(String txt)
    {
      descHdrText = txt;
      int ncols = hdrRow == null ? 1 : hdrRow.cells.Length;
      String[] vals = new String[ncols];
      for (int i = 0; i < vals.Length; i++) vals[i] = hdrRow.cells[i].valLab.Text;
      //if (descHdrText != null && descHdrText.Length > 0) addTableEntry(descHdrText, vals, 0);
      if (descHdrText != null && descHdrText.Length > 0) hdrRow.dscLab.Text = descHdrText;
    }
    private String getValueHeader()
    {
      return valHdrText == null ? "" : valHdrText;
    }
    private void setValueHeader(String txt)
    {
      valHdrText = txt;
      if (txt != null && txt.Length > 0)
      {
        valHdrList = txt.Split(new String[] { ";","\r\n" },StringSplitOptions.None);
      }
      else
      {
        valHdrList = new String[] { "" };
      }
      for (int i = 0; i < valHdrList.Length && i < numValueColumns; i++)
      {
        setValueHeader(valHdrList[i], i);
      }
    }
    private void setValueHeader(String txt, int valueColumn)
    {
      String dsctxt = tblLst.Count > 0 ? hdrRow.dscLab.Text : "Description";
      if (valueColumn == 0) hdrRow.dscLab.Text = dsctxt;
      if (valueColumn < 0) valueColumn = 0;
      if (valueColumn >= hdrRow.cells.Length) return;
      if (txt != null && txt.Length > 0)
      {
        hdrRow.cells[valueColumn].valLab.Text = txt;
      }
    }
    private void TableResized(object sender, EventArgs e)
    {
      if (Table == null) return;
      Boolean vsv = Table.VerticalScroll.Visible;
      if (vsv != scrollVisible)
      {
        int offset = vsv ? SystemInformation.VerticalScrollBarWidth / 4 : 0;
        setHeaderColumnSize(offset);
        scrollVisible = vsv;
      }
    }

    public AcopTable()
    {
      InitializeComponent();
      InitializeDBTableComponent();
      instance = this;
      SetDefaultUpdater(defaultUpdate);
      addTableEntry("Description", "Readback", "Input");
      Control[] ctrls = getTableControls();
      setControlsList(ctrls, true);
      allowDragAndDrop(ctrls);
      initialRowCount = Table.RowCount;
      SetLateInitializerProlog(initTransportInfo);
      SetLateInitializerEpilog(initComponentInfo);
      StartLateInitialization();
      DoubleBuffered = true;
      cm = new AcopTableContextMenu(this);
      im = new AcopTableInputMenu(this);
      //this.ContextMenu = cm;
      descriptionColumnVisible = true;
      blink = blinkColors;
    }
    private AcopTableUtil.DBTablLayoutPanel Header;
    private AcopTableUtil.DBTablLayoutPanel Table;
    private TableLayoutPanelCellBorderStyle tableBorderStyle = TableLayoutPanelCellBorderStyle.InsetDouble;
    private TableLayoutPanelCellBorderStyle hdrBorderStyle = TableLayoutPanelCellBorderStyle.InsetDouble;
    private void setTableBorderStyle(TableLayoutPanelCellBorderStyle value)
    {
      tableBorderStyle = value;
      Table.CellBorderStyle = value;
      Table.Refresh();
    }
    private void setHeaderBorderStyle(TableLayoutPanelCellBorderStyle value)
    {
      hdrBorderStyle = value;
      Header.CellBorderStyle = value;
      Header.Refresh();
    }
    private int hdrHeight = 30;
    private void InitializeDBTableComponent()
    {
      if (components == null)
        components = new System.ComponentModel.Container();
      this.Table = new AcopTableUtil.DBTablLayoutPanel(components);
      this.Header = new AcopTableUtil.DBTablLayoutPanel(components);
      this.SuspendLayout();
      hdrHeight = (new Label()).Height + 8;
      int hsize = hdrHeight;
      this.Header.Anchor = ((System.Windows.Forms.AnchorStyles)((((System.Windows.Forms.AnchorStyles.Top)
            | System.Windows.Forms.AnchorStyles.Left)
            | System.Windows.Forms.AnchorStyles.Right)));
      this.Header.CellBorderStyle = hdrBorderStyle;
      this.Header.ColumnCount = 2;
      this.Header.ColumnStyles.Add(new System.Windows.Forms.ColumnStyle(System.Windows.Forms.SizeType.Percent, 50F));
      this.Header.ColumnStyles.Add(new System.Windows.Forms.ColumnStyle(System.Windows.Forms.SizeType.Percent, 50F));
      this.Header.Location = new System.Drawing.Point(0, 0);
      this.Header.Name = "Header";
      this.Header.RowCount = 1;
      this.Header.RowStyles.Add(new System.Windows.Forms.RowStyle(SizeType.AutoSize));
      this.Header.RowStyles.Add(new System.Windows.Forms.RowStyle(SizeType.AutoSize));
      this.Header.Size = new System.Drawing.Size(150, hsize);
      this.Header.TabIndex = 0;
      this.Header.Padding = new Padding(1, 1, 1, 1);

      this.Table.Anchor = ((System.Windows.Forms.AnchorStyles)((((System.Windows.Forms.AnchorStyles.Top | System.Windows.Forms.AnchorStyles.Bottom)
            | System.Windows.Forms.AnchorStyles.Left)
            | System.Windows.Forms.AnchorStyles.Right)));
      this.Table.CellBorderStyle = tableBorderStyle;
      this.Table.ColumnCount = 2;
      this.Table.ColumnStyles.Add(new System.Windows.Forms.ColumnStyle(System.Windows.Forms.SizeType.Percent, 50F));
      this.Table.ColumnStyles.Add(new System.Windows.Forms.ColumnStyle(System.Windows.Forms.SizeType.Percent, 50F));
      this.Table.Location = new System.Drawing.Point(0, hsize);
      this.Table.Name = "Table";
      this.Table.RowCount = 2;
      this.Table.RowStyles.Add(new System.Windows.Forms.RowStyle(SizeType.AutoSize));
      this.Table.RowStyles.Add(new System.Windows.Forms.RowStyle(SizeType.AutoSize));
      this.Table.Size = new System.Drawing.Size(150, 150 - hsize);
      this.Table.TabIndex = 1;
      //int vertScrollWidth = SystemInformation.VerticalScrollBarWidth;
      this.Table.Padding = new Padding(1, 1, 1, 1);
      // 
      // AcopTable
      // 
      this.AutoScaleDimensions = new System.Drawing.SizeF(6F, 13F);
      this.AutoScaleMode = System.Windows.Forms.AutoScaleMode.Font;
      this.Controls.Add(this.Header);
      this.Controls.Add(this.Table);
      this.Name = "AcopTable";
      this.ResumeLayout(false);
    }
    private Boolean displayBits = false;
    private UInt32 displayMask = 0xffffffff;

    public enum DisplayBitsStyleSetting
    {
      on_off = 0,
      true_false = 1,
      yes_no = 2
    }
    private DisplayBitsStyleSetting displayBitsSetting = DisplayBitsStyleSetting.on_off;
    private String[] onstring = new String[] { "on", "true", "yes" };
    private String[] offstring = new String[] { "off", "false", "no" };
    private String[] getBitValues(String txt)
    {
      int v = 0;
      if (txt != null && txt.Length > 0)
      {
        Int32.TryParse(txt, out v);
      }
      String[] sv = new String[32];
      int n = 0, db;
      db = (int)displayBitsSetting;
      for (int i = 0; i < 32; i++)
      {
        if ((displayMask & (1 << i)) == 0) continue;
        sv[n++] = (v & (1 << i)) != 0 ? onstring[db] : offstring[db];
      }
      for (; n < 32; n++) sv[n] = "--";
      return sv;
    }
    private int numTableEntries = 0;
    /// <summary>
    /// Assigns the values to the table beginning at the row given
    /// </summary>
    /// <param name="values">the array or single value to be assigned to the row entry</param>
    /// <param name="row">the targeted table row</param>
    public void AssignTableEntries(object values, int row)
    {
      AssignTableEntries(values, row, 0);
    }
    /// <summary>
    /// Assigns the values to the table beginning at the row given
    /// </summary>
    /// <param name="values">the array or single value to be assigned to the row entry</param>
    /// <param name="row">the targeted table row</param>
    /// <param name="valueColumn">the targeted table value column</param>
    public void AssignTableEntries(object values, int row, int valueColumn)
    {
      String[] sval;
      object[] vals;
      int i;
      if (valueColumn < 0) valueColumn = 0;
      if (!(values is Array))
      {
        sval = new String[] { values.ToString() };
        vals = new object[] { values };
      }
      else
      {
        Array avals = (Array)values;
        vals = new object[avals.Length];
        for (i = 0; i < vals.Length; i++) vals[i] = avals.GetValue(i);
        sval = Array.ConvertAll<object, string>(vals, item => item.ToString());
      }
      if (displayBits && sval.Length == 1)
      {
        sval = getBitValues(sval[0]);
      }
      for (i = 0; i < sval.Length && row < tblLst.Count; i++)
      {
        if (!checkThresholds(vals, valueColumn, row, false)) checkStringMatch(sval, valueColumn, row, false);
        TableRow tr = tblLst.ElementAt(row++);
        if (valueColumn >= tr.cells.Length) continue;
        tr.cells[valueColumn].valLab.Text = sval[i];
      }
    }
    private int assignTableValue(AcopControl.AcopLink lnk, ref int idx, Boolean ismca)
    {
      return assignTableValue(lnk, ref idx, 0, ismca);
    }
    private int assignTableValue(AcopControl.AcopLink lnk, ref int idx, int valueColumn, Boolean ismca)
    {
      if (lnk == null || lnk.GetLinkId() < 0) return 0;
      if (valueColumn < 0) valueColumn = 0;
      String[] txt = lnk.GetOutputDataAsString();
      if (txt != null)
      {
        if (displayBits && !getLinkGrouped() && txt.Length == 1)
        {
          txt = getBitValues(txt[0]);
        }
        int i;
        for (i = 0; i < txt.Length && idx < tblLst.Count; i++)
        {
          if (!checkThresholds(lnk, valueColumn, idx, ismca)) checkStringMatch(txt, valueColumn, idx, ismca);
          TableRow tr = tblLst.ElementAt(idx++);
          if (valueColumn >= tr.cells.Length) continue;
          tr.cells[valueColumn].valLab.Text = txt[i];
          if (tr.linkAddr == null)
          {
            tr.linkAddr = lnk.GetAddress();
            tr.cells[valueColumn].linkAddr = tr.linkAddr;
          }
          if (tr.linkProperty == null) tr.cells[valueColumn].linkProperty = lnk.GetProperty();
          if (tr.valueType == null)
          {
            tr.cells[valueColumn].valueType = tr.valueType = getValueType(lnk);
          }
          if (tr.lnk == null) tr.lnk = lnk;
          if (tr.cells[valueColumn].lnk == null) tr.cells[valueColumn].lnk = lnk;
          if (!tr.cells[valueColumn].inptValuesSet)
          {
            tr.cells[valueColumn].inptTxt.Text = txt[i];
            tr.cells[valueColumn].inptValuesSet = true;
          }
        }
        return i;
      }
      return 0;
    }
    public void applyErrorStatus(String sts, ref int idx)
    {
      applyErrorStatus(sts, ref idx, 0);
    }
    public void applyErrorStatus(String sts, ref int idx, int valueColumn)
    {
      TableRow tr = tblLst.ElementAt(idx++);
      if (valueColumn < 0) valueColumn = 0;
      if (valueColumn >= tr.cells.Length) return;
      tr.cells[valueColumn].valLab.ForeColor = Color.White;
      tr.cells[valueColumn].valLab.BackColor = Color.DarkRed;
      tr.cells[valueColumn].valLab.Text = sts;
      tr.needResetColors = true;
      needResetColors = true;
    }
    private void applyTableSpecificLinkAddresses(String linkAddr, String[] names)
    {
      applyTableSpecificLinkAddresses(linkAddr, names, false);
    }
    private void applyTableSpecificLinkAddresses(String linkAddr, String[] names, Boolean isStruct)
    {
      if (names == null || names.Length == 0) return;
      if (linkAddr == null) return;
      String[] parts = linkAddr.Split(new char[] { '/' }, StringSplitOptions.RemoveEmptyEntries);
      if (parts.Length < 2) return;
      String baseAddr = "/"+parts[0]+"/"+parts[1]+"/";
      TableRow tr;
      String ttt;
      String[] dscArray = getDescArray();
      Boolean hasDescArray = (dscArray != null && dscArray.Length > 0);
      if (isStruct) baseAddr = linkAddr + ".";
      for (int i = 0; i < tblLst.Count && i < names.Length; i++)
      {
        tr = tblLst.ElementAt(i);
        if (!hasDescArray) tr.dscLab.Text = names[i];
        if (!isStruct)
        {
          tr.linkAddr = baseAddr + names[i];
          ttt = tr.linkAddr + "[" + LinkProperty + "]";
        }
        else
        {
          ttt = LinkProperty + "." + names[i];
        }
        applyToolTipText(tr.dscLab, ttt);
        foreach (TableCell c in tr.cells)
        {
          c.linkAddr = tr.linkAddr;
          if (!isStruct)
            ttt = tr.linkAddr + "[" + c.linkProperty +"]";
          else
            ttt = LinkProperty + "." + names[i];
          applyToolTipText(c.valLab, ttt);
        }
      }
    }
    private void initialUpdate(AcopControl.AcopLink lnk,Boolean applyNames,int len)
    {
      if (!displaySettingsAcquired) initTransportInfo();
      String[] dscArray = getDescArray();
      Boolean hasDescArray = (dscArray != null && dscArray.Length > 0);
      if (applyNames && acopTablelabels == null)
      { // last minute discovery of the proper channel names ?
        String[] names = lnk.GetOutputChannelNames();
        if (names != null)
        {
          if (names.Length < len)
          {
            String[] labs = new String[len];
            names.CopyTo(labs, 0);
            for (int i = names.Length; i < len; i++)
              labs[i] = "element " + i.ToString();
            names = labs;
          }
          if (!hasDescArray) applyTableLabels(names);
          if (getIsMCA()) applyTableSpecificLinkAddresses(lnk.GetAddress(), names);
        }
      }
      lnk.SetDisplayId(1);
    }
    private int lastGrpLstLength = 0;
    private Boolean isInsideUpdate = false;
    private Boolean needResetInput = true;
    private void defaultUpdate(AcopControl.AcopLink lnk)
    {
      Boolean isgrp = getLinkGrouped();
      if (!lateInitializationDone && !isgrp) return;
      if (!hasControlsList()) return; // we're too early ...
      if (needsAutoAttach())
      {
        hasAutoAttached = false;
        checkAutoAttach();
      }
      if (getAutoUpdate())
      {
        lock (tblLst)
        {
          isInsideUpdate = true;
          Boolean ismca = !isgrp && getIsMCA() && getDisplayAsMCA();
          Boolean isstruct = getIsStructured();
          if (!ismca && !isgrp) ismca = IsOutputArrayObject();
          if (!ismca) ismca = isstruct; // finesse
          int lnkSts = lnk.GetStatus();
          if (!isgrp)
          { // only single value column possible here !
            if (lnkSts != 0 && ErrorValue.Length == 0)
            {
              applyErrorStatus(lnk.GetStatusString(), lnkSts != lastLinkStatus);
              isInsideUpdate = false;
              return;
            }
            if (lnkSts != lastLinkStatus) applyToolTipText(null);
            if (lnk.GetDisplayId() < 0)
            {
              initialUpdate(lnk, ismca, GetOutputArrayLength());
              if (tableCapacityStale)
              {
                setControlsList(getTableControls(), true);
                tableCapacityStale = false;
                isInsideUpdate = false;
                return;
              }
            }
            int idx = 0; 
            numTableEntries = assignTableValue(lnk, ref idx, ismca);
          }
          else
          { // is grouped!
            int idx = 0, n = 0; 
            AcopControl.AcopLink[] lst = AcopControl.AcopLink.GetAcopLinkTable(this);
            if (lastGrpLstLength != lst.Length || !displaySettingsAcquired)
            { // late or otherwise new group members ?
              initTransportInfo();
              if (tableCapacityStale)
              {
                setControlsList(getTableControls(), true);
                tableCapacityStale = false;
                isInsideUpdate = false;
                needResetInput = true;
                return;
              }
            }
            ismca = getIsMCA() && getDisplayAsMCA();
            for (int i = 0; i < lst.Length; i++)
            {
              if (lst[i].GetDisplayId() < 0)
              {
                lst[i].SetDisplayId(1);
                if (ismca)
                {
                  String[] names = null;
                  if (acquiredLabels != null)
                    names = acquiredLabels;
                  else
                    names = lst[0].GetOutputChannelNames();
                  applyTableSpecificLinkAddresses(lst[i].GetAddress(), names);
                  setMCAHeaderRow(i);
                }
                else if (isstruct)
                {
                  String[] names = lst[0].GetOutputChannelNames();
                  applyTableSpecificLinkAddresses(lst[i].GetAddress(), names, true);
                  setStructHeaderRow(i,lst[i]);
                }
              }
              lnkSts = lst[i].GetStatus();
              if (lnkSts != 0 && ErrorValue.Length == 0)
              {
                applyErrorStatus(lst[i].GetStatusString(), ref idx);
                n++;
              }
              else
              {
                int col = ismca || isstruct ? i : 0;
                n += assignTableValue(lst[i], ref idx, col, ismca);
              }
              if (ismca || isstruct) idx = 0;
            }
            if (needResetInput)
            {
              selectedColumn = -1;
              resetInput();
            }
            numTableEntries = n;
          }
          if (needsResetSyncColors) resetSyncColors();
        }
        isInsideUpdate = false;
      }
    }
    private void checkStringMatch(String[] txt, int col, int index, Boolean ismca)
    { // index is table index and for MCA data 
      TableRow tr = tblLst.ElementAt(index);
      int arrayIndex = col;
      int row = index;
      if (!ismca)
      {
        arrayIndex = index;
        index = 0;
      }
      String match = getThresholdMatch(arrayIndex);
      if (txt == null) return;
      if (match.Length == 0 && !tr.needResetColors) return;
      if (index < 0 || index >= txt.Length) return;
      if (!TextMatches(match, txt[index]))
      {
        applyHiWarning(row, col);
      }
      else
      {
        resetColors(row, col);
        tr.needResetColors = false;
      }
    }
    double getThresholdLow(int index)
    {
      if (index < 0 || thresholds == null) return thresholdLow;
      if (index < thresholds.Length && thresholds[index].low != -MAX_THRESHOLD)
        return thresholds[index].low;
      return thresholdLow;
    }
    double getThresholdHigh(int index)
    {
      if (index < 0 || thresholds == null) return thresholdHigh;
      if (index < thresholds.Length && thresholds[index].hi != MAX_THRESHOLD)
        return thresholds[index].hi;
      return thresholdHigh;
    }
    String getThresholdMatch(int index)
    {
      if (index < 0 || thresholds == null) return thresholdMatch;
      if (index < thresholds.Length && thresholds[index].match != "")
        return thresholds[index].match;
      return thresholdMatch;
    }
    private Boolean checkThresholds(Object o, int col, int index, Boolean ismca)
    { // index is MCA index and for MCA data 
      if (o is AcopControl.AcopLink)
      {
        AcopControl.AcopLink lnk = (AcopControl.AcopLink)o;
        o = lnk.GetOutputDataObject();
      }
      try
      {
        int arrayIndex = col;
        int row = index;
        if (o is Array)
        {
          if (((Array)o).Length == 1)
          {
            if (!ismca) arrayIndex = index;
            index = 0;
          }
        }
        if (o is Boolean[])
        {
          Boolean[] bva = ((Boolean[])o);
          String[] bs = new String[bva.Length];
          for (int i = 0; i < bs.Length; i++) bs[i] = bva[i].ToString();
          checkStringMatch(bs, col, index, ismca);
          return true;
        }
        double v;
        if (o is double[]) v = ((double[])o)[index];
        else if (o is float[]) v = (double)((float[])o)[index];
        else if (o is Int16[]) v = (double)((Int16[])o)[index];
        else if (o is Int32[]) v = (double)((Int32[])o)[index];
        else if (o is Int64[]) v = (double)((Int64[])o)[index];
        else if (o is Double) v = (Double)o;
        else if (o is Single) v = (double)((Single)o);
        else if (o is Int16) v = (double)((Int16)o);
        else if (o is Int32) v = (double)((Int32)o);
        else if (o is Int64) v = (double)((Int64)o);
        else return false;

        if (v < getThresholdLow(arrayIndex))
        {
          applyLowWarning(row, col);
        }
        else if (v > getThresholdHigh(arrayIndex))
        {
          applyHiWarning(row, col);
        }
        else
        {
          resetColors(row, col);
        }
        return true;
      }
      catch (Exception)
      {
      }
      return false;
    }
    private void setTableFont(Font font)
    {
      TableRow tr;
      for (int row = 0; row < tblLst.Count; row++) 
      {
        tr = tblLst.ElementAt(row);
        foreach (TableCell c in tr.cells)
        {
          c.valLab.Font = font;
          if (c.inptTxt != null) c.inptTxt.Font = font;
        }
      }
      tableFont = font;
      this.Font = font;
    }
    private void setDescriptionFont(Font font)
    {
      TableRow tr;
      for (int row = 0; row < tblLst.Count; row++)
      {
        tr = tblLst.ElementAt(row);
        tr.dscLab.Font = font;
      }
      descFont = font;
      Table.Refresh();
    }

    private void setBlinkColors(TableCell c, Color fill, Color text)
    {
      c.valLab.BackColor = fill;
      c.valLab.ForeColor = text;
    }
    private Boolean isBlinkCycle = false;
    void blinkColors(Boolean blink)
    {
      if (blinkMode == BlinkMode.off) return;
      isBlinkCycle = blink;
      Boolean findBlinkColors = blink;
      TableRow tr;
      for (int row = 0; row < tblLst.Count; row++)
      {
        tr = tblLst.ElementAt(row);
        foreach (TableCell c in tr.cells)
        {
          findBlinkColors = blink;
          if (blinkMode == BlinkMode.onWarning && !c.isWarningCondition)
          {
            findBlinkColors = false;
          }
          if (findBlinkColors)
          {
            Color fc =
              blinkerIntensity > 0 ?
              ControlPaint.Light(c.currentBackColor, blinkerIntensity) :
              ControlPaint.Dark(c.currentBackColor, -blinkerIntensity);
            Color tc =
              blinkerIntensity > 0 ?
              ControlPaint.Dark(c.currentForeColor, blinkerIntensity) :
              ControlPaint.Light(c.currentForeColor, -blinkerIntensity);
            setBlinkColors(c, fc, tc);
          }
          else
          {
            setBlinkColors(c, c.currentBackColor, c.currentForeColor);
          }
        }
      }
    }
    private void resetColors()
    {
      foreach (TableRow tr in tblLst)
      {
        resetColors(tr.row);
        tr.needResetColors = false;
        needResetColors = false;
      }
    }
    private void resetColors(int index)
    {
      if (index < 0 || index >= tblLst.Count) return;
      TableRow tr = tblLst.ElementAt(index);
      if (tr.isSelected) return;
      foreach (TableCell c in tr.cells)
      {
        c.valLab.BackColor = getValueBackColor(tr, c.col);
        c.valLab.ForeColor = getValueForeColor(tr, c.col);
        c.currentBackColor = c.valLab.BackColor;
        c.currentForeColor = c.valLab.ForeColor;
        c.isWarningCondition = false;
      }
      if (IsDesignMode())
      {
        Table.Refresh();
      }
    }
    private void resetColors(int index, int valueColumn)
    {
      if (index < 0 || index >= tblLst.Count) return;
      TableRow tr = tblLst.ElementAt(index);
      if (tr.isSelected) return;
      if (valueColumn >= 0 && valueColumn < tr.cells.Length)
      {
        tr.cells[valueColumn].valLab.BackColor = getValueBackColor(tr,valueColumn);
        tr.cells[valueColumn].valLab.ForeColor = getValueForeColor(tr,valueColumn);
        tr.cells[valueColumn].currentBackColor = tr.cells[valueColumn].valLab.BackColor;
        tr.cells[valueColumn].currentForeColor = tr.cells[valueColumn].valLab.ForeColor;
        tr.cells[valueColumn].isWarningCondition = false;
      }
      tr.needResetColors = false;
    }
    private void setLabelForeColor(Color color)
    {
      TableRow tr = null;
      for (int row = 0; row < tblLst.Count; row++) 
      {
        tr = tblLst.ElementAt(row);
        foreach (TableCell c in tr.cells)
        {
          c.valLab.ForeColor = color;
          c.currentForeColor = color;
        }
      }
      labelForeColor = color;
    }
    private void setLabelBackColor(Color color)
    {
      TableRow tr = null;
      for (int row = 0; row < tblLst.Count; row++)
      {
        tr = tblLst.ElementAt(row);
        foreach (TableCell c in tr.cells)
        {
          c.valLab.BackColor = color;
          c.currentBackColor = color;
        }
      }
      labelBackColor = color;
    }
    private Boolean cellNeedsWarning(int row, int col)
    {
      if (row < 0 || row > tblLst.Count) return false;
      TableRow tr = tblLst.ElementAt(row);
      if (col >= 0 && col < tr.cells.Length)
      {
        return !tr.cells[col].isWarningCondition;
      }
      return false;
    }
    private void applyLowWarning(int index, int valueColumn)
    {
      if (isBlinkCycle)
      {
        if (!cellNeedsWarning(index,valueColumn)) return;
      }
      if (index < 0 || index >= tblLst.Count) return;
      TableRow tr = tblLst.ElementAt(index);
      if (tr.isSelected) return;
      Color bc = labelLowWarnBackColor != Color.Empty ? labelLowWarnBackColor : getValueBackColor(tr,valueColumn);
      Color fc = labelLowWarnForeColor != Color.Empty ? labelLowWarnForeColor : getValueForeColor(tr,valueColumn);
      if (valueColumn >= 0 && valueColumn < tr.cells.Length)
      {
        tr.cells[valueColumn].valLab.BackColor = bc;
        tr.cells[valueColumn].valLab.ForeColor = fc;
        tr.cells[valueColumn].currentBackColor = bc;
        tr.cells[valueColumn].currentForeColor = fc;
        tr.cells[valueColumn].isWarningCondition = true;
      }
    }
    private void applyHiWarning(int index, int valueColumn)
    {
//      if (isBlinkCycle) return;
      if (isBlinkCycle)
      {
        if (!cellNeedsWarning(index, valueColumn)) return;
      }
      if (index < 0 || index >= tblLst.Count) return;
      TableRow tr = tblLst.ElementAt(index);
      if (tr.isSelected) return;
      Color bc = labelHiWarnBackColor != Color.Empty ? labelHiWarnBackColor : getValueBackColor(tr,valueColumn);
      Color fc = labelHiWarnForeColor != Color.Empty ? labelHiWarnForeColor : getValueForeColor(tr,valueColumn);
      if (valueColumn >= 0 && valueColumn < tr.cells.Length)
      {
        tr.cells[valueColumn].valLab.BackColor = bc;
        tr.cells[valueColumn].valLab.ForeColor = fc;
        tr.cells[valueColumn].currentBackColor = bc;
        tr.cells[valueColumn].currentForeColor = fc;
        tr.cells[valueColumn].isWarningCondition = true;
      }
    }
    private void setThresholdHigh(double value)
    {
      if (value != MAX_THRESHOLD) thresholdHigh = value;
    }
    private void setThresholdLow(double value)
    {
      if (value != -MAX_THRESHOLD) thresholdLow = value;
    }
    private void setLabelLowWarnForeColor(Color value)
    {
      labelLowWarnForeColor = value;
    }

    private void setLabelHiWarnForeColor(Color value)
    {
      labelHiWarnForeColor = value;
    }

    private void setLabelLowWarnBackColor(Color value)
    {
      labelLowWarnBackColor = value;
    }

    private void setLabelHiWarnBackColor(Color value)
    {
      labelHiWarnBackColor = value;
    }

    private ContentAlignment valueLabelAlignment = ContentAlignment.MiddleCenter;
    private ContentAlignment descLabelAlignment = ContentAlignment.MiddleCenter;

    private void setValueLabelAligment(ContentAlignment value)
    {
      if (valueLabelAlignment == value) return;
      TableRow tr = null;
      for (int row = 0; row < tblLst.Count; row++)
      {
        tr = tblLst.ElementAt(row);
        foreach (TableCell c in tr.cells)
        {
          c.valLab.TextAlign = value;
        }
      }
      foreach (TableCell c in hdrRow.cells)
      {
        c.valLab.TextAlign = value;
        c.inptLab.TextAlign = value;
      }
      valueLabelAlignment = value;
    }
    private void setDescLabelAligment(ContentAlignment value)
    {
      if (descLabelAlignment == value) return;
      TableRow tr = null;
      for (int row = 0; row < tblLst.Count; row++)
      {
        tr = tblLst.ElementAt(row);
        tr.dscLab.TextAlign = value;
      }
      hdrRow.dscLab.TextAlign = value;
      descLabelAlignment = value;
    }

    private String[] descriptionList = new String[] { };
    private String[] getDescArray()
    {
      return descriptionList;
    }
    private void setDescArray(String[] items)
    {
      descriptionList = items;
      if (items == null || tblLst.Count == 0) return;
      TableRow tr;
      for (int row = 0; row < tblLst.Count && row < items.Length; row++)
      {
        tr = tblLst.ElementAt(row);
        tr.dscLab.Text = items[row];
      }
    }

    [Category("Acop.Transport")]
    [ToolboxItem(false)]
    [Browsable(false), EditorBrowsable(EditorBrowsableState.Never)]
    public new String Description { get; set; }

    /// <summary>
    /// Supply a specific array of description strings to use in the table Description column
    /// </summary>
    [Category("Acop.Display")]
    [Description("Provides a list of decription texts to be used in the description column (if visible)")]
    [DesignerSerializationVisibility(DesignerSerializationVisibility.Content)]
    [Editor("System.Windows.Forms.Design.StringArrayEditor, System.Design, Version = 2.0.0.0, Culture = neutral, PublicKeyToken = b03f5f7f11d50a3a", typeof(System.Drawing.Design.UITypeEditor))]
    public String[] DescriptionList { get { return getDescArray(); } set { setDescArray(value); } }

    /// <summary>
    /// Mask giving the relevant bits to use if DisplayAsBits is true
    /// </summary>
    [Category("Acop.Display")]
    [Description("Mask giving the relevant bits to use if DisplayAsBits is true")]
    public UInt32 DisplayBitMask { get { return displayMask; } set { displayMask = value; } }
    /// <summary>
    /// true: display bit breakdown (applies to a single link returning a number)
    /// </summary>
    [Category("Acop.Display")]
    [Description("true: display bit breakdown (applies to a single link returning a number)")]
    public Boolean DisplayAsBits { get { return displayBits; } set { displayBits = value; } }
    /// <summary>
    /// determines how bit values are displayed in the table (if DisplayAsBits is true)
    /// </summary>
    [Category("Acop.Display")]
    [Description("determines how bit values are displayed in the table (if DisplayAsBits is true)")]
    public DisplayBitsStyleSetting DisplayBitsStyle { get { return displayBitsSetting; } set { displayBitsSetting = value; } }

    /// <summary>
    /// Desired description column width in percent (if DescriptionHeader = true)
    /// </summary>
    [Category("Acop.Display")]
    [Description("Desired description column width in percent (if DescriptionHeader = true)")]
    public int DescriptionColumnSize { get { return descColSize; } set { setColumnDescriptionSize(value); } }

    /// <summary>
    /// Table reaback value header text
    /// </summary>
    [Category("Acop.Display")]
    [Description("Table reaback value header text")]
    [Editor(typeof(System.ComponentModel.Design.MultilineStringEditor), typeof(System.Drawing.Design.UITypeEditor))]
    public String ValueHeader { get { return getValueHeader(); } set { setValueHeader(value); } }

    /// <summary>
    /// Table description header text
    /// </summary>
    [Category("Acop.Display")]
    [Description("Table description header text")]
    public String DescriptionHeader {  get { return getDescriptionHeader(); } set { setDescriptionHeader(value); } }
    /// <summary>
    /// text aligment for table values
    /// </summary>
    [Category("Acop.Display")]
    [Description("Text aligment for table values")]
    public ContentAlignment TableValueAlignment { get { return valueLabelAlignment; } set { setValueLabelAligment(value); } }
    /// <summary>
    /// Text aligment for table description cells
    /// </summary>
    [Category("Acop.Display")]
    [Description("Text aligment for table description cells")]
    public ContentAlignment TableDescriptionAlignment { get { return descLabelAlignment; } set { setDescLabelAligment(value); } }
    /// <summary>
    /// initial number of table rows
    /// </summary>
    [Category("Acop.Display")]
    [Description("initial number of table rows")]
    public int NumberRows { get { return numTableRows; } set { setNumTableRows(value); } }

    /// <summary>
    /// initial number of table value columns
    /// </summary>
    [Category("Acop.Display")]
    [Description("initial number of table value columns")]
    public int NumberColumns { get { return numValueColumns; } set { setNumberValueColumns(value); } }

    [Category("Acop.Display")]
    [Description("initial number of table input columns (less than or equal to number of value columns)")]
    public int NumberInputColumns { get { return numInputColumns; } set { setNumberInputColumns(value); } }

    /// <summary>
    /// maximum number of table rows
    /// </summary>
    [Category("Acop.Display")]
    [Description("maximum number of table rows")]
    public int MaximumRows { get { return maxTableRows; } set { setMaxTableRows(value); } }
    /// <summary>
    /// true: show the description column
    /// </summary>
    [Category("Acop.Display")]
    [Description("true: show the description column")]
    public Boolean ShowDescriptionColumn { get { return showDescriptionColumn; } set { setShowDescriptionColumn(value); } }
    /// <summary>
    /// true: show the value input column
    /// </summary>
    [Category("Acop.Display")]
    [Description("true: show the value input column")]
    public Boolean ShowInputColumn { get { return showInputColumn; } set { setShowInputColumn(value); } }

    /// <summary>
    /// true: show the table header
    /// </summary>
    [Category("Acop.Display")]
    [Description("true: show the table")]
    public Boolean ShowHeader { get { return showHeader; } set { setShowHeader(value); } }
    /// <summary>
    /// Font to apply to table cells
    /// </summary>
    [Category("Acop.Display")]
    [Description("Font to apply to table cells")]
    public Font DisplayFont { get { return tableFont; } set { setTableFont(value); } }

    /// <summary>
    /// Font to apply to the description column
    /// </summary>
    [Category("Acop.Display")]
    [Description("Font to apply to the description column")]
    public Font DescriptionFont { get { return descFont; } set { setDescriptionFont(value); } }
    /// <summary>
    /// Fore Color to apply to the value cell
    /// </summary>
    [Category("Acop.Display")]
    [Description("ForeColor to apply to the value cell")]
    public Color LabelForeColor { get { return labelForeColor; } set { setLabelForeColor(value); } }
    /// <summary>
    /// Back Color to display when a cell is selected
    /// </summary>
    [Category("Acop.Display")]
    [Description("BackColor to display when a cell is selected")]
    public Color SelectionBackColor { get { return selectionBackColor; } set { selectionBackColor = value; } }
    /// <summary>
    /// ForeColor to display when a cell is selected
    /// </summary>
    [Category("Acop.Display")]
    [Description("ForeColor to display when a cell is selected")]
    public Color SelectionForeColor { get { return selectionForeColor; } set { selectionForeColor = value; } }
    /// <summary>
    /// Fore Color to display in the input column when it is shown
    /// </summary>
    [Category("Acop.Display")]
    [Description("ForeColor to display in the input column when it is shown")]
    public Color InputForeColor { get { return inputSyncForeColor; } set { setInputSyncForeColor(value); } }
    /// <summary>
    /// Fore Color to display in the input column when it is shown and the data contents are out of sync with the readback values
    /// </summary>
    [Category("Acop.Display")]
    [Description("ForeColor to display in the input column when it is shown and the data contents are out of sync with the readback values")]
    public Color InputForeColorOutOfSync { get { return inputOutofSyncColor; } set { inputOutofSyncColor = value; } }
    /// <summary>
    /// Back Color to display in the input column when it is shown
    /// </summary>
    [Category("Acop.Display")]
    [Description("BackColor to display in the input column when it is shown")]
    public Color InputBackColor { get { return inputBackColor; } set { setInputBackColor(value); } }

    /// <summary>
    /// true: allow selection of table value cells and fire Apply event in context menu
    /// </summary>
    [Category("Acop.Display")]
    [Description("true: allow selection of table value cells and fire Apply event in context menu")]
    public Boolean AllowSelections { get { return allowSelections; } set { setAllowSelections(value); } }
    /// <summary>
    /// true: allow scroll bars if the number of table rows exceeds the table layout size
    /// </summary>
    [Category("Acop.Display")]
    [Description("true: allow scroll bars if the number of table rows exceeds the table layout size")]
    public Boolean TableAutoScroll { get { return tableAutoScroll; } set { setTableAutoScroll(value); } }

    /// <summary>
    /// ForeColor to use if number value is less than ThresholdLowWarn or if string value equals ThresholdMatch
    /// </summary>
    [Category("Acop.Display")]
    [Description("ForeColor to use if number value is less than ThresholdLowWarn or if string value equals ThresholdMatch")]
    public Color LabelLowWarnForeColor { get { return labelLowWarnForeColor; } set { setLabelLowWarnForeColor(value); } }
    /// <summary>
    /// ForeColor to use if number value is greater than ThresholdHiWarn or if string value does not equal ThresholdMatch
    /// </summary>
    [Category("Acop.Display")]
    [Description("ForeColor to use if number value is greater than ThresholdHiWarn or if string value does not equal ThresholdMatch")]
    public Color LabelHiWarnForeColor { get { return labelHiWarnForeColor; } set { setLabelHiWarnForeColor(value); } }
    /// <summary>
    /// Back Color to apply to the value cell
    /// </summary>
    [Category("Acop.Display")]
    [Description("Back Color to apply to the value cell")]
    public Color LabelBackColor { get { return labelBackColor; } set { setLabelBackColor(value); } }
    /// <summary>
    /// Back Color to use if number value is less than ThresholdLowWarn or if string value equals ThresholdMatch
    /// </summary>
    [Category("Acop.Display")]
    [Description("BackColor to use if number value is less than ThresholdLowWarn or if string value equals ThresholdMatch")]
    public Color LabelLowWarnBackColor { get { return labelLowWarnBackColor; } set { setLabelLowWarnBackColor(value); } }
    /// <summary>
    /// Back Color to use if number value is greater than ThresholdHiWarn or if string value does not equal ThresholdMatch
    /// </summary>
    [Category("Acop.Display")]
    [Description("Back Color to use if number value is greater than ThresholdHiWarn or if string value does not equal ThresholdMatch")]
    public Color LabelHiWarnBackColor { get { return labelHiWarnBackColor; } set { setLabelHiWarnBackColor(value); } }
    /// <summary>
    /// if non-empty and the link data is non-numerical, then the string provided is used for comparisons with the monitored data
    /// </summary>
    [Category("Acop.Display")]
    [Description("if non-empty and the link data is non-numerical, then the string provided is used for comparisons with the monitored data")]
    public String ThresholdMatch { get { return thresholdMatch; } set { thresholdMatch = value; } }
    /// <summary>
    /// if the link data is numerical, then the threshold provided is used for comparisons with the monitored data
    /// </summary>
    [Category("Acop.Display")]
    [Description("if the link data is numerical, then the threshold provided is used for comparisons with the monitored data")]
    public double ThresholdHighWarn { get { return thresholdHigh; } set { setThresholdHigh(value); } }
    /// <summary>
    /// if the link data is numerical, then the threshold provided is used for comparisons with the monitored data
    /// </summary>
    [Category("Acop.Display")]
    [Description("if the link data is numerical, then the threshold provided is used for comparisons with the monitored data")]
    public double ThresholdLowWarn { get { return thresholdLow; } set { setThresholdLow(value); } }
    /// <summary>
    /// Supplies a list of threshold criteria to be used for a set of grouped links
    /// </summary>
    //[Category("Acop.Display")]
    //[Description("Supplies a list of threshold criteria to be used for a set of grouped links")]
    //public String ThresholdList { get { return getThresholdList(); } set { setThresholdList(value); } }

    [Category("Acop.Display")]
    [Description("If not empty, attach the target control property to the slider")]
    [Editor(typeof(ThresholdListEditor), typeof(System.Drawing.Design.UITypeEditor))]
    [DesignerSerializationVisibility(DesignerSerializationVisibility.Content)]
    [TypeConverter(typeof(ThresholdListEditorConverter))]
    public ThresholdListEditorValue ThresholdList { get { return getThresholdListEditorValue(); } set { setThresholdListEditorValue(value); } }

    /// <summary>
    /// Border style of this acop table
    /// </summary>
    [Category("Acop.Display")]
    [Description("Border style of this acop table")]
    public TableLayoutPanelCellBorderStyle TableBorderStyle {  get { return tableBorderStyle; } set { setTableBorderStyle(value); } }
    /// <summary>
    /// Border style of this acop table header
    /// </summary>
    [Category("Acop.Display")]
    [Description("Border style of this acop table header")]
    public TableLayoutPanelCellBorderStyle HeaderBorderStyle { get { return hdrBorderStyle; } set { setHeaderBorderStyle(value); } }
    /// <summary>
    /// Border style of the value label cells within the table
    /// </summary>
    [Category("Acop.Display")]
    [Description("Border style of the value label cells within the table")]
    public BorderStyle LabelBorderStyle { get { return labelBorderStyle; } set { setLabelBorderStyle(value); } }
    /// <summary>
    /// Specifies a list of foreground colors to use for value entries
    /// </summary>
    [Category("Acop.Display")]
    [Description("Value Label Forecolor List Editor")]
    [Editor(typeof(ColorListEditor), typeof(System.Drawing.Design.UITypeEditor))]
    [DesignerSerializationVisibility(DesignerSerializationVisibility.Content)]
    [TypeConverter(typeof(ColorListEditorConverter))]
    public ColorListEditorValue LabelForeColorList { get { return getColorListEditorValue(); } set { setColorListEditorValue(value); resetColors(); } }
    /// <summary>
    /// apply the BackColor List to column entries if NumberValueColumns > 1
    /// </summary>
    [Category("Acop.Display")]
    [Description("apply the BackColor List to column entries if NumberValueColumns > 1")]
    public Boolean ApplyBackColorListToColumns {  get { return applyBackColorListToColumns; } set { setApplyBackColorListToColumns(value); } }
    /// <summary>
    /// apply the ForeColor List to column entries if NumberValueColumns > 1
    /// </summary>
    [Category("Acop.Display")]
    [Description("apply the ForeColor List to column entries if NumberValueColumns > 1")]
    public Boolean ApplyForeColorListToColumns { get { return applyForeColorListToColumns; } set { setApplyForeColorListToColumns(value); } }
    /// <summary>
    /// Specifies a list of foreground colors to use for value entries
    /// </summary>
    [Category("Acop.Display")]
    [Description("Value Label Backcolor List Editor")]
    [Editor(typeof(ColorListEditor), typeof(System.Drawing.Design.UITypeEditor))]
    [DesignerSerializationVisibility(DesignerSerializationVisibility.Content)]
    [TypeConverter(typeof(ColorListEditorConverter))]
    public ColorListEditorValue LabelBackColorList { get { return getColorListEditorValue(BACKCOLOR_LIST); } set { setColorListEditorValue(BACKCOLOR_LIST,value); resetColors(); } }

    /// <summary>
    /// set the blink mode for the table cells
    /// </summary>
    [Category("Acop.Display")]
    [Description("set the blink mode for the table cells")]
    public BlinkMode Blink { get { return blinkMode; } set { setBlinkMode(value); } }
    /// <summary>
    /// set the blink intensity for the table cells (up to 100% lighter or -100% darker)
    /// </summary>
    [Category("Acop.Display")]
    [Description("set the blink intensity for the table cells (up to 100% lighter or -100% darker)")]
    public Int32 BlinkIntensity { get { return blinkerIntensity; } set { setBlinkerIntensity(value); } }
    /// <summary>
    /// set the blink interval (msec) for the table cells
    /// </summary>
    [Category("Acop.Display")]
    [Description("set the blink interval (msec) for the table cells")]
    public Int32 BlinkInterval { get { return blinkerInterval; } set { setBlinkerInterval(value); } }
    /// <summary>
    /// set the blink slot for the table cells (0 up to the number of slots)
    /// </summary>
    [Category("Acop.Display")]
    [Description("set the blink slot for the table cells (0 up to the number of slots)")]
    public Int32 BlinkSlot { get { return blinkerSlot; } set { setBlinkerSlot(value); } }
    /// <summary>
    /// set the number of blink slots for the table cells (at least 2)
    /// </summary>
    [Category("Acop.Display")]
    [Description("set the number of blink slots for the table cells (at least 2)")]
    public Int32 NumberBlinkSlots { get { return numberBlinkSlots; } set { setNumberBlinkerSlots(value); } }

  }
  [ToolboxItem(false)]
  public class AcopTableContextMenu : ContextMenu
  {
#   pragma warning disable 1591
    private AcopTable table;
    private MenuItem apply;
    private MenuItem clear;

    public AcopTableContextMenu(AcopTable table)
    {
      this.table = table;
      apply = new MenuItem("Apply Selection");
      apply.Click += new EventHandler(this.applySelections);
      clear = new MenuItem("Clear Selection");
      clear.Click += new EventHandler(this.clearSelections);

      this.MenuItems.Add(apply);
      this.MenuItems.Add(clear);
    }
    private void applySelections(object sender, EventArgs e)
    {
      table.handleApplySelection(sender);
    }
    private void clearSelections(object sender, EventArgs e)
    {
      table.ClearSelection();
    }
  }
  [ToolboxItem(false)]
  public class AcopTableInputMenu : ContextMenu
  {
#   pragma warning disable 1591
    private AcopTable table;
    private MenuItem apply;
    private MenuItem reset;

    public AcopTableInputMenu(AcopTable table)
    {
      this.table = table;
      apply = new MenuItem("Apply Input");
      apply.Click += new EventHandler(this.applySelections);
      reset = new MenuItem("Reset Input");
      reset.Click += new EventHandler(this.resetSelections);

      this.MenuItems.Add(apply);
      this.MenuItems.Add(reset);
    }
    private void applySelections(object sender, EventArgs e)
    {
      table.handleApplyInput(sender);
    }
    private void resetSelections(object sender, EventArgs e)
    {
      table.resetInput();
    }
  }
}
namespace AcopTableUtil
{
  [ToolboxItem(false)]
  [Browsable(false), EditorBrowsable(EditorBrowsableState.Never)]
  public class DBTablLayoutPanel : TableLayoutPanel
  {
#   pragma warning disable 1591
    public DBTablLayoutPanel(IContainer container)
    {
      container.Add(this);
      SetStyle(ControlStyles.AllPaintingInWmPaint |
        ControlStyles.OptimizedDoubleBuffer |
        ControlStyles.UserPaint, true);
    }
  }
}
