﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Drawing.Design;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Text.RegularExpressions;
using System.Windows.Forms;
using System.Windows.Forms.Design;
using System.Reflection;
using System.Threading;

namespace Acop
{
  [ToolboxItem(false)]
  public class AcopUserComponent : UserControl
  {
#   pragma warning disable 1591
    public static string acopVersion = "2.2.14";

    protected System.ComponentModel.IContainer components = null;
    private System.Windows.Forms.Timer lateInitializer;
    static AcopUserComponent instance = null;
    Component component = null;
    private ToolTip thisTT = null;
    private Control[] thisCtrlsList = null;
    private IDictionary<string, int> flagsDict = null;
    private AcopTransport.DisplaySettings linkDspSettings = null;
    private Type valueType = null;
    static DateTime utcThen = new DateTime(1970, 1, 1);
    private object notifyMutex = new object();

    private object linkDataOut = null;
    private object linkDataIn = null;
    private int linkFlags = 0;
    private String linkAddr = null;
    private String linkGroupList = null;

    private String linkProt = AcopControl.AcopLink.GetDefaultProtocol();
    private String linkFlagsString = "NONE";

    private Boolean isDesignMode = false;
    private Boolean initialPassDone = false;
    protected Boolean displaySettingsAcquired = false;
    protected Dictionary<AcopControl.AcopLink, Boolean> displaySettingsList = new Dictionary<AcopControl.AcopLink, bool>();

    private Boolean ctxStale = true;
    private Boolean srvStale = true;
    private Boolean devStale = true;
    private Boolean prpStale = false;
    private Boolean subStale = true;

    private Boolean hasPropertyQueryPrecedence = false;
    private Boolean hasDeviceQueryPrecedence = false;
    protected Boolean hasAutoAttached = false;

    private Boolean isGrouped = false;
    private Boolean autoUpdate = false;
    private Boolean autoAttach = false;
    private Boolean isMCA = false;
    private Boolean isStructured = false;
    private Boolean displayAsMCA = false;

    private Boolean valueIsInteger = false;
    private Boolean valueIsNumber = false;
    private Boolean valueAvailableAsNumber = false;
    private Boolean insideNotify = false;

    private String linkCtx = null;
    private String lastLinkCtx = null;
    private String linkSubsystem = null;
    private static String lastLinkSub = null;
    private String linkSrv = null;
    private String lastLinkSrv = null;
    private String linkDev = null;
    private String lastLinkDev = null;
    private String linkProp = null;
    private String lastLinkPrp = null;
    private String units = null;

    private String[] descriptionLabels = null;
    private String description = null;

    private const int AT_CHANNEL = 0x10;
    private const int AT_STRUCTURED = 0x40;
    public static Boolean isChannelArray(int arrayType)
    {
      return (arrayType & AT_CHANNEL) == AT_CHANNEL;
    }
    public static Boolean isStructure(int arrayType)
    {
      return (arrayType & AT_STRUCTURED) == AT_STRUCTURED;
    }
    private int defaultInterval = 1000;
    private int autoLinkCapacity = 0;
    private int precision = 0;
    protected int numAcopLinksPending = 0;
    protected int numAcopLinksInGroup = 0;

    private double resetValue = 0;
    private double readbackValue = 0;
    private String errorValue = "";
    private String linkInputValue = "";
    public const String UNAVAILABLE = "<unavailable>";

    public const double MAX_THRESHOLD = 1e100;

    public event AcopControl.AcopLink.acopTransportUpdate update = null;

    public delegate void defaultUpdaterType(AcopControl.AcopLink lnk);
    public delegate int defaultApplierType(AcopControl.AcopLink lnk);
    private delegate void notifyLinkUpdateDelegate(object sender, AcopControl.AcopLink lnk);
    //public delegate void acopUpdate(object sender, AcopControl.AcopLink lnk);
    public delegate void acopUpdate(object sender, Link lnk);
    public delegate void acopApplySetting(object sender, Link lnk);
    protected delegate void lateInitializerPrologDelegate();
    protected delegate void lateInitializerEpilogDelegate();

    public defaultUpdaterType updater = null;
    public defaultApplierType applier = null;
    // a lateInitializerProlog must attempt to acquire display settings !
    private lateInitializerPrologDelegate lateInitializerProlog = null;
    private lateInitializerEpilogDelegate lateInitializerEpilog = null;

    public Label defaultDescriptionLabel = null;
    public Label defaultUnitsLabel = null;

    /// <summary>
    /// An invoker helper for cases where InvokeRequired is needed
    /// </summary>
    /// <param name="updater">is the delegate to invoke</param>
    /// <param name="sender">is the sender object</param>
    /// <param name="lnk">is the Link object</param>
    public void LinkUpdateInvoker(acopUpdate updater, object sender, Link lnk)
    {
      try
      {
        Invoke(new acopUpdate(updater), sender, lnk);
      }
      catch (Exception) { }
    }

    protected void AddrDragDrop(object sender, DragEventArgs e)
    {
      if (isDesignMode)
      {
        if (instance == null) return;
        String ds = (String)e.Data.GetData(DataFormats.StringFormat);
        String[] parts = ds.Split(new char[] { '/' }, StringSplitOptions.RemoveEmptyEntries);
        int nparts = parts.Length;
        if (nparts > 0) instance.LinkProtocol = parts[0];
        if (nparts > 1)
        {
          instance.LinkContext = parts[1];
          linkDev = null; linkProp = null;
        }
        if (nparts > 2) instance.LinkServer = parts[2];
        if (nparts > 3) instance.LinkDevice = parts[3];
        if (nparts > 4) instance.LinkProperty = parts[4];
      }
    }
    protected void AddrDragEnter(object sender, DragEventArgs e)
    {
      e.Effect = DragDropEffects.All;
    }
    protected void AcopMouseDown(object sende, MouseEventArgs e)
    {
      string s = LinkProtocol;
      if (!LinkAddress.StartsWith("/")) s += "/";
      s += LinkAddress + "/" + LinkProperty;
      DoDragDrop(s, DragDropEffects.All);
    }
    protected void allowDragAndDrop(Control[] ctrls)
    {
      foreach (Control c in ctrls)
      {
        c.AllowDrop = true;
        if (c is AcopUserComponent)
        {
          c.DragDrop += new System.Windows.Forms.DragEventHandler(AddrDragDrop);
          c.DragEnter += new System.Windows.Forms.DragEventHandler(AddrDragEnter);
        }
        c.MouseDown += new System.Windows.Forms.MouseEventHandler(AcopMouseDown);
      }
    }
    public const String PRIMARY_COLOR_LIST = "Primary";
    protected class AcopColorList
    {
      public ColorListEditorValue editorValue = null;
      public String parseableString = "";
      public List<Color> colorList = new List<Color>();
      public AcopColorList(String listString, ColorListEditorValue value)
      {
        editorValue = value;
        parseableString = listString;
      }
      public AcopColorList(ColorListEditorValue value)
      {
        editorValue = value;
        parseableString = value != null ? value.ToString() : "";
      }
      public AcopColorList()
      {

      }
    }
    private Dictionary<String, AcopColorList> acopColorHash = new Dictionary<string, AcopColorList>();
    /// <summary>
    /// Adds a color (by string name) to the current color list
    /// </summary>
    /// <param name="color">the color (as a string) to add the color list</param>
    public void AddColorToList(String color)
    {
      if (currentColorListKey == null) currentColorListKey = PRIMARY_COLOR_LIST;
      AddColorToList(currentColorListKey, color);
    }
    public void AddColorToList(String key, String color)
    {
      if (color == null || color.Length == 0) return;
      Color c = Color.FromName(color);
      if (!acopColorHash.ContainsKey(key))
        acopColorHash.Add(key, new AcopColorList());
      AcopColorList lst;
      if (key == null) key = PRIMARY_COLOR_LIST;
      if (acopColorHash.TryGetValue(key,out lst))
        lst.colorList.Add(c);
    }
    /// <summary>
    /// Adds a color to the current color list
    /// </summary>
    /// <param name="color">the color to add to the color list</param>
    public void AddColorToList(Color color)
    {
      if (currentColorListKey == null) currentColorListKey = PRIMARY_COLOR_LIST;
      AddColorToList(currentColorListKey, color);
    }
    public void AddColorToList(String key, Color color)
    {
      if (color == null) return;
      if (!acopColorHash.ContainsKey(key))
        acopColorHash.Add(key, new AcopColorList());
      AcopColorList lst;
      if (key == null) key = PRIMARY_COLOR_LIST;
      if (acopColorHash.TryGetValue(key, out lst))
        lst.colorList.Add(color);
    }
    public void SetParseableColorList(String colorList)
    {
      if (currentColorListKey == null) currentColorListKey = PRIMARY_COLOR_LIST;
      SetParseableColorList(currentColorListKey, colorList);
    }
    public void SetParseableColorList(String key, String colorList)
    {
      if (!acopColorHash.ContainsKey(key))
        acopColorHash.Add(key, new AcopColorList());
      AcopColorList lst;
      if (key == null) key = PRIMARY_COLOR_LIST;
      if (acopColorHash.TryGetValue(key, out lst))
      {
        lst.parseableString = colorList;
        lst.colorList.Clear();
        if (colorList == null || colorList.Length == 0) return;
        String[] parts = colorList.Split(new char[] { ';' }, StringSplitOptions.RemoveEmptyEntries);
        foreach (String c in parts) AddColorToList(key, c);
      }
    }
    private String currentColorListKey = null;
    public String GetCurrentColorListKey()
    {
      return currentColorListKey != null ? currentColorListKey : PRIMARY_COLOR_LIST;
    }
    public String GetParseableColorList()
    {
      return GetParseableColorList(PRIMARY_COLOR_LIST);
    }
    public String GetParseableColorList(String key)
    {
      AcopColorList lst;
      if (key == null) key = PRIMARY_COLOR_LIST;
      if (acopColorHash.TryGetValue(key, out lst)) return lst.parseableString;
      return "";
    }
    protected ColorListEditorValue getColorListEditorValue()
    {
      return getColorListEditorValue(PRIMARY_COLOR_LIST);
    }
    protected ColorListEditorValue getColorListEditorValue(String key)
    {
      if (key == null) key = PRIMARY_COLOR_LIST;
      currentColorListKey = key;
      if (!acopColorHash.ContainsKey(key))
        acopColorHash.Add(key, new AcopColorList(new ColorListEditorValue("", key, this)));
      AcopColorList lst;
      if (acopColorHash.TryGetValue(key, out lst)) return lst.editorValue;
      return null;
    }
    protected void setColorListEditorValue(ColorListEditorValue value)
    {
      setColorListEditorValue(PRIMARY_COLOR_LIST, value);
    }
    protected void setColorListEditorValue(String key, ColorListEditorValue value)
    {
      if (key == null) key = PRIMARY_COLOR_LIST;
      currentColorListKey = key;
      if (value == null) value = new ColorListEditorValue("", key, this);
      AcopColorList lst;
      if (acopColorHash.TryGetValue(key, out lst))
        acopColorHash.Remove(key);
      acopColorHash.Add(key, new AcopColorList(value));
    }
    /// <summary>
    /// Gets the current color list as an array of colors
    /// </summary>
    /// <returns>the current color list as an array of colors</returns>
    public Color[] GetColorList()
    {
      return GetColorList(PRIMARY_COLOR_LIST);
    }
    public Color[] GetColorList(String key)
    {
      AcopColorList lst;
      if (key == null) key = PRIMARY_COLOR_LIST;
      if (acopColorHash.TryGetValue(key, out lst))
      {
        if (lst.colorList.Count == 0) return null;
        return lst.colorList.ToArray();
      }
      return null;
    }
    protected ThresholdListEditorValue thresholdListEditorValue = null;
    protected ThresholdListEditorValue getThresholdListEditorValue()
    {
      if (thresholdListEditorValue == null)
        thresholdListEditorValue = new ThresholdListEditorValue("", this);
      return thresholdListEditorValue;
    }
    protected void setThresholdListEditorValue(ThresholdListEditorValue value)
    {
      if (value == null) value = new ThresholdListEditorValue("", this);
      thresholdListEditorValue = value;
    }
    protected static AcopUserComponent GetInstance()
    {
      return instance;
    }
    #region BlinkerRoutines
    private System.Threading.Timer blinker = null;
    public enum BlinkMode
    {
      off = 0, onWarning = 1, always = 2
    }
    protected BlinkMode blinkMode = BlinkMode.off;
    protected void setBlinkMode(BlinkMode value)
    {
      if (value == blinkMode) return;
      blinkMode = value;
      if (value != BlinkMode.off) startBlinker();
    }
    protected Int32 blinkerIntensity = 50;
    protected void setBlinkerIntensity(Int32 value)
    {
      if (value == blinkerIntensity) return;
      if (value < -100) value = -100;
      if (value > 100) value = 100;
      blinkerIntensity = value;
    }
    protected Int32 blinkerInterval = 500;
    protected void setBlinkerInterval(Int32 value)
    {
      if (value == blinkerInterval) return;
      if (value < 100) value = 100;
      blinkerInterval = value;
      if (blinker != null) blinker.Change(0, value);
    }
    protected Int32 numberBlinkSlots = 2;
    protected void setNumberBlinkerSlots(Int32 value)
    {
      if (value == numberBlinkSlots) return;
      if (value < 2) value = 2;
      numberBlinkSlots = value;
    }
    protected Int32 blinkerSlot = 0;
    protected void setBlinkerSlot(Int32 value)
    {
      if (value == blinkerSlot) return;
      if (value < 0) value = 0;
      blinkerSlot = value;
    }
    protected void startBlinker()
    {
      if (IsDesignMode()) return;
      if (blinker == null) blinker = new System.Threading.Timer(blinkerUpdate, null, 0, blinkerInterval);
    }
    private Int32 currentBlink = 0;
    private delegate void blinkerUpdateDelegate(object state);
    private void blinkerUpdate(object state)
    {
      foreach (Control c in getControlsList())
        if (c.InvokeRequired)
        {
          try
          {
            Invoke(new blinkerUpdateDelegate(blinkerUpdate), state);
          }
          catch (Exception) { }
          return;
        }
      if (blink == null) return;
      currentBlink = (++currentBlink) % numberBlinkSlots;
      blink(blinkerSlot == currentBlink);
    }
    public delegate void BlinkerDelegate(Boolean blink);
    protected BlinkerDelegate blink = null;
    #endregion
    #region LateInitializationRoutines
    private static List<String> nonResponders = new List<String>();
    private int notRespondingCounts = 0;
    private static int allowedNotResponsdingCounts = 2;
    protected void SetLateInitializerProlog(lateInitializerPrologDelegate lateInitializerProlog)
    {
      this.lateInitializerProlog = lateInitializerProlog;
    }
    protected void SetLateInitializerEpilog(lateInitializerEpilogDelegate lateInitializerEpilog)
    {
      this.lateInitializerEpilog = lateInitializerEpilog;
    }
    protected Boolean lateInitializationDone = false;
    protected Boolean isLateInitializationDone(int status)
    {
      if (lateInitializationDone) return true;
      int limit = 100;
      if (status != 0) limit = 10;
      for (int tries = 0; tries < limit; tries++)
      {
        System.Threading.Thread.Sleep(10);
        if (lateInitializationDone) return true;
      }
      return false;
    }
    protected AcopControl.AcopLink earlyRespondingLink = null;
    private void lateInitializerUpdate(object sender, EventArgs e)
    {
      String key = AcopControl.AcopLink.GetConnectionHostAddress(linkProt, linkAddr, linkProp);
      if (lateInitializerProlog != null)
      {
        lateInitializerProlog();
        if (!displaySettingsAcquired)
        { // timeout getting info from source ?
          lateInitializer.Interval = 2000; // keep trying ...
          if (++notRespondingCounts == allowedNotResponsdingCounts)
          {
            if (key != null) lock (nonResponders) if (!nonResponders.Contains(key))
                {
                  nonResponders.Add(key);
                  String txt = "The target " + key + " appears to be off-line!";
                  MessageBox.Show(txt);
                }
          }
          if (notRespondingCounts > allowedNotResponsdingCounts)
          {
            hasAutoAttached = false;
            checkAutoAttach();
          }
          return;
        }
      }
      if (notRespondingCounts >= allowedNotResponsdingCounts)
      {
        if (key != null) lock (nonResponders) if (!nonResponders.Contains(key))
        {
          nonResponders.Remove(key);
        }
        notRespondingCounts = 0;
        hasAutoAttached = false;
        checkAutoAttach();
      }
      if (lateInitializerEpilog != null) lateInitializerEpilog();
      lateInitializer.Enabled = false;
      lateInitializationDone = true;
      if (earlyRespondingLink != null)
      {
        if (updater != null) updater(earlyRespondingLink);
        earlyRespondingLink = null;
      }
      if (needsAutoAttach())
      {
        hasAutoAttached = false;
        checkAutoAttach();
      }
    }
    protected void StartLateInitialization()
    {
      if (!isDesignMode)
      {
        lateInitializer.Enabled = true;
      }
    }
    #endregion
    private static String lastContextSetting = null;
    private static String lastServerSetting = null;
    private static String lastSubsystemSetting = null;
    private static String lastDeviceSetting = null;
    private static String lastPropertySetting = null;
    public static String GetLastContextSetting()
    {
      return lastContextSetting;
    }
    public static String GetLastServerSetting()
    {
      return lastServerSetting;
    }
    public static String GetLastDeviceSetting()
    {
      return lastDeviceSetting;
    }
    public static String GetLastPropertySetting()
    {
      return lastPropertySetting;
    }
    public AcopUserComponent()
    {
      Thread.CurrentThread.CurrentCulture = System.Globalization.CultureInfo.InvariantCulture;
      initialize();
      isDesignMode = (LicenseManager.UsageMode == LicenseUsageMode.Designtime);
      loadTransport(this);
      setNotifier(notifyLinkUpdate);
      if (IsDesignMode())
      {
        linkCtx = AcopControl.AcopLink.GetDefaultContext(LinkProtocol);
        linkSrv = AcopControl.AcopLink.GetDefaultServer(LinkProtocol);
        linkDev = AcopControl.AcopLink.GetDefaultDevice(LinkProtocol);
        linkProp = AcopControl.AcopLink.GetDefaultProperty(LinkProtocol);
        if (lastContextSetting != null) linkCtx = lastContextSetting;
        if (lastServerSetting != null) linkSrv = lastServerSetting;
        if (lastDeviceSetting != null) linkDev = lastDeviceSetting;
        if (lastPropertySetting != null) linkProp = lastPropertySetting;
      }
    }
    ~AcopUserComponent()
    {
      AcopControl.AcopLink.ClearAcopLinkTable(component);
    }

    protected bool IsDesignMode()
    {
      return isDesignMode;
    }

    protected void setNotifier(AcopControl.AcopLink.acopTransportUpdate notifier)
    {
      update = notifier;
    }

    protected Component GetComponentInstance()
    {
      return component;
    }
    protected void SetComponentInstance(Component control)
    {
      component = control;
    }

    protected void loadTransport(Component acopControl)
    {
      component = acopControl;
      instance = this;
    }
    void initTransport()
    {
      lastLinkSrv = null;
      lastLinkCtx = null;
      lastLinkDev = null;
      lastLinkPrp = null;
      lastLinkSub = null;
    }
    int getDefaultInterval() { return defaultInterval; }
    void setDefaultInterval(int value)
    {
      if (defaultInterval < 0) defaultInterval = 0;
      if (defaultInterval == 0)
        defaultInterval = AcopControl.AcopLink.GetDefaultInterval();
      defaultInterval = value;
    }
    object GetLinkDataOut() { return linkDataOut; }
    object GetLinkDataIn() { return linkDataIn; }
    int GetLinkFlags() { return linkFlags; }
    private String getLinkAddr()
    {
      instance = this;
      return linkAddr == null ? "" : linkAddr;
    }
    private void setLinkAddr(String adr)
    {
      instance = this;
      if (!adr.Equals(linkAddr, StringComparison.OrdinalIgnoreCase))
      {
        displaySettingsAcquired = false;
      }
      linkAddr = adr;
      validateLinkAddr();
      checkAutoAttach();
      applyToolTipText(null);
    }
    private void validateLinkAddr()
    {
      if (linkAddr == null) return;
      String[] parts = linkAddr.Split(new char[] { '/' }, StringSplitOptions.RemoveEmptyEntries);
      if (parts.Length > 2)
      {
        if (linkCtx == null || linkCtx.Length == 0) linkCtx = parts[0];
        if (linkSrv == null || linkSrv.Length == 0) linkSrv = parts[1];
        if (linkDev == null || linkDev.Length == 0) linkDev = parts[2];
      }
    }
    private String getLinkGroupList()
    {
      instance = this;
      return linkGroupList == null ? "" : linkGroupList;
    }

    private void setLinkGroupList(String lst)
    {
      instance = this;
      if (!lst.Equals(linkGroupList, StringComparison.OrdinalIgnoreCase))
      {
        displaySettingsAcquired = false;
      }
      linkGroupList = lst;
    }

    private Dictionary<String, LinkGroupListItem> grpLst = new Dictionary<String, LinkGroupListItem>();
    /// <summary>
    /// Gets a list of the grouped links which are attached to this control
    /// </summary>
    /// <returns>a list of the grouped links which are attached to this control</returns>
    public LinkGroupListItem[] GetAutoGroupLinks()
    {
      LinkGroupListItem grpLstItem = null;
      String[] lst = null;
      if (linkGroupList == null || linkGroupList.Length == 0) return null;
      lst = linkGroupList.Split(new char[] { ';' }, StringSplitOptions.RemoveEmptyEntries);
      foreach (String item in lst)
      {
        if (grpLst.ContainsKey(item)) continue;
        grpLstItem = new LinkGroupListItem(this, item);
        if (grpLstItem.IsWellFormed()) grpLst.Add(item, grpLstItem);
      }
      if (grpLst.Count > 0)
      { // and now add the current link settings if not already there ...
        String item = LinkAddress + "/" + LinkProperty;
        if (!grpLst.ContainsKey(item))
        {
          grpLstItem = new LinkGroupListItem(this, item);
          if (grpLstItem.IsWellFormed()) grpLst.Add(item, grpLstItem);
        }
      }
      return grpLst.Values.ToArray();
    }

    protected void setToolTip(ToolTip tto)
    {
      thisTT = tto;
    }
    delegate void ApplyToolTipDelegate(String statusString);
    protected void applyToolTipText(String statusString)
    {
      foreach (Control c in getControlsList())
        if (c.InvokeRequired)
        {
          try
          {
            Invoke(new ApplyToolTipDelegate(applyToolTipText), statusString);
          }
          catch (Exception) { }
          return;
        }
      if (thisTT == null) return;
      String ttt = linkAddr + "[" + linkProp + "]";
      if (statusString != null && statusString.Length > 0) ttt += " : " + statusString;
      foreach (Control uc in thisCtrlsList) thisTT.SetToolTip(uc, ttt);
    }
    protected void applyToolTipText(Control c, String ToolTipString)
    {
      if (thisTT == null) return;
      thisTT.SetToolTip(c, ToolTipString);
    }
    protected void setControlsList(Control[] list, Boolean addToolTip)
    {
      thisCtrlsList = list;
      if (addToolTip && thisTT == null) setToolTip(new ToolTip());
    }
    protected void setControlsList(Control[] list)
    {
      thisCtrlsList = list;
    }
    protected Control[] getControlsList()
    {
      return thisCtrlsList;
    }
    protected Boolean hasControlsList()
    {
      if (thisCtrlsList == null) return false;
      return thisCtrlsList.Length > 0;
    }
    /// <summary>
    /// Gets the integer flag representation of the input flags string
    /// </summary>
    /// <param name="flags"></param>
    /// <returns>the integer flag representation of the input flags string</returns>
    public int GetFlags(String flags)
    {
      if (flagsDict == null) flagsDict = AcopControl.AcopLink.GetFlagsList(linkProt);
      if (flagsDict == null || flags == null) return 0;
      if (!flagsDict.ContainsKey(flags)) return 0;
      int v = 0;
      flagsDict.TryGetValue(flags, out v);
      return v;
    }
    private String GetLinkFlagsString() { return linkFlagsString; }
    private void SetLinkFlagsString(String value) { linkFlagsString = value; }
    private String getEnumLinkFlags()
    {
      instance = this;
      return GetLinkFlagsString();
    }
    private void setEnumLinkFlags(String flags)
    {
      instance = this;
      SetLinkFlagsString(flags);
    }
    private String getLinkProt()
    {
      instance = this;
      return linkProt == null ? AcopControl.AcopLink.GetDefaultProtocol() : linkProt;
    }
    private void setLinkProt(String protocol)
    {
      instance = this;
      linkProt = protocol;
      browsingServicesAvailable = true;
      try
      {
        AcopControl.AcopLink lnk = AcopControl.AcopLink.GetAcopLink(linkProt, linkAddr, linkProp, component);
        if (lnk != null) lnk.SetProtocol(linkProt);
      }
      catch (Exception) { }
    }
    private String getLinkCtxSetting()
    {
      instance = this;
      return this.linkCtx;
    }
    private Boolean browsingServicesAvailable = true;

    [ToolboxItem(false)]
    [Browsable(false), EditorBrowsable(EditorBrowsableState.Never)]
    public Boolean BrowsingServiceAvailable { get { return browsingServicesAvailable; } set { browsingServicesAvailable = value; } }
    private String getFirstContext()
    {
      if (!IsDesignMode()) return null;
      if (lastLinkCtx != null) return lastLinkCtx;
      if (!BrowsingServiceAvailable) return UNAVAILABLE;
      String[] lst = AcopControl.AcopLink.GetContextList(linkProt);
      if (lst != null && lst[0] != null)
      {
        if (lst[0].Equals(UNAVAILABLE, StringComparison.CurrentCultureIgnoreCase))
          browsingServicesAvailable = false;
        return lst[0];
      }
      return "";
    }
    private String getLinkContext()
    {
      instance = this;
      return linkCtx == null ? getFirstContext() : linkCtx;
    }
    private void setLinkContext(String context)
    {
      instance = this;
      linkCtx = lastLinkCtx = context;
      linkSrv = null;
      srvStale = true; ctxStale = false;
      if (context != null && context.Equals(UNAVAILABLE,StringComparison.CurrentCultureIgnoreCase))
      {
        if (!initialPassDone)
          lastLinkCtx = context = AcopControl.AcopLink.GetDefaultContext(LinkProtocol);
        else
          BrowsingServiceAvailable = false;
      }
      if (!IsDesignMode())
      {
        if (linkSrv != null && linkDev != null)
        {
          linkAddr = AcopControl.AcopLink.GetConnectionAddress(linkProt, linkCtx, linkSubsystem, linkSrv, linkDev);
          setLinkAddr(linkAddr);
        }
      }
      else
      {
        lastContextSetting = context;
        if (initialPassDone)
        {
          lastLinkSrv = null;
          setLinkServer(getFirstServer());
          linkDev = lastLinkDev = null;
          linkProp = lastLinkPrp = null;
        }
      }
    }
    private String getFirstSubsystem()
    {
      if (!IsDesignMode()) return null;
      if (!ctxStale) return linkSubsystem;
      if (linkCtx == null) linkCtx = getFirstContext();
      if (lastLinkSub != null) return lastLinkSub;
      if (!BrowsingServiceAvailable) return UNAVAILABLE;
      String[] lst = AcopControl.AcopLink.GetSubSystemList(linkProt, linkCtx);
      if (lst != null && lst[0] != null) return lst[0];
      return "";
    }
    private String getLinkSubSystem()
    {
      instance = this;
      return linkSubsystem == null ? getFirstSubsystem() : linkSubsystem;
    }
    private void setLinkSubSystem(String subsystem)
    {
      instance = this;
      linkSubsystem = lastLinkSub = subsystem;
      srvStale = true; subStale = false;
    }
    private String getLinkSrvSetting()
    {
      instance = this;
      return this.linkSrv;
    }
    private String getFirstServer()
    {
      if (!IsDesignMode()) return null;
      if (!ctxStale && !srvStale && linkSrv != null) return linkSrv;
      if (linkCtx == null) linkCtx = getFirstContext();
      if (lastLinkSrv != null) return lastLinkSrv;
      if (!BrowsingServiceAvailable) return UNAVAILABLE;
      String[] lst = AcopControl.AcopLink.GetServerList(linkProt, linkCtx, linkSubsystem);
      if (lst != null && lst[0] != null) return lst[0];
      return "";
    }
    private String getLinkServer()
    {
      instance = this;
      return linkSrv == null ? getFirstServer() : linkSrv;
    }
    private Boolean isDesignTimeSetServer()
    {
      if (IsDesignMode())
      {
        if (linkDev.CompareTo(UNAVAILABLE) == 0 ||
            linkProp.CompareTo(UNAVAILABLE) == 0)
          return true;
      }
      return false;
    }
    private void setLinkServer(String server)
    {
      instance = this;
      linkSrv = lastLinkSrv = server;
      srvStale = false;
      if (server != null && server.Equals(UNAVAILABLE, StringComparison.CurrentCultureIgnoreCase))
      {
        BrowsingServiceAvailable = false;
      }
      if (IsDesignMode())
      {
        if (server != null && server.Length > 0)
          lastServerSetting = server;
      }
      if (initialPassDone || isDesignTimeSetServer())
      {
        linkDev = null; lastLinkDev = null;
        linkProp = lastLinkPrp = null;
        devStale = true;
        if (IsDesignMode())
        {
          hasPropertyQueryPrecedence = false;
          hasDeviceQueryPrecedence = false;
          setLinkDevice(getFirstDevice());
          if (hasPropertyQueryPrecedence)
          {
            linkDev = lastLinkDev = null;
            linkAddr = null;
            setLinkProp(getFirstProperty());
          }
        }
      }
      if (linkCtx != null && linkDev != null)
      {
        linkAddr = AcopControl.AcopLink.GetConnectionAddress(linkProt, linkCtx, linkSubsystem, linkSrv, linkDev);
        setLinkAddr(linkAddr);
      }
      checkAutoAttach(); // in initializeComponent this is alphabetically the last thing to happen
      initialPassDone = true;
    }
    private String getFirstDevice()
    {
      if (!IsDesignMode()) return null;
      if (linkDev != null && linkDev.CompareTo(UNAVAILABLE) == 0)
      {
        linkDev = lastLinkDev = null;
      }
      if (linkProp != null && linkProp.CompareTo(UNAVAILABLE) == 0)
      {
        if (linkDev != null)
        {
          linkProp = lastLinkPrp = null;
        }
      }
      if (!devStale && linkDev != null) return linkDev;
      if (lastLinkDev != null) return lastLinkDev;
      if (!BrowsingServiceAvailable) return UNAVAILABLE;
      if (linkCtx != null && linkSrv != null)
      {
        if (linkProp == null && !hasPropertyQueryPrecedence)
        {
          hasPropertyQueryPrecedence = AcopControl.AcopLink.HasPropertyQueryPrecedence(linkProt, linkCtx, linkSubsystem, linkSrv, linkProp);
        }
        if (hasPropertyQueryPrecedence)
        { // first query
          prpStale = true;
          if (linkProp == null) return UNAVAILABLE;
        }
        String[] lst = AcopControl.AcopLink.GetDeviceList(linkProt, linkCtx, linkSubsystem, linkSrv, linkProp);
        if (lst != null && lst[0] != null) return lst[0];
      }
      return "";
    }
    private String getLinkDevice()
    {
      instance = this;
      String dev = linkDev;
      if (dev == null)
      {
        linkCtx = this.getLinkCtxSetting();
        linkSrv = this.getLinkSrvSetting();
        linkDev = dev = getFirstDevice();
        if (linkCtx != null && linkSrv != null)
        {
          linkAddr = AcopControl.AcopLink.GetConnectionAddress(linkProt, linkCtx, linkSubsystem, linkSrv, dev);
          setLinkAddr(linkAddr);
        }
      }
      return dev == null ? UNAVAILABLE : dev;
    }
    private void setLinkDevice(String device)
    {
      instance = this;
      linkDev = lastLinkDev = device;
      linkCtx = this.getLinkCtxSetting();
      linkSrv = this.getLinkSrvSetting();
      //MessageBox.Show("setLinkDevice: /" + linkCtx + "/" + linkSrv);
      if (linkCtx != null && linkSrv != null)
      {
        linkAddr = AcopControl.AcopLink.GetConnectionAddress(linkProt, linkCtx, linkSubsystem, linkSrv, linkDev);
        setLinkAddr(linkAddr);
        if (IsDesignMode())
        {
          lastDeviceSetting = device;
        }
      }
      devStale = false;
      if (hasDeviceQueryPrecedence && !device.Contains("*"))
      {
        prpStale = true;
        if (IsDesignMode())
        {
          linkProp = lastLinkPrp = null;
          setLinkProp(getFirstProperty());
        }
      }
    }
    private String getLinkProp()
    {
      instance = this;
      String prp = linkProp;
      if (prp == null)
      {
        linkProp = prp = getFirstProperty();
      }
      return prp == null ? UNAVAILABLE : prp;
    }
    private void setLinkProp(String prp)
    {
      instance = this;
      if (!prp.Equals(linkProp, StringComparison.OrdinalIgnoreCase))
      {
        displaySettingsAcquired = false;
      }
      linkProp = lastLinkPrp = prp;
      if (IsDesignMode())
      {
        lastPropertySetting = prp;
      }
      checkAutoAttach();
      prpStale = false;
      if (hasPropertyQueryPrecedence)
      {
        devStale = true;
        if (IsDesignMode())
        {
          linkDev = lastLinkDev = null;
          setLinkDevice(getFirstDevice());
        }
      }
      applyToolTipText(null);
    }
    private String getFirstProperty()
    {
      if (!IsDesignMode()) return null;
      if (linkProp != null && linkProp.CompareTo(UNAVAILABLE) == 0)
      {
        linkProp = lastLinkPrp = null;
      }
      if (linkDev != null && linkDev.CompareTo(UNAVAILABLE) == 0)
      {
        if (linkProp != null)
        {
          linkDev = lastLinkDev = null;
        }
      }
      if (!prpStale && linkProp != null) return linkProp;
      if (linkCtx == null) linkCtx = getFirstContext();
      if (linkSrv == null) linkSrv = getFirstServer();
      if (linkDev == null) linkDev = getFirstDevice();
      if (lastLinkPrp != null) return lastLinkPrp;
      if (!BrowsingServiceAvailable) return UNAVAILABLE;
      if (!hasDeviceQueryPrecedence)
      {
        hasDeviceQueryPrecedence = AcopControl.AcopLink.HasDeviceQueryPrecedence(linkProt, linkCtx, linkSubsystem, linkSrv);
      }
      String[] lst = AcopControl.AcopLink.GetPropertyList(linkProt, linkCtx, linkSubsystem, linkSrv, linkDev);
      if (lst != null && lst[0] != null) return lst[0];
      return "";
    }
    /// <summary>
    /// Returns the output data object bound to the default transport link
    /// </summary>
    /// <returns>the output data object bound to the default transport link</returns>
    public object GetLinkOutput()
    {
      AcopControl.AcopLink lnk = AcopControl.AcopLink.GetAcopLink(linkProt, linkAddr, linkProp, component);
      if (lnk == null) return null;
      object o = lnk.GetOutputDataObject();
      valueType = (o is Array) ? ((Array)o).GetType().GetElementType() : o.GetType();
      return o;
    }
    /// <summary>
    /// Returns the appropriate X-Axis for the data bound to the default transport link (for use in the chart)
    /// </summary>
    /// <returns>the appropriate X-Axis for the data bound to the default transport link</returns>
    public double[] GetLinkOutputXAxis()
    {
      linkProt = getLinkProt().ToString();
      AcopControl.AcopLink lnk = AcopControl.AcopLink.GetAcopLink(linkProt, linkAddr, linkProp, component);
      if (lnk == null) return null;
      return lnk.GetOutputXAxis();
    }
    /// <summary>
    /// Returns the appropriate Y-Axis for the data bound to the default transport link (for use in the chart)
    /// </summary>
    /// <returns>the appropriate Y-Axis for the data bound to the default transport link</returns>
    public double[] GetLinkOutputYAxis()
    {
      AcopControl.AcopLink lnk = AcopControl.AcopLink.GetAcopLink(linkProt, linkAddr, linkProp, component);
      if (lnk == null) return null;
      return lnk.GetOutputYAxis();
    }
    /// <summary>
    /// Returns a list of discovered channel names for the data bound to the default transport link
    /// (In case the transport link is bound to a multi-channel array).
    /// </summary>
    /// <returns>a list of discovered channel names for the data bound to the default transport link</returns>
    public String[] GetLinkOutputChannelNames()
    {
      AcopControl.AcopLink lnk = AcopControl.AcopLink.GetAcopLink(linkProt, linkAddr, linkProp, component);
      if (lnk == null) return null;
      return lnk.GetOutputChannelNames();
    }
    protected AcopControl.AcopLink GetAttachedLink()
    {
      return AcopControl.AcopLink.GetAcopLink(linkProt, linkAddr, linkProp, linkDataOut, linkDataIn, linkFlags, component);
    }
    protected Boolean IsOutputArrayObject()
    {
      if (autoLinkCapacity > 1) return true;
      Object o = linkDataOut;
      if (o == null)
      {
        AcopControl.AcopLink lnk = AcopControl.AcopLink.GetAcopLink(linkProt, linkAddr, linkProp, component);
        o = AcopControl.AcopLink.GetOutputDataObject(lnk, linkAddr, linkProp);
      }
      if (o is Array)
      {
        valueType = ((Array)o).GetType().GetElementType();
        if (((Array)o).Length > 1) return true;
      }
      valueType = o == null ? null : o.GetType();
      return false;
    }
    protected int GetOutputArrayLength()
    {
      Object o = linkDataOut;
      if (o == null)
      {
        AcopControl.AcopLink lnk = AcopControl.AcopLink.GetAcopLink(linkProt, linkAddr, linkProp, component);
        o = AcopControl.AcopLink.GetOutputDataObject(lnk, linkAddr, linkProp);
      }
      if (o is Array)
      {
        valueType = ((Array)o).GetType().GetElementType();
        return ((Array)o).Length;
      }
      valueType = o == null ? null : o.GetType();
      return 1;
    }
    private void assignAddressParts(String Address, String Property)
    {
      linkAddr = Address;
      linkProp = Property;
      String[] parts = Address.Split(new char[] { '/' }, StringSplitOptions.RemoveEmptyEntries);
      if (parts != null && parts.Length != 0)
      {
        linkCtx = parts[0];
        if (parts.Length > 1) linkSrv = parts[1];
        if (parts.Length > 2) linkDev = parts[2];
      }
    }
    /// <summary>
    /// Reassigns the current bound default link to the address parameters given.
    /// </summary>
    /// <param name="Address">the desired new target link address</param>
    /// <param name="Property">the desired new target link property</param>
    /// <returns>a positive link handle to the new link upon success, otherwise -1.</returns>
    public int Reassign(String Address, String Property)
    {
      lock (notifyMutex)
      {
        AcopControl.AcopLink[] lst = AcopControl.AcopLink.GetAcopLinkTable(this);
        if (lst == null || lst.Length == 0)
        { // nobody in the link table !
          assignAddressParts(Address, Property);
          return Attach();
        }
        if (lst.Length > 1)
        {
          //throw new Exception("This Acop Component is attached to more than one Link !");
          for (int i = 1; i < lst.Length; i++)
          {
            lst[i].Detach();
            AcopControl.AcopLink.RmvAcopLink(lst[i].GetConnectionKey());
          }
        }
        // close the current monitor
        AcopControl.AcopLink lnk = lst[0];
        lnk.Detach();
        AcopControl.AcopLink.RmvAcopLink(lnk.GetConnectionKey());
        if (this is AcopChart)
        { // special instructions for chart
          AcopChart ac = (AcopChart)this;
          int id = lnk.GetDisplayId();
          if (id >= 0)
          {
            Color c = ac.histo[id].penColor;
            AcopChart.AcopPlotMode m = AcopChart.GetPlotMode(ac.histo[id].displayModeSetting);
            ac.ClearScreen(id);
            ac.ForegroundColor = c;
            ac.PlotMode = m;
          }
          for (int i = 1; i < lst.Length; i++)
          {
            ac.ClearScreen(lst[i].GetDisplayId());
            lst[i].SetDisplayId(-1);
          }
        }
        if (!descriptionSetByDesigner) description = null;
        displaySettingsAcquired = false;
        lnk.SetDisplayId(-1);
        //hasAutoAttached = false;
        //lateInitializationDone = false; //?
        //lateInitializer.Enabled = true; //?
        //
        assignAddressParts(Address, Property);
        applyToolTipText(null);
        return Attach();
      }
    }
    /// <summary>
    /// Attaches a transport link to the link target contained in the 
    /// LinkAddress and LinkProperty fields.  Will attempt to discover 
    /// and use the default output data object
    /// </summary>
    /// <returns>A positive link ID or -1 if the link cannot be established.</returns>
    public int Attach()
    {
      int flgs = GetFlags(linkFlagsString);
      Object inpt = null;
      if (linkInputValue != null && linkInputValue.Length > 0)
      {
        Int32 iv;
        Single fv;
        if (Int32.TryParse(linkInputValue, out iv)) inpt = new Int32[] { iv };
        else if (Single.TryParse(linkInputValue, out fv)) inpt = new Single[] { fv };
        else inpt = linkInputValue;
      }
      AcopControl.AcopLink lnk = AcopControl.AcopLink.GetAcopLink(linkProt, linkAddr, linkProp, component);
      return Attach(AcopControl.AcopLink.GetOutputDataObject(lnk, linkAddr, linkProp, null, autoLinkCapacity), inpt, flgs, DefaultMonitorInterval, null);
    }
    /// <summary>
    /// Attaches a transport link to the link target contained in the 
    /// LinkAddress and LinkProperty fields. 
    /// </summary>
    /// <param name="DataOut">A reference to the output data object which is to be filled</param>
    /// <param name="DataIn">A reference to an input data object to be sent to the link address</param>
    /// <param name="Flags">Any optional link flags</param>
    /// <param name="interval">An update interval (or timeout) in milliseconds</param>
    /// <param name="reference">Any reference which should be returned in the link's update event</param>
    /// <returns>A positive link ID or -1 if the link cannot be established.</returns>
    public int Attach(object DataOut, object DataIn, int Flags, int interval, object reference)
    {
      AcopControl.AcopLink lnk = AcopControl.AcopLink.GetAcopLink(linkProt, linkAddr, linkProp, DataOut, DataIn, Flags, component);
      if (lnk.IsAttached()) return lnk.GetLinkId();
      lnk.SetTransportUpdatedEvent(update);
      lnk.SetAcopControl(component);
      linkDataOut = DataOut; linkDataIn = DataIn; linkFlags = Flags;
      return lnk.Attach(DataOut, DataIn, Flags, interval, reference);
    }
    /// <summary>
    /// Attaches a transport link to the link target contained in the 
    /// LinkAddress and LinkProperty fields. 
    /// </summary>
    /// <param name="DataOut">A reference to the output data object which is to be filled</param>
    /// <param name="Offset">A specific offset into the output data object (array) reference. Output data will begin filling at this offset.</param>
    /// <param name="Length">The length of the output data object (array) which is to be filled in.</param>
    /// <param name="DataIn">A reference to an input data object to be sent to the link address</param>
    /// <param name="Flags">Any optional link flags</param>
    /// <param name="interval">An update interval (or timeout) in milliseconds</param>
    /// <param name="reference">Any reference which should be returned in the link's update event</param>
    /// <returns>A positive link ID or -1 if the link cannot be established.</returns>
    public int Attach(object DataOut, int Offset, int Length, object DataIn, int Flags, int interval, object reference)
    {
      AcopControl.AcopLink lnk = AcopControl.AcopLink.GetAcopLink(linkProt, linkAddr, linkProp, DataOut, DataIn, Flags, component);
      if (lnk.IsAttached()) return lnk.GetLinkId();
      lnk.SetTransportUpdatedEvent(update);
      lnk.SetAcopControl(component);
      linkDataOut = DataOut; linkDataIn = DataIn; linkFlags = Flags;
      return lnk.Attach(DataOut, Offset, Length, DataIn, Flags, interval, reference);
    }
    /// <summary>
    /// Detaches the link specifed by the link parameters from the acop control.
    /// </summary>
    /// <param name="Protocol">Control system protocol</param>
    /// <param name="Address">The link address</param>
    /// <param name="Property">The link property</param>
    /// <param name="DataOut">The output data object</param>
    /// <param name="DataIn">The input data object</param>
    /// <returns>0 upon success or -1 on error</returns>
    public int Detach(String Protocol, String Address, String Property, object DataOut, object DataIn)
    {
      AcopControl.AcopLink lnk = AcopControl.AcopLink.GetAcopLink(Protocol, Address, Property, DataOut, DataIn, linkFlags, component);
      if (lnk != null)
      {
        int rc = lnk.Detach();
        AcopControl.AcopLink.RmvAcopLink(lnk.GetConnectionKey());
        if (isGrouped && numAcopLinksPending > 0) numAcopLinksPending--;
        if (this is AcopChart)
        {
          AcopChart ac = (AcopChart)this;
          if (ac.AutoTrendUpdate || ac.AutoLinkAttach)
          {
            ac.ClearScreen(lnk.GetDisplayId());
            lnk.SetDisplayId(-1);
          }
        }
        return rc;
      }
      return -1;
    }
    /// <summary>
    /// Detaches the link specifed by the link parameters from the acop control.
    /// </summary>
    /// <param name="Address">The link address</param>
    /// <param name="Property">The link property</param>
    /// <param name="DataOut">The output data object</param>
    /// <param name="DataIn">The input data object</param>
    /// <returns>0 upon success or -1 on error</returns>
    public int Detach(String Address, String Property, object DataOut, object DataIn)
    {
      return Detach(linkProt, Address, Property, DataOut, DataIn);
    }
    /// <summary>
    /// Detaches the link specifed by the link parameters from the acop control.
    /// </summary>
    /// <param name="Protocol">Control system protocol</param>
    /// <param name="Address">The link address</param>
    /// <param name="Property">The link property</param>
    /// <returns>0 upon success or -1 on error</returns>
    public int Detach(String Protocol, String Address, String Property)
    {
      return Detach(Protocol, Address, Property, linkDataOut, linkDataIn);
    }
    /// <summary>
    /// Detaches the link specifed by the link parameters from the acop control.
    /// </summary>
    /// <param name="Address">The link address</param>
    /// <param name="Property">The link property</param>
    /// <returns>0 upon success or -1 on error</returns>
    public int Detach(String Address, String Property)
    {
      return Detach(linkProt.ToString(), Address, Property);
    }
    /// <summary>
    /// Detaches all links attached to the acop control.
    /// </summary>
    /// <returns>0 upon success or -1 on error</returns>
    public int Detach()
    {
      AcopControl.AcopLink.ClearAcopLinkTable(component);
      return 0;
    }
    /// <summary>
    /// Executes a single synchronous transport call to the link target contained in the 
    /// LinkAddress and LinkProperty fields.  Will attempt to discover 
    /// and use the default output data object
    /// </summary>
    /// <returns>0 upon success or -1 if there is a link error.</returns>
    public int Execute()
    {
      int flgs = GetFlags(linkFlagsString);
      AcopControl.AcopLink lnk = AcopControl.AcopLink.GetAcopLink(linkProt, linkAddr, linkProp, component);
      return Execute(AcopControl.AcopLink.GetOutputDataObject(lnk, linkAddr, linkProp), null, flgs, DefaultMonitorInterval);
    }
    /// <summary>
    /// Executes a single synchronous transport call to the link target contained in the 
    /// LinkAddress and LinkProperty fields.  
    /// </summary>
    /// <param name="DataOut">A reference to the output data object which is to be filled</param>
    /// <param name="DataIn">A reference to an input data object to be sent to the link address</param>
    /// <param name="Flags">Any optional link flags</param>
    /// <param name="timeout">timeout in milliseconds</param>
    /// <returns>0 upon success or -1 if there is a link error.</returns>
    public int Execute(object DataOut, object DataIn, int Flags, int timeout)
    {
      AcopControl.AcopLink lnk = AcopControl.AcopLink.GetAcopLink(linkProt, linkAddr, linkProp, DataOut, DataIn, Flags, component);
      linkDataOut = DataOut; linkDataIn = DataIn; linkFlags = Flags;
      return lnk.Execute(DataOut, DataIn, Flags, timeout);
    }
    private int Execute(AcopControl.AcopLink lnk, object DataOut, object DataIn, int Flags, int timeout)
    {
      linkDataOut = DataOut; linkDataIn = DataIn; linkFlags = Flags;
      return lnk.Execute(DataOut, DataIn, Flags, timeout);
    }
    /// <summary>
    /// Executes a single synchronous transport call to the link target contained in the 
    /// LinkAddress and LinkProperty fields.  
    /// </summary>
    /// <param name="DataOut">A reference to the output data object which is to be filled</param>
    /// <param name="Offset">A specific offset into the output data object (array) reference. Output data will begin filling at this offset.</param>
    /// <param name="Length">The length of the output data object (array) which is to be filled in.</param>
    /// <param name="DataIn">A reference to an input data object to be sent to the link address</param>
    /// <param name="Flags">Any optional link flags</param>
    /// <param name="timeout">timeout in milliseconds</param>
    /// <returns>0 upon success or -1 if there is a link error.</returns>
    public int Execute(object DataOut, int Offset, int Length, object DataIn, int Flags, int timeout)
    {
      AcopControl.AcopLink lnk = AcopControl.AcopLink.GetAcopLink(linkProt, linkAddr, linkProp, DataOut, DataIn, Flags, component);
      linkDataOut = DataOut; linkDataIn = DataIn; linkFlags = Flags;
      return lnk.Execute(DataOut, Offset, Length, DataIn, Flags, timeout);
    }
    /// <summary>
    /// Applies a single link update to the acop user component if the
    /// component is not already in a monitoring state.
    /// </summary>
    /// <returns>0 upon success or -1 if there is a link error.</returns>
    public int SingleUpdate()
    {
      int flgs = GetFlags(linkFlagsString);
      Object inpt = null;
      if (linkInputValue != null && linkInputValue.Length > 0)
      {
        Int32 iv;
        Single fv;
        if (Int32.TryParse(linkInputValue, out iv)) inpt = new Int32[] { iv };
        else if (Single.TryParse(linkInputValue, out fv)) inpt = new Single[] { fv };
        else inpt = linkInputValue;
      }
      AcopControl.AcopLink lnk = AcopControl.AcopLink.GetAcopLink(linkProt, linkAddr, linkProp, component);
      if (lnk == null) return -1;
      if (lnk.IsAttached()) return 0;
      int cc = Execute(lnk,AcopControl.AcopLink.GetOutputDataObject(lnk, linkAddr, linkProp, null, autoLinkCapacity), inpt, flgs, DefaultMonitorInterval);
      if (updater != null) updater(lnk);
      return cc;
    }
    protected Boolean needsAutoAttach()
    {
      return autoAttach && !hasAutoAttached;
    }
    private object checkAutoAttachMutex = new object();
    protected void checkAutoAttach()
    {
      if (!IsDesignMode()) lock(checkAutoAttachMutex)
      {
        if (autoAttach && !hasAutoAttached)
        {
          LinkGroupListItem[] lnks = GetAutoGroupLinks();
          if (lnks != null && lnks.Length > 0)
          {
            if (lnks.Length > 1) LinkGrouped = true;
            String p, pset = LinkProtocol;
            foreach (LinkGroupListItem item in lnks)
            { //TODO: for a group of links we need the settings info for each separately !
              if (item.IsAttached()) continue;
              linkAddr = item.GetAddress();
              linkProp = item.GetProperty();
              AcopTransport.DisplaySettings dsp =
                acquireAndApplyDisplaySettings(linkAddr, linkProp, ref descriptionLabels, true);
              if (dsp == null || dsp.SettingsStale()) return; // didn't get the settings ... 
              p = item.GetProtocol();
              if (!p.Equals(LinkProtocol, StringComparison.CurrentCultureIgnoreCase))
                LinkProtocol = p;
              Attach();
              item.IsAttached(true);
            }
            hasAutoAttached = true;
            LinkProtocol = pset;
            return;
          }
          if (linkAddr != null && linkProp != null)
          {
            AcopTransport.DisplaySettings dsp =
              acquireAndApplyDisplaySettings(linkAddr, linkProp, ref descriptionLabels, true);
            if (dsp == null || dsp.SettingsStale()) return; // didn't get the settings ... 
            Attach();
            hasAutoAttached = true;
          }
        }
      }
    }
    protected String[] acquiredLabels = null;
    protected AcopTransport.DisplaySettings acquireAndApplyDisplaySettings(String address, String property, ref String[] labels, Boolean acquireLabels)
    {
      if (!isGrouped && displaySettingsAcquired && (linkDspSettings != null))
      {
        labels = acquiredLabels;
        return linkDspSettings;
      }
      if (address == null || address.Length == 0) return null;
      AcopTransport.DisplaySettings dsp = AcopControl.AcopLink.GetDisplaySettings(address, property);
      if (dsp != null)
      {
        String[] addrParts = address.Split(new char[] { '/' }, StringSplitOptions.RemoveEmptyEntries);
        String ctx = "", srv = "", dev = "";
        if (addrParts.Length > 0) ctx = addrParts[0];
        if (addrParts.Length > 1) srv = addrParts[1];
        if (addrParts.Length > 2) dev = addrParts[2];
        if (units == null || units.Length == 0)
        {
          units = dsp.GetYUnits();
          if (units.Equals("none", StringComparison.CurrentCultureIgnoreCase)) units = "";
        }
        if (description == null || description.Length == 0)
        {
          description = dsp.GetDescription();
          if (dev.Length > 0)
          {
            if (!dev.Equals("keyword", StringComparison.CurrentCultureIgnoreCase) &&
                !dev.StartsWith("#"))
            {
              description += " of " + dev;
            }
          }
        }
        int atype = dsp.GetArrayType();
        if ((atype & AT_CHANNEL) == AT_CHANNEL)
        {
          isMCA = true;
          if (!displayAsMCA && AutoLinkCapacity == 0) AutoLinkCapacity = 1;
          if (acquireLabels)
          {
            if (addrParts != null && addrParts.Length > 2)
            { // if wildcard then append the wildcard to the server ...
              String tgt = (dev.Contains("*")) ? srv + "/" + dev : srv;
              labels = AcopControl.AcopLink.GetDeviceList(linkProt, ctx, linkSubsystem, tgt, property);
              if (labels != null)
              {
                int idx = 0;
                for (int i = 0; i < labels.Length; i++)
                {
                  if (labels[i].Equals(dev, StringComparison.CurrentCultureIgnoreCase))
                  {
                    idx = i;
                    break;
                  }
                }
                if (idx > 0)
                {
                  string[] labels_sub = new string[labels.Length - idx];
                  Array.Copy(labels, idx, labels_sub, 0, labels.Length - idx);
                  labels = labels_sub;
                }
              }
              acquiredLabels = labels;
            }
          }
        }
        if ((atype & AT_STRUCTURED) == AT_STRUCTURED)
        {
          if (dsp.GetCapacity() > 1) isStructured = true;
        }
        if (!dsp.SettingsStale())
        {
          displaySettingsAcquired = true;
          linkDspSettings = dsp;
        }
        if (dsp.GetYMaximum() == dsp.GetYMinimum())
        {
          dsp.SetYMaximum(dsp.GetYMinimum() + 100.0);
        }
      }
      String unitsString = ((units == null || units.Length == 0) ? "" : " (" + units + ") ");
      if (unitsString.Length > 0 && !description.Contains(unitsString) && !description.Contains(";"))
      {
        description += unitsString;
      }
      return dsp;
    }
    /// <summary>
    /// Gets the current transport link status as a string
    /// </summary>
    /// <returns>The current transport link status as a string</returns>
    public String GetStatusString()
    {
      AcopControl.AcopLink lnk = AcopControl.AcopLink.GetAcopLink(linkProt, linkAddr, linkProp, linkDataOut, linkDataIn, linkFlags, component);
      if (lnk != null) return lnk.GetStatusString();
      return "no link exists for this acop";
    }
    /// <summary>
    /// Gets the last delivered link output completion length.
    /// </summary>
    /// <returns>The last delivered link output completion length.</returns>
    public int GetOutputCompletionLength()
    {
      AcopControl.AcopLink lnk = AcopControl.AcopLink.GetAcopLink(linkProt, linkAddr, linkProp, linkDataOut, linkDataIn, linkFlags, component);
      if (lnk != null) return lnk.GetOutputCompletionLength();
      return 0;
    }
    /// <summary>
    /// Get the last link timestamp (as a UTC double).
    /// </summary>
    /// <returns>The last link timestamp (as a UTC double).</returns>
    public double GetTimeStamp()
    {
      AcopControl.AcopLink lnk = AcopControl.AcopLink.GetAcopLink(linkProt, linkAddr, linkProp, linkDataOut, linkDataIn, linkFlags, component);
      if (lnk != null) return lnk.GetTimeStamp();
      return 0;
    }
    /// <summary>
    /// Gets the last data tag (if any) associated with the output data.
    /// </summary>
    /// <returns>The last data tag (if any) associated with the output data.</returns>
    public int GetDataTag()
    {
      AcopControl.AcopLink lnk = AcopControl.AcopLink.GetAcopLink(linkProt, linkAddr, linkProp, linkDataOut, linkDataIn, linkFlags, component);
      if (lnk != null) return lnk.GetDataTag();
      return 0;
    }
    /// <summary>
    /// Gets the current link status as an integer value
    /// </summary>
    /// <returns>The current link status as an integer value</returns>
    public int GetStatus()
    {
      AcopControl.AcopLink lnk = AcopControl.AcopLink.GetAcopLink(linkProt, linkAddr, linkProp, linkDataOut, linkDataIn, linkFlags, component);
      if (lnk != null) return lnk.GetStatus();
      return -1;
    }
    protected void setAutoLinkCapacity(int value) { autoLinkCapacity = value; }
    protected int getAutoLinkCapacity() { return autoLinkCapacity; }
    protected void setLinkGrouped(bool value) { isGrouped = value; }
    protected bool getLinkGrouped() { return isGrouped; }

    protected void setAutoUpdate(bool value) { autoUpdate = value; }
    protected bool getAutoUpdate() { return autoUpdate; }

    protected void setAutoAttach(bool value) { autoAttach = value; }
    protected bool getAutoAttach() { return autoAttach; }

    protected String getUnits() { return units; }
    protected void setUnits(String value)
    {
      if (value != null) units = value;
      if (defaultUnitsLabel != null) defaultUnitsLabel.Text = units;
    }

    protected int getPrecision() { return precision; }
    protected void setPrecision(int value)
    {
      if (value >= 0) precision = value;
    }

    protected double GetReadbackValue()
    {
      return readbackValue;
    }
    protected void SetReadbackValue(double value)
    {
      readbackValue = value;
    }
    protected double GetResetValue()
    {
      return resetValue;
    }
    protected void SetResetValue(double value)
    {
      resetValue = value;
    }
    protected Boolean getValueIsInteger()
    {
      return valueIsInteger;
    }
    protected Boolean getValueIsNumber()
    {
      return valueIsNumber;
    }
    protected Boolean getValueAvailableAsNumber()
    {
      return valueAvailableAsNumber;
    }
    protected void setDisplayAsMCA(Boolean value) { displayAsMCA = value; }
    protected Boolean getDisplayAsMCA() { return displayAsMCA; }
    protected Boolean getIsMCA() { return isMCA; }
    protected String getMCADescription()
    {
      if (descriptionLabels == null) return null;
      if (!isMCA || !displayAsMCA) return null;
      StringBuilder sb = new StringBuilder();
      for (int i = 0; i < descriptionLabels.Length; i++) sb.AppendLine(descriptionLabels[i] + " ");
      return sb.ToString();
    }
    protected Boolean getIsStructured() { return isStructured; }
    private String getLinkInputValue() { return linkInputValue; }
    private void setLinkInputValue(String value)
    {
      if (value == null) value = "";
      linkInputValue = value;
    }
    private String getErrorValue() { return errorValue == null ? "" : errorValue; }
    private void setErrorValue(String value)
    {
      if (value == null) value = "";
      errorValue = value;
    }
    protected void applyErrorValue(AcopControl.AcopLink lnk)
    {
      if (errorValue.Length == 0) return;
      if (isGrouped)
      {
        AcopControl.AcopLink[] lst = AcopControl.AcopLink.GetAcopLinkTable(component);
        int gsize = lst.GetLength(0);
        foreach (AcopControl.AcopLink alnk in lst)
        {
          if (alnk.GetStatus() == 0) continue;
          alnk.FillDataObjectFromString(errorValue);
        }
      }
      else
      {
        if (lnk.GetStatus() == 0) return;
        lnk.FillDataObjectFromString(errorValue);
      }
    }
    private Boolean isLocked = false;
    protected void applyLock(CheckBox cb)
    {
      AcopControl.AcopLink lnk = AcopControl.AcopLink.GetAcopLink(linkProt, linkAddr, linkProp, component);
      if (lnk == null) return;
      if (cb.Checked && !isLocked)
      {
        if (lnk.LockAccess(LinkAddress) == 0) isLocked = true;
        else cb.Checked = false;
      }
      else
      {
        if (isLocked) lnk.UnlockAccess(LinkAddress);
        isLocked = false;
      }
    }
    protected String getLabelString(double v, int width)
    {
      if (valueIsInteger) return ((int)v).ToString("D");
      return v.ToString(getLabelFormat(width));
    }
    protected String getLabelString(double v, int width, Boolean general)
    {
      if (valueIsInteger) return ((int)v).ToString("D");
      if (precision == 0) general = true;
      String f = general ? "G" + width : "F" + precision;
      return v.ToString(f);
    }
    private String getLabelFormat(int width)
    {
      if (valueIsInteger)
      {
        return "D";
      }
      if (precision == 0)
      {
        return "G" + width;
      }
      return "F" + precision;
    }
    protected Type getValueType()
    {
      return valueType;
    }
    private Boolean isIntegerType(String value)
    {
      foreach (char c in value)
      {
        if (c == ' ' || c == '-' || c == '+') continue;
        if (c < '0' || c > '9') return false;
      }
      return true;
    }
    protected Type getValueType(String value)
    {
      if (isIntegerType(value)) return typeof(Int32);
      double v;
      if (Double.TryParse(value, out v)) return typeof(Double);
      return typeof(String);
    }
    protected Type getValueType(AcopControl.AcopLink lnk)
    {
      Object o = lnk.GetOutputDataObject();
      if (o == null) return null;
      return (o is Array) ? ((Array)o).GetType().GetElementType() : o.GetType();
    }
    protected Boolean ValueIsNumber()
    {
      if (valueType == null) return false;
      switch (Type.GetTypeCode(valueType))
      {
        case TypeCode.Double: return true;
        case TypeCode.Int16: return true;
        case TypeCode.Int32: return true;
        case TypeCode.Int64: return true;
        case TypeCode.Single: return true;
        case TypeCode.UInt16: return true;
        case TypeCode.UInt32: return true;
        case TypeCode.UInt64: return true;
      }
      return false;
    }
    protected double getValueAsDouble(AcopControl.AcopLink lnk)
    {
      double v = 0.0;
      try
      {
        Object o = lnk.GetOutputDataObject();
        valueType = (o is Array) ? ((Array)o).GetType().GetElementType() : o.GetType();
        if (o is Boolean[])
        {
          valueIsInteger = true;
          return ((Boolean[])o)[0] ? 1.0 : 0.0;
        }
        if (o is double[]) v = ((double[])o)[0];
        else if (o is float[]) v = (double)((float[])o)[0];
        else if (o is Int16[])
        {
          v = (double)((Int16[])o)[0];
          valueIsInteger = true;
        }
        else if (o is Int32[])
        {
          v = (double)((Int32[])o)[0];
          valueIsInteger = true;
        }
        else if (o is Int64[])
        {
          v = (double)((Int64[])o)[0];
          valueIsInteger = true;
        }
        else if (o is Double) v = (Double)o;
        else if (o is Single) v = (double)((Single)o);
        else if (o is Int16)
        {
          v = (double)((Int16)o);
          valueIsInteger = true;
        }
        else if (o is Int32)
        {
          v = (double)((Int32)o);
          valueIsInteger = true;
        }
        else if (o is Int64)
        {
          v = (double)((Int64)o);
          valueIsInteger = true;
        }
        else
        {
          v = lnk.GetOutputDataAsDouble()[0];
          valueAvailableAsNumber = true;
        }
        valueIsNumber = ValueIsNumber();
      }
      catch (Exception)
      {
        valueIsNumber = false;
      }
      if (valueIsNumber) valueAvailableAsNumber = true;
      return v;
    }
    protected Bitmap getDataAsImage(AcopControl.AcopLink lnk, ref String header)
    {
      Bitmap image = null;
      lnk.GetImage(ref image, ref header);
      return image;
    }
    protected Boolean TextMatches(String pattern, String txt)
    {
      if (txt == null || txt.Length == 0) return false;
      if (pattern == null || pattern.Length == 0) return true;
      if (pattern.StartsWith("-"))
      { // cut people a break with regular expressions ...
        // pattern = "^(?!.*XYZ).*$"; -> everything except XYZ matches
        pattern = "^(?!.*"+pattern.Substring(1)+ ").*$";
      }
      return Regex.Match(txt, pattern).Success;
    }
    protected TargetControlEditorValue targetControlEditorValue = null;
    protected TargetControlEditorValue getTargetControlEditorValue()
    {
      if (targetControlEditorValue == null)
        targetControlEditorValue = new TargetControlEditorValue("", this);
      return targetControlEditorValue;
    }
    protected void setTargetControlEditorValue(TargetControlEditorValue value)
    {
      if (value == null) value = new TargetControlEditorValue("", this);
      if (value != null) setTargetComponent(value.ToString());
      targetControlEditorValue = value;
    }
    private String targetComponent = null;
    public String getTargetComponent()
    {
      return targetComponent == null ? "" : targetComponent;
    }
    public void setTargetComponent(String value)
    {
      targetComponent = value;
      targetComponentProperty = null;
      targetComponentControl = null;
    }
    protected Boolean hasTargetComponent()
    {
      return (targetComponent != null && targetComponent.Length > 0);
    }
    private PropertyInfo targetComponentProperty = null;
    private Control targetComponentControl = null;
    protected Control getTargetComponentControl(String controlProperty)
    {
      if (controlProperty == null || controlProperty.Length == 0) return null;
      if (targetComponent.Equals(controlProperty, StringComparison.CurrentCultureIgnoreCase) &&
          targetComponentControl != null)
      {
        return targetComponentControl;
      }
      String[] parts = controlProperty.Split(new char[] { '.' }, StringSplitOptions.RemoveEmptyEntries);
      if (parts.Length < 2) return null;
      Control parent = this.Parent;
      foreach (var c in parent.Controls.OfType<Control>())
      {
        if (c.Name.Equals(parts[0], StringComparison.CurrentCultureIgnoreCase))
        { // found the desired control instance
          targetComponentControl = c;
          return c;
        }
      }
      return null;
    }
    protected PropertyInfo getTargetComponentProperty(String controlProperty)
    {
      Control c = getTargetComponentControl(controlProperty);
      if (c == null) return null;
      String[] parts = controlProperty.Split(new char[] { '.' }, StringSplitOptions.RemoveEmptyEntries);
      PropertyInfo[] prps = c.GetType().GetProperties();
      foreach (PropertyInfo p in prps)
      {
        if (p.Name.Equals(parts[1], StringComparison.CurrentCultureIgnoreCase))
        {
          targetComponentProperty = p;
          return p;
        }
      }
      return null;
    }
    protected void applyToControlProperty(String controlProperty, object value)
    {
      PropertyInfo p = getTargetComponentProperty(controlProperty);
      if (p == null) return;
      try
      {
        Control c = getTargetComponentControl(controlProperty);
        if (value is Array)
        {
          value = ((Array)value).GetValue(0);
        }
        Type vtype = value.GetType();
        if (p.PropertyType.IsPrimitive && vtype.IsPrimitive)
        {
          p.SetValue(c, value);
        }
        else if (p.PropertyType == vtype)
        {
          p.SetValue(c, value);
        }
        else
        {
          if (p.PropertyType == typeof(String))
          {
            p.SetValue(c, value.ToString());
          }
          else if (vtype == typeof(String))
          {
            if (p.PropertyType == typeof(Boolean))
            {
              Boolean bv = false;
              if (Boolean.TryParse((String)value, out bv))
              {
                p.SetValue(c, bv);
              }
            }
            else
            { // last chance ...
              double dv = 0;
              if (Double.TryParse((String)value, out dv))
              {
                p.SetValue(c, dv);
              }
            }
          }
        }
      }
      catch (Exception) {}
    }
    protected object getSettingFromControlProperty(String controlProperty)
    {
      PropertyInfo p = getTargetComponentProperty(controlProperty);
      if (p == null) return null;
      try
      {
        Control c = getTargetComponentControl(controlProperty);
        return p.GetValue(c);
      }
      catch (Exception) { }
      return null;
    }
    protected String getDescription() { return description == null ? "" : description; }
    private Boolean descriptionSetByDesigner = false;
    private void setDescription(String description)
    {
      this.description = description;
      if (description != null && description.Length > 0) descriptionSetByDesigner = true;
      if (defaultDescriptionLabel != null) defaultDescriptionLabel.Text = description;
    }

    protected String[] getDescriptionArray()
    {
      if (description == null) return new String[] { "" };
      return description.Split(new char[] { ';' }, StringSplitOptions.RemoveEmptyEntries);
    }
    /// <summary>
    /// Returns the current UTC time (seconds since 1970) as a double (with fractional seconds).
    /// </summary>
    /// <returns>the current UTC time (seconds since 1970) as a double (with fractional seconds)</returns>
    public static double getUtcNow()
    {
      return (DateTime.Now.ToUniversalTime() - utcThen).TotalSeconds;
    }
    private Boolean AreNotificationsPending()
    {
      if (isGrouped)
      {
        AcopControl.AcopLink[] lst = AcopControl.AcopLink.GetAcopLinkTable(component);
        int gsize = lst.GetLength(0);
        if (numAcopLinksPending == 0)
        { // 1st callback: set the notification parameters ...
          numAcopLinksPending = numAcopLinksInGroup = gsize;
        }
        if (gsize != numAcopLinksInGroup)
        { // something has changed ...
          numAcopLinksPending -= (numAcopLinksInGroup - gsize);
          numAcopLinksInGroup = gsize;
        }
        foreach (AcopControl.AcopLink lnk in lst)
          if (lnk.isCallbackPending()) return true;
        foreach (AcopControl.AcopLink lnk in lst) lnk.setCallbackPending(true);
        //if (--numAcopLinksPending > 0) return true;
        //numAcopLinksPending = 0;
      }
      return false;
    }
    protected void SetDefaultUpdater(defaultUpdaterType updater)
    {
      this.updater = updater;
    }
    protected void SetDefaultApplier(defaultApplierType applier)
    {
      this.applier = applier;
    }
    protected void notifyLinkUpdate(object sender, AcopControl.AcopLink lnk)
    {
      foreach (Control c in getControlsList())
        if (c.InvokeRequired)
        {
          try
          {
            Invoke(new notifyLinkUpdateDelegate(notifyLinkUpdate), sender, lnk);
          }
          catch (Exception) { }
          return;
        }
      lock (notifyMutex)
      {
        try
        {
          if (insideNotify) return;
          insideNotify = true;
          applyErrorValue(lnk);
          if (AreNotificationsPending())
          {
            insideNotify = false;
            return;
          }
          if (acopLinkUpdate != null)
          { // there is an event delegate : fire it
            acopLinkUpdate(this, Link.GetLink(lnk));
          }
          // check for automatic display updates:
          if (updater != null) updater(lnk);
          insideNotify = false;
        }
        catch (Exception)
        {
          insideNotify = false;
        }
      }
    }
    public enum ApplyEventSettingType
    {
      exclusive = 0,
      before = 1,
      after = 2
    }
    private ApplyEventSettingType applySettingType = ApplyEventSettingType.exclusive;
    private int applyAcopSetting(object setting, AcopControl.AcopLink lnk)
    {
      if (targetComponent != null && targetComponent.Length > 0)
      {
        applyToControlProperty(targetComponent, setting);
        return 0;
      }
      lock (notifyMutex)
      {
        int flgs = GetFlags("WRITE_EXCLUSIVE");
        int cc = lnk.Execute(null, setting, flgs);
        if (applier != null) applier(lnk);
        return cc;
      }
    }
    protected int applySetting(object sender, object setting, AcopControl.AcopLink lnk)
    {
      switch (applySettingType)
      {
        case ApplyEventSettingType.exclusive:
          if (acopApply != null)
          {
            acopApply(sender, Link.GetLink(lnk));
            return 0;
          }
          return applyAcopSetting(setting, lnk);
        case ApplyEventSettingType.before:
          if (acopApply != null) acopApply(sender, Link.GetLink(lnk));
          return applyAcopSetting(setting, lnk);
        case ApplyEventSettingType.after:
          int cc = applyAcopSetting(setting, lnk);
          if (acopApply != null) acopApply(sender, Link.GetLink(lnk));
          return cc;
        default:
          return -1;
      }
    }
    public class LinkGroupListItem
    {
      AcopUserComponent component = null;
      String protocol = null;
      String address = null;
      String property = null;
      Boolean isWellFormed = false;
      Boolean isAttached = false;
      public String GetProtocol() { return protocol; }
      public String GetAddress() { return address; }
      public String GetProperty() { return property; }
      public Boolean IsWellFormed() { return isWellFormed; }
      public Boolean IsAttached() { return isAttached; }
      public void IsAttached(Boolean value)
      {
        isAttached = value;
      }
      public LinkGroupListItem(AcopUserComponent component, String protocol, String address, String property)
      {
        this.component = component;
        this.protocol = protocol;
        this.address = address;
        this.property = property;
        if (protocol != null && address != null && property != null)
          isWellFormed = true;
      }
      public LinkGroupListItem(AcopUserComponent component, String fullAddress)
      {
        if (component == null || fullAddress == null) return;
        this.component = component;
        String[] parts = fullAddress.Split(new char[] { '/' }, StringSplitOptions.RemoveEmptyEntries);
        int adrIdx;
        if (fullAddress.StartsWith("/"))
        { // no protocol given in string: take the component's assigned value
          protocol = component.LinkProtocol;
          if (parts.Length < 2) return;
          adrIdx = 0;
        }
        else
        {
          if (parts.Length < 3) return;
          protocol = parts[0];
          adrIdx = 1;
        }
        property = parts[parts.Length - 1];
        StringBuilder sb = new StringBuilder("/"+parts[adrIdx]);
        for (int i = adrIdx + 1; i < parts.Length - 1; i++) sb.Append("/" + parts[i]);
        address = sb.ToString();
        isWellFormed = true;
      }
    }

    #region Converter
    public class LinkProtocolConverter : StringConverter
    {
      public override bool GetStandardValuesSupported(ITypeDescriptorContext context)
      { //true means show a combobox
        return true;
      }
      public override bool GetStandardValuesExclusive(ITypeDescriptorContext context)
      { //true will limit to list. false will show the list, but allow free-form entry
        return false;
      }
      public override
      System.ComponentModel.TypeConverter.StandardValuesCollection GetStandardValues(ITypeDescriptorContext context)
      {
        return new StandardValuesCollection(AcopControl.AcopLink.GetProtocolList());
      }
    }
    public class LinkContextConverter : StringConverter
    {
      private AcopUserComponent myAcop = GetInstance();
      public override bool GetStandardValuesSupported(ITypeDescriptorContext context) { return true; }
      public override bool GetStandardValuesExclusive(ITypeDescriptorContext context) { return true; }
      public override
      System.ComponentModel.TypeConverter.StandardValuesCollection GetStandardValues(ITypeDescriptorContext context)
      {
        return new StandardValuesCollection(AcopControl.AcopLink.GetContextList(myAcop.linkProt));
      }
    }
    public class LinkSubSystemConverter : StringConverter
    {
      private AcopUserComponent myAcop = null;
      public override bool GetStandardValuesSupported(ITypeDescriptorContext context) { return true; }
      public override bool GetStandardValuesExclusive(ITypeDescriptorContext context) { return true; }
      public override
      System.ComponentModel.TypeConverter.StandardValuesCollection GetStandardValues(ITypeDescriptorContext context)
      {
        myAcop = GetInstance();
        return new StandardValuesCollection(AcopControl.AcopLink.GetSubSystemList(myAcop.linkProt, myAcop.getLinkContext()));
      }
    }
    public class LinkServerConverter : StringConverter
    {
      private AcopUserComponent myAcop = null;
      public override bool GetStandardValuesSupported(ITypeDescriptorContext context) { return true; }
      public override bool GetStandardValuesExclusive(ITypeDescriptorContext context) { return true; }
      public override
      System.ComponentModel.TypeConverter.StandardValuesCollection GetStandardValues(ITypeDescriptorContext context)
      {
        myAcop = GetInstance();
        return new StandardValuesCollection(AcopControl.AcopLink.GetServerList(myAcop.linkProt, myAcop.getLinkContext(), myAcop.getLinkSubSystem()));
      }
    }
    public class LinkDeviceConverter : StringConverter
    {
      private AcopUserComponent myAcop = null;
      public override bool GetStandardValuesSupported(ITypeDescriptorContext context) { return true; }
      public override bool GetStandardValuesExclusive(ITypeDescriptorContext context) { return false; }
      public override
      System.ComponentModel.TypeConverter.StandardValuesCollection GetStandardValues(ITypeDescriptorContext context)
      {
        myAcop = GetInstance();
        myAcop.hasPropertyQueryPrecedence = AcopControl.AcopLink.HasPropertyQueryPrecedence(myAcop.linkProt, myAcop.getLinkContext(), myAcop.getLinkSubSystem(), myAcop.getLinkServer());
        String prp = myAcop.hasPropertyQueryPrecedence ? myAcop.linkProp : null;
        return new StandardValuesCollection(AcopControl.AcopLink.GetDeviceList(myAcop.linkProt, myAcop.getLinkContext(), myAcop.getLinkSubSystem(), myAcop.getLinkServer(), prp));
      }
    }
    public class LinkPropertyConverter : StringConverter
    {
      private AcopUserComponent myAcop = null;
      public override bool GetStandardValuesSupported(ITypeDescriptorContext context) { return true; }
      public override bool GetStandardValuesExclusive(ITypeDescriptorContext context) { return false; }
      public override
      System.ComponentModel.TypeConverter.StandardValuesCollection GetStandardValues(ITypeDescriptorContext context)
      {
        myAcop = GetInstance();
        myAcop.hasDeviceQueryPrecedence = AcopControl.AcopLink.HasDeviceQueryPrecedence(myAcop.linkProt, myAcop.getLinkContext(), myAcop.getLinkSubSystem(), myAcop.getLinkServer());
        return new StandardValuesCollection(AcopControl.AcopLink.GetPropertyList(myAcop.linkProt, myAcop.getLinkContext(), myAcop.getLinkSubSystem(), myAcop.getLinkServer(), myAcop.linkDev));
      }
    }
    public class LinkFlagsConverter : StringConverter
    {
      private AcopUserComponent myAcop = null;
      public override bool GetStandardValuesSupported(ITypeDescriptorContext context) { return true; }
      public override bool GetStandardValuesExclusive(ITypeDescriptorContext context) { return false; }
      public override
      System.ComponentModel.TypeConverter.StandardValuesCollection GetStandardValues(ITypeDescriptorContext context)
      {
        myAcop = GetInstance();
        myAcop.flagsDict = AcopControl.AcopLink.GetFlagsList(myAcop.linkProt);
        String[] sv = myAcop.flagsDict.Keys.ToArray();
        return new StandardValuesCollection(sv);
      }
    }
    #endregion

    #region Properties
    [Category("Acop.Transport")]
    [Description("Occurs when the underlying monitored link has updated its data")]
    public event acopUpdate acopLinkUpdate;

    [Category("Acop.Transport")]
    [Description("Occurs when the control attempts to apply the setting shown to the link address")]
    public event acopApplySetting acopApply;
    /// <summary>
    /// Determines the firing order of a user-defined acopApply delegate
    /// </summary>
    [Category("Acop.Transport")]
    [Description("Determines the firing order of a user-defined acopApply delegate")]
    [DefaultValue(ApplyEventSettingType.exclusive)]
    public ApplyEventSettingType ApplyEventOrder { get { return applySettingType; } set { applySettingType = value; } }

    /// <summary>
    /// The link control system protocol
    /// </summary>
    [Category("Acop.Transport")]
    [Description("The link control system protocol")]
    [Browsable(true)]
    [DefaultValue("TINE")]
    [TypeConverter(typeof(LinkProtocolConverter))]
    public String LinkProtocol { get { return getLinkProt(); } set { setLinkProt(value); } }

    /// <summary>
    /// The design-time browseable link context (part of the link address)
    /// </summary>
    [Category("Acop.Transport")]
    [Description("The design-time browseable link context (part of the link address)")]
    [Browsable(true)]
    [DefaultValue("SITE")]
    [TypeConverter(typeof(LinkContextConverter))]
    public String LinkContext { get { return getLinkContext(); } set { setLinkContext(value); } }

    /// <summary>
    /// The design-time browseable link subsystem
    /// </summary>
    [Category("Acop.Transport")]
    [Description("The design-time browseable link subsystem")]
    [Browsable(true)]
    [DefaultValue("ALL")]
    [TypeConverter(typeof(LinkSubSystemConverter))]
    public String LinkSubSystem { get { return getLinkSubSystem(); } set { setLinkSubSystem(value); } }

    /// <summary>
    /// The design-time browseable link server (part of the link address)
    /// </summary>
    [Category("Acop.Transport")]
    [Description("The design-time browseable link server (part of the link address)")]
    [Browsable(true)]
    [DefaultValue("ENS")]
    [TypeConverter(typeof(LinkServerConverter))]
    public String LinkServer { get { return getLinkServer(); } set { setLinkServer(value); } }

    /// <summary>
    /// The design-time browseable/editable link device (part of the link address)
    /// </summary>
    [Category("Acop.Transport")]
    [Description("The design-time browseable/editable link device (part of the link address)")]
    [Browsable(true)]
    [DefaultValue("DEFAULT")]
    [TypeConverter(typeof(LinkDeviceConverter))]
    public String LinkDevice { get { return getLinkDevice(); } set { setLinkDevice(value); } }

    /// <summary>
    /// Identifies the transport link target. The target end point is given by the LinkAddress and the LinkProperty.
    /// </summary>
    [Category("Acop.Transport")]
    [Description("Identifies the transport link target. The target end point is given by the LinkAddress and the LinkProperty.")]
    public String LinkAddress { get { return getLinkAddr(); } set { setLinkAddr(value); } }

    /// <summary>
    /// The design-time browseable/editable link property (suppliments the link address)"
    /// </summary>
    [Category("Acop.Transport")]
    [Description("The design-time browseable/editable link property (suppliments the link address)")]
    [Browsable(true)]
    [DefaultValue("NSERVERS")]
    [TypeConverter(typeof(LinkPropertyConverter))]
    public String LinkProperty { get { return getLinkProp(); } set { setLinkProp(value); } }

    /// <summary>
    /// A design-time list of targeted links.  If there is more than one target then 'LinkGrouped' is automatically set to 'true'
    /// </summary>
    [Category("Acop.Transport")]
    [Description("A design-time list of targeted links.  If there is more than one target then 'LinkGrouped' is automatically set to 'true'")]
    public String LinkList { get { return getLinkGroupList(); } set { setLinkGroupList(value); } }

    /// <summary>
    /// An optional design-time browseable flags set which can be applied to the underlying link
    /// </summary>
    [Category("Acop.Transport")]
    [Description("An optional design-time browseable flags set which can be applied to the underlying link")]
    [Browsable(true)]
    [DefaultValue("NONE")]
    [TypeConverter(typeof(LinkFlagsConverter))]
    public String LinkFlags { get { return getEnumLinkFlags(); } set { setEnumLinkFlags(value); } }

    /// <summary>
    /// if set to true, instructs ACOP to update this Label's contents upon receipt of new link data (provided no user-defined event handle is in play)
    /// </summary>
    [Category("Acop.Transport")]
    [Description("if set to true, instructs ACOP to update this Label's contents upon receipt of new link data (provided no user-defined event handle is in play)")]
    [DefaultValue(false)]
    public bool AutoLinkUpdate { get { return getAutoUpdate(); } set { setAutoUpdate(value); } }

    /// <summary>
    /// if set to true, instructs ACOP to automatically monitor the data provided by LinkAddress and LinkProperty
    /// </summary>
    [Category("Acop.Transport")]
    [Description("if set to true, instructs ACOP to automatically monitor the data provided by LinkAddress and LinkProperty")]
    [DefaultValue(false)]
    public bool AutoLinkAttach { get { return getAutoAttach(); } set { setAutoAttach(value); } }

    /// <summary>
    /// The default monitoring interval in msec
    /// </summary>
    [Category("Acop.Transport")]
    [DefaultValue(1000)]
    [Description("The default monitoring interval in msec")]
    public int DefaultMonitorInterval { get { return getDefaultInterval(); } set { setDefaultInterval(value); } }

    /// <summary>
    /// if non-zero, provides the requested data size to use when AutoLinkUpdate is set to true
    /// </summary>
    [Category("Acop.Transport")]
    [Description("if non-zero, provides the requested data size to use when AutoLinkUpdate is set to true")]
    [DefaultValue(0)]
    public int AutoLinkCapacity { get { return getAutoLinkCapacity(); } set { setAutoLinkCapacity(value); } }

    /// <summary>
    /// if non-empty, then the string provided will be used as the Caption text (else the property's discovered description)
    /// </summary>
    [Category("Acop.Display")]
    [Description("if non-empty, then the string provided will be used as the Caption text (else the property's discovered description)")]
    public String Description { get { return getDescription(); } set { setDescription(value); } }

    /// <summary>
    /// the value units (applied to the slider desciption)
    /// </summary>
    [Category("Acop.Display")]
    [Description("the value units (applied to the slider desciption)")]
    public String Units { get { return units != null ? units : ""; } set { setUnits(value); } }

    /// <summary>
    /// precision to use for floating point numbers (0 => general format)
    /// </summary>
    [Category("Acop.Display")]
    [Description("precision to use for floating point numbers (0 => general format)")]
    [DefaultValue(0)]
    public int Precision { get { return getPrecision(); } set { setPrecision(value); } }

    /// <summary>
    /// gets/sets whether all associated links have a grouped event
    /// </summary>
    [Category("Acop.Transport")]
    [Description("gets/sets whether all associated links have a grouped event")]
    [DefaultValue(false)]
    public Boolean LinkGrouped { get { return getLinkGrouped(); } set { setLinkGrouped(value); } }

    /// <summary>
    /// optional value to use as input for automatic links
    /// </summary>
    [Category("Acop.Transport")]
    [Description("optional value to use as input for automatic links")]
    public String LinkInputValue { get { return getLinkInputValue(); } set { setLinkInputValue(value); } }

    private AddressEditor addressEditorValue = new AddressEditor();
    [Category("Acop.Transport")]
    [Description("Address Editor"), Editor(typeof(LinkAddressEditor), typeof(UITypeEditor))]
    public AddressEditor AddressEditor { get { return this.addressEditorValue; } set { this.addressEditorValue = value; } }

    /// <summary>
    /// value to use when there are link errors
    /// </summary>
    [Category("Acop.Transport")]
    [Description("value to use when there are link errors")]
    public String ErrorValue { get { return getErrorValue(); } set { setErrorValue(value); } }

    /// <summary>
    /// if true, then the label contents will be a multi-line representation of all values if the targeted data is a multi-channel array
    /// </summary>
    [Category("Acop.Display")]
    [Description("if true, then the label contents will be a multi-line representation of all values if the targeted data is a multi-channel array")]
    [DefaultValue(false)]
    public Boolean DisplayAsMultiChannel { get { return getDisplayAsMCA(); } set { setDisplayAsMCA(value); } }
    #endregion

    private void initialize()
    {
      this.SuspendLayout();
      // 
      // AcopUserComponent
      // 
      //this.Name = "AcopUserComponent";
      if (components == null)
        components = new System.ComponentModel.Container();
      if (!isDesignMode)
      {
        lateInitializer = new System.Windows.Forms.Timer(components);
        lateInitializer.Tick += new System.EventHandler(lateInitializerUpdate);
        lateInitializer.Interval = 50;
      }
      this.ResumeLayout(false);
    }
  }
  /// <summary>
  /// Protocol independent flags which can be applied to the Attach and Execute
  /// methods when coding.  Otherwise reference AcopTransport and use protocol 
  /// specific flags.
  /// </summary>
  public enum StandardLinkFlags
  {
    NONE = 0,
    READ = 1,
    WRITE = 2
  }
  public class Link
  {
    static Dictionary<AcopControl.AcopLink, Link> links = new Dictionary<AcopControl.AcopLink, Link>();
    static Dictionary<Component, Link[]> tbl = new Dictionary<Component, Link[]>();
    public static Link GetLink(AcopControl.AcopLink lnk)
    {
      if (lnk == null) return null;
      lnk.GetAcopControl();
      Link val;
      if (links.TryGetValue(lnk, out val)) return val;
      return new Link(lnk);
    }

    AcopControl.AcopLink parent = null;
    public Link(AcopControl.AcopLink lnk)
    {
      if (lnk == null) return;
      parent = lnk;
      if (!links.ContainsKey(lnk)) links.Add(lnk, this);
    }
    /// <summary>
    /// gets the completion length of the last data transfer
    /// </summary>
    /// <returns>the completion length of the last data transfer</returns>
    public int GetOutputCompletionLength()
    {
      if (parent == null) return 0;
      return parent.GetOutputCompletionLength();
    }
    /// <summary>
    /// gets the timestamp (UTC) of the last data transfer
    /// </summary>
    /// <returns>the timestamp (UTC) of the last data transfer</returns>
    public double GetTimeStamp()
    {
      if (parent == null) return 0;
      return parent.GetTimeStamp();
    }
    /// <summary>
    /// gets the assigned integer data tag of the last data transfer
    /// </summary>
    /// <returns>the assigned integer data tag of the last data transfer</returns>
    public int GetDataTag()
    {
      if (parent == null) return 0;
      return parent.GetDataTag();
    }
    /// <summary>
    /// gets an array of numbers appropriate for an X (horizontal) axis in a display
    /// </summary>
    /// <returns>an array of numbers appropriate for an X (horizontal) axis in a display</returns>
    public double[] GetOutputXAxis()
    {
      if (parent == null) return null;
      return parent.GetOutputXAxis();
    }
    /// <summary>
    /// gets an array of numbers appropriate for an Y (vertical) axis in a display
    /// </summary>
    /// <returns>an array of numbers appropriate for an Y (vertical) axis in a display</returns>
    public double[] GetOutputYAxis()
    {
      if (parent == null) return null;
      return parent.GetOutputYAxis();
    }
    /// <summary>
    /// get the last transfered data as a string array
    /// </summary>
    /// <returns>the last transfered data as a string array</returns>
    public string[] GetOutputDataAsStringArray()
    {
      if (parent == null) return null;
      return parent.GetOutputDataAsString();
    }
    /// <summary>
    /// get the last transfered data as a string 
    /// </summary>
    /// <returns>the last transfered data as a string</returns>
    public string GetOutputDataAsString()
    {
      if (parent == null) return null;
      string[] sa = parent.GetOutputDataAsString();
      return sa == null ? null : sa[0];
    }
    /// <summary>
    /// get the last transfered data as a number array (doubles)
    /// </summary>
    /// <returns>the last transfered data as a number array</returns>
    public double[] GetOutputDataAsNumberArray()
    {
      if (parent == null) return null;
      return parent.GetOutputDataAsDouble();
    }
    /// <summary>
    /// get the last transfered data as a number
    /// </summary>
    /// <returns>the last transfered data as a number</returns>
    public double GetOutputDataAsNumber()
    {
      if (parent == null) return 0;
      double[] da = parent.GetOutputDataAsDouble();
      return da == null ? 0 : da[0];
    }
    /// <summary>
    /// get the last transfered data as an integer array
    /// </summary>
    /// <returns>the last transfered data as an integer array</returns>
    public Int32[] GetOutputDataAsIntegerArray()
    {
      if (parent == null) return null;
      return parent.GetOutputDataAsInteger();
    }
    /// <summary>
    /// get the last transfered data as an integer
    /// </summary>
    /// <returns>the last transfered data as an integer</returns>
    public Int32 GetOutputDataAsInteger()
    {
      if (parent == null) return 0;
      Int32[] ia = parent.GetOutputDataAsInteger();
      return ia == null ? 0 : ia[0];
    }
    /// <summary>
    /// get the last transfered data
    /// </summary>
    /// <returns>the last transfered data</returns>
    public object GetOutputDataObject()
    {
      if (parent == null) return null;
      return parent.GetOutputDataObject();
    }
    /// <summary>
    /// gets channel/array element names appropriate to the last transfered data
    /// </summary>
    /// <returns>channel/array element names appropriate to the last transfered data</returns>
    public string[] GetOutputChannelNames()
    {
      if (parent == null) return null;
      return parent.GetOutputChannelNames();
    }
    /// <summary>
    /// gets any link flags assigned to the last transfered data
    /// </summary>
    /// <returns>any link flags assigned to the last transfered data</returns>
    public int GetLinkFlags()
    {
      if (parent == null) return 0;
      return parent.GetLinkFlags();
    }
    /// <summary>
    /// gets the status code for the last data transfer
    /// </summary>
    /// <returns>the status code for the last data transfer</returns>
    public int GetStatus()
    {
      if (parent == null) return -1;
      return parent.GetStatus();
    }
    /// <summary>
    /// gets the status string for the last data transfer
    /// </summary>
    /// <returns>the status string for the last data transfer</returns>
    public string GetStatusString()
    {
      if (parent == null) return "no link exists for this acop transport";
      return parent.GetStatusString();
    }
    /// <summary>
    /// gets the assigned callback reference object 
    /// </summary>
    /// <returns>the assigned callback reference object</returns>
    public object GetCallbackReference()
    {
      if (parent == null) return null;
      return parent.GetCallbackReference();
    }
    private AcopControl.AcopLink[] lst = null;
    /// <summary>
    /// gets a link table containing any link group members 
    /// </summary>
    /// <returns>a link table containing any link group members</returns>
    public Link[] GetTransportLinkTable()
    {
      Component acop = parent.GetAcopControl();
      AcopControl.AcopLink[] lnks = AcopControl.AcopLink.GetAcopLinkTable(acop);
      Link[] tlnks = null;
      if (lnks == lst)
      {
        if (tbl.TryGetValue(acop, out tlnks)) return tlnks;
      }
      if (lnks == null || lnks.Length == 0) return null;
      tlnks = new Link[lnks.Length];
      for (int i = 0; i < lnks.Length; i++) tlnks[i] = new Link(lnks[i]);
      if (tbl.ContainsKey(acop)) tbl[acop] = tlnks; else tbl.Add(acop, tlnks);
      return tlnks;
    }
    /// <summary>
    /// gets an array of integer values (one per groupped link) in the last data transfer
    /// </summary>
    /// <returns>an array of integer values (one per groupped link) in the last data transfer</returns>
    public Int32[] GetTransportLinkTableDataAsIntegerArray()
    {
      Link[] lnks = GetTransportLinkTable();
      if (lnks == null) return null;
      Int32[] vals = new Int32[lnks.Length];
      for (int i = 0; i < lnks.Length; i++)
        vals[i] = lnks[i].GetOutputDataAsInteger();
      return vals;
    }
    /// <summary>
    /// gets an array of numbers (one per groupped link) in the last data transfer
    /// </summary>
    /// <returns>an array of numbers (one per groupped link) in the last data transfer</returns>
    public double[] GetTransportLinkTableDataAsNumberArray()
    {
      Link[] lnks = GetTransportLinkTable();
      if (lnks == null) return null;
      double[] vals = new double[lnks.Length];
      for (int i = 0; i < lnks.Length; i++)
        vals[i] = lnks[i].GetOutputDataAsNumber();
      return vals;
    }
    /// <summary>
    /// gets an array of string values (one per groupped link) in the last data transfer
    /// </summary>
    /// <returns>an array of string values (one per groupped link) in the last data transfer</returns>
    public String[] GetTransportLinkTableDataAsStringArray()
    {
      Link[] lnks = GetTransportLinkTable();
      if (lnks == null) return null;
      String[] vals = new String[lnks.Length];
      for (int i = 0; i < lnks.Length; i++)
        vals[i] = lnks[i].GetOutputDataAsString();
      return vals;
    }
    /// <summary>
    /// gets the assigned ACOP display id for this data link
    /// </summary>
    /// <returns>the assigned ACOP display id for this data link (-1 if not assigned)</returns>
    public int GetDisplayId()
    {
      if (parent == null) return -1;
      return parent.GetDisplayId();
    }
    /// <summary>
    /// gets the assigned transport link table id for this data link
    /// </summary>
    /// <returns>the assigned transport link table id for this data link (-1 if not assigned)</returns>
    public int GetLinkId()
    {
      if (parent == null) return -1;
      return parent.GetLinkId();
    }
  }
  public class AddressEditor
  {
  }
  public class LinkAddressEditor : UITypeEditor
  {
#   pragma warning disable 1591
    private Boolean valuesStale = false;
    public void SetValuesStale(Boolean value) { valuesStale = value; }
    public LinkAddressEditor() { }
    public override UITypeEditorEditStyle GetEditStyle(System.ComponentModel.ITypeDescriptorContext context)
    {
      return UITypeEditorEditStyle.Modal;
    }
    public override object EditValue(System.ComponentModel.ITypeDescriptorContext context, IServiceProvider provider, object value)
    {
      IWindowsFormsEditorService frmsvr = (IWindowsFormsEditorService)provider.GetService(typeof(IWindowsFormsEditorService));
      if (frmsvr == null) return null;
      LinkProperties propPage = new LinkProperties(context.Instance);
      propPage.SetControl((AcopUserComponent)context.Instance);
      propPage.SetEditor(this);
      frmsvr.ShowDialog(propPage);
      if (valuesStale) value = new AddressEditor();
      return base.EditValue(context, provider, value);
    }

  }
  public class LinkProperties : Form
  {
#   pragma warning disable 1591
    private LinkAddressEditor editor = null;
    private System.ComponentModel.IContainer components = null;
    public AcopUserComponent control = null;
    private String protocol;
    private String context;
    private String server;
    private String device;
    private String property;
    private String subsystem;
    private String linklist;
    private Boolean hasPropertyQueryPrecedence;
    private Boolean hasDeviceQueryPrecedence;
    Boolean hasInitialized = false;
    private Boolean hasGroup = true;
    private Boolean hasTrendInfo = false;
    private Boolean hasAutoScale = false;
    private Boolean hasErrorValue = true;
    public void SetEditor(LinkAddressEditor editor)
    {
      this.editor = editor;
    }
    public void SetControl(AcopUserComponent control)
    {
      this.control = control;
      //TODO: fill in fields from current settings ...

      protocol = control.LinkProtocol;
      context = control.LinkContext;
      subsystem = control.LinkSubSystem;
      server = control.LinkServer;
      device = control.LinkDevice;
      property = control.LinkProperty;

      String[] items = AcopControl.AcopLink.GetProtocolList();
      if (items != null)
        foreach (String item in items) this.ProtocolCombo.Items.Add(item);
      this.ProtocolCombo.Text = protocol;
      if (this.control.BrowsingServiceAvailable)
        items = AcopControl.AcopLink.GetContextList(protocol);
      else
        items = new String[] { AcopUserComponent.UNAVAILABLE };
      if (items != null)
      {
        foreach (String item in items) this.ContextCombo.Items.Add(item);
      }
      if (items == null || items[0].Equals(AcopUserComponent.UNAVAILABLE,StringComparison.CurrentCultureIgnoreCase))
      {
        this.control.BrowsingServiceAvailable = false;
      }
      this.ContextCombo.Text = context;
      if (this.control.BrowsingServiceAvailable)
        items = AcopControl.AcopLink.GetServerList(protocol, context, subsystem);
      else
        items = new String[] { AcopUserComponent.UNAVAILABLE };
      if (items != null)
      {
        foreach (String item in items) this.ServerCombo.Items.Add(item);
        if (this.control.BrowsingServiceAvailable)
        {
          hasDeviceQueryPrecedence = AcopControl.AcopLink.HasDeviceQueryPrecedence(protocol, context, subsystem, server);
          hasPropertyQueryPrecedence = AcopControl.AcopLink.HasPropertyQueryPrecedence(protocol, context, subsystem, server, property);
        }
      }
      this.ServerCombo.Text = server;

      if (this.control.BrowsingServiceAvailable)
        items = AcopControl.AcopLink.GetDeviceList(protocol, context, subsystem, server, property);
      else
        items = new String[] { AcopUserComponent.UNAVAILABLE };
      if (items != null) foreach (String item in items) this.DeviceCombo.Items.Add(item);
      this.DeviceCombo.Text = device;

      if (this.control.BrowsingServiceAvailable)
        items = AcopControl.AcopLink.GetPropertyList(protocol, context, subsystem, server, device);
      else
        items = new String[] { AcopUserComponent.UNAVAILABLE };
      if (items != null)
      {
        foreach (String item in items) this.PropertyCombo.Items.Add(item);
        String addr = "/" + context + "/" + server + "/" + device;
        if (this.control.BrowsingServiceAvailable)
        {
          dspSettings = AcopControl.AcopLink.GetDisplaySettings(addr, property);
          SetInfoLabelTT(dspSettings);
          MCALabel.BackColor = isMca() ? Color.Lime : default(Color);
        }
      }
      this.PropertyCombo.Text = property;

      if (this.control.BrowsingServiceAvailable)
        items = AcopControl.AcopLink.GetSubSystemList(protocol, context);
      else
        items = new String[] { AcopUserComponent.UNAVAILABLE };
      if (items != null) foreach (String item in items)
      {
        if (item.Length == 0) continue;
        this.SubsystemCombo.Items.Add(item);
      }
      this.SubsystemCombo.Text = subsystem;

      this.InputTextBox.Text = control.LinkInputValue;

      IDictionary<String, int> flagsDict = AcopControl.AcopLink.GetFlagsList(protocol);
      items = flagsDict.Keys.ToArray();
      if (items != null)
      {
        foreach (String item in items) this.FlagsCombo.Items.Add(item);
      }
      this.FlagsCombo.Text = control.LinkFlags;
      hasInitialized = true;

      linklist = control.LinkList;
      if (linklist != null && linklist.Length > 0)
      {
        items = control.LinkList.Split(new char[] { ';' }, StringSplitOptions.RemoveEmptyEntries);
        foreach (String item in items) AddressList.Items.Add(item);
      }

      GroupLinksCheckBox.Checked = this.control.LinkGrouped;
      AutoUpdateCheck.Checked = this.control.AutoLinkUpdate;
      AutoAttachCheck.Checked = this.control.AutoLinkAttach;
      IntervalBox.Text = this.control.DefaultMonitorInterval.ToString();
      Boolean useLinkCapacity = true;
      Boolean showTrendDepth = false;
      if (this.control is AcopChart)
      {
        AcopChart ac = (AcopChart)this.control;
        AcopChart.AutoScaleMode asm = ac.AutoScaleUpdate;
        this.AutoScaleCombo.SelectedIndex = (int)asm;
        this.AutoTrendCheck.Checked = ac.AutoTrendUpdate;
        if (ac.AutoTrendUpdate)
        {
          LinkCapacityBox.Text = ac.LinkTrendDepth.ToString();
          useLinkCapacity = false;
          showTrendDepth = true;
          TrendDepthTextBox.Text = ac.LinkTrendHours.ToString();
        }
      }
      TrendDepthTextBox.Visible = showTrendDepth;
      TrendDepthLabel.Visible = showTrendDepth;
      if (useLinkCapacity)
      {
        LinkCapacityBox.Text = this.control.AutoLinkCapacity.ToString();
      }
      if (hasErrorValue)
      {
        ErrorValueBox.Text = this.control.ErrorValue;
      }
    }
    protected override void Dispose(bool disposing)
    {
      if (disposing && (components != null))
      {
        components.Dispose();
      }
      base.Dispose(disposing);
    }
    public LinkProperties(object comp)
    {
      if (comp.GetType() == typeof(AcopWheelSwitch)) hasGroup = false;
      else if (comp.GetType() == typeof(AcopBitRegister)) hasGroup = false;
      else if (comp.GetType() == typeof(AcopSlider)) hasGroup = false;
      else if (comp.GetType() == typeof(AcopGauge)) hasGroup = false;
      else if (comp.GetType() == typeof(AcopButton))
      {
        hasGroup = false;
        hasErrorValue = false;
      }
      else if (comp.GetType() == typeof(AcopInputBox)) hasGroup = false;
      else if (comp.GetType() == typeof(AcopCheckBox))
      {
        hasGroup = false;
        hasErrorValue = false;
      }
      else if (comp.GetType() == typeof(AcopComboBox))
      {
        hasGroup = false;
        hasErrorValue = false;
      }
      else if (comp.GetType() == typeof(AcopChart))
      {
        hasTrendInfo = true;
        hasAutoScale = true;
      }
      else if (comp.GetType() == typeof(AcopRectangle))
      {
        hasGroup = false;
      }
      else if (comp.GetType() == typeof(AcopOval))
      {
        hasGroup = false;
      }
      else if (comp.GetType() == typeof(AcopTriangle))
      {
        hasGroup = false;
      }
      else if (comp.GetType() == typeof(AcopVideo))
      {
        hasGroup = false;
        hasErrorValue = false;
      }
      InitializePropertyPage();
    }

    private void InitializePropertyPage()
    {
      this.components = new System.ComponentModel.Container();
      this.ContextCombo = new System.Windows.Forms.ComboBox();
      this.ContextLabel = new System.Windows.Forms.Label();
      this.ServerCombo = new System.Windows.Forms.ComboBox();
      this.DeviceCombo = new System.Windows.Forms.ComboBox();
      this.PropertyCombo = new System.Windows.Forms.ComboBox();
      this.SubsystemCombo = new System.Windows.Forms.ComboBox();
      this.SubsystemLabel = new System.Windows.Forms.Label();
      this.ServerLabel = new System.Windows.Forms.Label();
      this.DeviceLabel = new System.Windows.Forms.Label();
      this.PropertyLabel = new System.Windows.Forms.Label();
      this.AddressList = new System.Windows.Forms.ListBox();
      this.GroupLinksCheckBox = new System.Windows.Forms.CheckBox();
      this.FlagsCombo = new System.Windows.Forms.ComboBox();
      this.FlagsLabel = new System.Windows.Forms.Label();
      this.InputTextBox = new System.Windows.Forms.TextBox();
      this.InputLabel = new System.Windows.Forms.Label();
      this.OKButton = new System.Windows.Forms.Button();
      this.CancelButton = new System.Windows.Forms.Button();
      this.AddButton = new System.Windows.Forms.Button();
      this.ProtocolCombo = new System.Windows.Forms.ComboBox();
      this.ProtocolLabel = new System.Windows.Forms.Label();
      this.ToolTip = new System.Windows.Forms.ToolTip(this.components);
      this.ErrorValueBox = new System.Windows.Forms.TextBox();
      this.ErrorValueLabel = new System.Windows.Forms.Label();
      this.LinkCapacityBox = new System.Windows.Forms.TextBox();
      this.LinkCapacityLabel = new System.Windows.Forms.Label();
      this.IntervalBox = new System.Windows.Forms.TextBox();
      this.IntervalLabel = new System.Windows.Forms.Label();
      this.AutoAttachCheck = new System.Windows.Forms.CheckBox();
      this.AutoUpdateCheck = new System.Windows.Forms.CheckBox();
      this.AutoTrendCheck = new System.Windows.Forms.CheckBox();
      this.AutoScaleCombo = new System.Windows.Forms.ComboBox();
      this.AutoScaleLabel = new System.Windows.Forms.Label();
      this.TrendDepthTextBox = new System.Windows.Forms.TextBox();
      this.TrendDepthLabel = new System.Windows.Forms.Label();
      this.MCALabel = new System.Windows.Forms.Label();
      this.MCACaption = new System.Windows.Forms.Label();
      this.InfoLabel = new System.Windows.Forms.Label();
      this.SuspendLayout();
      // 
      // ContextCombo
      // 
      this.ContextCombo.DropDownStyle = System.Windows.Forms.ComboBoxStyle.DropDownList;
      this.ContextCombo.FormattingEnabled = true;
      this.ContextCombo.Location = new System.Drawing.Point(111, 26);
      this.ContextCombo.Name = "ContextCombo";
      this.ContextCombo.Size = new System.Drawing.Size(92, 21);
      this.ContextCombo.Sorted = true;
      this.ContextCombo.TabIndex = 0;
      this.ContextCombo.DrawMode = DrawMode.OwnerDrawFixed;
      this.ContextCombo.DrawItem += new System.Windows.Forms.DrawItemEventHandler(this.ComboDrawItem);
      this.ContextCombo.DropDownClosed += new System.EventHandler(this.ComboDropDownClosed);
      this.ContextCombo.TextChanged += new System.EventHandler(this.ContextChanged);
      // 
      // ContextLabel
      // 
      this.ContextLabel.AutoSize = true;
      this.ContextLabel.Location = new System.Drawing.Point(113, 11);
      this.ContextLabel.Name = "ContextLabel";
      this.ContextLabel.Size = new System.Drawing.Size(43, 13);
      this.ContextLabel.TabIndex = 1;
      this.ContextLabel.Text = "Context";
      // 
      // ServerCombo
      // 
      this.ServerCombo.DropDownStyle = System.Windows.Forms.ComboBoxStyle.DropDownList;
      this.ServerCombo.FormattingEnabled = true;
      this.ServerCombo.Location = new System.Drawing.Point(9, 70);
      this.ServerCombo.Name = "ServerCombo";
      this.ServerCombo.Size = new System.Drawing.Size(96, 21);
      this.ServerCombo.Sorted = true;
      this.ServerCombo.TabIndex = 2;
      this.ServerCombo.DrawMode = DrawMode.OwnerDrawFixed;
      this.ServerCombo.DrawItem += new System.Windows.Forms.DrawItemEventHandler(this.ComboDrawItem);
      this.ServerCombo.DropDownClosed += new System.EventHandler(this.ComboDropDownClosed);
      this.ServerCombo.TextChanged += new System.EventHandler(this.ServerChanged);
      // 
      // DeviceCombo
      // 
      this.DeviceCombo.FormattingEnabled = true;
      this.DeviceCombo.Location = new System.Drawing.Point(111, 70);
      this.DeviceCombo.Name = "DeviceCombo";
      this.DeviceCombo.Size = new System.Drawing.Size(92, 21); // 110, 21
      this.DeviceCombo.TabIndex = 3;
      this.DeviceCombo.DrawMode = DrawMode.OwnerDrawFixed;
      this.DeviceCombo.DrawItem += new System.Windows.Forms.DrawItemEventHandler(this.ComboDrawItem);
      this.DeviceCombo.DropDownClosed += new System.EventHandler(this.ComboDropDownClosed);
      this.DeviceCombo.TextChanged += new System.EventHandler(this.DeviceChanged);
      // 
      // PropertyCombo
      // 
      this.PropertyCombo.FormattingEnabled = true;
      this.PropertyCombo.Location = new System.Drawing.Point(209, 70); // 227, 70
      this.PropertyCombo.Name = "PropertyCombo";
      this.PropertyCombo.Size = new System.Drawing.Size(92, 21); // 131, 21
      this.PropertyCombo.Sorted = true;
      this.PropertyCombo.TabIndex = 4;
      this.PropertyCombo.DrawMode = DrawMode.OwnerDrawFixed;
      this.PropertyCombo.DrawItem += new System.Windows.Forms.DrawItemEventHandler(this.ComboDrawItem);
      this.PropertyCombo.DropDownClosed += new System.EventHandler(this.ComboDropDownClosed);
      this.PropertyCombo.TextChanged += new System.EventHandler(this.PropertyChanged);
      // 
      // SubsystemCombo
      // 
      this.SubsystemCombo.FormattingEnabled = true;
      this.SubsystemCombo.Location = new System.Drawing.Point(209, 27);
      this.SubsystemCombo.Name = "SubsystemCombo";
      this.SubsystemCombo.Size = new System.Drawing.Size(86, 21);
      this.SubsystemCombo.Sorted = true;
      this.SubsystemCombo.TabIndex = 5;
      this.SubsystemCombo.DrawMode = DrawMode.OwnerDrawFixed;
      this.SubsystemCombo.DrawItem += new System.Windows.Forms.DrawItemEventHandler(this.ComboDrawItem);
      this.SubsystemCombo.DropDownClosed += new System.EventHandler(this.ComboDropDownClosed);
      this.SubsystemCombo.TextChanged += new System.EventHandler(this.SubsystemChanged);
      // 
      // SubsystemLabel
      // 
      this.SubsystemLabel.AutoSize = true;
      this.SubsystemLabel.Location = new System.Drawing.Point(212, 11);
      this.SubsystemLabel.Name = "SubsystemLabel";
      this.SubsystemLabel.Size = new System.Drawing.Size(58, 13);
      this.SubsystemLabel.TabIndex = 6;
      this.SubsystemLabel.Text = "Subsystem";
      // 
      // ServerLabel
      // 
      this.ServerLabel.AutoSize = true;
      this.ServerLabel.Location = new System.Drawing.Point(11, 54);
      this.ServerLabel.Name = "ServerLabel";
      this.ServerLabel.Size = new System.Drawing.Size(38, 13);
      this.ServerLabel.TabIndex = 7;
      this.ServerLabel.Text = "Server";
      // 
      // DeviceLabel
      // 
      this.DeviceLabel.AutoSize = true;
      this.DeviceLabel.Location = new System.Drawing.Point(115, 54);
      this.DeviceLabel.Name = "DeviceLabel";
      this.DeviceLabel.Size = new System.Drawing.Size(41, 13);
      this.DeviceLabel.TabIndex = 8;
      this.DeviceLabel.Text = "Device";
      // 
      // PropertyLabel
      // 
      this.PropertyLabel.AutoSize = true;
      this.PropertyLabel.Location = new System.Drawing.Point(212, 54); // 230 , 54
      this.PropertyLabel.Name = "PropertyLabel";
      this.PropertyLabel.Size = new System.Drawing.Size(46, 13);
      this.PropertyLabel.TabIndex = 9;
      this.PropertyLabel.Text = "Property";
      // 
      // AddressList
      // 
      this.AddressList.FormattingEnabled = true;
      this.AddressList.Location = new System.Drawing.Point(9, 238);
      this.AddressList.Name = "AddressList";
      this.AddressList.Size = new System.Drawing.Size(349, 69);
      this.AddressList.TabIndex = 10;
      this.AddressList.KeyDown += new System.Windows.Forms.KeyEventHandler(this.AddressListKeyPressed);
      // 
      // GroupLinksCheckBox
      // 
      this.GroupLinksCheckBox.AutoSize = true;
      this.GroupLinksCheckBox.Location = new System.Drawing.Point(272, 215);
      this.GroupLinksCheckBox.Name = "GroupLinksCheckBox";
      this.GroupLinksCheckBox.Size = new System.Drawing.Size(95, 17);
      this.GroupLinksCheckBox.TabIndex = 11;
      this.GroupLinksCheckBox.Text = "Grouped Links";
      this.GroupLinksCheckBox.UseVisualStyleBackColor = true;
      // 
      // FlagsCombo
      // 
      this.FlagsCombo.FormattingEnabled = true;
      this.FlagsCombo.Location = new System.Drawing.Point(10, 116);
      this.FlagsCombo.Name = "FlagsCombo";
      this.FlagsCombo.Size = new System.Drawing.Size(177, 21);
      this.FlagsCombo.TabIndex = 14;
      // 
      // FlagsLabel
      // 
      this.FlagsLabel.AutoSize = true;
      this.FlagsLabel.Location = new System.Drawing.Point(7, 100);
      this.FlagsLabel.Name = "FlagsLabel";
      this.FlagsLabel.Size = new System.Drawing.Size(81, 13);
      this.FlagsLabel.TabIndex = 15;
      this.FlagsLabel.Text = "Additional Flags";
      // 
      // InputTextBox
      // 
      this.InputTextBox.Location = new System.Drawing.Point(193, 116);
      this.InputTextBox.Name = "InputTextBox";
      this.InputTextBox.Size = new System.Drawing.Size(82, 20);
      this.InputTextBox.TabIndex = 16;
      // 
      // InputLabel
      // 
      this.InputLabel.AutoSize = true;
      //this.InputLabel.Font = new System.Drawing.Font("Microsoft Sans Serif", 8.25F, System.Drawing.FontStyle.Regular, System.Drawing.GraphicsUnit.Point, ((byte)(0)));
      this.InputLabel.Location = new System.Drawing.Point(190, 100);
      this.InputLabel.Name = "InputLabel";
      this.InputLabel.Size = new System.Drawing.Size(61, 13);
      this.InputLabel.TabIndex = 17;
      this.InputLabel.Text = "Input Value";
      // 
      // OKButton
      // 
      this.OKButton.Location = new System.Drawing.Point(301, 11);
      this.OKButton.Name = "OKButton";
      this.OKButton.Size = new System.Drawing.Size(57, 21);
      this.OKButton.TabIndex = 18;
      this.OKButton.Text = "OK";
      this.OKButton.UseVisualStyleBackColor = true;
      this.OKButton.Click += new System.EventHandler(this.OKClicked);
      // 
      // CancelButton
      // 
      this.CancelButton.Location = new System.Drawing.Point(301, 35);
      this.CancelButton.Name = "CancelButton";
      this.CancelButton.Size = new System.Drawing.Size(57, 21);
      this.CancelButton.TabIndex = 19;
      this.CancelButton.Text = "Cancel";
      this.CancelButton.UseVisualStyleBackColor = true;
      this.CancelButton.Click += new System.EventHandler(this.CancelClicked);
      // 
      // AddButton
      // 
      this.AddButton.Location = new System.Drawing.Point(9, 211);
      this.AddButton.Name = "AddButton";
      this.AddButton.Size = new System.Drawing.Size(57, 21);
      this.AddButton.TabIndex = 20;
      this.AddButton.Text = "Add";
      this.AddButton.UseVisualStyleBackColor = true;
      this.AddButton.Click += new System.EventHandler(this.AddClicked);
      // 
      // ProtocolCombo
      // 
      this.ProtocolCombo.DropDownStyle = System.Windows.Forms.ComboBoxStyle.DropDownList;
      this.ProtocolCombo.FormattingEnabled = true;
      this.ProtocolCombo.Location = new System.Drawing.Point(9, 26);
      this.ProtocolCombo.Name = "ProtocolCombo";
      this.ProtocolCombo.Size = new System.Drawing.Size(96, 21);
      this.ProtocolCombo.TabIndex = 21;
      this.ProtocolCombo.DrawMode = DrawMode.OwnerDrawFixed;
      this.ProtocolCombo.DrawItem += new System.Windows.Forms.DrawItemEventHandler(this.ComboDrawItem);
      this.ProtocolCombo.DropDownClosed += new System.EventHandler(this.ComboDropDownClosed);
      this.ProtocolCombo.TextChanged += new System.EventHandler(this.ProtocolChanged);
      // 
      // ProtocolLabel
      // 
      this.ProtocolLabel.AutoSize = true;
      this.ProtocolLabel.Location = new System.Drawing.Point(11, 11);
      this.ProtocolLabel.Name = "ProtocolLabel";
      this.ProtocolLabel.Size = new System.Drawing.Size(46, 13);
      this.ProtocolLabel.TabIndex = 22;
      this.ProtocolLabel.Text = "Protocol";
      // 
      // ErrorValueBox
      // 
      this.ErrorValueBox.Location = new System.Drawing.Point(281, 116);
      this.ErrorValueBox.Name = "ErrorValueBox";
      this.ErrorValueBox.Size = new System.Drawing.Size(77, 20);
      this.ErrorValueBox.TabIndex = 23;
      // 
      // ErrorValueLabel
      // 
      this.ErrorValueLabel.AutoSize = true;
      this.ErrorValueLabel.Font = new System.Drawing.Font("Microsoft Sans Serif", 8.25F, System.Drawing.FontStyle.Regular, System.Drawing.GraphicsUnit.Point, ((byte)(0)));
      this.ErrorValueLabel.Location = new System.Drawing.Point(278, 100);
      this.ErrorValueLabel.Name = "ErrorValueLabel";
      this.ErrorValueLabel.Size = new System.Drawing.Size(59, 13);
      this.ErrorValueLabel.TabIndex = 24;
      this.ErrorValueLabel.Text = "Error Value";
      // 
      // LinkCapacityBox
      // 
      this.LinkCapacityBox.Location = new System.Drawing.Point(9, 160);
      this.LinkCapacityBox.Name = "LinkCapacityBox";
      this.LinkCapacityBox.Size = new System.Drawing.Size(46, 20);
      this.LinkCapacityBox.TabIndex = 25;
      // 
      // LinkCapacityLabel
      // 
      this.LinkCapacityLabel.AutoSize = true;
      this.LinkCapacityLabel.Font = new System.Drawing.Font("Microsoft Sans Serif", 8.25F, System.Drawing.FontStyle.Regular, System.Drawing.GraphicsUnit.Point, ((byte)(0)));
      this.LinkCapacityLabel.Location = new System.Drawing.Point(7, 144);
      this.LinkCapacityLabel.Name = "LinkCapacityLabel";
      this.LinkCapacityLabel.Size = new System.Drawing.Size(48, 13);
      this.LinkCapacityLabel.TabIndex = 26;
      this.LinkCapacityLabel.Text = "Capacity";
      // 
      // IntervalBox
      // 
      this.IntervalBox.Location = new System.Drawing.Point(75, 160);
      this.IntervalBox.Name = "IntervalBox";
      this.IntervalBox.Size = new System.Drawing.Size(57, 20);
      this.IntervalBox.TabIndex = 27;
      this.IntervalBox.Text = "1000";
      // 
      // IntervalLabel
      // 
      this.IntervalLabel.AutoSize = true;
      this.IntervalLabel.Font = new System.Drawing.Font("Microsoft Sans Serif", 8.25F, System.Drawing.FontStyle.Regular, System.Drawing.GraphicsUnit.Point, ((byte)(0)));
      this.IntervalLabel.Location = new System.Drawing.Point(75, 145);
      this.IntervalLabel.Name = "IntervalLabel";
      this.IntervalLabel.Size = new System.Drawing.Size(42, 13);
      this.IntervalLabel.TabIndex = 28;
      this.IntervalLabel.Text = "Interval";
      // 
      // AutoAttachCheck
      // 
      this.AutoAttachCheck.AutoSize = true;
      this.AutoAttachCheck.Location = new System.Drawing.Point(272, 144);
      this.AutoAttachCheck.Name = "AutoAttachCheck";
      this.AutoAttachCheck.Size = new System.Drawing.Size(82, 17);
      this.AutoAttachCheck.TabIndex = 29;
      this.AutoAttachCheck.Text = "Auto-Attach";
      this.AutoAttachCheck.UseVisualStyleBackColor = true;
      // 
      // AutoUpdateCheck
      // 
      this.AutoUpdateCheck.AutoSize = true;
      this.AutoUpdateCheck.Location = new System.Drawing.Point(272, 167);
      this.AutoUpdateCheck.Name = "AutoUpdateCheck";
      this.AutoUpdateCheck.Size = new System.Drawing.Size(86, 17);
      this.AutoUpdateCheck.TabIndex = 30;
      this.AutoUpdateCheck.Text = "Auto-Update";
      this.AutoUpdateCheck.UseVisualStyleBackColor = true;
      // 
      // AutoTrendCheck
      // 
      this.AutoTrendCheck.AutoSize = true;
      this.AutoTrendCheck.Location = new System.Drawing.Point(272, 190);
      this.AutoTrendCheck.Name = "AutoTrendCheck";
      this.AutoTrendCheck.Size = new System.Drawing.Size(79, 17);
      this.AutoTrendCheck.TabIndex = 31;
      this.AutoTrendCheck.Text = "Auto-Trend";
      this.AutoTrendCheck.UseVisualStyleBackColor = true;
      this.AutoTrendCheck.Click += new System.EventHandler(this.AutoTrendClicked);
      // 
      // AutoScaleCombo
      // 
      this.AutoScaleCombo.DropDownStyle = System.Windows.Forms.ComboBoxStyle.DropDownList;
      this.AutoScaleCombo.FormattingEnabled = true;
      this.AutoScaleCombo.Location = new System.Drawing.Point(147, 160);
      this.AutoScaleCombo.Name = "AutoScaleCombo";
      this.AutoScaleCombo.Size = new System.Drawing.Size(68, 21);
      this.AutoScaleCombo.TabIndex = 34;
      // 
      // AutoScaleLabel
      // 
      this.AutoScaleLabel.AutoSize = true;
      this.AutoScaleLabel.Font = new System.Drawing.Font("Microsoft Sans Serif", 8.25F, System.Drawing.FontStyle.Regular, System.Drawing.GraphicsUnit.Point, ((byte)(0)));
      this.AutoScaleLabel.Location = new System.Drawing.Point(144, 144);
      this.AutoScaleLabel.Name = "AutoScaleLabel";
      this.AutoScaleLabel.Size = new System.Drawing.Size(59, 13);
      this.AutoScaleLabel.TabIndex = 35;
      this.AutoScaleLabel.Text = "Auto-Scale";
      // 
      // TrendDepthTextBox
      // 
      this.TrendDepthTextBox.Location = new System.Drawing.Point(215, 187);
      this.TrendDepthTextBox.Name = "TrendDepthTextBox";
      this.TrendDepthTextBox.Size = new System.Drawing.Size(28, 20);
      this.TrendDepthTextBox.TabIndex = 36;
      this.TrendDepthTextBox.Text = "24";
      // 
      // TrendDepthLabel
      // 
      this.TrendDepthLabel.AutoSize = true;
      this.TrendDepthLabel.Location = new System.Drawing.Point(246, 190);
      this.TrendDepthLabel.Name = "TrendDepthLabel";
      this.TrendDepthLabel.Size = new System.Drawing.Size(16, 13);
      this.TrendDepthLabel.TabIndex = 37;
      this.TrendDepthLabel.Text = "hr";
      // 
      // MCALabel
      // 
      this.MCALabel.AutoSize = true;
      this.MCALabel.BackColor = System.Drawing.Color.Lime;
      this.MCALabel.BorderStyle = System.Windows.Forms.BorderStyle.FixedSingle;
      this.MCALabel.Font = new System.Drawing.Font("Microsoft Sans Serif", 6F, System.Drawing.FontStyle.Regular, System.Drawing.GraphicsUnit.Point, ((byte)(0)));
      this.MCALabel.Location = new System.Drawing.Point(351, 76);
      this.MCALabel.Name = "MCALabel";
      this.MCALabel.Size = new System.Drawing.Size(11, 11);
      this.MCALabel.TabIndex = 38;
      this.MCALabel.Text = "  ";
      // 
      // MCACaption
      // 
      this.MCACaption.AutoSize = true;
      this.MCACaption.Font = new System.Drawing.Font("Microsoft Sans Serif", 6F, System.Drawing.FontStyle.Regular, System.Drawing.GraphicsUnit.Point, ((byte)(0)));
      this.MCACaption.Location = new System.Drawing.Point(327, 77);
      this.MCACaption.Name = "MCACaption";
      this.MCACaption.Size = new System.Drawing.Size(23, 9); this.MCACaption.TabIndex = 39;
      this.MCACaption.Text = "MCA";
      //
      // InfoPB
      //
      this.InfoLabel.AutoSize = true;
      this.InfoLabel.Font = new System.Drawing.Font("Times New Roman", 10F, System.Drawing.FontStyle.Bold, System.Drawing.GraphicsUnit.Point, ((byte)(0)));
      this.InfoLabel.Location = new System.Drawing.Point(305, 72);
      this.InfoLabel.Name = "InfoLabel";
      this.InfoLabel.Size = new System.Drawing.Size(20, 20);
      this.InfoLabel.TabIndex = 40;
      this.InfoLabel.Text = "i";
      this.InfoLabel.TextAlign = System.Drawing.ContentAlignment.MiddleCenter;
      // 
      // LinkProperties
      // 
      this.AllowDrop = true;
      this.AutoScaleDimensions = new System.Drawing.SizeF(6F, 13F);
      this.AutoScaleMode = System.Windows.Forms.AutoScaleMode.Font;
      this.ClientSize = new System.Drawing.Size(370, 314);
      this.Controls.Add(this.MCACaption);
      this.Controls.Add(this.MCALabel);
      this.Controls.Add(this.TrendDepthLabel);
      this.Controls.Add(this.TrendDepthTextBox);
      this.Controls.Add(this.AutoScaleLabel);
      this.Controls.Add(this.AutoScaleCombo);
      this.Controls.Add(this.AutoTrendCheck);
      this.Controls.Add(this.AutoUpdateCheck);
      this.Controls.Add(this.AutoAttachCheck);
      this.Controls.Add(this.IntervalLabel);
      this.Controls.Add(this.IntervalBox);
      this.Controls.Add(this.LinkCapacityLabel);
      this.Controls.Add(this.LinkCapacityBox);
      this.Controls.Add(this.ErrorValueLabel);
      this.Controls.Add(this.ErrorValueBox);
      this.Controls.Add(this.ProtocolLabel);
      this.Controls.Add(this.ProtocolLabel);
      this.Controls.Add(this.ProtocolCombo);
      this.Controls.Add(this.AddButton);
      this.Controls.Add(this.CancelButton);
      this.Controls.Add(this.OKButton);
      this.Controls.Add(this.InputLabel);
      this.Controls.Add(this.InputTextBox);
      this.Controls.Add(this.FlagsLabel);
      this.Controls.Add(this.FlagsCombo);
      this.Controls.Add(this.GroupLinksCheckBox);
      this.Controls.Add(this.AddressList);
      this.Controls.Add(this.PropertyLabel);
      this.Controls.Add(this.DeviceLabel);
      this.Controls.Add(this.ServerLabel);
      this.Controls.Add(this.SubsystemLabel);
      this.Controls.Add(this.SubsystemCombo);
      this.Controls.Add(this.PropertyCombo);
      this.Controls.Add(this.DeviceCombo);
      this.Controls.Add(this.ServerCombo);
      this.Controls.Add(this.ContextLabel);
      this.Controls.Add(this.ContextCombo);
      this.Controls.Add(this.InfoLabel);
      this.MaximizeBox = false;
      this.MinimizeBox = false;
      this.ControlBox = false;
      this.Name = "LinkProperties";
      this.Text = "Link Properties ...";
      this.DragDrop += new DragEventHandler(this.AddrDropped);
      this.DragEnter += new DragEventHandler(this.AddrDragEnter);
      this.ResumeLayout(false);
      this.PerformLayout();
      String[] items = new String[] { "off", "once", "always" }; // just reiterate them ...
      foreach (String s in items) AutoScaleCombo.Items.Add(s);
      GroupLinksCheckBox.Enabled = hasGroup;
      AddButton.Enabled = hasGroup;
      AddressList.Enabled = hasGroup;
      ErrorValueBox.Enabled = hasErrorValue;
      AutoTrendCheck.Enabled = hasTrendInfo;
      AutoScaleCombo.Enabled = hasAutoScale;
      AutoScaleLabel.Enabled = hasAutoScale;

      String tt = "optional value to use a input for automatic links";
      ToolTip.SetToolTip(this.InputTextBox, tt);
      ToolTip.SetToolTip(this.InputLabel, tt);

      tt = "an optional design-time browseable flags set which can be applied to underlying links";
      ToolTip.SetToolTip(this.FlagsLabel, tt);
      ToolTip.SetToolTip(this.FlagsCombo, tt);

      tt = "add the currently selected link address to an address list for grouped links";
      ToolTip.SetToolTip(this.AddButton, tt);

      tt = "to remove and item: select it and press the Delete key";
      ToolTip.SetToolTip(this.AddressList, tt);

      tt = "if checked: all links in the address list will fire a single update event";
      ToolTip.SetToolTip(this.GroupLinksCheckBox, tt);

      tt = "if checked: all links will automatically begin monitoring at program start time";
      ToolTip.SetToolTip(AutoAttachCheck, tt);

      tt = "if checked: the internal acop update delegate will display all monitored following an update event";
      ToolTip.SetToolTip(AutoUpdateCheck, tt);

      tt = "if checked: the internal acop update delegate will display incoming data to a trend chart display following an update event";
      ToolTip.SetToolTip(AutoTrendCheck, tt);

      tt = "if not empty, then the contents of the error value will be assigned to the monitored data in the event of any link error";
      ToolTip.SetToolTip(ErrorValueBox, tt);
      ToolTip.SetToolTip(ErrorValueLabel, tt);

      tt = "if non-zero the value indicates the prefered monitored data array size";
      if (hasTrendInfo) tt += " OR the trend chart depth if Auto-Trend is selected";
      ToolTip.SetToolTip(LinkCapacityBox, tt);
      ToolTip.SetToolTip(LinkCapacityLabel, tt);

      tt = "if non-zero the value gives the monitor polling interval in milliseconds";
      ToolTip.SetToolTip(IntervalBox, tt);
      ToolTip.SetToolTip(IntervalLabel, tt);

      tt = "gives the auto-scale charateristics of the chart display when Auto-Update is true";
      ToolTip.SetToolTip(AutoScaleCombo, tt);
      ToolTip.SetToolTip(AutoScaleLabel, tt);

      tt = "gives the trend chart depth in hours when Auto-Trend is true";
      ToolTip.SetToolTip(TrendDepthTextBox, tt);
      ToolTip.SetToolTip(TrendDepthLabel, tt);

      tt = "indicates whether or not the selected property is a multi-channel array";
      ToolTip.SetToolTip(MCACaption, tt);
      ToolTip.SetToolTip(MCALabel, tt);
    }
    public void AddrDropped(object sender, DragEventArgs e)
    {
      String ds = (String)e.Data.GetData(DataFormats.StringFormat);
      String[] parts = ds.Split(new char[] { '/' }, StringSplitOptions.RemoveEmptyEntries);
      int nparts = parts.Length;
      if (nparts > 0) ProtocolLabel.Text = parts[0];
      if (nparts > 1) ContextLabel.Text = parts[1];
      if (nparts > 2) ServerLabel.Text = parts[2];
      if (nparts > 3) DeviceLabel.Text = parts[3];
      if (nparts > 4) PropertyLabel.Text = parts[4];
    }
    public void AddrDragEnter(object sender, DragEventArgs e)
    {
      e.Effect = DragDropEffects.Copy;
    }
    private void ContextChanged(object sender, EventArgs e)
    {
      if (!hasInitialized) return;
      context = ContextCombo.Text;
      if (!this.control.BrowsingServiceAvailable) return;
      this.SubsystemCombo.Items.Clear();
      String[] items = AcopControl.AcopLink.GetSubSystemList(protocol, context);
      foreach (String item in items)
      {
        if (item.Length == 0) continue;
        this.SubsystemCombo.Items.Add(item);
      }
      this.SubsystemCombo.SelectedIndex = 0;

      this.ServerCombo.Items.Clear();
      items = AcopControl.AcopLink.GetServerList(protocol, context, subsystem);
      if (items == null) items = new String[] { AcopUserComponent.UNAVAILABLE };
      foreach (String item in items) this.ServerCombo.Items.Add(item);
      if (items[0].Equals(AcopUserComponent.UNAVAILABLE, StringComparison.CurrentCultureIgnoreCase))
      {
        this.control.BrowsingServiceAvailable = false;
      }

      this.ServerCombo.SelectedIndex = 0;
      this.ToolTip.SetToolTip(this.ContextCombo, context);
    }

    private void SubsystemChanged(object sender, EventArgs e)
    {
      if (!hasInitialized) return;
      subsystem = SubsystemCombo.Text;
      if (!this.control.BrowsingServiceAvailable) return;
      this.ServerCombo.Items.Clear();
      String[] items = AcopControl.AcopLink.GetServerList(protocol, context, subsystem);
      if (items != null)
        foreach (String item in items) this.ServerCombo.Items.Add(item);
      this.ServerCombo.SelectedIndex = 0;
    }

    private void ServerChanged(object sender, EventArgs e)
    {
      if (!hasInitialized) return;
      server = ServerCombo.Text;
      if (!this.control.BrowsingServiceAvailable) return;
      hasDeviceQueryPrecedence = AcopControl.AcopLink.HasDeviceQueryPrecedence(protocol, context, subsystem, server);
      hasPropertyQueryPrecedence = false;
      this.DeviceCombo.Items.Clear();
      String[] items = AcopControl.AcopLink.GetDeviceList(protocol, context, subsystem, server, null);
      if (items != null)
        foreach (String item in items) this.DeviceCombo.Items.Add(item);
      this.DeviceCombo.SelectedIndex = 0;
      this.ToolTip.SetToolTip(this.ServerCombo, server);
    }

    private void DeviceChanged(object sender, EventArgs e)
    {
      if (!hasInitialized) return;
      device = DeviceCombo.Text;
      if (device.Contains('*')) return;
      if (hasPropertyQueryPrecedence) return;
      if (!this.control.BrowsingServiceAvailable) return;
      this.ToolTip.SetToolTip(this.DeviceCombo, device);
      if (!DeviceCombo.Items.Contains(device)) return;
      this.PropertyCombo.Items.Clear();
      String[] items = AcopControl.AcopLink.GetPropertyList(protocol, context, subsystem, server, device);
      if (items != null)
        foreach (String item in items) this.PropertyCombo.Items.Add(item);
      this.PropertyCombo.SelectedIndex = 0;
      hasPropertyQueryPrecedence = AcopControl.AcopLink.HasPropertyQueryPrecedence(protocol, context, subsystem, server, property);
    }
    private AcopTransport.DisplaySettings dspSettings = null;
    private Boolean isMca()
    {
      if (dspSettings != null && AcopUserComponent.isChannelArray(dspSettings.GetArrayType()))
      {
        return true;
      }
      return false;
    }
    private Boolean isStruct()
    {
      if (dspSettings != null && AcopUserComponent.isStructure(dspSettings.GetArrayType()))
      {
        return true;
      }
      return false;
    }
    private void PropertyChanged(object sender, EventArgs e)
    {
      if (!hasInitialized) return;
      property = PropertyCombo.Text;
      if (!this.control.BrowsingServiceAvailable) return;
      if (!PropertyCombo.Items.Contains(property)) return;
      String addr = "/" + context + "/" + server + "/" + device;
      dspSettings = AcopControl.AcopLink.GetDisplaySettings(addr, property);
      SetInfoLabelTT(dspSettings);
      if (hasDeviceQueryPrecedence || !hasPropertyQueryPrecedence)
      {
        MCALabel.BackColor = isMca() ? Color.Lime : default(Color);
        return;
      }
      this.DeviceCombo.Items.Clear();
      String[] items = AcopControl.AcopLink.GetDeviceList(protocol, context, subsystem, server, property);
      if (items != null)
        foreach (String item in items) this.DeviceCombo.Items.Add(item);
      this.DeviceCombo.SelectedIndex = 0;
      this.ToolTip.SetToolTip(this.PropertyCombo, property);
      MCALabel.BackColor = isMca() ? Color.Lime : default(Color);
    }

    private void SetInfoLabelTT(AcopTransport.DisplaySettings dsp)
    {
      if (dsp == null) return;
      StringBuilder sb = new StringBuilder();
      sb.AppendLine("Property: " + property);
      sb.AppendLine("Description: " + dsp.GetDescription());
      sb.AppendLine("DataType: " + dsp.GetDataType());
      sb.AppendLine("Max Capacity: " + dsp.GetCapacity());

      ToolTip.SetToolTip(InfoLabel, sb.ToString());
    }

    private void OKClicked(object sender, EventArgs e)
    {
      this.control.LinkProtocol = protocol;
      this.control.LinkContext = context;
      this.control.LinkServer = server;
      this.control.LinkDevice = device;
      this.control.LinkProperty = property;
      this.control.LinkInputValue = InputTextBox.Text;
      this.control.LinkFlags = FlagsCombo.Text;
      if (AddressList.Items.Count > 0)
      {
        StringBuilder sb = new StringBuilder();
        foreach (var element in AddressList.Items)
        {
          sb.Append(element.ToString() + ";");
        }
        control.LinkList = sb.ToString();
        GroupLinksCheckBox.Checked = true;
      }
      else
      {
        control.LinkList = "";
      }
      this.control.LinkGrouped = GroupLinksCheckBox.Checked;
      this.control.AutoLinkUpdate = AutoUpdateCheck.Checked;
      this.control.AutoLinkAttach = AutoAttachCheck.Checked;
      int v = 0;
      if (Int32.TryParse(IntervalBox.Text, out v)) control.DefaultMonitorInterval = v;
      int cap = dspSettings.GetCapacity();
      Boolean useLinkCapacity = true;
      v = 0;
      if (control is AcopChart)
      {
        AcopChart ac = (AcopChart)this.control;
        ac.AutoScaleUpdate = (AcopChart.AutoScaleMode)AutoScaleCombo.SelectedIndex;
        ac.AutoTrendUpdate = AutoTrendCheck.Checked;
        if (ac.AutoTrendUpdate)
        {
          if (Int32.TryParse(LinkCapacityBox.Text, out v)) ac.LinkTrendDepth = v;
          useLinkCapacity = false;
          double dv;
          if (Double.TryParse(TrendDepthTextBox.Text, out dv)) ac.LinkTrendHours = dv;
        }
      }
      v = 0;
      if (useLinkCapacity)
      {
        if (Int32.TryParse(LinkCapacityBox.Text, out v)) control.AutoLinkCapacity = v;
      }
      if (control is AcopTable)
      {
        AcopTable at = (AcopTable)this.control;
        if (AddressList.Items.Count == 0)
        {
          if (v == 0 && cap > 1) at.NumberRows = cap;
          if (v == 1) at.DisplayAsMultiChannel = false;
          else if (isMca()) at.DisplayAsMultiChannel = true;
        }
        else
        {
          if (isMca())
          { // check multi column criteria ...
            Boolean multiColumns = true;
            String baseAddress = "/" + context + "/" + server + "/" + device;
            foreach (String item in AddressList.Items)
            {
              if (!item.StartsWith(baseAddress))
              {
                multiColumns = false;
                break;
              }
            }
            if (v == 1) multiColumns = false;
            if (multiColumns)
            {
              if (v == 0 && cap > 1) at.NumberRows = cap;
              at.DisplayAsMultiChannel = true;
              if (at.NumberColumns < AddressList.Items.Count)
                at.NumberColumns = AddressList.Items.Count;
            }
          }
          if (isStruct())
          {
            Boolean multiColumns = true;
            String baseAddress = "/" + context + "/" + server;
            foreach (String item in AddressList.Items)
            {
              if (!item.StartsWith(baseAddress) || !item.EndsWith(property))
              {
                multiColumns = false;
                break;
              }
            }
            if (multiColumns)
            {
              if (v == 0 && cap > 1) at.NumberRows = cap;
              at.DisplayAsMultiChannel = true;
              if (at.NumberColumns < AddressList.Items.Count)
                at.NumberColumns = AddressList.Items.Count;
            }
          }
        }
      }
      if (hasErrorValue)
      {
        control.ErrorValue = ErrorValueBox.Text;
      }
      editor.SetValuesStale(true);
      this.Dispose();
    }

    private void CancelClicked(object sender, EventArgs e)
    {
      editor.SetValuesStale(false);
      this.Dispose();
    }

    private void AddClicked(object sender, EventArgs e)
    {
      String txt = "/" + ContextCombo.Text + "/" + ServerCombo.Text + "/" + DeviceCombo.Text + "/" + PropertyCombo.Text;
      if (!AddressList.Items.Contains(txt)) AddressList.Items.Add(txt);
    }
    private void ProtocolChanged(object sender, EventArgs e)
    {
      if (!hasInitialized) return;
      protocol = ProtocolCombo.Text;
      this.control.BrowsingServiceAvailable = true;
      this.control.LinkProtocol = protocol;
      this.ContextCombo.Items.Clear();
      String[] items = AcopControl.AcopLink.GetContextList(protocol);
      if (items != null)
        foreach (String item in items) this.ContextCombo.Items.Add(item);
      if (items == null || items[0].Equals(AcopUserComponent.UNAVAILABLE,StringComparison.CurrentCultureIgnoreCase))
      {
        this.control.BrowsingServiceAvailable = false;
      }
      this.ContextCombo.SelectedIndex = 0;
    }
    private void AddressListKeyPressed(object sender, KeyEventArgs e)
    {
      if (AddressList.SelectedItems.Count == 0) return;
      if (e.KeyCode == Keys.Delete)
      {
        for (int i = AddressList.SelectedItems.Count - 1; i >= 0; i--)
          AddressList.Items.Remove(AddressList.SelectedItems[i]);
        if (AddressList.Items.Count > 0)
        {
          var element = AddressList.Items[AddressList.Items.Count - 1];
          String[] parts = element.ToString().Split(new char[] { '/' }, StringSplitOptions.RemoveEmptyEntries);
          if (parts.Length > 0) ContextCombo.Text = parts[0];
          if (parts.Length > 1) ServerCombo.Text = parts[1];
          if (parts.Length > 2) DeviceCombo.Text = parts[2];
          if (parts.Length > 3) PropertyCombo.Text = parts[3];
        }
      }
    }
    private void AutoTrendClicked(object sender, EventArgs e)
    {
      TrendDepthTextBox.Visible = AutoTrendCheck.Checked;
      TrendDepthLabel.Visible = AutoTrendCheck.Checked;
      if (AutoTrendCheck.Checked)
      {
        if (LinkCapacityBox.Text.Equals("0")) LinkCapacityBox.Text = "1200";
        else
        {
          AcopChart ac = (AcopChart)this.control;
          LinkCapacityBox.Text = ac.LinkTrendDepth.ToString();
        }
      }
      else
      {
        LinkCapacityBox.Text = this.control.AutoLinkCapacity.ToString();
      }
    }
    private void showComboToolTip(ComboBox cb, DrawItemEventArgs e)
    {
      if (e.Index < 0) return; // added this line thanks to Andrew's comment
      string text = cb.GetItemText(cb.Items[e.Index]);
      e.DrawBackground();
      using (SolidBrush br = new SolidBrush(e.ForeColor))
      {
        e.Graphics.DrawString(text, e.Font, br, e.Bounds);
      }
      if (!cb.DroppedDown) return;
      if ((e.State & DrawItemState.Selected) == DrawItemState.Selected)
      {
        int bottom = e.Bounds.Bottom;
        ToolTip.Show(text, cb, e.Bounds.Right, bottom);
        e.DrawFocusRectangle();
      }
    }
    private void ComboDrawItem(object sender, DrawItemEventArgs e)
    {
      if (sender is ComboBox)
      {
        ComboBox combo = (ComboBox)sender;
        showComboToolTip(combo, e);
      }
    }
    private void ComboDropDownClosed(object sender, EventArgs e)
    {
      if (sender is ComboBox)
      {
        ComboBox combo = (ComboBox)sender;
        ToolTip.Hide(combo);
      }
    }

    private System.Windows.Forms.ComboBox ContextCombo;
    private System.Windows.Forms.Label ContextLabel;
    private System.Windows.Forms.ComboBox ServerCombo;
    private System.Windows.Forms.ComboBox DeviceCombo;
    private System.Windows.Forms.ComboBox PropertyCombo;
    private System.Windows.Forms.ComboBox SubsystemCombo;
    private System.Windows.Forms.Label SubsystemLabel;
    private System.Windows.Forms.Label ServerLabel;
    private System.Windows.Forms.Label DeviceLabel;
    private System.Windows.Forms.Label PropertyLabel;
    private System.Windows.Forms.ListBox AddressList;
    private System.Windows.Forms.CheckBox GroupLinksCheckBox;
    private System.Windows.Forms.ComboBox FlagsCombo;
    private System.Windows.Forms.Label FlagsLabel;
    private System.Windows.Forms.TextBox InputTextBox;
    private System.Windows.Forms.Label InputLabel;
    private System.Windows.Forms.Button OKButton;
    private new System.Windows.Forms.Button CancelButton;
    private System.Windows.Forms.Button AddButton;
    private System.Windows.Forms.ComboBox ProtocolCombo;
    private System.Windows.Forms.Label ProtocolLabel;
    private System.Windows.Forms.ToolTip ToolTip;
    private System.Windows.Forms.TextBox ErrorValueBox;
    private System.Windows.Forms.Label ErrorValueLabel;
    private System.Windows.Forms.TextBox LinkCapacityBox;
    private System.Windows.Forms.Label LinkCapacityLabel;
    private System.Windows.Forms.TextBox IntervalBox;
    private System.Windows.Forms.Label IntervalLabel;
    private System.Windows.Forms.CheckBox AutoAttachCheck;
    private System.Windows.Forms.CheckBox AutoUpdateCheck;
    private System.Windows.Forms.CheckBox AutoTrendCheck;
    private System.Windows.Forms.ComboBox AutoScaleCombo;
    private System.Windows.Forms.Label AutoScaleLabel;
    private System.Windows.Forms.TextBox TrendDepthTextBox;
    private System.Windows.Forms.Label TrendDepthLabel;
    private System.Windows.Forms.Label MCALabel;
    private System.Windows.Forms.Label MCACaption;
    private System.Windows.Forms.Label InfoLabel;

  }

  [TypeConverter(typeof(ColorListEditorConverter))]
  public class ColorListEditorValue
  {
#   pragma warning disable 1591
    private AcopUserComponent acop = null;
    private String key = AcopUserComponent.PRIMARY_COLOR_LIST;
    private String valuesString = null;
    private void setValuesString(String value)
    {
      valuesString = value;
      acop.SetParseableColorList(key, value);
    }
    public String Colors { get { return valuesString; } set { setValuesString(value); } }
    public ColorListEditorValue(String value, String key, AcopUserComponent acop)
    {
      if (value == null) value = "";
      valuesString = value;
      if (acop != null) this.acop = acop;
      if (key != null) this.key = key;
    }
    public override string ToString()
    {
      return valuesString;
    }
  }
  public class ColorListEditorConverter : TypeConverter
  {
#   pragma warning disable 1591
    public override object ConvertTo(
        ITypeDescriptorContext context,
        System.Globalization.CultureInfo culture,
        object value, Type destinationType)
    {
      if (destinationType == typeof(string)) return value.ToString();
      return base.ConvertTo(context, culture, value, destinationType);
    }
    public override object ConvertFrom(
        ITypeDescriptorContext context,
        System.Globalization.CultureInfo culture,
        object value)
    {
      if (value.GetType() == typeof(string))
      {
        String key = null;
        AcopUserComponent acop = null;
        if (context is AcopUserComponent)
        {
          acop = (AcopUserComponent)context;
          key = acop.GetCurrentColorListKey();
        }
        return new ColorListEditorValue((string)(value), key, acop);
      }
      else
      {
        return base.ConvertFrom(context, culture, value);
      }
    }
    public override bool CanConvertTo(ITypeDescriptorContext context, Type destinationType)
    {
      if (destinationType == typeof(string)) return true;
      return base.CanConvertTo(context, destinationType);
    }
    public override bool CanConvertFrom(ITypeDescriptorContext context, Type sourceType)
    {
      if (sourceType == typeof(string)) return true;
      return base.CanConvertFrom(context, sourceType);
    }
  }
  public class ColorListEditor : UITypeEditor
  {
#   pragma warning disable 1591
    private Boolean valuesStale = false;
    private String valuesString = null;
    private String key = null;
    public ColorListEditor()
    {
    }
    public ColorListEditor(String value)
    {
      valuesString = value;
    }
    public override string ToString()
    {
      return valuesString;
    }
    public void SetValuesStale(Boolean value) { valuesStale = value; }
    public override UITypeEditorEditStyle GetEditStyle(System.ComponentModel.ITypeDescriptorContext context)
    {
      return UITypeEditorEditStyle.Modal;
    }
    public override object EditValue(ITypeDescriptorContext context, IServiceProvider provider, object value)
    {
      AcopUserComponent acop = (AcopUserComponent)context.Instance;
      IWindowsFormsEditorService frmsvr = (IWindowsFormsEditorService)provider.GetService(typeof(IWindowsFormsEditorService));
      if (frmsvr == null) return null;
      key = acop.GetCurrentColorListKey();
      valuesString = acop.GetParseableColorList(key);
      ColorSelector page = new ColorSelector(acop);
      page.SetEditor(this);
      frmsvr.ShowDialog(page);
      if (valuesStale)
      {
        valuesString = acop.GetParseableColorList(key);
        value = new ColorListEditorValue(valuesString, key, acop);
      }
      return base.EditValue(context, provider, value);
    }
  }
  public class ColorSelector : Form
  {
    private AcopUserComponent acop = null;
    private ColorListEditor editor = null;
    private ColorListEditorValue editorValue = null;
    private String key = null;
    public void SetEditor(ColorListEditor editor)
    {
      this.editor = editor;
    }
    /// <summary>
    /// Required designer variable.
    /// </summary>
    private System.ComponentModel.IContainer components = null;

    /// <summary>
    /// Clean up any resources being used.
    /// </summary>
    /// <param name="disposing">true if managed resources should be disposed; otherwise, false.</param>
    protected override void Dispose(bool disposing)
    {
      if (disposing && (components != null))
      {
        components.Dispose();
      }
      base.Dispose(disposing);
    }

    #region Windows Form Designer generated code

    /// <summary>
    /// Required method for Designer support - do not modify
    /// the contents of this method with the code editor.
    /// </summary>
    private void InitializeComponent()
    {
      this.SelectedColorsList = new System.Windows.Forms.ListBox();
      this.ColorPaletteCombo = new System.Windows.Forms.ComboBox();
      this.colorDialog1 = new System.Windows.Forms.ColorDialog();
      this.AddButton = new System.Windows.Forms.Button();
      this.OKButton = new System.Windows.Forms.Button();
      this.CancelButton = new System.Windows.Forms.Button();
      this.SuspendLayout();
      // 
      // SelectedColorsList
      // 
      this.SelectedColorsList.DrawMode = System.Windows.Forms.DrawMode.OwnerDrawFixed;
      this.SelectedColorsList.FormattingEnabled = true;
      this.SelectedColorsList.Location = new System.Drawing.Point(12, 39);
      this.SelectedColorsList.Name = "SelectedColorsList";
      this.SelectedColorsList.Size = new System.Drawing.Size(224, 82);
      this.SelectedColorsList.TabIndex = 1;
      this.SelectedColorsList.DrawItem += new System.Windows.Forms.DrawItemEventHandler(this.SelectedColorsList_DrawItem);
      this.SelectedColorsList.MeasureItem += new System.Windows.Forms.MeasureItemEventHandler(this.SelectedColorsList_MeasureItem);
      this.SelectedColorsList.KeyDown += new System.Windows.Forms.KeyEventHandler(this.SelectedColorsList_KeyDown);
      // 
      // ColorPaletteCombo
      // 
      this.ColorPaletteCombo.DrawMode = System.Windows.Forms.DrawMode.OwnerDrawFixed;
      this.ColorPaletteCombo.DropDownStyle = System.Windows.Forms.ComboBoxStyle.DropDownList;
      this.ColorPaletteCombo.FormattingEnabled = true;
      this.ColorPaletteCombo.Location = new System.Drawing.Point(12, 12);
      this.ColorPaletteCombo.Name = "ColorPaletteCombo";
      this.ColorPaletteCombo.Size = new System.Drawing.Size(183, 21);
      this.ColorPaletteCombo.TabIndex = 2;
      this.ColorPaletteCombo.DrawItem += new System.Windows.Forms.DrawItemEventHandler(this.ColorPaletteCombo_DrawItem);
      // 
      // AddButton
      // 
      this.AddButton.Location = new System.Drawing.Point(201, 14);
      this.AddButton.Name = "AddButton";
      this.AddButton.Size = new System.Drawing.Size(35, 19);
      this.AddButton.TabIndex = 3;
      this.AddButton.Text = "Add";
      this.AddButton.UseVisualStyleBackColor = true;
      this.AddButton.Click += new System.EventHandler(this.AddButton_Click);
      // 
      // OKButton
      // 
      this.OKButton.Location = new System.Drawing.Point(125, 127);
      this.OKButton.Name = "OKButton";
      this.OKButton.Size = new System.Drawing.Size(50, 24);
      this.OKButton.TabIndex = 4;
      this.OKButton.Text = "OK";
      this.OKButton.UseVisualStyleBackColor = true;
      this.OKButton.Click += new System.EventHandler(this.OKButton_Click);
      // 
      // CancelButton
      // 
      this.CancelButton.Location = new System.Drawing.Point(186, 127);
      this.CancelButton.Name = "CancelButton";
      this.CancelButton.Size = new System.Drawing.Size(50, 24);
      this.CancelButton.TabIndex = 5;
      this.CancelButton.Text = "Cancel";
      this.CancelButton.UseVisualStyleBackColor = true;
      this.CancelButton.Click += new System.EventHandler(this.CancelButton_Click);
      // 
      // ColorSelector
      // 
      this.AutoScaleDimensions = new System.Drawing.SizeF(6F, 13F);
      this.AutoScaleMode = System.Windows.Forms.AutoScaleMode.Font;
      this.ClientSize = new System.Drawing.Size(248, 157);
      this.ControlBox = false;
      this.Controls.Add(this.CancelButton);
      this.Controls.Add(this.OKButton);
      this.Controls.Add(this.AddButton);
      this.Controls.Add(this.ColorPaletteCombo);
      this.Controls.Add(this.SelectedColorsList);
      this.Name = "ColorSelector";
      this.Text = "Color Selector";
      this.ResumeLayout(false);

    }

    #endregion
    private System.Windows.Forms.ListBox SelectedColorsList;
    private System.Windows.Forms.ComboBox ColorPaletteCombo;
    private System.Windows.Forms.ColorDialog colorDialog1;
    private System.Windows.Forms.Button AddButton;
    private System.Windows.Forms.Button OKButton;
    private new System.Windows.Forms.Button CancelButton;
    public ColorSelector(AcopUserComponent acop)
    {
      InitializeComponent();
      this.acop = acop;
      key = acop.GetCurrentColorListKey();
      initSettings();
      editorValue = new ColorListEditorValue(acop.GetParseableColorList(key), key, acop);
    }
    private void initSettings()
    {
      KnownColor[] colors = (KnownColor[])Enum.GetValues(typeof(KnownColor));
      foreach (KnownColor c in colors)
      {
        ColorPaletteCombo.Items.Add(c.ToString());
      }
      ColorPaletteCombo.Text = "DarkBlue";
      String clst = acop.GetParseableColorList(key);
      String[] parts = clst.Split(new char[] { ';' }, StringSplitOptions.RemoveEmptyEntries);
      foreach (String c in parts)
        SelectedColorsList.Items.Add(c);
    }
    private void ColorPaletteCombo_DrawItem(object sender, DrawItemEventArgs e)
    {
      Graphics g = e.Graphics;
      Rectangle rect = e.Bounds;
      if (e.Index >= 0)
      {
        string n = ((ComboBox)sender).Items[e.Index].ToString();
        Font f = new Font("Arial", 9, FontStyle.Regular);
        Color c = Color.FromName(n);
        Brush b = new SolidBrush(c);
        g.DrawString(n, f, Brushes.Black, rect.X, rect.Top);
        g.FillRectangle(b, rect.X + 110, rect.Y + 5,
                        rect.Width - 10, rect.Height - 10);
      }
    }
    private void SelectedColorsList_DrawItem(object sender, DrawItemEventArgs e)
    {
      Graphics g = e.Graphics;
      Rectangle rect = e.Bounds;
      if (e.Index >= 0)
      {
        ListBox lb = ((ListBox)sender);
        string n = lb.Items[e.Index].ToString();
        Font f = new Font("Arial", 9, FontStyle.Regular);
        Color c = Color.FromName(n);
        Brush b = new SolidBrush(c);
        e.DrawBackground();
        if ((e.State & DrawItemState.Selected) == DrawItemState.Selected)
        {
          g.DrawString(n, f, SystemBrushes.HighlightText, rect.X, rect.Top);
        }
        else
        {
          g.DrawString(n, f, Brushes.Black, rect.X, rect.Top);
        }
        g.FillRectangle(b, rect.X + 110, rect.Y + 5,
                        rect.Width - 10, rect.Height - 10);
      }
    }
    private void SelectedColorsList_MeasureItem(object sender, MeasureItemEventArgs e)
    {
      // Get the ListBox and the item.
      ListBox lst = sender as ListBox;
      string txt = (string)lst.Items[e.Index];
      // Measure the string.
      SizeF txt_size = e.Graphics.MeasureString(txt, this.Font);
      // Set the required size.
      e.ItemHeight = (int)txt_size.Height + 2;
      e.ItemWidth = (int)txt_size.Width;
    }
    private void SelectedColorsList_KeyDown(object sender, KeyEventArgs e)
    {
      if (SelectedColorsList.SelectedItems.Count == 0) return;
      if (e.KeyCode == Keys.Delete)
      {
        for (int i = SelectedColorsList.SelectedItems.Count - 1; i >= 0; i--)
          SelectedColorsList.Items.Remove(SelectedColorsList.SelectedItems[i]);
      }
    }
    private void AddButton_Click(object sender, EventArgs e)
    {
      SelectedColorsList.Items.Add(ColorPaletteCombo.Text);
    }
    private void CancelButton_Click(object sender, EventArgs e)
    {
      editor.SetValuesStale(false);
      this.Dispose();
    }
    private void OKButton_Click(object sender, EventArgs e)
    {
      StringBuilder sb = new StringBuilder();
      foreach (String s in SelectedColorsList.Items)
      {
        sb.Append(s + ";");
      }
      String lst = sb.ToString();
      acop.SetParseableColorList(lst);
      editorValue.Colors = lst;
      editor.SetValuesStale(true);
      this.Dispose();
    }
  }

  [TypeConverter(typeof(TargetControlEditorConverter))]
  public class TargetControlEditorValue
  {
#   pragma warning disable 1591
    private AcopUserComponent acop = null;
    private String valuesString = null;
    private void setValuesString(String value)
    {
      valuesString = value;
      acop.setTargetComponent(value);
    }
    public String Component { get { return valuesString; } set { setValuesString(value); } }
    public TargetControlEditorValue(String value, AcopUserComponent acop)
    {
      if (value == null) value = "";
      valuesString = value;
      this.acop = acop;
    }
    public override string ToString()
    {
      return valuesString;
    }
  }

  public class TargetControlEditorConverter : TypeConverter
  {
#   pragma warning disable 1591
    public override object ConvertTo(
        ITypeDescriptorContext context,
        System.Globalization.CultureInfo culture,
        object value, Type destinationType)
    {
      if (destinationType == typeof(string)) return value.ToString();
      return base.ConvertTo(context, culture, value, destinationType);
    }
    public override object ConvertFrom(
        ITypeDescriptorContext context,
        System.Globalization.CultureInfo culture,
        object value)
    {
      if (value.GetType() == typeof(string))
      {
        return new TargetControlEditorValue((string)(value), (AcopUserComponent)context);
      }
      else
      {
        return base.ConvertFrom(context, culture, value);
      }
    }
    public override bool CanConvertTo(ITypeDescriptorContext context, Type destinationType)
    {
      if (destinationType == typeof(string)) return true;
      return base.CanConvertTo(context, destinationType);
    }
    public override bool CanConvertFrom(ITypeDescriptorContext context, Type sourceType)
    {
      if (sourceType == typeof(string)) return true;
      return base.CanConvertFrom(context, sourceType);
    }
  }

  public class TargetControlEditor : UITypeEditor
  {
#   pragma warning disable 1591
    private Boolean valuesStale = false;
    private String valuesString = null;

    public TargetControlEditor()
    {
    }
    public TargetControlEditor(String value)
    {
      valuesString = value;
    }
    public override string ToString()
    {
      return valuesString;
    }
    public void SetValuesStale(Boolean value) { valuesStale = value; }
    public override UITypeEditorEditStyle GetEditStyle(System.ComponentModel.ITypeDescriptorContext context)
    {
      return UITypeEditorEditStyle.Modal;
    }
    public override object EditValue(ITypeDescriptorContext context, IServiceProvider provider, object value)
    {
      AcopUserComponent acop = (AcopUserComponent)context.Instance;
      IWindowsFormsEditorService frmsvr = (IWindowsFormsEditorService)provider.GetService(typeof(IWindowsFormsEditorService));
      if (frmsvr == null) return null;
      valuesString = acop.getTargetComponent();
      TargetControlSelector page = new TargetControlSelector(acop);
      page.SetEditor(this);
      frmsvr.ShowDialog(page);
      if (valuesStale)
      {
        valuesString = acop.getTargetComponent();
        value = new TargetControlEditorValue(valuesString, acop);
      }
      return base.EditValue(context, provider, value);
    }
  }
  public class TargetControlSelector : Form
  {
    private AcopUserComponent acop = null;
    private TargetControlEditor editor = null;
    private TargetControlEditorValue editorValue = null;
    /// <summary>
    /// Required designer variable.
    /// </summary>
    private System.ComponentModel.IContainer components = null;

    /// <summary>
    /// Clean up any resources being used.
    /// </summary>
    /// <param name="disposing">true if managed resources should be disposed; otherwise, false.</param>
    protected override void Dispose(bool disposing)
    {
      if (disposing && (components != null))
      {
        components.Dispose();
      }
      base.Dispose(disposing);
    }
    public void SetEditor(TargetControlEditor editor)
    {
      this.editor = editor;
    }
    public TargetControlSelector(AcopUserComponent acop)
    {
      InitializeComponent();
      this.acop = acop;
      initSettings();
      editorValue = new TargetControlEditorValue(acop.getTargetComponent(), acop);
    }

    private Dictionary<String, Control> ctrlLst = new Dictionary<string, Control>();
    public Control[] getControls()
    {
      Control parent = acop.Parent;
      ctrlLst.Clear();
      foreach (var c in parent.Controls.OfType<Control>())
      {
        ctrlLst.Add(c.Name, c);
      }
      return ctrlLst.Values.ToArray();
    }
    public String[] getControlsByName()
    { // called at design time by chooserEditor : use (acop).Site.Name
      Control parent = acop.Parent;
      ctrlLst.Clear();
      foreach (var c in parent.Controls.OfType<Control>())
      {
        ctrlLst.Add(c.Site.Name, (Control)c);
      }
      return ctrlLst.Keys.ToArray();
    }
    private Control getControl(String name)
    {
      Control value = null;
      if (ctrlLst.TryGetValue(name, out value))
        return value;
      return null;
    }
    public PropertyInfo[] getProperties(Control control)
    {
      if (control == null) return null;
      return control.GetType().GetProperties();
    }

    private void initSettings()
    {
      // fill in controls combo
      String[] ctrls = getControlsByName();
      if (ctrls != null)
      {
        foreach (String s in ctrls)
          ControlsCombo.Items.Add(s);
        ControlsCombo.Text = (String)ControlsCombo.Items[0];
      }
      // fill in properties combo (in combo change)
      String tgt = acop.getTargetComponent();
      if (tgt != null && tgt.Length > 0)
      {
        TargetTextBox.Text = tgt;
        String[] parts = tgt.Split(new char[] { '.' }, StringSplitOptions.RemoveEmptyEntries);
        if (ControlsCombo.FindStringExact(parts[0]) != -1)
        {
          ControlsCombo.Text = parts[0];
          if (parts.Length > 1 && PropertiesCombo.Items.Contains(parts[1]))
          {
            if (PropertiesCombo.FindStringExact(parts[1]) != -1)
            {
              PropertiesCombo.Text = parts[1];
            }
          }
        }
      }
    }
    #region Windows Form Designer generated code

    /// <summary>
    /// Required method for Designer support - do not modify
    /// the contents of this method with the code editor.
    /// </summary>
    private void InitializeComponent()
    {
      this.ControlsCombo = new System.Windows.Forms.ComboBox();
      this.PropertiesCombo = new System.Windows.Forms.ComboBox();
      this.TargetLabelCaption = new System.Windows.Forms.Label();
      this.TargetTextBox = new System.Windows.Forms.TextBox();
      this.OKBtn = new System.Windows.Forms.Button();
      this.CancelBtn = new System.Windows.Forms.Button();
      this.SuspendLayout();
      // 
      // ControlsCombo
      // 
      this.ControlsCombo.DropDownStyle = System.Windows.Forms.ComboBoxStyle.DropDownList;
      this.ControlsCombo.FormattingEnabled = true;
      this.ControlsCombo.Location = new System.Drawing.Point(15, 24);
      this.ControlsCombo.Name = "ControlsCombo";
      this.ControlsCombo.Size = new System.Drawing.Size(120, 21);
      this.ControlsCombo.Sorted = true;
      this.ControlsCombo.TabIndex = 0;
      this.ControlsCombo.TextChanged += new System.EventHandler(this.TargetControlChanged);
      // 
      // PropertiesCombo
      // 
      this.PropertiesCombo.DropDownStyle = System.Windows.Forms.ComboBoxStyle.DropDownList;
      this.PropertiesCombo.FormattingEnabled = true;
      this.PropertiesCombo.Location = new System.Drawing.Point(141, 24);
      this.PropertiesCombo.Name = "PropertiesCombo";
      this.PropertiesCombo.Size = new System.Drawing.Size(120, 21);
      this.PropertiesCombo.Sorted = true;
      this.PropertiesCombo.TabIndex = 1;
      this.PropertiesCombo.TextChanged += new System.EventHandler(this.TargetPropertyChanged);
      // 
      // TargetLabelCaption
      // 
      this.TargetLabelCaption.AutoSize = true;
      this.TargetLabelCaption.Location = new System.Drawing.Point(13, 64);
      this.TargetLabelCaption.Name = "TargetLabelCaption";
      this.TargetLabelCaption.Size = new System.Drawing.Size(122, 13);
      this.TargetLabelCaption.TabIndex = 2;
      this.TargetLabelCaption.Text = "Target Control.Property :";
      // 
      // TargetTextBox
      // 
      this.TargetTextBox.Location = new System.Drawing.Point(141, 60);
      this.TargetTextBox.Name = "TargetTextBox";
      this.TargetTextBox.Size = new System.Drawing.Size(120, 20);
      this.TargetTextBox.TabIndex = 3;
      // 
      // OKBtn
      // 
      this.OKBtn.Location = new System.Drawing.Point(156, 96);
      this.OKBtn.Name = "OKBtn";
      this.OKBtn.Size = new System.Drawing.Size(49, 28);
      this.OKBtn.TabIndex = 4;
      this.OKBtn.Text = "OK";
      this.OKBtn.UseVisualStyleBackColor = true;
      this.OKBtn.Click += new System.EventHandler(this.OKClicked);
      // 
      // CancelBtn
      // 
      this.CancelBtn.Location = new System.Drawing.Point(211, 96);
      this.CancelBtn.Name = "CancelBtn";
      this.CancelBtn.Size = new System.Drawing.Size(49, 28);
      this.CancelBtn.TabIndex = 5;
      this.CancelBtn.Text = "Cancel";
      this.CancelBtn.UseVisualStyleBackColor = true;
      this.CancelBtn.Click += new System.EventHandler(this.CancelClicked);
      // 
      // TargetEditor
      // 
      this.AutoScaleDimensions = new System.Drawing.SizeF(6F, 13F);
      this.AutoScaleMode = System.Windows.Forms.AutoScaleMode.Font;
      this.ClientSize = new System.Drawing.Size(284, 138);
      this.ControlBox = false;
      this.Controls.Add(this.CancelBtn);
      this.Controls.Add(this.OKBtn);
      this.Controls.Add(this.TargetTextBox);
      this.Controls.Add(this.TargetLabelCaption);
      this.Controls.Add(this.PropertiesCombo);
      this.Controls.Add(this.ControlsCombo);
      this.Name = "TargetEditor";
      this.Text = "Target Control Editor";
      this.ResumeLayout(false);
      this.PerformLayout();
    }
    #endregion
    private System.Windows.Forms.ComboBox ControlsCombo;
    private System.Windows.Forms.ComboBox PropertiesCombo;
    private System.Windows.Forms.Label TargetLabelCaption;
    private System.Windows.Forms.TextBox TargetTextBox;
    private System.Windows.Forms.Button OKBtn;
    private System.Windows.Forms.Button CancelBtn;
    private void TargetControlChanged(object sender, EventArgs e)
    {
      Control c = getControl(ControlsCombo.Text);
      if (c == null) return;
      PropertyInfo[] prps = getProperties(c);
      PropertiesCombo.Items.Clear();
      foreach (PropertyInfo p in prps)
        PropertiesCombo.Items.Add(p.Name);
      PropertiesCombo.Text = (String)PropertiesCombo.Items[0];
    }
    private void TargetPropertyChanged(object sender, EventArgs e)
    {
      TargetTextBox.Text = ControlsCombo.Text + "." + PropertiesCombo.Text;
    }
    private void OKClicked(object sender, EventArgs e)
    {
      editorValue.Component = TargetTextBox.Text;
      editor.SetValuesStale(true);
      this.Dispose();
    }
    private void CancelClicked(object sender, EventArgs e)
    {
      editor.SetValuesStale(false);
      this.Dispose();
    }
  }

  [TypeConverter(typeof(ThresholdListEditorConverter))]
  public class ThresholdListEditorValue
  {
#   pragma warning disable 1591
    private AcopUserComponent acop = null;
    private AcopTable at = null;
    private String valuesString = null;
    private void setValuesString(String value)
    {
      valuesString = value;
      if (at != null) at.SetThresholdListString(valuesString);
    }
    public String Thresholds { get { return valuesString; } set { setValuesString(value); } }
    public ThresholdListEditorValue(String value, AcopUserComponent acop)
    {
      if (value == null) value = "";
      valuesString = value;
      if (acop != null)
      {
        this.acop = acop;
        if (acop is AcopTable) at = (AcopTable)acop;
      }
    }
    public override string ToString()
    {
      return valuesString;
    }
  }
  public class ThresholdListEditorConverter : TypeConverter
  {
#   pragma warning disable 1591
    public override object ConvertTo(
        ITypeDescriptorContext context,
        System.Globalization.CultureInfo culture,
        object value, Type destinationType)
    {
      if (destinationType == typeof(string)) return value.ToString();
      return base.ConvertTo(context, culture, value, destinationType);
    }
    public override object ConvertFrom(
        ITypeDescriptorContext context,
        System.Globalization.CultureInfo culture,
        object value)
    {
      if (value.GetType() == typeof(string))
      {
        AcopUserComponent acop = null;
        if (context is AcopUserComponent)
        {
          acop = (AcopUserComponent)context;
        }
        return new ThresholdListEditorValue((string)(value), acop);
      }
      else
      {
        return base.ConvertFrom(context, culture, value);
      }
    }
    public override bool CanConvertTo(ITypeDescriptorContext context, Type destinationType)
    {
      if (destinationType == typeof(string)) return true;
      return base.CanConvertTo(context, destinationType);
    }
    public override bool CanConvertFrom(ITypeDescriptorContext context, Type sourceType)
    {
      if (sourceType == typeof(string)) return true;
      return base.CanConvertFrom(context, sourceType);
    }
  }
  public class ThresholdListEditor : UITypeEditor
  {
#   pragma warning disable 1591
    private Boolean valuesStale = false;
    private String valuesString = "";
    public ThresholdListEditor()
    {
    }
    public ThresholdListEditor(String value)
    {
      valuesString = value;
    }
    public override string ToString()
    {
      return valuesString;
    }
    public void SetValuesStale(Boolean value) { valuesStale = value; }
    public override UITypeEditorEditStyle GetEditStyle(System.ComponentModel.ITypeDescriptorContext context)
    {
      return UITypeEditorEditStyle.Modal;
    }
    public override object EditValue(ITypeDescriptorContext context, IServiceProvider provider, object value)
    {
      AcopUserComponent acop = (AcopUserComponent)context.Instance;
      IWindowsFormsEditorService frmsvr = (IWindowsFormsEditorService)provider.GetService(typeof(IWindowsFormsEditorService));
      if (frmsvr == null) return null;
      AcopTable at = null;
      if (acop is AcopTable) at = (AcopTable)acop;
      if (at != null) valuesString = at.GetThresholdListString();
      ThresholdSelector page = new ThresholdSelector(acop);
      page.SetEditor(this);
      frmsvr.ShowDialog(page);
      if (valuesStale)
      {
        valuesString = at != null ? at.GetThresholdListString() : "";
        value = new ThresholdListEditorValue(valuesString, acop);
      }
      return base.EditValue(context, provider, value);
    }
  }
  public class ThresholdSelector : Form
  {
    private AcopUserComponent acop = null;
    private AcopTable at = null;
    private ThresholdListEditor editor = null;
    private ThresholdListEditorValue editorValue = null;
    public void SetEditor(ThresholdListEditor editor)
    {
      this.editor = editor;
    }
    /// <summary>
    /// Required designer variable.
    /// </summary>
    private System.ComponentModel.IContainer components = null;

    /// <summary>
    /// Clean up any resources being used.
    /// </summary>
    /// <param name="disposing">true if managed resources should be disposed; otherwise, false.</param>
    protected override void Dispose(bool disposing)
    {
      if (disposing && (components != null))
      {
        components.Dispose();
      }
      base.Dispose(disposing);
    }

    #region Windows Form Designer generated code

    /// <summary>
    /// Required method for Designer support - do not modify
    /// the contents of this method with the code editor.
    /// </summary>
    private void InitializeComponent()
    {
      this.components = new System.ComponentModel.Container();
      this.OKButton = new System.Windows.Forms.Button();
      this.CancelButton = new System.Windows.Forms.Button();
      this.ToolTip = new System.Windows.Forms.ToolTip(this.components);
      this.ThresholdLow = new System.Windows.Forms.TextBox();
      this.ThresholdLowLabel = new System.Windows.Forms.Label();
      this.ThresholdHighLabel = new System.Windows.Forms.Label();
      this.ThresholdHigh = new System.Windows.Forms.TextBox();
      this.ThresholdMatchLabel = new System.Windows.Forms.Label();
      this.ThresholdMatch = new System.Windows.Forms.TextBox();
      this.AddressLabel = new System.Windows.Forms.Label();
      this.AddressCombo = new System.Windows.Forms.ComboBox();
      this.ApplyButton = new System.Windows.Forms.Button();
      this.SuspendLayout();
      // 
      // OKButton
      // 
      this.OKButton.Location = new System.Drawing.Point(98, 102);
      this.OKButton.Name = "OKButton";
      this.OKButton.Size = new System.Drawing.Size(50, 24);
      this.OKButton.TabIndex = 5;
      this.OKButton.Text = "OK";
      this.OKButton.UseVisualStyleBackColor = true;
      this.OKButton.Click += new System.EventHandler(this.OKButton_Click);
      // 
      // CancelButton
      // 
      this.CancelButton.Location = new System.Drawing.Point(154, 102);
      this.CancelButton.Name = "CancelButton";
      this.CancelButton.Size = new System.Drawing.Size(50, 24);
      this.CancelButton.TabIndex = 6;
      this.CancelButton.Text = "Cancel";
      this.CancelButton.UseVisualStyleBackColor = true;
      this.CancelButton.Click += new System.EventHandler(this.CancelButton_Click);
      // 
      // ThresholdLow
      // 
      this.ThresholdLow.Location = new System.Drawing.Point(92, 40);
      this.ThresholdLow.Name = "ThresholdLow";
      this.ThresholdLow.Size = new System.Drawing.Size(69, 20);
      this.ThresholdLow.TabIndex = 7;
      // 
      // ThresholdLowLabel
      // 
      this.ThresholdLowLabel.AutoSize = true;
      this.ThresholdLowLabel.Location = new System.Drawing.Point(12, 43);
      this.ThresholdLowLabel.Name = "ThresholdLowLabel";
      this.ThresholdLowLabel.Size = new System.Drawing.Size(80, 13);
      this.ThresholdLowLabel.TabIndex = 8;
      this.ThresholdLowLabel.Text = "Threshold Low:";
      // 
      // ThresholdHighLabel
      // 
      this.ThresholdHighLabel.AutoSize = true;
      this.ThresholdHighLabel.Location = new System.Drawing.Point(167, 43);
      this.ThresholdHighLabel.Name = "ThresholdHighLabel";
      this.ThresholdHighLabel.Size = new System.Drawing.Size(32, 13);
      this.ThresholdHighLabel.TabIndex = 9;
      this.ThresholdHighLabel.Text = "High:";
      // 
      // ThresholdHigh
      // 
      this.ThresholdHigh.Location = new System.Drawing.Point(199, 40);
      this.ThresholdHigh.Name = "ThresholdHigh";
      this.ThresholdHigh.Size = new System.Drawing.Size(72, 20);
      this.ThresholdHigh.TabIndex = 10;
      // 
      // ThresholdMatchLabel
      // 
      this.ThresholdMatchLabel.AutoSize = true;
      this.ThresholdMatchLabel.Location = new System.Drawing.Point(12, 76);
      this.ThresholdMatchLabel.Name = "ThresholdMatchLabel";
      this.ThresholdMatchLabel.Size = new System.Drawing.Size(84, 13);
      this.ThresholdMatchLabel.TabIndex = 11;
      this.ThresholdMatchLabel.Text = "Matching String:";
      // 
      // ThresholdMatch
      // 
      this.ThresholdMatch.Location = new System.Drawing.Point(102, 76);
      this.ThresholdMatch.Name = "ThresholdMatch";
      this.ThresholdMatch.Size = new System.Drawing.Size(170, 20);
      this.ThresholdMatch.TabIndex = 12;
      // 
      // AddressLabel
      // 
      this.AddressLabel.AutoSize = true;
      this.AddressLabel.Location = new System.Drawing.Point(12, 9);
      this.AddressLabel.Name = "AddressLabel";
      this.AddressLabel.Size = new System.Drawing.Size(51, 13);
      this.AddressLabel.TabIndex = 13;
      this.AddressLabel.Text = "Apply to :";
      // 
      // AddressCombo
      // 
      this.AddressCombo.DropDownStyle = System.Windows.Forms.ComboBoxStyle.DropDownList;
      this.AddressCombo.FormattingEnabled = true;
      this.AddressCombo.Location = new System.Drawing.Point(69, 6);
      this.AddressCombo.Name = "AddressCombo";
      this.AddressCombo.Size = new System.Drawing.Size(203, 21);
      this.AddressCombo.TabIndex = 14;
      this.AddressCombo.SelectedIndexChanged += new System.EventHandler(this.AddressChanged);
      // 
      // ApplyButton
      // 
      this.ApplyButton.Location = new System.Drawing.Point(222, 102);
      this.ApplyButton.Name = "ApplyButton";
      this.ApplyButton.Size = new System.Drawing.Size(50, 24);
      this.ApplyButton.TabIndex = 15;
      this.ApplyButton.Text = "Apply";
      this.ApplyButton.UseVisualStyleBackColor = true;
      this.ApplyButton.Click += new System.EventHandler(this.ApplyClicked);
      // 
      // ThresholdSelector
      // 
      this.AutoScaleDimensions = new System.Drawing.SizeF(6F, 13F);
      this.AutoScaleMode = System.Windows.Forms.AutoScaleMode.Font;
      this.ClientSize = new System.Drawing.Size(284, 132);
      this.ControlBox = false;
      this.Controls.Add(this.ApplyButton);
      this.Controls.Add(this.AddressCombo);
      this.Controls.Add(this.AddressLabel);
      this.Controls.Add(this.ThresholdMatch);
      this.Controls.Add(this.ThresholdMatchLabel);
      this.Controls.Add(this.ThresholdHigh);
      this.Controls.Add(this.ThresholdHighLabel);
      this.Controls.Add(this.ThresholdLowLabel);
      this.Controls.Add(this.ThresholdLow);
      this.Controls.Add(this.CancelButton);
      this.Controls.Add(this.OKButton);
      this.Name = "ThresholdSelector";
      this.Text = "Threshold Editor";
      this.Click += new System.EventHandler(this.ApplyClicked);

      String tt = "currently assigned link addresses";
      ToolTip.SetToolTip(this.AddressCombo, tt);
      ToolTip.SetToolTip(this.AddressLabel, tt);
      tt = "apply the current threshold conditions to the selected link address";
      ToolTip.SetToolTip(this.ApplyButton, tt);
      tt = "Low warning threshold to apply to number value links";
      ToolTip.SetToolTip(this.ThresholdLow, tt);
      ToolTip.SetToolTip(this.ThresholdLowLabel, tt);
      tt = "High warning threshold to apply to number value links";
      ToolTip.SetToolTip(this.ThresholdHigh, tt);
      ToolTip.SetToolTip(this.ThresholdHighLabel, tt);
      tt = "regular expression (value must contain) to use for (high) warning for sting value links";
      ToolTip.SetToolTip(this.ThresholdMatch, tt);
      ToolTip.SetToolTip(this.ThresholdMatchLabel, tt);

      this.ResumeLayout(false);
      this.PerformLayout();
    }

    #endregion

    private System.Windows.Forms.Button OKButton;
    private new System.Windows.Forms.Button CancelButton;
    private System.Windows.Forms.ToolTip ToolTip;
    private System.Windows.Forms.TextBox ThresholdLow;
    private System.Windows.Forms.Label ThresholdLowLabel;
    private System.Windows.Forms.Label ThresholdHighLabel;
    private System.Windows.Forms.TextBox ThresholdHigh;
    private System.Windows.Forms.Label ThresholdMatchLabel;
    private System.Windows.Forms.TextBox ThresholdMatch;
    private System.Windows.Forms.Label AddressLabel;
    private System.Windows.Forms.ComboBox AddressCombo;
    private System.Windows.Forms.Button ApplyButton;
    public ThresholdSelector(AcopUserComponent acop)
    { // for now: only support AcopTable ...
      InitializeComponent();
      this.acop = acop;
      if (acop is AcopTable) at = (AcopTable)acop;
      initSettings();
      if (at != null) editorValue = new ThresholdListEditorValue(at.GetThresholdListString(), acop);
    }
    private void OKButton_Click(object sender, EventArgs e)
    {
      // convert contents of ThresholdList to info for property grid setting
      StringBuilder sb = new StringBuilder();
      foreach (String s in thresholdArray)
      {
        sb.Append(s + ";");
      }
      thresholdList = sb.ToString();
      if (at != null) at.SetThresholdListString(thresholdList);
      editor.SetValuesStale(true);
      this.Dispose();
    }
    private void CancelButton_Click(object sender, EventArgs e)
    {
      editor.SetValuesStale(false);
      this.Dispose();
    }
    private String thresholdList = "";
    private String[] thresholdArray;
    private void initSettings()
    {
      // set thresholdList from control setting
      if (at != null) thresholdList = at.GetThresholdListString();
      thresholdArray = thresholdList.Split(new char[] { ';' });
      // populate the address combo with LinkAddress + LinkProperty
      if (acop != null)
      {
        AcopUserComponent.LinkGroupListItem[] lnks = acop.GetAutoGroupLinks();
        foreach(AcopUserComponent.LinkGroupListItem lnk in lnks)
        {
          AddressCombo.Items.Add(lnk.GetAddress() + " " + lnk.GetProperty());
        }
      }
      if (AddressCombo.Items.Count == 0) return;
      AddressCombo.Text = AddressCombo.Items[0] as String;
      // can we know if pointing to a number or string ?

    }
    private void ApplyClicked(object sender, EventArgs e)
    {
      int i = AddressCombo.SelectedIndex;
      int len = thresholdArray != null ? thresholdArray.Length : 0;
      if (i >= len)
      {
        Array.Resize<String>(ref thresholdArray, i + 1);
        for (int k = len; k < thresholdArray.Length; k++)
          thresholdArray[k] = "";
      }
      String threshold = ThresholdMatch.Text;
      if (threshold.Length == 0)
        threshold = ThresholdLow.Text.Trim() + ":" + ThresholdHigh.Text.Trim();
      thresholdArray[i] = threshold;
    }

    private void AddressChanged(object sender, EventArgs e)
    {
      int i = AddressCombo.SelectedIndex;
      String threshold = i < thresholdArray.Length ? thresholdArray[i] : "";
      if (threshold.Contains(':'))
      {
        String[] parts = threshold.Split(new char[] { ':' }, StringSplitOptions.RemoveEmptyEntries);
        ThresholdLow.Text = parts[0];
        ThresholdHigh.Text = parts[1];
      }
      else
      {
        ThresholdMatch.Text = threshold;
        ThresholdLow.Text = "";
        ThresholdHigh.Text = "";
      }
    }
  }
}