﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Drawing.Imaging;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace Acop
{
  [ToolboxItem(true)]
  [ToolboxBitmap(typeof(PictureBox))]
  public partial class AcopVideo : AcopUserComponent
  {
#   pragma warning disable 1591
    private PictureBox imageBox = new PictureBox();
    private Panel xPanel = new Panel();
    private Panel yPanel = new Panel();
    private Label msgLabel = new Label();
    private FrmProperties prop;

    private ContextMenu cm = new ContextMenu();
    private MenuItem miPref = new MenuItem("Preferences");

    private int lastLinkStatus = 0;
    private Boolean keepARatio = false;
    private Boolean isZoomed = false;
    private Boolean mouseMoved = false;
    private Rectangle mouseRect = new Rectangle(0, 0, 0, 0);
    private Rectangle posRect = new Rectangle(0, 0, 0, 0);
    private Size imageSize = new Size(0, 0);
    private Boolean sGrid = false;
    private int gSize = 30;
    private Color gColor = Color.LightGray;
    private Boolean iAnalysis = false;

    private Pen grayPen = new Pen(Color.Gray, 0.5f);
    private Pen redPen = new Pen(Color.Red, 0.5f);

    public enum colorMapEnum
    {
      NONE,
      GRAYSCALE,
//      PITZ,
      JET
    }
    private colorMapEnum cMap = colorMapEnum.NONE;

    public AcopVideo()
    {
      InitializeComponent();
      SetDefaultUpdater(DefaultUpdate);
      setControlsList(new Control[] { imageBox }, true);
    }

    private void InitializeComponent()
    {
      this.Size = new Size(200, 200);
      this.MinimumSize = new Size(50, 50);

      imageBox.Size = new Size(200, 200);
      imageBox.Location = new Point(0, 0);
      imageBox.Dock = DockStyle.Fill;
      imageBox.SizeMode = PictureBoxSizeMode.StretchImage;
      imageBox.Paint += new PaintEventHandler(Painting);
      imageBox.MouseDown += new MouseEventHandler(MDown);
      imageBox.MouseUp += new MouseEventHandler(MUp);
      imageBox.MouseMove += new MouseEventHandler(MMove);
      imageBox.MouseDoubleClick += new MouseEventHandler(MDoubleClick);

      xPanel.BackColor = Color.White;
      yPanel.BackColor = Color.White;

      miPref.Click += new EventHandler(this.ApplyPrefClicked);
      cm.MenuItems.Add(miPref);

      imageBox.ContextMenu = cm;
      this.Controls.Add(imageBox);
    }

    private void ApplyPrefClicked(object sender, EventArgs e)
    {
      prop = new FrmProperties(this);
      prop.ShowDialog();
    }

    private void SetKeepARatio(Boolean value)
    {
      if (keepARatio == value) return;
      keepARatio = value;
      if(keepARatio)
      {
        imageBox.SizeMode = PictureBoxSizeMode.Zoom;
      }
      else
      {
        imageBox.SizeMode = PictureBoxSizeMode.StretchImage;
      }
    }

    private void SetShowGrid(Boolean value)
    {
      if (sGrid == value) return;
      sGrid = value;
      this.Refresh();
    }
    private void SetGridSize(int value)
    {
      if (gSize == value) return;
      gSize = value;
    }
    private void SetGridColor(Color value)
    {
      if (gColor == value) return;
      gColor = value;
    }

    private void SetColorMap(colorMapEnum value)
    {
      if (cMap == value) return;
      cMap = value;
    }

    private void SetAnalysis(Boolean value)
    {
      iAnalysis = value;
      if (iAnalysis)
      {
        double xSize = this.Width * 0.2;
        double ySize = this.Height * 0.2;
        imageBox.Size = new Size(this.Width - (int)xSize, this.Height - (int)ySize);
        yPanel.Location = new Point(this.Width - (int)xSize, 0);
        yPanel.Size = new Size((int)xSize, this.Height - (int)ySize);
        xPanel.Location = new Point(0, this.Height - (int)ySize);
        xPanel.Size = new Size(this.Width - (int)xSize, (int)ySize);

        imageBox.Dock = DockStyle.None;
        this.Controls.Add(yPanel);
        this.Controls.Add(xPanel);
      }
      else
      {
        this.Controls.Remove(yPanel);
        this.Controls.Remove(xPanel);
        imageBox.Dock = DockStyle.Fill;
      }
    }

    //protected override void OnResize(EventArgs e)
    //{
    //  if (iAnalysis)
    //  {
    //    double xSize = this.Width * 0.2;
    //    double ySize = this.Height * 0.2;
    //    imageBox.Size = new Size(this.Width - (int)xSize, this.Height - (int)ySize);
    //    yPanel.Location = new Point(this.Width - (int)xSize, 0);
    //    yPanel.Size = new Size((int)xSize, this.Height - (int)ySize);
    //    xPanel.Location = new Point(0, this.Height - (int)ySize);
    //    xPanel.Size = new Size(this.Width - (int)xSize, (int)ySize);

    //    imageBox.Dock = DockStyle.None;
    //  }
    //  else
    //  {
    //    imageBox.Dock = DockStyle.Fill;
    //  }
    //}

    private void DefaultUpdate(AcopControl.AcopLink lnk)
    {
      if (getAutoUpdate())
      {
        int lnkSts = lnk.GetStatus();

        if (lnkSts != 0 && ErrorValue.Length == 0)
        {
          //applyErrorStatus(lnk.GetStatusString(), lnkSts != lastLinkStatus);
        }
        else
        {
          if (lnkSts != lastLinkStatus) applyToolTipText(null);

          String header = null;
          Bitmap oImage = getDataAsImage(lnk, ref header);
          imageSize.Height = oImage.Height;
          imageSize.Width = oImage.Width;

          if (cMap != colorMapEnum.NONE)
          {            
            switch (cMap)
            {
              case colorMapEnum.GRAYSCALE:
                Bitmap tmpBitmap = new Bitmap(oImage.Width, oImage.Height);
                Graphics g = Graphics.FromImage(tmpBitmap);
                ImageAttributes iAttributes = new ImageAttributes();
                ColorMatrix cMatrix = new ColorMatrix(
                  new float[][]
                  {
                    new float[] {.29f, .29f, .29f, 0, 0},
                    new float[] {.59f, .59f, .59f, 0, 0},
                    new float[] {.11f, .11f, .11f, 0, 0},
                    new float[] {0, 0, 0, 1, 0},
                    new float[] {0, 0, 0, 0, 1}
                  });
                iAttributes.SetColorMatrix(cMatrix);
                g.DrawImage(oImage, new Rectangle(0, 0, oImage.Width, oImage.Height), 0, 0, oImage.Width, oImage.Height, GraphicsUnit.Pixel, iAttributes);
                g.Dispose();
                oImage = tmpBitmap;
                break;
//              case colorMapEnum.PITZ:
//                break;
              case colorMapEnum.JET:
                ColorPalette cp = oImage.Palette;
                Color[] c = cp.Entries;
                for (int i = 0; i < 256; i++)
                {
                  if(i < 32)
                  {
                    c[i] = Color.FromArgb((byte)0, (byte)0, (byte)(127 + i * 4));
                  }
                  if(i >= 32 && i < 96)
                  {
                    c[i] = Color.FromArgb((byte)0, (byte)(3 + (i - 32) * 4), (byte)255);
                  }
                  if(i >= 96 && i < 160)
                  {
                    c[i] = Color.FromArgb((byte)(3 + (i - 96) * 4), (byte)255, (byte)(255 - (i - 96) * 4));
                  }
                  if(i >= 160 && i < 224)
                  {
                    c[i] = Color.FromArgb((byte)255, (byte)(255 - (i - 160) * 4), (byte)0);
                  }
                  if(i >= 224)
                  {
                    c[i] = Color.FromArgb((byte)(255 - (i - 224) * 4), (byte)0, (byte)0);
                  }
                }
                oImage.Palette = cp;
                break;
            }         
          }
          if (isZoomed)
          {
            imageBox.Image = ZoomImage(oImage);
          }
          else
          {
            imageBox.Image = oImage;
          }
          imageBox.Refresh();

          if (lnk.GetDisplayId() < 0)
          {
            lnk.SetDisplayId(1);
          }
        }
      }
    }

    private Bitmap ZoomImage(Bitmap img)
    {
      float factor_X = (float)imageSize.Width / (float)imageBox.Width;
      float factor_Y = (float)imageSize.Height / (float)imageBox.Height;
      Rectangle zoomImgRect = new Rectangle((int)(posRect.X * factor_X), (int)(posRect.Y * factor_Y), (int)(posRect.Width * factor_X), (int)(posRect.Height * factor_Y));
      Bitmap zoomBitmap = new Bitmap(img.Width, img.Height);
      Rectangle dstRect = new Rectangle(0, 0, zoomBitmap.Width, zoomBitmap.Height);
      Graphics g = Graphics.FromImage(zoomBitmap);
      g.InterpolationMode = System.Drawing.Drawing2D.InterpolationMode.NearestNeighbor;
      g.DrawImage(img, dstRect, zoomImgRect, GraphicsUnit.Pixel);
      g.Dispose();
      return zoomBitmap;
    }

    //private void AnalysisImage(Bitmap image)
    //{
    //  BitmapData bData = image.LockBits(new Rectangle(0, 0, image.Width, image.Height), ImageLockMode.ReadOnly, image.PixelFormat);
    //  int bytesPerPixel = Bitmap.GetPixelFormatSize(image.PixelFormat) / 8;
    //  int byteCount = bData.Stride * image.Height;
    //  byte[] pixels = new byte[byteCount];
    //  IntPtr ptrFirstPixel = bData.Scan0;
    //  System.Runtime.InteropServices.Marshal.Copy(ptrFirstPixel, pixels, 0, pixels.Length);
    //  int pixelHeight = bData.Height;
    //  int pixelWidth = bData.Height * bytesPerPixel;
    //  double[] sideViewHeight = new double[pixelHeight];
    //  double[] sideViewWidth = new double[pixelWidth];
    //}

    private void MDown(object sender, MouseEventArgs e)
    {
      if (e.Button != MouseButtons.Left || isZoomed) return;
      mouseRect.X = e.X;
      mouseRect.Y = e.Y;
    }
    private void MUp(object sender, MouseEventArgs e)
    {
      if (e.Button != MouseButtons.Left && !mouseMoved) return;
      if (mouseRect.Width == 0 || mouseRect.Height == 0) return;
      isZoomed = true;
      mouseMoved = false;
    }
    private void MMove(object sender, MouseEventArgs e)
    {
      if (e.Button != MouseButtons.Left || isZoomed) return;
      if (mouseRect.X != e.X || mouseRect.Y != e.Y)
      {
        mouseMoved = true;
        mouseRect.Width = e.X - mouseRect.X;
        mouseRect.Height = e.Y - mouseRect.Y;
        Refresh();
      }
    }
    private void MDoubleClick(object sender, MouseEventArgs e)
    {
      if (e.Button != MouseButtons.Left) return;
      isZoomed = false;
      mouseRect.X = 0;
      mouseRect.Y = 0;
      mouseRect.Width = 0;
      mouseRect.Height = 0;
    }

    private void Painting(object sender, PaintEventArgs e)
    {
      Graphics g = e.Graphics;
      Rectangle rect = new Rectangle(0, 0, imageBox.Width - 1, imageBox.Height - 1);
      if (sGrid)
      {
        Pen gridPen = new Pen(gColor, 0.5f);
        int x = imageBox.Width;
        int y = imageBox.Height;
        for (int i = 0; i < imageBox.Height; i += gSize)
        {
          g.DrawLine(gridPen, new Point(0, i), new Point(x, i));
        }
        for (int j = 0; j < imageBox.Width; j += gSize)
        {
          g.DrawLine(gridPen, new Point(j, 0), new Point(j, y));
        }
      }
      if (mouseMoved)
      {
        if (mouseRect.Width > 0)
        {
          posRect.X = mouseRect.X;
          posRect.Width = mouseRect.Width;
        }
        else
        {
          posRect.X = mouseRect.X + mouseRect.Width;
          posRect.Width = -mouseRect.Width;
        }
        if (mouseRect.Height > 0)
        {
          posRect.Y = mouseRect.Y;
          posRect.Height = mouseRect.Height;
        }
        else
        {
          posRect.Y = mouseRect.Y + mouseRect.Height;
          posRect.Height = -mouseRect.Height;
        }

        g.DrawRectangle(redPen, posRect);
      }
      g.DrawRectangle(grayPen, rect);
    }

    #region Properties
    //[Category("Acop.Video")]
    //[DefaultValue(false)]
    //public Boolean ShowAnalysis { get { return iAnalysis; } set { SetAnalysis(value); } }
    [Category("Acop.Video")]
    [Description("Keep aspect ratio")]
    [DefaultValue(false)]
    public Boolean KeepAspectRatio { get { return keepARatio; } set { SetKeepARatio(value); } }
    [Category("Acop.Video")]
    [Description("Show grid on image")]
    [DefaultValue(false)]
    public Boolean GridVisible { get { return sGrid; } set { SetShowGrid(value); } }
    [Category("Acop.Video")]
    [Description("The horizontal and vertical grid spacing")]
    [DefaultValue(30)]
    public int GridSize { get { return gSize; } set { SetGridSize(value); } }
    [Category("Acop.Video")]
    [Description("The grid color")]
    [DefaultValue(typeof(Color), "0xD3D3D3")]
    public Color GridColor { get { return gColor; } set { SetGridColor(value); } }
    [Category("Acop.Video")]
    [Description("The color palette used by the image")]
    [DefaultValue(colorMapEnum.NONE)]
    public colorMapEnum ColorMap { get { return cMap; } set { SetColorMap(value); } }

    [Category("Acop.Display")]
    [Browsable(false), EditorBrowsable(EditorBrowsableState.Never)]
    public new int Precision { get; set; }
    [Category("Acop.Display")]
    [Browsable(false), EditorBrowsable(EditorBrowsableState.Never)]
    public new String Description { get; set; }
    [Category("Acop.Display")]
    [Browsable(false), EditorBrowsable(EditorBrowsableState.Never)]
    public new String Units { get; set; }
    [Category("Acop.Display")]
    [Browsable(false), EditorBrowsable(EditorBrowsableState.Never)]
    public new Boolean DisplayAsMultiChannel { get; set; }
    [Category("Acop.Transport")]
    [Browsable(false), EditorBrowsable(EditorBrowsableState.Never)]
    public new Boolean LinkGrouped { get; set; }
    [Category("Acop.Transport")]
    [Browsable(false), EditorBrowsable(EditorBrowsableState.Never)]
    public new ApplyEventSettingType ApplyEventOrder { get; set; }
    #endregion
  }

  public partial class FrmProperties : Form
  {
    private AcopVideo vid;

    public FrmProperties(AcopVideo vid)
    {
      this.vid = vid;
      InitializeComponent();
      propertyGrid1.SelectedObject = vid;
      propertyGrid1.BrowsableAttributes = new AttributeCollection(new CategoryAttribute("Acop.Video"));
    }

    private void btn_Cancel_Click(object sender, EventArgs e)
    {
      this.Dispose();
    }

    private void InitializeComponent()
    {
      this.propertyGrid1 = new System.Windows.Forms.PropertyGrid();
      this.btn_Cancel = new System.Windows.Forms.Button();
      this.SuspendLayout();
      // 
      // propertyGrid1
      // 
      this.propertyGrid1.LineColor = System.Drawing.SystemColors.ControlDark;
      this.propertyGrid1.Location = new System.Drawing.Point(3, 12);
      this.propertyGrid1.Name = "propertyGrid1";
      this.propertyGrid1.PropertySort = System.Windows.Forms.PropertySort.Alphabetical;
      this.propertyGrid1.Size = new System.Drawing.Size(285, 576);
      this.propertyGrid1.TabIndex = 0;
      this.propertyGrid1.ToolbarVisible = false;
      // 
      // btn_Cancel
      // 
      this.btn_Cancel.Location = new System.Drawing.Point(206, 602);
      this.btn_Cancel.Name = "btn_Cancel";
      this.btn_Cancel.Size = new System.Drawing.Size(75, 23);
      this.btn_Cancel.TabIndex = 2;
      this.btn_Cancel.Text = "Close";
      this.btn_Cancel.UseVisualStyleBackColor = true;
      this.btn_Cancel.Click += new EventHandler(this.btn_Cancel_Click);
      // 
      // FrmProperties
      // 
      this.AutoScaleDimensions = new System.Drawing.SizeF(6F, 13F);
      this.AutoScaleMode = System.Windows.Forms.AutoScaleMode.Font;
      this.ClientSize = new System.Drawing.Size(293, 637);
      this.ControlBox = false;
      this.Controls.Add(this.btn_Cancel);
      this.Controls.Add(this.propertyGrid1);
      this.Name = "FrmProperties";
      this.Text = "Preferences";
      this.ResumeLayout(false);
    }
    private System.Windows.Forms.PropertyGrid propertyGrid1;
    private System.Windows.Forms.Button btn_Cancel;
  }
}
