﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Timers;
using System.Windows.Forms;
using Acop.AcopWheelSwitchUtil;

namespace Acop
{
  [ToolboxItem(true)]
  [ToolboxBitmap(typeof(NumericUpDown))]
  public partial class AcopWheelSwitch : AcopUserComponent
  {
#   pragma warning disable 1591
    private AcopWheelSwitchContextMenu cm;

    private Label descLabel = new Label();
    private Label barLine = new Label();
    private Label rbvLabel = new Label();
    private Label rbvValueLabel = new Label();
    private Label refLabel = new Label();
    private Label refValueLabel = new Label();
    private Label unitLabel = new Label();

    private Panel headerPanel = new Panel();
    private Panel digitsPanel = new Panel();

    private CheckBox lockCheck = new CheckBox();
    private CheckBox continuousCheck = new CheckBox();

    private Button setBtn = new Button();
    private Button refBtn = new Button();
    private Button resetBtn = new Button();

    private System.Timers.Timer holdTimer = new System.Timers.Timer();

    private double thresholdHigh = MAX_THRESHOLD;
    private double thresholdLow = -MAX_THRESHOLD;

    private const int NORMAL_DIGIT_X_SIZE = 15;
    private const int NORMAL_DIGIT_Y_SIZE = 30;
    private const int NORMAL_DIGIT_Y_POS = 5;
    private const int NORMAL_BTN_X_SIZE = 12;
    private const int NORMAL_BTN_Y_SIZE = 20;
    private const int NORMAL_DIGITS_FONTSIZE = 16;
    private const int NORMAL_UNITLABEL_FONTSIZE = 12;
    private const int SMALLER_DIGIT_X_SIZE = 10;
    private const int SMALLER_DIGIT_Y_SIZE = 18;
    private const int SMALLER_DIGIT_Y_POS = 6;
    private const int SMALLER_BTN_X_SIZE = 10;
    private const int SMALLER_BTN_Y_SIZE = 15;
    private const int SMALLER_DIGITS_FONTSIZE = 9;
    private const int SMALLER_UNITLABEL_FONTSIZE = 8;
    private const int LARGER_DIGIT_X_SIZE = 20;
    private const int LARGER_DIGIT_Y_SIZE = 40;
    private const int LARGER_DIGIT_X_POS = 4;
    private const int LARGER_BTN_X_SIZE = 16;
    private const int LARGER_BTN_Y_SIZE = 24;
    private const int LARGER_DIGITS_FONTSIZE = 22;
    private const int LARGER_UNITLABEL_FONTSIZE = 18;
    private const int LARGEST_DIGIT_X_SIZE = 25;
    private const int LARGEST_DIGIT_Y_SIZE = 50;
    private const int LARGEST_DIGIT_Y_POS = 3;
    private const int LARGEST_BTN_X_SIZE = 20;
    private const int LARGEST_BTN_Y_SIZE = 28;
    private const int LARGEST_DIGITS_FONTSIZE = 30;
    private const int LARGEST_UNITLABEL_FONTSIZE = 26;

    private int digits = 4;
    private int fractionalDigits = 1;
    private int digitBtnGap = 5;
    private int digitXSize = NORMAL_DIGIT_X_SIZE;
    private int digitYSize = NORMAL_DIGIT_Y_SIZE;
    private int digitYPos = NORMAL_DIGIT_Y_POS;
    private int btnXSize = NORMAL_BTN_X_SIZE;
    private int btnYSize = NORMAL_BTN_Y_SIZE;
    private int digitsFontSize = NORMAL_DIGITS_FONTSIZE;
    private int unitLabelFontSize = NORMAL_UNITLABEL_FONTSIZE;

    private int xLocation = 0;
    private int lastLinkStatus = 0;

    private int deadtime = 50;

    private Boolean isNegValue = false;
    private Boolean showHeader = true;
    private Boolean rbvStartValue = true;
    private Boolean isDownButton = false;
    private Boolean hasLnkError = false;
    private Boolean applyContinuous = true;
    private Boolean hasRangeError = false;

    private StringBuilder sbFormat = new StringBuilder();

    private WsBtn upBtn = new WsBtn(false);
    private WsBtn downBtn = new WsBtn(true);

    private List<WsLabel> digitsList = new List<WsLabel>();
    private WsLabel signLb = new WsLabel();
    private WsLabel pointLb = new WsLabel();

    private Color labelInSyncForeColor = Color.DarkBlue;
    private Color labelOutOfSyncForeColor = Color.DarkRed;
    private Color readbackForeColor = Color.Black;
    private Color referenceForeColor = Color.Black;

    private delegate void DigitsUpdateDelegate(double value);
    private delegate void TimerUpdateDelegate(object sender, ElapsedEventArgs e);

    public enum wheelDigitSizeEnum
    {
      Smaller,
      Normal,
      Larger,
      Largest
    }
    private wheelDigitSizeEnum digitSize = wheelDigitSizeEnum.Normal;

    public AcopWheelSwitch()
    {
      InitializeComponent();
      SetDefaultUpdater(defaultUpdate);
      SetDefaultApplier(defaultApplier);
      allowDragAndDrop(new Control[] { this, rbvValueLabel, descLabel });
      setControlsList(new Control[] { rbvValueLabel, refValueLabel, descLabel, unitLabel }, true);
      applyToolTipText(continuousCheck, "if checked: send setting with each value change (independent of 'set' button)");
      applyToolTipText(lockCheck, "if checked: apply an exclusive (only this application is allowed) access lock to the set value");
      applyToolTipText(rbvLabel, "the current readback value");
      applyToolTipText(setBtn, "send the set value to the target");
      applyToolTipText(refBtn, "set the reference value to the current wheelswitch value");
      applyToolTipText(resetBtn, "send the reference value to the target");
      holdTimer.Interval = deadtime;
      holdTimer.Enabled = false;
      holdTimer.Elapsed += new ElapsedEventHandler(this.holdTimerEvent);
      holdTimer.AutoReset = true;
      SetLateInitializerProlog(initTransportInfo);
      StartLateInitialization();
      cm = new AcopWheelSwitchContextMenu(this);
      this.ContextMenu = cm;
    }
    private void handleMouseDown(MouseEventArgs e)
    {
      if (showHeader) return;
      if (e.Button == MouseButtons.Right)
      {
        Point pos = this.PointToClient(Cursor.Position);
        cm.Show(this, pos, LeftRightAlignment.Right);
      }
    }
    private void WheelSwitchMouseDown(object sender, MouseEventArgs e)
    {
      handleMouseDown(e);
    }

    private void InitializeComponent()
    {
      this.Size = new Size(160, 160);
      this.MinimumSize = new Size(160, 160);

      headerPanel.Size = new Size(this.Size.Width, 100);
      headerPanel.Location = new Point(0, 0);
      headerPanel.Anchor = (AnchorStyles.Left | AnchorStyles.Top | AnchorStyles.Right);

      digitsPanel.Size = new Size(this.Size.Width, 60);
      digitsPanel.Location = new Point(0, 100);
      digitsPanel.Anchor = (AnchorStyles.Left | AnchorStyles.Top | AnchorStyles.Right);

      descLabel.Location = new Point(1, 5);
      descLabel.Font = new Font(FontFamily.GenericSansSerif, 8.25f, FontStyle.Bold);
      descLabel.Text = "description";
      descLabel.AutoSize = true;
      descLabel.Anchor = (AnchorStyles.Left | AnchorStyles.Top | AnchorStyles.Right);

      barLine.Location = new Point(1, 20);
      barLine.AutoSize = true;
      barLine.Font = new Font(FontFamily.GenericSansSerif, 1F, FontStyle.Regular, GraphicsUnit.Point, ((byte)(0)));
      barLine.Size = new Size(121, 2);
      barLine.Text = "-----------------------------------------------------------------------------------------------------------------------";

      rbvLabel.Location = new Point(1, 25);
      rbvLabel.Font = new Font(FontFamily.GenericSansSerif, 8.25f, FontStyle.Italic);
      rbvLabel.Text = "rbv:";
      rbvLabel.AutoSize = true;
      rbvLabel.Anchor = AnchorStyles.Top;

      rbvValueLabel.Location = new Point(26, 25);
      rbvValueLabel.Text = "0000.0";
      rbvValueLabel.ForeColor = readbackForeColor;
      rbvValueLabel.AutoSize = true;
      rbvValueLabel.Anchor = AnchorStyles.Top;

      refLabel.Location = new Point(81, 25);
      refLabel.Text = "ref:";
      refLabel.Font = new Font(FontFamily.GenericSansSerif, 8.25f, FontStyle.Italic);
      refLabel.AutoSize = true;
      refLabel.Anchor = AnchorStyles.Top;

      refValueLabel.Location = new Point(106, 25);
      refValueLabel.Text = "0000.0";
      refValueLabel.AutoSize = true;
      refValueLabel.Anchor = AnchorStyles.Top;

      lockCheck.Location = new Point(50, 40);
      lockCheck.Text = "lock access";
      lockCheck.Anchor = AnchorStyles.Top;
      lockCheck.MouseClick += new MouseEventHandler(this.applyLockClick);

      continuousCheck.Location = new Point(1, 40);
      continuousCheck.Text = "cont.";
      continuousCheck.Checked = true;
      continuousCheck.Anchor = AnchorStyles.Top;
      continuousCheck.MouseClick += new MouseEventHandler(this.applyContinousClick);

      setBtn.Location = new Point(1, 65);
      setBtn.Text = "set";
      setBtn.Size = new Size(40, 20);
      setBtn.Anchor = AnchorStyles.Top;
      setBtn.MouseClick += new MouseEventHandler(this.applySetClick);

      refBtn.Location = new Point(42, 65);
      refBtn.Text = "ref";
      refBtn.Size = new Size(40, 20);
      refBtn.Anchor = AnchorStyles.Top;
      refBtn.MouseClick += new MouseEventHandler(this.applySaveRefClick);

      resetBtn.Location = new Point(84, 65);
      resetBtn.Text = "reset";
      resetBtn.Size = new Size(40, 20);
      resetBtn.Anchor = AnchorStyles.Top;
      resetBtn.MouseClick += new MouseEventHandler(this.applyResetClick);

      unitLabel.Text = "unit";
      unitLabel.AutoSize = true;
      unitLabel.Anchor = (AnchorStyles.Left | AnchorStyles.Top | AnchorStyles.Right);
      unitLabel.Font = new Font(FontFamily.GenericSansSerif, unitLabelFontSize, FontStyle.Regular);
      unitLabel.TextAlign = ContentAlignment.MiddleCenter;

      upBtn.MouseDown += new MouseEventHandler(this.AcopButtonUpMouseDown);
      upBtn.MouseUp += new MouseEventHandler(this.AcopButtonMouseUp);

      downBtn.MouseDown += new MouseEventHandler(this.AcopButtonDownMouseDown);
      downBtn.MouseUp += new MouseEventHandler(this.AcopButtonMouseUp);

      headerPanel.Controls.Add(descLabel);
      headerPanel.Controls.Add(barLine);
      headerPanel.Controls.Add(rbvLabel);
      headerPanel.Controls.Add(rbvValueLabel);
      headerPanel.Controls.Add(refLabel);
      headerPanel.Controls.Add(refValueLabel);
      headerPanel.Controls.Add(lockCheck);
      headerPanel.Controls.Add(continuousCheck);
      headerPanel.Controls.Add(setBtn);
      headerPanel.Controls.Add(refBtn);
      headerPanel.Controls.Add(resetBtn);

      this.Controls.Add(headerPanel);
      this.Controls.Add(digitsPanel);

      this.MouseDown += new MouseEventHandler(this.WheelSwitchMouseDown);

      setDigitLabels();
      addDynamicComp();
    }

    private void setWheelSwitch()
    {
      removeDynamicComp();
      setDigitLabels();
      setDigitsFontSize();
      addDynamicComp();
    }

    private void setDigitLabels()
    {
      signLb.Size = new Size(digitXSize, digitYSize);
      signLb.Location = new Point(0, digitYPos);
      signLb.Text = "+";
      xLocation += digitXSize;
      sbFormat.Append("#");
      signLb.Visible = false;

      for (int i=0; i<digits; i++)
      {
        WsLabel lb = new WsLabel();
        lb.Location = new Point(xLocation, digitYPos);
        lb.Size = new Size(digitXSize, digitYSize);
        lb.Text = "0";
        lb.MouseClick += new MouseEventHandler(this.AcopMouseClick);
        xLocation += digitXSize;
        digitsList.Add(lb);
        if (i == 0)
        {
          lb.FirstDigit = true;
          lb.PrevDigit = signLb;
        }
        else lb.PrevDigit = digitsList[i - 1];
        sbFormat.Append("0");
      }
      if(fractionalDigits > 0)
      {
        pointLb.Size = new Size(digitXSize, digitYSize);
        pointLb.Location = new Point(xLocation, digitYPos);
        pointLb.Text = ".";
        xLocation += digitXSize;
        sbFormat.Append(".");
        for (int i=digits; i<digits+fractionalDigits; i++)
        {
          WsLabel lb = new WsLabel();
          lb.Location = new Point(xLocation, digitYPos);
          lb.Size = new Size(digitXSize, digitYSize);
          lb.Text = "0";
          lb.MouseClick += new MouseEventHandler(this.AcopMouseClick);
          xLocation += digitXSize;
          digitsList.Add(lb);
          lb.PrevDigit = digitsList[i - 1];
          sbFormat.Append("0");
        }
      }
    }

    private void addDynamicComp()
    {
      upBtn.Location = new Point(xLocation+digitBtnGap, 0);
      upBtn.Size = new Size(btnXSize, btnYSize);
      upBtn.setEnabled(false);

      downBtn.Location = new Point(xLocation + digitBtnGap, btnYSize + 1);
      downBtn.Size = new Size(btnXSize, btnYSize);
      downBtn.setEnabled(false);

      unitLabel.Font = new Font(FontFamily.GenericSansSerif, unitLabelFontSize, FontStyle.Regular);
      unitLabel.Location = new Point(xLocation + digitBtnGap + btnXSize + 2, 7);

      digitsPanel.Controls.Add(signLb);
      if(fractionalDigits > 0) digitsPanel.Controls.Add(pointLb);
      digitsPanel.Controls.Add(upBtn);
      digitsPanel.Controls.Add(downBtn);
      digitsPanel.Controls.Add(unitLabel);

      foreach (WsLabel lb in digitsList) digitsPanel.Controls.Add(lb);
    }

    private void removeDynamicComp()
    {
      digitsPanel.Controls.Remove(signLb);
      digitsPanel.Controls.Remove(pointLb);
      digitsPanel.Controls.Remove(upBtn);
      digitsPanel.Controls.Remove(downBtn);
      digitsPanel.Controls.Remove(unitLabel);

      foreach (WsLabel lb in digitsList)
      {
        digitsPanel.Controls.Remove(lb);
        lb.Dispose();
      }
      digitsList.Clear();
      sbFormat.Clear();
      xLocation = 0;
    }

    private void clearSelection()
    {
      if (digitsList.Count > 0)
      {
        foreach (WsLabel lb in digitsList) lb.Selected = false;
      }
    }

    private void setDigitValue(double value)
    {
      String valueAsString = value.ToString(sbFormat.ToString());
      int i = 0;

      if(this.InvokeRequired)
      {
        Invoke(new DigitsUpdateDelegate(setDigitValue), value);
        return;
      }
      
      foreach (char c in valueAsString)
      {
        try
        {
          if (hasRangeError)
          {
            digitsList.ForEach(AddMouseEvent);
            hasRangeError = false;
          }
          if (c == '.') continue;
          if (c == '+' || c == '-')
          {
            if (c == '-')
            {
              isNegValue = true;
              signLb.Visible = true;
            }
            else
            {
              signLb.Visible = false;
            }
            signLb.Text = c.ToString();
          }
          else
          {
            digitsList[i].Text = c.ToString();
            i++;
          }
        }
        catch(Exception ex)
        {
          digitsList.ForEach(SetErrorStatus);
          hasRangeError = true;
        }
      }
    }
    private void SetErrorStatus(WsLabel lb)
    {
      lb.Text = "x";
      lb.Selected = false;
      lb.MouseClick -= this.AcopMouseClick;
    }
    private void AddMouseEvent(WsLabel lb)
    {
      lb.MouseClick += new MouseEventHandler(this.AcopMouseClick);
    }
    private double getDigitValue()
    {
      StringBuilder tmp = new StringBuilder();
      double value = 0;

      tmp.Append(signLb.Text);
      for(int i = 0; i < digits; i++)
      {
        tmp.Append(digitsList[i].Text);
      }
      if (fractionalDigits > 0)
      {
        tmp.Append(".");
        for(int i = digits; i < digits + fractionalDigits; i++)
        {
          tmp.Append(digitsList[i].Text);
        }
      }
      value = Convert.ToDouble(tmp.ToString());
      return value;
    }

    private String getDigitValueString()
    {
      StringBuilder stringValue = new StringBuilder();;

      stringValue.Append(signLb.Text);
      for (int i = 0; i < digits; i++)
      {
        stringValue.Append(digitsList[i].Text);
      }
      if (fractionalDigits > 0)
      {
        stringValue.Append(".");
        for (int i = digits; i < digits + fractionalDigits; i++)
        {
          stringValue.Append(digitsList[i].Text);
        }
      }
      return stringValue.ToString();
    }

    private Boolean isInRange(double value)
    {
      if (value < thresholdLow)
      {
        holdTimer.Enabled = false;
        setDigitValue(thresholdLow);
        return false;
      }
      if(value > thresholdHigh)
      {
        holdTimer.Enabled = false;
        setDigitValue(thresholdHigh);
        return false;
      }
      return true;
    }

    private void setThresholdHigh(double value)
    {
      if (value != MAX_THRESHOLD) thresholdHigh = value;
    }
    private void setThresholdLow(double value)
    {
      if (value != -MAX_THRESHOLD) thresholdLow = value;
    }

    private void setDigits(int value)
    {
      digits = value;
      setWheelSwitch();
    }

    private void setFractionalDigits(int value)
    {
      fractionalDigits = value;
      setWheelSwitch();
    }

    private void setWheelValue(double value)
    {
      if (value > thresholdHigh) value = thresholdHigh;
      if (value < thresholdLow) value = thresholdLow;
      if (value < 0) isNegValue = true;
      setDigitValue(value);
    }

    private double getWheelValue()
    {
      return getDigitValue();
    }

    private void setRbvStartValue(Boolean value)
    {
      rbvStartValue = value;
    }

    private void setShowHeader(Boolean value)
    {
      showHeader = value;
      setComponentSize();
    }

    private void setComponentSize()
    {
      if (showHeader)
      {
        digitsPanel.Location = new Point(0, 100);
        this.Controls.Add(headerPanel);
      }
      else
      {
        this.Controls.Remove(headerPanel);
        digitsPanel.Location = new Point(0, 0);
      }

      switch(digitSize)
      {
        case wheelDigitSizeEnum.Normal:
          if (showHeader)
          {
            this.MinimumSize = new Size(160, 142);
          }
          else
          {
            this.MinimumSize = new Size(160, 42);
            this.Size = new Size(this.Width, 42);
          }
          break;
        case wheelDigitSizeEnum.Smaller:
          if (showHeader)
          {
            this.MinimumSize = new Size(160, 132);
          }
          else
          {
            this.MinimumSize = new Size(120, 32);
            this.Size = new Size(this.Width, 32);
          }
          break;
        case wheelDigitSizeEnum.Larger:
          if (showHeader)
          {
            this.MinimumSize = new Size(160, 149);
          }
          else
          {
            this.MinimumSize = new Size(220, 49);
            this.Size = new Size(this.Width, 49);
          }
          break;
        case wheelDigitSizeEnum.Largest:
          if (showHeader)
          {
            this.MinimumSize = new Size(160, 159);
          }
          else
          {
            this.MinimumSize = new Size(285, 59);
            this.Size = new Size(this.Width, 59);
          }
          break;
      }
    }

    private void setDeadTime(int value)
    {
      deadtime = value;
      holdTimer.Interval = deadtime;
    }

    private void setLabelInSyncForeColor(Color value)
    {
      labelInSyncForeColor = value;
    }
    private void setLabelOutOfSyncForeColor(Color value)
    {
      labelOutOfSyncForeColor = value;
    }

    private void setReadbackForeColor(Color value)
    {
      readbackForeColor = value;
      rbvValueLabel.ForeColor = value;
    }
    private void setReferenceForeColor(Color value)
    {
      referenceForeColor = value;
      refValueLabel.ForeColor = value;
    }

    private void setWheelDigitSize(wheelDigitSizeEnum value)
    {
      digitSize = value;
      switch (digitSize)
      {
        case wheelDigitSizeEnum.Normal:
          digitXSize = NORMAL_DIGIT_X_SIZE;
          digitYSize = NORMAL_DIGIT_Y_SIZE;
          digitYPos = NORMAL_DIGIT_Y_POS;
          btnXSize = NORMAL_BTN_X_SIZE;
          btnYSize = NORMAL_BTN_Y_SIZE;
          digitsFontSize = NORMAL_DIGITS_FONTSIZE;
          unitLabelFontSize = NORMAL_UNITLABEL_FONTSIZE;
          break;
        case wheelDigitSizeEnum.Larger:
          digitXSize = LARGER_DIGIT_X_SIZE;
          digitYSize = LARGER_DIGIT_Y_SIZE;
          digitYPos = LARGER_DIGIT_X_POS;
          btnXSize = LARGER_BTN_X_SIZE;
          btnYSize = LARGER_BTN_Y_SIZE;
          digitsFontSize = LARGER_DIGITS_FONTSIZE;
          unitLabelFontSize = LARGER_UNITLABEL_FONTSIZE;
          break;
        case wheelDigitSizeEnum.Largest:
          digitXSize = LARGEST_DIGIT_X_SIZE;
          digitYSize = LARGEST_DIGIT_Y_SIZE;
          digitYPos = LARGEST_DIGIT_Y_POS;
          btnXSize = LARGEST_BTN_X_SIZE;
          btnYSize = LARGEST_BTN_Y_SIZE;
          digitsFontSize = LARGEST_DIGITS_FONTSIZE;
          unitLabelFontSize = LARGEST_UNITLABEL_FONTSIZE;
          break;
        case wheelDigitSizeEnum.Smaller:
          digitXSize = SMALLER_DIGIT_X_SIZE;
          digitYSize = SMALLER_DIGIT_Y_SIZE;
          digitYPos = SMALLER_DIGIT_Y_POS;
          btnXSize = SMALLER_BTN_X_SIZE;
          btnYSize = SMALLER_BTN_Y_SIZE;
          digitsFontSize = SMALLER_DIGITS_FONTSIZE;
          unitLabelFontSize = SMALLER_UNITLABEL_FONTSIZE;
          break;
      }
      setWheelSwitch();
      setComponentSize();
    }

    private void setDigitsFontSize()
    {
      signLb.setFontSize(digitsFontSize);
      pointLb.setFontSize(digitsFontSize);

      foreach (WsLabel lb in digitsList)
      {
        lb.setFontSize(digitsFontSize);
      }
    }

    private void setButtonsEnabled(Boolean enabled)
    {
      upBtn.setEnabled(enabled);
      downBtn.setEnabled(enabled);
      upBtn.Invalidate();
      upBtn.Update();
      downBtn.Invalidate();
      downBtn.Update();
    }

    private void setDigitsInsync(Boolean insync)
    {
      Color tmpColor;
      if (insync) tmpColor = labelInSyncForeColor;
      else tmpColor = LabelOutOfSyncForeColor;

      signLb.ForeColor = tmpColor;
      pointLb.ForeColor = tmpColor;

      foreach (WsLabel lb in digitsList)
      {
        lb.ForeColor = tmpColor;
      }
    }

    private void AcopMouseClick(object sender, MouseEventArgs e)
    {
      if (e.Button != MouseButtons.Left) return;
      WsLabel digit = sender as WsLabel;
      if(digit.Selected)
      {
        clearSelection();
        setButtonsEnabled(false);
      }
      else
      {
        clearSelection();
        digit.Selected = true;
        setButtonsEnabled(true);
      }
    }

    private void AcopButtonUpMouseDown(object sender, MouseEventArgs e)
    {
      if (e.Button != MouseButtons.Left) return;
      isDownButton = false;
      holdTimer.Enabled = true;
    }
    private void AcopButtonDownMouseDown(object sender, MouseEventArgs e)
    {
      if (e.Button != MouseButtons.Left) return;
      isDownButton = true;
      holdTimer.Enabled = true;
    }
    private void AcopButtonMouseUp(object sender, MouseEventArgs e)
    {
      if (e.Button != MouseButtons.Left) return;
      holdTimer.Enabled = false;
    }

    private void holdTimerEvent(object sender, ElapsedEventArgs e)
    {
      if (this.InvokeRequired)
      {
        Invoke(new TimerUpdateDelegate(holdTimerEvent), sender, e);
        return;
      }
      if (hasLnkError) return;
      foreach (WsLabel lb in digitsList)
      {
        if (lb.Selected)
        {
          if (isDownButton)
          {
            if (lb.isAllPrevZero() && lb.isZero())
            {
              signLb.setText("-");
              signLb.Visible = true;
              isNegValue = true;
            }
            if (isNegValue) lb.plusOne();
            else lb.minusOne();
          }
          else
          {
            if(lb.isAllPrevZero() && lb.isZero())
            {
              signLb.setText("+");
              signLb.Visible = false;
              isNegValue = false;
            }
            if (isNegValue) lb.minusOne();
            else lb.plusOne();
          }
          if (isInRange(getDigitValue()))
          {
            if (applyContinuous)
            {
              if (AutoLinkAttach || hasTargetComponent())
              {
                int cc = applyWheelSettings(sender);
                if (cc != 0) hasLnkError = true;
                else hasLnkError = false;
              }
            }
          }
        }
      }
      System.Threading.Thread.Yield();
      Application.DoEvents();
    }

    public int applyWheelSettings(object sender)
    {
      object inpt = null;
      AcopControl.AcopLink lnk = null;
      if (hasTargetComponent())
      {
        inpt = getDigitValue();
      }
      else
      {
        Type type = getValueType(); // as long as getValueAsDouble() is called in the updater this will be something ...
        String s = getDigitValueString();
        if (type == null) type = getValueType(s);
        inpt = AcopControl.AcopLink.GetDataObjectFromString(type, s);
        lnk = AcopControl.AcopLink.GetAcopLink(LinkProtocol, LinkAddress, LinkProperty, null, inpt, 0, this);
      }
      return applySetting(sender, inpt, lnk);
    }

    private void applyContinousClick(object sender, EventArgs e)
    {
      applyContinuous = continuousCheck.Checked;
    }
    private void applySetClick(object sender, EventArgs e)
    {
      int cc = applyWheelSettings(sender);
    }

    public void applySaveRefClick(object sender, EventArgs e)
    {
      double resetValue = getDigitValue();
      SetResetValue(resetValue);
      refValueLabel.Text = getLabelString(resetValue, 6);
    }

    public void applyResetClick(object sender, EventArgs e)
    {
      setDigitValue(GetResetValue());
      if (rbvValueLabel.Text.CompareTo(refValueLabel.Text) == 0) return;
      int cc = applyWheelSettings(sender);
    }

    public void applyLockClick(object sender, EventArgs e)
    {
      applyLock(lockCheck);
    }

    public void applyErrorStatus(String sts, Boolean applyToTTT)
    {
      rbvValueLabel.Text = sts;
      rbvValueLabel.ForeColor = Color.Red;
      if (applyToTTT) applyToolTipText(sts);
    }

    private void defaultUpdate(AcopControl.AcopLink lnk)
    {
      if (!lateInitializationDone) return;
      if (getAutoUpdate())
      {
        int lnkSts = lnk.GetStatus();
        if (lnkSts != 0 && ErrorValue.Length == 0)
        {
          applyErrorStatus(lnk.GetStatusString(), lnkSts != lastLinkStatus);
          hasLnkError = true;
        }
        else
        {
          if (lnkSts != lastLinkStatus) applyToolTipText(null);
          hasLnkError = false;
          double v = getValueAsDouble(lnk);
          String txt;
          if (getValueAvailableAsNumber())
          {
            txt = getLabelString(v, 6);
          }
          else
          {
            String[] txtlines = lnk.GetOutputDataAsString();
            txt = txtlines[0];
          }
          if (lnk.GetDisplayId() < 0)
          {
            if (!displaySettingsAcquired) initTransportInfo();
            lnk.SetDisplayId(1);
            SetResetValue(v);
            applyToolTipText(descLabel, Description);
            refValueLabel.Text = getLabelString(v, 6);
          }
          SetReadbackValue(v);
          rbvValueLabel.ForeColor = readbackForeColor;
          rbvValueLabel.Text = txt;
          if(rbvValueLabel.Text.CompareTo(getDigitValue().ToString()) != 0)
          {
            setDigitsInsync(false);
          }
          else
          {
            setDigitsInsync(true);
          }
           
          if(rbvStartValue)
          {
            rbvStartValue = false;
            setDigitValue(v);
          }
        }
      }
    }

    private void initTransportInfo()
    {
      if (!IsDesignMode())
      {
        if (hasTargetComponent())
        {
          try
          { // casting errors ?
            setWheelValue((double)getSettingFromControlProperty(getTargetComponent()));
          }
          catch (Exception) { }
          descLabel.Text = Description;
          unitLabel.Text = Units;
          displaySettingsAcquired = true;
          return;
        }
        String[] labels = null;
        AcopTransport.DisplaySettings dsp = null;
        if (!AutoLinkAttach && !AutoLinkUpdate)
        {
          descLabel.Text = Description;
          unitLabel.Text = Units;
          displaySettingsAcquired = true;
          return;
        }
        dsp = acquireAndApplyDisplaySettings(LinkAddress, LinkProperty, ref labels, false);
        if (dsp == null || dsp.SettingsStale()) return;
        displaySettingsAcquired = true;
        if (Units == "") unitLabel.Text = dsp.GetYUnits();
        else unitLabel.Text = Units;
        if (thresholdHigh == MAX_THRESHOLD) thresholdHigh = dsp.GetYMaximum();
        if (thresholdLow == -MAX_THRESHOLD) thresholdLow = dsp.GetYMinimum();
        if (thresholdHigh == thresholdLow) thresholdHigh = thresholdLow + 100.0;
        descLabel.Text = (getIsMCA() && getDisplayAsMCA()) ? getMCADescription() : getDescription();
      }
    }

    private int defaultApplier(AcopControl.AcopLink lnk)
    {
      int cc = lnk.GetStatus();
      if (getValueAvailableAsNumber() && cc == 0)
      {
        double v = getDigitValue();
        rbvValueLabel.Text = getLabelString(v, 6);
      }
      if (cc != 0)
      {
        rbvValueLabel.ForeColor = Color.Red;
        rbvValueLabel.Text = lnk.GetStatusString();
      }
      else
      {
        rbvValueLabel.ForeColor = readbackForeColor;
      }
      return cc;
    }

    #region Properties
    [Category("Acop.Display")]
    [Description("The maximum setting value")]
    [DefaultValue(MAX_THRESHOLD)]
    public double MaximumValue { get { return thresholdHigh; } set { setThresholdHigh(value); } }
    [Category("Acop.Display")]
    [Description("The minimum setting value")]
    [DefaultValue(-MAX_THRESHOLD)]
    public double MinimumValue { get { return thresholdLow; } set { setThresholdLow(value); } }
    [Category("Acop.Display")]
    [Description("The number of digits")]
    [DefaultValue(4)]
    public int Digits { get { return digits; } set { setDigits(value); } }
    [Category("Acop.Display")]
    [Description("The number of fractional digits")]
    [DefaultValue(1)]
    public int FractionalDigits { get { return fractionalDigits; } set { setFractionalDigits(value); } }
    [Category("Acop.Display")]
    [Description("The set value at initialization")]
    public double WheelSwitchValue { get { return getWheelValue(); } set { setWheelValue(value); } }
    [Category("Acop.Display")]
    [Description("Set the size of the wheel digits")]
    [DefaultValue(wheelDigitSizeEnum.Normal)]
    public wheelDigitSizeEnum WheelDigitsSize { get { return digitSize; } set { setWheelDigitSize(value); } }
    [Category("Acop.Display")]
    [Description("Set the first readback value as start value")]
    [DefaultValue(true)]
    public Boolean StartValueFromRbv { get { return rbvStartValue; } set { setRbvStartValue(value); } }
    [Category("Acop.Display")]
    [Description("Show the information header")]
    [DefaultValue(true)]
    public Boolean ShowHeader { get { return showHeader; } set { setShowHeader(value); } }
    [Category("Acop.Display")]
    [Description("Set the Apply Deadtime Interval for continuous mode")]
    [DefaultValue(50)]
    public int ApplyDeadTimeInterval { get { return deadtime; } set { setDeadTime(value); } }
    [Category("Acop.Display")]
    [Description("ForeColor to use if setting is synchronized with the current readback value")]
    [DefaultValue(typeof(Color), "0x00008B")]
    public Color LabelInSyncForeColor { get { return labelInSyncForeColor; } set { setLabelInSyncForeColor(value); } }
    [Category("Acop.Display")]
    [Description("ForeColor to use if setting is not synchronized with the current readback value")]
    [DefaultValue(typeof(Color), "0x8B0000")]
    public Color LabelOutOfSyncForeColor { get { return labelOutOfSyncForeColor; } set { setLabelOutOfSyncForeColor(value); } }
    [Category("Acop.Display")]
    [Description("Default ForeColor to use for the Readback label")]
    [DefaultValue(typeof(Color), "0x000000")]
    public Color ReadbackForeColor { get { return readbackForeColor; } set { setReadbackForeColor(value); } }
    [Category("Acop.Display")]
    [Description("Default ForeColor to use for the Reference label")]
    [DefaultValue(typeof(Color), "0x000000")]
    public Color ReferenceForeColor { get { return referenceForeColor; } set { setReferenceForeColor(value); } }

    //[Category("Acop.Display")]
    //[Description("If not empty, should be in the form of 'ControlName.Property' : apply setting will set the control property to the wheel switch value")]
    //public String TargetComponent { get { return getTargetComponent(); } set { setTargetComponent(value); } }

    [Category("Acop.Display")]
    [Description("If not empty, attach the target control property to the wheel switch")]
    [Editor(typeof(TargetControlEditor), typeof(System.Drawing.Design.UITypeEditor))]
    [DesignerSerializationVisibility(DesignerSerializationVisibility.Content)]
    [TypeConverter(typeof(TargetControlEditorConverter))]
    public TargetControlEditorValue Target { get { return getTargetControlEditorValue(); } set { setTargetControlEditorValue(value); } }

    [Category("Acop.Transport")]
    [Browsable(false), EditorBrowsable(EditorBrowsableState.Never)]
    public new Boolean LinkGrouped { get; set; }
    #endregion
  }

  [ToolboxItem(false)]
  public class AcopWheelSwitchContextMenu : ContextMenu
  {
#   pragma warning disable 1591
    private AcopWheelSwitch wheel;
    private MenuItem apply;
    private MenuItem reset;
    private MenuItem save;

    public AcopWheelSwitchContextMenu(AcopWheelSwitch wheel)
    {
      this.wheel = wheel;
      apply = new MenuItem("Apply");
      apply.Click += new EventHandler(this.applyChanges);
      save = new MenuItem("Set Reference");
      save.Click += new EventHandler(this.applySave);
      reset = new MenuItem("Restore Reference");
      reset.Click += new EventHandler(this.applyReset);

      this.MenuItems.Add(apply);
      this.MenuItems.Add(save);
      this.MenuItems.Add(reset);
    }
    private void applyChanges(object sender, EventArgs e)
    {
      wheel.applyWheelSettings(sender);
    }
    private void applySave(object sender, EventArgs e)
    {
      wheel.applySaveRefClick(sender, e);
    }
    private void applyReset(object sender, EventArgs e)
    {
      wheel.applyResetClick(sender, e);
    }
  }

}
