﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace Acop.AcopWheelSwitchUtil
{
  [ToolboxItem(false)]
  public partial class WsBtn : Button
  {
#   pragma warning disable 1591
    private SolidBrush blackBrush = new SolidBrush(Color.Black);
    private SolidBrush grayBrush = new SolidBrush(Color.Gray);
    private Pen blackPen = new Pen(Color.Black, 0.5f);
    private Pen grayPen = new Pen(Color.Gray, 0.5f);

    private Boolean mirrored = false;
    private Boolean enabled = true;
    
    private delegate void ButtonEnabledDelegate(Boolean value);

    public WsBtn(Boolean mirrored)
    {
      this.mirrored = mirrored;
      FlatStyle = FlatStyle.Flat;
      this.FlatAppearance.BorderSize = 0;
    }

    public void setEnabled(Boolean enabled)
    {
      if(this.InvokeRequired)
      {
        Invoke(new ButtonEnabledDelegate(setEnabled), enabled);
        return;
      }
      this.Enabled = enabled;
      this.enabled = enabled;
    }

    public Boolean isMirrored()
    {
      return mirrored;
    }

    protected override void OnPaint(PaintEventArgs e)
    {
      base.OnPaint(e);

      FillMode fm = FillMode.Winding;
      PointF p1, p2, p3;

      if (mirrored)
      {
        p1 = new PointF(this.Width / 2, this.Height - 3);
        p2 = new PointF(2, 2);
        p3 = new PointF(this.Width - 3, 2);
      }
      else
      {
        p1 = new PointF(this.Width / 2, 2);
        p2 = new PointF(this.Width - 3, this.Height - 3);
        p3 = new PointF(2, this.Height - 3);
      }
      PointF[] points = { p1, p2, p3 };

      Graphics g = e.Graphics;
      g.SmoothingMode = SmoothingMode.AntiAlias;
      if(enabled)
      {
        g.FillPolygon(blackBrush, points, fm);
        g.DrawRectangle(blackPen, 0, 0, this.Width - 1, this.Height - 1);
      }
      else
      {
        g.FillPolygon(grayBrush, points, fm);
        g.DrawRectangle(grayPen, 0, 0, this.Width - 1, this.Height - 1);
      }
      
    }
  }
}
