﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using System.Drawing;

namespace Acop.AcopWheelSwitchUtil
{
  [ToolboxItem(false)]
  public partial class WsLabel : Label
  {
#   pragma warning disable 1591
    private Boolean isSelected = false;
    private Boolean isNumber = true;
    private Boolean isFirstDigit = false;

    private Pen blackPen = new Pen(Color.Gray, 0.5f);

    private WsLabel prev;

    private delegate void PlusMinusUpdateDelegate();
    private delegate void TextUpdateDelegate(String text);

    public WsLabel()
    {
      this.Font = new Font(FontFamily.GenericMonospace, 16, FontStyle.Regular);
      this.TextAlign = ContentAlignment.MiddleCenter;
    }

    public void setFontSize(int value)
    {
      this.Font = new Font(FontFamily.GenericMonospace, value, FontStyle.Regular);
    }

    private void setSelected(Boolean selected)
    {
      this.isSelected = selected;
      if (isSelected) this.BackColor = Color.LightBlue;
      else this.BackColor = Color.FromKnownColor(KnownColor.Control);
      this.Refresh();
    }

    private void setNumber(Boolean number)
    {
      this.isNumber = number;
    }

    private void setPrevDigit(WsLabel prev)
    {
      this.prev = prev;
    }

    private void setFirstDigit(Boolean first)
    {
      this.isFirstDigit = first;
    }

    public void plusOne()
    {
      if(this.InvokeRequired)
      {
        Invoke(new PlusMinusUpdateDelegate(plusOne));
        return;
      }
      int num = Convert.ToInt16(this.Text);
      if(num == 9)
      {
        num = 0;
        if (!isFirstDigit)
        {
          prev.plusOne();
          this.Text = num.ToString();
        }
        else
        {
          prev.Text = "+";
          this.Text = num.ToString();
        }
      }
      else
      {
        num++;
        this.Text = num.ToString();
      }
    }
    public void minusOne()
    {
      if (this.InvokeRequired)
      {
        Invoke(new PlusMinusUpdateDelegate(minusOne));
        return;
      }
      int num = Convert.ToInt16(this.Text);
      if (num == 0)
      {
        num = 9;
        if (!isFirstDigit)
        {
          prev.minusOne();
          this.Text = num.ToString();
        }
        else
        {
          prev.Text = "-";
          this.Text = num.ToString();
        }
      }
      else
      {
        num--;
        this.Text = num.ToString();
      }
    }

    public Boolean isAllPrevZero()
    {
      if(!isFirstDigit)
      {
        if (prev.Text == "0") return prev.isAllPrevZero();
        else return false;
      }
      if(this.Text == "0") return true;
      return false;
    }

    public Boolean isZero()
    {
      if (this.Text == "0") return true;
      else return false;
    }

    public void setText(String text)
    {
      if (this.InvokeRequired)
      {
        Invoke(new TextUpdateDelegate(setText), text);
        return;
      }
      this.Text = text;
    }

    public Boolean Selected { get { return isSelected; } set { setSelected(value); } }
    public Boolean Number { get { return isNumber; } set { setNumber(value); } }
    public WsLabel PrevDigit { get { return prev; } set { setPrevDigit(value); } }
    public Boolean FirstDigit { get { return isFirstDigit; } set { setFirstDigit(value); } }

    protected override void OnPaint(PaintEventArgs e)
    {
      base.OnPaint(e);
      Graphics g = e.Graphics;
      
      g.DrawRectangle(blackPen, 0, 0, this.Width - 1, this.Height - 1);
    }
  }
}
