﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace AcopTransport
{
  public interface AcopConnection
  {
    //public AcopConnection(String Address, String Property, String Protocol);
    // address methods:
    AcopConnection GetConnection(String Protocol, String Address, String Property, object DataOut, object DataIn, int Flags, object acop);
    AcopConnection GetConnection(String Protocol, String Address, String Property, object DataOut, int offset, int length, object DataIn, int Flags, object acop);
    AcopLink[] GetConnectionList();
    string GetConnectionKey();
    int GetFlags();
    string GetConnectionKey(String Protocol, String Address, String Property, object DataOut, object DataIn, int Flags, object acop);
    string GetConnectionAddress(String Protocol, String Context, String Subsystem, String Server, String Device);
    string GetConnectionProtocol();
    string GetConnectionAddress();
    string GetConnectionProperty();
    object GetConnectionReference();
    string GetConnectionHostAddress(String Protocol, String Address, String Property);
    // acquisition methods:
    int Attach(int Flags, int interval, object reference, AcopLinkCallback cb);
    int Detach();
    int Execute(int Flags, int timeout);
    int GetHistoryPoints(double[][] data, ref int length, double StartTime, double StopTime, int Flags, int timeout);
    int GetImage(ref Bitmap bt, ref String header);
    int GetLinkStatus();
    string GetLinkStatusString();
    bool HasLinkStatusChanged();
    double GetLinkTimestamp();
    int GetLinkCompletionSize();
    int GetLinkTag();
    int LockAccess(String Address);
    int UnlockAccess(String Address);
    // query methods:
    double[] GetXAxis();
    double[] GetYAxis();
    string[] GetDataAsString();
    double[] GetDataAsNumber();
    Int32[] GetDataAsInteger();
    string[] GetChannelNames();
    object GetOutputDataObject(String Address, String Property);
    object GetOutputDataObject(String Address, String Property, Object Reference, Int32 Capacity);
    DisplaySettings GetDisplaySettings(String Address, String Property);
    string[] GetProtocolList();
    string[] GetContextList(string protocol);
    string[] GetSubSystemList(string protocol, string context);
    string[] GetServerList(string protocol, string context, string subsystem);
    string[] GetDeviceList(string protocol, string context, string subsystem, string server, string property);
    string[] GetPropertyList(string protocol, string context, string subsystem, string server, string device);
    Boolean HasPropertyQueryPrecedence(string protocol, string context, string subsystem, string server, string property);
    Boolean HasPropertyQueryPrecedence(string protocol, string context, string subsystem, string server);
    Boolean HasDeviceQueryPrecedence(string protocol, string context, string subsystem, string server);
    string[] GetConnectionList(string protocol, Boolean activeOnly);
    string[] GetTransportMessages(string protocol);
    int GetTransportStatus(string protocol);
    void StartDebuggingSession(string protocol);
    object GetDataObjectFromString(Type type, String data);
    object GetDataObjectFromStringArray(Type type, String[] data);
    void FillDataObjectFromString(object src, String data);
    String GetDefaultContext(String Protocol);
    String GetDefaultServer(String Protocol);
    String GetDefaultDevice(String Protocol);
    String GetDefaultProperty(String Protocol);
    IDictionary<string, int> GetFlagsList(String Protocol);
  }
  public interface AcopLinkCallback
  {
    void callback(AcopConnection lnk);
  }
  public enum simFlags
  {
    NONE = 0,
  }
  public enum tineFlags
  {
    NONE = 0,
    WRITE = tine.Access.CA_WRITE,
    READ_EXCLUSIVE = (tine.Access.CA_READ | tine.Access.CA_NOCALLBACKS),
    WRITE_EXCLUSIVE = (tine.Access.CA_WRITE | tine.Access.CA_NOCALLBACKS),
    SINGLE_UPDATE = (tine.Modes.CM_SINGLE << 16),
    DATACHANGE_UPDATE = (tine.Modes.CM_DATACHANGE << 16),
    TIMER_UPDATE = (tine.Modes.CM_TIMER << 16),
    EVENT_UPDATE = (tine.Modes.CM_EVENT << 16),
    DATACHANGE_NETWORK = ((tine.Modes.CM_DATACHANGE | tine.Modes.CM_NETWORK) << 16),
    TIMER_NETWORK = ((tine.Modes.CM_TIMER | tine.Modes.CM_NETWORK) << 16),
    EVENT_NETWORK = ((tine.Modes.CM_EVENT | tine.Modes.CM_NETWORK) << 16),
    SINGLE_CONNECT = ((tine.Modes.CM_SINGLE | tine.Modes.CM_CONNECT) << 16),
    WRITE_CONNECT = tine.Access.CA_WRITE | ((tine.Modes.CM_SINGLE | tine.Modes.CM_CONNECT) << 16),
    WRITE_CONNECT_EXLUSIVE  = (tine.Access.CA_WRITE|tine.Access.CA_NOCALLBACKS) | ((tine.Modes.CM_SINGLE|tine.Modes.CM_CONNECT) << 16),
    DATACHANGE_CONNECT = ((tine.Modes.CM_DATACHANGE | tine.Modes.CM_CONNECT) << 16),
    TIMER_CONNECT = ((tine.Modes.CM_TIMER | tine.Modes.CM_CONNECT) << 16),
    EVENT_CONNECT = ((tine.Modes.CM_EVENT | tine.Modes.CM_CONNECT) << 16),
  }
  public class SelectList
  {
    public static IDictionary<string, int> Of<T>()
    {
      Type t = typeof(T);
      if (t.IsEnum)
      {
        var values = Enum.GetValues(t);
        var result = new Dictionary<string, int>();
        foreach (var value in values)
        {
          result.Add(value.ToString(), (int)value);
        }
        return result;
//        return Enum.GetValues(t).Cast<Enum>();
      }
      return null;
    }
  }
  public class DisplaySettings
  {
    public enum DataType
    {
      EMPTY = 0,
      NUMBER = 1,
      STRING = 2,
      STRUCTURE = 3,
      COMPOUND = 4,
      IMAGE = 5
    }
    public DisplaySettings(double ymin, double ymax, double xmin, double xmax, String yunits, String xunits, int arraytype, String description)
    {
      YMin = ymin; YMax = ymax; XMin = xmin; XMax = xmax;
      if (yunits == null) yunits = "";
      int idx = yunits.IndexOf('\0'); if (idx < 0) idx = yunits.Length;
      YUnits = new String(yunits.ToCharArray(), 0, idx);
      if (xunits == null) xunits = "";
      idx = xunits.IndexOf('\0'); if (idx < 0) idx = xunits.Length;
      XUnits = new String(xunits.ToCharArray(), 0, idx);
      if (description == null) description = "";
      idx = description.IndexOf('\0'); if (idx < 0) idx = description.Length;
      Description = new String(description.ToCharArray(), 0, idx);
      ArrayType = arraytype;
      Boolean ylog = (ymax - ymin) > 10E4;
      Boolean xlog = (xmax - xmin) > 10E4;
      if (!ylog && !xlog) GraphType = 0;
      else if (ylog && !xlog) GraphType = 1;
      else if (!ylog && xlog) GraphType = 2;
      else GraphType = 3;
    }
    private Boolean settingsStale = true;
    public void SettingsStale(Boolean value)
    {
      settingsStale = value;
    }
    public Boolean SettingsStale()
    {
      return settingsStale;
    }
    private DataType dataType = DataType.EMPTY;
    public DataType GetDataType() { return dataType; }
    public void SetDataType(DataType value)
    {
      dataType = value;
    }
    private int capacity = 0;
    public int GetCapacity() { return capacity; }
    public void SetCapacity(int value)
    {
      if (value > 0) capacity = value;
    }
    private double YMin;
    public double GetYMinimum() { return YMin; }
    public void SetYMinimum(double value) { YMin = value; }
    private double YMax;
    public double GetYMaximum() { return YMax; }
    public void SetYMaximum(double value) { YMax = value; }
    private double XMin;
    public double GetXMinimum() { return XMin; }
    public void SetXMinimum(double value) { XMin = value; }
    private double XMax;
    public double GetXMaximum() { return XMax; }
    public void SetXMaximum(double value) { XMax = value; }
    private double XShift = 0.0;
    public double GetXShift() { return XShift; }
    public void SetXShift(double value) { XShift = value; }
    private double XScale = 1.0;
    public double GetXScale() { return XScale; }
    public void SetXScale(double value) { XScale = value; }
    private String YUnits;
    public String GetYUnits() { return YUnits; }
    public void SetYUnits(String value) { YUnits = value; }
    private String XUnits;
    public String GetXUnits() { return XUnits; }
    public void SetXUnits(String value) { XUnits = value; }
    private int ArrayType;
    public int GetArrayType() { return ArrayType; }
    public void SetArrayType(int value) { ArrayType = value; }
    private int GraphType;
    public int GetGraphType() { return GraphType; }
    public void SetGraphType(int value) { GraphType = value; }
    private String Description;
    public String GetDescription() { return Description; }
    public void SetDescription(String value) { Description = value; }
  }
}
