﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Reflection;
using tine;
using System.IO;
using System.Drawing;

namespace AcopTransport
{ // protocol dependent methods ...
  public class AcopLink : AcopConnection
  {
    private class AcopTLinkCb
    {
      AcopLink thisAcopLink = null;
      public AcopTLinkCb(AcopLink alnk)
      {
        thisAcopLink = alnk;
      }
      public void cb(TLink lnk)
      {
        AcopLink alnk = AcopLinkFactory.GetAcopLink(thisAcopLink.linkKey);
        alnk.stsLastUpdate = alnk.sts;
        alnk.sts = lnk.GetStatus();
        if (tine.Errors.HasData(alnk.sts)) alnk.sts = 0; // make acop happy!
        alnk.lastLinkCompletionSize = lnk.GetCompletionDataSize();
        alnk.SetLinkTimestamp(lnk.GetCompletionDataTimeStamp());
        alnk.SetLinkTag((int)lnk.GetCompletionSystemStamp());
        alnk.SetLinkStatusString(lnk.GetStatusString());
        if (alnk.acopCallback != null) alnk.acopCallback.callback(alnk);
        if (tine.Modes.GetBaseMode((ushort)lnk.GetAccessMode()) <= tine.Modes.CM_SINGLE)
        { // no longer active after this callback
          alnk.lid = -1;
          thisAcopLink.ctrlLnk.SetIdle();
        }
      }
    }
    AcopControl.AcopLink ctrlLnk = null;
    public void SetAcopControlLink(AcopControl.AcopLink lnk)
    {
      ctrlLnk = lnk;
    }
    static AcopLink instance = new AcopLink();
    //AcopLink instance = null;
    private static System.Timers.Timer cbTimer = null;
    //public delegate void acopTransportUpdate(object sender, AcopEvent e);
    // public event acopTransportUpdate acopUpdated;
    private String linkKey = null;
    private String addr = null;  // address
    public String GetAddress() { return addr; }
    private String prop = null;  // property
    public String GetProperty() { return prop; }
    private String prot = AcopLinkFactory.GetDefaultProtocol();  // protocol
    public String GetProtocol() { return prot; }
    public void SetProtocol(String protocol) { prot = protocol; }
    private int lid = -1;      // link id
    public String GetDefaultContext(String Protocol)
    {
      if (prot.Equals("TINE", StringComparison.OrdinalIgnoreCase))
      {
        return "SITE";
      }
      return "";
    }
    public String GetDefaultServer(String Protocol)
    {
      if (prot.Equals("TINE", StringComparison.OrdinalIgnoreCase))
      {
        return "ENS";
      }
      return "";
    }
    public String GetDefaultDevice(String Protocol)
    {
      if (prot.Equals("TINE", StringComparison.OrdinalIgnoreCase))
      {
        return "DEFAULT";
      }
      return "";
    }
    public String GetDefaultProperty(String Protocol)
    {
      if (prot.Equals("TINE", StringComparison.OrdinalIgnoreCase))
      {
        return "NSERVERS";
      }
      return "";
    }
    public int GetLinkId() { return lid; }
    private object cbref = null;     // callback reference object
    public object GetConnectionReference() { return cbref; }
    private int intvl = AcopLinkFactory.GetDefaultInterval();    // monitor interval or timeout
    private int sts = -1;      // status
    private int stsLastUpdate = -1;
    public int GetLinkStatus() { return sts; }
    private String stsString = "";
    public String GetLinkStatusString() { return stsString; }
    public bool HasLinkStatusChanged()
    {
      return (sts == stsLastUpdate);
    }
    public void SetLinkStatusString(String sts) { stsString = sts; }
    private double ts = 0;    // timestamp
    public void SetLinkTimestamp(double timestamp) { ts = timestamp; }
    public double GetLinkTimestamp() { return ts; }
    private int tag = 0;      // tag stamp (e.g. cycle no., pulse no.)
    public int GetLinkTag() { return tag; }
    public void SetLinkTag(int value) { tag = value; }
    private int lastLinkCompletionSize = 0;
    public int GetLinkCompletionSize() { return lastLinkCompletionSize; }
    public void SetLinkCompletionSize(int length) { lastLinkCompletionSize = length; }
    private int flgs = 0;     // link flags
    private object dout = null;  // data output object
    private object din = null;   // data input object
    private int doutOffset = -1; // specific data output object offset (relevant for arrays)
    private int doutLength = -1; // specific data output object length (relevant for arrays)
    private DisplaySettings dsp = null;
    private AcopLinkCallback acopCallback = null;
    private AcopTLinkCb acopTineCallback = null;
    private TLink.TLinkCallback tineCallback = null; //acopTineCallback.cb;
    private TLink tineLink = null;
    private int tineLinkMode = 0;
    private int tineLinkAccess = 0;
    private int tineLinkFormat = 0;
    private string[] protlst = null;
    public AcopLink()
    {
      // TODO: read from a manifest ? Put in the LinkFactory ...
      protlst = new String[] { "SIMULATE", "TINE" };
      acopTineCallback = new AcopTLinkCb(this);
      tineCallback = acopTineCallback.cb;
    }
    public AcopLink[] GetConnectionList()
    {
      return AcopLinkFactory.GetAcopLinks();
    }
    public int GetConnectionInterval() { return intvl; }
    private static string makeObjectTag(object d)
    {
      if (d == (Object)null) return "null";
      Type dtype;
      int len = 1;
      if (d is Array)
      {
        dtype = d.GetType().GetElementType();
        Array da = (Array)d;
        int dim = da.Rank;
        while (--dim >= 0) len *= da.GetLength(dim);
      }
      else
      {
        dtype = d.GetType();
      }
      return dtype.ToString() + "[" + len + "]@" + d.GetHashCode();
    }
    public static string MakeConnectionKey(String Protocol, String Address, String Property, object DataOut, object DataIn, int Flags, object acop)
    { // TODO: work DataOut and DataIn and Flags into this
      string acopkey = acop != null ? acop.GetHashCode().ToString() : "unbound";
      string dkey = "o:" + makeObjectTag(DataOut) + "i:" + makeObjectTag(DataIn);
      string key = acopkey + "://" + Protocol + "/" + Address + "/" + Property + "::" + dkey;
      // if DataOut and DataIn == null, Flags = 0 then look for 'something' and if found return it
      return key;
    }
    public string GetConnectionKey()
    {
      return linkKey;
    }
    public string GetConnectionKey(String Protocol, String Address, String Property, object DataOut, object DataIn, int Flags, object acop)
    { // add the link key 'after the fact'
      return linkKey = MakeConnectionKey(Protocol, Address, Property, DataOut, DataIn, Flags, acop);
    }
    public string GetConnectionAddress(String Protocol, String Context, String Subsystem, String Server, String Device)
    {
      if (Protocol == null) Protocol = prot;
      if (Protocol.Equals("TINE", StringComparison.OrdinalIgnoreCase))
      {
        if (Context == null || Context.Length == 0) throw new ArgumentNullException("<Context> cannot be omitted!");
        if (Server == null || Server.Length == 0) throw new ArgumentNullException("<Server> cannot be omitted!");
        if (Device == null || Device.Length == 0) return "/" + Context + "/" + Server;
        return "/" + Context + "/" + Server + "/" + Device;
      }
      if (Protocol.Equals("SIMULATE", StringComparison.OrdinalIgnoreCase))
      {
        if (Context == null) Context = "";
        if (Subsystem == null) Subsystem = "";
        if (Server == null) Server = "";
        if (Device == null) Device = "";
        return "/" + Context + "/" + Subsystem + "/" + Server + "/" + Device;
      }
      throw new NotImplementedException();
    }
    public string GetConnectionHostAddress(String Protocol, String Address, String Property)
    {
      if (Protocol == null) Protocol = prot;
      if (Protocol.Equals("TINE", StringComparison.OrdinalIgnoreCase))
      {
        if (Address == null || Address.Length == 0) return null;
        String[] parts = Address.Split(new String[] { "/" }, StringSplitOptions.RemoveEmptyEntries);
        if (parts.Length < 2) return null;
        return "/"+parts[0]+"/"+parts[1];
      }
      if (Protocol.Equals("SIMULATE", StringComparison.OrdinalIgnoreCase))
      {
        return Address;
      }
      throw new NotImplementedException();

    }
    public int GetFlags()
    {
      return flgs;
    }
    private char[] addrSplitTok = new char[] { '/' };
    public int LockAccess(String Address)
    {
      if (prot.Equals("TINE", StringComparison.OrdinalIgnoreCase))
      {
        if (Address == null || !Address.StartsWith("/")) return Errors.invalid_name;
        String[] parts = Address.Split(addrSplitTok, StringSplitOptions.RemoveEmptyEntries);
        int lockType = AccessLock.LOCK_PERSISTENT | AccessLock.LOCK_NOCALLBACKS;
        int cc = TKernel.SetAccessLock(parts[0], parts[1], lockType, AccessLock.MAX_LOCK_DURATION);
        if (cc != 0)
        {
          Char[] usr = new Char[32], ip = new Char[32], tleft = new Char[32];
          String msg = "Cannot acquire lock : " + Errors.ToString(cc);
          if (TKernel.GetAccessLockInformation(parts[0], parts[1], usr, ip, tleft) == 0)
          {
            int len = 0; while (usr[++len] != 0) ;
            String usrs = new string(usr, 0, len);
            if (len > 0)
            {
              len = 0; while (ip[++len] != 0) ;
              String ips = new string(ip, 0, len);
              msg = Address + " is locked by user " + usrs + " at address " + ips;
            }
          }
          System.Windows.Forms.MessageBox.Show(msg);
        }
        return cc;
      }
      return -1;
    }
    public int UnlockAccess(String Address)
    {
      if (prot.Equals("TINE", StringComparison.OrdinalIgnoreCase))
      {
        if (Address == null || !Address.StartsWith("/")) return Errors.invalid_name;
        String[] parts = Address.Split(addrSplitTok, StringSplitOptions.RemoveEmptyEntries);
        return TKernel.FreeAccessLock(parts[0], parts[1]);
      }
      return -1;
    }
    private Dictionary<String, DisplaySettings> dspDict = new Dictionary<string, DisplaySettings>();
    private String getDispDictKey(String Address,String Property)
    {
      if (Address == null || Address.Length == 0) return null;
      if (Property == null || Property.Length == 0) return null;
      String[] parts = Address.Split(new char[] { '/' }, StringSplitOptions.RemoveEmptyEntries);
      if (parts.Length < 2) return null;
      return prot + "/" + parts[0] + "/" + parts[1] + "[" + Property + "]";
    }
    public DisplaySettings GetDisplaySettings(String Address, String Property)
    {
      if (prot.Equals("SIMULATE", StringComparison.OrdinalIgnoreCase))
      {
        double ymin = 0, ymax = 100.0, xmin = 0, xmax = 0;
        String yunits = "", xunits = "", dsc = "";
        int atype = 0;
        return new DisplaySettings(ymin, ymax, xmin, xmax, yunits, xunits, atype, dsc);
      }
      if (prot.Equals("TINE", StringComparison.OrdinalIgnoreCase))
      {
        double ymin = 0, ymax = 100.0, xmin = 0, xmax = 0;
        double shift = 0, scale = 1;
        String yunits = "", xunits = "", dsc = "";
        int atype = 0;
        AcopLink lnk = AcopLinkFactory.GetAcopLink(GetConnectionKey());
        if (lnk != null && lnk.dsp != null)
        {
          int fmt = TDataType.GetTineFormat(lnk.dout);
          xmax = (lnk.dout is Array) ? ((Array)lnk.dout).Length : 1;
          if (Formats.IsSimpleString(fmt)) xmin = xmax = 1;
          try
          {
            if (Address.Equals(addr, StringComparison.OrdinalIgnoreCase) &&
                Property.Equals(prop, StringComparison.OrdinalIgnoreCase))
            {
              return lnk.dsp;
            }
          }
          catch (Exception) { }
        }
        String key = getDispDictKey(Address, Property);
        if (key != null)
        {
          DisplaySettings value = null;
          if (dspDict.TryGetValue(key, out value)) return value;
        }
        PrpQueryStruct[] pqs = new PrpQueryStruct[1];
        int n = 1, cc = -1;
        uint to = 100;
        // explicit retry as this is important ! 
        // -> use the acop component lateInitialization timer to retry ?
        for (int tries = 0; cc != 0 && tries < 2; tries++)
        {
          cc = TQuery.GetPropertyInformation(Address, Property, ref pqs, ref n, to);
          to += 100;
        }
        uint cap = 0;
        DisplaySettings.DataType typ = DisplaySettings.DataType.EMPTY;
        if (cc == 0)
        {
          ymin = pqs[0].prpMinValue; ymax = pqs[0].prpMaxValue;
          xmin = pqs[0].rngMinValue; xmax = pqs[0].rngMaxValue;
          if (xmin == 0 && xmax == 0) xmax = pqs[0].prpSize;
          yunits = new String(pqs[0].prpUnits); xunits = new String(pqs[0].rngUnits);
          atype = pqs[0].prpArrayType;
          dsc = new String(pqs[0].prpDescription);
          shift = xmin;
          scale = (xmax - xmin) / (pqs[0].prpSize);
          cap = pqs[0].prpSize;
          byte fmt = pqs[0].prpFormat;
          if (pqs[0].prpFormat == Formats.CF_STRUCT)
          { // finesse the structure information this way ...
            atype |= ArrayType.AT_STRUCTURED;
            NAME64DBLDBL[] ndd = new NAME64DBLDBL[512];
            // Encoding.GetEncoding("UTF-8").GetBytes(chararray);
            sbyte[] tag = new sbyte[16];
            for (int i = 0; i < 16; i++) tag[i] = (sbyte)pqs[0].prpTag[i];
            TLink slnk = new TLink(Address, "STRUCTFORMAT", new TDataType(ndd), new TDataType(tag), Access.CA_READ|Access.CA_NOCALLBACKS);
            if (slnk.Execute() == 0)
            {
              int nflds = slnk.GetCompletionDataSize();
              if (nflds > 1)
              {
                nflds--; // subtract sealing info
                cap = (uint)nflds;
              }
            }
          } 
          switch (fmt)
          {
            case Formats.CF_NULL: break;
            case Formats.CF_INT16: goto case Formats.CF_DOUBLE;
            case Formats.CF_INT32: goto case Formats.CF_DOUBLE;
            case Formats.CF_INT64: goto case Formats.CF_DOUBLE;
            case Formats.CF_FLOAT: goto case Formats.CF_DOUBLE;
            case Formats.CF_DOUBLE:
              typ = DisplaySettings.DataType.NUMBER;
              break;
            case Formats.CF_USTRING:
              typ = DisplaySettings.DataType.STRING;
              break;
            case Formats.CF_STRUCT:
              typ = DisplaySettings.DataType.STRUCTURE;
              break;
            case Formats.CF_AIMAGE: goto case Formats.CF_IMAGE;
            case Formats.CF_IMAGE:
              typ = DisplaySettings.DataType.IMAGE;
              break;
            default:
              if (Formats.IsSimpleString(fmt))
              {
                typ = DisplaySettings.DataType.STRING;
                break;
              }
              typ = DisplaySettings.DataType.COMPOUND;
              break;
          }
        }
        if (lnk != null && Formats.IsSimpleString(TDataType.GetTineFormat(lnk.dout))) xmin = xmax = 1;
        DisplaySettings dsp = new DisplaySettings(ymin, ymax, xmin, xmax, yunits, xunits, atype, dsc);
        if (scale != 1) dsp.SetXScale(scale);
        if (shift != 0) dsp.SetXShift(shift);
        if (cap > 0) dsp.SetCapacity((int)cap);
        dsp.SetDataType(typ);
        if (lnk != null) lnk.dsp = dsp;
        dsp.SettingsStale(cc != 0);
        if (cc == 0 && key != null) dspDict.Add(key, dsp);
        return dsp;
      }
      return null;
    }
    private Boolean[] int2bool(Object o)
    {
      if (o is Boolean[]) return (Boolean[])o;
      if (!(o is Int32[])) return null;
      Int32[] vals = (Int32[])o;
      Boolean[] bo = new Boolean[vals.Length];
      for (int i = 0; i < vals.Length; i++) bo[i] = vals[i] == 0 ? false : true;
      return bo;
    }
    public object GetOutputDataObject(String Address, String Property, Object Reference, Int32 Capacity)
    {
      if (prot.Equals("TINE", StringComparison.OrdinalIgnoreCase))
      {
        if (Address == null || Property == null) return null;
        AcopLink lnk = AcopLinkFactory.GetAcopLink(GetConnectionKey());
        ushort fmt = Reference == null ? Formats.CF_DOUBLE : TDataType.GetTineFormat(Reference);
        int siz = Capacity <= 0 ? 256 : Capacity;
        if (lnk != null)
        {
          try
          {
            if (Address.Equals(addr, StringComparison.OrdinalIgnoreCase) &&
                Property.Equals(prop, StringComparison.OrdinalIgnoreCase))
            {
              if (lnk.tineLinkFormat == Formats.CF_BOOLEAN)
              {
                Boolean[] bo = int2bool(lnk.dout);
                if (bo != null) return bo;
              }
              return lnk.dout;
            }
          }
          catch (Exception) { }
          if (Reference == null) fmt = TDataType.GetTineFormat(lnk.dout);
          if (Capacity <= 0) siz = (lnk.dout is Array) ? ((Array)lnk.dout).Length : 1;
        }
        string tag = null;
        if (Reference == null || Capacity <= 0)
        {
          PrpQueryStruct[] pqs = new PrpQueryStruct[1];
          int n = 1;
          //TKernel.SetSuspendCallbacks(true);
          int cc = TQuery.GetPropertyInformation(Address, Property, ref pqs, ref n);
          //TKernel.SetSuspendCallbacks(false);
          //if (cc != 0) throw new Exception("cannot acquire property information : " + Errors.ToString(cc));
          if (cc == 0)
          {
            if (Reference == null) fmt = pqs[0].prpFormat;
            if (Capacity <= 0 || fmt == Formats.CF_CHAR) siz = (int)pqs[0].prpSize;
            tag = new String(pqs[0].prpTag).Trim(new char[] { '\0' });
            if (fmt == Formats.CF_STRUCT && tag.Length > 0)
            {
              TQuery.AcquireAndRegisterStruct(Address, tag, (int)pqs[0].prpSize);
            }
          }
        }
        tineLinkFormat = fmt;
        object o = TDataType.MakeDataObject(fmt, siz, tag);
        if (fmt == Formats.CF_BOOLEAN)
        {
          Boolean[] bo = int2bool(o);
          if (bo != null) return bo;
        }
        return o;
      }
      return null;
    }
    public object GetOutputDataObject(String Address, String Property)
    {
      return GetOutputDataObject(Address, Property, null, 0);
    }
    public override string ToString()
    {
      if (linkKey != null) return linkKey;
      return "unitialized @" + GetHashCode();
    }
    public Boolean IsInitialized()
    {
      return linkKey != null;
    }
    public AcopConnection GetConnection(String Protocol, String Address, String Property, object DataOut, object DataIn, int Flags, object acop)
    {
      return GetConnection(Protocol, Address, Property, DataOut, -1, -1, DataIn, Flags, acop);
    }
    public AcopConnection GetConnection(String Protocol, String Address, String Property, object DataOut, int offset, int length, object DataIn, int Flags, object acop)
    {
      String key = MakeConnectionKey(Protocol, Address, Property, DataOut, DataIn, Flags, acop);
      AcopLink lnk = AcopLinkFactory.GetAcopLink(key);
      if (lnk != null)
      {
        lnk.din = DataIn;
        lnk.flgs = Flags;
        lnk.dout = DataOut;
        return lnk;
      }
      linkKey = key;
      prot = Protocol;
      addr = Address;
      prop = Property;
      dout = DataOut;
      doutOffset = offset;
      doutLength = length;
      din = DataIn;
      flgs = Flags;
      AcopLinkFactory.AddAcopLink(key, this);
      return this;
    }
    private static DateTime DateTimeOrigin = new DateTime(1970, 1, 1, 0, 0, 0, 0);
    public static DateTime ConvertFromUnixTimestamp(double timestamp)
    {
      return DateTimeOrigin.AddSeconds(timestamp);
    }

    public static double ConvertToUnixTimestamp(DateTime date)
    {
      DateTime origin = new DateTime(1970, 1, 1, 0, 0, 0, 0);
      TimeSpan diff = date.ToUniversalTime() - origin;
      return Math.Floor(diff.TotalSeconds);
    }
    public void UpdateData(Object data)
    { // simulate
      if (data.GetType() == dout.GetType())
      {
        if (data.GetType().IsArray)
        {
          if (data.GetType().GetElementType() != dout.GetType().GetElementType()) return;
          if (doutOffset < 0 || doutLength < 1)
          {
            long len = Math.Min(((Array)data).GetLength(0), ((Array)dout).GetLength(0));
            Array.Copy((Array)data, (Array)dout, len);
          }
          else
          {
            long len = Math.Min(((Array)data).GetLength(0), doutLength);
            Array.Copy((Array)data, 0, (Array)dout, doutOffset, len);
          }
          return;
        }
        dout = data;
      }
    }
    public double[] GetXAxis()
    {
      if (prot.Equals("TINE", StringComparison.OrdinalIgnoreCase))
      {
        return TDataType.GetXAxis(GetOutputDataObject(addr, prop), lastLinkCompletionSize);
      }
      return null;
    }
    public double[] GetYAxis()
    {
      if (prot.Equals("TINE", StringComparison.OrdinalIgnoreCase))
      {
        return TDataType.GetYAxis(GetOutputDataObject(addr, prop), lastLinkCompletionSize);
      }
      return null;
    }
    public string[] GetDataAsString()
    {
      if (prot.Equals("TINE", StringComparison.OrdinalIgnoreCase))
      {
        object dout = GetOutputDataObject(addr, prop);
        if (dout is DUSTRING[])
        {
          DUSTRING[] uout = (DUSTRING[])dout;
          float[] fout = new float[uout.Length];
          for (int i = 0; i < uout.Length; i++) fout[i] = (float)uout[i].GetPrimaryFloat();
          dout = fout;
        }
        return TDataType.GetDataAsStringArray(dout);
      }
      return null;
    }
    public double[] GetDataAsNumber()
    {
      if (prot.Equals("TINE", StringComparison.OrdinalIgnoreCase))
      {
        object dout = GetOutputDataObject(addr, prop);
        if (dout == null) return null;
        Type type = (dout is Array) ? ((Array)dout).GetType().GetElementType() : dout.GetType();
        if (!ValueIsNumber(type)) return TDataType.GetYAxis(dout); //return null;
        try
        {
          if (!(dout is Array)) return new double[] { Convert.ToDouble(dout) };
          Array va = (Array)dout;
          double[] vals = new double[va.Length];
          for (int i = 0; i < vals.Length; i++)
          {
            vals[i] = Convert.ToDouble(va.GetValue(i));
          }
          return vals;
        }
        catch (Exception) { }
      }
      return null;
    }
    public Int32[] GetDataAsInteger()
    {
      if (prot.Equals("TINE", StringComparison.OrdinalIgnoreCase))
      {
        object dout = GetOutputDataObject(addr, prop);
        if (dout == null) return null;
        Type type = (dout is Array) ? ((Array)dout).GetType().GetElementType() : dout.GetType();
        if (!ValueIsNumber(type)) return null;
        try
        {
          if (!(dout is Array)) return new Int32[] { Convert.ToInt32(dout) };
          Array va = (Array)dout;
          Int32[] vals = new Int32[va.Length];
          for (int i = 0; i < vals.Length; i++)
          {
            vals[i] = Convert.ToInt32(va.GetValue(i));
          }
          return vals;
        }
        catch (Exception) { }
      }
      return null;
    }
    private static Boolean ValueIsNumber(Type type)
    {
      switch (Type.GetTypeCode(type))
      {
        case TypeCode.Double: return true;
        case TypeCode.Int16: return true;
        case TypeCode.Int32: return true;
        case TypeCode.Int64: return true;
        case TypeCode.Single: return true;
        case TypeCode.UInt16: return true;
        case TypeCode.UInt32: return true;
        case TypeCode.UInt64: return true;
      }
      return false;
    }
    public void FillDataObjectFromString(object src, String data)
    {
      if (src == null) return;
      if (data == null) data = "";
      Type type = (src is Array) ? ((Array)src).GetType().GetElementType() : src.GetType();
      try
      {
        if (prot.Equals("TINE", StringComparison.OrdinalIgnoreCase))
        {
          if (ValueIsNumber(type))
          {
            double v = Double.Parse(data);
            object vo = Convert.ChangeType(v, type);
            if (src is Array)
            {
              int len = ((Array)src).Length;
              for (int i = 0; i < len; i++) ((Array)src).SetValue(vo, i);
            }
            else
            {
              src = vo;
            }
            return;
          }
          if (Type.GetTypeCode(type) == TypeCode.SByte)
          {
            char[] co = data.ToCharArray();
            sbyte[] sbo = new sbyte[co.Length];
            for (int i = 0; i < co.Length; i++) sbo[i] = (sbyte)co[i];
            if (src is Array)
            {
              int len = ((Array)src).Length;
              sbyte sbv = (sbyte)0;
              for (int i = 0; i < len; i++)
              {
                sbv = (i < co.Length) ? sbo[i] : (sbyte)0;
                ((Array)src).SetValue(sbv, i);
              }
            }
            else
            {
              src = sbo[0];
            }
            return;
          }
          else if (Type.GetTypeCode(type) == TypeCode.Char)
          {
            char[] co = data.ToCharArray();
            if (src is Array)
            {
              int len = ((Array)src).Length;
              char cv = (char)0;
              for (int i = 0; i < len && i < co.Length; i++)
              {
                cv = (i < co.Length) ? co[i] : (char)0;
                ((Array)src).SetValue(cv, i);
              }
            }
            else
            {
              src = co[0];
            }
            return;
          }
          else
          {
            object vo = Activator.CreateInstance(type, data);
            if (src is Array)
            {
              int len = ((Array)src).Length;
              for (int i = 0; i < len; i++) ((Array)src).SetValue(vo, i);
            }
            else
            {
              src = vo;
            }
            return;
          }
        }
      }
      catch (Exception) { }
      return;
    }
    public object GetDataObjectFromStringArray(Type type, String[] data)
    {
      if (data == null || data.Length == 0) return null;
      if (type == null) return null;
      try
      {
        if (prot.Equals("TINE", StringComparison.OrdinalIgnoreCase))
        {
          int dlen = data.Length;
          if (ValueIsNumber(type))
          {
            double v = 0;
            var ov = Array.CreateInstance(type, dlen);
            for (int i = 0; i < dlen; i++)
            {
              v = Double.Parse(data[i]);
              object vo = Convert.ChangeType(v, type);
              ov.SetValue(vo, i);
            }
            return ov;
          }
          if (Type.GetTypeCode(type) == TypeCode.SByte)
          { // just do the 1st array element for now (fix this later?)
            char[] co = data[0].ToCharArray();
            sbyte[] sbo = new sbyte[co.Length];
            for (int i = 0; i < co.Length; i++) sbo[i] = (sbyte)co[i];
            return sbo;
          }
          else if (Type.GetTypeCode(type) == TypeCode.Char)
          { // just do the 1st array element for now (fix this later?)
            return data[0].ToCharArray();
          }
          else if (Type.GetTypeCode(type) == TypeCode.Boolean)
          {
            Boolean[] bv = new Boolean[dlen];
            for (int i = 0; i < dlen; i++)
            {
              bv[i] = false;
              if (data[i].Equals("TRUE", StringComparison.OrdinalIgnoreCase) ||
                  data[i].Equals("ON", StringComparison.OrdinalIgnoreCase) ||
                  data[i].Equals("YES", StringComparison.OrdinalIgnoreCase))
              {
                bv[i] = true;
              }
            }
            return bv;
          }
          else
          {
            AcopLink lnk = AcopLinkFactory.GetAcopLink(GetConnectionKey());
            if (type == typeof(TTaggedStruct) || lnk.tineLinkFormat == Formats.CF_STRUCT)
            {
              TTaggedStruct[] tts = new TTaggedStruct[1];
              String dtag = lnk.tineLink.GetOutputDataType().GetDataTag();
              tts[0] = new TTaggedStruct(dtag, data);
              return tts;
            }
            var ov = Array.CreateInstance(type, dlen);
            for (int i = 0; i < dlen; i++)
            {
              object vo = Activator.CreateInstance(type, data[i]);
              ov.SetValue(vo, i);
            }
            return ov;
          }
        }
      }
      catch (Exception) { }
      return null;
    }
    public object GetDataObjectFromString(Type type, String data)
    {
      if (data == null) return null;
      if (type == null) return null;
      try
      {
        if (prot.Equals("TINE", StringComparison.OrdinalIgnoreCase))
        {
          if (ValueIsNumber(type))
          {
            double v = Double.Parse(data);
            object vo = Convert.ChangeType(v, type);
            var ov = Array.CreateInstance(type, 1);
            ov.SetValue(vo, 0);
            return ov;
          }
          if (Type.GetTypeCode(type) == TypeCode.SByte)
          {
            char[] co = data.ToCharArray();
            sbyte[] sbo = new sbyte[co.Length];
            for (int i = 0; i < co.Length; i++) sbo[i] = (sbyte)co[i];
            return sbo;
          }
          else if (Type.GetTypeCode(type) == TypeCode.Char)
          {
            return data.ToCharArray();
          }
          else if (Type.GetTypeCode(type) == TypeCode.Boolean)
          {
            Boolean bv = false;
            if (data.Equals("TRUE", StringComparison.OrdinalIgnoreCase) ||
                data.Equals("ON", StringComparison.OrdinalIgnoreCase) ||
                data.Equals("YES", StringComparison.OrdinalIgnoreCase))
            {
              bv = true;
            }
            return new Boolean[] { bv };
          }
          else
          {
            object vo = Activator.CreateInstance(type, data);
            var ov = Array.CreateInstance(type, 1);
            ov.SetValue(vo, 0);
            return ov;
          }
        }
      }
      catch (Exception) { }
      return null;
    }
    public string[] GetChannelNames()
    {
      if (prot.Equals("TINE", StringComparison.OrdinalIgnoreCase))
      {
        String[] nams = TDataType.GetChannelNames(GetOutputDataObject(addr, prop), lastLinkCompletionSize);
        if (nams == null)
        {
          NAME64[] vals = new NAME64[lastLinkCompletionSize];
          try
          {
            nams = TQuery.GetSystemDevices(addr, prop);
          }
          catch (Exception)
          {
            nams = null;
          }
        }
        return nams;
      }
      return null;
    }
    private static void OnTimedCbEvent(Object source, System.Timers.ElapsedEventArgs e)
    { // simulate data update:
      // update timestamp:
      instance.SetLinkTimestamp(ConvertToUnixTimestamp(System.DateTime.UtcNow));
      instance.SetLinkCompletionSize(100);
      // update output data:
      double[] vals = new double[100];
      Random r = new Random();
      for (int i = 0; i < 100; i++) vals[i] = 100.0 * r.NextDouble();
      instance.UpdateData(vals);
      if (instance.acopCallback != null) instance.acopCallback.callback(instance);
    }
    public int Attach(int Flags, int interval, object reference, AcopLinkCallback cb)
    {
      flgs |= Flags;
      intvl = interval;
      instance = (AcopLink)reference;
      acopCallback = cb;

      if (prot.Equals("SIMULATE", StringComparison.OrdinalIgnoreCase))
      {
        if (cbTimer == null)
        {
          cbTimer = new System.Timers.Timer(500);
          cbTimer.Elapsed += OnTimedCbEvent;
          cbTimer.Enabled = true;
        }
        return 0;
      }
      if (prot.Equals("TINE", StringComparison.OrdinalIgnoreCase))
      {
        int access = (flgs & 0xff);
        access &= ~Access.CA_NOCALLBACKS; // callback MUST be enabled !
        if (!Access.IsWrite((ushort)access)) access |= Access.CA_READ; // make sure of this
        int mode = (flgs >> 16);
        if ((mode & 0x0f) == 0) mode |= Modes.CM_TIMER;
        mode &= ~Modes.CM_GROUPED; // make sure this is there!  groupping is done via acop!
        if (!Modes.IsSingle((UInt16)mode))
        { // WRITE must be a single link 
          access &= ~Access.CA_WRITE;
        }
        tineLinkMode = mode;
        tineLinkAccess = access;
        if (tineLink == null)
        {
          TDataType tdt = null;
          if (doutOffset < 0 && doutLength < 1)
          {
            tdt = new TDataType(dout);
          }
          else
            tdt = new TDataType((Array)dout, doutOffset, doutLength);
          tineLink = new TLink(addr, prop, tdt, new TDataType(din), (ushort)tineLinkAccess);
        }
        int lid = tineLink.Attach((uint)mode, tineCallback, intvl);
        if (lid < 0) sts = -lid;
        return lid;
      }
      throw new NotImplementedException("transport protocol " + prot + " is not supported");
    }
    public int Detach()
    {
      int cc = 0;
      if (prot.Equals("TINE", StringComparison.OrdinalIgnoreCase))
      {
        if (tineLink != null)
        {
          cc = tineLink.Close();
          tineLink = null;
          AcopLinkFactory.removeAcopLink(linkKey);
        }
      }
      return cc;
    }
    public int Execute(int Flags, int timeout)
    {
      int cc = 0;
      if (prot.Equals("SIMULATE", StringComparison.OrdinalIgnoreCase))
      {
        // update timestamp:
        instance.SetLinkTimestamp(ConvertToUnixTimestamp(System.DateTime.UtcNow));
        // update output data:
        double[] vals = new double[100];
        Random r = new Random();
        for (int i = 0; i < 100; i++) vals[i] = 100.0 * r.NextDouble();
        instance.UpdateData(vals);
        return sts = 0;
      }
      if (prot.Equals("TINE", StringComparison.OrdinalIgnoreCase))
      {
        int access = (flgs & 0xff);
        if (!Access.IsWrite((ushort)access)) access |= Access.CA_READ; // make sure of this
        tineLinkAccess = access;
        int mode = (flgs >> 16);
        if ((mode & 0x0f) == 0) mode |= Modes.CM_SINGLE;
        tineLinkMode = mode;
        if (tineLink == null)
        {
          TDataType tdt = null;
          if (doutOffset < 0 && doutLength < 1)
            tdt = new TDataType(dout);
          else
            tdt = new TDataType((Array)dout, doutOffset, doutLength);
          access |= Access.CA_NOCALLBACKS;
          tineLink = new TLink(addr, prop, tdt, new TDataType(din), (ushort)access);
        }
        stsLastUpdate = sts;
        sts = tineLink.Execute((ushort)mode, (uint)timeout);
        ts = tineLink.GetCompletionDataTimeStamp();
        SetLinkCompletionSize(tineLink.GetCompletionDataSize());
        SetLinkTag((int)tineLink.GetCompletionSystemStamp());
        stsString = tineLink.GetStatusString();
        return sts;
      }
      return sts = cc;
    }

    public int GetHistoryPoints(double[][] data, ref int length, double StartTime, double StopTime, int Flags, int timeout)
    {
      int dlen = 0;
      if (data.Length != 2) return -1;
      dlen = data[0].Length;
      if (data[1].Length != dlen) return -1;
      if (length < dlen) dlen = length;
      int cc = 0;
      if (prot.Equals("SIMULATE", StringComparison.OrdinalIgnoreCase))
      {
        // update timestamp:
        instance.SetLinkTimestamp(ConvertToUnixTimestamp(System.DateTime.UtcNow));
        // update output data:
        double[] vals = new double[100];
        Random r = new Random();
        int raster = (int)(StopTime - StartTime) / dlen;
        for (int i = 0; i < dlen; i++)
        {
          data[0][i] = 100.0 * r.NextDouble();
          data[1][i] = StartTime + i * raster;
        }
        return 0;
      }
      if (prot.Equals("TINE", StringComparison.OrdinalIgnoreCase))
      {
        if (tineLink != null)
        {
          if (addr == null || addr.Length == 0) return Errors.invalid_parameter;
          String[] parts = addr.Split(new String[] { "/" }, StringSplitOptions.RemoveEmptyEntries);
          if (parts.Length < 3) return Errors.invalid_name;
          THistory.HstSrc src = new THistory.HstSrc("", "", 0);
          DBLDBL[] hdout = new DBLDBL[dlen];
          int sample = 0;
          int index = 0;
          int fieldindex = 0;
          if ((Flags & 0x10000) == 0x10000) sample = 1; // no raster
          index = (Flags & 0xffff);
          try
          {
            if (tineLink.GetOutputDataType().GetDataFormat() == Formats.CF_STRUCT)
            {
              fieldindex = index;
              index = 0;
            }
          }
          catch (Exception) {}
          TKernel.SetSuspendCallbacks(true);
          cc = THistory.GetArchivedData(parts[0], parts[1], prop, parts[2], index, fieldindex, sample, StartTime, StopTime, hdout, src, (uint)timeout, false);
          TKernel.SetSuspendCallbacks(false);
          //cc = THistory.GetArchivedData(parts[0], parts[1], prop, parts[2], StartTime, StopTime, hdout, src, (uint)timeout, false);
          // hdout to double[][] ....
          if (cc == 0)
          {
            length = src.getNumberReturned();
            for (int i=0; i<length; i++)
            {
              data[0][i] = hdout[i].d1val;
              data[1][i] = hdout[i].d2val;
            }
          }
          return cc;
        }
        return Errors.invalid_link;
      }
      return cc;
    }

    public int GetImage(ref Bitmap bt, ref String header)
    {
      int cc = 0;

      if (prot.Equals("TINE", StringComparison.OrdinalIgnoreCase))
      {
        AcopLink lnk = AcopLinkFactory.GetAcopLink(GetConnectionKey());
        if (lnk != null)
        {
          if(TDataType.GetTineFormat(lnk.dout) == Formats.CF_IMAGE)
          {
            DIMAGE[] dimg = (DIMAGE[])lnk.dout;
            byte[] fb = dimg[0].GetFrameBuffer();
            TImageHeader hdr = dimg[0].GetFrameHeader();
            header = hdr.ToString();

            switch (hdr.imageFormat)
            {
              case ImageFormat.IF_GRAY:
                if(hdr.bytesPerPixel == 1 && hdr.effectiveBitsPerPixel == 8)
                {
                  unsafe
                  {
                    fixed (byte* ptr = fb)
                    {
                      bt = new Bitmap(hdr.sourceWidth, hdr.sourceHeight, hdr.sourceWidth, System.Drawing.Imaging.PixelFormat.Format8bppIndexed, new IntPtr(ptr));
                    }
                  }
                  System.Drawing.Imaging.ColorPalette cp = bt.Palette;
                  Color[] c = cp.Entries;
                  for (int i = 0; i < 256; i++)
                  {
                    c[i] = Color.FromArgb((byte)i, (byte)i, (byte)i);
                  }
                  bt.Palette = cp;
                }
                else if(hdr.bytesPerPixel == 2 && hdr.effectiveBitsPerPixel >= 9 && hdr.effectiveBitsPerPixel <= 16)
                {
                  bt = ConvertGrayScale(fb, hdr.sourceWidth, hdr.sourceHeight, hdr.effectiveBitsPerPixel);
                }
                else
                {
                  lnk.SetLinkStatusString("Unknown image format");
                  return 512;
                }
                break;
              case ImageFormat.IF_JPEG:
                MemoryStream ms = new MemoryStream(fb, 0, fb.Length);
                ms.Write(fb, 0, fb.Length);
                Image img = Image.FromStream(ms);
                ms.Dispose();
                bt = new Bitmap(img, hdr.sourceWidth, hdr.sourceHeight);
                break;
              case ImageFormat.IF_RGB:
                //TODO
                break;
              default:
                lnk.SetLinkStatusString("Unknown image format");
                return 512;
            }
          }
          return cc;
        }
        return Errors.invalid_link;
      }
      return cc;
    }
    //----------------------------------------
    // from: https://stackoverflow.com/questions/6832948/c-sharp-converting-8-bit-or-16-bit-grayscale-raw-pixel-data
    private Bitmap ConvertGrayScale(byte[] input, int width, int height, int bits)
    {
      Bitmap bMap= new Bitmap(width, height, System.Drawing.Imaging.PixelFormat.Format32bppArgb);
      Rectangle dim = new Rectangle(0, 0, width, height);

      uint[] lut = CreatLUT(bits);
      System.Drawing.Imaging.BitmapData bData = bMap.LockBits(dim, System.Drawing.Imaging.ImageLockMode.WriteOnly, bMap.PixelFormat);
      ConvertCore(width, height, bits, input, bData, lut);
      bMap.UnlockBits(bData);

      return bMap;
    }
    private uint[] CreatLUT(int bits)
    {
      int maxInput = 1 << bits;
      uint[] lut = new uint[maxInput];

      for(int i = 0; i < maxInput; i++)
      {
        byte intensity = (byte)((i * 0xFF) / maxInput);
        lut[i] = (uint)(0xFF000000L | (intensity * 0x00010101L));
      }

      return lut;
    }
    private unsafe void ConvertCore(int width, int height, int bits, byte[] input, System.Drawing.Imaging.BitmapData output, uint[] lut)
    {
      ushort mask = (ushort)((1 << bits) - 1);
      int in_stride = output.Stride;
      int out_stride = width * 2;
      byte* out_data = (byte*)output.Scan0;

      fixed (byte* in_data = input)
      {
        for (int y = 0; y < height; y++)
        {
          uint* out_row = (uint*)(out_data + (y * in_stride));
          ushort* in_row = (ushort*)(in_data + (y * out_stride));

          for (int x = 0; x < width; x++)
          {
            ushort in_pixel = (ushort)(in_row[x] & mask);
            out_row[x] = lut[in_pixel];
          }
        }
      }
    }
    //----------------------------------------
    public const String UNAVAILABLE = "<unavailable>";
    public string GetConnectionProtocol()
    {
      return prot;
    }
    public string GetConnectionAddress()
    {
      return addr;
    }
    public string GetConnectionProperty()
    {
      return prop;
    }
    public string[] GetProtocolList()
    {
      return protlst;
    }
    private static String lastQueriedServer = null;
    //private static String[] lastPropertyList = null;
    //private static String[] lastDeviceList = null;
    public string[] GetContextList(string protocol)
    {
      if (protocol == null) protocol = prot;
      if (protocol.Equals("TINE", StringComparison.OrdinalIgnoreCase))
      { // get TINE contexts
        try
        {
          return TQuery.GetSystemContexts();
        }
        catch (Exception)
        {
          return new string[] { UNAVAILABLE };
        }
      }
      if (protocol.Equals("SIMULATE", StringComparison.OrdinalIgnoreCase))
      {
        return new string[1] { "TEST" };
      }
      throw new NotImplementedException();
    }
    public string[] GetSubSystemList(string protocol, string context)
    {
      if (protocol == null) protocol = prot;
      if (protocol.Equals("TINE", StringComparison.OrdinalIgnoreCase))
      { // get TINE subsystems for context
        if (context == null || context.Length == 0) throw new Exception("no context specified!");
        NAME16[] n16 = new NAME16[64];
        TDataType dout = new TDataType(n16);
        TLink lnk = new TLink("ENS/" + context, "SUBSYSTEMS", dout, null, Access.CA_READ);
        int len = (lnk.Execute(500) != 0) ? 0 : lnk.GetCompletionDataSize();
        string[] s = new string[len + 1]; s[0] = "ALL";
        for (int i = 0; i < len; i++) s[i + 1] = n16[i].GetName();
        return s;
      }
      if (protocol.Equals("SIMULATE", StringComparison.OrdinalIgnoreCase))
      {
        return new string[1] { "TEST" };
      }
      throw new NotImplementedException();
    }
    public string[] GetServerList(string protocol, string context, string subsystem)
    {
      if (protocol == null) protocol = prot;
      if (protocol.Equals("TINE", StringComparison.OrdinalIgnoreCase))
      { // get TINE servers for context and subsystem
        if (context == null || context.Length == 0) context = "DEFAULT";
        if (subsystem == null || subsystem.Length == 0 || subsystem.CompareTo("ALL") == 0)
        {
          try
          {
            return TQuery.GetSystemServers(context);
          }
          catch (Exception)
          {
            return new string[] { UNAVAILABLE };
          }
        }
        NAME32[] n32 = new NAME32[1024];
        NAME16[] ss = new NAME16[] { new NAME16(subsystem) };
        TDataType dout = new TDataType(n32);
        TDataType din = new TDataType(ss);
        TLink lnk = new TLink("ENS/" + context, "SERVERS", dout, din, Access.CA_READ);
        if (lnk.Execute(500) != 0) return new string[1] { "" };
        int len = lnk.GetCompletionDataSize(); //(int)dout.GetDataArrayLength();
        string[] s = new string[len];
        for (int i = 0; i < len; i++) s[i] = n32[i].GetName();
        return s;
      }
      if (protocol.Equals("SIMULATE", StringComparison.OrdinalIgnoreCase))
      {
        return new string[1] { "SimulatedServer" };
      }
      throw new NotImplementedException();
    }
    static DateTime utcThen = new DateTime(1970, 1, 1);
    public static double getUtcNow()
    {
      return (DateTime.Now.ToUniversalTime() - utcThen).TotalSeconds;
    }
    private String UnavailTarget = "";
    public string[] GetDeviceList(string protocol, string context, string subsystem, string server, string property)
    {
      if (protocol == null) protocol = prot;
      if (protocol.Equals("TINE", StringComparison.OrdinalIgnoreCase))
      { // get TINE devices for context, subsystem, server and property
        if (server == null) return new string[] { UNAVAILABLE };
        if (lastQueriedServer == null || lastQueriedServer.CompareTo(server) != 0)
        {
          lastQueriedServer = String.Copy(server);
        }
        //else if (lastDeviceList != null) return lastDeviceList;
        String target = context == null ? server : "/" + context + "/" + server;
        //if (target.CompareTo(UnavailTarget) == 0) return new string[] { UNAVAILABLE };
        //lastDeviceList = TQuery.GetSystemDevices(target, property);
        try
        {
          return TQuery.GetSystemDevices(target, property);
        }
        catch (Exception)
        {
          UnavailTarget = target;
          return new string[] { UNAVAILABLE };
        }
      }
      if (protocol.Equals("SIMULATE", StringComparison.OrdinalIgnoreCase))
      {
        return new string[1] { "SimulatedDevice" };
      }
      throw new NotImplementedException();
    }
    public Boolean HasPropertyQueryPrecedence(string protocol, string context, string subsystem, string server, string property)
    {
      if (protocol == null) protocol = prot;
      if (protocol.Equals("TINE", StringComparison.OrdinalIgnoreCase))
      { // get TINE devices for context, subsystem, server and property
        if (lastQueriedServer == null || lastQueriedServer.CompareTo(server) != 0)
        {
          lastQueriedServer = String.Copy(server);
        }
        String target = context == null ? server : "/" + context + "/" + server;
        //if (target.CompareTo(UnavailTarget) == 0) return false;
        Boolean p = false;
        try
        {
          UnavailTarget = target;
          TQuery.GetSystemDevices(target, null, ref p);
        }
        catch (Exception) { }
        return p;
      }
      return false;
    }
    public Boolean HasPropertyQueryPrecedence(string protocol, string context, string subsystem, string server)
    {
      if (protocol == null) protocol = prot;
      if (protocol.Equals("TINE", StringComparison.OrdinalIgnoreCase))
      { // get TINE devices for context, subsystem, server and property
        if (lastQueriedServer == null || lastQueriedServer.CompareTo(server) != 0)
        {
          lastQueriedServer = String.Copy(server);
        }
        String target = context == null ? server : "/" + context + "/" + server;
        //if (target.CompareTo(UnavailTarget) == 0) return false;
        Boolean p = false;
        try
        {
          TQuery.GetSystemDevices(target, null, ref p);
        }
        catch (Exception)
        {
          UnavailTarget = target;
        }
        return p;
      }
      return false;
    }
    public Boolean HasDeviceQueryPrecedence(string protocol, string context, string subsystem, string server)
    {
      if (protocol == null) protocol = prot;
      if (protocol.Equals("TINE", StringComparison.OrdinalIgnoreCase))
      { // get TINE devices for context, subsystem, server and property
        if (lastQueriedServer == null || lastQueriedServer.CompareTo(server) != 0)
        {
          lastQueriedServer = String.Copy(server);
        }
        String target = context == null ? server : "/" + context + "/" + server;
        //if (target.CompareTo(UnavailTarget) == 0) return false;
        Boolean p = false;
        try
        {
          TQuery.GetSystemProperties(target, ref p);
        }
        catch (Exception)
        {
          UnavailTarget = target;
        }
        return p;
      }
      return false;
    }
    public string[] GetPropertyList(string protocol, string context, string subsystem, string server, string device)
    {
      if (protocol == null) protocol = prot;
      if (protocol.Equals("TINE", StringComparison.OrdinalIgnoreCase))
      { // get TINE properties for context, subsystem, server and device
        String target = context == null ? server : "/" + context + "/" + server;
        if (device != null && device.Length > 0) target += "/" + device;
        //if (target.CompareTo(UnavailTarget) == 0) return new string[] { UNAVAILABLE };
        try
        {
          String[] prps = TQuery.GetSystemProperties(target);
          prps = prps.Where(v => !v.Contains(".HIST") && !v.Contains(".ARCH") && !v.Contains(".EGU") && !v.Contains(".NAM")).ToArray();
          return prps;
        }
        catch (Exception)
        {
          UnavailTarget = target;
          return new string[] { UNAVAILABLE };
        }
      }
      if (protocol.Equals("SIMULATE", StringComparison.OrdinalIgnoreCase))
      {
        return new string[1] { "SimulatedProperty" };
      }
      throw new NotImplementedException();
    }
    public IDictionary<string, int> GetFlagsList(String Protocol)
    {
      IDictionary<string, int> flagsEnum = null;
      if (Protocol.Equals("TINE", StringComparison.OrdinalIgnoreCase))
      {
        return AcopTransport.SelectList.Of<AcopTransport.tineFlags>();
      }
      if (Protocol.Equals("SIMULATE", StringComparison.OrdinalIgnoreCase))
      {
        return AcopTransport.SelectList.Of<AcopTransport.simFlags>();
      }
      return flagsEnum;
    }
    private string GetKey(ref ConTblInfo cti)
    {
      char[] keychars = new char[256];
      int i,k=0;
      keychars[k++] = '/';
      for (i = 0; i < 32; i++)
      {
        if (cti.context[i] == 0) break;
        keychars[k++] = cti.context[i];
      }
      keychars[k++] = '/';
      for (i = 0; i < 32; i++)
      {
        if (cti.expName[i] == 0) break;
        keychars[k++] = cti.expName[i];
      }
      keychars[k++] = '/';
      for (i = 0; i < 64; i++)
      {
        if (cti.devName[i] == 0) break;
        keychars[k++] = cti.devName[i];
      }
      keychars[k++] = '[';
      for (i = 0; i < 64; i++)
      {
        if (cti.prpName[i] == 0) break;
        keychars[k++] = cti.prpName[i];
      }
      keychars[k++] = ']';
      keychars[k] = '\0';
      return new String(keychars).Substring(0,k);
    }
    public string[] GetConnectionList(string protocol, Boolean activeOnly)
    {
      if (protocol == null) protocol = prot;
      if (protocol.Equals("TINE", StringComparison.OrdinalIgnoreCase))
      { 
        if (activeOnly)
        {
          ConTblInfo[] infTbl = TKernel.GetConnectionTable();
          String[] slst = new String[infTbl.Length];
          String key;
          for (int i = 0; i < slst.Length; i++)
          {
            key = GetKey(ref infTbl[i]);
            slst[i] = key.Trim() + " : " + Errors.ToString(infTbl[i].linkStatus);
          }
          return slst;
        }
        else
        {
          return TKernel.GetConnectionList();
        }
      }
      throw new NotImplementedException();
    }
    public string[] GetTransportMessages(string protocol)
    {
      if (protocol == null) protocol = prot;
      if (protocol.Equals("TINE", StringComparison.OrdinalIgnoreCase))
      { 
        return TKernel.GetMessageList();
      }
      throw new NotImplementedException();
    }
    public int GetTransportStatus(string protocol)
    {
      int n = 0;
      if (protocol.Equals("TINE", StringComparison.OrdinalIgnoreCase))
      { 
        ConTblInfo[] infTbl = TKernel.GetConnectionTable();
        foreach (ConTblInfo cti in infTbl)
        {
          if (!Errors.HasData(cti.linkStatus)) n++;
        }
      }
      return n;
    }
    public void StartDebuggingSession(string protocol)
    {
      if (protocol.Equals("TINE", StringComparison.OrdinalIgnoreCase))
      {
        native.LaunchRuntimeDebugWindow();
      }
    }
  }
}
