﻿using System;
using System.Collections.Generic;
using System.Collections.Specialized;
using System.Linq;
using System.ComponentModel;
using System.Windows.Forms;
using System.Text;
using System.Threading.Tasks;
using System.Drawing;

namespace AcopControl
{ // accessed from each Acop control (no protocol references)
  public partial class AcopLink
  {
    private static AcopLink anchor = null;
    //static AcopLink thisLink = null;
    //static Assembly asm = null;
    AcopTransport.AcopLink alnk = null;
    //object alnk = null;
    private Component thisAcop = null;
    public Component GetAcopControl() { return thisAcop; }
    public void SetAcopControl(Component acop) { thisAcop = acop; }
    public delegate void acopTransportUpdate(object sender, AcopLink lnk);
    public event acopTransportUpdate acopUpdated;
    private static int defaultInterval = 1000;
    public static int GetDefaultInterval() { return defaultInterval; }
    private static String defaultProtocol = null;
    public static String GetDefaultProtocol()
    {
      if (anchor == null) anchor = new AcopLink();
      if (defaultProtocol == null && anchor.alnk != null)
        defaultProtocol = anchor.alnk.GetProtocol();
      return defaultProtocol;
    }
    public static String GetDefaultContext(String protocol)
    {
      if (anchor == null) anchor = new AcopLink();
      if (defaultProtocol == null) defaultProtocol = anchor.alnk.GetProtocol();
      if (protocol == null) protocol = defaultProtocol;
      return anchor.alnk.GetDefaultContext(protocol);
    }
    public static String GetDefaultServer(String protocol)
    {
      if (anchor == null) anchor = new AcopLink();
      if (defaultProtocol == null) defaultProtocol = anchor.alnk.GetProtocol();
      if (protocol == null) protocol = defaultProtocol;
      return anchor.alnk.GetDefaultServer(protocol);
    }
    public static String GetDefaultDevice(String protocol)
    {
      if (anchor == null) anchor = new AcopLink();
      if (defaultProtocol == null) defaultProtocol = anchor.alnk.GetProtocol();
      if (protocol == null) protocol = defaultProtocol;
      return anchor.alnk.GetDefaultDevice(protocol);
    }
    public static String GetDefaultProperty(String protocol)
    {
      if (anchor == null) anchor = new AcopLink();
      if (defaultProtocol == null) defaultProtocol = anchor.alnk.GetProtocol();
      if (protocol == null) protocol = defaultProtocol;
      return anchor.alnk.GetDefaultProperty(protocol);
    }
    private String addr;  // address
    public String GetAddress() { return addr; }
    private String prop;  // property
    public String GetProperty() { return prop; }
    private String prot = null;  // protocol
    public String GetProtocol() { return prot; }
    public void SetProtocol(String protocol)
    {
      prot = protocol;
      if (alnk != null) alnk.SetProtocol(prot);
    }
    private int did = -1;  // display id
    public int GetDisplayId() { return did; }
    public void SetDisplayId(int id) { did = id; }
    private int lid = -1;      // link id
    public void SetIdle()
    {
      lid = -1;
    }
    public int GetLinkId() { return lid; }
    public Boolean IsAttached() { return lid < 0 ? false : true; }
    private Boolean callbackPending = false;
    public Boolean isCallbackPending() { return callbackPending; }
    public void setCallbackPending(Boolean value)
    {
      callbackPending = value;
    }
    private object cbref;     // callback reference object
    private int intvl;    // monitor interval or timeout
    private int sts = -1;      // status
    private int stsLastUpdate = -1;
    private double ts;    // timestamp
    private int tag;      // tag stamp (e.g. cycle no., pulse no.)
    private int flgs;     // link flags
    private object dout;  // data output object
    private object din;   // data input object
    public Boolean TransportAvailable()
    {
      return alnk == null ? false : true;
    }
    public int GetStatus()
    {
      if (alnk != null) sts = alnk.GetLinkStatus();
      return sts;
    }
    public string GetStatusString()
    {
      if (alnk != null) return alnk.GetLinkStatusString();
      return "no link exists for this acop";
    }
    public bool HasLinkStatusChanged()
    {
      if (alnk != null) return alnk.HasLinkStatusChanged();
      return false;
    }
    public object GetOutputDataObject() { return dout; }
    public object GetInputDataObject() { return din; }
    public object GetCallbackReference() { return cbref; }
    public static String GetConnectionAddress(String protocol, String context, String subsystem, String server, String device)
    {
      if (anchor == null) anchor = new AcopLink(protocol);
      return anchor.alnk.GetConnectionAddress(protocol, context, subsystem, server, device);
    }
    public static String GetConnectionHostAddress(String protocol, String address, String property)
    {
      if (anchor == null) anchor = new AcopLink(protocol);
      return anchor.alnk.GetConnectionHostAddress(protocol, address, property);
    }
    public static String[] GetProtocolList()
    {
      if (anchor == null) anchor = new AcopLink();
      return anchor.alnk.GetProtocolList();
    }
    public static String[] GetContextList(String protocol)
    {
      if (anchor == null) anchor = new AcopLink(protocol);
      String[] lst = null;
      try
      {
        lst = anchor.alnk.GetContextList(protocol);
        Array.Sort(lst);
      }
      catch (Exception)
      {
        lst = new String[] { "<unavailable>" };
      }
      return lst;
    }
    public static String[] GetSubSystemList(String protocol, String context)
    {
      if (anchor == null) anchor = new AcopLink(protocol);
      String[] lst = null;
      try
      {
        lst = anchor.alnk.GetSubSystemList(protocol, context);
        Array.Sort(lst);
      }
      catch (Exception)
      {
        lst = new String[] { "<unavailable>" };
      }
      return lst;
    }
    public static String[] GetServerList(String protocol, String context, String subsystem)
    {
      if (anchor == null) anchor = new AcopLink(protocol);
      String[] lst = null;
      try
      {
        lst = anchor.alnk.GetServerList(protocol, context, subsystem);
        Array.Sort(lst);
      }
      catch (Exception)
      {
        lst = new String[] { "<unavailable>" };
      }
      return lst;
    }
    public static String[] GetPropertyList(String protocol, String context, String subsystem, String server, String device)
    {
      if (anchor == null) anchor = new AcopLink(protocol);
      String[] lst = null;
      try
      {
        lst = anchor.alnk.GetPropertyList(protocol, context, subsystem, server, device);
        Array.Sort(lst);
      }
      catch (Exception)
      {
        lst = new String[] { "<unavailable>" };
      }
      return lst;
    }
    public static String[] GetDeviceList(String protocol, String context, String subsystem, String server, String property)
    {
      if (anchor == null) anchor = new AcopLink(protocol);
      String[] lst = null;
      try
      {
        lst = anchor.alnk.GetDeviceList(protocol, context, subsystem, server, property);
      }
      catch (Exception)
      {
        lst = new String[] { "<unavailable>" };
      }
      return lst;
    }
    public static IDictionary<string, int> GetFlagsList(String protocol)
    {
      if (anchor == null) anchor = new AcopLink(protocol);
      IDictionary<string, int> e = AcopTransport.SelectList.Of<AcopTransport.simFlags>();
      try
      {
        return anchor.alnk.GetFlagsList(protocol);
      }
      catch (Exception)
      {
      }
      return e;
    }
    public static Boolean HasPropertyQueryPrecedence(String protocol, String context, String subsystem, String server, String property)
    {
      if (anchor == null) anchor = new AcopLink(protocol);
      try
      {
        return anchor.alnk.HasPropertyQueryPrecedence(protocol, context, subsystem, server, property);
      }
      catch (Exception)
      {
        return false;
      }
    }
    public static Boolean HasPropertyQueryPrecedence(String protocol, String context, String subsystem, String server)
    {
      if (anchor == null) anchor = new AcopLink(protocol);
      try
      {
        return anchor.alnk.HasPropertyQueryPrecedence(protocol, context, subsystem, server);
      }
      catch (Exception)
      {
        return false;
      }
    }
    public static Boolean HasDeviceQueryPrecedence(String protocol, String context, String subsystem, String server)
    {
      if (anchor == null) anchor = new AcopLink(protocol);
      try
      {
        return anchor.alnk.HasDeviceQueryPrecedence(protocol, context, subsystem, server);
      }
      catch (Exception)
      {
        return false;
      }
    }
    public void SetTransportUpdatedEvent(acopTransportUpdate evn)
    {
      if (acopUpdated == null) acopUpdated = evn;
    }
    private class acopCallback : AcopTransport.AcopLinkCallback
    {
      private AcopLink thisAcopLink = null;
      private Component thisAcopControl = null;
      public acopCallback(AcopLink alink)
      {
        thisAcopLink = alink;
        if (alink != null) thisAcopControl = alink.GetAcopControl();
      }
      private delegate void callbackDelegate(AcopTransport.AcopConnection lnk);
      public void callback(AcopTransport.AcopConnection lnk)
      {
        if (thisAcopControl != null && thisAcopControl is UserControl)
        {
          UserControl us = (UserControl)thisAcopControl;
          if (us.InvokeRequired)
          {
            us.Invoke(new callbackDelegate(callback), new object[] { lnk });
            return;
          }
        }
        lock (this)
        {
          thisAcopLink.setCallbackPending(false);
		  thisAcopLink.dout = lnk.GetOutputDataObject(lnk.GetConnectionAddress(), lnk.GetConnectionProperty());																								   
          if (thisAcopLink.acopUpdated != null) thisAcopLink.acopUpdated((object)thisAcopLink, thisAcopLink);
        }
      }
    }
    public static object GetOutputDataObject(AcopLink link, string address, string property, object Reference, int Capacity)
    {
      if (anchor == null) anchor = new AcopLink();
      Boolean hasLink = link != null && link.alnk != null;
      AcopTransport.AcopLink lnk = hasLink ? link.alnk : anchor.alnk;
      if (hasLink && link.dout != null) return link.dout;
      object obj = lnk.GetOutputDataObject(address, property, Reference, Capacity);
      if (hasLink) link.dout = obj;
      return obj;
    }
    public static object GetOutputDataObject(string address, string property, object Reference, int Capacity)
    {
      return GetOutputDataObject(null, address, property, Reference, Capacity);
    }
    public static object GetOutputDataObject(AcopLink link, string address, string property)
    {
      if (anchor == null) anchor = new AcopLink();
      Boolean hasLink = link != null && link.alnk != null;
      AcopTransport.AcopLink lnk = hasLink ? link.alnk : anchor.alnk;
      if (hasLink && link.dout != null) return link.dout;
      object obj = lnk.GetOutputDataObject(address, property);
      if (hasLink) link.dout = obj;
      return obj;
    }
    public static object GetOutputDataObject(string address, string property)
    {
      return GetOutputDataObject(null, address, property);
    }
    public static object GetDataObjectFromString(object reference, String data)
    {
      if (reference == null) return null;
      Type valueType = (reference is Array) ? ((Array)reference).GetType().GetElementType() : reference.GetType();
      return GetDataObjectFromString(valueType, data);
    }
    public static object GetDataObjectFromString(Type type, String data)
    {
      return GetDataObjectFromString(null, type, data);
    }
    public static object GetDataObjectFromString(AcopLink link, Type type, String data)
    {
      if (anchor == null) anchor = new AcopLink();
      Boolean hasLink = link != null && link.alnk != null;
      AcopTransport.AcopLink lnk = hasLink ? link.alnk : anchor.alnk;
      return lnk.GetDataObjectFromString(type, data);
    }
    public static object GetDataObjectFromStringArray(Type type, String[] data)
    {
      return GetDataObjectFromStringArray(null, type, data);
    }
    public static object GetDataObjectFromStringArray(AcopLink link, Type type, String[] data)
    {
      if (anchor == null) anchor = new AcopLink();
      Boolean hasLink = link != null && link.alnk != null;
      AcopTransport.AcopLink lnk = hasLink ? link.alnk : anchor.alnk;
      return lnk.GetDataObjectFromStringArray(type, data);
    }

    public static void FillDataObjectFromString(object dataObject, String dataString)
    {
      if (anchor == null) anchor = new AcopLink();
      anchor.alnk.FillDataObjectFromString(dataObject, dataString);
      return;
    }
    public void FillDataObjectFromString(String dataString)
    {
      FillDataObjectFromString(dout, dataString);
    }
    public int GetOutputCompletionLength()
    {
      if (alnk == null) return 0;
      return alnk.GetLinkCompletionSize();
    }
    public double GetTimeStamp()
    {
      if (alnk == null) return 0;
      return alnk.GetLinkTimestamp();
    }
    public int GetDataTag()
    {
      if (alnk == null) return 0;
      return alnk.GetLinkTag();
    }
    public static AcopTransport.DisplaySettings GetDisplaySettings(string address, string property)
    {
      if (anchor == null) anchor = new AcopLink();
      if (anchor.alnk == null) return null;
      return anchor.alnk.GetDisplaySettings(address, property);
    }
    public double[] GetOutputXAxis()
    {
      if (alnk == null) return null;
      return alnk.GetXAxis();
    }
    public double[] GetOutputYAxis()
    {
      if (alnk == null) return null;
      return alnk.GetYAxis();
    }
    public string[] GetOuputDataAsString()
    {
      if (alnk == null) return null;
      return alnk.GetDataAsString();
    }
    public double[] GetOutputDataAsDouble()
    {
      if (alnk == null) return null;
      return alnk.GetDataAsNumber();
    }
    public Int32[] GetOutputDataAsInteger()
    {
      if (alnk == null) return null;
      return alnk.GetDataAsInteger();
    }
    private string[] channelNames = null;
    public string[] GetOutputChannelNames()
    {
      if (alnk == null) return null;
      if (!alnk.IsInitialized())
      {
        alnk = AcopTransport.AcopLinkFactory.GetAcopLink(this.linkKey);
        if (alnk == null)
        {
          AcopControl.AcopLink acl = GetAcopLink(this.GetAcopControl());
          if (acl != null) return acl.GetOutputChannelNames();
        }
      }
      if (channelNames != null) return channelNames;
      if (alnk == null) return null;
      string[] nams = alnk.GetChannelNames();
      channelNames = nams;
      return nams;
    }
    private acopCallback thisCallback = null;
    public int Attach()
    {
      return Attach(null, null, 0, defaultInterval, null);
    }
    public int Attach(object DataOut)
    {
      return Attach(DataOut, null, 0, defaultInterval, null);
    }
    public int Attach(object DataOut, int Flags)
    {
      return Attach(DataOut, null, Flags, defaultInterval, null);
    }
    public int Detach()
    {
      if (alnk == null) throw new Exception("link target does not exist or is not yet initialized");
      lid = -1;
      return alnk.Detach();
    }
    public int Execute()
    {
      return Execute(null, null, 0, defaultInterval);
    }
    public int Execute(object DataOut)
    {
      return Execute(DataOut, null, 0, defaultInterval);
    }
    public int Execute(object DataOut, int Flags)
    {
      return Execute(DataOut, null, Flags, defaultInterval);
    }
    public int Execute(object DataOut, object DataIn, int Flags)
    {
      return Execute(DataOut, DataIn, Flags, defaultInterval);
    }
    public int Execute(object DataOut, object DataIn, int Flags, int timeout)
    {
      if (alnk == null) throw new Exception("link target does not exist or is not yet initialized");
      dout = DataOut;
      din = DataIn;
      flgs |= Flags;
      intvl = timeout;
      cbref = null;
      alnk.GetConnection(prot, addr, prop, dout, din, flgs, thisAcop);
      alnk.SetAcopControlLink(this);
      return sts = alnk.Execute(Flags, timeout);
    }
    public int Execute(object DataOut, int Offset, int Length, object DataIn, int Flags, int timeout)
    {
      if (alnk == null) throw new Exception("link target does not exist or is not yet initialized");
      dout = DataOut;
      din = DataIn;
      flgs |= Flags;
      intvl = timeout;
      cbref = null;
      alnk.GetConnection(prot, addr, prop, dout, Offset, Length, din, flgs, thisAcop);
      alnk.SetAcopControlLink(this);
      return alnk.Execute(Flags, timeout);
    }
    public int Attach(object DataOut, object DataIn, int Flags, int interval, object reference)
    {
      if (alnk == null) throw new Exception("link target does not exist or is not yet initialized");
      dout = DataOut;
      din = DataIn;
      flgs |= Flags;
      if (interval > 0) intvl = interval;
      cbref = reference;
      alnk.GetConnection(prot, addr, prop, dout, din, flgs, thisAcop);
      alnk.SetAcopControlLink(this);
      lid = alnk.Attach(flgs, intvl, alnk, thisCallback);
      return lid;
    }
    public int Attach(object DataOut, int Offset, int Length, object DataIn, int Flags, int interval, object reference)
    {
      if (alnk == null) throw new Exception("link target does not exist or is not yet initialized");
      dout = DataOut;
      din = DataIn;
      flgs |= Flags;
      if (interval > 0) intvl = interval;
      cbref = reference;
      alnk.GetConnection(prot, addr, prop, dout, Offset, Length, din, flgs, thisAcop);
      alnk.SetAcopControlLink(this);
      lid = alnk.Attach(flgs, intvl, alnk, thisCallback);
      return lid;
    }
    public int GetHistoryPoints(double[][] data, ref int Length, double StartTime, double StopTime, int Flags, int timeout)
    {
      if (alnk == null) throw new Exception("link target does not exist or is not yet initialized");
      alnk.GetConnection(prot, addr, prop, dout, din, flgs, thisAcop);
      return alnk.GetHistoryPoints(data, ref Length, StartTime, StopTime, Flags, timeout);
    }
    public int GetImage(ref Bitmap bt, ref String header)
    {
      if (alnk == null) throw new Exception("link target does nor exist or is not yet initialized");
      return alnk.GetImage(ref bt, ref header);
    }
    public int GetLinkFlags()
    {
      if (alnk == null) throw new Exception("link target does not exist or is not yet initialized");
      return alnk.GetFlags();
    }
    public void SetConnectionAddress(string address)
    {
      SetConnectionAddress(null, address, null);
    }
    public void SetConnectionAddress(string address, string property)
    {
      SetConnectionAddress(null, address, property);
    }
    public void SetConnectionAddress(string protocol, string address, string property)
    {
      if (address == null) throw new Exception("the 'address' parameter cannot be empty!");
      if (address.StartsWith("//"))
      {
        string p, s = address.Substring(2);
        int idx;
        if ((idx = s.IndexOf('/')) < 0) throw new Exception("the 'address' parameter specifies a protocol but no address!");
        p = s.Substring(0, idx);
        if (p.Length > 0 && (protocol == null || protocol.Length == 0)) protocol = p;
        address = s.Substring(idx);
        if (address.Length == 0) throw new Exception("the 'address' parameter specifies a protocol but no address!");
      }
      if (protocol != null) SetProtocol(protocol);
      addr = address;
      if (property != null) prop = property;
    }
    public int LockAccess(string address)
    {
      if (alnk != null)
      {
        return alnk.LockAccess(address);
      }
      return -1;
    }
    public int UnlockAccess(string address)
    {
      if (alnk != null)
      {
        return alnk.UnlockAccess(address);
      }
      return -1;
    }
    static OrderedDictionary olnkTbl = new OrderedDictionary();
    static Dictionary<String, AcopLink> lnkTbl = new Dictionary<String, AcopLink>();
    static Dictionary<Component, AcopLink[]> lstTbl = new Dictionary<Component, AcopLink[]>();
    static Boolean lnkTblStale = false;
    static Dictionary<Component, Boolean> lstStale = new Dictionary<Component, bool>();
    public static AcopLink[] GetAcopLinkTable()
    {
      lock (lnkTbl)
      {
        return lnkTbl.Values.ToArray();
      }
    }
    public static AcopLink[] GetAcopLinkTable(Component target)
    {
      lock (lnkTbl)
      {
        AcopLink[] newlst = null;
        if (!isLinkTableStale(target))
        {
          if (lstTbl.TryGetValue(target, out newlst)) return newlst;
        }
        lnkTblStale = false;
        AcopLink[] lst = lnkTbl.Values.ToArray();
        if (lst == null)
        {
          newlst = new AcopLink[] { null };
        }
        else
        {
          newlst = new AcopLink[lst.GetLength(0)];
          int n = 0;
          foreach (AcopLink al in lst)
          {
            if (al.thisAcop != target) continue;
            if (al.lid < 0) continue; // just the active links
            newlst[n++] = al;
          }
          Array.Resize<AcopLink>(ref newlst, n);
        }
        Array.Sort(newlst, delegate(AcopLink a1, AcopLink a2) {
          return a1.orderedId.CompareTo(a2.orderedId);
        });
        AcopLink[] oldlst = null;
        if (lstTbl.TryGetValue(target,out oldlst))
        {
          Boolean modified = (newlst.Count() != oldlst.Count());
          if (!modified)
          {
            foreach (AcopLink al in oldlst)
            {
              if (!newlst.Contains(al))
              {
                modified = true;
                break;
              }
            }
          }
          if (!modified) return oldlst;
        }
        if (lstTbl.ContainsKey(target)) lstTbl.Remove(target);
        lstTbl.Add(target, newlst);
        if (lstStale.ContainsKey(target)) lstStale[target] = false;
        return newlst;
      }
    }
    public static AcopLink GetAcopLink(String key)
    {
      lock (lnkTbl)
      {
        if (key == null) return null;
        if (lnkTbl.ContainsKey(key)) return lnkTbl[key];
      }
      return null;
    }
    public static AcopLink GetAcopLink(object acop)
    {
      if (acop == null) return null;
      String hash = acop.GetHashCode().ToString();
      String[] keys = lnkTbl.Keys.ToArray<String>();
      foreach (String k in keys)
      {
        if (k.StartsWith(hash) &&  
            lnkTbl[k].alnk != null && // anything else ?
            !lnkTbl[k].alnk.ToString().StartsWith("unitialized"))
          return lnkTbl[k];
      }
      return null;
    }
    private static Boolean isLinkTableStale(Component acop)
    {
      if (!lnkTblStale) return false; // can't be stale !
      if (!lstStale.ContainsKey(acop)) return false;
      return lstStale[acop];
    }
    private static void markAsStale(AcopLink lnk)
    {
      lnkTblStale = true;
      Component acop = lnk.GetAcopControl();
      if (acop == null) return;
      if (!lstStale.ContainsKey(acop))
      {
        lstStale.Add(acop, true);
      }
      else
      {
        lstStale[acop] = true;
      }
    }
    public static void AddAcopLink(String key, AcopLink lnk)
    {
      lock (lnkTbl)
      {
        if (lnkTbl.ContainsKey(key)) return;
        lnkTbl.Add(key, lnk);
        markAsStale(lnk);
      }
    }
    public static void RmvAcopLink(String key)
    {
      if (key == null) return;
      lock (lnkTbl)
      {
        if (lnkTbl.ContainsKey(key))
        {
          AcopLink lnk = lnkTbl[key];
          markAsStale(lnk);
          lnkTbl.Remove(key);
        }
      }
    }
    public static void ClearAcopLinkTable()
    {
      lock (lnkTbl)
      {
        foreach (AcopLink lnk in lnkTbl.Values)
        {
          lnk.Detach();
          markAsStale(lnk);
        }
        lnkTbl.Clear();
      }
    }
    public static void ClearAcopLinkTable(Component target)
    {
      lock (lnkTbl)
      {
        Dictionary<string, AcopLink> clrlst = new Dictionary<string, AcopLink>();
        foreach (AcopLink lnk in lnkTbl.Values)
        {
          if (lnk.GetAcopControl() == target)
          {
            clrlst.Add(lnk.GetConnectionKey(), lnk);
          }
        }
        foreach (AcopLink lnk in clrlst.Values)
        {
          lnk.Detach();
          lnkTbl.Remove(lnk.GetConnectionKey());
          markAsStale(lnk);
        }
      }
    }
    private String linkKey = null;
    public String GetConnectionKey()
    {
      return linkKey;
    }
    public void SetConnectionKey(String key)
    {
      if (linkKey == null) linkKey = key;
    }
    public String MakeConnectionKey(string Protocol, string Address, string Property, object DataOut, object DataIn, int Flags)
    {
      String key = AcopTransport.AcopLink.MakeConnectionKey(Protocol, Address, Property, DataOut, DataIn, Flags, thisAcop);
      linkKey = thisAcop.GetHashCode().ToString() + key;
      return linkKey;
    }
    public static void RmvAcopLink(string protocol, string address, string property, object acop)
    {
      String key = AcopTransport.AcopLink.MakeConnectionKey(protocol, address, property, null, null, 0, acop);
      RmvAcopLink(key);
    }
    public static AcopLink GetAcopLink(string protocol, string address, string property, object acop)
    {
      //AcopLink al = GetAcopLink(acop);
      //String key = al != null ? al.GetConnectionKey() : null;
      //if (al == null || key == null)
      //{
      //  key = AcopTransport.AcopLink.MakeConnectionKey(protocol, address, property, null, null, 0, acop);
      //  al = GetAcopLink(key);
      //}
      String key = AcopTransport.AcopLink.MakeConnectionKey(protocol, address, property, null, null, 0, acop);
      AcopLink al = GetAcopLink(key);
      if (al != null && al.alnk != null && al.alnk.ToString().StartsWith("unitialized"))
      {
        AcopLink alt = GetAcopLink(acop);
        if (alt != null)
        {
          if ((alt.addr.Equals(address, StringComparison.CurrentCultureIgnoreCase) &&
               alt.prop.Equals(property, StringComparison.CurrentCultureIgnoreCase))
               || alt.lid < 0)
          {
            al = alt;
          }
        }
      }
      if (al == null)
      {
        if (key == null) return null;
        al = new AcopLink();
        al.SetConnectionAddress(protocol, address, property);
        al.SetConnectionKey(key);
        al.SetAcopControl((Component)acop);
        AddAcopLink(key, al);
      }
      return al;
    }
    public static AcopLink GetAcopLink(string protocol, string address, string property, object dataOut, object dataIn, int flags, object acop)
    {
      String key = AcopTransport.AcopLink.MakeConnectionKey(protocol, address, property, dataOut, dataIn, flags, acop);
      AcopLink al = GetAcopLink(key);
      if (al == null)
      {
        if (key == null) return null;
        al = new AcopLink();
        al.SetConnectionAddress(protocol, address, property);
        al.SetConnectionKey(key);
        al.SetAcopControl((Component)acop);
        AddAcopLink(key, al);
      }
      return al;
    }
    private static uint orderedRefCounter = 0;
    private uint orderedId = 0;
    public AcopLink()
      : this(null)
    {
    }
    ~AcopLink()
    {
      if (alnk != null) Detach();
    }
    public AcopLink(String protocol)
    {
      try
      {
        if (alnk == null) alnk = new AcopTransport.AcopLink();
        SetProtocol(protocol == null ? alnk.GetConnectionProtocol() : protocol);
        intvl = alnk.GetConnectionInterval();
        thisCallback = new acopCallback(this);
        if (anchor == null) anchor = this;
        orderedId = ++orderedRefCounter;
      }
      catch (Exception)
      {
        //throw new Exception("cannot create Acop Link", x); 
        return;
      }
    }
  }
}
