using System;
using System.Runtime.InteropServices;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using tine;
using System.Windows.Forms;

// note: there is no GUI components to this example server.
// add your own 'buttons' etc. to check on the current status of things, etc.
namespace csSineServer
{
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct SineInfo
  {
    public float amplitude;
    public float frequency;
    public float noise;
    public float phase;
    public int   numberCalls;
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = 64)]
      public char[] description;
  };

  public partial class Form1 : Form
  {
    private const int PRP_SINE = 1;
    private const int PRP_AMPLITUDE = 2;
    private const int PRP_FREQUENCY = 3;
    private const int PRP_PHASE = 4;
    private const int PRP_NOISE = 5;
    private const int PRP_INFO = 6;
    private const int PRP_STRINGS = 7;
    private const int NUM_DEVICES = 10;
    private const int NUM_VALUES = (1024*8);
    float[,] sinbuf = new float[NUM_DEVICES,NUM_VALUES];
    private SineInfo[] sineInfoTable = new SineInfo[NUM_DEVICES];
    private String[] SineStrings = new String[NUM_DEVICES];
    public Int32 sineqm(string dev, string prp, TDataType dout, TDataType din, UInt16 acc,TEquipmentModule eqm)
    {
      Console.WriteLine("eqm called: "+dev+" "+prp);
      float fval = 0;
      float[] mcarray = new float[NUM_DEVICES];
      int cc;
      int devnr = eqm.GetDeviceNumber(dev, prp);
      int prpid = eqm.GetPropertyId(prp); 
      switch (prpid)
      {
        case PRP_SINE:
          if ((acc & Access.CA_WRITE) != 0) return Errors.illegal_read_write;
          if ((cc = dout.PutData(sinbuf,devnr,NUM_VALUES,0)) != 0) return cc;
          sineInfoTable[devnr].numberCalls++;
          return 0;
        case PRP_AMPLITUDE:
          if (din.GetDataArrayLength() > 0)
          { /* input data => require write access */
            if ((acc & Access.CA_WRITE) != Access.CA_WRITE) return Errors.illegal_read_write;
            if ((cc = din.GetData(ref fval)) != 0) return cc;
            if (fval < 1 || fval > 1000) return Errors.out_of_range;
            sineInfoTable[devnr].amplitude = fval;
          }
          if (dout.GetDataArrayLength() > 0)
          { /* prepare multichannel array */
            if (dout.GetDataFormat() == Formats.CF_FLTINT)
            {
              FLTINT[] fi = new FLTINT[NUM_DEVICES];
              for (int i = 0; i < NUM_DEVICES; i++)
              {
                fi[i] = new FLTINT();
                fi[i].fval = sineInfoTable[i].amplitude;
                fi[i].ival = sineInfoTable[i].numberCalls;
              }
              dout.PutData(fi, NUM_DEVICES, devnr);
              return 0;
            }
            for (int i = 0; i < NUM_DEVICES; i++)
              mcarray[i] = sineInfoTable[i].amplitude;
            if ((cc = dout.PutData(mcarray,NUM_DEVICES,devnr)) != 0) return cc;
          }
          return 0;
        case PRP_FREQUENCY:
          if (din.GetDataArrayLength() > 0)
          { /* input data => require write access */
            if ((acc & Access.CA_WRITE) != Access.CA_WRITE) return Errors.illegal_read_write;
            if ((cc = din.GetData(ref fval)) != 0) return cc;
            if (fval < 1 || fval > 100) return Errors.out_of_range;
            sineInfoTable[devnr].frequency = fval;
          }
          if (dout.GetDataArrayLength() > 0)
          { /* prepare multichannel array */
            for (int i = 0; i < NUM_DEVICES; i++)
              mcarray[i] = sineInfoTable[i].frequency;
            if ((cc = dout.PutData(mcarray,NUM_DEVICES,devnr)) != 0) return cc;
          }
          return 0;
        case PRP_PHASE:
          if (din.GetDataArrayLength() > 0)
          { /* input data => require write access */
            if ((acc & Access.CA_WRITE) != Access.CA_WRITE) return Errors.illegal_read_write;
            if ((cc = din.GetData(ref fval)) != 0) return cc;
            if (fval < 0 || fval > 512) return Errors.out_of_range;
            sineInfoTable[devnr].phase = fval;
          }
          if (dout.GetDataArrayLength() > 0)
          { /* prepare multichannel array */
            for (int i = 0; i < NUM_DEVICES; i++)
              mcarray[i] = sineInfoTable[i].phase;
            if ((cc = dout.PutData(mcarray,NUM_DEVICES,devnr)) != 0) return cc;
          }
          return 0;
        case PRP_NOISE:
          if (din.GetDataArrayLength() > 0)
          { /* input data => require write access */
            if ((acc & Access.CA_WRITE) != Access.CA_WRITE) return Errors.illegal_read_write;
            if ((cc = din.GetData(ref fval)) != 0) return cc;
            if (fval < 0 || fval > 100) return Errors.out_of_range;
            sineInfoTable[devnr].noise = fval;
          }
          if (dout.GetDataArrayLength() > 0)
          { /* prepare multichannel array */
            for (int i = 0; i < NUM_DEVICES; i++)
              mcarray[i] = sineInfoTable[i].noise;
            if ((cc = dout.PutData(mcarray,NUM_DEVICES,devnr)) != 0) return cc;
          }
          return 0;
        case PRP_STRINGS:
          if (din.GetDataArrayLength() > 0)
          { /* input data => require write access */
            if ((acc & Access.CA_WRITE) != Access.CA_WRITE) return Errors.illegal_read_write;
            String[] inpt = new String[10];
            if ((cc = din.GetData(inpt)) != 0) return cc;
            int len = (int)din.GetDataArrayLength();
            for (int i = 0; i < 10 && i<len; i++) SineStrings[i] = inpt[i];
          }
          if (dout.GetDataArrayLength() > 0)
          { /* prepare multichannel array */
             if ((cc = dout.PutData(SineStrings, NUM_DEVICES, devnr)) != 0) return cc;
          }
          return 0;
        case PRP_INFO:
          if (devnr == 33)
          {
            eqm.SetCompletion("what a stupid device number");
            return 512;
          }
          if (din.GetDataArrayLength() > 0)
          { /* allow an atomic set of all relevant values this way : */
            if (din.GetDataFormat() != Formats.CF_STRUCT) return Errors.illegal_format;
            if (din.GetDataTag().CompareTo("SineInfo") != 0) return Errors.invalid_structure_tag;
            SineInfo sinf = new SineInfo();
            if ((cc = din.GetData(sinf)) != 0) return cc;
            if (sinf.amplitude < 1 || sinf.amplitude > 1000) return Errors.out_of_range;
            if (sinf.frequency < 1 || sinf.frequency > 100) return Errors.out_of_range;
            if (sinf.phase < 0 || sinf.phase > 512) return Errors.out_of_range;
            if (sinf.noise < 0 || sinf.noise > 100) return Errors.out_of_range;
            sineInfoTable[devnr].amplitude = sinf.amplitude;
            sineInfoTable[devnr].frequency = sinf.frequency;
            sineInfoTable[devnr].phase = sinf.phase;
            sineInfoTable[devnr].noise = sinf.noise;
          }
          if (dout.GetDataArrayLength() > 0)
          {
            if (dout.GetDataFormat() != Formats.CF_STRUCT) return Errors.illegal_format;
            if (dout.GetDataTag().CompareTo("SineInfo") != 0) return Errors.invalid_structure_tag;
            if (dout.GetDataArrayLength() > NUM_DEVICES) dout.SetDataArrayLength(NUM_DEVICES);
            if ((cc = dout.PutData(sineInfoTable, NUM_DEVICES, devnr)) != 0) return cc;
         }
          return 0;
      }      
      return 0;
    }
    void updateSine(int devnr)
    { /* used internally -> don't bother with bounds checking */
      int i, k;
      float f = sineInfoTable[devnr].frequency;
      int n = (int)sineInfoTable[devnr].noise;
      int p = (int)sineInfoTable[devnr].phase;
      float a = sineInfoTable[devnr].amplitude;
      Random rnd = new Random();
      for (i=0; i<NUM_VALUES; i++)
      {
        k = (i + p) % NUM_VALUES;
        sinbuf[devnr, k] = (float)(rnd.NextDouble())*n + (float)(Math.Sin(f * i * 6.2832 / (NUM_VALUES / 8)) * a);
      }
      return;
    }
    void registerStructs()
    {
      TTaggedStruct tts = new TTaggedStruct(sineInfoTable);
    }
    // the equipment module registers this routine as a 'background task'
    // to be called every 1000 msecs
    public void sinbkg()
    { // update the sine curves
      for (int i = 0; i < NUM_DEVICES; i++) updateSine(i);
    }
    // the equipment module registers this routine as an 'initialization task'
    public void sinini()
    {
      TKernel.SrvLog("mylog", "sinini", "just testing the new log routine");

      Console.WriteLine("function called");
      for (int i = 0; i < NUM_DEVICES; i++)
      {
        sineInfoTable[i].amplitude = 256;
        sineInfoTable[i].frequency = 1;
        sineInfoTable[i].phase = 0;
        sineInfoTable[i].noise = (float)5.0;
        String dsc = "Sine device " + i + " at your service";
        sineInfoTable[i].description = new char[64];
        Array.Copy(dsc.ToCharArray(),sineInfoTable[i].description,Math.Min(dsc.Length,64));
      }
    }
    // the equipment module registers this routine as an 'exit task'
    public void sinexi()
    {
      Console.WriteLine("server stopping");
    }
    public Form1()
    {
      InitializeComponent();
    }
    public TEquipmentModule myeqm;
    private void Form1_Load(object sender, EventArgs e)
    {

      //String[] ctx = TQuery.GetSystemContexts();
      //foreach (String s in ctx) System.Console.WriteLine(s);
      TKernel.InitializeServer();

      String[] stra = new String[32];
      TLink lnk = new TLink("/TEST/UnitServer/UnitDevice0", "StringArray", new TDataType(stra), null, Access.CA_READ);
      int cc = lnk.Execute();
      System.Console.Write(stra[0]);

      registerStructs();

      // to register everything via configuration database use this form:
      // myeqm = new TEquipmentModule("", "SINEQM", 0, sineqm, sinini, sinbkg, 1000, sinexi); <- like this will use .csv config database files
      // else hardwire the 'names' via the API calls:
      // register the FEC name and other information (important: portOffset must be host-wide unique)
      // alternative: use a fecid.csv file (or fec.xml)
      TKernel.RegisterFecInformation("DotNetFec", "TEST", "TEST", "Dot Net Test Fec", "My office", "none", "Me", 15);
      // register the equipment module and export it as 'DotNetSine'
      // alternative: use the exports.csv or fec.xml to register export names and properties)
      myeqm = new TEquipmentModule("DotNetSine", "DNTEQM", 10, sineqm, sinini, sinbkg, 1000, sinexi); // like this is hard coded names

      //TKernel.InitializeServer();

      // register the properties (alternative: use exports.csv or fec.xml)
      // first: property 'Sine' is a trace waveform which can return up to 8192 float values (no input):
      TDataType dtout = new TDataType(new float[8192]);
      TKernel.RegisterPropertyInformation("DNTEQM", "Sine", dtout, null, Access.CA_READ, ArrayType.AT_TRACE, 8192, "[-512:512 V]Sine Curve", PRP_SINE, "");

      // the following properties are 'multi-channel' properties
      // one can read all values for the 10 sine curves at once, but only set 1 at a time:
      dtout = new TDataType(new float[10]);
      TDataType dtin = new TDataType(new float[1]);
      TKernel.RegisterPropertyInformation("DNTEQM", "Amplitude", dtout, dtin, Access.CA_READ | Access.CA_WRITE, ArrayType.AT_CHANNEL, 10, "[-512:512 V]Sine Curve Amplitude", PRP_AMPLITUDE, "");
      TKernel.RegisterPropertyInformation("DNTEQM", "Frequency", dtout, dtin, Access.CA_READ | Access.CA_WRITE, ArrayType.AT_CHANNEL, 10, "[1:20 Hz]Sine Curve Frequency", PRP_FREQUENCY, "");
      TKernel.RegisterPropertyInformation("DNTEQM", "Phase", dtout, dtin, Access.CA_READ | Access.CA_WRITE, ArrayType.AT_CHANNEL, 10, "[0:6.28]Sine curve Phase", PRP_PHASE, "");
      TKernel.RegisterPropertyInformation("DNTEQM", "Noise", dtout, dtin, Access.CA_READ | Access.CA_WRITE, ArrayType.AT_CHANNEL, 10, "[0:50 V]Sine curve Noise", PRP_NOISE, "");

      // the property 'SineInfo' is a tagged structure
      // returning all information atomically
      // also allow the setting of everything via the same structure, 
      // but only for one sine curve at a time:
      dtout = new TDataType(new SineInfo[10]);
      dtin = new TDataType(new SineInfo[1]);
      TKernel.RegisterPropertyInformation("DNTEQM", "SineInfo", dtout, dtin, Access.CA_READ | Access.CA_WRITE, ArrayType.AT_UNKNOWN, 10, "Sine curve Information", PRP_INFO, "");

      dtout = new TDataType(SineStrings);
      dtin = new TDataType(null);
      TKernel.RegisterPropertyInformation("DNTEQM", "Strings", dtout, dtin, Access.CA_READ | Access.CA_WRITE, ArrayType.AT_CHANNEL, 10, "[0:50 V]Sine curve Noise", PRP_STRINGS, "");

      for (int i = 0; i < 10; i++)
      { // put in some device names
        TKernel.RegisterDeviceName("DNTEQM", "SineGen" + i, i);
        SineStrings[i] = ""+i+"Super-duper string number " + i + "is ready for duty";
      }
      // now start the server ...
      TKernel.StartServices();
    }
  }
}