using System;
using System.IO;
using System.Runtime.InteropServices;
using System.Collections.Generic;
using System.Text;
using System.Security;
using System.Security.Permissions;
using System.Reflection;
using System.Net;

namespace tine
{
  /// <summary>
  /// Systematic string length constants 
  /// </summary>
  public static class Constants
  {
#   pragma warning disable 1591
    public const int FEC_NAME_SIZE = 16;
    public const int EXPORT_NAME_SIZE = 32;
    public const int EQM_NAME_SIZE = 8;
    public const int EQM_NAME_SHORTSIZE = 6;
    public const int CONTEXT_NAME_SIZE = 32;
    public const int STATUS_SIZE = 192;
    public const int SUBSYSTEM_NAME_SIZE = 16;
    public const int TAG_NAME_SIZE = 16;
    public const int FIELD_NAME_SIZE = 16;
    public const int USERNAME_SIZE = 16;
    public const int DEVICE_NAME_SIZE = 64;
    public const int DEVICE_NAME_MAXSIZE = 1024;
    public const int DEVICE_REDIR_SIZE = STATUS_SIZE;
    public const int PROPERTY_NAME_SIZE = 64;
    public const int PROPERTY_DESC_SIZE = 64;
    public const int PROPERTY_REDIR_SIZE = STATUS_SIZE;
    public const int UNITS_SIZE = 16;
    public const int ADDR_SIZE = 16;
    public const int ALARM_TAG_SIZE = 32;
    public const int ALARM_TEXT_SIZE = 64;
    public const int ALARM_TEXT_LONGSIZE = 128;
#   pragma warning restore 1591
  }
  /// <summary>
  /// Available Link Transfer Modes
  /// </summary>
  public static class Modes
  {
#   pragma warning disable 1591
    public const UInt16 CM_NULL = 0x00;
    public const UInt16 CM_CANCEL = 0x00;
    public const UInt16 CM_SINGLE = 0x01;
    public const UInt16 CM_REFRESH = 0x02;
    public const UInt16 CM_DATACHANGE = 0x02;
    public const UInt16 CM_POLL = 0x03;
    public const UInt16 CM_TIMER = 0x03;
    public const UInt16 CM_REGISTER = 0x04;
    public const UInt16 CM_RECEIVE = 0x05;
    public const UInt16 CM_EVENT = 0x06;
    public const UInt16 CM_NODUPLICATES = 0x10;
    public const UInt16 CM_SEEKPATTERN = 0x20;
    public const UInt16 CM_MISSPATTERN = 0x40;
    public const UInt16 CM_STREAM = 0x80;
    public const UInt16 CM_BCAST = 0x100;
    public const UInt16 CM_NETWORK = 0x200;
    public const UInt16 CM_MCAST = 0x200;
    public const UInt16 CM_USE_ON_ERROR = 0x400;
    public const UInt16 CM_USEONERROR = 0x400;
    public const UInt16 CM_RETRY = 0x800;
    public const UInt16 CM_GROUPED = 0x1000;
    public const UInt16 CM_CONNECT = 0x2000;
    public const UInt16 CM_WAIT = 0x4000;
    public const UInt16 CM_SYNCNOTIFY = 0x8000;
    /// <summary>
    /// returns the primary mode contained in the first nibble
    /// </summary>
    /// <param name="mode">is the full transport mode specification</param>
    /// <returns></returns>
    public static UInt16 GetBaseMode(UInt16 mode)
    {
      return (UInt16)(mode & 0x0f);
    }
    public static Boolean IsSingle(UInt16 mode)
    {
      return ((mode & 0x0f) == CM_SINGLE) ;
    }
    public static Boolean IsDataChange(UInt16 mode)
    {
      return ((mode & 0x0f) == CM_DATACHANGE);
    }
    public static Boolean IsTimer(UInt16 mode)
    {
      return ((mode & 0x0f) == CM_TIMER);
    }
    public static Boolean IsEvent(UInt16 mode)
    {
      return ((mode & 0x0f) == CM_EVENT);
    }
  }
  /// <summary>
  /// Available Data Array Types
  /// </summary>
  public static class ArrayType
  {
#   pragma warning disable 1591
    public const UInt16 AT_UNKNOWN = 0x00;
    public const UInt16 AT_SCALAR = 0x01;
    public const UInt16 AT_SINGLE = 0x02;
    public const UInt16 AT_DOUBLE = 0x04;
    public const UInt16 AT_BOOLEAN = 0x08;
    public const UInt16 AT_CHANNEL = 0x10;
    public const UInt16 AT_SPECTRUM = 0x20;
    public const UInt16 AT_TRACE = 0x20;
    public const UInt16 AT_WAVEFORM = 0x20;
    public const UInt16 AT_STRUCTURED = 0x40;
    public const UInt16 AT_COLLECTION = 0x40;
    public const UInt16 AT_UNSTRUCTURED = 0x80;
    public const UInt16 AT_BLOB = 0x80;
    public const UInt16 AT_IMAGE = 0x80;
    public const UInt16 AT_JPEG = 0x100;
    public const UInt16 AT_MPEG = 0x200;
    public const UInt16 AT_BMP = 0x400;
    public const UInt16 AT_GIF = 0x800;
    public const UInt16 AT_WAV = 0x1000;
    public const UInt16 AT_MP3 = 0x2000;
    public static Boolean IsChannel(UInt16 code)
    {
      return ((code & AT_CHANNEL) == AT_CHANNEL);
    }
  }
  /// <summary>
  /// Available Access Flags
  /// </summary>
  public static class Access
  {
    public const UInt16 CA_NULL = 0x00;
    public const UInt16 CA_READ = 0x01;
    public const UInt16 CA_WRITE = 0x02;
    public const UInt16 CA_REPEAT = 0x04;
    public const UInt16 CA_SYNC = 0x08;
    public const UInt16 CA_ALARM = 0x10;
    public const UInt16 CA_HIST = 0x20;
    public const UInt16 CA_FIRST = 0x40;
    public const UInt16 CA_LAST = 0x80;
    public const UInt16 CA_NOCALLBACKS = 0x80; // used only in client-side sync call
    public const UInt16 CA_RETRY = 0x100;
    public const UInt16 CA_NETWORK = 0x200;
    public const UInt16 CA_QUERY = 0x400;
    public const UInt16 CA_LOCKED = 0x800;// used only in server-side eqm dispatch
    public const UInt16 CA_STREAM = 0x800;// used only in client-side sync call
    public const UInt16 CA_XREAD = 0x800; // used only in property registration
    public const UInt16 CA_NORETRY = 0x1000;
    public const UInt16 CA_CONNECT = 0x2000;
    public const UInt16 CA_MUTABLE = 0x4000;
    public const UInt16 CA_SYNCNOTIFY = 0x8000;
    public static Boolean IsRead(UInt16 code)
    {
      return ((code & CA_READ) == CA_READ);
    }
    public static Boolean IsWrite(UInt16 code)
    {
      return ((code & CA_WRITE) == CA_WRITE);
    }
    public static Boolean IsExclusiveRead(UInt16 code)
    {
      return ((code & CA_XREAD) == CA_XREAD);
    }
    public static Boolean IsFirst(UInt16 code)
    {
      return ((code & CA_FIRST) == CA_FIRST);
    }
    public static Boolean IsLast(UInt16 code)
    {
      return ((code & CA_LAST) == CA_LAST);
    }
    public static Boolean IsLocked(UInt16 code)
    {
      return ((code & CA_LOCKED) == CA_LOCKED);
    }
    public static Boolean IsMutable(UInt16 code)
    {
      return ((code & CA_MUTABLE) == CA_MUTABLE);
    }
  }
  /// <summary>
  /// Access Lock definitions
  /// </summary>
  public static class AccessLock
  {
    public const int LOCK_UNLOCKED = 0;
    public const int LOCK_PREEMPTIVE = 1;
    public const int LOCK_PERSISTENT = 2;
    public const int LOCK_ABORT = 3;
    public const int LOCK_XREAD = 0x100;
    public const int LOCK_NOCALLBACKS = 0x200;
    public const int MAX_LOCK_DURATION = 60;
#   pragma warning restore 1591
  }
  /// <summary>
  /// Available Data Format Types
  /// </summary>
  public static class Formats
  {
#   pragma warning disable 1591
    public const byte CF_DOUBLE = 0;
    public const byte CF_INT16  = 1;
    public const byte CF_BYTE = 2;
    public const byte CF_INT8 = 2;
    public const byte CF_INT32 = 3;
    public const byte CF_CHAR = 4;
    public const byte CF_TEXT = 4;
    public const byte CF_FLOAT = 5;
    public const byte CF_INT64 = 6;
    public const byte CF_STRUCT = 7;
    public const byte CF_NAME8 = 8;
    public const byte CF_NAME16 = 9;
    public const byte CF_NAME16FI = 10;
    public const byte CF_NAME16FLTINT = 10;
    public const byte CF_BIT = 11;
    public const byte CF_NAME32 = 13;
    public const byte CF_FLTINT = 14;
    public const byte CF_INTINT = 15;
    public const byte CF_INTFLTINT = 18;
    public const byte CF_NAME48 = 19;
    public const byte CF_NAME8I = 20;
    public const byte CF_NAME16I = 21;
    public const byte CF_NAME32I = 22;
    public const byte CF_NAME48I = 23;
    public const byte CF_USTRING  = 24;
    public const byte CF_FLTFLTINT = 25;
    public const byte CF_FLTINTINT = 26;
    public const byte CF_FILTER = 27;
    public const byte CF_INTFLTFLTFLT = 27;
    public const byte CF_SPECTRUM = 28;
    public const byte CF_ADDRESS = 29;
    public const byte CF_INTINTINTINT = 29;
    public const byte CF_WINDOW = 30;
    public const byte CF_TTII = 30;
    public const byte CF_POINT = 31;
    public const byte CF_XY = 31;
    public const byte CF_FLTFLT = 31;
    public const byte CF_INTFLTSHTSHT = 32;
    public const byte CF_BOOLEAN = 33;
    public const byte CF_INTINTFLTFLT = 34;
    public const byte CF_FWINDOW = 34;
    public const byte CF_FLTINTFLTINT = 35;
    public const byte CF_NAME64 = 36;
    public const byte CF_NAME64I = 37;
    public const byte CF_DBLDBL = 40;
    public const byte CF_NAME16DBLDBL = 41;
    public const byte CF_NAME16II = 42;
    public const byte CF_NAME16INTINT = 42;
    public const byte CF_INTINTINT  = 43;
    public const byte CF_NAME32DBLDBL  = 44;
    public const byte CF_NAME64DBLDBL  = 45;
    public const byte CF_INTFLTFLTFLTNAME16 = 46;
    public const byte CF_UNAME = 46;
    public const byte CF_XML = 47;
    public const byte CF_DBLDBLDBL = 48;
    public const byte CF_NAME64DBLDBLDBL = 49;
    public const byte CF_BITFIELD8 = 50;
    public const byte CF_BITFIELD16 = 51;
    public const byte CF_BITFIELD32 = 52;
    public const byte CF_BITFIELD64 = 53;
    public const byte CF_NAME64DBL = 54;
    public const byte CF_IMAGE = 55;
    public const byte CF_HISTORY = 56;
    public const byte CF_STRING = 57;
    public const byte CF_ASPECTRUM = 58;
    public const byte CF_AIMAGE = 59;

    public const byte CF_DEFAULT = 254;
    public const byte CF_NULL = 255;

    public const short LFMT_OFFSET = 512;
#   pragma warning restore 1591
    /// <summary>
    /// returns the size in bytes of the specified tine format
    /// </summary>
    /// <param name="fmt">a tine format code</param>
    /// <returns></returns>
    public static Int32 GetSizeInBytes(Int32 fmt)
    {
      return native.GetFormatSize(LFMT_OFFSET + (fmt%256));
    }
    /// <summary>
    /// 'true' if the tine format code corresponds to a primitive type
    /// </summary>
    /// <param name="fmt"></param>
    /// <returns></returns>
    public static Boolean IsPrimitive(Int32 fmt)
    {
      return (fmt <= CF_INT64);
    }
    /// <summary>
    /// 'true' if the tine format code corresponds to a simple string type
    /// </summary>
    /// <param name="fmt">a tine format code</param>
    /// <returns></returns>
    public static Boolean IsSimpleString(Int32 fmt)
    {
      switch (fmt)
      {
        case CF_TEXT:
        case CF_STRING:
        case CF_NAME8:
        case CF_NAME16:
        case CF_NAME32:
        case CF_NAME48:
        case CF_NAME64:
          return true;
        default:
          return false;
      }
    }
  }
  /// <summary>
  /// Flags which can be used when specifying a Data Notification Tolerance
  /// </summary>
  public static class DataNotificationFlags
  {
#   pragma warning disable 1591
    public const UInt32 DTF_NONE = 0x00;
    public const UInt32 DTF_SUPPRESS = 0x01;
    public const UInt32 DTF_NOTIFY = 0x02;
#   pragma warning restore 1591
  }
  /// <summary>
  /// TINE Error definitions and flags
  /// </summary>
  public static class Errors
  {
#   pragma warning disable 1591
    public const UInt16 CE_REDIRECTED  = 0x8000;
    public const UInt16 CE_SENDDATA    = 0x4000;
    public const UInt16 CE_RENEGOTIATE = 0x2000;

    public const UInt16 operation_success = 0; // successful completion
    public const UInt16 illegal_line = 1;  // illegal line number */
    public const UInt16 illegal_format = 2;  // illegal format or data type */
    public const UInt16 illegal_arithmetic = 3;  // illegal arithmetic expression */
    public const UInt16 ambiguous = 4;  // ambigous input information */
    public const UInt16 illegal_delimiter = 5;  // illegal delimiter in expression */
    public const UInt16 zero_divide =  6;  // divide by zero */
    public const UInt16 work_area_full = 7;  // working area buffer is full */
    public const UInt16 non_existent = 8;  // requested target does not exist */
    public const UInt16 invalid_transport = 9;  // transport medium is not valid */
    public const UInt16 link_exhausted = 10; // link exhausted (unused) */

    public const UInt16 not_properly = 11; // expression not properly terminated (unused) */
    public const UInt16 un_allocated = 12; // requested target is not allocated in the given context */
    public const UInt16 no_such_line = 13; // expected line not found in database */
    public const UInt16 illegal_data_size = 14; // give data size is illegal or not valid */
    public const UInt16 io_error = 15; // bus or network io error */
    public const UInt16 illegal_context = 16; // given context is illegal */
    public const UInt16 runtime_error = 17; // runtime error */
    public const UInt16 system_error = 18; // system internal error */
    public const UInt16 hardware_busy = 19; // hardware is busy */
    public const UInt16 argument_list_error = 20; // input argument or parameter is invalid (alias for invalid_parameter */
    public const UInt16 invalid_parameter = 20; // input argument or parameter is invalid */

    public const UInt16 file_error =   21; // file access error */
    public const UInt16 use_stream_transport = 22; // datagram message size exceeds the allowable mtu (reserved) */
    public const UInt16 dimension_error = 23; // array dimension error */
    public const UInt16 square_root_negative = 24; // square root of a negative number */
    public const UInt16 buffer_too_small = 25; // supplied buffer is too small to handle the request */
    public const UInt16 string_too_long = 26; // supplied string is too long to buffer and cannot be truncated */
    public const UInt16 ipx_socket_error = 27; // ipx socket error (reserved) */
    public const UInt16 net_read_error = 28; // network read error */
    public const UInt16 not_ready = 29; // call to equipment module dispatch is not ready (internal, reserved) */
    public const UInt16 invalid_transport_size = 30; // total message size exceeds the configured allowable message size */

    public const UInt16 log_negative = 31; // log of a negative or zero number */
    public const UInt16 device_not_connected = 32; // requested device is not connected */
    public const UInt16 unauthorised_action = 33; // requested operation is not authorized */
    public const UInt16 hardware_error = 34; // hardware error */
    public const UInt16 illegal_equipment_number = 35; // requested device does not have a table entry (alias for illegal_device_number) */
    public const UInt16 illegal_device_number = 35; // requested device does not have a table entry */
    public const UInt16 illegal_device = 35; // requested device is illegal */
    public const UInt16 illegal_property = 36; // requested property is illegal */
    public const UInt16 out_of_range = 37; // given value is out of range */
    public const UInt16 not_implemented = 38; // requested operation has not been implemented */
    public const UInt16 no_such_computer = 39; // requested host computer does not exist */
    public const UInt16 struct_sealed = 40; // tagged structure has already been sealed (reserved) */

    public const UInt16 syntax_error = 41; // expression syntax error (unused) */
    public const UInt16 no_such_file = 42; // requested file does not exist */
    public const UInt16 file_already_exists = 43; // requested file already exists */
    public const UInt16 no_file_space = 44; // not enough capacity in pre-allocated random access file to complete the requested operation */
    public const UInt16 link_not_open = 45; // presistent link has timed out (alias for link_timeout) */
    public const UInt16 link_timeout = 45; // presistent link has timed out */
    public const UInt16 remitted_data_lost = 46; // returned only a portion of the requested data set (reserved) */
    public const UInt16 end_of_file =  47; // attempt to read past the end of a file (unused) */
    public const UInt16 archive_busy = 48; // archive operation is in progress */
    public const UInt16 server_name_in_use = 49; // device server name is already being used */
    public const UInt16 no_such_column = 50; // expected column not found in database */

    public const UInt16 out_of_client_memory = 51; // a client call could not allocate sufficient memory */
    public const UInt16 mcast_access_required = 52; // access requires CA_NETWORK which was not supplied! */
    public const UInt16 address_unresolved = 53; // requested target address could not be resolved */
    public const UInt16 invalid_property = 54; // requested property is not properly formed */
    public const UInt16 address_unknown = 55; // requested target address could not be found */
    public const UInt16 name_unknown = 56; // requested name could not be found */
    public const UInt16 invalid_keyword = 57; // input globals keyword is invalid */
    public const UInt16 invalid_link = 58; // input link id is invalid */
    public const UInt16 string_expected = 59; // database file read encountered an end-of-line while expecting a string */       
    public const UInt16 out_of_local_memory =    60; // a local call to allocated memory was not successful */

    public const UInt16 out_of_shared_memory = 61; // a call to obtain a shared-memory buffer was not successful */
    public const UInt16 invalid_structure_tag = 62; // the given structure tag is not in the structure registry */
    public const UInt16 invalid_index = 63; // the index given does not reference a table entry */
    public const UInt16 illegal_equipment_name = 64; // the equipment name given is not valid */
    public const UInt16 link_error = 65; // internal error processing data link (reserved) */
    public const UInt16 code_failure = 66; // internal failure due to unhandled exceptions, unavailable resource, OS failure, etc. */
    public const UInt16 non_existent_server = 67; // requested server does not exist */
    public const UInt16 deprecated = 68; // requested action or functionality has been deprecated */
    public const UInt16 not_supported = 69; // requested action or functionality not supported */
    public const UInt16 illegal_character =  70; // input character is illegal (unused) */

    public const UInt16 illegal_operator = 71; // input arithmetic or logical operator is not allowed */
    public const UInt16 not_allowed = 72; // requested operation is not allowed */
    public const UInt16 illegal_read_write = 73; // requested access is not allowed (typically CA_WRITE) */
    public const UInt16 out_of_server_memory = 74; // a server call could not allocate sufficient memory */
    public const UInt16 database_not_loaded = 75; // requested operation requires a database which has not been loaded */
    public const UInt16 illegal_command = 76; // requested command is illegal */
    public const UInt16 resources_exhausted = 77; // requested operation would exceed the configured capacity of an allocation table (clients, contracts, connections, etc.) */
    public const UInt16 file_not_open = 78; // requested file io could not be made as the file is not open (unused) */
    public const UInt16 sedac_error =  79; // sedac bus error */
    public const UInt16 no_such_loop = 80; // (unused) */

    public const UInt16 semaphore_error = 81; // a semaphore or mutex could not be obtained or released */
    public const UInt16 driver_not_installed = 82; // request requires a hardware driver which is not installed */
    public const UInt16 port_not_available = 83; // request requires access to a port whichis not available */
    public const UInt16 scan_error = 84; // (unused) */
    public const UInt16 semaphore_busy = 85; // a semahpre or mutex could not be obtained within the requested time */
    public const UInt16 non_existent_elem =  86; // the requested equipment modules does not exist in the registry */
    public const UInt16 non_existent_fec = 87; // the requested FEC does not exist */
    public const UInt16 non_existent_client = 88; // (unused) */
    public const UInt16 cannot_lock = 89; // not able to obtain an access lock or lock resources */
    public const UInt16 not_running = 90; // the requested resource is not running */

    public const UInt16 not_posted = 91; // a dispatch routine did not post a reply */
    public const UInt16 not_accepted = 92; // the requested operation was refused */
    public const UInt16 operation_timeout =  93; // the requested operation did not complete within the time limit specified */
    public const UInt16 illegal_protocol = 94; // the incoming tine protocol is not recognized (reserved, internal) */
    public const UInt16 gpib_error = 95; // gpib bus error */
    public const UInt16 rs232_error = 96; // rs232 bus error */
    public const UInt16 operation_busy = 97; // the requested operation is busy and needs more time to complete */
    public const UInt16 connection_timeout = 98; // a synchronous link did not return within the time limit specified */
    public const UInt16 illegal_mode = 99; // the requested access mode is not valid */
    public const UInt16 not_owner = 100; // (unused) */

    public const UInt16 not_defined = 101; // the requested operation has not yet been defined */
    public const UInt16 net_write_error = 102; // unable to send data on the network */
    public const UInt16 invalid_data = 103; // the input data are not valid */
    public const UInt16 software_error = 104; // general software error */
    public const UInt16 access_denied = 105; // call was refused by the security subsystem */
    public const UInt16 tcp_not_supported = 106; // an attempt to use TCP/IP access was made and is not supported on this host (reserved) */
    public const UInt16 ipx_not_supported	 = 107; // an attempt to use IPX access was made and is not supported on this host (reserved) */
    public const UInt16 host_not_resolved	 = 108; // the requested host could not be resolved */
    public const UInt16 tcp_connect_error	 = 109; // unable to connect to a TCP/IP socket (reserved) */
    public const UInt16 tcp_socket_error 	 = 110; // general TCP/IP socket error (reserved) */

    public const UInt16 configuration_error = 111; // resource configuration error (e.g. redirection information) */
    public const UInt16 invalid_return_code = 112; // indicates a systematic error in processing the equipment dispatch handler (reserved) */
    public const UInt16 link_blacklisted 	 = 113; // requested link parameters cannot succeed (due to initial return of non_existent_elem, address_unknown, etc.) */
    public const UInt16 link_exists = 114; // requested link already exists in the link table indicating a dependent link (reserved,internal) */
    public const UInt16 max_alarms_exceeded = 115; // attempt to add an alarm to the local alarm system whose capacity has been reached */
    public const UInt16 not_exported = 116; // (unused) */
    public const UInt16 is_switching = 117; // switching transition is in progress (unused) */
    public const UInt16 at_limit = 118; // setting is at a maximum or minimum (unused) */
    public const UInt16 get_subscription_id =   119; // used by access mode CM_NETWORK (reserved, internal) */
    public const UInt16 renegotiate_contract =  120; // used to repackage output data when few bytes are returned than requested (reserved, internal) */

    public const UInt16 server_redirection =    121; // used to supply redirection information (reserved, internal) */
    public const UInt16 value_too_high = 122; // returned value is too high */
    public const UInt16 value_too_low = 123; // returned value is too low  */
    public const UInt16 warn_too_high = 124; // returned value is approaching too high  */
    public const UInt16 warn_too_low = 125; // returned value is approaching too low  */
    public const UInt16 completion_waiting =    126; // dependent link is waiting for parent to complete */
    public const UInt16 cannot_select = 127; // select() operation failed (reserved) */
    public const UInt16 has_query_function =    128; // query call to PROPERTIES or DEVICES has completed but indicated a hierachical preference (reserved) */
    public const UInt16 not_signalled = 129; // contract is waiting to call the equipment module dispatch routine (reserved, internal) */
    public const UInt16 call_redirection =  130; // nominal alias for server_redirection */

    public const UInt16 udp_socket_error 	 =    131; // general UDP socket error (reserved) */
    public const UInt16 mutex_error = 132; // a mutex could not be obtained or released */
    public const UInt16 data_not_local = 133; // indicates that a device withing a wildcard call needs to span multiple servers (reserved, internal) */
    public const UInt16 name_already_exists =   134; // a structure or bitfield field name already exists */
    public const UInt16 already_assigned =  135; // the requested resource has already been assigned */
    public const UInt16 invalid_field = 136; // a structure or bitfield field is invalid */
    public const UInt16 mcast_init_error =  137; // could not initialize a multicast socket (reserved) */
    public const UInt16 invalid_mcast_group =   138; // multicast group address is invalid */
    public const UInt16 non_existent_function = 139; // an equipment module's dispatch routine is not registered */
    public const UInt16 property_is_mca = 140; // requested property is a multi-channel array (reserved, internal) */

    public const UInt16 invalid_name = 141; // input name is invalid */
    public const UInt16 invalid_value = 142; // input value is invalid */
    public const UInt16 device_error = 143; // requested device is in an error state */
    public const UInt16 device_not_ready = 144; // requested device is not ready */
    public const UInt16 device_busy = 145; // requested device is busy */
    public const UInt16 invalid_interval = 146; // requested polling interval is invalid */
    public const UInt16 notification_pending =  147; // a link is pending notification (reserved, internal) */
    public const UInt16 thread_error = 148; // a thread operation failed */
    public const UInt16 is_terminating = 149; // the device server is terminating (reserved, internal) */
    public const UInt16 name_server_unknown =   150; // the equipment name server address is unknown (reserved, internal) */

    public const UInt16 lock_required = 151; // an access lock is required to process the operation */
    public const UInt16 not_initialized = 152; // server is not yet initialized */
    public const UInt16 invalid_structure_size = 153; // the input structure size does not match the size in the structure registry */
    public const UInt16 canbus_error = 154; // can bus error */
    public const UInt16 profibus_error = 155; // profibus error */
    public const UInt16 data_stale =  156; // contract's data are stale (reserved) */
    public const UInt16 has_bitfield_tag = 157; // reqested contract needs to request a bitfield (reserved, internal) */
    public const UInt16 not_locked =  158; // the requested resource does not have an access lock (reserved) */
    public const UInt16 lock_expired = 159; // the requested access lock has expired (reserved) */

    public const UInt16 not_registered = 160; // the requested property (e.g. SystemScheduleProperty()) has not been registered */
    public const UInt16 non_existent_property = 161; // the requested property does not exist */
    public const UInt16 memory_overwrite = 162; // a heap error within a dispatch routine has been detected */
    public const UInt16 server_idle = 163; // the server is in an idle state */
    public const UInt16 not_applicable = 164; // the requested opertion is not applicable in the current context */
    public const UInt16 access_locked = 165; // access temporarily suspended due to access lock */
    public const UInt16 invalid_reference = 166; // invalid data reference handle */
    public const UInt16 has_structure_tag = 167; // reqested contract needs to request a structure (reserved, internal) */
    public const UInt16 warn_disk_space = 168; // warn that disk space is low (local alarm system) */
    public const UInt16 low_disk_space = 169; // signal that disk space is very low (local alarm system)  */

    public const UInt16 information_static = 170; // signal that static information is being 'polled'  */
    public const UInt16 reset_mca_property = 171; // signal to re-establish all mca single element links */
    public const UInt16 record_too_long = 172; // record length is too long to process within kernel */
    public const UInt16 unix_socket_error = 173; // an error condition on a unix (pipe) socket */
    public const UInt16 pipe_error = 173; // an error condition on a unix (pipe) socket */
    public const UInt16 shm_error = 174; // an error occured accessing shared memory */
    public const UInt16 tcp_connection_closed = 175; // tcp peer has closed the connection */
    public const UInt16 service_unavailable = 176; // requested service is unavailable */
    public const UInt16 device_offline = 177; // requested device is offline */
    public const UInt16 out_of_sequence = 178; // data value is out of sequence */
    public const UInt16 invalid_datarequest = 179; // requested data length or type is invalid (contract coercion) */

    public const UInt16 property_is_mca_elevated = 180; // requested property is a multi-channel array with elevated format */
    public const UInt16 name_not_found = 181; // named resource was not found in database */
    public const UInt16 out_of_tolerance = 182; // data out of tolerance */
#   pragma warning restore 1591

    private static sbyte[] errstr = new sbyte[64];
    /// <summary>
    /// Converts a TINE error code to a string
    /// </summary>
    /// <param name="err">a TINE error code</param>
    /// <returns>The string representation of the given error code</returns>
    public static String ToString(int err)
    {
      native.GetLastLinkError((short)err, errstr);
      char[] errchr = new char[64];
      for (int i = 0; i < 64 && errstr[i] != 0; i++) errchr[i] = (char)errstr[i];
      return new String(errchr);
    }
    /// <summary>
    /// Checks whether the given status/error code also carries returned data
    /// </summary>
    /// <param name="err">The error/status code as returned from a TLink operation</param>
    /// <returns>true if data were also returned in the TLink operation.</returns>
    public static Boolean HasData(int err)
    {
      if ((err & 0xfff) == 0) return true;
      if ((err & CE_SENDDATA) == CE_SENDDATA) return true;
      return false;
    }
    /// <summary>
    /// Gives the (base) error/status code contained in the error/status returned from a TLink operation.
    /// </summary>
    /// <param name="err">The error/status code as returned from a TLink operation</param>
    /// <returns>the (base) error/status code contained in the error/status returned from a TLink operation.</returns>
    public static UInt16 GetErrorCode(int err)
    {
      return (UInt16)(err & 0xfff);
    }
    /// <summary>
    /// Checks whether the given status/error code also carries the 'out_of_tolerance' flag.
    /// </summary>
    /// <param name="err">The error/status code as returned from a TLink operation</param>
    /// <returns>true if data are out of tolerance.</returns>
    public static Boolean IsOutOfTolerance(int err)
    {
      return ((err & 0xfff) == out_of_tolerance);
    }
  }
  /// <summary>
  /// TINE Image format definitions
  /// </summary>
  public static class ImageFormat
  {
#pragma warning disable 1591
    public const int IF_GRAY = 0;
    public const int IF_RGB = 1;
    public const int IF_RGBA = 2;
    public const int IF_RGB15 = 3;
    public const int IF_RGB16 = 4;
    public const int IF_JPEG = 5;
    public const int IF_TIFF = 6;
    public const int IF_YUV411 = 7;
    public const int IF_YUV412 = 8;
    public const int IF_YUV444 = 9;
    public const int IF_HUFFYUV = 10;
    public const int IF_BITMAP = 11;
    public const int IF_BAYER = 12;
  }
  public static class AlarmCodeFlags
  {
    public const uint ALM_APPLYNAME = 0x80000000;
    public const uint ALM_DISKSPACE = 0x40000000;
    public const uint ALM_LINK_ERROR = 0x20000000;
    public const uint ALM_SYSTEM_DEFAULT = 0x10000000;
    public const uint ALM_SYSTEM_HIDE = 0x08000000;
  }
  /// <summary>
  /// TINE Image flags
  /// </summary>
  public static class ImageFlags
  {
    public const int IFLG_LITTLE_ENDIAN = 0x01;
    public const int IFLG_LOSSLESS_COMP = 0x02;
    public const int IFLG_HORIZONTAL_FLIP = 0x04;
    public const int IFLG_VERTICAL_FLIP = 0x08;
#pragma warning restore 1591
  }
#pragma warning disable 1591
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct TImageSourceHeader
  {
    public UInt32 baseTag;
    public UInt32 cameraPortId;
    public UInt32 versionTag;
    public UInt32 totalLength;
    public Int32 timestampSeconds;
    public Int32 timestampMicroseconds;
    public fixed byte cameraPortName[64];
//    private const int sizeInBytes = 88;
    public TImageSourceHeader(String inpt)
    {
      baseTag = 0x33565356; cameraPortId = 0x0;
      versionTag = 0x1; totalLength = 0;
      timestampSeconds = 0; timestampMicroseconds = 0;
      SetCameraPortName("unknown");
      try
      {
        String[] vals = inpt.Split(StrUtils.parserDelimiter, StringSplitOptions.RemoveEmptyEntries);
        baseTag = UInt32.Parse(vals[0]);
        cameraPortId = UInt32.Parse(vals[1]);
        versionTag = UInt32.Parse(vals[2]);
        totalLength = UInt32.Parse(vals[3]);
        timestampSeconds = Int32.Parse(vals[4]);
        timestampMicroseconds = Int32.Parse(vals[5]);
        SetCameraPortName(vals[6]);
      }
      catch (Exception) { }
    }
    //public byte[] ToByteArray()
    //{
    //  int p = 0;
    //  byte[] ba = new byte[sizeInBytes];
    //  byte[] bap;
    //  bap = BitConverter.GetBytes(baseTag); Array.Copy(bap, 0, ba, p, 4); p += 4;
    //  bap = BitConverter.GetBytes(cameraPortId); Array.Copy(bap, 0, ba, p, 4); p += 4;
    //  bap = BitConverter.GetBytes(versionTag); Array.Copy(bap, 0, ba, p, 4); p += 4;
    //  bap = BitConverter.GetBytes(totalLength); Array.Copy(bap, 0, ba, p, 4); p += 4;
    //  bap = BitConverter.GetBytes(timestampSeconds); Array.Copy(bap, 0, ba, p, 4); p += 4;
    //  bap = BitConverter.GetBytes(timestampMicroseconds); Array.Copy(bap, 0, ba, p, 4); p += 4;
    //  fixed (byte* bp = cameraPortName) for (int i = 0; i < 64; i++) ba[p] = bp[i];
    //  return ba;
    //}
    //public void ToStruct(byte[] src)
    //{
    //  int p = 0;
    //  baseTag = BitConverter.ToUInt32(src, p); p += 4;
    //  cameraPortId = BitConverter.ToUInt32(src, p); p += 4;
    //  versionTag = BitConverter.ToUInt32(src, p); p += 4;
    //  totalLength = BitConverter.ToUInt32(src, p); p += 4;
    //  timestampSeconds = BitConverter.ToInt32(src, p); p += 4;
    //  timestampMicroseconds = BitConverter.ToInt32(src, p); p += 4;
    //  fixed (byte* bp = cameraPortName) for (int i = 0; i < 64; i++) bp[i] = src[p];
    //  return;
    //}
    public UInt32 GetBaseTag()
    {
      return baseTag;
    }
    public void SetBaseTag(UInt32 value)
    {
      baseTag = value;
    }
    public UInt32 GetCameraPortId()
    {
      return cameraPortId;
    }
    public void SetCameraPortId(UInt32 value)
    {
      cameraPortId = value;
    }
    public UInt32 GetVersionTag()
    {
      return versionTag;
    }
    public void SetVersionTag(UInt32 value)
    {
      versionTag = value;
    }
    public UInt32 GetTotalLength()
    {
      return totalLength;
    }
    public void SetTotalLength(UInt32 value)
    {
      totalLength = value;
    }
    public Int32 GetTimestampSeconds()
    {
      return timestampSeconds;
    }
    public void SetTimestampSeconds(Int32 value)
    {
      timestampSeconds = value;
    }
    public Int32 GetTimestampMicroseconds()
    {
      return timestampMicroseconds;
    }
    public void SetTimestampMicroseconds(Int32 value)
    {
      timestampMicroseconds = value;
    }
    public string GetCameraPortName()
    {
      fixed (byte* pca = cameraPortName) return StrUtils.NameToString(pca, 64);
    }
    public void SetCameraPortName(string str)
    {
      fixed (byte* pca = cameraPortName) StrUtils.StringToName(str, pca, 64);
    }

    public override string ToString()
    {
      StringBuilder sb = new StringBuilder();
      sb.Append("baseTag = ").Append(baseTag);
      sb.Append(", cameraPortId = ").Append(cameraPortId);
      sb.Append(", cameraPortName = ").Append(GetCameraPortName());
      sb.Append(", timestampMicroseconds = ").Append(timestampMicroseconds);
      sb.Append(", timestampSeconds = ").Append(timestampSeconds);
      sb.Append(", totalLength = ").Append(totalLength);
      sb.Append(", versionTag = ").Append(versionTag);
      return sb.ToString();
    }
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct TImageFrameHeader
  {
    public Int32 sourceWidth;
    public Int32 sourceHeight;
    public Int32 aoiWidth;
    public Int32 aoiHeight;
    public Int32 xStart;
    public Int32 yStart;
    public Int32 bytesPerPixel;
    public Int32 effectiveBitsPerPixel;
    public Int32 horizontalBinning;
    public Int32 verticalBinning;
    public Int32 sourceFormat;
    public Int32 imageFormat;
    public Int32 frameNumber;
    public Int32 eventNumber;
    public float xScale;
    public float yScale;
    public float imageRotation;
    public float fspare1;
    public float fspare2;
    public float fspare3;
    public Int32 imageFlags;
    public Int32 ispare1;
    public Int32 ispare2;
    public Int32 ispare3;
    public Int32 appendedFrameSize;

    public TImageFrameHeader(String inpt)
    {
      sourceWidth = -1; sourceHeight = -1;
      aoiWidth = -1; aoiHeight = -1;
      xStart = 0; yStart = 0;
      bytesPerPixel = -1; effectiveBitsPerPixel = 0;
      horizontalBinning = 0; verticalBinning = 0;
      sourceFormat = -1; imageFormat = 0;
      frameNumber = 0; eventNumber = 0;
      xScale = -1; yScale = -1;
      imageRotation = 0; fspare1 = -1;
      fspare2 = -1; fspare3 = -1;
      imageFlags = 0x1; ispare1 = -1;
      ispare2 = -1; ispare3 = -1;
      appendedFrameSize = -1;
      try
      {
        String[] vals = inpt.Split(StrUtils.parserDelimiter, StringSplitOptions.RemoveEmptyEntries);
        sourceWidth = Int32.Parse(vals[0]);
        sourceHeight = Int32.Parse(vals[1]);
        aoiWidth = Int32.Parse(vals[2]);
        aoiHeight = Int32.Parse(vals[3]);
        xStart = Int32.Parse(vals[4]);
        yStart = Int32.Parse(vals[5]);
        bytesPerPixel = Int32.Parse(vals[6]);
        effectiveBitsPerPixel = Int32.Parse(vals[7]);
        horizontalBinning = Int32.Parse(vals[8]);
        verticalBinning = Int32.Parse(vals[9]);
        sourceFormat = Int32.Parse(vals[10]);
        imageFormat = Int32.Parse(vals[11]);
        frameNumber = Int32.Parse(vals[12]);
        eventNumber = Int32.Parse(vals[13]);
        xScale = float.Parse(vals[14]);
        yScale = float.Parse(vals[15]);
        imageRotation = float.Parse(vals[16]);
        fspare1 = float.Parse(vals[17]);
        fspare2 = float.Parse(vals[18]);
        fspare3 = float.Parse(vals[19]);
        imageFlags = Int32.Parse(vals[20]);
        ispare1 = Int32.Parse(vals[21]);
        ispare2 = Int32.Parse(vals[22]);
        ispare3 = Int32.Parse(vals[23]);
        appendedFrameSize = Int32.Parse(vals[24]);
      }
      catch (Exception) { }
    }

    public Int32 GetSourceWidth()
    {
      return sourceWidth;
    }
    public void SetSourceWidth(Int32 value)
    {
      sourceWidth = value;
    }
    public Int32 GetSourceHeight()
    {
      return sourceHeight;
    }
    public void SetSourceHeight(Int32 value)
    {
      sourceHeight = value;
    }
    public Int32 GetAoiWidth()
    {
      return aoiWidth;
    }
    public void SetAoiWidth(Int32 value)
    {
      aoiWidth = value;
    }
    public Int32 GetAoiHeight()
    {
      return aoiHeight;
    }
    public void SetAoiHeight(Int32 value)
    {
      aoiHeight = value;
    }
    public Int32 GetXStart()
    {
      return xStart;
    }
    public void SetXStart(Int32 value)
    {
      xStart = value;
    }
    public Int32 GetYStart()
    {
      return yStart;
    }
    public void SetYStart(Int32 value)
    {
      yStart = value;
    }
    public Int32 GetBytesPerPixel()
    {
      return bytesPerPixel;
    }
    public void SetBytesPerPixel(Int32 value)
    {
      bytesPerPixel = value;
    }
    public Int32 GetEffectiveBitsPerPixel()
    {
      return effectiveBitsPerPixel;
    }
    public void SetEffectiveBitsPerPixel(Int32 value)
    {
      effectiveBitsPerPixel = value;
    }
    public Int32 GetHorizontalBinning()
    {
      return horizontalBinning;
    }
    public void SetHorizontalBinning(Int32 value)
    {
      horizontalBinning = value;
    }
    public Int32 GetVerticalBinning()
    {
      return verticalBinning;
    }
    public void SetVerticalBinning(Int32 value)
    {
      verticalBinning = value;
    }
    public Int32 GetSourceFormat()
    {
      return sourceFormat;
    }
    public void SetSourceFormat(Int32 value)
    {
      sourceFormat = value;
    }
    public Int32 GetImageFormat()
    {
      return imageFormat;
    }
    public void SetImageFormat(Int32 value)
    {
      imageFormat = value;
    }
    public Int32 GetFrameNumber()
    {
      return frameNumber;
    }
    public void SetFrameNumber(Int32 value)
    {
      frameNumber = value;
    }
    public Int32 GetEventNumber()
    {
      return eventNumber;
    }
    public void SetEventNumber(Int32 value)
    {
      eventNumber = value;
    }
    public float GetXScale()
    {
      return xScale;
    }
    public void SetXScale(float value)
    {
      xScale = value;
    }
    public float GetYScale()
    {
      return yScale;
    }
    public void SetYScale(float value)
    {
      yScale = value;
    }
    public float GetImageRotation()
    {
      return imageRotation;
    }
    public void SetImageRotation(float value)
    {
      imageRotation = value;
    }
    public float GetFspare1()
    {
      return fspare1;
    }
    public void SetFspare1(float value)
    {
      fspare1 = value;
    }
    public float GetFspare2()
    {
      return fspare2;
    }
    public void SetFspare2(float value)
    {
      fspare2 = value;
    }
    public float GetFspare3()
    {
      return fspare3;
    }
    public void SetFspare3(float value)
    {
      fspare3 = value;
    }
    public Int32 GetImageFlags()
    {
      return imageFlags;
    }
    public void SetImageFlags(Int32 value)
    {
      imageFlags = value;
    }
    public Int32 GetIspare1()
    {
      return ispare1;
    }
    public void SetIspare1(Int32 value)
    {
      ispare1 = value;
    }
    public Int32 GetIspare2()
    {
      return ispare2;
    }
    public void SetIspare2(Int32 value)
    {
      ispare2 = value;
    }
    public Int32 GetIspare3()
    {
      return ispare3;
    }
    public void SetIspare3(Int32 value)
    {
      ispare3 = value;
    }
    public Int32 GetAppendedFrameSize()
    {
      return appendedFrameSize;
    }
    public void SetAppendedFrameSize(Int32 value)
    {
      appendedFrameSize = value;
    }

    public override string ToString()
    {
      StringBuilder sb = new StringBuilder();
      sb.Append("aoiHeight = ").Append(aoiHeight);
      sb.Append(", aoiWidth = ").Append(aoiWidth);
      sb.Append(", appendedFrameSize = ").Append(appendedFrameSize);
      sb.Append(", bytesPerPixel = ").Append(bytesPerPixel);
      sb.Append(", effectiveBitsPerPixel = ").Append(effectiveBitsPerPixel);
      sb.Append(", eventNumber = ").Append(eventNumber);
      sb.Append(", frameNumber = ").Append(frameNumber);
      sb.Append(", fspare1 = ").Append(fspare1);
      sb.Append(", fspare2 = ").Append(fspare2);
      sb.Append(", fspare3 = ").Append(fspare3);
      sb.Append(", horizontalBinning = ").Append(horizontalBinning);
      sb.Append(", imageFlags = ").Append(imageFlags);
      sb.Append(", imageFormat = ").Append(imageFormat);
      sb.Append(", imageRotation = ").Append(imageRotation);
      sb.Append(", ispare1 = ").Append(ispare1);
      sb.Append(", ispare2 = ").Append(ispare2);
      sb.Append(", ispare3 = ").Append(ispare3);
      sb.Append(", sourceFormat = ").Append(sourceFormat);
      sb.Append(", sourceHeight = ").Append(sourceHeight);
      sb.Append(", sourceWidth = ").Append(sourceWidth);
      sb.Append(", verticalBinning = ").Append(verticalBinning);
      sb.Append(", xScale = ").Append(xScale);
      sb.Append(", xStart = ").Append(xStart);
      sb.Append(", yScale = ").Append(yScale);
      sb.Append(", yStart = ").Append(yStart);
      return sb.ToString();
    }
  }
  [StructLayout(LayoutKind.Sequential, Pack = 8, CharSet = CharSet.Ansi)]
  public unsafe struct TImageHeader
  { // source header
    public UInt32 baseTag;
    public UInt32 cameraPortId;
    public UInt32 versionTag;
    public UInt32 totalLength;
    public Int32 timestampSeconds;
    public Int32 timestampMicroseconds;
    public fixed byte cameraPortName[64];
    // frame header
    public Int32 sourceWidth;
    public Int32 sourceHeight;
    public Int32 aoiWidth;
    public Int32 aoiHeight;
    public Int32 xStart;
    public Int32 yStart;
    public Int32 bytesPerPixel;
    public Int32 effectiveBitsPerPixel;
    public Int32 horizontalBinning;
    public Int32 verticalBinning;
    public Int32 sourceFormat;
    public Int32 imageFormat;
    public Int32 frameNumber;
    public Int32 eventNumber;
    public float xScale;
    public float yScale;
    public float imageRotation;
    public float fspare1;
    public float fspare2;
    public float fspare3;
    public Int32 imageFlags;
    public Int32 ispare1;
    public Int32 ispare2;
    public Int32 ispare3;
    public Int32 appendedFrameSize;
    // extras
    public UInt32 frameBufferSize;
    public IntPtr frameBuffer;

    public override string ToString()
    {
      StringBuilder sb = new StringBuilder();
      sb.Append("aoiHeight = ").Append(aoiHeight);
      sb.Append(", aoiWidth = ").Append(aoiWidth);
      sb.Append(", appendedFrameSize = ").Append(appendedFrameSize);
      sb.Append(", bytesPerPixel = ").Append(bytesPerPixel);
      sb.Append(", effectiveBitsPerPixel = ").Append(effectiveBitsPerPixel);
      sb.Append(", eventNumber = ").Append(eventNumber);
      sb.Append(", frameNumber = ").Append(frameNumber);
      sb.Append(", fspare1 = ").Append(fspare1);
      sb.Append(", fspare2 = ").Append(fspare2);
      sb.Append(", fspare3 = ").Append(fspare3);
      sb.Append(", horizontalBinning = ").Append(horizontalBinning);
      sb.Append(", imageFlags = ").Append(imageFlags);
      sb.Append(", imageFormat = ").Append(imageFormat);
      sb.Append(", imageRotation = ").Append(imageRotation);
      sb.Append(", ispare1 = ").Append(ispare1);
      sb.Append(", ispare2 = ").Append(ispare2);
      sb.Append(", ispare3 = ").Append(ispare3);
      sb.Append(", sourceFormat = ").Append(sourceFormat);
      sb.Append(", sourceHeight = ").Append(sourceHeight);
      sb.Append(", sourceWidth = ").Append(sourceWidth);
      sb.Append(", verticalBinning = ").Append(verticalBinning);
      sb.Append(", xScale = ").Append(xScale);
      sb.Append(", xStart = ").Append(xStart);
      sb.Append(", yScale = ").Append(yScale);
      sb.Append(", yStart = ").Append(yStart);
      return sb.ToString();
    }
  }
  [StructLayout(LayoutKind.Sequential, Pack = 8, CharSet = CharSet.Ansi)]
  public unsafe struct DIMAGE
  {
    public const int headerSizeInBytes = 188;
    public TImageHeader header;
    public UInt32 frameBufferSize;
    public byte[] frameBuffer;
    public GCHandle gchFrame;
    public DIMAGE(int length)
    {
      header = new TImageHeader();
      frameBufferSize = (UInt32)length;
      frameBuffer = new byte[length];
      header.frameBufferSize = Convert.ToUInt32(length);
      header.frameBuffer = TDataType.GetDataPtr(frameBuffer);
      gchFrame = GCHandle.Alloc(header.frameBuffer, GCHandleType.Pinned);
    }
    public DIMAGE(DIMAGE reference)
    {
      header = reference.header;
      frameBufferSize = reference.frameBufferSize;
      frameBuffer = reference.frameBuffer;
      gchFrame = reference.gchFrame;
    }
    public byte[] GetFrameBuffer()
    {
      return frameBuffer;
    }
    public TImageHeader GetFrameHeader()
    {
      return header;
    }
    public void SetFrameHeader(TImageHeader hdr)
    {
      IntPtr fbp = header.frameBuffer;
      uint fbs = header.frameBufferSize;
      header = hdr;
      header.frameBuffer = fbp; // absolutely not allowed to change ...
      header.frameBufferSize = fbs;
    }
  }
#pragma warning restore 1591

#pragma warning disable 1591
  public static class StrUtils
  {
    public static string[] parserDelimiter = new string[] { " ", "," };
    public unsafe static string NameToString(byte* ba,int len)
    {
      char[] n = new char[len];
      int nc;
      for (nc = 0; nc < len && ba[nc] != 0; nc++) n[nc] = (char)ba[nc];
      return new String(n, 0, nc).Trim();
    }
    public unsafe static void StringToName(string s, byte* ba, int len)
    {
      int i = 0;
      if (s == null) return;
      foreach (char c in s)
      {
        ba[i++] = (byte)c;
        if (i >= len) break;
      }
    }
    public static string Trim(string srcString)
    {
      if (srcString == null) return null;
      return Trim(srcString.ToCharArray());
    }
    public static string Trim(char[] srcChars)
    {
      if (srcChars == null) return null;
      int nc;
      for (nc = 0; nc < srcChars.Length && srcChars[nc] != 0; nc++) ;
      return new String(srcChars, 0, nc).Trim();
    }
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct NAME8
  {
    public const int sizeInBytes = 8;
    public const int format = Formats.CF_NAME8;
    public fixed byte name[sizeInBytes];
    public NAME8(String name)
    {
      SetName(name);
    }
    public string GetName()
    {
      fixed (byte* pca = name) return StrUtils.NameToString(pca, sizeInBytes);
    }
    public void SetName(string name)
    {
      fixed (byte* pca = this.name) StrUtils.StringToName(name, pca, sizeInBytes);
    }
    public void SetName(byte[] name)
    {
      fixed (byte* pca = this.name) for (int i = 0; i < sizeInBytes && i < name.Length; i++) pca[i] = name[i];
    }
    public override String ToString()
    {
      return GetName();
    }
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct NAME16
  {
    public const int sizeInBytes = 16;
    public const int format = Formats.CF_NAME16;
    public fixed byte name[sizeInBytes];
    public NAME16(String name)
    {
      SetName(name);
    }
    public string GetName()
    {
      fixed (byte* pca = name) return StrUtils.NameToString(pca, sizeInBytes);
    }
    public void SetName(string name)
    {
      fixed (byte* pca = this.name) StrUtils.StringToName(name, pca, sizeInBytes);
    }
    public void SetName(byte[] name)
    {
      fixed (byte* pca = this.name) for (int i = 0; i < sizeInBytes && i < name.Length; i++) pca[i] = name[i];
    }
    public override String ToString()
    {
      return GetName();
    }
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct NAME32
  {
    public const int sizeInBytes = 32;
    public const int format = Formats.CF_NAME32;
    public fixed byte name[sizeInBytes];
    public NAME32(String name)
    {
      SetName(name);
    }
    public string GetName()
    {
      fixed (byte* pca = name) return StrUtils.NameToString(pca, sizeInBytes);
    }
    public void SetName(string name)
    {
      fixed (byte* pca = this.name) StrUtils.StringToName(name, pca, sizeInBytes);
    }
    public void SetName(byte[] name)
    {
      fixed (byte* pca = this.name) for (int i = 0; i < sizeInBytes && i < name.Length; i++) pca[i] = name[i];
    }
    public override String ToString()
    {
      return GetName();
    }
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct NAME48
  {
    public const int sizeInBytes = 48;
    public const int format = Formats.CF_NAME48;
    public fixed byte name[sizeInBytes];
    public NAME48(String name)
    {
      SetName(name);
    }
    public string GetName()
    {
      fixed (byte* pca = name) return StrUtils.NameToString(pca, sizeInBytes);
    }
    public void SetName(string name)
    {
      fixed (byte* pca = this.name) StrUtils.StringToName(name, pca, sizeInBytes);
    }
    public void SetName(byte[] name)
    {
      fixed (byte* pca = this.name) for (int i = 0; i < sizeInBytes && i < name.Length; i++) pca[i] = name[i];
    }
    public override String ToString()
    {
      return GetName();
    }
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct NAME64
  {
    public const int sizeInBytes = 64;
    public const int format = Formats.CF_NAME64;
    public fixed byte name[sizeInBytes];
    public NAME64(String name)
    {
      SetName(name);
    }
    public string GetName()
    {
      fixed (byte* pca = name) return StrUtils.NameToString(pca, sizeInBytes);
    }
    public void SetName(string name)
    {
      fixed (byte* pca = this.name) StrUtils.StringToName(name, pca, sizeInBytes);
    }
    public void SetName(byte[] name)
    {
      fixed (byte* pca = this.name) for (int i = 0; i < sizeInBytes && i < name.Length; i++) pca[i] = name[i];
    }
    public override String ToString()
    {
      return GetName();
    }
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct FLTINT
  {
    public const int sizeInBytes = 8;
    public const int format = Formats.CF_FLTINT;
    public float fval;
    public Int32 ival;
    public FLTINT(String inpt)
    {
      fval = 0; ival = 0;
      try
      {
        String[] vals = inpt.Split(StrUtils.parserDelimiter, StringSplitOptions.RemoveEmptyEntries);
        fval = Single.Parse(vals[0]);
        ival = Int32.Parse(vals[1]);
      }
      catch (Exception) { }
    }
    public FLTINT(float value1, int value2)
    {
      fval = value1;
      ival = value2;
    }
    public float GetFloat()
    {
      return fval;
    }
    public void SetFloat(float value)
    {
      fval = value;
    }
    public Int32 GetInteger()
    {
      return ival;
    }
    public void SetInteger(Int32 value)
    {
      ival = value;
    }
    public override String ToString()
    {
      return "" + fval + " " + ival;
    }
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct INTINT
  {
    public const int sizeInBytes = 8;
    public const int format = Formats.CF_INTINT;
    public Int32 i1val;
    public Int32 i2val;
    public INTINT(String inpt)
    {
      i1val = 0; i2val = 0;
      try
      {
        String[] vals = inpt.Split(StrUtils.parserDelimiter, StringSplitOptions.RemoveEmptyEntries);
        i1val = Int32.Parse(vals[0]);
        i2val = Int32.Parse(vals[1]);
      }
      catch (Exception) { }
    }
    public Int32 GetPrimaryInteger()
    {
      return i1val;
    }
    public void SetPrimaryInteger(Int32 value)
    {
      i1val = value;
    }
    public Int32 GetInteger()
    {
      return i2val;
    }
    public void SetInteger(Int32 value)
    {
      i2val = value;
    }
    public override String ToString()
    {
      return "" + i1val + " " + i2val;
    }
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct FLTFLT
  {
    public const int sizeInBytes = 8;
    public const int format = Formats.CF_FLTFLT;
    public float f1val;
    public float f2val;
    public FLTFLT(String inpt)
    {
      f1val = 0; f2val = 0;
      try
      {
        String[] vals = inpt.Split(StrUtils.parserDelimiter, StringSplitOptions.RemoveEmptyEntries);
        f1val = Single.Parse(vals[0]);
        f2val = Single.Parse(vals[1]);
      }
      catch (Exception) { }
    }
    public float GetPrimaryFloat()
    {
      return f1val;
    }
    public void SetPrimaryFloat(float value)
    {
      f1val = value;
    }
    public float GetSecondaryFloat()
    {
      return f2val;
    }
    public void SetSecondaryFloat(float value)
    {
      f2val = value;
    }
    public override String ToString()
    {
      return "" + f1val + " " + f2val;
    }
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct DBLDBL
  {
    public const int sizeInBytes = 16;
    public const int format = Formats.CF_DBLDBL;
    public double d1val;
    public double d2val;
    public DBLDBL(String inpt)
    {
      d1val = 0; d2val = 0;
      try
      {
        String[] vals = inpt.Split(StrUtils.parserDelimiter, StringSplitOptions.RemoveEmptyEntries);
        d1val = Double.Parse(vals[0]);
        d2val = Double.Parse(vals[1]);
      }
      catch (Exception) { }
    }
    public DBLDBL(double value1, double value2)
    {
      d1val = value1;
      d2val = value2;
    }
    public double GetPrimaryDouble()
    {
      return d1val;
    }
    public void SetPrimaryDouble(double value)
    {
      d1val = value;
    }
    public double GetSecondaryDouble()
    {
      return d2val;
    }
    public void SetSecondaryDouble(double value)
    {
      d2val = value;
    }
    public override String ToString()
    {
      return "" + d1val + " " + d2val;
    }
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct NAME8I
  {
    public const int sizeInBytes = 12;
    public const int format = Formats.CF_NAME8I;
    public fixed byte name[sizeInBytes - 4];
    public Int32 ival;
    public NAME8I(String inpt)
    {
      ival = 0;
      try
      {
        String[] vals = inpt.Split(StrUtils.parserDelimiter, StringSplitOptions.RemoveEmptyEntries);
        SetName(vals[0]);
        ival = Int32.Parse(vals[1]);
      }
      catch (Exception) { }
    }
    public string GetName()
    {
      fixed (byte* pca = name) return StrUtils.NameToString(pca, sizeInBytes - 4);
    }
    public void SetName(string name)
    {
      fixed (byte* pca = this.name) StrUtils.StringToName(name, pca, sizeInBytes - 4);
    }
    public void SetName(byte[] name)
    {
      fixed (byte* pca = this.name) for (int i = 0; i < sizeInBytes - 4 && i < name.Length; i++) pca[i] = name[i];
    }
    public Int32 GetInteger()
    {
      return ival;
    }
    public void SetInteger(Int32 value)
    {
      ival = value;
    }
    public override String ToString()
    {
      return GetName() + " " + ival;
    }
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct NAME16I
  {
    public const int sizeInBytes = 20;
    public const int format = Formats.CF_NAME16I;
    public fixed byte name[sizeInBytes - 4];
    public Int32 ival;
    public NAME16I(String inpt)
    {
      ival = 0;
      try
      {
        String[] vals = inpt.Split(StrUtils.parserDelimiter, StringSplitOptions.RemoveEmptyEntries);
        SetName(vals[0]);
        ival = Int32.Parse(vals[1]);
      }
      catch (Exception) { }
    }
    public string GetName()
    {
      fixed (byte* pca = name) return StrUtils.NameToString(pca, sizeInBytes - 4);
    }
    public void SetName(string name)
    {
      fixed (byte* pca = this.name) StrUtils.StringToName(name, pca, sizeInBytes - 4);
    }
    public void SetName(byte[] name)
    {
      fixed (byte* pca = this.name) for (int i = 0; i < sizeInBytes - 4 && i < name.Length; i++) pca[i] = name[i];
    }
    public Int32 GetInteger()
    {
      return ival;
    }
    public void SetInteger(Int32 value)
    {
      ival = value;
    }
    public override String ToString()
    {
      return GetName() + " " + ival;
    }
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct NAME32I
  {
    public const int sizeInBytes = 36;
    public const int format = Formats.CF_NAME32I;
    public fixed byte name[sizeInBytes - 4];
    public Int32 ival;
    public NAME32I(String inpt)
    {
      ival = 0;
      try
      {
        String[] vals = inpt.Split(StrUtils.parserDelimiter, StringSplitOptions.RemoveEmptyEntries);
        SetName(vals[0]);
        ival = Int32.Parse(vals[1]);
      }
      catch (Exception) { }
    }
    public string GetName()
    {
      fixed (byte* pca = name) return StrUtils.NameToString(pca, sizeInBytes - 4);
    }
    public void SetName(string name)
    {
      fixed (byte* pca = this.name) StrUtils.StringToName(name, pca, sizeInBytes - 4);
    }
    public void SetName(byte[] name)
    {
      fixed (byte* pca = this.name) for (int i = 0; i < sizeInBytes - 4 && i < name.Length; i++) pca[i] = name[i];
    }
    public Int32 GetInteger()
    {
      return ival;
    }
    public void SetInteger(Int32 value)
    {
      ival = value;
    }
    public override String ToString()
    {
      return GetName() + " " + ival;
    }
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct NAME48I
  {
    public const int sizeInBytes = 52;
    public const int format = Formats.CF_NAME48I;
    public fixed byte name[sizeInBytes - 4];
    public Int32 ival;
    public NAME48I(String inpt)
    {
      ival = 0;
      try
      {
        String[] vals = inpt.Split(StrUtils.parserDelimiter, StringSplitOptions.RemoveEmptyEntries);
        SetName(vals[0]);
        ival = Int32.Parse(vals[1]);
      }
      catch (Exception) { }
    }
    public string GetName()
    {
      fixed (byte* pca = name) return StrUtils.NameToString(pca, sizeInBytes - 4);
    }
    public void SetName(string name)
    {
      fixed (byte* pca = this.name) StrUtils.StringToName(name, pca, sizeInBytes - 4);
    }
    public void SetName(byte[] name)
    {
      fixed (byte* pca = this.name) for (int i = 0; i < sizeInBytes - 4 && i < name.Length; i++) pca[i] = name[i];
    }
    public Int32 GetInteger()
    {
      return ival;
    }
    public void SetInteger(Int32 value)
    {
      ival = value;
    }
    public override String ToString()
    {
      return GetName() + " " + ival;
    }
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct NAME64I
  {
    public const int sizeInBytes = 68;
    public const int format = Formats.CF_NAME64I;
    public fixed byte name[sizeInBytes - 4];
    public Int32 ival;
    public NAME64I(String inpt)
    {
      ival = 0;
      try
      {
        String[] vals = inpt.Split(StrUtils.parserDelimiter, StringSplitOptions.RemoveEmptyEntries);
        SetName(vals[0]);
        ival = Int32.Parse(vals[1]);
      }
      catch (Exception) { }
    }
    public string GetName()
    {
      fixed (byte* pca = name) return StrUtils.NameToString(pca, sizeInBytes - 4);
    }
    public void SetName(string name)
    {
      fixed (byte* pca = this.name) StrUtils.StringToName(name, pca, sizeInBytes - 4);
    }
    public void SetName(byte[] name)
    {
      fixed (byte* pca = this.name) for (int i = 0; i < sizeInBytes - 4 && i < name.Length; i++) pca[i] = name[i];
    }
    public Int32 GetInteger()
    {
      return ival;
    }
    public void SetInteger(Int32 value)
    {
      ival = value;
    }
    public override String ToString()
    {
      return GetName() + " " + ival;
    }
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct NAME64DBL
  {
    public const int sizeInBytes = 72;
    public const int format = Formats.CF_NAME64DBL;
    public fixed byte name[sizeInBytes - 8];
    public double dval;
    public NAME64DBL(String inpt)
    {
      dval = 0;
      try
      {
        String[] vals = inpt.Split(StrUtils.parserDelimiter, StringSplitOptions.RemoveEmptyEntries);
        SetName(vals[0]);
        dval = Double.Parse(vals[1]);
      }
      catch (Exception) { }
    }
    public string GetName()
    {
      fixed (byte* pca = name) return StrUtils.NameToString(pca, sizeInBytes - 8);
    }
    public void SetName(string name)
    {
      fixed (byte* pca = this.name) StrUtils.StringToName(name, pca, sizeInBytes - 8);
    }
    public void SetName(byte[] name)
    {
      fixed (byte* pca = this.name) for (int i = 0; i < sizeInBytes - 8 && i < name.Length; i++) pca[i] = name[i];
    }
    public double GetDouble()
    {
      return dval;
    }
    public void SetDouble(double value)
    {
      dval = value;
    }
    public override String ToString()
    {
      return GetName() + " " + dval;
    }
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct NAME16FI
  {
    public const int sizeInBytes = 24;
    public const int format = Formats.CF_NAME16FI;
    public fixed byte name[sizeInBytes - 8];
    public float fval;
    public Int32 ival;
    public NAME16FI(String inpt)
    {
      fval = 0; ival = 0;
      try
      {
        String[] vals = inpt.Split(StrUtils.parserDelimiter, StringSplitOptions.RemoveEmptyEntries);
        SetName(vals[0]);
        fval = Single.Parse(vals[1]);
        ival = Int32.Parse(vals[2]);
      }
      catch (Exception) { }
    }
    public string GetName()
    {
      fixed (byte* pca = name) return StrUtils.NameToString(pca, sizeInBytes - 8);
    }
    public void SetName(string name)
    {
      fixed (byte* pca = this.name) StrUtils.StringToName(name, pca, sizeInBytes - 8);
    }
    public void SetName(byte[] name)
    {
      fixed (byte* pca = this.name) for (int i = 0; i < sizeInBytes - 8 && i < name.Length; i++) pca[i] = name[i];
    }
    public float GetFloat()
    {
      return fval;
    }
    public void SetFloat(float value)
    {
      fval = value;
    }
    public Int32 GetInteger()
    {
      return ival;
    }
    public void SetInteger(Int32 value)
    {
      ival = value;
    }
    public override String ToString()
    {
      return GetName() + " " + fval + " " + ival;
    }
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct NAME16II
  {
    public const int sizeInBytes = 24;
    public const int format = Formats.CF_NAME16INTINT;
    public fixed byte name[sizeInBytes - 8];
    public Int32 i1val;
    public Int32 i2val;
    public NAME16II(String inpt)
    {
      i1val = 0; i2val = 0;
      try
      {
        String[] vals = inpt.Split(StrUtils.parserDelimiter, StringSplitOptions.RemoveEmptyEntries);
        SetName(vals[0]);
        i1val = Int32.Parse(vals[1]);
        i2val = Int32.Parse(vals[2]);
      }
      catch (Exception) { }
    }
    public string GetName()
    {
      fixed (byte* pca = name) return StrUtils.NameToString(pca, sizeInBytes - 8);
    }
    public void SetName(string name)
    {
      fixed (byte* pca = this.name) StrUtils.StringToName(name, pca, sizeInBytes - 8);
    }
    public void SetName(byte[] name)
    {
      fixed (byte* pca = this.name) for (int i = 0; i < sizeInBytes - 8 && i < name.Length; i++) pca[i] = name[i];
    }
    public float GetPrimaryInteger()
    {
      return i1val;
    }
    public void SetPrimaryInteger(Int32 value)
    {
      i1val = value;
    }
    public Int32 GetSecondaryInteger()
    {
      return i2val;
    }
    public void SetSecondaryInteger(Int32 value)
    {
      i2val = value;
    }
    public override String ToString()
    {
      return GetName() + " " + i1val + " " + i2val;
    }
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct NAME16DBLDBL
  {
    public const int sizeInBytes = 32;
    public const int format = Formats.CF_NAME16DBLDBL;
    public fixed byte name[sizeInBytes - 16];
    public double d1val;
    public double d2val;
    public NAME16DBLDBL(String inpt)
    {
      d1val = 0; d2val = 0;
      try
      {
        String[] vals = inpt.Split(StrUtils.parserDelimiter, StringSplitOptions.RemoveEmptyEntries);
        SetName(vals[0]);
        d1val = Double.Parse(vals[1]);
        d2val = Double.Parse(vals[2]);
      }
      catch (Exception) { }
    }
    public string GetName()
    {
      fixed (byte* pca = name) return StrUtils.NameToString(pca, sizeInBytes - 16);
    }
    public void SetName(string name)
    {
      fixed (byte* pca = this.name) StrUtils.StringToName(name, pca, sizeInBytes - 16);
    }
    public void SetName(byte[] name)
    {
      fixed (byte* pca = this.name) for (int i = 0; i < sizeInBytes - 16 && i < name.Length; i++) pca[i] = name[i];
    }
    public double GetPrimaryDouble()
    {
      return d1val;
    }
    public void SetPrimaryDouble(double value)
    {
      d1val = value;
    }
    public double GetSecondaryDouble()
    {
      return d2val;
    }
    public void SetSecondaryDouble(double value)
    {
      d2val = value;
    }
    public override String ToString()
    {
      return GetName() + " " + d1val + " " + d2val;
    }
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct NAME32DBLDBL
  {
    public const int sizeInBytes = 48;
    public const int format = Formats.CF_NAME32DBLDBL;
    public fixed byte name[sizeInBytes - 16];
    public double d1val;
    public double d2val;
    public NAME32DBLDBL(String inpt)
    {
      d1val = 0; d2val = 0;
      try
      {
        String[] vals = inpt.Split(StrUtils.parserDelimiter, StringSplitOptions.RemoveEmptyEntries);
        SetName(vals[0]);
        d1val = Double.Parse(vals[1]);
        d2val = Double.Parse(vals[2]);
      }
      catch (Exception) { }
    }
    public string GetName()
    {
      fixed (byte* pca = name) return StrUtils.NameToString(pca, sizeInBytes - 16);
    }
    public void SetName(string name)
    {
      fixed (byte* pca = this.name) StrUtils.StringToName(name, pca, sizeInBytes - 16);
    }
    public void SetName(byte[] name)
    {
      fixed (byte* pca = this.name) for (int i = 0; i < sizeInBytes - 16 && i < name.Length; i++) pca[i] = name[i];
    }
    public double GetPrimaryDouble()
    {
      return d1val;
    }
    public void SetPrimaryDouble(double value)
    {
      d1val = value;
    }
    public double GetSecondaryDouble()
    {
      return d2val;
    }
    public void SetSecondaryDouble(double value)
    {
      d2val = value;
    }
    public override String ToString()
    {
      return GetName() + " " + d1val + " " + d2val;
    }
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct NAME64DBLDBL
  {
    public const int sizeInBytes = 80;
    public const int format = Formats.CF_NAME64DBLDBL;
    public fixed byte name[sizeInBytes - 16];
    public double d1val;
    public double d2val;
    public NAME64DBLDBL(String inpt)
    {
      d1val = 0; d2val = 0;
      try
      {
        String[] vals = inpt.Split(StrUtils.parserDelimiter, StringSplitOptions.RemoveEmptyEntries);
        SetName(vals[0]);
        d1val = Double.Parse(vals[1]);
        d2val = Double.Parse(vals[2]);
      }
      catch (Exception) { }
    }
    public string GetName()
    {
      fixed (byte* pca = name) return StrUtils.NameToString(pca, sizeInBytes - 16);
    }
    public void SetName(string name)
    {
      fixed (byte* pca = this.name) StrUtils.StringToName(name, pca, sizeInBytes - 16);
    }
    public void SetName(byte[] name)
    {
      fixed (byte* pca = this.name) for (int i = 0; i < sizeInBytes - 16 && i < name.Length; i++) pca[i] = name[i];
    }
    public double GetPrimaryDouble()
    {
      return d1val;
    }
    public void SetPrimaryDouble(double value)
    {
      d1val = value;
    }
    public double GetSecondaryDouble()
    {
      return d2val;
    }
    public void SetSecondaryDouble(double value)
    {
      d2val = value;
    }
    public override String ToString()
    {
      return GetName() + " " + d1val + " " + d2val;
    }
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct INTFLTINT
  {
    public const int sizeInBytes = 12;
    public const int format = Formats.CF_INTFLTINT;
    public Int32 i1val;
    public float fval;
    public Int32 i2val;
    public INTFLTINT(String inpt)
    {
      i1val = 0; fval = 0; i2val = 0;
      try
      {
        String[] vals = inpt.Split(StrUtils.parserDelimiter, StringSplitOptions.RemoveEmptyEntries);
        i1val = Int32.Parse(vals[1]);
        fval = Single.Parse(vals[1]);
        i2val = Int32.Parse(vals[2]);
      }
      catch (Exception) { }
    }
    public float GetFloat()
    {
      return fval;
    }
    public void SetFloat(float value)
    {
      fval = value;
    }
    public Int32 GetPrimaryInteger()
    {
      return i1val;
    }
    public void SetPrimaryInteger(Int32 value)
    {
      i1val = value;
    }
    public Int32 GetSecondaryInteger()
    {
      return i2val;
    }
    public void SetSecondaryInteger(Int32 value)
    {
      i2val = value;
    }
    public override String ToString()
    {
      return "" + i1val + " " + fval + " " + i2val;
    }
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct FLTFLTINT
  {
    public const int sizeInBytes = 12;
    public const int format = Formats.CF_FLTFLTINT;
    public float f1val;
    public float f2val;
    public Int32 ival;
    public FLTFLTINT(String inpt)
    {
      f1val = 0; f2val = 0; ival = 0;
      try
      {
        String[] vals = inpt.Split(StrUtils.parserDelimiter, StringSplitOptions.RemoveEmptyEntries);
        f1val = Single.Parse(vals[1]);
        f2val = Single.Parse(vals[1]);
        ival = Int32.Parse(vals[2]);
      }
      catch (Exception) { }
    }
    public float GetPrimaryFloat()
    {
      return f1val;
    }
    public void SetPrimaryFloat(float value)
    {
      f1val = value;
    }
    public float GetSecondaryFloat()
    {
      return f2val;
    }
    public void SetSecondaryFloat(float value)
    {
      f2val = value;
    }
    public Int32 GetInteger()
    {
      return ival;
    }
    public void SetInteger(Int32 value)
    {
      ival = value;
    }
    public override String ToString()
    {
      return "" + f1val + " " + f2val + " " + ival;
    }
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct FLTINTINT
  {
    public const int sizeInBytes = 12;
    public const int format = Formats.CF_FLTINTINT;
    public float fval;
    public Int32 i1val;
    public Int32 i2val;
    public FLTINTINT(String inpt)
    {
      fval = 0; i1val = 0; i2val = 0;
      try
      {
        String[] vals = inpt.Split(StrUtils.parserDelimiter, StringSplitOptions.RemoveEmptyEntries);
        fval = Single.Parse(vals[1]);
        i1val = Int32.Parse(vals[1]);
        i2val = Int32.Parse(vals[2]);
      }
      catch (Exception) { }
    }
    public float GetFloat()
    {
      return fval;
    }
    public void SetFloat(float value)
    {
      fval = value;
    }
    public Int32 GetPrimaryInteger()
    {
      return i1val;
    }
    public void SetPrimaryInteger(Int32 value)
    {
      i1val = value;
    }
    public Int32 GetSecondaryInteger()
    {
      return i2val;
    }
    public void SetSecondaryInteger(Int32 value)
    {
      i2val = value;
    }
    public override String ToString()
    {
      return "" + fval + " " + i1val + " " + i2val;
    }
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct INTINTINT
  {
    public const int sizeInBytes = 12;
    public const int format = Formats.CF_INTINTINT;
    public Int32 i1val;
    public Int32 i2val;
    public Int32 i3val;
    public INTINTINT(String inpt)
    {
      i1val = 0; i2val = 0; i3val = 0; 
      try
      {
        String[] vals = inpt.Split(StrUtils.parserDelimiter, StringSplitOptions.RemoveEmptyEntries);
        i1val = Int32.Parse(vals[0]);
        i2val = Int32.Parse(vals[1]);
        i3val = Int32.Parse(vals[2]);
      }
      catch (Exception) { }
    }
    public Int32 GetTertiaryInteger()
    {
      return i3val;
    }
    public void SetTertiaryInteger(Int32 value)
    {
      i3val = value;
    }
    public Int32 GetPrimaryInteger()
    {
      return i1val;
    }
    public void SetPrimaryInteger(Int32 value)
    {
      i1val = value;
    }
    public Int32 GetSecondaryInteger()
    {
      return i2val;
    }
    public void SetSecondaryInteger(Int32 value)
    {
      i2val = value;
    }
    public override String ToString()
    {
      return "" + i1val + " " + i2val + " " + i3val;
    }
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct DBLDBLDBL
  {
    public const int sizeInBytes = 12;
    public const int format = Formats.CF_DBLDBLDBL;
    public double d1val;
    public double d2val;
    public double d3val;
    public DBLDBLDBL(String inpt)
    {
      d1val = 0; d2val = 0; d3val = 0;
      try
      {
        String[] vals = inpt.Split(StrUtils.parserDelimiter, StringSplitOptions.RemoveEmptyEntries);
        d1val = Double.Parse(vals[0]);
        d2val = Double.Parse(vals[1]);
        d3val = Double.Parse(vals[2]);
      }
      catch (Exception) { }
    }
    public double GetTertiaryInteger()
    {
      return d3val;
    }
    public void SetTertiaryInteger(double value)
    {
      d3val = value;
    }
    public double GetPrimaryInteger()
    {
      return d1val;
    }
    public void SetPrimaryInteger(double value)
    {
      d1val = value;
    }
    public double GetSecondaryInteger()
    {
      return d2val;
    }
    public void SetSecondaryInteger(double value)
    {
      d2val = value;
    }
    public override String ToString()
    {
      return "" + d1val + " " + d2val + " " + d3val;
    }
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct INTFLTFLTFLT
  {
    public const int sizeInBytes = 16;
    public const int format = Formats.CF_INTFLTFLTFLT;
    public Int32 i1val;
    public float f1val;
    public float f2val;
    public float f3val;
    public INTFLTFLTFLT(String inpt)
    {
      i1val = 0; f1val = 0; f2val = 0; f3val = 0;
      try
      {
        String[] vals = inpt.Split(StrUtils.parserDelimiter, StringSplitOptions.RemoveEmptyEntries);
        i1val = Int32.Parse(vals[0]);
        f1val = Single.Parse(vals[1]);
        f2val = Single.Parse(vals[2]);
        f3val = Single.Parse(vals[3]);
      }
      catch (Exception) { }
    }
    public Int32 GetPrimaryInteger()
    {
      return i1val;
    }
    public void SetPrimaryInteger(Int32 value)
    {
      i1val = value;
    }
    public float GetPrimaryFloat()
    {
      return f1val;
    }
    public void SetPrimaryFloat(float value)
    {
      f1val = value;
    }
    public float GetSecondaryFloat()
    {
      return f2val;
    }
    public void SetSecondaryFloat(float value)
    {
      f2val = value;
    }
    public float GetTertiaryFloat()
    {
      return f3val;
    }
    public void SetTertiaryFloat(float value)
    {
      f3val = value;
    }
    public override String ToString()
    {
      return "" + i1val + " " + f1val + " " + f2val + " " + f3val;
    }
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct INTINTINTINT
  {
    public const int sizeInBytes = 16;
    public const int format = Formats.CF_INTINTINTINT;
    public Int32 i1val;
    public Int32 i2val;
    public Int32 i3val;
    public Int32 i4val;
    public INTINTINTINT(String inpt)
    {
      i1val = 0; i2val = 0; i3val = 0; i4val = 0;
      try
      {
        String[] vals = inpt.Split(StrUtils.parserDelimiter, StringSplitOptions.RemoveEmptyEntries);
        i1val = Int32.Parse(vals[0]);
        i2val = Int32.Parse(vals[1]);
        i3val = Int32.Parse(vals[2]);
        i4val = Int32.Parse(vals[3]);
      }
      catch (Exception) { }
    }
    public INTINTINTINT(int value1, int value2, int value3, int value4)
    {
      i1val = value1;
      i2val = value2;
      i3val = value3;
      i4val = value4;
    }
    public Int32 GetQuaternaryInteger()
    {
      return i4val;
    }
    public void SetQuaternaryInteger(Int32 value)
    {
      i4val = value;
    }
    public Int32 GetTertiaryInteger()
    {
      return i3val;
    }
    public void SetTertiaryInteger(Int32 value)
    {
      i3val = value;
    }
    public Int32 GetPrimaryInteger()
    {
      return i1val;
    }
    public void SetPrimaryInteger(Int32 value)
    {
      i1val = value;
    }
    public Int32 GetSecondaryInteger()
    {
      return i2val;
    }
    public void SetSecondaryInteger(Int32 value)
    {
      i2val = value;
    }
    public override String ToString()
    {
      return "" + i1val + " " + i2val + " " + i3val + " " + i4val;
    }
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct INTINTFLTFLT
  {
    public const int sizeInBytes = 16;
    public const int format = Formats.CF_INTINTFLTFLT;
    public Int32 i1val;
    public Int32 i2val;
    public float f1val;
    public float f2val;
    public INTINTFLTFLT(String inpt)
    {
      i1val = 0; f1val = 0; f2val = 0; i2val = 0;
      try
      {
        String[] vals = inpt.Split(StrUtils.parserDelimiter, StringSplitOptions.RemoveEmptyEntries);
        i1val = Int32.Parse(vals[0]);
        i2val = Int32.Parse(vals[1]);
        f1val = Single.Parse(vals[2]);
        f2val = Single.Parse(vals[3]);
      }
      catch (Exception) { }
    }
    public Int32 GetPrimaryInteger()
    {
      return i1val;
    }
    public void SetPrimaryInteger(Int32 value)
    {
      i1val = value;
    }
    public float GetPrimaryFloat()
    {
      return f1val;
    }
    public void SetPrimaryFloat(float value)
    {
      f1val = value;
    }
    public float GetSecondaryFloat()
    {
      return f2val;
    }
    public void SetSecondaryFloat(float value)
    {
      f2val = value;
    }
    public float GetSecondaryInteger()
    {
      return i2val;
    }
    public void SetSecondaryInteger(Int32 value)
    {
      i2val = value;
    }
    public override String ToString()
    {
      return "" + i1val + " " + i2val + " " + f1val + " " + f2val;
    }
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct FLTINTFLTINT
  {
    public const int sizeInBytes = 16;
    public const int format = Formats.CF_FLTINTFLTINT;
    public float f1val;
    public Int32 i1val;
    public float f2val;
    public Int32 i2val;
    public FLTINTFLTINT(String inpt)
    {
      i1val = 0; f1val = 0; f2val = 0; i2val = 0;
      try
      {
        String[] vals = inpt.Split(StrUtils.parserDelimiter, StringSplitOptions.RemoveEmptyEntries);
        f1val = Single.Parse(vals[0]);
        i1val = Int32.Parse(vals[1]);
        f2val = Single.Parse(vals[2]);
        i2val = Int32.Parse(vals[3]);
      }
      catch (Exception) { }
    }
    public Int32 GetPrimaryInteger()
    {
      return i1val;
    }
    public void SetPrimaryInteger(Int32 value)
    {
      i1val = value;
    }
    public float GetPrimaryFloat()
    {
      return f1val;
    }
    public void SetPrimaryFloat(float value)
    {
      f1val = value;
    }
    public float GetSecondaryFloat()
    {
      return f2val;
    }
    public void SetSecondaryFloat(float value)
    {
      f2val = value;
    }
    public float GetSecondaryInteger()
    {
      return i2val;
    }
    public void SetSecondaryInteger(Int32 value)
    {
      i2val = value;
    }
    public override String ToString()
    {
      return "" + f1val + " " + i1val + " " + f2val + " " + i2val;
    }
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct DWINDOW
  {
    public const int sizeInBytes = 16;
    public const int format = Formats.CF_WINDOW;
    public Int32 t1val;
    public Int32 t2val;
    public Int32 i1val;
    public Int32 i2val;
    public DWINDOW(String inpt)
    {
      i1val = 0; i2val = 0; t1val = 0; t2val = 0;
      try
      {
        String[] vals = inpt.Split(StrUtils.parserDelimiter, StringSplitOptions.RemoveEmptyEntries);
        t1val = Int32.Parse(vals[0]);
        t2val = Int32.Parse(vals[1]);
        i1val = Int32.Parse(vals[2]);
        i2val = Int32.Parse(vals[3]);
      }
      catch (Exception) { }
    }
    public Int32 GetPrimaryUTC()
    {
      return t1val;
    }
    public void SetPrimaryUTC(Int32 value)
    {
      t1val = value;
    }
    public Int32 GetSecondaryUTC()
    {
      return t2val;
    }
    public void SetSecondaryUTC(Int32 value)
    {
      t2val = value;
    }
    public Int32 GetPrimaryInteger()
    {
      return i1val;
    }
    public void SetPrimaryInteger(Int32 value)
    {
      i1val = value;
    }
    public Int32 GetSecondaryInteger()
    {
      return i2val;
    }
    public void SetSecondaryInteger(Int32 value)
    {
      i2val = value;
    }
    public override String ToString()
    {
      return "" + t1val + " " + t2val + " " + i1val + " " + i2val;
    }
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct NAME64DBLDBLDBL
  {
    public const int sizeInBytes = 88;
    public const int format = Formats.CF_NAME64DBLDBLDBL;
    public fixed byte name[sizeInBytes - 24];
    public double d1val;
    public double d2val;
    public double d3val;
    public NAME64DBLDBLDBL(String inpt)
    {
      d1val = 0; d2val = 0; d3val = 0;
      try
      {
        String[] vals = inpt.Split(StrUtils.parserDelimiter, StringSplitOptions.RemoveEmptyEntries);
        SetName(vals[0]);
        d1val = Double.Parse(vals[1]);
        d2val = Double.Parse(vals[2]);
        d3val = Double.Parse(vals[3]);
      }
      catch (Exception) { }
    }
    public string GetName()
    {
      fixed (byte* pca = name) return StrUtils.NameToString(pca, sizeInBytes - 24);
    }
    public void SetName(string name)
    {
      fixed (byte* pca = this.name) StrUtils.StringToName(name, pca, sizeInBytes - 24);
    }
    public void SetName(byte[] name)
    {
      fixed (byte* pca = this.name) for (int i = 0; i < sizeInBytes - 24 && i < name.Length; i++) pca[i] = name[i];
    }
    public double GetPrimaryDouble()
    {
      return d1val;
    }
    public void SetPrimaryDouble(double value)
    {
      d1val = value;
    }
    public double GetSecondaryDouble()
    {
      return d2val;
    }
    public void SetSecondaryDouble(double value)
    {
      d2val = value;
    }
    public double GetTertiaryDouble()
    {
      return d3val;
    }
    public void SetTertiaryDouble(double value)
    {
      d3val = value;
    }
    public override String ToString()
    {
      return GetName() + " " + d1val + " " + d2val + " " + d3val;
    }
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct DUSTRING
  {
    public const int sizeInBytes = 96;
    public const int format = Formats.CF_USTRING;
    Int32 i1val;
    float f1val;
    float f2val;
    Int32 i2val;
    public fixed byte str[80];
    public DUSTRING(String inpt)
    {
      i1val = 0; f1val = 0; f2val = 0; i2val = 0;
      try
      {
        String[] vals = inpt.Split(StrUtils.parserDelimiter, StringSplitOptions.RemoveEmptyEntries);
        i1val = Int32.Parse(vals[0]);
        f1val = Single.Parse(vals[1]);
        f2val = Single.Parse(vals[2]);
        i2val = Int32.Parse(vals[3]);
        SetName(vals[4]);
      }
      catch (Exception) { }
    }
    public string GetName()
    {
      fixed (byte* pca = str) return StrUtils.NameToString(pca, 80);
    }
    public void SetName(string str)
    {
      fixed (byte* pca = this.str) StrUtils.StringToName(str, pca, 80);
    }
    public double GetPrimaryFloat()
    {
      return f1val;
    }
    public void SetPrimaryFloat(float value)
    {
      f1val = value;
    }
    public double GetSecondaryFloat()
    {
      return f2val;
    }
    public void SetSecondaryFloat(float value)
    {
      f2val = value;
    }
    public Int32 GetPrimaryInteger()
    {
      return i1val;
    }
    public void SetPrimaryInteger(Int32 value)
    {
      i1val = value;
    }
    public Int32 GetSecondaryInteger()
    {
      return i2val;
    }
    public void SetSecondaryInteger(Int32 value)
    {
      i2val = value;
    }
    public override String ToString()
    {
      return "" + i1val + " " + f1val + " " + f2val + " " + i2val + " " + GetName();
    }
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct DUNAME
  {
    public const int sizeInBytes = 32;
    public const int format = Formats.CF_UNAME;
    Int32 i1val;
    float f1val;
    float f2val;
    float f3val;
    public fixed byte str[16];
    public DUNAME(String inpt)
    {
      i1val = 0; f1val = 0; f2val = 0; f3val = 0;
      try
      {
        String[] vals = inpt.Split(StrUtils.parserDelimiter, StringSplitOptions.RemoveEmptyEntries);
        i1val = Int32.Parse(vals[0]);
        f1val = Single.Parse(vals[1]);
        f2val = Single.Parse(vals[2]);
        f3val = Single.Parse(vals[3]);
        SetName(vals[4]);
      }
      catch (Exception) { }
    }
    public string GetName()
    {
      fixed (byte* pca = str) return StrUtils.NameToString(pca, 80);
    }
    public void SetName(string str)
    {
      fixed (byte* pca = this.str) StrUtils.StringToName(str, pca, 80);
    }
    public double GetPrimaryFloat()
    {
      return f1val;
    }
    public void SetPrimaryFloat(float value)
    {
      f1val = value;
    }
    public double GetSecondaryFloat()
    {
      return f2val;
    }
    public void SetSecondaryFloat(float value)
    {
      f2val = value;
    }
    public Int32 GetPrimaryInteger()
    {
      return i1val;
    }
    public void SetPrimaryInteger(Int32 value)
    {
      i1val = value;
    }
    public double GetTertiaryFloat()
    {
      return f3val;
    }
    public void SetTertiaryFloat(float value)
    {
      f3val = value;
    }
    public override String ToString()
    {
      return "" + i1val + " " + f1val + " " + f2val + " " + f3val + " " + GetName();
    }
  }
  /// <summary>
  /// Due to the unfixed float array this type is NOT blitable
  /// </summary>
  [StructLayout(LayoutKind.Auto, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct DSPECTRUM
  {
    public const int SPECTRUM_STD_SIZE = 4096;
    public const int headerSizeInBytes = 96;
    private int length;
    private int sizeInBytes;
    public const int format = Formats.CF_SPECTRUM;
    public fixed byte str[80];
    Int32 i1val;
    float f1val;
    float f2val;
    Int32 i2val;
    float[]  d_spect_array;
    public DSPECTRUM(String inpt)
    {
      i1val = 0; f1val = 0; f2val = 0; i2val = 0;
      length = 0; sizeInBytes = headerSizeInBytes;
      d_spect_array = null;
      try
      {
        String[] vals = inpt.Split(StrUtils.parserDelimiter, StringSplitOptions.RemoveEmptyEntries);
        length = vals.Length - 5;
        sizeInBytes = headerSizeInBytes + length * 4;
        d_spect_array = new float[length];
        SetName(vals[0]);
        i1val = Int32.Parse(vals[1]);
        f1val = Single.Parse(vals[2]);
        f2val = Single.Parse(vals[3]);
        i2val = Int32.Parse(vals[4]);
        for (int i = 0; i < length; i++) d_spect_array.SetValue(Single.Parse(vals[5 + i]), i);
      }
      catch (Exception) { }
    }
    public int GetLength()
    {
      return length;
    }
    public string GetName()
    {
      fixed (byte* pca = str) return StrUtils.NameToString(pca, 80);
    }
    public void SetName(string str)
    {
      fixed (byte* pca = this.str) StrUtils.StringToName(str, pca, 80);
    }
    public double GetPrimaryFloat()
    {
      return f1val;
    }
    public void SetPrimaryFloat(float value)
    {
      f1val = value;
    }
    public double GetSecondaryFloat()
    {
      return f2val;
    }
    public void SetSecondaryFloat(float value)
    {
      f2val = value;
    }
    public Int32 GetPrimaryInteger()
    {
      return i1val;
    }
    public void SetPrimaryInteger(Int32 value)
    {
      i1val = value;
    }
    public Int32 GetSecondaryInteger()
    {
      return i2val;
    }
    public void SetSecondaryInteger(Int32 value)
    {
      i2val = value;
    }
    public float[] GetSpectrum()
    {
      return d_spect_array;
    }
    public DSPECTRUM(int length)
    {
      i1val = 0; i2val = 0; f1val = 0; f2val = 0;
      this.length = length;
      sizeInBytes = headerSizeInBytes + length * sizeof(float);
      d_spect_array = new float[length];
    }
    public DSPECTRUM(int tm, float s_start, float s_inc, int status, float[] s_array)
    {
      i1val = tm;
      i2val = status;
      f1val = s_start;
      f2val = s_inc;
      if (s_array == null)
      {
        length = SPECTRUM_STD_SIZE;
        d_spect_array = new float[length];
      }
      else
      {
        length = s_array.Length;
        d_spect_array = s_array;
      }
      sizeInBytes = headerSizeInBytes + length * sizeof(float);
    }
    public override String ToString()
    {
      return "" + i1val + " " + f1val + " " + f2val + " " + i2val + " " + GetName() + "(Data: " + length + " float values)";
    }
    public byte[] ToByteArray()
    {
      int p = 0;
      byte[] ba = new byte[sizeInBytes];
      byte[] bap;
      fixed (byte *bp = str) for (p = 0; p < 80; p++) ba[p] = bp[p];
      bap = BitConverter.GetBytes(i1val); Array.Copy(bap, 0, ba, p, 4); p += 4;
      bap = BitConverter.GetBytes(f1val); Array.Copy(bap, 0, ba, p, 4); p += 4;
      bap = BitConverter.GetBytes(f2val); Array.Copy(bap, 0, ba, p, 4); p += 4;
      bap = BitConverter.GetBytes(i2val); Array.Copy(bap, 0, ba, p, 4); p += 4;
      for (int i = 0; i < length; i++)
      {
        bap = BitConverter.GetBytes(d_spect_array[i]);
        Array.Copy(bap, 0, ba, p, 4); p += 4;
      }
      return ba;
    }
    public void ToStruct(byte[] src)
    {
      int p = 0;
      length = (src.Length - headerSizeInBytes) / sizeof(float);
      fixed (byte* bp = str) for (p = 0; p < 80; p++) bp[p] = src[p];
      i1val = BitConverter.ToInt32(src, p); p += 4;
      f1val = BitConverter.ToSingle(src, p); p += 4;
      f2val = BitConverter.ToSingle(src, p); p += 4;
      i2val = BitConverter.ToInt32(src, p); p += 4;
      for (int i = 0; i < length; i++)
      {
        d_spect_array[i] = BitConverter.ToSingle(src, p); p += 4;
      }
      return;
    }
  }
  //[StructLayout(LayoutKind.Explicit,Pack = 1,Size = 40+sizeof(IntPtr))]
  //public struct DTYPE
  //{
  //  [FieldOffset(0)]    int dArrayLength;
  //  [FieldOffset(4)]    short dFormat;
  //  [FieldOffset(6)]    short xferReason;
  //  [FieldOffset(8)]    int dStamp;
  //  [FieldOffset(12)]    int sysStamp;
  //  [FieldOffset(16)]    double dTimeStamp;
  //  [FieldOffset(24)]    char dTag;
  //  [FieldOffset(40)]    IntPtr data;
  //}
  //[StructLayout(LayoutKind.Sequential, Pack = 1)]
  //public unsafe struct DTYPE
  //{
  //  public int dArrayLength;
  //  public short dFormat;
  //  public short xferReason;
  //  public int dStamp;
  //  public int sysStamp;
  //  public double dTimeStamp;
  //  public fixed char dTag[Constants.TAG_NAME_SIZE];
  //  public IntPtr data;
  //}
  public static class AlarmDescriptors
  {
    public const UInt16 almNEWALARM = 0x01;
    public const UInt16 almHEARTBEAT = 0x02;
    public const UInt16 almOSCILLATION = 0x04;
    public const UInt16 almDATACHANGE = 0x08;
    public const UInt16 almTRANSIENT = 0x10;
    public const UInt16 almDISABLED = 0x20;
    public const UInt16 almTERMINATE = 0x40;
    public const UInt16 almTEST = 0x80;
    public const UInt16 almSUPPRESS = 0x80;
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct DTYPE
  {
    public UInt32 dArrayLength;
    public short dFormat;
    public short xferReason;
    public int dStamp;
    public int sysStamp;
    public double dTimeStamp;
    //public fixed byte dTag[Constants.TAG_NAME_SIZE];
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = Constants.TAG_NAME_SIZE)]
      public char[] dTag;
    public IntPtr data;
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct ALM_THRESHOLDS
  {
    public Int32 cnt;       // number successive threshold exceeded before applying alarm */
    public Int32 hisev;     // high severity */
    public Int32 hiwarnsev; // high warning severity */
    public Int32 losev;     // low severity */
    public Int32 lowarnsev; // low warning severity */
    public float hi;      // high threshold */
    public float hiwarn;  // high threshold warning (approaching high threshold) */
    public float lo;      // low threshold */
    public float lowarn;  // low threshold warning (approaching low threshold) */
    public UInt32 mask;   // if non-zero -> apply mask and check vs. normal value */
    public UInt32 normal; // normal valid data value */
    public Int32 normalIsAlarm; // apply a NOT to the 'normal value' */
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct ADS    /* Alarm Definition Structure */
  {
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = Constants.ALARM_TAG_SIZE)]
      public char[] alarmTag;
    public UInt32 alarmCode;                // alarm code ID */
    public UInt32 alarmMask;                // alarm mask */
    public UInt16 alarmSystem;              // alarm system ID (0 -> use registered default) */
    public Int16 alarmSeverity;            // alarm severity (0 - 15) */
    public byte alarmDataFormat;            // alarm data format (simple code) */
    public byte alarmDataArraySize;         // alarm data array length */
    public byte alarmOscillationWindow;     // alarm oscillation window (in clear counts) */
    public byte alarmOscillationWindowPinned;// binary: 0 = not pinned */
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = Constants.ALARM_TEXT_SIZE)]
      public char[] alarmText; // long text description for alarm tag */
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = Constants.ALARM_TEXT_SIZE)]
      public char[] deviceText;// text description for device type */
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = Constants.ALARM_TEXT_SIZE)]
      public char[] dataText;  // text description for alarm data contents */
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = Constants.ALARM_TEXT_LONGSIZE)]
      public char[] url;   // url link to further information */
    public void initialize()
    {
      if (alarmTag == null) alarmTag = new char[Constants.ALARM_TAG_SIZE];
      if (alarmText == null) alarmText = new char[Constants.ALARM_TEXT_SIZE];
      if (deviceText == null) deviceText = new char[Constants.ALARM_TEXT_SIZE];
      if (dataText == null) dataText = new char[Constants.ALARM_TEXT_SIZE];
      if (url == null) url = new char[Constants.ALARM_TEXT_LONGSIZE];
    }
    public String GetAlarmTag()
    {
      return new String(alarmTag).Trim();
    }
    public void SetAlarmTag(String tag)
    {
      initialize();
      Array.Clear(alarmTag, 0, Constants.ALARM_TAG_SIZE);
      if (tag == null) tag = "";
      if (tag.Length > Constants.ALARM_TAG_SIZE) tag = tag.Substring(0, Constants.ALARM_TAG_SIZE);
      tag.ToCharArray().CopyTo(alarmTag, 0);
    }
    public String GetAlarmText()
    {
      return new String(alarmText).Trim();
    }
    public void SetAlarmText(String text)
    {
      initialize();
      Array.Clear(alarmText, 0, Constants.ALARM_TEXT_SIZE);
      if (text == null) text = "";
      if (text.Length > Constants.ALARM_TEXT_SIZE) text = text.Substring(0, Constants.ALARM_TEXT_SIZE);
      text.ToCharArray().CopyTo(alarmText, 0);
    }
    public String GetDeviceText()
    {
      return new String(deviceText).Trim();
    }
    public void SetDeviceText(String text)
    {
      initialize();
      Array.Clear(deviceText, 0, Constants.ALARM_TEXT_SIZE);
      if (text == null) text = "";
      if (text.Length > Constants.ALARM_TEXT_SIZE) text = text.Substring(0, Constants.ALARM_TEXT_SIZE);
      text.ToCharArray().CopyTo(deviceText, 0);
    }
    public String GetDataText()
    {
      return new String(dataText).Trim();
    }
    public void SetDataText(String text)
    {
      initialize();
      Array.Clear(dataText, 0, Constants.ALARM_TEXT_SIZE);
      if (text == null) text = "";
      if (text.Length > Constants.ALARM_TEXT_SIZE) text = text.Substring(0, Constants.ALARM_TEXT_SIZE);
      text.ToCharArray().CopyTo(dataText, 0);
    }
    public String GetUrl()
    {
      return new String(url).Trim();
    }
    public void SetUrl(String url)
    {
      initialize();
      Array.Clear(this.url, 0, Constants.ALARM_TEXT_LONGSIZE);
      if (url == null) url = "";
      if (url.Length > Constants.ALARM_TEXT_LONGSIZE) url = url.Substring(0, Constants.ALARM_TEXT_LONGSIZE);
      url.ToCharArray().CopyTo(this.url, 0);
    }
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct HistorySpecification
  {
    public Int32 pollingRate;      // polling rate in msec */
    public Int32 archiveRate;      // archive rate in sec */
    public Int32 depthShort;       // for short term storage */
    public Int32 depthLong;        // for long term storage */
    public Int32 heartbeat;        // archive heartbeat in sec */
    public float pTolerance;      // percent tolerance */
    public float aTolerance;      // absolute tolerance */
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = Constants.EXPORT_NAME_SIZE)]
    public char[] rhsServerName;   // Remote Server Name */
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = Constants.PROPERTY_NAME_SIZE)]
    public char[] rhsPropertyName; // Remote Property Name */
    public void initialize()
    {
      if (rhsServerName == null) rhsServerName = new char[Constants.EXPORT_NAME_SIZE];
      if (rhsPropertyName == null) rhsPropertyName = new char[Constants.PROPERTY_NAME_SIZE];
    }
    public void SetRemoteHistoryServerName(string ServerName)
    {
      initialize();
      if (ServerName == null) ServerName = "";
      if (ServerName.Length > Constants.EXPORT_NAME_SIZE) ServerName = ServerName.Substring(0, Constants.EXPORT_NAME_SIZE);
      ServerName.ToCharArray().CopyTo(rhsServerName, 0);
    }
    public void SetRemoteHistoryPropertyName(string PropertyName)
    {
      initialize();
      if (PropertyName == null) PropertyName = "";
      if (PropertyName.Length > Constants.PROPERTY_NAME_SIZE) PropertyName = PropertyName.Substring(0, Constants.PROPERTY_NAME_SIZE);
      PropertyName.ToCharArray().CopyTo(rhsPropertyName, 0);
    }
  } 
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct sockaddr_in
  {
    public Int16 sin_family;
    public UInt16 sin_port;
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = 4)]
      public byte[] sin_addr;
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = 8)]
     public char[] sin_zero;
    public void initialize()
    {
      if (sin_addr == null) sin_addr = new byte[4];
      if (sin_zero == null) sin_zero = new char[8];
    }
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct ClnInfo
  {
    public UInt16 accessMode; // this client's requested access mode: CM_SINGLE, CM_DATACHANGE, CM_TIMER, etc. */
    public UInt16 reserved1;
    public Int32 pollingRate;// this client's requested polling rate */
    public UInt32 starttime;  // this client's start of initial subscription */
    public UInt16 counter;    // this client's counts remaining in subscription(s) */
    public UInt16 reserved2;
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = Constants.USERNAME_SIZE)]
      public char[] userName; // this client's user name */
    public sockaddr_in addr; // this client's ip address */
    public short ncontracts;  // this client's total number of contracts */
    public short tineProtocol;// this client's tine protocol level */
    public short inetProtocol;// this client's transport protocol: UDP, TCP, STREAM , etc.  */
    public UInt16 reserved3;
    public void initialize()
    {
      if (userName == null) userName = new char[Constants.USERNAME_SIZE];
      addr.initialize();
    }
    public String GetUserName()
    {
      if (userName == null) return null;
      int len = 0;
      while (len < Constants.USERNAME_SIZE && userName[len] != 0) len++;
      return new String(userName,0,len);
    }
    public String GetNetworkAddr()
    {
      if (addr.sin_addr == null) return null;
      ushort p = addr.sin_port;
      p = (ushort)(((p & 0xffff) << 8) + (p >> 8));
      String ip = "" + addr.sin_addr[0] + "." + addr.sin_addr[1] + "." + addr.sin_addr[2] + "." + addr.sin_addr[3] + ":"+p;
      return ip;
    }
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct PrpQueryStruct
  {
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = Constants.PROPERTY_NAME_SIZE)]
      public char[] prpName; // the property name */
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = Constants.PROPERTY_DESC_SIZE)]
      public char[] prpDescription; // the property description */
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = Constants.PROPERTY_REDIR_SIZE)]
      public char[] prpRedirection; // a redirection string if the property is redirected to another server */
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = Constants.TAG_NAME_SIZE)]
      public char[] prpTag;   // the data tag for output data sets (CF_STRUCT, CF_BITFIELD, CF_HISTORY, etc.) */
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = Constants.TAG_NAME_SIZE)]
      public char[] prpTagIn; // the data tag for input data sets (CF_STRUCT, CF_BITFIELD, CF_HISTORY, etc.) */
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = Constants.UNITS_SIZE)]
      public char[] prpUnits;   // a string containing the natural units of the data returned by the property */
    public float prpMinValue;    // the lower limit of valid data associated with the property */
    public float prpMaxValue;    // the upper limit of valid data associated with the property */
    public UInt32 prpSize;       // the maximum allowed size of an output data set associated with the property */
    public UInt32 prpSizeIn;     // the maximum allowed size of an input data set associated with the property */
    public UInt32 prpNumOverloads; // the number of overloads associated with this property name  */
    public UInt16 prpHistoryDepthShort; // the depth of the short term history kept for this property */
    public UInt16 prpHistoryDepthLong; // the depth of the long term history kept for this property */
    public byte prpFormat;       // the default output data format accepted for this property */
    public byte prpFormatIn;     // is the default input data format accepted for this property */
    public byte prpAccess;       // the allowed data access for this property (either CA_READ, CA_WRITE or both) */
    public byte prpGraphType;    // is the preferred graph type for output data associated with this property, e.g. GT_LIN, GT_LOG, etc */
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = Constants.UNITS_SIZE)]
       public char[] rngUnits;   // a string containing the natural units of the x-axis (range) in case of a spectrum type array */
    public float rngMinValue;    // the lower limit of x-axis range associated with the property  */
    public float rngMaxValue;    // the upper limit of x-axis range associated with the property  */
    public UInt16 numRows;       // the number of rows in an array property */
    public UInt16 rowSize;       // the size of a row in an array property */
    public UInt16 prpArrayType;  // the property array type (AT_NONE, AT_SCALAR, AT_CHANNEL, AT_SPECTRUM, etc.) */
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = 3)]
      public UInt16[] reserved;   // reserved field) */
    public static string getStructTag()
    {
      return "PRPQSr4";
    }
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct ExpDataStruct
  {
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = Constants.FEC_NAME_SIZE)]
    public char[] FecName;         // FEC Name */
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = Constants.SUBSYSTEM_NAME_SIZE)]
    public char[] SubSystem; // sub-system name */
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = Constants.EXPORT_NAME_SIZE)]
    public char[] ExportName;   // Export Name */
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = Constants.EQM_NAME_SIZE)]
    public char[] EqmName;         // Equipment name */
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = Constants.CONTEXT_NAME_SIZE)]
    public char[] EqmContext;  // EqmContext */
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct FecDataStruct
  {
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = Constants.FEC_NAME_SIZE)]
    public char[] FecName;      // FEC Name */
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = 4)]
    public byte[] IPXNetwork;  // IPX addr info (or internal) */
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = 6)]
    public byte[] IPXNode;     // IPX addr info (or internal)  */
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = 6)]
    public byte[] IPXImmediateAddress; // IPX addr info (or internal)  */
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = Constants.ADDR_SIZE)]
    public char[] IP;          // IP address as string */
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = 4)]
    public byte[] IPh_addr;    // IP haddr (byte representation) */
    public Int32 portOffset;   // port offset (applied to all listening ports) */
    public Int32 inetProtocol; // inet protocol (UDP, TCP, IPX, MMF, PIPE, etc.) */
    public Int32 tineProtocol; // tine protocol level (highest level accepted) */
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct FecAddrStruct
  {
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = Constants.FEC_NAME_SIZE)]
    public char[] fecName;      // FEC Name */
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = 16)]
    public byte[] netAdr;  // IPX addr info (or internal) */
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = 8)]
    public byte[] lclAdr;     // IPX addr info (or internal)  */
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = 40)]
    public char[] strAdr;          // IP address as string */
    public Int32 portOffset;   // port offset (applied to all listening ports) */
    public Int32 inetProtocol; // inet protocol (UDP, TCP, IPX, MMF, PIPE, etc.) */
    public Int32 tineProtocol; // tine protocol level (highest level accepted) */
    public Int32 reserved; // tine protocol level (highest level accepted) */
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct ConTblInfo
  {
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = Constants.EXPORT_NAME_SIZE)]
    public char[] expName; /* eqm function name */
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = Constants.PROPERTY_NAME_SIZE)]
    public char[] prpName; /* eqm property */
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = Constants.DEVICE_NAME_SIZE)]
    public char[] devName; /* eqm device name */
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = Constants.CONTEXT_NAME_SIZE)]
    public char[] context; /* eqm context */
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = Constants.EQM_NAME_SIZE)]
    public char[] eqm;
    public UInt16 mode;          /* used by subscriptions */
    public short access;         /* data access */
    public short formatIn;       /* eqp data format down */
    public short formatOut;      /* eqp data format up */
    public UInt32 sizeIn;        /* data array size down */
    public UInt32 sizeOut;       /* data array size up */
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = Constants.TAG_NAME_SIZE)]
    public char[] strTagIn; /* eqm structure tag down */
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = Constants.TAG_NAME_SIZE)]
    public char[] strTagOut; /* eqm structure tag up */
    public UInt32 pollingInterval; /* polling rate for connectioned clients */
    public UInt32 heartbeat;     /* used by subscriptions */
    public UInt32 starttime;     /* entry into table (used by subscriptions) */
    public UInt32 dtimestamp;    /* arriving TINE timestamp */
    public int dtimestampMSEC;   /* arriving TINE timestamp (msec fraction) */
    public int  counter;         /* current subscription counter */
    public short linkStatus;     /* current link status */
    public short reserved;
    public void initialize()
    {
      expName = new char[Constants.EXPORT_NAME_SIZE];
      prpName = new char[Constants.PROPERTY_NAME_SIZE];
      devName = new char[Constants.DEVICE_NAME_SIZE];
      context = new char[Constants.CONTEXT_NAME_SIZE];
      eqm = new char[Constants.EQM_NAME_SIZE];
      strTagIn = new char[Constants.TAG_NAME_SIZE];
      strTagOut = new char[Constants.TAG_NAME_SIZE];
    }
  }
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct MsgType
  {
    public double ts;
    public int status;
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = 128)]
    public char[] msg;
  }
  class MemMgr
  {
    public object mmRef; // managed memory reference to object
    public unsafe IntPtr umRef; // unmanaged pointer reference;
    public MemMgr(object objRef, int dataSize) 
    {
      mmRef = objRef;
      umRef = Marshal.AllocHGlobal(dataSize);
    }
    public bool Equals(MemMgr mm) { return mm == null ? false : mm.mmRef == this.mmRef; }
  }
  /// <summary>
  /// TINE Data Type object: All data transfered client between and server must be contained in a TDataType object.
  /// </summary>
  public class TDataType
  {
    private static List<MemMgr> memLst = new List<MemMgr>();
    public static IntPtr GetIntPtrFromManagedDataObject(object data,int dataSize)
    {
      MemMgr mm = memLst.Find(delegate(MemMgr m) { return m.mmRef == data; });
      if (mm == null)
      {
        mm = new MemMgr(data, dataSize);
        memLst.Add(mm);
      }
      return mm.umRef;
    }
    /// <summary>
    /// gets the TINE data format code for the object passed
    /// </summary>
    /// <param name="data">is the data object whose TINE data format is desired.</param>
    /// <returns>the TINE data format code assigned to the object specified</returns>
    public static byte GetTineFormat(Object data)
    {
      return GetTineType(data);
    }
    /// <summary>
    /// gets the TINE data format code for the Type passed
    /// </summary>
    /// <param name="type">is the .NET Type whose TINE data format is desired.</param>
    /// <returns>the TINE data format code assigned to the object specified</returns>
    public static byte GetTineFormat(Type type)
    {
      switch (Type.GetTypeCode(type))
      {
        case TypeCode.Boolean: return Formats.CF_BOOLEAN;
        case TypeCode.Byte: return Formats.CF_BYTE;
        case TypeCode.Char: return Formats.CF_CHAR;
        case TypeCode.Double: return Formats.CF_DOUBLE;
        case TypeCode.Empty: return Formats.CF_NULL;
        case TypeCode.Int16: return Formats.CF_INT16;
        case TypeCode.Int32: return Formats.CF_INT32;
        case TypeCode.Int64: return Formats.CF_INT64;
        case TypeCode.Single: return Formats.CF_FLOAT;
        case TypeCode.String: return Formats.CF_STRING;
        case TypeCode.UInt16: return Formats.CF_INT16;
        case TypeCode.UInt32: return Formats.CF_INT32;
        case TypeCode.UInt64: return Formats.CF_INT64;
        default: break;
      }
      return Formats.CF_NULL;
    }
    private static bool isSystemStruct(string cname)
    {
      if (cname.CompareTo("PrpQueryStruct") == 0) return true;
      return false;
    }
    private static string getSystemStructName(string cname)
    {
      if (native.GetStructSize(cname) > 0) return cname;
      // check system struct registry here :
      if (cname.CompareTo("PrpQueryStruct") == 0) return PrpQueryStruct.getStructTag();
      return null;
    }
    private static string getStructName(Object data)
    {
      if (data == (Object)null) return null;
      string cname;
      if (data is TTaggedStruct) cname = ((TTaggedStruct)data).GetTag();
      else if (data is TTaggedStruct[]) cname = ((TTaggedStruct[])data)[0].GetTag();
      else if (data is Array) cname = ((Array)data).GetType().GetElementType().Name;
      else cname = data.GetType().Name;
      return getSystemStructName(cname);
    }
    private static byte GetTineType(Type dtype)
    {
      switch (Type.GetTypeCode(dtype))
      {
        case TypeCode.Boolean: return Formats.CF_BOOLEAN;
        case TypeCode.Byte: return Formats.CF_BYTE;
        case TypeCode.SByte: return Formats.CF_CHAR;
        //case TypeCode.Char: return Formats.CF_CHAR;
        case TypeCode.Double: return Formats.CF_DOUBLE;
        case TypeCode.Empty: return Formats.CF_NULL;
        case TypeCode.Int16: return Formats.CF_INT16;
        case TypeCode.Int32: return Formats.CF_INT32;
        case TypeCode.Int64: return Formats.CF_INT64;
        case TypeCode.Single: return Formats.CF_FLOAT;
        case TypeCode.String: return Formats.CF_STRING;
        case TypeCode.UInt16: return Formats.CF_INT16;
        case TypeCode.UInt32: return Formats.CF_INT32;
        case TypeCode.UInt64: return Formats.CF_INT64;
        case TypeCode.Object:
          String nam = dtype.GetTypeInfo().FullName;
          if (nam.CompareTo("tine.NAME8") == 0) return Formats.CF_NAME8;
          if (nam.CompareTo("tine.NAME16") == 0) return Formats.CF_NAME16;
          if (nam.CompareTo("tine.NAME32") == 0) return Formats.CF_NAME32;
          if (nam.CompareTo("tine.NAME48") == 0) return Formats.CF_NAME48;
          if (nam.CompareTo("tine.NAME64") == 0) return Formats.CF_NAME64;
          if (nam.CompareTo("tine.FLTINT") == 0) return Formats.CF_FLTINT;
          if (nam.CompareTo("tine.INTINT") == 0) return Formats.CF_INTINT;
          if (nam.CompareTo("tine.FLTFLT") == 0) return Formats.CF_FLTFLT;
          if (nam.CompareTo("tine.DBLDBL") == 0) return Formats.CF_DBLDBL;
          if (nam.CompareTo("tine.NAME8I") == 0) return Formats.CF_NAME8I;
          if (nam.CompareTo("tine.NAME16I") == 0) return Formats.CF_NAME16I;
          if (nam.CompareTo("tine.NAME32I") == 0) return Formats.CF_NAME32I;
          if (nam.CompareTo("tine.NAME48I") == 0) return Formats.CF_NAME48I;
          if (nam.CompareTo("tine.NAME64I") == 0) return Formats.CF_NAME64I;
          if (nam.CompareTo("tine.NAME64DBL") == 0) return Formats.CF_NAME64DBL;
          if (nam.CompareTo("tine.NAME16FI") == 0) return Formats.CF_NAME16FI;
          if (nam.CompareTo("tine.NAME16II") == 0) return Formats.CF_NAME16INTINT;
          if (nam.CompareTo("tine.NAME16DBLDBL") == 0) return Formats.CF_NAME16DBLDBL;
          if (nam.CompareTo("tine.NAME32DBLDBL") == 0) return Formats.CF_NAME32DBLDBL;
          if (nam.CompareTo("tine.NAME64DBLDBL") == 0) return Formats.CF_NAME64DBLDBL;
          if (nam.CompareTo("tine.INTFLTINT") == 0) return Formats.CF_INTFLTINT;
          if (nam.CompareTo("tine.FLTFLTINT") == 0) return Formats.CF_FLTFLTINT;
          if (nam.CompareTo("tine.FLTINTINT") == 0) return Formats.CF_FLTINTINT;
          if (nam.CompareTo("tine.INTINTINT") == 0) return Formats.CF_INTINTINT;
          if (nam.CompareTo("tine.DBLDBLDBL") == 0) return Formats.CF_DBLDBLDBL;
          if (nam.CompareTo("tine.INTINTINTINT") == 0) return Formats.CF_INTINTINTINT;
          if (nam.CompareTo("tine.INTINTFLTFLT") == 0) return Formats.CF_INTINTFLTFLT;
          if (nam.CompareTo("tine.INTFLTFLTFLT") == 0) return Formats.CF_INTFLTFLTFLT;
          if (nam.CompareTo("tine.FLTINTFLTINT") == 0) return Formats.CF_FLTINTFLTINT;
          if (nam.CompareTo("tine.DWINDOW") == 0) return Formats.CF_WINDOW;
          if (nam.CompareTo("tine.NAME64DBLDBLDBL") == 0) return Formats.CF_NAME64DBLDBLDBL;
          if (nam.CompareTo("tine.DUSTRING") == 0) return Formats.CF_USTRING;
          if (nam.CompareTo("tine.DUNAME") == 0) return Formats.CF_UNAME;
          if (nam.CompareTo("tine.DSPECTRUM") == 0) return Formats.CF_SPECTRUM;
          if (nam.CompareTo("tine.TTaggedStruct") == 0) return Formats.CF_STRUCT;
          if (nam.CompareTo("tine.DIMAGE") == 0) return Formats.CF_IMAGE;
          if (nam.CompareTo("tine.TImageHeader") == 0) return Formats.CF_IMAGE;
          break;
        default: break;
      }
      return Formats.CF_NULL;
    }
    private static byte GetTineType(Object data)
    {
      if (data == (Object)null) return Formats.CF_NULL;
      if (getStructName(data) != null) return Formats.CF_STRUCT;

      Type dtype;
      if (data is Array)
      {
        dtype = data.GetType().GetElementType();
      }
      else
      {
        dtype = data.GetType();
      }
      return GetTineType(dtype);
    }
    public static object MakeDataObject(Type typ, String inpt)
    {
      if (inpt == null) return null;
      String[] sa = inpt.Split(Environment.NewLine.ToCharArray());
      byte fmt = GetTineType(typ);
      int siz = sa.Length;
      return MakeDataObject(fmt, siz, null);
    }

    public static object MakeDataObject(ushort fmt, int siz)
    {
      return MakeDataObject(fmt, siz, null);
    }
    public static object MakeDataObject(ushort fmt, int siz, String tag)
    {
      if (siz <= 0) return null;
      switch (fmt)
      {
        case Formats.CF_BYTE: return new byte[siz];
        case Formats.CF_INT16: return new Int16[siz];
        case Formats.CF_INT32: return new Int32[siz];
        case Formats.CF_INT64: return new Int64[siz];
        case Formats.CF_FLOAT: return new float[siz];
        case Formats.CF_DOUBLE: return new double[siz];
        case Formats.CF_BOOLEAN: return new Int32[siz];
        case Formats.CF_CHAR: return new sbyte[siz];
        case Formats.CF_NAME8: return new NAME8[siz];
        case Formats.CF_NAME16: return new NAME16[siz];
        case Formats.CF_NAME32: return new NAME32[siz];
        case Formats.CF_NAME48: return new NAME48[siz];
        case Formats.CF_NAME64: return new NAME64[siz];
        case Formats.CF_NAME8I: return new NAME8I[siz];
        case Formats.CF_NAME16I: return new NAME16I[siz];
        case Formats.CF_NAME32I: return new NAME32I[siz];
        case Formats.CF_NAME48I: return new NAME48I[siz];
        case Formats.CF_NAME64I: return new NAME64I[siz];
        case Formats.CF_NAME64DBL: return new NAME64DBL[siz];
        case Formats.CF_INTINT: return new INTINT[siz];
        case Formats.CF_FLTINT: return new FLTINT[siz];
        case Formats.CF_FLTFLT: return new FLTFLT[siz];
        case Formats.CF_DBLDBL: return new DBLDBL[siz];
        case Formats.CF_NAME16FI: return new NAME16FI[siz];
        case Formats.CF_NAME16II: return new NAME16II[siz];
        case Formats.CF_NAME16DBLDBL: return new NAME16DBLDBL[siz];
        case Formats.CF_NAME32DBLDBL: return new NAME32DBLDBL[siz];
        case Formats.CF_NAME64DBLDBL: return new NAME64DBLDBL[siz];
        case Formats.CF_DBLDBLDBL: return new DBLDBLDBL[siz];
        case Formats.CF_XML: return new sbyte[siz];
        case Formats.CF_STRING: return new String[siz];
        case Formats.CF_BITFIELD8: return new byte[siz];
        case Formats.CF_BITFIELD16: return new UInt16[siz];
        case Formats.CF_BITFIELD32: return new UInt32[siz];
        case Formats.CF_BITFIELD64: return new UInt64[siz];
        case Formats.CF_USTRING: return new DUSTRING[siz];
        case Formats.CF_UNAME: return new DUNAME[siz];
        case Formats.CF_FLTFLTINT: return new FLTFLTINT[siz];
        case Formats.CF_FLTINTINT: return new FLTINTINT[siz];
        case Formats.CF_INTINTINT: return new INTINTINT[siz];
        case Formats.CF_INTFLTFLTFLT: return new INTFLTFLTFLT[siz];
        case Formats.CF_SPECTRUM: return new DSPECTRUM(siz);
        case Formats.CF_IMAGE:
          { // this method is used via Acop
            DIMAGE[] img = new DIMAGE[] { new DIMAGE(siz) };
            return img;
          }
        case Formats.CF_INTINTINTINT: return new INTINTINTINT[siz];
        case Formats.CF_STRUCT:
          if (native.GetStructSize(tag) > 0)
          { // structure is registered
            //Assembly a = typeof(TTaggedStruct).Assembly;
            //Type t = a.GetType(tag);
            //object o = Activator.CreateInstance(t);
            return new TTaggedStruct[] { new TTaggedStruct(tag,siz) };
          }
          break;
      }
      return null;
    }
    public static double[] GetXAxis(Object data)
    {
      if (!(data is Array)) return null;
      return GetXAxis(data, ((Array)data).GetLength(0));
    }
    public static double[] GetXAxis(Object data, int len)
    {
      if (!(data is Array)) return null;
      int olen = ((Array)data).GetLength(0);
      if (len < 1 || len > olen) len = olen;
      if (data is double[])
      {
        double[] da = (double[])data;
        if (da.Length == len) return da;
        double[] nda = new double[len];
        Array.Copy(da, nda, len);
        return nda;
      }
      double[] x = new double[len];
      if (data is float[]) for (int i = 0; i < len; i++) x[i] = (float[])data == null ? 0 : ((float[])data)[i];
      else if (data is Int16[]) for (int i = 0; i < len; i++) x[i] = ((Int16[])data)[i];
      else if (data is Int32[]) for (int i = 0; i < len; i++) x[i] = ((Int32[])data)[i];
      else if (data is Int64[]) for (int i = 0; i < len; i++) x[i] = (double)((Int64[])data)[i];
      else if (data is INTINT[]) for (int i = 0; i < len; i++) x[i] = (INTINT[])data == null ? 0 : ((INTINT[])data)[i].i2val;
      else if (data is FLTINT[]) for (int i = 0; i < len; i++) x[i] = (FLTINT[])data == null ? 0 : ((FLTINT[])data)[i].ival;
      else if (data is FLTFLT[]) for (int i = 0; i < len; i++) x[i] = (FLTFLT[])data == null ? 0 : ((FLTFLT[])data)[i].f2val;
      else if (data is DBLDBL[]) for (int i = 0; i < len; i++) x[i] = (DBLDBL[])data == null ? 0 : ((DBLDBL[])data)[i].d2val;
      else if (data is INTFLTINT[]) for (int i = 0; i < len; i++) x[i] = (INTFLTINT[])data == null ? 0 : ((INTFLTINT[])data)[i].i1val;
      else if (data is FLTFLTINT[]) for (int i = 0; i < len; i++) x[i] = (FLTFLTINT[])data == null ? 0 : ((FLTFLTINT[])data)[i].f2val;
      else if (data is FLTINTINT[]) for (int i = 0; i < len; i++) x[i] = (FLTINTINT[])data == null ? 0 : ((FLTINTINT[])data)[i].i1val;
      else if (data is INTINTINT[]) for (int i = 0; i < len; i++) x[i] = (INTINTINT[])data == null ? 0 : ((INTINTINT[])data)[i].i2val;
      else if (data is INTINTINTINT[]) for (int i = 0; i < len; i++) x[i] = (INTINTINTINT[])data == null ? 0 : ((INTINTINTINT[])data)[i].i2val;
      else if (data is NAME8I[]) for (int i = 0; i < len; i++) x[i] = (NAME8I[])data == null ? 0 : ((NAME8I[])data)[i].ival;
      else if (data is NAME16I[]) for (int i = 0; i < len; i++) x[i] = (NAME16I[])data == null ? 0 : ((NAME16I[])data)[i].ival;
      else if (data is NAME32I[]) for (int i = 0; i < len; i++) x[i] = (NAME32I[])data == null ? 0 : ((NAME32I[])data)[i].ival;
      else if (data is NAME48I[]) for (int i = 0; i < len; i++) x[i] = (NAME48I[])data == null ? 0 : ((NAME48I[])data)[i].ival;
      else if (data is NAME64I[]) for (int i = 0; i < len; i++) x[i] = (NAME64I[])data == null ? 0 : ((NAME64I[])data)[i].ival;
      else if (data is NAME16II[]) for (int i = 0; i < len; i++) x[i] = (NAME16II[])data == null ? 0 : ((NAME16II[])data)[i].i2val;
      else if (data is NAME16FI[]) for (int i = 0; i < len; i++) x[i] = (NAME16FI[])data == null ? 0 : ((NAME16FI[])data)[i].ival;
      else if (data is NAME16DBLDBL[]) for (int i = 0; i < len; i++) x[i] = (NAME16DBLDBL[])data == null ? 0 : ((NAME16DBLDBL[])data)[i].d2val;
      else if (data is NAME32DBLDBL[]) for (int i = 0; i < len; i++) x[i] = (NAME32DBLDBL[])data == null ? 0 : ((NAME32DBLDBL[])data)[i].d2val;
      else if (data is NAME64DBLDBL[]) for (int i = 0; i < len; i++) x[i] = (NAME64DBLDBL[])data == null ? 0 : ((NAME64DBLDBL[])data)[i].d2val;
      else if (data is DUSTRING[]) for (int i = 0; i < len; i++) x[i] = (DUSTRING[])data == null ? 0 : ((DUSTRING[])data)[i].GetSecondaryFloat();
      return x;
    }
    public static double[] GetYAxis(Object data)
    {
      if (!(data is Array)) return null;
      return GetYAxis(data, ((Array)data).GetLength(0));
    }
    public static double[] GetYAxis(Object data, int len)
    {
      if (data == null) return null;
      double[] x;
      if (data is DSPECTRUM)
      {
        x = new double[len];
        float[] spec = ((DSPECTRUM)data).GetSpectrum();
        for (int i = 0; i < len && i < spec.Length; i++) x[i] = (double)spec[i];
        data = x;
      }
      if (!(data is Array)) return null;
      int olen = ((Array)data).GetLength(0);
      if (len < 1 || len > olen) len = olen;
      if (data is double[])
      {
        x = (double[])data;
        if (x.Length == len) return x;
        double[] nda = new double[len];
        Array.Copy(x, nda, len);
        return nda;
      }
      x = new double[len];
      if (data is float[]) for (int i = 0; i < len; i++) x[i] = (float[])data == null ? 0 : ((float[])data)[i];
      else if (data is Int16[]) for (int i = 0; i < len; i++) x[i] = ((Int16[])data)[i];
      else if (data is Int32[]) for (int i = 0; i < len; i++) x[i] = ((Int32[])data)[i];
      else if (data is Int64[]) for (int i = 0; i < len; i++) x[i] = (double)((Int64[])data)[i];
      else if (data is INTINT[]) for (int i = 0; i < len; i++) x[i] = (INTINT[])data == null ? 0 : ((INTINT[])data)[i].i1val;
      else if (data is FLTINT[]) for (int i = 0; i < len; i++) x[i] = (FLTINT[])data == null ? 0 : ((FLTINT[])data)[i].fval;
      else if (data is FLTFLT[]) for (int i = 0; i < len; i++) x[i] = (FLTFLT[])data == null ? 0 : ((FLTFLT[])data)[i].f1val;
      else if (data is DBLDBL[]) for (int i = 0; i < len; i++) x[i] = (DBLDBL[])data == null ? 0 : ((DBLDBL[])data)[i].d1val;
      else if (data is INTFLTINT[]) for (int i = 0; i < len; i++) x[i] = (INTFLTINT[])data == null ? 0 : ((INTFLTINT[])data)[i].fval;
      else if (data is FLTFLTINT[]) for (int i = 0; i < len; i++) x[i] = (FLTFLTINT[])data == null ? 0 : ((FLTFLTINT[])data)[i].f1val;
      else if (data is FLTINTINT[]) for (int i = 0; i < len; i++) x[i] = (FLTINTINT[])data == null ? 0 : ((FLTINTINT[])data)[i].fval;
      else if (data is INTINTINT[]) for (int i = 0; i < len; i++) x[i] = (INTINTINT[])data == null ? 0 : ((INTINTINT[])data)[i].i1val;
      else if (data is INTINTINTINT[]) for (int i = 0; i < len; i++) x[i] = (INTINTINTINT[])data == null ? 0 : ((INTINTINTINT[])data)[i].i1val;
      else if (data is NAME16II[]) for (int i = 0; i < len; i++) x[i] = (NAME16II[])data == null ? 0 : ((NAME16II[])data)[i].i1val;
      else if (data is NAME16FI[]) for (int i = 0; i < len; i++) x[i] = (NAME16FI[])data == null ? 0 : ((NAME16FI[])data)[i].fval;
      else if (data is NAME16DBLDBL[]) for (int i = 0; i < len; i++) x[i] = (NAME16DBLDBL[])data == null ? 0 : ((NAME16DBLDBL[])data)[i].d1val;
      else if (data is NAME32DBLDBL[]) for (int i = 0; i < len; i++) x[i] = (NAME32DBLDBL[])data == null ? 0 : ((NAME32DBLDBL[])data)[i].d1val;
      else if (data is NAME64DBLDBL[]) for (int i = 0; i < len; i++) x[i] = (NAME64DBLDBL[])data == null ? 0 : ((NAME64DBLDBL[])data)[i].d1val;
      else if (data is DUSTRING[]) for (int i = 0; i < len; i++) x[i] = (DUSTRING[])data == null ? 0 : ((DUSTRING[])data)[i].GetPrimaryFloat();
      return x;
    }
    public static String[] GetChannelNames(Object data)
    {
      if (!(data is Array)) return null;
      return GetChannelNames(data, ((Array)data).GetLength(0));
    }
    public static String[] GetChannelNames(Object data, int len)
    {
      if (GetTineType(data) == Formats.CF_STRUCT)
      { // special case for structs: return a list of the fields
        String tag = getStructName(data);
        int ssiz = native.GetStructSize(tag);
        int cc = Errors.name_not_found;
        if (ssiz > 0)
        {
          char[] stringbuf = new char[ssiz * 32];
          String rs = null;
          String[] rsa = null;
          if ((cc = native.GetStructureAsString(tag, TDataType.GetTineDataPtr(data), stringbuf, stringbuf.Length)) == 0)
          {
            rs = new String(stringbuf).Trim(new char[] { '\0' });
            rsa = rs.Split(new char[] { '\n' }, StringSplitOptions.RemoveEmptyEntries);
            String[] rssa = new String[rsa.Length - 2];
            for (int i = 0; i < rssa.Length; i++)
            {
              rssa[i] = rsa[i + 1].Substring(0, rsa[i + 1].IndexOf(':')).Trim();
            }
            rsa = rssa;
            return rsa;
          }
        }
      }
      if (!(data is Array)) return null;
      int olen = ((Array)data).GetLength(0);
      if (len < 1 || len > olen) len = olen;
      if (data is String[]) return (String[])data;
      String[] n = new String[len];
      if (data is NAME8[]) for (int i = 0; i < len; i++) n[i] = (NAME8[])data == null ? "" : ((NAME8[])data)[i].GetName();
      else if (data is NAME16[]) for (int i = 0; i < len; i++) n[i] = (NAME16[])data == null ? "" : ((NAME16[])data)[i].GetName();
      else if (data is NAME32[]) for (int i = 0; i < len; i++) n[i] = (NAME32[])data == null ? "" : ((NAME32[])data)[i].GetName();
      else if (data is NAME48[]) for (int i = 0; i < len; i++) n[i] = (NAME48[])data == null ? "" : ((NAME48[])data)[i].GetName();
      else if (data is NAME64[]) for (int i = 0; i < len; i++) n[i] = (NAME64[])data == null ? "" : ((NAME64[])data)[i].GetName();
      else if (data is NAME8I[]) for (int i = 0; i < len; i++) n[i] = (NAME8I[])data == null ? "" : ((NAME8I[])data)[i].GetName();
      else if (data is NAME16I[]) for (int i = 0; i < len; i++) n[i] = (NAME16I[])data == null ? "" : ((NAME16I[])data)[i].GetName();
      else if (data is NAME32I[]) for (int i = 0; i < len; i++) n[i] = (NAME32I[])data == null ? "" : ((NAME32I[])data)[i].GetName();
      else if (data is NAME48I[]) for (int i = 0; i < len; i++) n[i] = (NAME48I[])data == null ? "" : ((NAME48I[])data)[i].GetName();
      else if (data is NAME64I[]) for (int i = 0; i < len; i++) n[i] = (NAME64I[])data == null ? "" : ((NAME64I[])data)[i].GetName();
      else if (data is NAME16II[]) for (int i = 0; i < len; i++) n[i] = (NAME16II[])data == null ? "" : ((NAME16II[])data)[i].GetName();
      else if (data is NAME16FI[]) for (int i = 0; i < len; i++) n[i] = (NAME16FI[])data == null ? "" : ((NAME16FI[])data)[i].GetName();
      else if (data is NAME16DBLDBL[]) for (int i = 0; i < len; i++) n[i] = (NAME16DBLDBL[])data == null ? "" : ((NAME16DBLDBL[])data)[i].GetName();
      else if (data is NAME32DBLDBL[]) for (int i = 0; i < len; i++) n[i] = (NAME32DBLDBL[])data == null ? "" : ((NAME32DBLDBL[])data)[i].GetName();
      else if (data is NAME64DBLDBL[]) for (int i = 0; i < len; i++) n[i] = (NAME64DBLDBL[])data == null ? "" : ((NAME64DBLDBL[])data)[i].GetName();
      else if (data is DUSTRING[]) for (int i = 0; i < len; i++) n[i] = (DUSTRING[])data == null ? "" : ((DUSTRING[])data)[i].GetName();
      else if (data is DUNAME[]) for (int i = 0; i < len; i++) n[i] = (DUNAME[])data == null ? "" : ((DUNAME[])data)[i].GetName();
      else
      { // type does not contain names !
        return null;
      }
      return n;
    }
    public static unsafe IntPtr GetDataPtr(Object data)
    {
      return GetTineDataPtr(data, null);
    }
    public static unsafe IntPtr GetDataPtr(Object data,string tag)
    {
      return GetTineDataPtr(data,tag);
    }
    private static unsafe IntPtr GetTineDataPtr(Object data)
    {
      return GetTineDataPtr(data, null);
    }
    private static unsafe IntPtr GetTineDataPtr(Object data, string cname)
    {
      if (data == null) return IntPtr.Zero;
      Boolean isArray = data is Array;
      if (isArray && ((Array)data).Rank != 1)
      {
        throw new Exception("object is NOT a 1 dimensional array !");
      }
      Type dtype = data.GetType().GetElementType();
      switch (Type.GetTypeCode(dtype))
      {
        //case TypeCode.Boolean:
        //  fixed (double* p = (double[])data) return new IntPtr((void*)p);
        case TypeCode.Byte:
          fixed (byte* p = (byte[])data) return new IntPtr((void*)p);
        case TypeCode.Char:
          fixed (char* p = (char[])data) return new IntPtr((void*)p);
        case TypeCode.Double:
          fixed (double *p = (double[])data) return new IntPtr((void *)p);
        case TypeCode.Empty:
          if (data is TTaggedStruct)
          {
            cname = ((TTaggedStruct)data).GetTag();
            goto case TypeCode.Object;
          }
          if (data != null && !isArray)
          {
            throw new Exception("data object argument is NOT an array and is not passed by reference !");
          }
          return IntPtr.Zero;
        case TypeCode.Int16:
          fixed (Int16* p = (Int16[])data) return new IntPtr((void*)p);
        case TypeCode.Int32:
          fixed (Int32* p = (Int32[])data) return new IntPtr((void*)p);
        case TypeCode.Int64:
          fixed (Int64* p = (Int64[])data) return new IntPtr((void*)p);
        case TypeCode.Single:
          fixed (float* p = (float[])data) return new IntPtr((void*)p);
        case TypeCode.String:
          {
            String[] sa = (String[])data;
            IntPtr p = GetIntPtrFromManagedDataObject(sa, GetSizeOfStringData(sa));
            void *vp = p.ToPointer();
            byte **bp = (byte **)vp; // have to use bytes as the underlying native strings are ascii chars 
            *bp = (byte *)((byte **)vp + sa.Length);
            byte *bc;
            char[] ca;
            for (int i = 0; i < sa.Length; i++)
            {
              bc = *bp++;
              if (sa[i] != null)
              { // indicates a 'put operation'
                ca = sa[i].ToCharArray();
                for (int k = 0; k < ca.Length; k++) *bc++ = (byte)ca[k];
                *bc++ = 0;
                if (ca.Length % 2 == 0) *bc++ = 0;
              }
              if (i < sa.Length - 1) *bp = (byte *)bc;
            }
            return p;
          }
        case TypeCode.UInt16:
          fixed (UInt16* p = (UInt16[])data) return new IntPtr((void*)p);
        case TypeCode.UInt32:
          fixed (UInt32* p = (UInt32[])data) return new IntPtr((void*)p);
        case TypeCode.UInt64:
          fixed (UInt64* p = (UInt64[])data) return new IntPtr((void*)p);
        case TypeCode.Object:
          if (data is NAME8[]) fixed (NAME8* p = &(((NAME8[])data)[0])) { return new IntPtr((void*)p); }
          if (data is NAME16[]) fixed (NAME16* p = &(((NAME16[])data)[0])) { return new IntPtr((void*)p); }
          if (data is NAME32[]) fixed (NAME32* p = &(((NAME32[])data)[0])) { return new IntPtr((void*)p); }
          if (data is NAME48[]) fixed (NAME48* p = &(((NAME48[])data)[0])) { return new IntPtr((void*)p); }
          if (data is NAME64[]) fixed (NAME64* p = &(((NAME64[])data)[0])) { return new IntPtr((void*)p); }
          if (data is FLTINT[]) fixed (FLTINT* p = &(((FLTINT[])data)[0])) { return new IntPtr((void *)p); }
          if (data is INTINT[]) fixed (INTINT* p = &(((INTINT[])data)[0])) { return new IntPtr((void*)p); }
          if (data is FLTFLT[]) fixed (FLTFLT* p = &(((FLTFLT[])data)[0])) { return new IntPtr((void*)p); }
          if (data is DBLDBL[]) fixed (DBLDBL* p = &(((DBLDBL[])data)[0])) { return new IntPtr((void*)p); }
          if (data is NAME8I[]) fixed (NAME8I* p = &(((NAME8I[])data)[0])) { return new IntPtr((void*)p); }
          if (data is NAME16I[]) fixed (NAME16I* p = &(((NAME16I[])data)[0])) { return new IntPtr((void*)p); }
          if (data is NAME32I[]) fixed (NAME32I* p = &(((NAME32I[])data)[0])) { return new IntPtr((void*)p); }
          if (data is NAME48I[]) fixed (NAME48I* p = &(((NAME48I[])data)[0])) { return new IntPtr((void*)p); }
          if (data is NAME64I[]) fixed (NAME64I* p = &(((NAME64I[])data)[0])) { return new IntPtr((void*)p); }
          if (data is NAME64DBL[]) fixed (NAME64DBL* p = &(((NAME64DBL[])data)[0])) { return new IntPtr((void*)p); }
          if (data is NAME16FI[]) fixed (NAME16FI* p = &(((NAME16FI[])data)[0])) { return new IntPtr((void*)p); }
          if (data is NAME16II[]) fixed (NAME16II* p = &(((NAME16II[])data)[0])) { return new IntPtr((void*)p); }
          if (data is NAME16DBLDBL[]) fixed (NAME16DBLDBL* p = &(((NAME16DBLDBL[])data)[0])) { return new IntPtr((void*)p); }
          if (data is NAME32DBLDBL[]) fixed (NAME32DBLDBL* p = &(((NAME32DBLDBL[])data)[0])) { return new IntPtr((void*)p); }
          if (data is NAME64DBLDBL[]) fixed (NAME64DBLDBL* p = &(((NAME64DBLDBL[])data)[0])) { return new IntPtr((void*)p); }
          if (data is INTFLTINT[]) fixed (INTFLTINT* p = &(((INTFLTINT[])data)[0])) { return new IntPtr((void*)p); }
          if (data is FLTFLTINT[]) fixed (FLTFLTINT* p = &(((FLTFLTINT[])data)[0])) { return new IntPtr((void*)p); }
          if (data is FLTINTINT[]) fixed (FLTINTINT* p = &(((FLTINTINT[])data)[0])) { return new IntPtr((void*)p); }
          if (data is INTINTINT[]) fixed (INTINTINT* p = &(((INTINTINT[])data)[0])) { return new IntPtr((void*)p); }
          if (data is DBLDBLDBL[]) fixed (DBLDBLDBL* p = &(((DBLDBLDBL[])data)[0])) { return new IntPtr((void*)p); }
          if (data is INTINTINTINT[]) fixed (INTINTINTINT* p = &(((INTINTINTINT[])data)[0])) { return new IntPtr((void*)p); }
          if (data is INTFLTFLTFLT[]) fixed (INTFLTFLTFLT* p = &(((INTFLTFLTFLT[])data)[0])) { return new IntPtr((void*)p); }
          if (data is INTINTFLTFLT[]) fixed (INTINTFLTFLT* p = &(((INTINTFLTFLT[])data)[0])) { return new IntPtr((void*)p); }
          if (data is FLTINTFLTINT[]) fixed (FLTINTFLTINT* p = &(((FLTINTFLTINT[])data)[0])) { return new IntPtr((void*)p); }
          if (data is DWINDOW[]) fixed (DWINDOW* p = &(((DWINDOW[])data)[0])) { return new IntPtr((void*)p); }
          if (data is NAME64DBLDBLDBL[]) fixed (NAME64DBLDBLDBL* p = &(((NAME64DBLDBLDBL[])data)[0])) { return new IntPtr((void*)p); }
          if (data is DUSTRING[]) fixed (DUSTRING* p = &(((DUSTRING[])data)[0])) { return new IntPtr((void*)p); }
          if (data is TTaggedStruct[])
          {
            TTaggedStruct[] tts = (TTaggedStruct[])data;
            byte[] ba;
            if ((ba = tts[0].GetStructByteArray()) != null)
            {
              fixed (byte* p = (byte[])ba) return new IntPtr((void*)p);
            }
          }
          if (cname != null || dtype.IsValueType)
          {
            if (cname == null) cname = dtype.IsArray ? dtype.GetElementType().Name : dtype.Name;
            int ssiz;
            if ((ssiz = native.GetStructSize(cname)) > 0)
            {
              if (isArray)
              {
                return GetIntPtrFromManagedDataObject(data, ssiz * ((Array)data).Length);
              }
              else
              {
                return GetIntPtrFromManagedDataObject(data, ssiz);
              }
            }
          }
          break;
        //  default: return Formats.CF_NULL;
      }
      return IntPtr.Zero;
    }
    private UInt32 GetTineArrayLength(Object data)
    {
      if (data == null) return 0;
      if (data is System.Array)
      {
        Array da = (Array)data;
        //if (da.Rank != 1) return 0;
        //return (UInt32)da.GetLength(0);
        int dim = da.Rank;
        int len = 1;
        while (--dim >= 0) len *= da.GetLength(dim);
        return (UInt32)len;
      }
      return 1;
    }
    private unsafe DTYPE tdts = new DTYPE();
    private Object data;
    private DIMAGE image;
    public Boolean isByRef = false;
    public Object GetDataObject() { return data; }
    public double[] GetXAxis()
    {
      return GetXAxis(data);
    }
    public double[] GetYAxis()
    {
      return GetYAxis(data);
    }
    public static String GetDataAsString(Object data)
    {
      if (data == null) return "(null)";
      if (data is String) return (String)data;
      if (data is Array) return data.ToString();
      switch (GetTineType(data))
      {
        case Formats.CF_NULL: return "(null)";
        case Formats.CF_CHAR: return new String((char[])data);
        case Formats.CF_INT8: return "" + ((Byte)data);
        case Formats.CF_INT16: return "" + ((Int16)data);
        case Formats.CF_INT32: return "" + ((Int32)data);
        case Formats.CF_INT64: return "" + ((Int64)data);
        case Formats.CF_FLOAT: return "" + ((Single)data);
        case Formats.CF_DOUBLE: return "" + ((Double)data);
        case Formats.CF_NAME8: return ((NAME8)data).ToString();
        case Formats.CF_NAME16: return ((NAME16)data).ToString();
        case Formats.CF_NAME32: return ((NAME32)data).ToString();
        case Formats.CF_NAME48: return ((NAME48)data).ToString();
        case Formats.CF_NAME64: return ((NAME64)data).ToString();
        case Formats.CF_INTINT: return ((INTINT)data).ToString();
        case Formats.CF_INTINTINT: return ((INTINTINT)data).ToString();
        case Formats.CF_INTINTINTINT: return ((INTINTINTINT)data).ToString();
        case Formats.CF_FLTINT: return ((FLTINT)data).ToString();
        case Formats.CF_FLTFLT: return ((FLTFLT)data).ToString();
        case Formats.CF_INTFLTFLTFLT: return ((INTFLTFLTFLT)data).ToString();
        case Formats.CF_INTFLTINT: return ((INTFLTINT)data).ToString();
        case Formats.CF_INTINTFLTFLT: return ((INTINTFLTFLT)data).ToString();
        case Formats.CF_DBLDBL: return ((DBLDBL)data).ToString();
        case Formats.CF_DBLDBLDBL: return ((DBLDBLDBL)data).ToString();
        case Formats.CF_FLTFLTINT: return ((FLTFLTINT)data).ToString();
        case Formats.CF_FLTINTFLTINT: return ((FLTINTFLTINT)data).ToString();
        case Formats.CF_FLTINTINT: return ((FLTINTINT)data).ToString();
        case Formats.CF_NAME8I: return ((NAME8I)data).ToString();
        case Formats.CF_NAME16I: return ((NAME16I)data).ToString();
        case Formats.CF_NAME32I: return ((NAME32I)data).ToString();
        case Formats.CF_NAME48I: return ((NAME48I)data).ToString();
        case Formats.CF_NAME64I: return ((NAME64I)data).ToString();
        case Formats.CF_NAME16FI: return ((NAME16FI)data).ToString();
        case Formats.CF_NAME16II: return ((NAME16II)data).ToString();
        case Formats.CF_NAME16DBLDBL: return ((NAME16DBLDBL)data).ToString();
        case Formats.CF_NAME32DBLDBL: return ((NAME32DBLDBL)data).ToString();
        case Formats.CF_NAME64DBLDBL: return ((NAME64DBLDBL)data).ToString();
        case Formats.CF_NAME64DBLDBLDBL: return ((NAME64DBLDBLDBL)data).ToString();
        case Formats.CF_USTRING: return ((DUSTRING)data).ToString();
        case Formats.CF_UNAME: return ((DUNAME)data).ToString();
        case Formats.CF_STRUCT:
          String tag = getStructName(data);
          int ssiz = native.GetStructSize(tag);
          if (ssiz > 0)
          {
            char[] stringbuf = new char[ssiz * 32];
            if (native.GetStructureAsString(tag, TDataType.GetTineDataPtr(data), stringbuf, stringbuf.Length) == 0)
            {
              return new String(stringbuf).Trim(new char[] { '\0' });
            }
          }
          break;
        default: break;
      }
      return "(unknown)";
    }
    public static String[] GetDataAsStringArray(Object data)
    {
      return GetDataAsStringArray(data, 0);
    }
    public static String[] GetDataAsStringArray(Object data, int precision)
    {
      if (data == null) return new String[]{"(null)"};
      byte dataType = GetTineType(data);
      if (!(data is Array))
      {
        String rs = GetDataAsString(data);
        String[] rsa = rs.Split(new char[] { '\n' }, StringSplitOptions.RemoveEmptyEntries);
        if (dataType == Formats.CF_STRUCT)
        { // single structure -> return array of field values
          String[] rssa = new String[rsa.Length - 2];
          for (int i=0; i<rssa.Length; i++)
          {
            rssa[i] = rsa[i + 1].Substring(rsa[i + 1].IndexOf(':') + 1).Trim();
          }
          rsa = rssa;
        }
        return rsa;
      }
      int len = ((Array)data).GetLength(0);
      String[] sa = new String[len];
      String sf = precision == 0 ? "G6" : "F" + precision;
      switch (dataType)
      {
        case Formats.CF_NULL: return new String[] { "(null)" };
        case Formats.CF_CHAR: 
          sbyte[] sba = (sbyte[])data;
          char[] ca = new char[len];
          for (int i = 0; i < len; i++) ca[i] = (char)sba[i];
          return new String[] { new String(ca)};
        case Formats.CF_BOOLEAN: for (int i = 0; i < len; i++) sa[i] = "" + ((Boolean[])data)[i]; break;
        case Formats.CF_INT8: for (int i = 0; i < len; i++) sa[i] = "" + ((Byte[])data)[i]; break;
        case Formats.CF_INT16: for (int i = 0; i < len; i++) sa[i] = "" + ((Int16[])data)[i]; break;
        case Formats.CF_INT32: for (int i = 0; i < len; i++) sa[i] = "" + ((Int32[])data)[i]; break;
        case Formats.CF_INT64: for (int i = 0; i < len; i++) sa[i] = "" + ((Int64[])data)[i]; break;
        case Formats.CF_FLOAT:
          for (int i = 0; i < len; i++)
            sa[i] = ((Single[])data)[i].ToString(sf);
          break;
        case Formats.CF_DOUBLE:
          for (int i = 0; i < len; i++)
            sa[i] = ((Double[])data)[i].ToString(sf);
          break;
        case Formats.CF_NAME8:
          for (int i = 0; i < len; i++) sa[i] = "" + ((NAME8[])data)[i].ToString(); break;
        case Formats.CF_NAME16:
          for (int i = 0; i < len; i++) sa[i] = "" + ((NAME16[])data)[i].ToString(); break;
        case Formats.CF_NAME32:
          for (int i = 0; i < len; i++) sa[i] = "" + ((NAME32[])data)[i].ToString(); break;
        case Formats.CF_NAME48:
          for (int i = 0; i < len; i++) sa[i] = "" + ((NAME48[])data)[i].ToString(); break;
        case Formats.CF_NAME64:
          for (int i = 0; i < len; i++) sa[i] = "" + ((NAME64[])data)[i].ToString(); break;
        case Formats.CF_INTINT: 
          for (int i = 0; i < len; i++) sa[i] = "" + ((INTINT[])data)[i].ToString(); break;
        case Formats.CF_INTINTINT: 
          for (int i = 0; i < len; i++) sa[i] = "" + ((INTINTINT[])data)[i].ToString(); break;
        case Formats.CF_INTINTINTINT:
          for (int i = 0; i < len; i++) sa[i] = "" + ((INTINTINTINT[])data)[i].ToString(); break;
        case Formats.CF_FLTINT:
          for (int i = 0; i < len; i++) sa[i] = "" + ((FLTINT[])data)[i].ToString(); break;
        case Formats.CF_FLTFLT:
          for (int i = 0; i < len; i++) sa[i] = "" + ((FLTFLT[])data)[i].ToString(); break;
        case Formats.CF_INTFLTFLTFLT:
          for (int i = 0; i < len; i++) sa[i] = "" + ((INTFLTFLTFLT[])data)[i].ToString(); break;
        case Formats.CF_INTFLTINT:
          for (int i = 0; i < len; i++) sa[i] = "" + ((INTFLTINT[])data)[i].ToString(); break;
        case Formats.CF_INTINTFLTFLT:
          for (int i = 0; i < len; i++) sa[i] = "" + ((INTINTFLTFLT[])data)[i].ToString(); break;
        case Formats.CF_DBLDBL:
          for (int i = 0; i < len; i++) sa[i] = "" + ((DBLDBL[])data)[i].ToString(); break;
        case Formats.CF_DBLDBLDBL:
          for (int i = 0; i < len; i++) sa[i] = "" + ((DBLDBLDBL[])data)[i].ToString(); break;
        case Formats.CF_FLTFLTINT:
          for (int i = 0; i < len; i++) sa[i] = "" + ((FLTFLTINT[])data)[i].ToString(); break;
        case Formats.CF_FLTINTFLTINT:
          for (int i = 0; i < len; i++) sa[i] = "" + ((FLTINTFLTINT[])data)[i].ToString(); break;
        case Formats.CF_FLTINTINT:
          for (int i = 0; i < len; i++) sa[i] = "" + ((FLTINTINT[])data)[i].ToString(); break;
        case Formats.CF_NAME8I:
          for (int i = 0; i < len; i++) sa[i] = "" + ((NAME8I[])data)[i].ToString(); break;
        case Formats.CF_NAME16I:
          for (int i = 0; i < len; i++) sa[i] = "" + ((NAME16I[])data)[i].ToString(); break;
        case Formats.CF_NAME32I:
          for (int i = 0; i < len; i++) sa[i] = "" + ((NAME32I[])data)[i].ToString(); break;
        case Formats.CF_NAME48I:
          for (int i = 0; i < len; i++) sa[i] = "" + ((NAME48I[])data)[i].ToString(); break;
        case Formats.CF_NAME64I:
          for (int i = 0; i < len; i++) sa[i] = "" + ((NAME64I[])data)[i].ToString(); break;
        case Formats.CF_NAME16FI:
          for (int i = 0; i < len; i++) sa[i] = "" + ((NAME16FI[])data)[i].ToString(); break;
        case Formats.CF_NAME16II:
          for (int i = 0; i < len; i++) sa[i] = "" + ((NAME16II[])data)[i].ToString(); break;
        case Formats.CF_NAME16DBLDBL:
          for (int i = 0; i < len; i++) sa[i] = "" + ((NAME16DBLDBL[])data)[i].ToString(); break;
        case Formats.CF_NAME32DBLDBL:
          for (int i = 0; i < len; i++) sa[i] = "" + ((NAME32DBLDBL[])data)[i].ToString(); break;
        case Formats.CF_NAME64DBLDBL:
          for (int i = 0; i < len; i++) sa[i] = "" + ((NAME64DBLDBL[])data)[i].ToString(); break;
        case Formats.CF_NAME64DBLDBLDBL:
          for (int i = 0; i < len; i++) sa[i] = "" + ((NAME64DBLDBLDBL[])data)[i].ToString(); break;
        case Formats.CF_USTRING:
          for (int i = 0; i < len; i++) sa[i] = "" + ((DUSTRING[])data)[i].ToString(); break;
        case Formats.CF_UNAME:
          for (int i = 0; i < len; i++) sa[i] = "" + ((DUNAME[])data)[i].ToString(); break;
        case Formats.CF_STRING:
          for (int i = 0; i < len; i++) sa[i] = "" + ((String[])data)[i]; break;
        case Formats.CF_STRUCT: 
          String tag = getStructName(data);
          int ssiz = native.GetStructSize(tag);
          int cc = Errors.name_not_found;
          if (ssiz > 0)
          {
            char[] stringbuf = new char[ssiz * 32];
            String rs;
            String[] rsa;
            if ((cc = native.GetStructureAsString(tag, TDataType.GetTineDataPtr(data), stringbuf, stringbuf.Length)) == 0)
            {
              rs = new String(stringbuf).Trim(new char[] { '\0' });
              rsa = rs.Split(new char[] { '\n' }, StringSplitOptions.RemoveEmptyEntries);
              String[] rssa = new String[rsa.Length - 2];
              for (int i = 0; i < rssa.Length; i++)
              {
                rssa[i] = rsa[i + 1].Substring(rsa[i + 1].IndexOf(':') + 1).Trim();
              }
              rsa = rssa;
              return rsa;
            }
          }
          return new String[] { Errors.ToString(cc) };
        default: return new String[] { "(format not supported)" };
      }
      return sa;
    }
    public String[] GetDataAsStringArray()
    {
      return GetDataAsStringArray(data);
    }
    public String[] GetChannelNames()
    {
      return GetChannelNames(data);
    }
    public unsafe DTYPE GetDTYPE() 
    {
      return tdts; 
    }
    private unsafe IntPtr GetTineDataPtr(ref double data)
    {
      fixed (double* p = &data) return new IntPtr((void*)p);
    }
    private unsafe IntPtr GetTineDataPtr(ref float data)
    {
      fixed (float* p = &data) return new IntPtr((void*)p);
    }
    private unsafe IntPtr GetTineDataPtr(ref byte data)
    {
      fixed (byte* p = &data) return new IntPtr((void*)p);
    }
    private unsafe IntPtr GetTineDataPtr(ref Int16 data)
    {
      fixed (Int16* p = &data) return new IntPtr((void*)p);
    }
    private unsafe IntPtr GetTineDataPtr(ref UInt16 data)
    {
      fixed (UInt16* p = &data) return new IntPtr((void*)p);
    }
    private unsafe IntPtr GetTineDataPtr(ref Int32 data)
    {
      fixed (Int32* p = &data) return new IntPtr((void*)p);
    }
    private unsafe IntPtr GetTineDataPtr(ref UInt32 data)
    {
      fixed (UInt32* p = &data) return new IntPtr((void*)p);
    }
    private unsafe IntPtr GetTineDataPtr(ref Int64 data)
    {
      fixed (Int64* p = &data) return new IntPtr((void*)p);
    }
    private unsafe IntPtr GetTineDataPtr(ref UInt64 data)
    {
      fixed (UInt64* p = &data) return new IntPtr((void*)p);
    }
    /// <summary>
    /// makes a TDataType from the .NET reference passed
    /// </summary>
    /// <param name="data">a reference to the the scalar object which should be bound to the TDataType</param>
    public unsafe TDataType(ref double data)
    {
      gchData = GCHandle.Alloc(data, GCHandleType.Pinned);
      this.data = data;
      this.isByRef = true;
      tdts.dFormat = (short)(Formats.LFMT_OFFSET + Formats.CF_DOUBLE);
      tdts.dArrayLength = 1;
      tdts.data = GetTineDataPtr(ref data);
      tdts.dTag = new char[Constants.TAG_NAME_SIZE];
    }
    /// <summary>
    /// makes a TDataType from the .NET reference passed
    /// </summary>
    /// <param name="data">a reference to the the scalar object which should be bound to the TDataType</param>
    public unsafe TDataType(ref float data)
    {
      gchData = GCHandle.Alloc(data, GCHandleType.Pinned); //n.b. this only pins a new object boxing the value
      this.data = data;
      this.isByRef = true;
      tdts.dFormat = (short)(Formats.LFMT_OFFSET + Formats.CF_FLOAT);
      tdts.dArrayLength = 1;
      tdts.data = GetTineDataPtr(ref data);
      tdts.dTag = new char[Constants.TAG_NAME_SIZE];
    }
    /// <summary>
    /// makes a TDataType from the .NET reference passed
    /// </summary>
    /// <param name="data">a reference to the the scalar object which should be bound to the TDataType</param>
    public unsafe TDataType(ref UInt16 data)
    {
      gchData = GCHandle.Alloc(data, GCHandleType.Pinned);
      this.data = data;
      this.isByRef = true;
      tdts.dFormat = (short)(Formats.LFMT_OFFSET + Formats.CF_INT16);
      tdts.dArrayLength = 1;
      tdts.data = GetTineDataPtr(ref data);
      tdts.dTag = new char[Constants.TAG_NAME_SIZE];
    }
    /// <summary>
    /// makes a TDataType from the .NET reference passed
    /// </summary>
    /// <param name="data">a reference to the the scalar object which should be bound to the TDataType</param>
    public unsafe TDataType(ref Int16 data)
    {
      gchData = GCHandle.Alloc(data, GCHandleType.Pinned);
      this.data = data;
      this.isByRef = true;
      tdts.dFormat = (short)(Formats.LFMT_OFFSET + Formats.CF_INT16);
      tdts.dArrayLength = 1;
      tdts.data = GetTineDataPtr(ref data);
      tdts.dTag = new char[Constants.TAG_NAME_SIZE];
    }
    /// <summary>
    /// makes a TDataType from the .NET reference passed
    /// </summary>
    /// <param name="data">a reference to the the scalar object which should be bound to the TDataType</param>
    public unsafe TDataType(ref UInt32 data)
    {
      gchData = GCHandle.Alloc(data, GCHandleType.Pinned);
      this.data = data;
      this.isByRef = true;
      tdts.dFormat = (short)(Formats.LFMT_OFFSET + Formats.CF_INT32);
      tdts.dArrayLength = 1;
      tdts.data = GetTineDataPtr(ref data);
      tdts.dTag = new char[Constants.TAG_NAME_SIZE];
    }
    /// <summary>
    /// makes a TDataType from the .NET reference passed
    /// </summary>
    /// <param name="data">a reference to the the scalar object which should be bound to the TDataType</param>
    public unsafe TDataType(ref Int32 data)
    {
      gchData = GCHandle.Alloc(data, GCHandleType.Pinned);
      this.data = data;
      this.isByRef = true;
      tdts.dFormat = (short)(Formats.LFMT_OFFSET + Formats.CF_INT32);
      tdts.dArrayLength = 1;
      tdts.data = GetTineDataPtr(ref data);
      tdts.dTag = new char[Constants.TAG_NAME_SIZE];
    }
    /// <summary>
    /// makes a TDataType from the .NET reference passed
    /// </summary>
    /// <param name="data">a reference to the the scalar object which should be bound to the TDataType</param>
    public unsafe TDataType(ref UInt64 data)
    {
      gchData = GCHandle.Alloc(data, GCHandleType.Pinned);
      this.data = data;
      this.isByRef = true;
      tdts.dFormat = (short)(Formats.LFMT_OFFSET + Formats.CF_INT64);
      tdts.dArrayLength = 1;
      tdts.data = GetTineDataPtr(ref data);
      tdts.dTag = new char[Constants.TAG_NAME_SIZE];
    }
    /// <summary>
    /// makes a TDataType from the .NET reference passed
    /// </summary>
    /// <param name="data">a reference to the the scalar object which should be bound to the TDataType</param>
    public unsafe TDataType(ref Int64 data)
    {
      gchData = GCHandle.Alloc(data, GCHandleType.Pinned);
      this.data = data;
      this.isByRef = true;
      tdts.dFormat = (short)(Formats.LFMT_OFFSET + Formats.CF_INT64);
      tdts.dArrayLength = 1;
      tdts.data = GetTineDataPtr(ref data);
      tdts.dTag = new char[Constants.TAG_NAME_SIZE];
    }
    private void AssignCompoundDataType(object data,byte fmt,int siz)
    {
      gchData = GCHandle.Alloc(data, GCHandleType.Pinned);
      this.data = data;
      tdts.dFormat = (short)(Formats.LFMT_OFFSET + fmt);
      tdts.dArrayLength = 1;
      tdts.data = GetIntPtrFromManagedDataObject(data, siz);
      tdts.dTag = new char[Constants.TAG_NAME_SIZE];
    }
    public unsafe TDataType(ref NAME64 data)
    {
      AssignCompoundDataType(data, Formats.CF_NAME64, NAME64.sizeInBytes);
    }
    public unsafe TDataType(ref NAME32 data)
    {
      AssignCompoundDataType(data, Formats.CF_NAME32, NAME32.sizeInBytes);
    }
    public unsafe TDataType(ref NAME16 data)
    {
      AssignCompoundDataType(data, Formats.CF_NAME16, NAME16.sizeInBytes);
    }
    public unsafe TDataType(ref NAME8 data)
    {
      AssignCompoundDataType(data, Formats.CF_NAME8, NAME8.sizeInBytes);
    }
    public unsafe TDataType(ref FLTINT data)
    {
      AssignCompoundDataType(data, Formats.CF_FLTINT, FLTINT.sizeInBytes);
    }
    public unsafe TDataType(ref INTINT data)
    {
      AssignCompoundDataType(data, Formats.CF_INTINT, INTINT.sizeInBytes);
    }
    public unsafe TDataType(ref FLTFLT data)
    {
      AssignCompoundDataType(data, Formats.CF_FLTFLT, FLTFLT.sizeInBytes);
    }
    public unsafe TDataType(ref DBLDBL data)
    {
      AssignCompoundDataType(data, Formats.CF_DBLDBL, DBLDBL.sizeInBytes);
    }
    public unsafe TDataType(ref NAME64I data)
    {
      AssignCompoundDataType(data, Formats.CF_NAME64I, NAME64I.sizeInBytes);
    }
    public unsafe TDataType(ref NAME64DBL data)
    {
      AssignCompoundDataType(data, Formats.CF_NAME64DBL, NAME64DBL.sizeInBytes);
    }
    public unsafe TDataType(ref NAME32I data)
    {
      AssignCompoundDataType(data, Formats.CF_NAME32I, NAME32I.sizeInBytes);
    }
    public unsafe TDataType(ref NAME16I data)
    {
      AssignCompoundDataType(data, Formats.CF_NAME16I, NAME16I.sizeInBytes);
    }
    public unsafe TDataType(ref NAME16FI data)
    {
      AssignCompoundDataType(data, Formats.CF_NAME16FI, NAME16FI.sizeInBytes);
    }
    public unsafe TDataType(ref NAME64DBLDBL data)
    {
      AssignCompoundDataType(data, Formats.CF_NAME64DBLDBL, NAME64DBLDBL.sizeInBytes);
    }
    public unsafe TDataType(ref NAME32DBLDBL data)
    {
      AssignCompoundDataType(data, Formats.CF_NAME32DBLDBL, NAME32DBLDBL.sizeInBytes);
    }
    public unsafe TDataType(ref INTFLTINT data)
    {
      AssignCompoundDataType(data, Formats.CF_INTFLTINT, INTFLTINT.sizeInBytes);
    }
    public unsafe TDataType(ref FLTFLTINT data)
    {
      AssignCompoundDataType(data, Formats.CF_FLTFLTINT, FLTFLTINT.sizeInBytes);
    }
    public unsafe TDataType(ref FLTINTINT data)
    {
      AssignCompoundDataType(data, Formats.CF_FLTINTINT, FLTINTINT.sizeInBytes);
    }
    public unsafe TDataType(ref INTINTINTINT data)
    {
      AssignCompoundDataType(data, Formats.CF_INTINTINTINT, INTINTINTINT.sizeInBytes);
    }
    public unsafe TDataType(ref DUSTRING data)
    {
      AssignCompoundDataType(data, Formats.CF_USTRING, DUSTRING.sizeInBytes);
    }
    public unsafe TDataType(DIMAGE[] data)
    {
      if (data == null)
      {
        SetAsNullType();
        return;
      }
      if (data.Length != 1) throw new Exception("type DIMAGE must be passed as an array of 1 object");
      int tsize = Marshal.SizeOf(data[0].header);
      gchData = GCHandle.Alloc(data[0].header, GCHandleType.Pinned);
      this.data = data;
      this.image = data[0];
      tdts.data = GetIntPtrFromManagedDataObject(data[0].header, tsize);
      Marshal.StructureToPtr(data[0].header, tdts.data, true);
      gchRef = data[0].header;
      tdts.dArrayLength = data[0].header.frameBufferSize; // length = embedded frame buffer length
      tdts.dFormat = (short)(Formats.LFMT_OFFSET + Formats.CF_IMAGE);
      return;
    }
    public unsafe TDataType(ref DIMAGE data)
    {
      int tsize = Marshal.SizeOf(data.header);
      gchData = GCHandle.Alloc(data.header, GCHandleType.Pinned);
      this.data = data;
      this.image = data;
      tdts.data = GetIntPtrFromManagedDataObject(data.header, tsize);
      Marshal.StructureToPtr(data.header, tdts.data, true);
      gchRef = data.header;
      tdts.dArrayLength = data.header.frameBufferSize; // length = embedded frame buffer length
      tdts.dFormat = (short)(Formats.LFMT_OFFSET + Formats.CF_IMAGE);
      return;
    }
    private GCHandle gchData;
    protected Object gchDataObject = null;
    public Object GetIopDataObject()
    {
      return gchDataObject;
    }
    private Object gchRef = null;
    public Object GetReferencedDataObject()
    {
      return gchRef;
    }
    private void SetAsNullType()
    {
      tdts.dFormat = (short)(Formats.LFMT_OFFSET + Formats.CF_NULL);
      tdts.dArrayLength = 0;
      tdts.data = IntPtr.Zero;
      tdts.dTag = new char[Constants.TAG_NAME_SIZE];
    }
    private static TDataType dtNull = new TDataType(null);
    public static TDataType empty()
    {
      return dtNull;
    }
    /// <summary>
    /// makes a TDataType from the .NET object passed
    /// </summary>
    /// <param name="data">the .NET object which should be bound to the TDataType</param>
    public TDataType(Object data)
    {
      if (data == null)
      {
        SetAsNullType();
        return;
      }
      if (data is DIMAGE[])
      {
        unsafe
        {
          if (((DIMAGE[])data).Length != 1) throw new Exception("type DIMAGE must be passed as an array of 1 object");
          int tsize = Marshal.SizeOf(((DIMAGE[])data)[0].header);
          gchData = GCHandle.Alloc(((DIMAGE[])data)[0].header, GCHandleType.Pinned);
          this.data = data;
          this.image = ((DIMAGE[])data)[0];
          tdts.data = GetIntPtrFromManagedDataObject(((DIMAGE[])data)[0].header, tsize);
          Marshal.StructureToPtr(((DIMAGE[])data)[0].header, tdts.data, true);
          gchRef = ((DIMAGE[])data)[0].header;
          tdts.dArrayLength = ((DIMAGE[])data)[0].header.frameBufferSize; // length = embedded frame buffer length
          tdts.dFormat = (short)(Formats.LFMT_OFFSET + Formats.CF_IMAGE);
        }
        return;
      }
      //if (!(data is Array)) throw exception
      int fmt = GetTineType(data);
      if (fmt == Formats.CF_NULL)
      {
        SetAsNullType();
        return;
      }
      tdts.dFormat = (short)(Formats.LFMT_OFFSET + fmt);
      tdts.dArrayLength = GetTineArrayLength(data);
      if (data is TTaggedStruct[])
      {
        char[] ch = ((TTaggedStruct[])data)[0].GetTag().ToCharArray();
        byte[] bd = ((TTaggedStruct[])data)[0].GetStructByteArray();
        if (gchData == null || !gchData.IsAllocated) gchData = GCHandle.Alloc(bd, GCHandleType.Pinned);
        this.data = bd;
        gchRef = bd;
        tdts.data = gchData.AddrOfPinnedObject();
        tdts.dTag = new char[Constants.TAG_NAME_SIZE];
        Array.Copy(ch, tdts.dTag, Math.Min(ch.Length, Constants.TAG_NAME_SIZE));
        return;
      }
      if (data is String[])
      {
        tdts.data = GetIntPtrFromManagedDataObject(data, GetSizeOfStringData((String[])data));
        gchRef = data;
        this.data = data;
        return;
      }
      if (data is Boolean[])
      {
        Int32[] idata = new int[tdts.dArrayLength];
        for (int i = 0; i < tdts.dArrayLength; i++)
        { // map initial data set 
          idata[i] = ((Boolean[])data)[i] == true ? 1 : 0;
        }
        if (gchData == null || !gchData.IsAllocated)
          gchData = GCHandle.Alloc(idata, GCHandleType.Pinned);
        gchDataObject = idata;
        tdts.data = gchData.AddrOfPinnedObject();
        gchRef = data;
        this.data = data;
        return;
      }
      if (data is DSPECTRUM)
      { // not an array !
        DSPECTRUM spc = (DSPECTRUM)data;
        byte[] bd = spc.ToByteArray();
        if (gchData == null || !gchData.IsAllocated) gchData = GCHandle.Alloc(bd, GCHandleType.Pinned);
        this.data = data;
        gchRef = bd;
        tdts.data = gchData.AddrOfPinnedObject();
        tdts.dArrayLength = (uint)spc.GetLength(); // length = embedded float array length
        return;
      }
      if (data is Array)
      { // whatever the rank, do it this way ...
        gchData = GCHandle.Alloc(data, GCHandleType.Pinned);
        tdts.data = gchData.AddrOfPinnedObject();
      }
      else
      {
        tdts.data = GetTineDataPtr(data);
      }
      tdts.dTag = new char[Constants.TAG_NAME_SIZE];
      if (fmt == Formats.CF_STRUCT)
      {
        Type dtype = data.GetType();
        string cname = getStructName(data);
        if (cname == null) return;
        char[] ch = cname.ToCharArray();
        for (int i=0; i<ch.Length && i<Constants.TAG_NAME_SIZE; i++) tdts.dTag[i] = ch[i];
        gchRef = data;
      }
      else if (gchData == null || !gchData.IsAllocated)
      {
        gchData = GCHandle.Alloc(data, GCHandleType.Pinned);
      }
      this.data = data;
    }
    private static int GetSizeOfStringData(String[] strArray)
    {
      int siz = 128 + IntPtr.Size;
      for (int i = 0; strArray != null && i < strArray.Length; i++)
      {
        siz += 128;
        if (strArray[i] != null && strArray[i].Length > 128) siz += strArray[i].Length;
        siz += IntPtr.Size;
      }
      return siz;
    }
    public TDataType()
    {
      SetAsNullType();
      return;
    }
    public TDataType(Array data)
    {
      if (data == null)
      {
        SetAsNullType();
        return;
      }
      int fmt = GetTineType(data);
      if (fmt == Formats.CF_NULL)
      {
        SetAsNullType();
        return;
      }
      tdts.dFormat = (short)(Formats.LFMT_OFFSET + fmt);
      tdts.dArrayLength = GetTineArrayLength(data);
      tdts.dTag = new char[Constants.TAG_NAME_SIZE];

      if (fmt == Formats.CF_STRUCT)
      {
        tdts.data = GetTineDataPtr(data);
        Type dtype = data.GetType();
        string cname = getStructName(data);
        if (cname == null) return;
        char[] ch = cname.ToCharArray();
        for (int i = 0; i < ch.Length && i < Constants.TAG_NAME_SIZE; i++) tdts.dTag[i] = ch[i];
        gchRef = data;
      }
      else if (fmt == Formats.CF_STRING)
      {
        tdts.data = GetIntPtrFromManagedDataObject(data, GetSizeOfStringData((String[])data));
        gchRef = data;
      }
      else if (fmt == Formats.CF_BOOLEAN)
      {
        Int32[] idata = new int[tdts.dArrayLength];
        if (gchData == null || !gchData.IsAllocated)
          gchData = GCHandle.Alloc(idata, GCHandleType.Pinned);
        tdts.data = gchData.AddrOfPinnedObject();
        gchRef = data;
      }
      else
      {
        gchData = GCHandle.Alloc(data, GCHandleType.Pinned);
        tdts.data = gchData.AddrOfPinnedObject();
      }
      this.data = data;
    }
    public TDataType(Array data, int offset, int length)
    { // this probably won't work ...
      if (data == null)
      {
        SetAsNullType();
        return;
      }
      int fmt = GetTineType(data);
      if (fmt == Formats.CF_NULL)
      {
        SetAsNullType();
        return;
      }
      if (data.Rank > 1)
      {
        throw new Exception("arrays with more than 2 dimensions not supported with this call!");
      }
      if (offset >= data.GetLength(0))
      {
        throw new Exception("offset is larger than the length !");
      }
      IntPtr dp = IntPtr.Zero;
      if (fmt == Formats.CF_STRING)
      {
        dp = GetIntPtrFromManagedDataObject(data, GetSizeOfStringData((String[])data));
        gchRef = data;
      }
      else if (fmt == Formats.CF_BOOLEAN)
      {
          Int32[] idata = new int[tdts.dArrayLength];
          if (gchData == null || !gchData.IsAllocated)
            gchData = GCHandle.Alloc(idata, GCHandleType.Pinned);
          dp = gchData.AddrOfPinnedObject();
          gchRef = data;
      }
      else
      {
        gchData = GCHandle.Alloc(data, GCHandleType.Pinned);
        dp = gchData.AddrOfPinnedObject();
      }
      int lfmt = (Formats.LFMT_OFFSET + fmt);
      string tag = "";
      char[] ctag = new char[Constants.TAG_NAME_SIZE];
      if (fmt == Formats.CF_STRUCT)
      {
        Type dtype = data.GetType();
        string cname = getStructName(data);
        if (cname == null) return;
        char[] ch = cname.ToCharArray();
        for (int i = 0; i < ch.Length && i < Constants.TAG_NAME_SIZE; i++) ctag[i] = ch[i];
        gchRef = data;
        tag = new string(ctag);
      }
      native.AssignDataToDTYPE(ref tdts, dp, offset, lfmt, length, tag);
      this.data = data;
    }    
    private TDataType(Array data,int row,int offset,int length)
    { // this probably won't work ...
      if (data == null)
      {
        SetAsNullType();
        return;
      }
      int fmt = GetTineType(data);
      if (fmt == Formats.CF_NULL)
      {
        SetAsNullType();
        return;
      }
      if (!Formats.IsPrimitive(fmt))
      {
        if (row != 0 || offset != 0)
        {
          throw new Exception("Cannot parse a non-primitive data type !");
        }
      }
      if (data.Rank > 2)
      {
        throw new Exception("arrays with more than 2 dimensions not supported!");
      }
      int rowlen = data.GetLength(row);
      if (row >= data.Rank) row = data.Rank - 1;
      if (length > rowlen) length = rowlen;
      if (offset >= length)
      {
        throw new Exception("offset is larger than the row length !");
      }
      tdts.dFormat = (short)(Formats.LFMT_OFFSET + fmt);
      tdts.dArrayLength = GetTineArrayLength(data);
      if (length < tdts.dArrayLength) tdts.dArrayLength = (UInt32)length;
      gchData = GCHandle.Alloc(data, GCHandleType.Pinned);
      tdts.data = gchData.AddrOfPinnedObject();
      int p = row * rowlen + offset;
//      tdts.data = tdts.data + (IntPtr)p;

      tdts.dTag = new char[Constants.TAG_NAME_SIZE];
      if (fmt == Formats.CF_STRUCT)
      {
        Type dtype = data.GetType();
        string cname = getStructName(data);
        if (cname == null) return;
        char[] ch = cname.ToCharArray();
        for (int i = 0; i < ch.Length && i < Constants.TAG_NAME_SIZE; i++) tdts.dTag[i] = ch[i];
        gchRef = data;
      }
      this.data = data;
    }
    public TDataType(DTYPE dt)
    {
      tdts.dFormat = dt.dFormat;
      tdts.dArrayLength = dt.dArrayLength;
      tdts.dStamp = dt.dStamp;
      tdts.dTimeStamp = dt.dTimeStamp;
      tdts.sysStamp = dt.sysStamp;
      tdts.dTag = new char[Constants.TAG_NAME_SIZE];
      if (dt.dTag != null)
      {
        Int32 siz = Constants.TAG_NAME_SIZE;
        if (dt.dTag.Length < siz) siz = dt.dTag.Length;
        Array.Copy(dt.dTag, tdts.dTag, siz);
      }
      tdts.data = dt.data; // the underlying pointer
      this.data = dt.data; // the app's pointer
    }
    public TDataType(DTYPE dt,object dobj)
    {
      tdts.dFormat = dt.dFormat;
      tdts.dArrayLength = dt.dArrayLength;
      tdts.dStamp = dt.dStamp;
      tdts.dTimeStamp = dt.dTimeStamp;
      tdts.sysStamp = dt.sysStamp;
      tdts.dTag = new char[Constants.TAG_NAME_SIZE];
      if (dt.dTag != null)
      {
        Int32 siz = Constants.TAG_NAME_SIZE;
        if (dt.dTag.Length < siz) siz = dt.dTag.Length;
        Array.Copy(dt.dTag, tdts.dTag, siz);
        gchRef = dobj;
      }
      tdts.data = dt.data;
      this.data = dt.data;
    }
    public TDataType(int dataSize, byte dataFmt, string dataTag)
    {
      tdts.dFormat = (short)(Formats.LFMT_OFFSET + dataFmt);
      tdts.dArrayLength = (uint)dataSize;
      tdts.dTag = new char[Constants.TAG_NAME_SIZE];
      if (dataTag != null)
      {
        Int32 siz = Constants.TAG_NAME_SIZE;
        if (dataTag.Length < siz) siz = dataTag.Length;
        Array.Copy(dataTag.ToCharArray(), tdts.dTag, siz);
      }
    //  tdts.data = IntPtr.Zero;
    //  this.data = IntPtr.Zero;
      data = new byte[dataSize * Formats.GetSizeInBytes(dataFmt)];
      gchData = GCHandle.Alloc(data, GCHandleType.Pinned);
      tdts.data = GetTineDataPtr(data);
    }
    ~TDataType()
    {
      GCHandle? gch = null;
      if (data is DIMAGE) gch = ((DIMAGE)data).gchFrame;
      if (data is DIMAGE[] && ((DIMAGE[])data).Length == 1) gch = ((DIMAGE[])data)[0].gchFrame;
      if (gch.HasValue) gch.Value.Free();
      if (gchData != null && gchData.IsAllocated) gchData.Free();
    }
    public Int32 GetData(ref Int16 data) 
    {
      Int16[] v = new Int16[1];
      Int32 cc = native.GetValuesAsShort(ref tdts,v, 1);
      data = v[0];
      return cc;
    }
    public Int32 GetData(ref UInt16 data) 
    {
      Int16[] sdata = new Int16[1];
      Int32 cc = native.GetValuesAsShort(ref tdts,sdata, 1);
      data = (UInt16)sdata[0];
      return cc;
    }
    public Int32 GetData(ref Int32 data) 
    { 
      Int32[] v = new Int32[1];
      Int32 cc = native.GetValuesAsLong(ref tdts, v, 1);
      data = v[0];
      return cc;
    }
    public Int32 GetData(ref UInt32 data) 
    {
      Int32[] sdata = new Int32[1];
      Int32 cc = native.GetValuesAsLong(ref tdts, sdata, 1);
      data = (UInt32)sdata[0];
      return cc;
    }
    public Int32 GetData(ref float data) 
    {
      float[] v = new float[1];
      Int32 cc = native.GetValuesAsFloat(ref tdts, v, 1);
      data = v[0];
      return cc;
    }
    public Int32 GetData(ref double data) 
    {
      double[] v = new double[1];
      Int32 cc = native.GetValuesAsDouble(ref tdts, v, 1);
      data = v[0];
      return cc;
    }
    public Int32 GetData(Object data)
    {
      if (data is Array)
      { // handle arrays here ...
        return GetData(data, ((Array)data).Length, 0);
      }
      int cc = Errors.illegal_format;
      int fmt = GetTineType(data);
      Type dtype = data.GetType();
      if (!dtype.IsByRef && fmt != Formats.CF_IMAGE) return Errors.invalid_parameter;

      switch (fmt)
      {
        case Formats.CF_STRUCT:
          string cname = getStructName(data);
          int ssiz;
          if ((ssiz = native.GetStructSize(cname)) > 0)
          {
            if (new string(tdts.dTag).CompareTo(cname) != 0) return Errors.invalid_structure_tag;
            try
            {
              TTaggedStruct tts = data as TTaggedStruct;
              if (tts.GetClassName() != null)
              {
                data = Marshal.PtrToStructure(tdts.data, dtype);
              }
            }
            catch (Exception) { }
          }
          cc = 0;
          break;
        case Formats.CF_IMAGE:
          data = (TImageHeader)Marshal.PtrToStructure(tdts.data, dtype);
          image.SetFrameHeader((TImageHeader)data);
          if (this.data is Array) ((DIMAGE[])this.data)[0] = image;
          else this.PutData(image);
          cc = 0;
          break;
        case Formats.CF_BOOLEAN:
          if (data is Boolean[]) // hmmm ... I can't ever be here ...  (arrays go thru another path)
          {
            Boolean[] bdata = (Boolean[])data;
            int len = bdata.Length;
            Int32[] idata = new Int32[len];
            IntPtr value = GetTineDataPtr(idata);
            cc = native.GetValuesAsAny(ref tdts, value, (short)(Formats.LFMT_OFFSET + fmt), 4, 1, 0);
            for (int i = 0; i < len; i++) bdata[i] = idata[i] != 0;
            break;
          }
          goto default;
        default:
          {
            int siz = Formats.GetSizeInBytes(fmt);
            IntPtr value = GetTineDataPtr(data);
            cc = native.GetValuesAsAny(ref tdts, value, (short)(Formats.LFMT_OFFSET + fmt), siz, 1, 0);
            break;
          }
      }
      return cc;
    }
    public Int32 GetData(Object data,int length,int offset)
    {
      if (length <= 0) return Errors.dimension_error;
      if (!(data is Array)) return Errors.dimension_error;
      if (((Array)data).Rank != 1) return Errors.dimension_error;
      int cc = Errors.illegal_format;
      int fmt = GetTineType(data);

      switch (fmt)
      {
        case Formats.CF_BYTE:
          { // check for some special cases before hand ...
            if (tdts.dFormat == (Formats.CF_SPECTRUM+Formats.LFMT_OFFSET))
            {
              DSPECTRUM spc = (DSPECTRUM)(this.data);
              spc.ToStruct((byte[])data);
              break;
            }
            if (((Array)data).Rank != 1) return Errors.dimension_error;
            if (offset == 0) return native.GetValuesAsByte(ref tdts, (byte[])data, length);
            byte[] values = new byte[length];
            cc = native.GetValuesAsByte(ref tdts, values, length);
            System.Array.Copy(values, 0, (byte[])data, offset, length);
          }
          break;
        case Formats.CF_CHAR:
          {
            if (((Array)data).Rank != 1) return Errors.dimension_error;
            if (offset == 0) return native.GetValuesAsStringEx(ref tdts, (char[])data, Formats.CF_CHAR, length, 0);
            char[] values = new char[length];
            cc = native.GetValuesAsStringEx(ref tdts, values, Formats.CF_CHAR, length, 0);
            System.Array.Copy(values, 0, (char[])data, offset, length);
          }
          break;
        case Formats.CF_INT16:
          {
            if (offset == 0) return native.GetValuesAsShort(ref tdts, (Int16[])data, length);
            Int16[] values = new Int16[length];
            cc = native.GetValuesAsShort(ref tdts, values, length);
            System.Array.Copy(values, 0, (Int16[])data, offset, length);
          }
          break;
        case Formats.CF_INT32:
          {
            if (offset == 0) return native.GetValuesAsLong(ref tdts, (Int32[])data, length);
            Int32[] values = new Int32[length];
            cc = native.GetValuesAsLong(ref tdts, values, length);
            System.Array.Copy(values, 0, (Int32[])data, offset, length);
          }
          break;
        case Formats.CF_INT64:
          {
            IntPtr values = GetTineDataPtr(data);
            cc = native.GetValuesAsAny(ref tdts, values, (short)(Formats.LFMT_OFFSET + Formats.CF_INT64), 8, length, offset);
          }
          break;
        case Formats.CF_FLOAT:
          {
            if (offset == 0) return native.GetValuesAsFloat(ref tdts, (float[])data, length);
            float[] values = new float[length];
            cc = native.GetValuesAsFloat(ref tdts, values, length);
            System.Array.Copy(values, 0, (float[])data, offset, length);
          }
          break;
        case Formats.CF_DOUBLE:
          {
            if (offset == 0) return native.GetValuesAsDouble(ref tdts, (double[])data, length);
            double[] values = new double[length];
            cc = native.GetValuesAsDouble(ref tdts, values, length);
            System.Array.Copy(values, 0, (double[])data, offset, length);
          }
          break;
        case Formats.CF_STRUCT:
          string cname = getStructName(data);
          int ssiz;
          if ((ssiz = native.GetStructSize(cname)) > 0)
          {
            if (new string(tdts.dTag).CompareTo(cname) != 0) return Errors.invalid_structure_tag;
            int n = 0, alen = ((Array)data).Length;
            object obj;
            IntPtr p = tdts.data;
            try
            {
              for (int i = 0; i < alen; i++)
              {
                obj = ((Array)data).GetValue((i + offset) % length);
                if (obj is TTaggedStruct)
                {
                  TTaggedStruct tts;
                  tts = ((Array)data).GetValue((i + offset) % length) as TTaggedStruct;
                  if (tts != null && tts.GetClassName() != null) break;
                }
                ((Array)data).SetValue(Marshal.PtrToStructure(p, obj.GetType()), i);
                p = (IntPtr)((long)p + ssiz);
                if (n++ >= tdts.dArrayLength) break;
              }
            }
            catch (Exception) { }
          }
          cc = 0;
          break;
        case Formats.CF_STRING:
          unsafe {
            IntPtr values = tdts.data; //GetTineDataPtr(data);
            //cc = native.GetValuesAsAny(ref tdts, values, (short)(Formats.LFMT_OFFSET + fmt), 1, length, offset);
            String[] sa = (String[])data;
            void* vp = values.ToPointer();
            sbyte **bp = (sbyte **)vp; // have to use bytes as the underlying native strings are ascii chars 
            sbyte *bc;
            for (int i = 0; i < sa.Length && i < tdts.dArrayLength; i++)
            {
              if (*bp == null) break;
              bc = *bp++;
              if (bc == null) break;
              sa[i] = new String(bc);
            }
            cc = 0;
            break;
          }
        default:
          {
            int siz = Formats.GetSizeInBytes(fmt);
            IntPtr values = GetTineDataPtr(data);
            cc = native.GetValuesAsAny(ref tdts, values, (short)(Formats.LFMT_OFFSET + fmt), siz, length, offset);
            break;
          }
      }
      return cc;
    }
    public Int32 PutData(DIMAGE data)
    {
      this.data = data;
      return 0;
    }
    public Int32 PutData()
    {
      if (gchRef == null) return 0;
      return PutData(gchRef,(int)tdts.dArrayLength,0);
    }
    public Int32 PutData(Object data)
    {
      int len = 1;
      if (data is Array) len = ((Array)data).GetLength(0);
      return PutData(data, len, 0);
    }
    public Int32 PutData(Object data,int length,int offset)
    {
      int cc = Errors.illegal_format;
      int fmt = GetTineType(data);
      switch (fmt)
      {
        case Formats.CF_BYTE:
          if (data is Array)
          {
            if (((Array)data).Rank != 1) return Errors.dimension_error;
            cc = native.PutValuesFromByteEx(ref tdts, (byte[])data, length, offset);
          }
          break;
        case Formats.CF_CHAR:
          if (data is Array)
          {
            if (((Array)data).Rank != 1) return Errors.dimension_error;
            sbyte[] sbdata = (sbyte[])data;
            char[] cdata = new char[sbdata.Length];
            for (int i = 0; i < sbdata.Length; i++) cdata[i] = (char)sbdata[i];
            string values = new String(cdata);
            cc = native.PutValuesFromString(ref tdts, values, Formats.LFMT_OFFSET+Formats.CF_CHAR, length, offset);
          }
          break;
        case Formats.CF_INT16:
          if (data is Array)
          {
            if (((Array)data).Rank != 1) return Errors.dimension_error;
            if (data is Int16[])
            {
              cc = native.PutValuesFromShortEx(ref tdts, (Int16[])data, length, offset);
            }
            else if (data is UInt16[])
            {
              cc = native.PutValuesFromUnsignedShortEx(ref tdts, (UInt16[])data, length, offset);
            }
          }
          else
          {
            if (data is Int16)
            {
              Int16[] value = {(Int16)data};
              cc = native.PutValuesFromShortEx(ref tdts, value, 1, 0);
            }
            else if (data is UInt16)
            {
              UInt16[] value = {(UInt16)data};
              cc = native.PutValuesFromUnsignedShortEx(ref tdts, value, 1, 0);
            }
          }
          break;
        case Formats.CF_INT32:
          if (data is Array)
          {
            if (((Array)data).Rank != 1) return Errors.dimension_error;
            if (data is Int32[])
            {
              cc = native.PutValuesFromLongEx(ref tdts, (Int32[])data, length, offset);
            }
            else if (data is UInt32[])
            {
              cc = native.PutValuesFromUnsignedLongEx(ref tdts, (UInt32[])data, length, offset);
            }
          }
          else
          {
            if (data is Int32)
            {
              Int32[] value = {(Int32)data};
              cc = native.PutValuesFromLongEx(ref tdts, value, 1, 0);
            }
            else if (data is UInt32)
            {
              UInt32[] value = {(UInt32)data};
              cc = native.PutValuesFromUnsignedLongEx(ref tdts, value, 1, 0);
            }
          }
          break;
        case Formats.CF_INT64:
          if (data is Array)
          {
            if (((Array)data).Rank != 1) return Errors.dimension_error;
            IntPtr values = GetTineDataPtr(data);
            cc = native.PutValuesFromAny(ref tdts, values, (short)(Formats.LFMT_OFFSET + Formats.CF_INT64), 1, 8, length, offset);
          }
          break;
        case Formats.CF_FLOAT:
          if (data is Array)
          {
            if (((Array)data).Rank != 1) return Errors.dimension_error;
            cc = native.PutValuesFromFloatEx(ref tdts, (float[])data, length, offset);
          }
          else
          {
            float[] value = { (float)data };
            cc = native.PutValuesFromFloatEx(ref tdts, value, 1, 0);
          }
          break;
        case Formats.CF_DOUBLE:
          if (data is Array)
          {
            if (((Array)data).Rank != 1) return Errors.dimension_error;
          //  double[] values = (double[])data;
            cc = native.PutValuesFromDoubleEx(ref tdts, (double[])data, length, offset);
          }
          else
          {
            double[] value = { (double)data };
            cc = native.PutValuesFromDoubleEx(ref tdts, value, 1, 0);
          }
          break;
        case Formats.CF_STRUCT:
          Type dtype = data.GetType();
          string cname = getStructName(data);
          int ssiz;
          if ((ssiz = native.GetStructSize(cname)) > 0)
          {
            if (new string(tdts.dTag).CompareTo(cname) != 0) return Errors.invalid_structure_tag;
            object obj;
            if (dtype.IsArray)
            {
              int n = 0, alen = ((Array)data).Length;
              IntPtr p = tdts.data;
              for (int i = 0; i < alen; i++)
              {
                obj = ((Array)data).GetValue((i + offset) % length);
                if (obj is TTaggedStruct)
                {
                  TTaggedStruct tts;
                  tts = obj as TTaggedStruct;
                  if (tts != null && tts.GetClassName() == null)
                  {
                    n = (int)tdts.dArrayLength;
                    break;
                  }
                }
                Marshal.StructureToPtr(obj, p, true);
                p = (IntPtr)((long)p + ssiz);
                if (n++ >= tdts.dArrayLength) break;
              }
              if (n < length) length = n;
            }
            else
            {
              Marshal.StructureToPtr(data, tdts.data, true);
            }
          }
          cc = 0;
          break;
        default:
          int siz = Formats.GetSizeInBytes(fmt);
          if (data is Array)
          {
            if (((Array)data).Rank != 1) return Errors.dimension_error;
            IntPtr values = GetTineDataPtr(data);
            cc = native.PutValuesFromAny(ref tdts, values, (short)(Formats.LFMT_OFFSET + fmt), 1, siz, length, offset);
          }
          else
          {
            IntPtr value = GetTineDataPtr(data);
            cc = native.PutValuesFromAny(ref tdts, value, (short)(Formats.LFMT_OFFSET + fmt), 1, siz, length, offset);
          }
          break;
      }
      if (length < tdts.dArrayLength) tdts.dArrayLength = (UInt32)length;
      return cc;
    }
    public Int32 PutData(Object data, int row, int length, int offset)
    {
      int cc = Errors.illegal_format;
      if (!(data is Array)) return Errors.dimension_error;
      if (((Array)data).Rank != 2) return Errors.dimension_error;
      switch (GetTineType(data))
      {
        case Formats.CF_BYTE:
          {
            byte[] values = new byte[length];
            for (int i = 0; i < length; i++) values[i] = ((byte[,])data)[row, i];
            cc = native.PutValuesFromByteEx(ref tdts, values, length, offset);
          }
          break;
        //case Formats.CF_CHAR:
        //  {
        //    char[,] values = (char[,])data;
        //    string s = new String(values[row, 0]);
        //    cc = native.PutValuesFromString(ref tdts, s, Formats.CF_CHAR, length, offset);
        //  }
        //  break;
        case Formats.CF_INT16:
          if (data is Int16[,])
          {
            Int16[] values = new Int16[length];
            for (int i = 0; i < length; i++) values[i] = ((Int16[,])data)[row, i];
            cc = native.PutValuesFromShortEx(ref tdts, values, length, offset);
          }
          else if (data is UInt16[,])
          {
            UInt16[] values = new UInt16[length];
            for (int i = 0; i < length; i++) values[i] = ((UInt16[,])data)[row, i];
            cc = native.PutValuesFromUnsignedShortEx(ref tdts, values, length, offset);
          }
          break;
        case Formats.CF_INT32:
          if (data is Int32[,])
          {
            Int32[] values = new Int32[length];
            for (int i = 0; i < length; i++) values[i] = ((Int32[,])data)[row, i];
            cc = native.PutValuesFromLongEx(ref tdts, values, length, offset);
          }
          else if (data is UInt32[,])
          {
            UInt32[] values = new UInt32[length];
            for (int i = 0; i < length; i++) values[i] = ((UInt32[,])data)[row, i];
            cc = native.PutValuesFromUnsignedLongEx(ref tdts, values, length, offset);
          }
          break;
        case Formats.CF_INT64:
          cc = Errors.not_implemented;
          //{
          //  IntPtr values = GetTineDataPtr(data);
          //  cc = native.PutValuesFromAny(ref tdts, values, (short)(Formats.LFMT_OFFSET + Formats.CF_INT64), 1, 8, length, offset);
          //}
          break;
        case Formats.CF_FLOAT:
          {
            float[] values = new float[length];
            for (int i = 0; i < length; i++) values[i] = ((float[,])data)[row, i];
            cc = native.PutValuesFromFloatEx(ref tdts, values, length, offset);
          }
          break;
        case Formats.CF_DOUBLE:
          {
            double[] values = new double[length];
            for (int i = 0; i < length; i++) values[i] = ((double[,])data)[row, i];
            cc = native.PutValuesFromDoubleEx(ref tdts, values, length, offset);
          }
          break;
        default:
          break;
      }
      if (length < tdts.dArrayLength) tdts.dArrayLength = (UInt32)length;
      return cc;
    }
    public double GetDataTimeStamp() { return tdts.dTimeStamp; }
    public void SetDataTimeStamp(double timestamp) { tdts.dTimeStamp = timestamp; }
    public Int32 GetSystemStamp() { return tdts.sysStamp; }
    public void SetSystemStamp(Int32 sysStamp) { tdts.sysStamp = sysStamp; }
    public Int32 GetUserStamp() { return tdts.dStamp; }
    public void SetUserStamp(Int32 userStamp) { tdts.dStamp = userStamp; }
    public UInt32 GetDataArrayLength() { return tdts.dArrayLength; }
    private bool isBounded = false;
    public void SetBounded(Boolean value) { isBounded = value; }
    public Boolean GetBounded() { return isBounded; }
    public void SetDataArrayLength(UInt32 len) 
    {
      if (isBounded && len > tdts.dArrayLength) return; // don't allow !
      tdts.dArrayLength = len; 
    }
    public int GetDataFormat() { return tdts.dFormat%256; }
    public string GetDataTag() 
    {
      int n = 0;
      foreach (char c in tdts.dTag) if (c != 0) n++; else break;
      return tdts.dTag != null ? new string(tdts.dTag,0,n) : ""; 
    }
    public void SetDataTag(string tag)
    {
      for (int i=0; i<Constants.TAG_NAME_SIZE; i++) tdts.dTag[i] = (char)0;
      if (tag == null) return;
      char[] tagch = tag.ToCharArray();
      for (int i=0; i<Constants.TAG_NAME_SIZE && i<tagch.Length; i++) tdts.dTag[i] = tagch[i];
      return;
    }
    private static List<TTaggedStruct> structLst = new List<TTaggedStruct>();
    public static int RegisterStructure(object tstruct)
    {
      structLst.Add(new TTaggedStruct(tstruct));
      return 0;
    }
  }
  public class native
  {
    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern Int32 LaunchRuntimeDebugWindow();

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern void SystemSetCycleTimer();

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern Int32 _SystemInit(Int32 netChk);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern Int32 _SystemCycle(Int32 chkCmd);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int AssignImageFrameBuffer(ref TImageHeader img, IntPtr buf, int buflen);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int ExecLinkEx(string devName, string devProperty, ref DTYPE dout, ref DTYPE din, Int16 access, UInt16 timeout);

    [UnmanagedFunctionPointer(CallingConvention.Cdecl)]
    public delegate void TLinkCallback(int id, int cc, IntPtr cbRef);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int AttachLinkEx2(string devName,string devProperty,ref DTYPE dout,ref DTYPE din,
         Int16 access, Int32 pollingRate, TLinkCallback cb, Int32 mode, UInt32 cbID, IntPtr cbRef);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern void CloseLink(Int32 lnkId);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int GetDataFromLinkId(int linkId, ref DTYPE dout, ref UInt16 status);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern double GetCurrentDataTimeStamp(int linkId);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern double GetCurrentDataTimeStampFromCallbackId(int linkId);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int GetSystemDataStamp(int linkId);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int GetSystemDataStampFromCallbackId(int linkId);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int GetDataStamp(int linkId);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int GetDataStampFromCallbackId(int linkId);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int GetCompletionDataSize(int linkId);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int GetCompletionDataSizeFromCallbackId(int cbId);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int GetCompletionDataType(int linkId);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int GetCompletionDataTypeFromCallbackId(int cbId);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int GetCurrentLinkStatus(int linkId);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int SetAccessLock(string context, string server, int lockType, int lockDuration);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int FreeAccessLock(string context, string server);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int GetAccessLockStatus(string context, string server);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int GetAccessLockInformation(string context, string server, IntPtr callerName, IntPtr callerIp, IntPtr timeLeft);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int FillCompletionStatusString(int linkId, sbyte[] errstr);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int GetCurrentLinkStatusFromCallbackId(int cbId);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern IntPtr GetLastLinkError(short cc, sbyte[] errstr);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern IntPtr GetSystemUserName();

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int _SystemSchedulePropertyEx(string eqm, string prp, Int32 scope);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern UInt32 SetWorkAreaSize(UInt32 size);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern Int32 GetFormatSize(Int32 fmt);

    [UnmanagedFunctionPointer(CallingConvention.Cdecl)]
    public delegate Int32 TEqmFcn(string dev,string prp,ref DTYPE dout,ref DTYPE din,UInt16 acc,IntPtr eqmRef);

    [UnmanagedFunctionPointer(CallingConvention.Cdecl)]
    public delegate void TEqmHook(IntPtr cbRef);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int RegisterEquipmentModuleEx(string expName,string eqmName,UInt32 numdevices,
      TEqmFcn fcn, TEqmHook ini, TEqmHook tsk, Int32 rate, TEqmHook exi, IntPtr eqmRef);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int RegisterFecInformation(string name, string sub, string cntxt, 
      string dsc, string loc, string hdw, string resp, UInt16 poff);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int GetPortOffset(string fecName);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int RegisterPropertyInformation(string eqm, string property, ref DTYPE dout, ref DTYPE din,
                                       UInt16 acc, UInt16 atype, UInt16 rowlength, string dsc, Int32 propId, string rdr);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int RedirectProperty(string eqm,string property,string rdr);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int RegisterDeviceName(string eqm,string devname,Int32 devnr);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int SetDeviceDescription(string eqm, string devname, string description);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int SetDeviceLocation(string eqm, string devname, string description);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int SetDeviceMask(string eqm, string devname, int mask);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int SetDeviceOfflineStatus(string eqm, string devname, int offline);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int SetDeviceZPosition(string eqm, string devname, float zpos);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int SetDeviceRegion(string eqm, string devname, string region);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int RedirectDeviceName(string eqm, string devname, string rdr);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern void SetHistoryFilesRepository(string path);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern void SetAppVersion(int major, int minor, int revision);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern void SetAppDate(IntPtr appdate);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int GetPropertyId(string eqm, string prpName);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int GetDeviceNumberEx(string eqm, string devname, string prpname);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int GetDeviceNumber(string eqm, string devname);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int GetCallerInformation(string eqm, ref ClnInfo clnInfoList, ref Int32 num);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern void SetEqmCompletion(string eqm,string stsstr);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int AddFieldToStruct(string tag, Int32 addr, Int32 size, Int32 fmt, string field);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int SealTaggedStruct(string tag, Int32 size, Int32 number);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int GetStructSize(string tag);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int GetStructFormat(string tag, IntPtr sfn64, ref Int32 num);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int AcquireAndRegisterStruct(string devName, string tag,int num);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int AssignNumDevices(string eqm, Int32 num);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int AssignDeviceListToProperty(string eqm, string prp, ref NAME64 devlst, Int32 lstlen);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int feclog(string txt);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern void SetFeclogDepth(int value);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int GetFeclogDepth();

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern UInt32 GetWorkAreaSize();

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern Int32 _SystemAssignBufferSpaceEx(UInt32 clnRcvBufferSpace, UInt32 srvRcvBufferSpace, UInt32 clnSndBufferSpace, UInt32 srvSndBufferSpace);

    //[DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    //public static extern void setToThree(ref float v);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int srvlog(string filename, string tag, string txt);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int PutValuesFromFloatEx(ref DTYPE d, float[] val, Int32 num, Int32 offset);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int PutValuesFromDoubleEx(ref DTYPE d, double[] val, Int32 num, Int32 offset);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int PutValuesFromLongEx(ref DTYPE d, Int32[] val, Int32 num, Int32 offset);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int PutValuesFromUnsignedLongEx(ref DTYPE d, UInt32[] val, Int32 num, Int32 offset);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int PutValuesFromShortEx(ref DTYPE d, Int16[] val, Int32 num, Int32 offset);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int PutValuesFromUnsignedShortEx(ref DTYPE d, UInt16[] val, Int32 num, Int32 offset);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int PutValuesFromByteEx(ref DTYPE d, byte[] val, Int32 num, Int32 offset);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int PutValuesFromString(ref DTYPE d, string str, Int32 fmt, Int32 num, Int32 offset);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int PutValuesFromAny(ref DTYPE d, IntPtr val, Int16 fmt, Int32 sgn, Int32 objSizeInBytes, Int32 num, Int32 offset);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int GetValuesAsDouble(ref DTYPE d, double[] val, int num);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int GetValuesAsFloat(ref DTYPE d, float[] val, int num);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int GetValuesAsLong(ref DTYPE d, Int32[] val, int num);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int GetValuesAsShort(ref DTYPE d, Int16[] val, int num);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int GetValuesAsByte(ref DTYPE d, byte[] val, int num);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int GetValuesAsStringEx(ref DTYPE d, char[] val, int fmt, int num, int offset);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int GetValuesAsAny(ref DTYPE d, IntPtr val, Int16 fmt, Int32 objSizeInBytes, Int32 num, Int32 offset);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int GetValuesArrayAsString(ref DTYPE d, char[] val, ref int siz, string sep);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int GetStructureAsString(string tag, IntPtr structref, [In, Out]char[] stringbuf, Int32 length);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int AssignDataToDTYPE(ref DTYPE d, IntPtr vals, Int32 offset, Int32 format, Int32 length, string tag);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern void SetDataTimeStamp(double ts);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern double GetDataTimeStamp();

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern double MakeDataTimeStamp();

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern void SetSystemDataStamp(int value);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int MakeSystemDataStamp();

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int GetUserStamp(int linkId);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int SetAlarmEx(string eqm, Int16 devNr, Int32 almCode, IntPtr almData, byte flags);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int SetDeviceAlarm(string eqm, string devname, Int32 almCode, IntPtr almData, byte flags);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern void RemoveDeviceAlarm(string eqm, string devname, Int32 code);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern void ClearDeviceAlarm(string eqm, string devname);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern void ClearAlarm(string eqm, Int16 devNr);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int AppendAlarmWatchTable(string eqm, string prp, string dev, Int32 siz, Int32 fmt, Int32 atyp, Int32 sev, Int32 sys, ref ALM_THRESHOLDS thr);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int AppendAlarmInfoTable(string eqm, ref ADS ads);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int AppendHistoryInformation(string eqm, string prp, string dev, Int32 len, Int32 fmt, Int32 idx, ref HistorySpecification hspec);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int AppendHistoryData(string eqm, string prp, string dev, ref DTYPE d);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int GetSystemContexts(IntPtr clist, ref int num);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int GetSystemServers(string context, IntPtr slist, ref int num);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int GetSystemProperties(string server,IntPtr plist,ref int num);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern short IsStockProperty(string property);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int GetSystemDevices(string server,string property,IntPtr dlist,ref int num);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int GetServerAddr(string context, string expName, ref ExpDataStruct srvaddr, ref FecAddrStruct fecaddr);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int GetServerCycleState(sbyte[] cycleStateString);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern void SetMinimumDiskSpaceInBlocks(uint spaceInBlocks);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern void SetDieOnAddressInUse(int value);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern void SetDieOnSocketError(int value);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern void SetDieOnFecIsAlias(int value);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int SetFecHome(string fecHomePath);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int SetFecLogPath(string path);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int IsStandAlone();

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int GetConnectionTable([In, Out]ConTblInfo[] infoTbl, ref int size);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern void SetConnectionTableCapacity(int value);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int GetMsgTable([In, Out]MsgType[] msgTbl, ref int size);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int GetMessageList([In, Out]char[] msgLst, int sizeInBytes);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int GetConnectionList([In, Out]char[] conLst, int sizeInBytes);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int GetArchiveAnnotation(string ctx, string srv, string prp, double target, [In, Out]char[] buf, int bufsiz, int timeout);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int SetArchiveAnnotation(string ctx, string srv, string prp, double target, string text, int keepAlive, int timeout);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int GetArchiveAnnotationTimes(string ctx, string srv, string prp, double start, double stop, IntPtr buf, int buffmt, ref int bufsiz, int timeout);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int SetNotificationToleranceEx(int linkId, float toleranceAbsolute, float tolerancePercent, int offset, int length, int flags);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern void SetSuspendCallbacks(int value);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern int GetSuspendCallbacks();

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern void SetRespondToServiceRequests(int value);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern void SetAllowNetworkAddressResolution(int value);

    [DllImport("tinemt", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
    public static extern IntPtr GetSystemVersionString();

    private static bool isInitialized = false;
    private static DTYPE nulltype = new DTYPE();
    private static void initializeSettings()
    { // make sure this happens on linux as well !
      if (isInitialized) return;
      nulltype.dFormat = Formats.CF_NULL;
      SystemSetCycleTimer();
      isInitialized = true;
    }
    public static unsafe int ExecLink(string devName, string devProperty, TDataType dout, TDataType din, Int16 access, UInt16 timeout)
    {
      if (!isInitialized) initializeSettings();
      DTYPE p1 = dout != null ? dout.GetDTYPE() : nulltype;
      DTYPE p2 = din != null ? din.GetDTYPE() : nulltype;
      if (din != null) // && din.GetDataFormat() == Formats.CF_STRUCT)
      {
        //din.PutData();
        din.PutData(din.GetDataObject());
      }
      //Int16 acc = (Int16)(access | Access.CA_MUTABLE);
      int cc = (int)ExecLinkEx(devName, devProperty, ref p1, ref p2, access, timeout);
      if (dout != null)
      {
        dout.SetUserStamp(p1.dStamp);
        dout.SetSystemStamp(p1.sysStamp);
        dout.SetDataTimeStamp(p1.dTimeStamp);
        dout.SetDataArrayLength(p1.dArrayLength);
      }
      return cc;
    }
    public static unsafe int AttachLink(string devName, string devProperty, TDataType dout, TDataType din,
      Int16 access, Int32 pollingInterval, TLinkCallback cb, Int32 mode, UInt32 cbID, IntPtr cbRef)
    {
      if (!isInitialized) initializeSettings();
      DTYPE p1 = dout != null ? dout.GetDTYPE() : nulltype;
      DTYPE p2 = din != null ? din.GetDTYPE() : nulltype;
      return (int)AttachLinkEx2(devName, devProperty, ref p1, ref p2, access, pollingInterval, cb, mode, cbID, cbRef);
    }

    native()
    {
      PermissionSet pSet = new PermissionSet(PermissionState.Unrestricted);
      pSet.AddPermission(new SecurityPermission(SecurityPermissionFlag.UnmanagedCode));
      pSet.AddPermission(new FileIOPermission(
          FileIOPermissionAccess.Write |
          FileIOPermissionAccess.Read |
          FileIOPermissionAccess.PathDiscovery, "*.*"));
      initializeSettings();
    }
    //private static TLink llnk;
    private static Dictionary<String,TLink> llnkset = new Dictionary<String,TLink>();
    public static void fixLink(TLink lnk)
    {
      if (llnkset.ContainsValue(lnk)) return;
      llnkset.Add("link:"+lnk.GetHashCode(), lnk);
      //llnk = lnk;
      //GCHandle.Alloc(cb, GCHandleType.Pinned);
    }
    public static void unfixLink(TLink lnk)
    {
      if (!llnkset.ContainsValue(lnk)) return;
      llnkset.Remove("link:" + lnk.GetHashCode());
    }
  }
  public class TLinkGroup
  {
    private static LinkedList<TLinkGroup> grpTbl = new LinkedList<TLinkGroup>();
    public static TLinkGroup GetGroup(TLink.TLinkCallback lcb)
    {
      foreach (TLinkGroup g in grpTbl)
      {
        if (g.lcb == lcb) return g;
      }
      TLinkGroup gn = new TLinkGroup(lcb);
      grpTbl.AddLast(gn);
      return gn;
    }
    public TLinkGroup(TLink.TLinkCallback lcb)
    {
      this.lcb = lcb;
      lcbdg = lclCb;
    }
    private LinkedList<TLink> members = new LinkedList<TLink>();
    public void AddMember(TLink lnk)
    {
      lock (members)
      {
        if (members.Contains(lnk)) return;
        members.AddLast(lnk);
      }
    }
    public void RemoveMember(TLink lnk)
    {
      lock (members)
      {
        if (!members.Contains(lnk)) return;
        members.Remove(lnk);
      }
    }
    private TLink.TLinkCallback lcb;
    public Boolean HasGroupCallback(TLink.TLinkCallback gcb)
    {
      if (lcb == null) return false;
      return gcb == lcb;
    }
    public TLink[] GetMembers()
    {
      lock (members)
      {
        if (members.Count == 0) return null;
        TLink[] mems = new TLink[members.Count];
        members.CopyTo(mems, 0);
        return mems;
      }
    }
    private Int32 lnkStatus;
    /// <summary>
    /// The current link status as an integer value.
    /// </summary>
    /// <returns>The current link status.</returns>
    public Int32 GetStatus() { return lnkStatus; }
    private Int32 cbId = 0; // for those who insist
    public Int32 GetCallbackId() { return cbId; }
    private IntPtr cbRef = IntPtr.Zero;  // for those who insist
    public IntPtr GetCallbackReference() { return cbRef; }
    private Int32 lnkId = -1;
    public Int32 GetLinkId() { return lnkId; }
    public native.TLinkCallback lcbdg;
    public unsafe void lclCb(Int32 id, Int32 cc, IntPtr cbRef)
    { // will only be called from below when they're all there !
      this.lnkStatus = cc;
      this.cbId = id;
      this.cbRef = cbRef;
      lock (members)
      {
        foreach (TLink m in members)
        {
          m.SetStatus(native.GetCurrentLinkStatus(m.GetLinkId()));
          m.memupd();
        }
        if (lcb != null && members.Count > 0) lcb(members.First.Value);
      }
    }
  }
  /// <summary>
  /// Defines a connection between the local client and
  /// the remote server's targeted end point
  /// </summary>
  /// <remarks>
  /// TLink is a TINE Link object which defines a connection between the local client and
  /// the remote server's targeted end point. Data may be exchanged either synchronously
  /// (via the Execute() method) or asynchronously (via the Attach() method).  Simple 
  /// commands which do not involve data can likewise be issued synchronously.
  /// </remarks>
  public class TLink
  {
    /// <summary>
    /// Combines at TINE access code and a TINE transfer mode code into a single 'link flags'
    /// value for use in the ACOP transport layer.
    /// </summary>
    /// <param name="access">A code or combination of codes from the Access constants</param>
    /// <param name="mode">A code or combination of codes from the Modes constants</param>
    /// <returns>The encoded value for use in the ACOP transport layer</returns>
    /// <remarks>
    /// Using the ACOP Attach() method without passing any flags (i.e. flags = 0) defaults to
    /// mode = Modes.CM_TIMER and access = Access.CA_READ.  Using the ACOP Execute() method 
    /// with flags = 0 defaults to mode = Modes.CM_SINGLE and access = Access.CA_READ.  
    /// The simplest way of issuing a WRITE command is to pass Access.CA_WRITE directly to 
    /// the ACOP Execute() method as the 'flags' parameter.
    /// </remarks>
    public static UInt32 EncodeLinkFlagsForAcop(UInt16 access, UInt16 mode)
    {
      return (UInt32)(access) + (UInt32)(mode << 16);
    }
    public static UInt16 GetAccessFromEncodedFlag(UInt32 flag)
    {
      return (UInt16)(flag & 0xffff);
    }
    public static UInt16 GetModeFromEncodedFlag(UInt32 flag)
    {
      return (UInt16)((flag >> 16) & 0xffff);
    }
    public Boolean IsGrouped()
    {
      return grp == null ? false : true;
    }
    public Boolean Equals(TLink lnk)
    {
      return lnk == this;
    }
    private sbyte[] lnkStatusString = new sbyte[Constants.STATUS_SIZE];
    private String lnkDevName;
    public String GetDeviceName() { return lnkDevName; }
    private String lnkPrpName;
    public String GetProperty() { return lnkPrpName; }
    public String GetKey() 
    {
      if (lnkDevName == null || lnkPrpName == null) return "[empty]";
      return lnkDevName + "[" + lnkPrpName + "]"; 
    }
    private TDataType lnkDout;
    public TDataType GetOutputDataType() { return lnkDout; }
    private TDataType lnkDin;
    public TDataType GetInputDataType() { return lnkDin; }
    private UInt16 lnkAccess;
    public UInt16 GetAccess() { return lnkAccess; }
    private Int32 lnkStatus;
    /// <summary>
    /// The current link status as an integer value.
    /// </summary>
    /// <returns>The current link status.</returns>
    public Int32 GetStatus() { return lnkStatus; }
    public void SetStatus(Int32 status) { lnkStatus = status; }
    private Int32 lnkAccessRate = 1000;
    public Int32 GetAccessRate() { return lnkAccessRate; }
    private UInt32 lnkAccessMode;
    public UInt32 GetAccessMode() { return lnkAccessMode; }
    private UInt32 cbId = 0; // for those who insist
    public Int32 GetCallbackId() { return (Int32)cbId; }
    private IntPtr cbRef = IntPtr.Zero;  // for those who insist
    public IntPtr GetCallbackReference() { return cbRef; }
    private TLinkGroup grp = null;
    public TLinkGroup GetGroup()
    {
      return grp;
    }
    public void DetachFromGroup()
    {
      if (grp != null) grp.RemoveMember(this);
      grp = null;
    }
    private float toleranceAbsolute = 0;
    private float tolerancePercent = 0;
    private int toleranceOffset = 0;
    private int toleranceLength = 0;
    private uint toleranceFlags = DataNotificationFlags.DTF_NONE;
    private Boolean toleranceApplied = false;
    private Boolean needsTolerance = false;
    private uint lastSyncCompletionSize = 0;
    private double lastSyncCompletionTimestamp = 0;
    private double lastSyncSystemStamp = 0;
    private double lastSyncUserStamp = 0;
    private Int32 lnkId = -1;
    public Int32 GetLinkId() { return lnkId; }
    public delegate void TLinkCallback(TLink lnk);
    private TLinkCallback lcb;
    private native.TLinkCallback lnkCb;
    private int _execute(string devName, string devProperty, TDataType dout, TDataType din, Int16 access, UInt32 timeout)
    {
      lastSyncCompletionSize = 0;
      if (dout == null) dout = new TDataType();
      uint arrayLength = dout.GetDataArrayLength();
      int cc = native.ExecLink(devName, devProperty, dout, din, (short)(access|Access.CA_MUTABLE), (UInt16)timeout);
      lastSyncCompletionSize = dout.GetDataArrayLength();
      lastSyncCompletionTimestamp = dout.GetDataTimeStamp();
      lastSyncSystemStamp = (double)dout.GetSystemStamp();
      lastSyncUserStamp = (double)dout.GetUserStamp();
      if (!Access.IsMutable((UInt16)access))
      { // restore this 
        dout.SetDataArrayLength(arrayLength); 
      }
      if (Errors.HasData(cc))
      {
        Object dobj;
        if (dout != null && (dobj = dout.GetReferencedDataObject()) != null)
        {
          dout.GetData(dobj);
        }
      }
      lnkStatus = cc;
      return cc;
    }
    public int Execute(string devName, string devProperty, TDataType dout, TDataType din, Int16 access, UInt32 timeout)
    {
      return _execute(devName, devProperty, dout, din, access, (UInt16)timeout);
    }
    /// <summary>
    /// Executes a synchronous link.
    /// </summary>
    /// <remarks>
    /// Synchronous data exchange. ExecLink() does not complete until the data transfer has completed or 
    /// a timeout has ensued. This call uses the default timeout of 1000 msec.
    /// </remarks>
    /// <returns>0 if successful, otherwise a TINE completion code.</returns>
    /// \b Example:
    /// \include eg_Execute.cs
    public int Execute()
    {
      return _execute(lnkDevName, lnkPrpName, lnkDout, lnkDin, (Int16)lnkAccess, (UInt16)1000);
    }
    /// <summary>
    /// Executes a synchronous link.
    /// </summary>
    /// <remarks>
    /// Synchronous data exchange. ExecLink() does not complete until the data transfer has completed or 
    /// a timeout has ensued.
    /// </remarks>
    /// <param name="timeout">is the designated timeout in milliseconds.</param>
    /// <returns>0 if successful, otherwise a TINE completion code.</returns>
    public int Execute(UInt32 timeout)
    {
      return _execute(lnkDevName, lnkPrpName, lnkDout, lnkDin, (Int16)lnkAccess, (UInt16)timeout);
    }
    /// <summary>
    /// Executes a synchronous link.
    /// </summary>
    /// <remarks>
    /// Synchronous data exchange. ExecLink() does not complete until the data transfer has completed or 
    /// a timeout has ensued.
    /// </remarks>
    /// <param name="mode">is a mode specification to force standard TCP communication (CM_CONNECT) or STREAM communication (CM_STREAM).</param>
    /// <param name="timeout">is the designated timeout in milliseconds.</param>
    /// <returns>0 if successful, otherwise a TINE completion code.</returns>
    public int Execute(UInt16 mode, UInt32 timeout)
    {
      UInt16 acc = 0;
      switch (mode)
      {
        case Modes.CM_CONNECT: acc = Access.CA_CONNECT; break;
        case Modes.CM_STREAM: acc = Access.CA_STREAM; break;
      }
      if (acc != 0) lnkAccess |= acc;
      return _execute(lnkDevName, lnkPrpName, lnkDout, lnkDin, (Int16)lnkAccess, (UInt16)timeout);
    }
    public int GetCompletionDataSize()
    {
      if (lnkId < 0) return (int)lastSyncCompletionSize;
      return native.GetCompletionDataSize(lnkId);
    }
    public double GetCompletionDataTimeStamp()
    {
      if (lnkId < 0) return lastSyncCompletionTimestamp;
      return native.GetCurrentDataTimeStamp(lnkId);
    }
    public double GetCompletionDataStamp()
    {
      if (lnkId < 0) return lastSyncUserStamp;
      return native.GetDataStamp(lnkId);
    }
    public double GetCompletionSystemStamp()
    {
      if (lnkId < 0) return lastSyncSystemStamp;
      return native.GetSystemDataStamp(lnkId);
    }
    public unsafe void memupd()
    {
      object rdo = lnkDout.GetReferencedDataObject();
      if (rdo == null) return;
      if (rdo is TImageHeader)
      {
        
        lnkDout.GetData((TImageHeader)rdo);
        return;
      }
      if (!rdo.GetType().IsArray) return;
      int fmt = lnkDout.GetDataFormat();
      if (fmt == Formats.CF_BOOLEAN)
      {
        if (rdo is Boolean[])
        {
          Boolean[] bd = (Boolean[])rdo;
          Object id = lnkDout.GetIopDataObject();
          if (id != null && id is Int32[])
          {
            Int32[] io = (Int32[])id;
            for (int i = 0; i < bd.Length && i < io.Length; i++)
              bd[i] = io[i] != 0;
            return;
          }
        }
      }
      if (fmt == Formats.CF_STRING)
      {
        if (rdo is String[]) lnkDout.GetData(rdo);
        return;
      }
      if (fmt == Formats.CF_IMAGE)
      {
        if (rdo is DIMAGE[]) lnkDout.GetData(rdo);
        return;
      }
      if (fmt == Formats.CF_SPECTRUM)
      {
        lnkDout.GetData(rdo);
        return;
      }
      if (fmt != Formats.CF_STRUCT) return;
      if (rdo is byte[]) return; // not bound to any Assembly 
      IntPtr ptr = lnkDout.GetDTYPE().data;
      Array tts = (Array)rdo;
      Type t = tts.GetValue(0).GetType();
      int slen = native.GetStructSize(lnkDout.GetDataTag());
      for (int i = 0; i < (int)lnkDout.GetDataArrayLength(); i++)
      {
        ptr = new IntPtr(ptr.ToInt64() + (i * slen));
        object o = Marshal.PtrToStructure(ptr, t);
        tts.SetValue(o, i);
      }
    }
    public unsafe void lclCb(Int32 id, Int32 cc, IntPtr cbRef)
    {
      this.lnkStatus = cc;
      this.cbId = (UInt32)id;
      this.cbRef = cbRef;
      memupd();
      if (lcb != null) lcb(this);
    }
    public int Attach(string devName, string devProperty, TDataType dout, TDataType din, UInt16 access, Int32 accessRate, TLinkCallback cb, UInt32 mode, UInt32 cbId, IntPtr cbRef)
    {
      lnkDevName = devName;
      lnkPrpName = devProperty;
      lnkDout = dout;
      lnkDin = din;
      lnkAccess = access;
      this.cbId = cbId;
      this.cbRef = cbRef;
      return Attach(mode, cb, accessRate);
    }
    /// <summary>
    /// Initiates an asynchronous link.
    /// </summary>
    /// <remarks>
    /// Asynchronous data exchange. Attach() returns immediately with a positive link index if the device 
    /// name can be resolved and there are sufficient resources on the client side. Otherwise, a call to 
    /// Attach() returns a negative completion code.
    /// This method automatically choses to monitor the endpoint in CM_TIMER mode at an access rate
    /// of 1 Hz (1000 msec polling interval).
    /// </remarks>
    /// <param name="cb">is the event callback routine which is fired when new data arrive or an error
    /// conditions arrises. This must have the prototype:
    /// void TLinkCallback(TLink lnk);
    /// </param>
    /// <returns>a positive link index or the negative of a TINE completion code.</returns>
    public int Attach(TLinkCallback cb)
    {
      return Attach(Modes.CM_TIMER, cb, 1000);
    }
    /// <summary>
    /// Initiates an asynchronous link.
    /// </summary>
    /// <remarks>
    /// Asynchronous data exchange. Attach() returns immediately with a positive link index if the device 
    /// name can be resolved and there are sufficient resources on the client side. Otherwise, a call to 
    /// Attach() returns a negative completion code.
    /// This method uses an access rate of 1 Hz (1000 msec polling interval).
    /// </remarks>
    /// <param name="mode">is the desired access mode, i.e. CM_TIMER, CM_DATACHANGE, CM_EVENT, etc. with
    /// possible modifiers (see Modes).</param>
    /// <param name="cb">is the event callback routine which is fired when new data arrive or an error
    /// conditions arrises. This must have the prototype:
    /// void TLinkCallback(TLink lnk);
    /// </param>
    /// <returns>a positive link index or the negative of a TINE completion code.</returns>
    public int Attach(UInt32 mode, TLinkCallback cb)
    {
      return Attach(mode, cb, 1000);
    }
    /// <summary>
    /// Initiates an asynchronous link.
    /// </summary>
    /// <remarks>
    /// Asynchronous data exchange. Attach() returns immediately with a positive link index if the device 
    /// name can be resolved and there are sufficient resources on the client side. Otherwise, a call to 
    /// Attach() returns a negative completion code.
    /// </remarks>
    /// <param name="mode">is the desired access mode, i.e. CM_TIMER, CM_DATACHANGE, CM_EVENT, etc. with
    /// possible modifiers (see Modes).</param>
    /// <param name="cb">is the event callback routine which is fired when new data arrive or an error
    /// conditions arrises. This must have the prototype:
    /// void TLinkCallback(TLink lnk);
    /// </param>
    /// <param name="accessRate">is the desired access polling interval in milliseconds. This interval is
    /// maintained at the server and is used to trigger the server's scheduler.</param>
    /// <returns>a positive link index or the negative of a TINE completion code.</returns>
    /// \b Example:
    /// \include eg_Attach.cs
    public int Attach(UInt32 mode, TLinkCallback cb, Int32 accessRate)
    {
      if (lnkDout != null && lnkDout.isByRef) throw new Exception("invalid output data reference: asynconchronous link on non-array type not allowed");
      lcb = cb;
      lnkCb = lclCb;
      native.fixLink(this);
      if ((mode & Modes.CM_GROUPED) == Modes.CM_GROUPED)
      {
        TLinkGroup g = TLinkGroup.GetGroup(cb);
        if (g != null)
        {
          g.AddMember(this);
          grp = g;
          lnkCb = g.lcbdg; // g.lclCb;
        }
        else
        {
          UInt32 mask = Modes.CM_GROUPED;
          mode &= (~mask);
        }
      }
      lnkAccessMode = mode;
      lnkAccessRate = accessRate;
      //if (lnkDout.GetDataFormat() == Formats.CF_STRING) lnkAccess |= Access.CA_SYNCNOTIFY;
      lnkId = native.AttachLink(lnkDevName, lnkPrpName, lnkDout, lnkDin, (Int16)lnkAccess, accessRate, lnkCb, (Int32)mode, cbId, cbRef);
      if (needsTolerance)
      {
        needsTolerance = false;
        if (native.SetNotificationToleranceEx(lnkId, toleranceAbsolute, tolerancePercent, toleranceOffset, toleranceLength, (int)toleranceFlags) == 0)
        {
          toleranceApplied = true;
        }
      }
      return lnkId;
    }
    /// <summary>
    ///  \brief Allows the caller to apply a tolerance pertaining to link notification.
    /// 
    /// When a TINE client subscribes to data using CM_DATACHANGE mode, the data is sent from server to client
    /// when the data change without regard to any applied tolerance (zero tolerance).  The client application
    /// may nonetheless wish to be notified only when the data have changed outside a given tolerance.
    /// This can be achieved by calling this routine.  The same applies to CM_TIMER mode, where the data are
    /// sent server-to-client according to the schedule.  In the latter case, all data comparisons are made 
    /// entirely at the client side.  Note that if this feature is desired in CM_TIMER mode, the CM_NODUPLICATES 
    /// switch must be applied to the mode parameter in the call to AttachLink() or AttachLinkEx().  This
    /// switch will be appended automatically if SetSuppressHeartbeatNotification(TRUE) is called.
    /// </summary>
    /// <param name="toleranceAbsolute">
    /// if non-zero is a tolerance in the data units of the in-coming data.  It is
    /// checked against irrespective of the current data values.
    /// </param>
    /// <param name="tolerancePercent">
    /// if non-zero is a tolerance expressed as a percent, and refers to the previous
    /// data set.
    /// </param>
    /// <param name="offset">
    /// specifies the array offset to be used in data comparison (DEFAULT: 0).  If the given offset
    /// is less then 0 or greater than the length of the acquired data array (i.e. > 1 if a single value is acquired)
    /// it will be automatically set to 0.
    /// </param>
    /// <param name="length">
    /// specifes the length of the data array (beginning at the given offset) to be used in data
    /// comparison (DEFAULT: length of the acquire data array, or 1 if a single value is acquired). If the given
    /// value exceeds the array bounds, then it will be adjusted.
    /// </param> 
    /// <param name="flags">
    /// is one of DTF_SUPPRESS or DTF_NOTIFY and specifies the desired form of the the 
    /// 'out of tolerance' notification (DEFAULT: DTF_SUPPRESS).  If DTF_SUPPRESS is specified, then callbacks will
    /// be suppressed unless there is an error or the most recent data set is out of tolerance.  
    /// If DTF_NOTIFY is specified, then callbacks will not be suppressed. However, if the most recent data set
    /// is out of tolerance, then an otherwise successful callback completion code will be flagged as 
    /// CE_SENDDATA + out_of_tolerance.  The caller can then make use of this form of 'out of tolerance' signal to
    /// suit his needs.
    /// </param> 
    /// <remarks>
    /// If both an absolute tolerance and a percent tolerance are used, the absolute value of both contributions
    /// are added together to define the total tolerance which should be checked against.
    /// </remarks>
    /// <returns> 0 upon success</returns>
    public int SetNotificationTolerance(float toleranceAbsolute, float tolerancePercent, Int32 offset, Int32 length, Int32 flags)
    {
      if (flags <= DataNotificationFlags.DTF_SUPPRESS || flags > DataNotificationFlags.DTF_NOTIFY)
        flags = (int)DataNotificationFlags.DTF_SUPPRESS;
      this.toleranceAbsolute = toleranceAbsolute;
      this.tolerancePercent = tolerancePercent;
      this.toleranceOffset = offset;
      this.toleranceLength = length;
      this.toleranceFlags = (uint)flags;
      if (lnkId < 0)
      { // link not yet 'attached'
        needsTolerance = true;
        toleranceApplied = false;
      }
      else if (native.SetNotificationToleranceEx(lnkId, toleranceAbsolute, tolerancePercent, offset, length, flags) == 0)
      {
        toleranceApplied = true;
        needsTolerance = false;
      }
      return 0;
    }
    /// <summary>
    ///  \brief Allows the caller to apply a tolerance pertaining to link notification.
    /// 
    /// When a TINE client subscribes to data using CM_DATACHANGE mode, the data is sent from server to client
    /// when the data change without regard to any applied tolerance (zero tolerance).  The client application
    /// may nonetheless wish to be notified only when the data have changed outside a given tolerance.
    /// This can be achieved by calling this routine.  The same applies to CM_TIMER mode, where the data are
    /// sent server-to-client according to the schedule.  In the latter case, all data comparisons are made 
    /// entirely at the client side.  Note that if this feature is desired in CM_TIMER mode, the CM_NODUPLICATES 
    /// switch must be applied to the mode parameter in the call to AttachLink() or AttachLinkEx().  This
    /// switch will be appended automatically if SetSuppressHeartbeatNotification(TRUE) is called.
    /// </summary>
    /// <param name="toleranceAbsolute">
    /// if non-zero is a tolerance in the data units of the in-coming data.  It is
    /// checked against irrespective of the current data values.
    /// </param>
    /// <param name="tolerancePercent">
    /// if non-zero is a tolerance expressed as a percent, and refers to the previous
    /// data set.
    /// </param>
    /// <param name="flags">
    /// is one of DTF_SUPPRESS or DTF_NOTIFY and specifies the desired form of the the 
    /// 'out of tolerance' notification (DEFAULT: DTF_SUPPRESS).  If DTF_SUPPRESS is specified, then callbacks will
    /// be suppressed unless there is an error or the most recent data set is out of tolerance.  
    /// If DTF_NOTIFY is specified, then callbacks will not be suppressed. However, if the most recent data set
    /// is out of tolerance, then an otherwise successful callback completion code will be flagged as 
    /// CE_SENDDATA + out_of_tolerance.  The caller can then make use of this form of 'out of tolerance' signal to
    /// suit his needs.
    /// </param> 
    /// <remarks>
    /// If both an absolute tolerance and a percent tolerance are used, the absolute value of both contributions
    /// are added together to define the total tolerance which should be checked against.
    /// </remarks>
    /// <returns> 0 upon success</returns>
    public int SetNotificationTolerance(float toleranceAbsolute, float tolerancePercent, Int32 flags)
    {
      return SetNotificationTolerance(toleranceAbsolute, tolerancePercent, 0, 0, flags);
    }
    /// <summary>
    ///  \brief Allows the caller to apply a tolerance pertaining to link notification.
    /// 
    /// When a TINE client subscribes to data using CM_DATACHANGE mode, the data is sent from server to client
    /// when the data change without regard to any applied tolerance (zero tolerance).  The client application
    /// may nonetheless wish to be notified only when the data have changed outside a given tolerance.
    /// This can be achieved by calling this routine.  The same applies to CM_TIMER mode, where the data are
    /// sent server-to-client according to the schedule.  In the latter case, all data comparisons are made 
    /// entirely at the client side.  Note that if this feature is desired in CM_TIMER mode, the CM_NODUPLICATES 
    /// switch must be applied to the mode parameter in the call to AttachLink() or AttachLinkEx().  This
    /// switch will be appended automatically if SetSuppressHeartbeatNotification(TRUE) is called.
    /// </summary>
    /// <param name="toleranceAbsolute">
    /// if non-zero is a tolerance in the data units of the in-coming data.  It is
    /// checked against irrespective of the current data values.
    /// </param>
    /// <param name="tolerancePercent">
    /// if non-zero is a tolerance expressed as a percent, and refers to the previous
    /// data set.
    /// </param>
    /// <remarks>
    /// If both an absolute tolerance and a percent tolerance are used, the absolute value of both contributions
    /// are added together to define the total tolerance which should be checked against.
    /// </remarks>
    /// <returns> 0 upon success</returns>
    public int SetNotificationTolerance(float toleranceAbsolute, float tolerancePercent)
    {
      return SetNotificationTolerance(toleranceAbsolute, tolerancePercent, 0, 0, (int)DataNotificationFlags.DTF_SUPPRESS);
    }
    /// <summary>
    /// Closes a TLink.
    /// </summary>
    /// <remarks>
    /// If the link is not currently attached to an input, this call has no effect.
    /// </remarks>
    /// <returns>0 upon success or a TINE error code.</returns>
    public int Close()
    {
      if (lnkId < 0) return -1;
      if (grp != null) DetachFromGroup();
      native.CloseLink(lnkId);
      native.unfixLink(this);
      lnkId = -1;
      return 0;
    }
    /// <summary>
    /// The current link status as a string
    /// </summary>
    /// <returns>The current link status as a string.</returns>
    public string GetStatusString()
    {
      //native.GetLastLinkError((short)lnkStatus, lnkStatusString);
      native.FillCompletionStatusString(lnkId, lnkStatusString);
      char[] errchr = new char[Constants.STATUS_SIZE];
      int i;
      for (i = 0; i < Constants.STATUS_SIZE && lnkStatusString[i] != 0; i++) errchr[i] = (char)lnkStatusString[i];
      return new String(errchr,0,i);
    }
    public TLink()
    {
      cbId = (UInt32)this.GetHashCode();
    }
    /// <summary>
    /// Establishes a link endpoint to the targeted server.
    /// </summary>
    /// <remarks>
    /// Creating a TLink instance will only register the desired link
    /// parameters at the client side.  It will NOT exchange data or 
    /// otherwise establish a connection.
    /// 
    /// Note that supplying an input data object does NOT constitute a
    /// 'WRITE' command to the targeted server!  Input data is simply 'input
    /// data'.  The access of the call is determined by the the access parameter.
    /// </remarks>
    /// <param name="devName">is the desired device name of the target, in the form of
    /// '/context/server/device'.
    /// </param>
    /// <param name="devProperty">is the desired property of the target</param>
    /// <param name="dout">is the desired output data object (data returned from the server)</param>
    /// <param name="din">is the given input data object (data sent to the server)</param>
    /// <param name="access">is the desired access of this link object</param>
    public TLink(string devName, string devProperty, TDataType dout, TDataType din, UInt16 access)
    {
      lnkDevName = devName;
      lnkPrpName = devProperty;
      lnkDout = dout;
      lnkDin = din;
      lnkAccess = access;
      cbId = (UInt32)this.GetHashCode();
    }
    ~TLink()
    { // the Link object is now a null pointer !
      Close(); // just in case ...
      //Console.WriteLine("link destructor is being called");
      GC.Collect();
    }
  }
  public class TServerAddress
  {
    private string srv;
    private string ctx;
    private string eqm;
    private string fec;
    private IPAddress ip;
    private Int32 po;
    public TServerAddress(string server, string context, string eqmName, string fecName, IPAddress ipAddr, Int32 portOffset)
    {
      srv = server == null ? "" : new string(server.ToCharArray());
      ctx = context == null ? "" : new string(context.ToCharArray());
      eqm = eqmName == null ? "" : new string(eqmName.ToCharArray());
      fec = fecName == null ? "" : new string(fecName.ToCharArray());
      ip = ipAddr;
      po = portOffset;
    }
    public TServerAddress(char[] server, char[] context, char[] eqmName, char[] fecName, IPAddress ipAddr, Int32 portOffset)
    {
      srv = server == null ? "" : new string(server);
      ctx = context == null ? "" : new string(context);
      eqm = eqmName == null ? "" : new string(eqmName);
      fec = fecName == null ? "" : new string(fecName);
      ip = ipAddr;
      po = portOffset;
    }
    public string GetServer() { return srv; }
    public void SetServer(string serverName) 
    { 
      srv = serverName == null ? "" : new string(serverName.ToCharArray()); 
    }
    public string GetContext() { return ctx; }
    public void SetContext(string context) 
    {
      ctx = context == null ? "" : new string(context.ToCharArray()); 
    }
    public string GetEqm() { return eqm; }
    public string GetFec() { return fec; }
    public IPAddress GetIpAddress() { return ip; }
    public Int32 GetPortOffset() { return po; }
  }
  public class TQuery
  {
    /// <summary>
    /// Retrieves a list of server contexts via query to the Equipment Name Server. 
    /// </summary>
    /// This call retrieves a list of all server context managed by the Equipment Name Server
    /// <returns>A string array containing the desired context list.  On error, throw an exception containing 
    /// the reason for the exception as a string, as well as the error code (embedded between angle brackets).
    /// </returns>
    public static String[] GetSystemContexts()
    {
      NAME32[] clst = new NAME32[100];
      Int32 n = 100;
      IntPtr p = TDataType.GetDataPtr(clst);
      int cc = native.GetSystemContexts(p, ref n);
      if (cc != 0) throw new Exception("query failed: <" + cc + "> " + Errors.ToString(cc));
      if (n <= 0) return null;
      String[] ctxs = new String[n];
      for (int i = 0; i < n; i++) ctxs[i] = clst[i].GetName();
      return ctxs;
    }
    /// <summary>
    /// Retrieves a list of device servers via query to the Equipment Name Server. 
    /// </summary>
    /// This call retrieves a list of all Front End Computer Names (FECs) associated 
    /// with the server context given. In addition, the context string can carry a 
    /// targeted subsystem for a more specific list.
    /// <param name="context">is the context for which the query is to be made. A narrower 
    /// query can also be made, where only servers belonging to a specific subsystem are 
    /// returned. If this is desired, then pass a context string of the form 
    /// 'context/subsystem' (e.g. "PETRA/DIAG"). </param>
    /// <returns>A string array containing the desired server list.  On error, throw an exception containing 
    /// the reason for the exception as a string, as well as the error code (embedded between angle brackets).
    /// </returns>
    public static String[] GetSystemServers(String context)
    {
      NAME32[] lst = new NAME32[1024];
      Int32 n = 1024;
      IntPtr p = TDataType.GetDataPtr(lst);
      int cc = native.GetSystemServers(context, p, ref n);
      if (cc != 0) throw new Exception("query failed: <" + cc + "> " + Errors.ToString(cc));
      if (n <= 0) return null;
      String[] srvs = new String[n];
      for (int i = 0; i < n; i++) srvs[i] = lst[i].GetName();
      return srvs;
    }
    /// <summary>
    /// Retrieves a list of properties via query to the server given. 
    /// </summary>
    /// This call retrieves a list of all properties registered on the server 
    /// associated with full device server name specified. A classic device server 
    /// will export its behavior through its properties, and export its "instances" 
    /// through its "device names", according to the motto "All devices found on this 
    /// server have the same set of properties, and all properties apply to any device 
    /// found on this server". In this case, simply specifying the device server name 
    /// as the first input parameter will return the set of properties valid for all 
    /// devices registered on the server. Under some circumstances, however, the set 
    /// of properties associated with a particular device might be different from device 
    /// to device. If this is the case, the full device server name should be used as the 
    /// first parameter, where the target device name is also sent as part of the query.
    /// <param name="server">is the full device server name for which the property query is to be made </param>
    /// <returns>A string array containing the desired property list.  On error, throw an exception containing 
    /// the reason for the exception as a string, as well as the error code (embedded between angle brackets).
    /// </returns>
    public static String[] GetSystemProperties(String server)
    {
      Boolean p = false;
      return GetSystemProperties(server, ref p);
    }
    /// <summary>
    /// Retrieves a list of properties via query to the server given. 
    /// </summary>
    /// This call retrieves a list of all properties registered on the server 
    /// associated with full device server name specified. A classic device server 
    /// will export its behavior through its properties, and export its "instances" 
    /// through its "device names", according to the motto "All devices found on this 
    /// server have the same set of properties, and all properties apply to any device 
    /// found on this server". In this case, simply specifying the device server name 
    /// as the first input parameter will return the set of properties valid for all 
    /// devices registered on the server. Under some circumstances, however, the set 
    /// of properties associated with a particular device might be different from device 
    /// to device. If this is the case, the full device server name should be used as the 
    /// first parameter, where the target device name is also sent as part of the query.
    /// <param name="server">is the full device server name for which the property query is to be made </param>
    /// <param name="hasDeviceQueryPrecedence">indicates whether the target server is using device query precedence</param>
    /// <returns>A string array containing the desired property list.  On error, throw an exception containing 
    /// the reason for the exception as a string, as well as the error code (embedded between angle brackets).
    /// </returns>
    public static String[] GetSystemProperties(String server, ref Boolean hasDeviceQueryPrecedence)
    {
      NAME64[] lst = new NAME64[2048];
      Int32 n = 2048;
      IntPtr p = TDataType.GetDataPtr(lst);
      int cc = native.GetSystemProperties(server, p, ref n);
      hasDeviceQueryPrecedence = (Errors.GetErrorCode(cc) == Errors.has_query_function);
      if ((cc & Errors.CE_SENDDATA) != 0) cc = 0;
      if (cc != 0) throw new Exception("query failed: <"+cc+"> "+Errors.ToString(cc));
      if (n <= 0) return null;
      String[] prps = new String[n];
      for (int i = 0; i < n; i++) prps[i] = lst[i].GetName();
      return prps;
    }
    /// <summary>
    /// Retrieves a list of device names associated with a given property via query to the server specified. 
    /// </summary>
    /// <remarks>
    /// This call retrieves a list of all device names registered on the server associated with the 
    /// property specified in the call. A classic device server will export its behavior through its properties, 
    /// and export its "instances" through its "device names", according to the motto "All devices found on this 
    /// server have the same set of properties, and all properties apply to any device found on this server". 
    /// For certain classes of server (in particular middle layer servers) the properties exported might refer 
    /// to 'apples' and 'oranges' simply because the server is archiving them. Or there might be other reasons why 
    /// the device list associated with a particular property has little or nothing to do with that associated with 
    /// some other property. Use this call to obtain the device list associated with a particular property.
    /// </remarks>
    /// <param name="server"> is the full device server name for which the property query is to be made </param>
    /// <param name="property">is the property for which the device list is desired. </param>
    /// 
    /// <returns>A string array containing the desired device list.  On error, throw an exception containing 
    /// the reason for the exception as a string, as well as the error code (embedded between angle brackets).
    /// </returns>
    public static String[] GetSystemDevices(String server, String property)
    {
      Boolean p = false;
      return GetSystemDevices(server, property, ref p);
    }
    /// <summary>
    /// Retrieves a list of device names associated with a given property via query to the server specified. 
    /// </summary>
    /// <remarks>
    /// This call retrieves a list of all device names registered on the server associated with the 
    /// property specified in the call. A classic device server will export its behavior through its properties, 
    /// and export its "instances" through its "device names", according to the motto "All devices found on this 
    /// server have the same set of properties, and all properties apply to any device found on this server". 
    /// For certain classes of server (in particular middle layer servers) the properties exported might refer 
    /// to 'apples' and 'oranges' simply because the server is archiving them. Or there might be other reasons why 
    /// the device list associated with a particular property has little or nothing to do with that associated with 
    /// some other property. Use this call to obtain the device list associated with a particular property.
    /// </remarks>
    /// <param name="server"> is the full device server name for which the property query is to be made </param>
    /// <param name="property">is the property for which the device list is desired. </param>
    /// <param name="hasPropertyQueryPrecedence">denotes whether the target server is using property query precedence.</param>
    /// 
    /// <returns>A string array containing the desired device list.  On error, throw an exception containing 
    /// the reason for the exception as a string, as well as the error code (embedded between angle brackets).
    /// </returns>
    public static String[] GetSystemDevices(String server, String property, ref Boolean hasPropertyQueryPrecedence)
    {
      int cc, n = 2048;
      String[] devs = null;
      if (server != null && property != null  && property.Length > 0 && server.Contains("*"))
      { // a cursed wildcard ...
        String[] parts = server.Split(new char[] { '/' }, StringSplitOptions.RemoveEmptyEntries);
        if (parts.Length > 2 && parts[2].Contains("*"))
        { // just making sure
          DUSTRING[] ulst = new DUSTRING[2048];
          TDataType dout = new TDataType(ulst);
          TLink lnk = new TLink(server, property, dout, null, Access.CA_READ|Access.CA_MUTABLE|Access.CA_NOCALLBACKS);
          cc = lnk.Execute(200);
          if (cc != 0) throw new Exception("query failed: <" + cc + "> " + Errors.ToString(cc));
          n = (int)dout.GetDataArrayLength();
          devs = new String[n];
          for (int i = 0; i < n; i++) devs[i] = ulst[i].GetName();
          return devs;
        }
      }
      NAME64[] lst = new NAME64[2048];
      IntPtr p = TDataType.GetDataPtr(lst);
      cc = native.GetSystemDevices(server, property, p, ref n);
      hasPropertyQueryPrecedence = (Errors.GetErrorCode(cc) == Errors.has_query_function);
      if ((cc & Errors.CE_SENDDATA) != 0) cc = 0;
      if (cc != 0) throw new Exception("query failed: <" + cc + "> " + Errors.ToString(cc));
      if (n <= 0) return null;
      devs = new String[n];
      for (int i = 0; i < n; i++) devs[i] = lst[i].GetName();
      return devs;
    }
    public static int GetPropertyInformation(string dev, string prp, ref PrpQueryStruct[] pqs, ref int num)
    {
      return GetPropertyInformation(dev, prp, ref pqs, ref num, 200);
    }
    private static int GetNumberOf(string dev, string tgt, uint timeout)
    {
      TLink lnk;
      TDataType dout;
      int n = 0, cc;
      dout = new TDataType(ref n); // make 'a' call in order to be sure the client cycle is initialized
      lnk = new TLink(dev, "N" + tgt, dout, null, Access.CA_READ | Access.CA_NOCALLBACKS);
      cc = lnk.Execute(timeout);
      if (!Errors.HasData(cc)) return -cc;
      return n;
    }
    public static int GetPropertyInformation(string dev, string prp, ref PrpQueryStruct[] pqs, ref int num, uint timeout)
    {
      TLink lnk;
      TDataType dout, din, dinptr = null;
      DTYPE dtout;
      int cc = 0;
      int n, nprp = 0;
      PrpQueryStruct[] prpqs = null;

      if (dev == null || pqs == null) return Errors.invalid_parameter;
      String tgt = (prp != null && TKernel.IsStockProperty(prp)) ? "STOCKPROPS" : "PROPERTIES";
      nprp = GetNumberOf(dev, tgt, timeout);
      if (nprp < 1) return -nprp;
      if ((n = num) <= 0) n = prp == null ? nprp : 10;
      if ((prpqs = pqs) == null)
      { /* caller wants me to allocate this */
        prpqs = new PrpQueryStruct[n];
      }
      Boolean iswildcard = false;
      NAME64[] n64 = new NAME64[1];
      if (prp != null)
      {
        n64[0] = new NAME64();
        n64[0].SetName(prp);
        din = new TDataType(n64);
        dinptr = din;
        if (dev.Contains("*"))
        {
          String[] parts = dev.Split(new char[] { '/' }, StringSplitOptions.RemoveEmptyEntries);
          if (parts.Length > 2 && parts[2].Contains("*"))
          {
            iswildcard = true;
          }
        }
      }
      dtout = new DTYPE();
      dtout.dArrayLength = (uint)n;
      dtout.dFormat = Formats.LFMT_OFFSET + Formats.CF_STRUCT;
      dtout.data = TDataType.GetDataPtr(prpqs, "PRPQSr4");
      dtout.dTag = "PRPQSr4".ToCharArray();
      dout = new TDataType(dtout,prpqs);
      lnk = new TLink(dev, tgt, dout, dinptr, Access.CA_READ | Access.CA_NOCALLBACKS);
      if ((cc = lnk.Execute(timeout)) != 0)
      {
        if (cc == Errors.non_existent_property && !iswildcard)
        { // maybe a structure field ?
          int idx = prp.LastIndexOf('.');
          if (idx > 1)
          {
            String bprp = prp.Substring(0, idx);
            n64[0] = new NAME64(bprp);
            lnk = new TLink(dev, tgt, dout, dinptr, Access.CA_READ | Access.CA_NOCALLBACKS);
            if (lnk.Execute(timeout) == 0)
            {
              if (prpqs[0].prpFormat == Formats.CF_STRUCT)
              {
                if (AcquireAndRegisterStruct(dev, new String(prpqs[0].prpTag), (int)prpqs[0].prpSize) == 0)
                {
                  // TODO: we can't as yet get the struct info out of native !
                  prpqs[0].prpFormat = Formats.CF_DOUBLE;
                  prpqs[0].prpArrayType = ArrayType.AT_SINGLE;
                  prpqs[0].prpSize = 1;
                  num = 1;
                  // 
                }
              }
            }
          }
        }
        if (iswildcard) cc = 0; else return cc;
      }
      if (cc == 0 || (cc & Errors.CE_SENDDATA) != 0)
      { /* modern server understood the query */
        if (iswildcard)
        { // doocs nonsense ...
          prpqs[0].prpFormat = Formats.CF_USTRING;
          prpqs[0].prpArrayType = ArrayType.AT_CHANNEL | ArrayType.AT_SINGLE;
          int ndevs = GetNumberOf(dev, "DEVICES", timeout);
          prpqs[0].prpSize = ndevs < 0 ? 0 : (uint)ndevs;
          num = 1;
        }
        else
        {
          num = (int)dout.GetDataArrayLength();
        }
        pqs = prpqs;
        return 0;
      }
      return cc;
    }
    public static TServerAddress GetServerAddress(string context, string server)
    {
      TServerAddress sa = null;
      ExpDataStruct ea = new ExpDataStruct();
      FecAddrStruct fa = new FecAddrStruct();
      int cc = 0;
      if ((cc=native.GetServerAddr(context, server, ref ea, ref fa)) == 0)
      { // TODO: ipv6 branch (the IPAddress constructor makes a decision based on the length of the byte array!)
        byte[] ipb = new byte[4];
        for (int i = 0; i < 4; i++) ipb[i] = fa.netAdr[i];
        IPAddress ip = new IPAddress(ipb);
        sa = new TServerAddress(ea.EqmContext,ea.ExportName,ea.EqmName,fa.fecName,ip,fa.portOffset);
      }

      return sa;
    }
    public static int AcquireAndRegisterStruct(string devName,string tag,int number)
    {
      return native.AcquireAndRegisterStruct(devName, tag, number);
    }
  }

  public class THistory
  {
    public class HstSrc
    {
      public const int CENTRAL = 1;
      public const int LOCAL = 2;
      private int sourceType;
      private String name;
      private String server;
      private int numberOfPoints;
      private int numberOfPointsAlternate;
      private int numberReturned = 0;
      /// <summary>
      /// sets the source type (used internally)
      /// </summary>
      /// <param name="type"></param>
      public void SetSourceType(int type)
      {
        if (type != CENTRAL) type = LOCAL;
        sourceType = type;
      }
      /// <summary>
      /// either CENTRAL or LOCAL
      /// </summary>
      /// <returns>the source type</returns>
      public int GetSourceType()
      {
        return sourceType;
      }
      /// <summary>
      /// either "central" or "local"
      /// </summary>
      /// <returns>the source type as string</returns>
      public String getName()
      {
        return name;
      }
      /// <summary>
      /// The server delivering the data
      /// </summary>
      /// <returns>The server delivering the data</returns>
      public String getServer()
      {
        return server;
      }
      /// <summary>
      /// Number of stored points in the time range
      /// </summary>
      /// <returns>Number of stored points in the time range</returns>
      public int getNumberOfPoints()
      {
        return numberOfPoints;
      }
      /// <summary>
      /// Number of stored points in the time range from the alternate source
      /// </summary>
      /// <returns>Number of stored points in the time range from the alternate source</returns>
      public int getNumberOfPointsAlternate()
      {
        return numberOfPointsAlternate;
      }
      public HstSrc(String srcName, String srvName, int numPoints)
      {
        server = srvName;
        name = srcName;
        numberOfPoints = numPoints;
      }
      /// <summary>
      /// Used internally
      /// </summary>
      /// <param name="srcName"></param>
      /// <param name="srvName"></param>
      /// <param name="numPoints"></param>
      public void SetSource(String srcName, String srvName, int numPoints)
      {
        server = srvName;
        name = srcName;
        numberOfPoints = numPoints;
      }
      /// <summary>
      /// Number of points returned in the archive data acquisition
      /// </summary>
      /// <returns>Number of points returned in the archive data acquisition</returns>
      public int getNumberReturned()
      {
        return numberReturned;
      }
      /// <summary>
      /// used internally
      /// </summary>
      /// <param name="numberReturned"></param>
      public void setNumberReturned(int numberReturned)
      {
        this.numberReturned = numberReturned;
      }
      /// <summary>
      /// used internally
      /// </summary>
      /// <param name="number"></param>
      public void setNumberPoints(int number)
      {
        numberOfPoints = number;
      }
      /// <summary>
      /// used internally
      /// </summary>
      /// <param name="number"></param>
      public void setNumberPointsAlternate(int number)
      {
        numberOfPointsAlternate = number;
      }
    }
    private const int HEARTBEAT = 60 * 15;
    public enum ExclusiveHistorySource
    {
      none = 0,
      central = 1,
      local = 2
    }
    private static ExclusiveHistorySource exHstSrc = ExclusiveHistorySource.none;
    /// <summary>
    /// Gets the current exclusive history source setting
    /// </summary>
    /// <returns>the current exclusive history source setting</returns>
    public static ExclusiveHistorySource GetExclusiveHistorySource()
    {
      return exHstSrc;
    }
    /// <summary>
    /// Sets the current exclusive history source setting
    /// </summary>
    /// <param name="value">the desired setting</param>
    public static  void SetExclusiveHistorySource(ExclusiveHistorySource value)
    {
      exHstSrc = value;
    }
    private static Boolean isCentralArchiveServer(String server)
    {
      return server.EndsWith("HISTORY");
    }
    private static Boolean isCentralArchiver(String server)
    {
      return server.EndsWith("ARCHIVER");
    }
    private static Boolean isArchiveQueryProperty(String property)
    {
      if (property.EndsWith(".HIST")) return true;
      if (property.EndsWith(".HST")) return true;
      if (property.EndsWith(".ARCH")) return true;
      if (property.EndsWith(".ARC")) return true;
      if (property.EndsWith(".AR")) return true;
      return false;
    }
    private static String getQueryPropertyRoot(String property)
    {
      int slen = property.Length;
      if (property.EndsWith(".HIST") || property.EndsWith(".ARCH"))
        return property.Substring(0, slen - 5);
      if (property.EndsWith(".HST") || property.EndsWith(".ARC"))
        return property.Substring(0, slen - 4);
      if (property.EndsWith(".AR"))
        return property.Substring(0, slen - 3);
      return property;
    }
    private static Boolean isScheduledProperty(String property)
    {
      int idx = property.LastIndexOf('.');
      if (idx < 0) return false;
      String dec = property.Substring(idx + 1);
      if (dec == null) return false;
      return dec.ToUpper().StartsWith("SCH");
    }
    private static int makeStartStopInput(double start, double stop, int index, int fieldindex, int sample, Boolean iscentral, int[] startstop, double[] dstartstop)
    {
      int inptLength = 0;
      if (startstop != null && startstop.Length >= 8)
      {
        startstop[0] = (int)start;
        startstop[1] = (int)Math.Ceiling(stop);
        startstop[2] = index;
        startstop[3] = sample;
        inptLength = (index == 0 && sample == 0) ? 2 : 4;
        if (fieldindex >= 0)
        {
          if (iscentral)
          {
            startstop[4] = fieldindex;
            inptLength = 5;
          }
          else
          {
            startstop[6] = startstop[5] = startstop[4] = 0;
            startstop[7] = fieldindex;
          }
        }
      }
      if (dstartstop != null && dstartstop.Length >= 8)
      {
        dstartstop[0] = start;
        dstartstop[1] = stop;
        dstartstop[2] = (double)index;
        dstartstop[3] = (double)sample;
        inptLength = (index == 0 && sample == 0) ? 2 : 4;
        if (fieldindex >= 0)
        {
          if (iscentral)
          {
            dstartstop[4] = (double)fieldindex;
            inptLength = 5;
          }
          else
          {
            dstartstop[6] = dstartstop[5] = dstartstop[4] = 0;
            dstartstop[7] = (double)fieldindex;
            inptLength = 8;
          }
        }
      }
      return inptLength;
    }
    private static String getTargetDevice(String context, String server, String device, String property)
    {
      String ctx = context, srv = server, prp = property, dev = device;
      // TODO: get redirection info is not yet exposed out of tinemt
      //if (TLink.isRedirected(context, server, device, property))
      //{
      //  RedirectedItem ri = TLinkFactory.getRedirectionInformation(context, server, device, property);
      //  ctx = ri.getDstContext();
      //  srv = ri.getDstServer();
      //  prp = ri.getDstProperty();
      //  dev = ri.getDstDevice();
      //}
      int idx = -1;
      if ((idx = prp.LastIndexOf(".HIST")) > 0 || (idx = prp.LastIndexOf(".HST")) > 0)
      {
        prp = prp.Substring(0, idx);
      }
      return dev;
    }
    private static String[] getArchivedProperties(String context, String server, String subsystem)
    {
      if (server == null && context == null) return null;
      if (context == null) context = "DEFAULT";
      if (server == null) server = "HISTORY";
      Boolean isCentral = isCentralArchiveServer(server);
      TDataType dtsubs;
      if (subsystem != null && subsystem.Length > 0 &&
          !subsystem.StartsWith("ALL") && !subsystem.StartsWith("All"))
      {
        dtsubs = new TDataType(subsystem);
      }
      else
      {
        dtsubs = new TDataType();
      }
      NAME32[] nlist = null; //TODO:  why not NAME64 ?
      int[] nprps = new int[1];
      String ntagsString, tagsString;
      TDataType dtnprps = new TDataType(nprps);
      TDataType dtprps;
      tagsString = isCentral ? "TAGS" : "HISTORIES";
      ntagsString = "N" + tagsString;

      String dname = "/" + context + "/" + server;
      TLink tl = new TLink(dname, ntagsString, dtnprps, dtsubs, Access.CA_READ);
      int cc = tl.Execute(800);
      if (cc != 0)
      { // TODO: throw an exception here
        return null;
      }
      if (nprps[0] > 0)
      {
        nlist = new NAME32[nprps[0]];
        dtprps = new TDataType(nlist);
        tl = new TLink(dname, tagsString, dtprps, dtsubs, Access.CA_READ);
        cc = tl.Execute(800);
        if (cc != 0)
        { // TODO: throw an exception here
          return null;
        }
      }
      if (nlist == null) return null;
      String nam;
      List<String> v = new List<String>(nlist.Length);
      int n = 0;
      for (int i = 0; i < nlist.Length; i++)
      {
        nam = nlist[i].GetName();
        if (nam.EndsWith(".NAM")) continue;
        if (nam.EndsWith(".ENUM")) continue;
        if (nam.EndsWith(".SRCADDR")) continue;
        if (nam.EndsWith(".DESC")) continue;
        v.Add(nam);
        n++;
      }
      return v.ToArray();
    }
    private static Boolean preferCentralSource = false; //TODO: this is of course not thread-safe
    private static int getArchivedNumPointsInInterval(String context, String server, String property, String device, double start, double stop, uint timeout)
    {
      int cc = 0;
      int[] startstop = new int[2];
      int[] npts = new int[1];
      TDataType dout, din;

      preferCentralSource = false;
      if (server == null && context == null) return Errors.argument_list_error;
      if (context == null) context = "DEFAULT";
      if (server == null) server = "HISTORY"; // assume the standard central archive server for the given context
      startstop[0] = (int)start;
      startstop[1] = (int)Math.Ceiling(stop);
      din = new TDataType(startstop);
      dout = new TDataType(npts);

      if (isCentralArchiver(server))
      { // trap a potential redirection problem here:
        if (device.Equals("keyword",StringComparison.CurrentCultureIgnoreCase)) device = "#0";
      }
      String dname = "/" + context + "/" + server + "/" + device;
      String pname = property;
      if (!isCentralArchiveServer(server))
      { // not a call to a central archive server
        if (!isArchiveQueryProperty(property))
          pname = pname + ".HIST"; // meta extension not given, try this ...
      }
      try
      {
        TLink tl = new TLink(dname, pname, dout, din, Access.CA_READ);
        cc = tl.Execute(timeout);
        if (Errors.HasData(cc))
        {
          if (Errors.GetErrorCode(cc) == Errors.at_limit) preferCentralSource = true;
          cc = 0;
        }
      }
      catch (Exception)
      {
        cc = Errors.address_unresolved;
      }
      return cc > 0 ? -cc : npts[0];
    }
    private static int getArchivedDataArray(String context, String server, String property, String device, int index, int fieldindex, int sample, double start, double stop, object data, uint timeout)
    {
      int cc = 0, inptLength = 4;
      int[] startstop = new int[8];
      TDataType dout, din;

      if (server == null && context == null) return -Errors.argument_list_error;
      if (data == null) return -Errors.argument_list_error;
      if (context == null) context = "DEFAULT";
      if (server == null) server = "HISTORY"; // assume the standard central archive server for the given context
      if (timeout < 100) timeout = 100;
      inptLength = makeStartStopInput(start, stop, index, fieldindex, sample, isCentralArchiveServer(server), startstop, null);
      if (startstop[0] == 0 && startstop[1] == 0)
      {
        din = new TDataType();
      }
      else if (data is INTFLTINT[])
      {
        INTINTINTINT[] twnd = new INTINTINTINT[1];
        twnd[0] = new INTINTINTINT(startstop[0], startstop[1], 0, 0);
        din = new TDataType(twnd);
      }
      else
      { 
        din = new TDataType(startstop, 0, inptLength);
      }
      dout = new TDataType(data);

      if (isCentralArchiver(server))
      { // trap a potential redirection problem here:
        if (device.Equals("keyword",StringComparison.CurrentCultureIgnoreCase)) device = "#0";
      }
      String pname = property;
      String rname = getQueryPropertyRoot(property);
      Boolean tryAlternateProperty = false;
      if (!isCentralArchiveServer(server))
      { // not a call to a central archive server
        device = getTargetDevice(context, server, device, property);
        tryAlternateProperty = true;
        if (!isArchiveQueryProperty(property))
          pname = pname + ".HIST"; // meta extension not given, try this ...
      }
      String dname = "/" + context + "/" + server + "/" + device;
      TLink tl = new TLink(dname, pname, dout, din, Access.CA_READ);
      cc = tl.Execute(timeout);
      if (cc == Errors.un_allocated && tryAlternateProperty)
      {
        String[] hists = getArchivedProperties(context, server, null);
        if (hists != null)
        { // there are histories available ....
          if (isScheduledProperty(rname))
          { // look for unscheduled alternative
            foreach (String h in hists)
            {
              if (rname.StartsWith(h))
              {
                pname = h + ".HIST";
                tl = new TLink(dname, pname, dout, din, Access.CA_READ);
                cc = tl.Execute(timeout);
                break;
              }
            }
          }
          else
          { // look for scheduled alternative
            foreach (String h in hists)
            {
              if (isScheduledProperty(h))
              {
                if (h.StartsWith(rname))
                {
                  pname = h + ".HIST";
                  tl = new TLink(dname, pname, dout, din, Access.CA_READ);
                  cc = tl.Execute(timeout);
                  break;
                }
              }
            }
          }
        }
      }
      int npts = cc == 0 ? tl.GetCompletionDataSize() : 0;
      if (npts == 0 && isCentralArchiver(server))
      { // special 2nd chance !
        if (isArchiveQueryProperty(property))
        {
          pname = rname + ".ARCH";
        }
        else
        {
          pname = property + ".ARCH";
        }
        tl = new TLink(dname, pname, dout, din, Access.CA_READ);
        cc = tl.Execute(timeout);
      }
      return cc == 0 ? npts : -cc;
    }
    private static int getDataArray(String context, String server, String property, String device, int index, int fieldindex, int sample, double start, double stop, object data, HstSrc src, uint timeout)
    {
      if (data == null || !(data is Array)) return Errors.invalid_parameter;
      int dlen = ((Array)data).Length;
      int idx, cc = 0, cc1 = 0, cc2 = 0;
      // s1, p1 will refer to the central archive repository
      // s2, p2 will refer to the local history at the source server
      String s1 = server;
      String p1 = property;
      if (s1 == null) s1 = "HISTORY";
      if ((idx = property.IndexOf(".HIST")) >= 0 ||
          (idx = property.IndexOf(".HST")) >= 0 ||
          (idx = property.IndexOf(".AR")) >= 0)
      { // passed with .HIST appended, remove it!
        property = property.Substring(0, idx);
      }
      String s2 = s1; // server 2nd source
      if (!isCentralArchiveServer(s1))
      { // a specific server was passed
        p1 = property + ".ARCH"; // redirect to same channel at the central archiver
      }
      else
      { // 2nd source must redirect back to find the target server's local histories
        s2 = "ARCHIVER";
      }
      double npstart, npstop;
      npstart = start;
      npstop = stop;
      int npts1 = -Errors.not_applicable;
      if (exHstSrc != ExclusiveHistorySource.local)
        npts1 = getArchivedNumPointsInInterval(context, s1, p1, device, npstart, npstop, 500);
      if (npts1 < 0)
      { // a problem with this entry
        cc1 = -npts1;
        npts1 = -1; // was 0
      }
      else
      { // central archive is alive and knows about the record
        String thisDataKey = "/" + context + "[" + property + "]" + start + "->" + stop;
      }
      String p2 = property + ".HIST";
      String dev2 = getTargetDevice(context, server, device, property);
      int npts2 = -Errors.not_applicable;
      if (exHstSrc != ExclusiveHistorySource.central)
        npts2 = getArchivedNumPointsInInterval(context, s2, p2, dev2, npstart, npstop, 500);
      Boolean preferCentral = preferCentralSource;
      if (npts2 < 0)
      { // a problem with this entry
        if (npts2 == -Errors.un_allocated)
        {
          String[] hists = getArchivedProperties(context, s2, null);
          if (hists != null)
          { // there are histories available ....
            if (isScheduledProperty(property))
            { // look for unscheduled alternative
              foreach (String h in hists)
              {
                if (property.StartsWith(h))
                {
                  p2 = h + ".HIST";
                  npts2 = getArchivedNumPointsInInterval(context, s2, p2, dev2, npstart, npstop, 500);
                  break;
                }
              }
            }
            else
            { // look for scheduled alternative
              foreach (String h in hists)
              {
                if (isScheduledProperty(h))
                {
                  if (h.StartsWith(property))
                  {
                    p2 = h + ".HIST";
                    npts2 = getArchivedNumPointsInInterval(context, s2, p2, dev2, npstart, npstop, 500);
                    break;
                  }
                }
              }
            }
          }
        }
        if (npts2 < 0)
        {
          cc2 = -npts2;
          npts2 = -1; // was 0
        }
      }
      if (npts1 <= 0) preferCentral = false;
      Boolean done = false;
      int nret = 0;
      if (npts1 > 500 || npts1 > npts2 || preferCentral)
      { // central archive has 'enough' points or at least more than the local history
        nret = getArchivedDataArray(context, s1, p1, device, index, fieldindex, sample, start, stop, data, timeout);
        if (nret > 0)
        { // everything's okay
          if (src != null)
          {
            src.SetSourceType(HstSrc.CENTRAL);
            src.setNumberReturned(nret);
          }
          done = true;
        }
        else
        {
          cc = -nret;
          cc1 = cc;
        }
      }
      if (!done && npts2 > 0)
      { // too few (or no) points upstairs, so try the local history ...
        nret = getArchivedDataArray(context, s2, p2, dev2, index, fieldindex, sample, start, stop, data, timeout);
        if (src != null) src.SetSourceType(HstSrc.LOCAL);
        if (nret > 0)
        {
          if (src != null) src.setNumberReturned(nret);
          done = true;
        }
        else
        {
          cc = -nret;
        }
        if (cc != 0 || nret == 0)
        { // error (local or remote -> did not get back any points after all!)
          cc2 = cc;
          if (npts1 > 0)
          { // would have been a nice place for a 'goto' ....
            nret = getArchivedDataArray(context, s1, p1, device, index, fieldindex, sample, start, stop, data, timeout);
            if (nret > 0)
            { // everything's okay
              if (src != null)
              {
                src.SetSourceType(HstSrc.CENTRAL);
                src.setNumberReturned(nret);
              }
              done = true;
            }
          }
        }
      }
      if (!done && npts2 <= 0 &&
          !(data is INTFLTINT[]) &&
          !isCentralArchiveServer(server) &&
          !isCentralArchiver(server) && 
          exHstSrc != ExclusiveHistorySource.central)
      { // last desperate attempt (possible legacy doocs server ?)
        int ifi_len = dlen;
        if (ifi_len < 2000) ifi_len = 2000;
        INTFLTINT[] ifi = new INTFLTINT[ifi_len];
        int rst = 1, dn = 0, n = 0;
        double interim = 0;
        while (start < stop)
        {
          nret = getArchivedDataArray(context, s2, p2, device, index, -1, 0, start, stop, ifi, timeout);
          if (nret > 0)
          {
            if (dn == 0)
            {
              dn = (ifi[nret - 1].i1val - (int)start);
              if (dn <= 0) dn = 1;
              rst = (int)(stop - start) / dn;
              if (rst < 1) rst = 1;
            }
            for (int i = 0; n < dlen && i < nret; i += rst)
            {
              if (data is FLTINT[])
              ((FLTINT[])data)[n] = new FLTINT(ifi[i].fval, ifi[i].i1val);
              else if (data is DBLDBL[])
                ((DBLDBL[])data)[n] = new DBLDBL((double)ifi[i].fval, (double)ifi[i].i1val);
              n++;
            }
            interim = (double)(ifi[nret - 1].i1val + 1);
            if (interim <= start) break;
            start = interim;
            done = true;
            if (rst == 1) break;
          }
          else
          {
            break;
          }
        }
        if (src != null)
        {
          src.SetSourceType(HstSrc.LOCAL);
          src.setNumberPoints(n);
          src.setNumberReturned(n);
        }
      }
      if (src != null)
      {
        if (src.GetSourceType() == HstSrc.CENTRAL)
        {
          src.SetSource("Central", s1, src.getNumberReturned());
          src.setNumberPointsAlternate(npts2);          
        }
        else
        {
          src.SetSource("Local", s2, src.getNumberReturned());
          src.setNumberPointsAlternate(npts1);
        }
      }
      if (nret == 0)
      {
    	  if (cc1 == Errors.connection_timeout || cc1 == Errors.link_timeout) 
    		  cc = cc1; 
    	  else if (cc2 != 0 && cc2 != Errors.un_allocated) 
    		  cc = cc2;
    	  else if (cc2 == 0)
    		  cc = 0;
    	  else
    		  cc = cc1;
      }
      return cc;
    }
    /// <summary>
    /// Gets archived data for the parameters given over the time range specified.
    /// </summary>
    /// <param name="context"></param>
    /// <param name="server"></param>
    /// <param name="property"></param>
    /// <param name="device"></param>
    /// <param name="index"></param>
    /// <param name="fieldindex"></param>
    /// <param name="sample"></param>
    /// <param name="start"></param>
    /// <param name="stop"></param>
    /// <param name="data"></param>
    /// <param name="src"></param>
    /// <param name="timeout"></param>
    /// <param name="extendedRange"></param>
    /// <returns></returns>
    public static int GetArchivedData(String context, String server, String property, String device, int index, int fieldindex, int sample, double start, double stop, Object data, HstSrc src, uint timeout, Boolean extendedRange)
    {
      if (extendedRange)
      {
        start -= HEARTBEAT * 2;
        stop += HEARTBEAT * 2;
      }
      return getDataArray(context, server, property, device, index, fieldindex, sample, start, stop, data, src, timeout);
    }
    /// <summary>
    /// Gets archived data for the parameters given over the time range specified.
    /// </summary>
    /// <param name="context"></param>
    /// <param name="server"></param>
    /// <param name="property"></param>
    /// <param name="device"></param>
    /// <param name="start"></param>
    /// <param name="stop"></param>
    /// <param name="data"></param>
    /// <param name="src"></param>
    /// <param name="timeout"></param>
    /// <param name="extendedRange"></param>
    /// <returns></returns>
    public static int GetArchivedData(String context, String server, String property, String device, double start, double stop, Object data, HstSrc src, uint timeout, Boolean extendedRange)
    {
      if (extendedRange)
      {
        start -= HEARTBEAT * 2;
        stop += HEARTBEAT * 2;
      }
      return getDataArray(context, server, property, device, 0, 0, 0, start, stop, data, src, timeout);
    }
    /// <summary>
    /// Gets archived data for the parameters given at the time specified
    /// </summary>
    /// <param name="context"></param>
    /// <param name="server"></param>
    /// <param name="property"></param>
    /// <param name="device"></param>
    /// <param name="targetTime"></param>
    /// <param name="data"></param>
    /// <param name="fieldindex"></param>
    /// <returns></returns>
    public static int GetArchivedDataSnapshot(String context, String server, String property, String device, [In, Out] double targetTime, Object data, int fieldindex)
    {
      int cc = 0, inptLength = 2;
      double[] dstartstop = new double[8];
      TDataType dout, din;

      if (server == null && context == null) return Errors.argument_list_error;
      if (data == null) return Errors.argument_list_error;
      if (context == null) context = "DEFAULT";
      if (server == null) server = "HISTORY"; // assume the standard central archive server for the given context
      Boolean iscentral = isCentralArchiveServer(server);
      inptLength = makeStartStopInput(targetTime, targetTime, 0, fieldindex, 0, iscentral, null, dstartstop);
      din = new TDataType(dstartstop, 0, inptLength);
      //if (data is DIMAGE)
      //{
      //  dout = new TDataType((DIMAGE)data);
      //}
      //else 
      if (data is Array) 
      {
        dout = new TDataType((Array)data);
      }
      else
      {
        return Errors.illegal_format;
      }

      String dname = "/" + context + "/" + server + "/" + device;
      String[] pname = new String[2];
      pname[0] = property;
      int ntries = 1;
      if (!iscentral)
      { // not a call to a central archive server
        Boolean isqp = isArchiveQueryProperty(property);
        if (exHstSrc == ExclusiveHistorySource.central)
        {
          if (!isqp)
          {
            pname[0] = property + ".ARCH";
          }
          else if (property.EndsWith(".HIST"))
          {
            return Errors.not_applicable;
          }
        }
        else if (!isqp)
        { // meta extension not given, try this ...
          pname[0] = property + ".HIST";
          pname[1] = property + ".ARCH";
          ntries = 2;
        }
      }
      int dlen = 0;
      for (int i = 0; i < ntries; i++)
      {
        TLink tl = new TLink(dname, pname[i], dout, din, Access.CA_READ);
        cc = tl.Execute(2000);
        targetTime= tl.GetCompletionDataTimeStamp();
        if (cc == 0 && (dlen=tl.GetCompletionDataSize()) > 0) break;
      }
      return cc;
    }
  }
  /// <summary>
  /// TKernel is the TINE Kernel (singleton) class, whose static methods are used to manage 
  /// the TINE engine.
  /// </summary>
  public class TKernel
  {
    private static sbyte[] cycleStateStringBytes = new sbyte[Constants.STATUS_SIZE];
    private static String cycleStateString = null;
    private static int cycleStateCode = -1;
    public static String GetServerCycleState()
    {
      char[] stschr = new char[Constants.STATUS_SIZE];
      int sts = native.GetServerCycleState(cycleStateStringBytes);
      if (sts == cycleStateCode) return cycleStateString;
      cycleStateCode = sts;
      int i;
      for (i = 0; i < Constants.STATUS_SIZE && cycleStateStringBytes[i] != 0; i++) stschr[i] = (char)cycleStateStringBytes[i];
      cycleStateString = new String(stschr,0,i);
      return cycleStateString;
    }
    public static void SetSuspendCallbacks(Boolean value)
    {
      native.SetSuspendCallbacks(value ? -1 : 0);
    }
    public static Boolean GetSuspendCallbacs()
    {
      return native.GetSuspendCallbacks() == 0 ? false : true;
    }
    public static void SetRespondToServiceRequests(Boolean value)
    {
      native.SetRespondToServiceRequests(value ? -1 : 0);
    }
    public static void SetAllowNetworkAddressResolution(Boolean value)
    {
      native.SetAllowNetworkAddressResolution(value ? -1 : 0);
    }

    public static String GetVersion()
    {
      IntPtr vp = native.GetSystemVersionString();
      return Marshal.PtrToStringAnsi(vp);
    }

    /// <summary>
    /// Sets the server-side work area size which is coupled to the maximum transport size for 
    /// supported by a single-threaded server.
    /// </summary>
    /// <remarks>
    /// Server-side equipment function calls are double buffered, one buffer being a dynamic allocation
    /// of the caller's requested data size, the other buffer being in the case of single threaded servers,
    /// a temporary use of the server's work area.  In order to allow transfer of 'large' data sets, this
    /// size needs to be adjusted accordingly prior to server initialization.  Note that multi-threaded
    /// servers will dynamically allocate both buffers on a call by call basis.
    /// </remarks>
    /// <param name="sizeInBytes">is the size in bytes to be reserved for the server's work area and hence the maximum
    /// allowed transport size for any property supported by the server.
    /// (default: 64 KBytes on most operating systems).</param>
    /// <returns>The current setting for the server's work area.</returns>
    public static UInt32 SetWorkAreaSize(UInt32 sizeInBytes)
    {
      return native.SetWorkAreaSize(sizeInBytes);
    }
    /// <summary>
    /// Gets the server-side work area size which is coupled to the maximum transport size for 
    /// supported by a single-threaded server.
    /// </summary>
    /// <remarks>
    /// Server-side equipment function calls are double buffered, one buffer being a dynamic allocation
    /// of the caller's requested data size, the other buffer being in the case of single threaded servers,
    /// a temporary use of the server's work area.  In order to allow transfer of 'large' data sets, this
    /// size needs to be adjusted accordingly prior to server initialization.  Note that multi-threaded
    /// servers will dynamically allocate both buffers on a call by call basis.
    /// </remarks>
    /// <returns>The current setting for the server's work area.</returns>
    public static UInt32 GetWorkAreaSize()
    {
      return native.GetWorkAreaSize();
    }
    /// <summary>
    /// Assigns specific buffer capacity for certain configurable TINE sockets.
    /// </summary>
    /// <remarks>
    ///The standard TINE client and server sockets (both UDP, TCP, and STREAM) are initiated with a 
    ///reasonable default amount of receive (64 Kbyte) and send (32 Kbyte) buffer capacity.  If for any
    ///performance reason, more or less buffer capactiy is needed, this routine should be called prior to
    ///any calls to SystemInit() or SystemCycle().  For example, VxWorks builds which do not have a large 
    ///amount of network resources might want to set these values smaller than the defaults.  Video system
    ///clients expecting extreme amounts of incoming data might want to set these values to a much larger
    ///amount.
    ///
    ///This call allows the setting of the server side receive buffer space as well as the client side
    ///send buffer space. This is potentially applicable in rare cases (middle layer servers) where large
    ///data payloads are sent as input from a client to a server.
    ///
    ///As in the case of _SystemAssignBufferSpace(), the UDP specific send buffer size is kept to a reasonable
    ///value, no larger than twice the registered packet MTU.  UDP datagram transmission is not inhibited by
    ///any client-server flow control which might cause the send buffer space to stack up.
    /// </remarks>
    /// <param name="clnRcvBufferSpace">is the amount of receive buffer space (in bytes) to assign to the configurable 
    /// TINE client sockets. A value of '0' will assign the default receive buffer capacity.</param>
    /// <param name="srvRcvBufferSpace">is the amount of receive buffer space (in bytes) to assign to the configurable TINE
    /// server sockets. A value of '0' will assign the same value as clnRcvBufferSpace (and damping this value by a
    /// factor of 2 should it exceed the default receive buffer capacity).</param>
    /// <param name="clnSndBufferSpace">s the amount of send buffer space (in bytes) to assign to the configurable TINE 
    /// client sockets. A value of '0' will assign the same value as srvSndBufferSpace (and damping this value by a 
    /// factor of 2 should it exceed the default receive buffer capacity).</param>
    /// <param name="srvSndBufferSpace">is the amount of send buffer space (in bytes) to assign to the configurable  
    /// TINE server TCP or STREAM sockets. 
    /// The UDP socket buffer space will be at least as large as twice the assigned UDP packet MTU.
    /// A value of 0' will assign the default send buffer capacity.</param>
    /// 
    /// <returns>0 on success or a TINE error code</returns>
    public static Int32 AssignBufferSpace(UInt32 clnRcvBufferSpace, UInt32 srvRcvBufferSpace, UInt32 clnSndBufferSpace, UInt32 srvSndBufferSpace)
    {
      return native._SystemAssignBufferSpaceEx(clnRcvBufferSpace,srvRcvBufferSpace,clnSndBufferSpace,srvSndBufferSpace);
    }
    /// <summary>
    /// Assigns specific buffer capacity for certain configurable TINE sockets.
    /// </summary>
    /// <remarks>
    /// The standard TINE client and server sockets (both UDP, TCP, and STREAM) are initiated with a 
    /// reasonable default amount of receive (64 Kbyte) and send (32 Kbyte) buffer capacity.  If for any
    /// performance reason, more or less buffer capactiy is needed, this routine should be called prior to
    /// any calls to SystemInit() or SystemCycle().  For example, VxWorks builds which do not have a large 
    /// amount of network resources might want to set these values smaller than the defaults.  Video system
    /// clients expecting extreme amounts of incoming data might want to set these values to a much larger
    /// amount.
    ///
    /// The principal benificiaries of the 'receive buffer space' are the client sockets (UDP,TCP and SREAM), whereas
    /// the principal benificiaries of the 'send buffer space' are the server sockets (TCP and STREAM). The
    /// remaining socket buffer space (e.g. client-side send buffer space or server side receive buffer space)
    /// will be held at 'reasonable' values, the assumption being that the dominant data delivery will generally
    /// server (sending) to client (receiving).
    /// </remarks>
    /// <param name="rcvBufferSpace">s the amount of receive buffer space (in bytes) to assign to the configurable TINE sockets.
    /// A value of '0' will assign the default receive buffer capacity.</param>
    /// <param name="sndBufferSpace">sndBufferSpace is the amount of send buffer space (in bytes) to assign to the configurable TINE TCP or  
    /// STREAM sockets.  The UDP socket send buffer space if not influenced by this parameter and 
    /// will be at least as large as twice the assigned UDP packet MTU. See _SystemAssignUdpSendBufferSpace().
    /// A value of 0' will assign the default send buffer capacity.</param>
    /// <returns>0 on success or a TINE error code</returns>
    public static Int32 AssignBufferSpace(UInt32 rcvBufferSpace, UInt32 sndBufferSpace)
    {
      return native._SystemAssignBufferSpaceEx(rcvBufferSpace, 0, 0, sndBufferSpace);
    }
    /// <summary>
    /// Appends the given text to the FEC process's log file.
    /// </summary>
    /// <remarks>
    /// Each log entry is prefixed by the FEC process name (i.e. the 'FECNAME') and the time
    /// and date.  The location of the log file (named 'fec.log' and rotated once into 'fec.bak')
    /// is determined by the environment variable FEC_LOG (first choice) or FEC_HOME. If
    /// neither environment variable is set, then the fec.log file will be written to the
    /// current working directory.
    /// </remarks>
    /// <param name="text">is the desired text to be appended to the FEC log file</param>
    /// <returns>0 upon success or a TINE error code (e.g. file_error)</returns>
    public static Int32 FecLog(string text)
    {
      return native.feclog(text);
    }
    /// <summary>
    /// Sets the approximate maximum number of log entries in the FEC process's log file.
    /// </summary>
    /// <remarks>
    /// The number of long entries is approximately deterimed by assuming the the average
    /// line of text contains 160 characters.  If the number so deterimed exceeds the
    /// given FEC log depth, the current log file 'fec.log' will be rotated into
    /// 'fec.bak'.
    /// </remarks>
    /// <param name="depth">
    /// is the maximum number of lines the FEC's log file is allowed to have.
    /// (default = 1000)
    /// </param>
    public static void SetFecLogDepth(Int32 depth)
    {
      native.SetFeclogDepth(depth);
    }
    /// <summary>
    /// Retrieves the current setting for the maximum number of log entries in the FEC process's log file.
    /// </summary>
    /// <returns>The current setting.</returns>
    public static Int32 GetFecLogDepth()
    {
      return native.GetFeclogDepth();
    }
    /// <summary>
    /// Appends the given text to the designated log file.
    /// </summary>
    /// <param name="filename">
    /// is the file name to which the log entry is to be appended.  The file will be created in the
    /// 'FEC_LOG' or 'FEC_HOME' directory (as in the case of fec.log), and rotated using the same
    /// log depth established via SetFecLogDepth().
    /// </param>
    /// <param name="tag">
    /// is a 'procedure tag' which will prefix the given log entry.  If ommitted, the tag 'logger' will
    /// be used.
    /// </param>
    /// <param name="text">
    /// is the desired log entry text.
    /// </param>
    /// <returns>
    /// 0 upon success or a TINE error code.
    /// </returns>
    public static Int32 SrvLog(string filename, string tag, string text)
    {
      return native.srvlog(filename,tag,text);
    }
    /// <summary>
    /// Initializes the TINE server engine.
    /// </summary>
    /// <remarks>
    /// This method MUST be invoked for the TINE server engine to function property.
    /// Should be called prior to the start of the main TINE server engine. Any deviations from default system parameters should be made before 
    /// server initialization. This might include server settings which fix the size of the local connection table entries, the maximum date transport 
    /// size, etc. Note that InitializeServer() will begin allocating space for its connection tables, so modifying these values away from their defaults 
    /// will have no effect after InitializeServer() has been called. Also, if the FEC Name is given by calling RegisterFecName() then this should also 
    /// preceed a call to InitializeServer().
    /// </remarks>
    /// <returns>0 if successful, otherwise a TINE completion code.</returns>
    public static Int32 InitializeServer()
    {
      return native._SystemInit(1);
    }
    /// <summary>
    /// Required call for the TINE engine to function property.
    /// </summary>
    /// <remarks>
    /// On a multithreaded platform (e.g. all .NET implementations) loops around  _SystemCycle().
    /// In a single thread environment,e.g., MSDOS blocks the main process.
    /// </remarks>
    public static void StartServices()
    {
      native.SystemSetCycleTimer();
    }
    /// <summary>
    /// Sets the intrinsic data timestamp to the value given. 
    /// </summary>
    /// <remarks>
    /// This call is typically used inside an equipment 
    /// function to set the data timestamp associated with the request being made. For instance, when the data 
    /// are taken from the hardware IO loop, a call to makeDataTimeStamp() is made to get the current system timestamp. 
    /// Then when a request for data is made the caller can know the precise time of data acquisition. 
    /// If this routine is not called, then the current system time will be used to stamp the data.
    /// </remarks>
    /// <param name="timestamp">The timestamp to be associated with the next data set.</param>
    public static void SetDataTimeStamp(double timestamp)
    {
      native.SetDataTimeStamp(timestamp);
    }
    /// <summary>
    /// Provides the last established data timestamp.
    /// </summary>
    /// <remarks>
    /// A call to getDataTimeStamp() immediately following a call to ExecLink() or within a 
    /// callback routine will yield the data timestamp of the call.
    /// Note: This call is not thread-safe. Synchronous data acquisition on two different threads 
    /// could conceivably interfer with the value returned from this routine. A thread-safe way of 
    /// obtaining the data timestamp associated with a call is to use the .dTimestamp field of the DTYPE 
    /// object for synchronous calls such as ExecLink() or ExecLinkEx() or to use getCurrentDataTimeStampFromCallbackId() 
    /// or getCurrentDataTimeStamp() (which take a link identifier as input parameter) for asynchronous links.
    /// </remarks>
    /// <returns>The data timestamp currently in place.</returns>
    public static double GetDataTimeStamp()
    {
      return native.GetDataTimeStamp();
    }
    /// <summary>
    /// Makes a data timestamp according to the current system time.
    /// </summary>
    /// <returns>A tine data timestamp.</returns>
    public static double MakeDataTimeStamp()
    {
      return native.MakeDataTimeStamp();
    }
    /// <summary>
    /// Provides the last established system data stamp
    /// </summary>
    /// <remarks>
    /// A server will automatically obtain and update the cannonical system data stamp for its context
    /// if a CYCLER is available for that context.  The system data stamp logically represents a
    /// 'cycle number' or 'pulse number' which is used to systematically tag data objects.  If no
    /// CYCLER exists, then the value for the system data stamp will always be = 0.
    /// </remarks>
    /// <returns>The system data stamp current in place.</returns>
    public static int GetSystemDataStamp()
    {
      return native.GetSystemDataStamp(-1);
    }
    /// <summary>
    /// A server will automatically obtain and update the cannonical system data stamp for its context
    /// if a CYCLER is available for that context.  The system data stamp logically represents a
    /// 'cycle number' or 'pulse number' which is used to systematically tag data objects.  If no
    /// CYCLER exists, then the value for the system data stamp will always be = 0.  This call 
    /// is in parallel with MakeDataTimeStamp() and is nominally equivalent to GetSystemDataStamp.
    /// </summary>
    /// <returns>The system data stamp current in place.</returns>
    public static int MakeSystemDataStamp()
    {
      return native.MakeSystemDataStamp();
    }
    /// <summary>
    /// A server will automatically obtain and update the cannonical system data stamp for its context
    /// if a CYCLER is available for that context.  The system data stamp logically represents a
    /// 'cycle number' or 'pulse number' which is used to systematically tag data objects.  If no
    /// CYCLER exists, then the value for the system data stamp will always be = 0.  A server may need to
    /// record the current system stamp at the time of e.g. a hardware trigger and apply the value in
    /// subsequent calls to access data.  This method can be used toward these ends.
    /// </summary>
    /// <param name="value">
    /// is the value of the system data stamp to apply to current data access.
    /// </param>
    public static void SetSystemDataStamp(int value)
    {
      native.SetSystemDataStamp(value);
    }
    /// <summary>
    /// Assigns a FEC Name and descriptive information to the server process. 
    /// </summary>
    /// <remarks>
    /// Servers must be assigned to a Front End Computer (FEC), which for many operating systems corresponds 
    /// to a process running on the computer rather than the computer itself (meaning there can in some cases 
    /// be more than one servers with different 'FEC' names running on the same computer. There can be several 
    /// device servers attached to the same FEC (i.e. sharing the same address space), so the FEC name is a distinct 
    /// quantity from the device server name (see RegisterExport()), and this name must also be system-wide unique. 
    /// The call RegisterFecInformation() supercedes RegisterFecNameEx() and RegisterFecName() and differs from them 
    /// in that it accepts 'Context' and 'SubSystem' as parameters, both of which will be applied to all device 
    /// servers attaching to the same FEC name.
    /// The FEC name can also be (and frequently is) registered via the presence of an 'fecid.csv' startup database file.
    /// </remarks>
    /// <param name="fecName">The FEC name indentifying the Front End Computer and to which all registered equipment modules are bound.</param>
    /// <param name="defaultSubSystem">The subsystem to which attached device servers belong. Note that the subsystem (unlike context) is itself 
    /// not part of the name space, although the name space can be queried for device servers belonging to a subsystem. 
    /// This information is now deduced automatically from the library build.</param>
    /// <param name="defaultContext">A 32-character string giving the context of all registered equipment modules. 
    /// (earlier versions of TINE truncated at 16 characters).</param>
    /// <param name="fecDescription">A 64-character desciption of the FECs server duties. A 'subsystem' can be associated with ALL equipment modules 
    /// found on the FEC by prefixing the description with a subsystem tag. This information will be parsed by the Equipment name 
    /// server and used for sub-system specific queries. For example, the associate the FEC "HEPHF" with the subsystem "RF", the description field 
    /// might be "[RF]Hera Proton HF Control".</param>
    /// <param name="fecLocation">A 32-character string giving the phyical location of the FEC.</param>
    /// <param name="attachedHardware">A 32-character brief description of the IO hardware found on the FEC.</param>
    /// <param name="responsible">A 32-character string listing the developer(s) responsible for the FEC. 
    /// Note: The Equipment Name Server will allow the removal of FECs and associated equipment modules to the user(s) specified here.</param>
    /// <param name="portOffset">The 'Port Offset' to be applied to the FEC. This parameter plays an important role where more that 
    /// one FEC is to run on a machine running an operating system using virtual memory. In such cases 'Front End Computer' is a misnomer, 
    /// since "FEC" really refers to a process running on the computer to which one or more equipment modules are bound. For operating systems 
    /// where all processes run in the same address space (such as VxWorks, MS-DOS, Win16, NIOS) there is in fact only one such processes managing 
    /// all requistered equipment modules. For systems using virtual memory (such as Linux, Solaris, HP-UX, Win32, etc.), you can have many such 
    /// processes running independently of one another. Since each such process must listen on a unique set of server ports, you must see to it that 
    /// all such FEC processes are registered with a unique 'Port Offset'.</param>
    /// <returns>0 if successful, otherwise a TINE completion code which can be interpreted by a call to GetLastLinkError()</returns>
    public static Int32 RegisterFecInformation(String fecName,String defaultSubSystem,String defaultContext,String fecDescription,String fecLocation,String attachedHardware,String responsible,UInt16 portOffset)
    {
      return native.RegisterFecInformation(fecName, defaultSubSystem, defaultContext, fecDescription, fecLocation, attachedHardware, responsible, portOffset);
    }
    /// <summary>
    /// Obtains the FEC port offset appropriate for the give FEC name.
    /// </summary>
    /// <remarks>
    /// In cases where the pure registration API routines are used to register FEC and server information (as opposed to a common fecid.csv 
    /// configuration file) it is often a more cumbersome task to ensure a unique FEC port on the host machine. This routine can be used to 
    /// this end. It will search the existing FEC manifest on the current host in order to find the matching entry for the given FEC name. 
    /// If the entry is not found, then the 'next free port' is returned. In this way the FEC can always register with its own proper FEC port, 
    /// without the need to hard code this number.
    /// 
    /// If this routine is called after the server is running and the FEC name is not found in the manifest the -name_unknown is returned.
    ///
    /// </remarks>
    /// <param name="fecName"></param>
    /// <returns>
    /// the port offset applicable for the given FEC name or the a negative number giving the negative of the TINE return code indicating an error.
    /// </returns>
    public static Int32 GetPortOffset(String fecName)
    {
      return native.GetPortOffset(fecName);
    }
    /// <summary>
    /// Assigns pertinent information for the specified property. 
    /// </summary>
    /// <remarks>
    /// Servers export their behavior through Properties, which can be thought of as 'properties' 
    /// in the Object Oriented sense with Get and Set methods. That is, any read or write access 
    /// for a given property will call into the associated equipment module and be handled there. 
    /// The general behavior of the property (including descriptions) can be specified with this call. 
    /// Properties can also be (and frequently are) registered via the presence of an 'exports.csv' 
    /// startup database file.
    /// Note: Properties which are overloaded can and should be registered more than once with the 
    /// appropriate characteristics supplied in the registration call.
    /// </remarks>
    /// <param name="eqm">A local equipment module name (maximum 6 characters in length) For example: "BPMEQM".</param>
    /// <param name="prpName">Property name in question (up to 32 characters, preferably 16 or less).</param>
    /// <param name="dtOut">Atemplate data object giving the default format size, and structure tag 
    /// (if any) of the output data set to be returned by the server to the caller. </param>
    /// <param name="dtIn">A template data object giving the default format size, and structure tag 
    /// (if any) of the input data set to be supplied by the caller. </param>
    /// <param name="accessMode">the data 'access' (e.g. CA_READ or CA_WRITE) accepted by the server. This parameter can 
    /// be used to coerce client-side links into using preferred link attributes. For example, by applying the CA_NETWORK 
    /// flag (CA_READ|CA_NETWORK) to the access, asynchronous NETWORK (i.g. multicast acquisition) will be enforced at the 
    /// client side. Applying the CA_XREAD access flag will likewise allow exclusive READ access regarding the registered property 
    /// (i.e. the same security rules applied to CA_WRITE will be in force for READ acquisitions, either permanently - no CA_READ flag, 
    /// or temporarily in case of an applied access lock - if both CA_READ and CA_XREAD are applied). Applying the CA_SAVERESTORE flag 
    /// will apply the internal save-and-restore logic to the registered property (restoring the last property settings upon server 
    /// startup, and saving any changed property settings).</param>
    /// <param name="arrayType">Data 'array type' (AT_SCALAR, AT_SINGLE, etc.) associated with the returned data.</param>
    /// <param name="rowLength">if > 0 gives the row length of the returned array data. This is most relevant for double arrays, where 
    /// the number of rows is then given by the output data size divided by the row length.</param>
    /// <param name="prpDescription">The 64-character description of what the property reads or writes. </param>
    /// <param name="prpId">The property identifier to be associated with the property name returned in calls to GetPropertyId()</param>
    /// <param name="redirectionTarget">A redirection string, if the property is to be redirected to another server. 
    /// Most properties are not redirected, and this parameter is usually NULL. If used, it will be parsed according to 
    /// device server/device name (device property), where the 'device name' and 'device property' are optional. 
    /// If 'device name' is not present, all devices will be redirected to the specified device server. If 'device property' 
    /// is not present then the identical property name as registered will be redirected to the device server specified. 
    /// If on the other hand, the device property is present in the string to be parsed, then the property registered 
    /// will be redirected to the device server specified as well as to the device property specified.</param>
    /// <returns>The (positive) associated property identifier if successful, otherwise the negative of a TINE return code.</returns>
    public static Int32 RegisterPropertyInformation(String eqm, String prpName, TDataType dtOut, TDataType dtIn, UInt16 accessMode, UInt16 arrayType, UInt16 rowLength, String prpDescription, Int32 prpId, String redirectionTarget)
    {
      DTYPE dout = new DTYPE();
      DTYPE din = new DTYPE();
      if (dtOut != null)
      {
        dout.dArrayLength = dtOut.GetDataArrayLength();
        dout.dFormat = (Int16)dtOut.GetDataFormat();
        dout.dTag = new char[Constants.TAG_NAME_SIZE];
        String dtag = dtOut.GetDataTag();
        Int32 tlen = dtag != null ? Math.Min(Constants.TAG_NAME_SIZE, dtag.Length) : 0;
        if (tlen > 0 && dtag != null)
        {
          Array.Copy(dtag.ToCharArray(), dout.dTag, (int)Math.Min(Constants.TAG_NAME_SIZE, dtag.Length));
        }
      }
      if (dtIn != null)
      {
        din.dArrayLength = dtIn.GetDataArrayLength();
        din.dFormat = (Int16)dtIn.GetDataFormat();
        din.dTag = new char[Constants.TAG_NAME_SIZE];
        String dtag = dtIn.GetDataTag();
        Int32 tlen = dtag != null ? Math.Min(Constants.TAG_NAME_SIZE, dtag.Length) : 0;
        if (tlen > 0 && dtag != null)
        {
          Array.Copy(dtag.ToCharArray(), din.dTag, (int)Math.Min(Constants.TAG_NAME_SIZE, dtag.Length));
        }
      }
      return native.RegisterPropertyInformation(eqm, prpName, ref dout, ref din, accessMode, arrayType, rowLength, prpDescription, prpId, redirectionTarget);
    }    
    /// <summary>
    /// Assigns a device name to the specified device number.
    /// </summary>
    /// <remarks>
    /// Many servers handle instances of a given device type, such as a set of vacuum pumps, 
    /// beam position monitors, magnet power supply controllers, etc. It is frequently very efficient 
    /// to code in terms of device numbers which might be an index into an array, but to specify modules 
    /// in calls by their human-readable names. For such purposes a device name can be assigned to a number 
    /// at initialization via a call to RegisterDeviceName() (alias: RegisterModuleName()). An alternative 
    /// is to provide a startup database file 'devices.csv' containing a cross-reference for numbers and names. 
    /// Internally, a hash table is maintained for fast lookups inside equipment module routines.
    /// </remarks>
    /// <param name="eqm">The local equipment module name (maximum 6 characters in length) For example: "BPMEQM".</param>
    /// <param name="devName">Up to 64 characters device name to be assigned</param>
    /// <param name="devNr">The device number associated with the device name specified.</param>
    /// <returns>0 if successful, otherwise a TINE completion code</returns>
    public static Int32 RegisterDeviceName(String eqm, String devName,Int32 devNr)
    {
      return native.RegisterDeviceName(eqm, devName, devNr);
    }
    /// <summary>
    /// Assigns a description to a registered device
    /// </summary>
    /// <param name="eqm">The local equipment module name (maximum 6 characters in length) For example: "BPMEQM".</param>
    /// <param name="devName">Up to 64 characters device name for which the description is to be assigned</param>
    /// <param name="description">The device description to assign to the device</param>
    /// <returns>0 if successful, otherwise a TINE completion code</returns>
    public static Int32 SetDeviceDescription(String eqm, String devName, String description)
    {
        return native.SetDeviceDescription(eqm, devName, description);
    }
    /// <summary>
    /// Assigns a location to a registered device
    /// </summary>
    /// <param name="eqm">The local equipment module name (maximum 6 characters in length) For example: "BPMEQM".</param>
    /// <param name="devName">Up to 64 characters device name for which the location is to be assigned</param>
    /// <param name="location">The device location to assign to the device</param>
    /// <returns>0 if successful, otherwise a TINE completion code</returns>
    public static Int32 SetDeviceLocation(String eqm, String devName, String location)
    {
        return native.SetDeviceLocation(eqm, devName, location);
    }
    /// <summary>
    ///  Assigns an alarm region to a registered device
    /// </summary>
    /// <param name="eqm">The local equipment module name (maximum 6 characters in length) For example: "BPMEQM".</param>
    /// <param name="devName">Up to 64 characters device name for which the region is to be assigned</param>
    /// <param name="region">The alarm region to assign to the device</param>
    /// <returns>0 if successful, otherwise a TINE completion code</returns>
    public static Int32 SetDeviceRegion(String eqm, String devName, String region)
    {
        return native.SetDeviceRegion(eqm, devName, region);
    }
    /// <summary>
    /// Assigns a mask to a registered device
    /// </summary>
    /// <param name="eqm">The local equipment module name (maximum 6 characters in length) For example: "BPMEQM".</param>
    /// <param name="devName">Up to 64 characters device name for which the location is to be assigned</param>
    /// <param name="mask">The device mask to assign to the device</param>
    /// <returns>0 if successful, otherwise a TINE completion code</returns>
    public static Int32 SetDeviceMask(String eqm, String devName, int mask)
    {
        return native.SetDeviceMask(eqm, devName, mask);
    }
    /// <summary>
    /// Assigns an offline status to a registered device
    /// </summary>
    /// <param name="eqm">The local equipment module name (maximum 6 characters in length) For example: "BPMEQM".</param>
    /// <param name="devName">Up to 64 characters device name for which the location is to be assigned</param>
    /// <param name="offline">The offline status: 'true' => device is offline, 'false' => device is online</param>
    /// <returns>0 if successful, otherwise a TINE completion code</returns>
    public static Int32 SetDeviceOfflineStatus(String eqm, String devName, Boolean offline)
    {
        return native.SetDeviceOfflineStatus(eqm, devName, (offline ? -1 : 0));
    }
    /// <summary>
    /// Assigns an Z-position to a registered device
    /// </summary>
    /// <param name="eqm">The local equipment module name (maximum 6 characters in length) For example: "BPMEQM".</param>
    /// <param name="devName">Up to 64 characters device name for which the location is to be assigned</param>
    /// <param name="zposition">The Z (longitundinal) position of the device in the machine</param>
    /// <returns>0 if successful, otherwise a TINE completion code</returns>
    public static Int32 SetDeviceZPosition(String eqm, String devName, float zposition)
    {
        return native.SetDeviceZPosition(eqm, devName, zposition);
    }
    public static void SetHistoryFilesRepository(String path)
    {
      native.SetHistoryFilesRepository(path);
    }
    public static void SetAppVersion(int major, int minor, int revision)
    {
      native.SetAppVersion(major, minor, revision);
    }
    private static DateTime epoch = new DateTime(1970, 1, 1, 0, 0, 0, 0);
    public static void SetAppDate(DateTime dt)
    {
      IntPtr t=(IntPtr)0;
      TimeSpan span = (dt - epoch);
      t = (IntPtr)span.TotalSeconds;
      native.SetAppDate(t);
    }
    public static int AppendHistoryInformation(String eqm, string prp, string dev, Int32 len, Int32 fmt, Int32 idx, ref HistorySpecification hspec)
    {
      return native.AppendHistoryInformation(eqm, prp, dev, len, fmt, idx, ref hspec);
    }
    public static int AppendAlarmInfoTable(String eqm, ref ADS ads)
    {
      return native.AppendAlarmInfoTable(eqm, ref ads);
    }
    public static void SetMinimumDiskSpaceInBlocks(uint spaceInBlocks)
    {
      native.SetMinimumDiskSpaceInBlocks(spaceInBlocks);
    }
    public static void SetDieOnAddressInUse(Boolean value)
    {
      native.SetDieOnAddressInUse(value ? 1 : 0);
    }
    public static void SetDieOnSocketError(Boolean value)
    {
      native.SetDieOnSocketError(value ? 1 : 0);
    }
    public static void SetDieOnFecIsAlias(Boolean value)
    {
      native.SetDieOnFecIsAlias(value ? 1 : 0);
    }
    public static Int32 SetFecHome(string path)
    {
      return native.SetFecHome(path);
    }
    public static Int32 SetFecLogPath(string path)
    {
      return native.SetFecLogPath(path);
    }
    public static Boolean IsStandAlone()
    {
      return native.IsStandAlone() == 0 ? false : true;
    }
    public static void SetConnectionTableCapacity(int size)
    {
      native.SetConnectionTableCapacity(size);
    }
    public static ConTblInfo[] GetConnectionTable()
    { 
      int nc = 1024; //TODO: make this dynamic
      ConTblInfo[] ret = null;
      unsafe 
      {
        ConTblInfo[] inf = new ConTblInfo[1024];
        if (native.GetConnectionTable(inf, ref nc) != 0) return null;
        ret = new ConTblInfo[nc];
        Array.Copy(inf, ret, nc);
      }
      return ret;
    }
    public static String[] GetConnectionList()
    {
      int nc = 1024; //TODO: make this dynamic
      unsafe
      {
        char[] cbuf = new char[nc*128];
        if (native.GetConnectionList(cbuf, nc*128) != 0) return null;
        String s = new String(cbuf);
        return s.Split("\n".ToCharArray());
      }
    }
    public static String[] GetMessageList()
    {
      int nc = 512; //TODO: make this dynamic
      unsafe
      {
        char[] cbuf = new char[512 * 256];
        if (native.GetMessageList(cbuf, nc * 256) != 0) return null;
        String s = new String(cbuf);
        return s.Split("\n".ToCharArray());
      }
    }
    public static MsgType[] GetMessageTable()
    {
      int nc = 1024; //TODO: make this dynamic
      MsgType[] ret = null;
      unsafe
      {
        MsgType[] tbl = new MsgType[1024];
        if (native.GetMsgTable(tbl, ref nc) != 0) return null;
        ret = new MsgType[nc];
        Array.Copy(tbl, ret, nc);
      }
      return ret;
    }
    public static String GetArchiveAnnotation(String context,String server,String property,double targetTime)
    {
      String annoStr = null;
      int siz = 512;
      char[] buf = new char[512];
      int cc = native.GetArchiveAnnotation(context, server, property, targetTime, buf, siz, 1000);
      if (cc == 0)
      {
        //annoStr = new String(buf).Trim();
        annoStr = StrUtils.Trim(buf);
      }
      return annoStr;
    }
    public static double[] GetArchiveAnnotationTimes(String context,String server,String property,double start, double stop)
    {
      unsafe
      {
        double[] times = new double[256];
        int siz = 256;
        int cc = native.GetArchiveAnnotationTimes(context, server, property, start, stop, TDataType.GetDataPtr(times), Formats.LFMT_OFFSET+Formats.CF_DOUBLE, ref siz, 1000);
        if (cc == 0)
        {
          double[] rtimes = new double[siz];
          Array.Copy(times, rtimes, siz);
          return rtimes;
        }
      }
      return null;
    }
    public static int SetArchiveAnnotation(String context,String server,String property,double targetTime,string txt,Boolean keepAlive)
    {
      int ka = keepAlive ? 1 : 0;
      int cc = native.SetArchiveAnnotation(context, server, property, targetTime, txt, ka, 1000);
      return cc;
    }
    public static Boolean IsStockProperty(String property)
    {
      return native.IsStockProperty(property) != 0;
    }
    public static int SetAccessLock(String context, String server, int lockType, int lockDuration)
    {
      return native.SetAccessLock(context, server, lockType, lockDuration);
    }
    public static int FreeAccessLock(String context, String server)
    {
      return native.FreeAccessLock(context, server);
    }
    public static int GetAccessLockStatus(String context, String server)
    {
      return native.GetAccessLockStatus(context, server);
    }
    public static int GetAccessLockInformation(String context, String server, [In, Out] Char[] callerName, [In, Out] Char[] callerIp, [In, Out] Char[] timeLeft)
    {
      NAME32[] cllrName = new NAME32[] { new NAME32("") };
      NAME32[] cllrIp = new NAME32[] { new NAME32("") };
      NAME32[] timeLft = new NAME32[] { new NAME32("") };
      IntPtr pName = TDataType.GetDataPtr(cllrName);
      IntPtr pIp = TDataType.GetDataPtr(cllrIp);
      IntPtr pLft = TDataType.GetDataPtr(timeLft);
      int cc = native.GetAccessLockInformation(context, server, pName, pIp, pLft);
      cllrName[0].GetName().ToCharArray().CopyTo(callerName, 0);
      cllrIp[0].GetName().ToCharArray().CopyTo(callerIp, 0);
      timeLft[0].GetName().ToCharArray().CopyTo(timeLeft, 0);
      return cc;
    }

  }
  /// <summary>
  /// Represents a TINE tagged structure.
  /// </summary>
  /// <remarks>
  /// A tagged structure is a user-defined data type that can carry any set of 
  /// TINE data types (including other tagged structures) as an atomic unit.
  /// </remarks>
  public class TTaggedStruct 
  {
    private string tag;
    public string GetTag() { return tag; }
    public void SetTag(string structTag) 
    { 
      tag = structTag;
      if (tag != null && tag.Length > Constants.TAG_NAME_SIZE) 
        tag = tag.Substring(0, Constants.TAG_NAME_SIZE); 
    }
    private string className;
    public string GetClassName() { return className; }
    /// <summary>
    /// The primary constructor for a TINE tagged structure.
    /// </summary>
    /// <remarks>
    /// Using this constructor will assign the structure's 'tag' name from the
    /// .NET structure's class name and the field names from the associated field 
    /// names.
    /// But note: A structure's 'tag' name as well as the field names within a TINE tagged 
    /// structure cannot exceed 16 characters. 
    /// An alternative is to pass a null object and assign the structure's tag name and
    /// fields manually.
    /// </remarks>
    /// <param name="tstruct">is a .NET structure which will be examined and converted
    /// into a TINE tagged structure.  
    /// </param>
    /// \b Example:
    /// \include eg_RegisterStructure.cs
    public TTaggedStruct(object tstruct)
    {
      fixStructLayout(tstruct, null);
    }
    /// <summary>
    /// The primary constructor for a TINE tagged structure.
    /// </summary>
    /// <remarks>
    /// Using this constructor will assign the structure's 'tag' name according to the
    /// tag parameter passed.  If a null or empty string is passed then the constructor
    /// reverts to its simpler form using the 
    /// .NET structure's class name and the field names from the associated field 
    /// names.
    /// But note: A structure's 'tag' name as well as the field names within a TINE tagged 
    /// structure cannot exceed 16 characters. 
    /// An alternative is to pass a null object and assign the structure's tag name and
    /// fields manually.
    /// </remarks>
    /// <param name="tstruct">is a .NET structure which will be examined and converted
    /// into a TINE tagged structure.  
    /// </param>
    /// <param name="tstag">is a structure tag name to be assigned to the structure registered.
    /// </param>
    /// \b Example:
    /// \include eg_RegisterStructure.cs
    public TTaggedStruct(object tstruct, string tstag)
    {
      fixStructLayout(tstruct, tstag);
    }
    private int tsnum = 0;
    private byte[] tsbytes = null;
    public byte[] GetStructByteArray() { return tsbytes; }
    public TTaggedStruct(string server, string tstag, int num)
    {
      int cc;
      if ((cc=TQuery.AcquireAndRegisterStruct(server, tstag, 1)) != 0)
      {
        throw new Exception("cannot register structure "+tstag+" : "+Errors.ToString(cc));
      }
      int tssiz = native.GetStructSize(tstag);
      tag = tstag;
      tsnum = num;
      tsbytes = new byte[tssiz * tsnum];
    }
    public TTaggedStruct(string tstag,int num)
    {
      int tssiz = 0;
      if ((tssiz=native.GetStructSize(tstag)) > 0)
      { // struct is known !
        tsnum = num;
        tag = tstag;
        tsbytes = new byte[tssiz*tsnum];
      }
      else
      {
        throw new Exception("cannot register structure " + tstag + " : structure is unknown");
      }
    }
    private byte[] parseFieldToBytes(string field, Int32 fmt, Int32 siz)
    {
      int fsiz = Formats.GetSizeInBytes(fmt);
      byte[] b = null;
      byte[] ba = new byte[siz * fsiz];
      switch (fmt)
      {
        case Formats.CF_INT16:
          {
            Int16 v = 0;
            for (int i = 0; i < siz; i++)
            {
              if (Int16.TryParse(field, out v)) b = BitConverter.GetBytes(v);
              Array.Copy(b, 0, ba, i * fsiz, fsiz);
            }
          }
          break;
        case Formats.CF_INT32:
          {
            Int32 v = 0;
            for (int i = 0; i < siz; i++)
            {
              if (Int32.TryParse(field, out v)) b = BitConverter.GetBytes(v);
              Array.Copy(b, 0, ba, i * fsiz, fsiz);
            }
          }
          break;
        case Formats.CF_INT64:
          {
            Int64 v = 0;
            for (int i = 0; i < siz; i++)
            {
              if (Int64.TryParse(field, out v)) b = BitConverter.GetBytes(v);
              Array.Copy(b, 0, ba, i * fsiz, fsiz);
            }
          }
          break;
        case Formats.CF_FLOAT:
          {
            float v = 0;
            for (int i = 0; i < siz; i++)
            {
              if (Single.TryParse(field, out v)) b = BitConverter.GetBytes(v);
              Array.Copy(b, 0, ba, i * fsiz, fsiz);
            }
          }
          break;
        case Formats.CF_DOUBLE:
          {
            double v = 0;
            for (int i = 0; i < siz; i++)
            {
              if (Double.TryParse(field,out v)) b = BitConverter.GetBytes(v);
              Array.Copy(b, 0, ba, i * fsiz, fsiz);
            }
          }
          break;
        case Formats.CF_TEXT:
          char[] ca = field.ToCharArray();
          for (int i=0; i< siz && i<ca.Length; i++)
          {
            ba[i] = (byte)ca[i];
          }
          break;
      }
      return ba;
    }
    public TTaggedStruct(string tstag, string[] fieldStrings)
    {
      int tssiz = 0;
      if (fieldStrings == null || fieldStrings.Length == 0)
      {
        throw new Exception("field string array is either null or of zero length");
      }
      if ((tssiz = native.GetStructSize(tstag)) > 0)
      { // struct is known !
        tag = tstag;
        tsbytes = new byte[tssiz];
        // get the struct structure (native call ?)
        Int32 n = Math.Max(fieldStrings.Length + 1, 100);
        NAME64DBLDBL[] sf = new NAME64DBLDBL[n];
        IntPtr psf = TDataType.GetDataPtr(sf);
        if (native.GetStructFormat(tstag, psf, ref n) != 0)
        {
          throw new Exception("cannot get structure information for "+tstag);
        }
        Int32 fmt, siz;
        byte[] ba = null;
        int didx = 0;
        // run thru and set the fields according to string parsing
        for (int i = 0; i < n - 1; i++)
        {
          // field, size, fmt
          siz = (Int32)sf[i].d1val;
          fmt = (Int32)sf[i].d2val;
          fmt -= Formats.LFMT_OFFSET;
          // parse fieldStrings[i] according to fmt
          ba = parseFieldToBytes(fieldStrings[i], fmt, siz);
          // fill in tsbytes // not constructing from known .NET struct object
          Array.Copy(ba, 0, tsbytes, didx, ba.Length);
          didx += ba.Length;
        }
      }
      else
      {
        throw new Exception("cannot register structure " + tstag + " : structure is unknown");
      }

    }
    private TTaggedStruct() { }
    private static TTaggedStruct instance = null;
    public static int RegisterStructure(object tstruct,string tstag)
    {
      if (tstruct == null) return Errors.argument_list_error;
      if (instance == null) instance = new TTaggedStruct(); // create an empty one
      Type t = (tstruct is Array) ? ((Array)tstruct).GetType().GetElementType() : tstruct.GetType();
      string cn = instance.GetClassName();
      if (cn != null && cn.CompareTo(t.Name) == 0) return 0; // done (same as current instance)
      return instance.fixStructLayout(tstruct, tstag);
    }
    public static int RegisterStructure(object tstruct)
    {
      return RegisterStructure(tstruct, null);
    }
    private int fixStructLayout(object tstruct, string stag)
    {
      if (tstruct == null) return Errors.argument_list_error;
      Type stype = tstruct.GetType();
      int siz = 1;
      if (tstruct is Array)
      {
        siz = ((Array)tstruct).Length;
        stype = ((Array)tstruct).GetType().GetElementType();
      }
      className = stype.Name;
      tag = (stag != null && stag.Length > 0) ? stag : stype.Name;
      int fsiz = 1;
      Type ftype;
      FieldInfo[] fields = stype.GetFields();
      int off0, off1, fmt;
      for (int i = 0; i < fields.Length; i++)
      {
        ftype = fields[i].FieldType;
        fmt = TDataType.GetTineFormat(ftype);
        if (fields[i].FieldType.IsArray)
        {
          ftype = ftype.GetElementType();
          fmt = TDataType.GetTineFormat(ftype);
          off0 = (int)Marshal.OffsetOf(fields[i].DeclaringType, fields[i].Name);
          off1 = i < fields.Length - 1 ? ((int)Marshal.OffsetOf(fields[i + 1].DeclaringType, fields[i + 1].Name)) : Marshal.SizeOf(stype);
          fsiz = (off1 - off0) / Formats.GetSizeInBytes(fmt);
        }
        else
        {
          fsiz = 1;
        }
        AddField(fsiz, fmt, fields[i].Name);
      }
      return Seal(Marshal.SizeOf(stype), siz);
    }
    private int addrOffset = 0;
    public int AddField(Int32 siz, Int32 fmt, string field)
    {
      int cc = AddField(tag, addrOffset, siz, fmt, field);
      if (cc == 0) addrOffset += siz * Formats.GetSizeInBytes(fmt);
      return cc;
    }
    public int Seal(Int32 sizeInBytes,Int32 number)
    {
      //addrOffset = 0;
      return SealTaggedStruct(tag, sizeInBytes, number);
    }
    private static int AddField(string tag,Int32 addr,Int32 siz,Int32 fmt,string field)
    {
      fmt = Formats.LFMT_OFFSET + (fmt%256);
      return native.AddFieldToStruct(tag,addr,siz,fmt,field);
    }
    private static int SealTaggedStruct(string tag,Int32 sizeInBytes,Int32 number)
    {
      return native.SealTaggedStruct(tag, sizeInBytes, number);
    }
  }
  /// <summary>
  /// TEquipmentModule is the a TINE server's equipment module class.  
  /// </summary>
  /// <remarks>
  /// An equipment module represents a device server.  This is ideally a class which manages
  /// like instances of a set of devices.  The equipment module is responsible for
  /// handling dispatch actions pertaining to registered properties ('READ' or 'WRITE'
  /// actions). 
  /// </remarks>
  public class TEquipmentModule
  {
    private CallHandler eqmHndlr;
    private native.TEqmFcn eqmFcn;
    private FcnDispatch eqmIniDispatch;
    private native.TEqmHook eqmIni;
    private FcnDispatch eqmBkgDispatch;
    private native.TEqmHook eqmBkg;
    private FcnDispatch eqmExiDispatch;
    private native.TEqmHook eqmExi;
    private string eqmName;
    public unsafe Int32 lclHndlr(string dev, string prp, ref DTYPE dout, ref DTYPE din, UInt16 acc, IntPtr eqmRef)
    {
      TDataType dtdout = new TDataType(dout);
      TDataType dtdin = new TDataType(din);
      dtdout.SetBounded(true);
      Int32 cc = eqmHndlr(dev, prp, dtdout, dtdin, acc, this);
      dout.dArrayLength = dtdout.GetDataArrayLength();
      dout.dTimeStamp = dtdout.GetDataTimeStamp();
      dout.dStamp = dtdout.GetUserStamp();
      dout.sysStamp = dtdout.GetSystemStamp();
      return cc;
    }
    public unsafe void lclIni(IntPtr eqmRef)
    {
      eqmIniDispatch();
    }
    public unsafe void lclBkg(IntPtr eqmRef)
    {
      eqmBkgDispatch();
    }
    public unsafe void lclExi(IntPtr eqmRef)
    {
      eqmExiDispatch();
    }
    /// <summary>
    /// Primary constructor for an equipment module object.
    /// </summary>
    /// <remarks>
    /// A FEC (front-end controller) process can contain more than one equipment module (which is exported
    /// as a device server or property server).  However, a FEC process should contain at least one 
    /// equipment module in order to be a useful element in the control system.
    /// </remarks>
    /// <param name="exp">is the equipment module's exported device or property server name 
    /// (maximum 32 characters).  This name must be unique within a given context.
    /// If a 'null' parameter is passed, then the exported name must be given by an underlying 
    /// configuration file such as 'export.csv' or 'fec.xml'</param>
    /// <param name="eqm">is the equipment module's internal 'local' name (maximum 6-characters).
    /// This name must be unique within the process only, and is used for internal de-referencing.
    /// </param>
    /// <param name="ndev">gives the maximum device capacity (maximum number of device instances
    /// supported by this device server.</param>
    /// <param name="hndlr">is the equipment module's dispatch call handler.  This must be a function
    /// of the form:
    /// int CallHandler(string dev, string prp, TDataType dout, TDataType din, UInt16 acc, TEquipmentModule eqm);
    /// </param>
    /// <param name="ini">is an optional initialization routine for the equipment module.  This must be
    /// a function of the form:
    /// void FcnDispatch();
    /// </param>
    /// <param name="bkg">is an optional background routine for the equipment module, which will be called
    /// at the given background task interval.  This must be
    /// a function of the form:
    /// void FcnDispatch();
    /// </param>
    /// <param name="bkgInterval">is the background task interval in play if a background task is registered.</param>
    /// <param name="exi">is an optional background routine for the equipment module, which will be called
    /// at the given background task interval.  This must be
    /// a function of the form:
    /// void FcnDispatch();
    /// </param>
    /// \b Example:
    /// \include eg_RegisterEqm.cs
    public TEquipmentModule(string exp, string eqm, Int32 ndev, CallHandler hndlr, FcnDispatch ini, FcnDispatch bkg, Int32 bkgInterval, FcnDispatch exi)
    {
      eqmHndlr = hndlr;
      eqmFcn = lclHndlr;
      eqmIniDispatch = ini;
      eqmIni = lclIni;
      eqmBkgDispatch = bkg;
      eqmBkg = lclBkg;
      eqmExiDispatch = exi;
      eqmExi = lclExi;
      eqmName = (string)eqm.Clone();
      native.RegisterEquipmentModuleEx(exp, eqmName, (UInt32)ndev, eqmFcn, eqmIni, eqmBkg, bkgInterval, eqmExi, IntPtr.Zero);
    }
    /// <summary>
    /// Assigns pertinent information for the specified property. 
    /// </summary>
    /// <remarks>
    /// Servers export their behavior through Properties, which can be thought of as 'properties' 
    /// in the Object Oriented sense with Get and Set methods. That is, any read or write access 
    /// for a given property will call into the associated equipment module and be handled there. 
    /// The general behavior of the property (including descriptions) can be specified with this call. 
    /// Properties can also be (and frequently are) registered via the presence of an 'exports.csv' 
    /// startup database file.
    /// Note: Properties which are overloaded can and should be registered more than once with the 
    /// appropriate characteristics supplied in the registration call.
    /// </remarks>
    /// <param name="property">Property name in question (up to 32 characters, preferably 16 or less).</param>
    /// <param name="dout">Atemplate data object giving the default format size, and structure tag 
    /// (if any) of the output data set to be returned by the server to the caller. </param>
    /// <param name="din">A template data object giving the default format size, and structure tag 
    /// (if any) of the input data set to be supplied by the caller. </param>
    /// <param name="acc">the data 'access' (e.g. CA_READ or CA_WRITE) accepted by the server. This parameter can 
    /// be used to coerce client-side links into using preferred link attributes. For example, by applying the CA_NETWORK 
    /// flag (CA_READ|CA_NETWORK) to the access, asynchronous NETWORK (i.g. multicast acquisition) will be enforced at the 
    /// client side. Applying the CA_XREAD access flag will likewise allow exclusive READ access regarding the registered property 
    /// (i.e. the same security rules applied to CA_WRITE will be in force for READ acquisitions, either permanently - no CA_READ flag, 
    /// or temporarily in case of an applied access lock - if both CA_READ and CA_XREAD are applied). Applying the CA_SAVERESTORE flag 
    /// will apply the internal save-and-restore logic to the registered property (restoring the last property settings upon server 
    /// startup, and saving any changed property settings).</param>
    /// <param name="atype">Data 'array type' (AT_SCALAR, AT_SINGLE, etc.) associated with the returned data.</param>
    /// <param name="rowLength">if > 0 gives the row length of the returned array data. This is most relevant for double arrays, where 
    /// the number of rows is then given by the output data size divided by the row length.</param>
    /// <param name="dsc">The 64-character description of what the property reads or writes. </param>
    /// <param name="propId">The property identifier to be associated with the property name returned in calls to GetPropertyId()</param>
    /// <param name="rdr">A redirection string, if the property is to be redirected to another server. 
    /// Most properties are not redirected, and this parameter is usually NULL. If used, it will be parsed according to 
    /// device server/device name (device property), where the 'device name' and 'device property' are optional. 
    /// If 'device name' is not present, all devices will be redirected to the specified device server. If 'device property' 
    /// is not present then the identical property name as registered will be redirected to the device server specified. 
    /// If on the other hand, the device property is present in the string to be parsed, then the property registered 
    /// will be redirected to the device server specified as well as to the device property specified.</param>
    /// <returns>The (positive) associated property identifier if successful, otherwise the negative of a TINE return code.</returns>
    /// \b Example:
    /// \include eg_RegisterProperty.cs
    /// 
    public int RegisterPropertyInformation(string property, TDataType dout, TDataType din, UInt16 acc, UInt16 atype, int rowLength, string dsc, int propId, string rdr)
    {
      return TKernel.RegisterPropertyInformation(eqmName, property, dout, din, acc, atype, (ushort)rowLength, dsc, propId, rdr);
    }
    /// <summary>
    /// Assigns a device name to the specified device number.
    /// </summary>
    /// <remarks>
    /// Many servers handle instances of a given device type, such as a set of vacuum pumps, 
    /// beam position monitors, magnet power supply controllers, etc. It is frequently very efficient 
    /// to code in terms of device numbers which might be an index into an array, but to specify modules 
    /// in calls by their human-readable names. For such purposes a device name can be assigned to a number 
    /// at initialization via a call to RegisterDeviceName() (alias: RegisterModuleName()). An alternative 
    /// is to provide a startup database file 'devices.csv' containing a cross-reference for numbers and names. 
    /// Internally, a hash table is maintained for fast lookups inside equipment module routines.
    /// </remarks>
    /// <param name="devName">Up to 64 characters device name to be assigned</param>
    /// <param name="devNr">The device number associated with the device name specified.</param>
    /// <returns>0 if successful, otherwise a TINE completion code</returns>
    /// \b Example:
    /// \include eg_RegisterDevice.cs
    public int RegisterDeviceName(String devName, Int32 devNr)
    {
      return TKernel.RegisterDeviceName(eqmName, devName, devNr);
    }
    public int AssignNumberDevices(Int32 num)
    {
      return native.AssignNumDevices(eqmName, num);
    }
    /// <summary>
    /// Inserts an alarm into the local alarm server table.
    /// </summary>
    /// <remarks>
    /// A server can set an alarm for a specific device by giving its device number. 
    /// For further information, please see the discussion of the local alarm server.
    /// </remarks>
    /// <param name="devNr">is the device number of the device for which the alarm is to be set.</param>
    /// <param name="almCode">is the registered alarm code which defines the alarm.</param>
    /// <returns></returns>
    public int SetAlarm(Int32 devNr, Int32 almCode)
    {
      return native.SetAlarmEx(eqmName, (Int16)devNr, almCode, IntPtr.Zero, (byte)0);
    }
    /// <summary>
    /// Inserts an alarm into the local alarm server table.
    /// </summary>
    /// <remarks>
    /// A server can set an alarm for a specific device by giving its device number. 
    /// This methods also allows passing associated alarm data and specific alarm flags.
    /// For further information, please see the discussion of the local alarm server.
    /// </remarks>
    /// <param name="devNr">is the device number of the device for which the alarm is to be set.</param>
    /// <param name="almCode">is the registered alarm code which defines the alarm.</param>
    /// <param name="almData">is (up to 64-byte) the optional alarm data set, which supplies additional information for the alarm.</param>
    /// <param name="almFlags">is the optional set of alarm flags which is used in the alarm descriptor information.</param>
    /// <returns></returns>
    public int SetAlarm(Int32 devNr, Int32 almCode, object almData, byte almFlags)
    {
      IntPtr ptr;
      if (almData is string)
      {
        char[] cdata = ((string)almData).ToCharArray();
        byte[] bdata = new byte[64];
        for (int i = 0; i < 64 && i < cdata.Length; i++) bdata[i] = (byte)cdata[i];
        ptr = TDataType.GetDataPtr(bdata);
      }
      else
      {
        ptr = TDataType.GetDataPtr(almData);
      }
      return native.SetAlarmEx(eqmName, (Int16)devNr, almCode, ptr, almFlags);
    }
    /// <summary>
    /// Inserts an alarm into the local alarm server table.
    /// </summary>
    /// <remarks>
    /// A server can set an alarm for a specific device by giving its device name
    /// (as opposed to its device number).
    /// This methods also allows passing associated alarm data and specific alarm flags.
    /// For further information, please see the discussion of the local alarm server.
    /// </remarks>
    /// <param name="devName">is the device name for which the alarm is to be set.</param>
    /// <param name="almCode">is the registered alarm code which defines the alarm.</param>
    /// <param name="almData">is (up to 64-byte) the optional alarm data set, which supplies additional information for the alarm.</param>
    /// <param name="almFlags">is the optional set of alarm flags which is used in the alarm descriptor information.</param>
    /// <returns></returns>
    public int SetAlarm(string devName, Int32 almCode, object almData, byte almFlags)
    {
      IntPtr ptr;
      if (almData is string)
      {
        char[] cdata = ((string)almData).ToCharArray();
        byte[] bdata = new byte[64];
        for (int i = 0; i < 64 && i < cdata.Length; i++) bdata[i] = (byte)cdata[i];
        ptr = TDataType.GetDataPtr(bdata);
      }
      else
      {
        ptr = TDataType.GetDataPtr(almData);
      }
      return native.SetDeviceAlarm(eqmName, devName, almCode, ptr, almFlags);
    }
    /// <summary>
    /// Instructs the local alarm server table that the given alarm is to be cleared. 
    /// </summary>
    /// <remarks>
    /// A server can clear alarms for specific device by giving its device number. 
    /// Note that all alarms for the device specified are cleared. 
    /// Also note, that "clearing" an alarm means that its "clear" counter is incremented. 
    /// It is not removed immediately from the local alarm list. 
    /// For further information, please see the discussion of the local alarm server.
    /// </remarks>
    /// <param name="devNr">is the device number of the device for which all alarms is to be cleared.</param>
    /// 
    public void ClearAlarm(Int32 devNr)
    {
      native.ClearAlarm(eqmName,(Int16)devNr);
    }
    /// <summary>
    /// Instructs the local alarm server table that the given alarm is to be cleared. 
    /// </summary>
    /// <remarks>
    /// A server can clear alarms for specific device by giving its device name. 
    /// Note that all alarms for the device specified are cleared. 
    /// Also note, that "clearing" an alarm means that its "clear" counter is incremented. 
    /// It is not removed immediately from the local alarm list. 
    /// For further information, please see the discussion of the local alarm server.
    /// </remarks>
    /// <param name="devName">is the registered device name of the device for which all alarms is to be cleared.</param>
    public void ClearAlarm(string devName)
    {
      native.ClearDeviceAlarm(eqmName, devName);
    }
    /// <summary>
    /// Instructs the local alarm server table that the given alarm is to be marked for removal.
    /// </summary>
    /// <remarks>
    /// A server can remove alarms for specific device by giving its registered device name. 
    /// Note that only the alarm specified for the device specified is removed. 
    /// Also note, that "removing" an alarm means that termination bit will be set. 
    /// It is subsequently marked for removal from the local alarm list. 
    /// For further information, please see the discussion of the local alarm server.
    /// </remarks>
    /// <param name="devName"></param>
    /// <param name="almCode"></param>
    public void RemoveAlarm(string devName,Int32 almCode)
    {
      native.RemoveDeviceAlarm(eqmName, devName, almCode);
    }
    /// <summary>
    /// Inserts an alarm definition into the alarm definition table.
    /// </summary>
    /// <remarks>
    /// As an alternative to the 'local name'-alarms.csv configuration file, the front end 
    /// server can make use of this API call in order to fill in the alarm definition table 
    /// describing locally generated alarms. This is particularly useful for embedded platforms 
    /// where there is no file system, or where a TINE server is used as a translation layer and 
    /// needs to map a given alarm system onto the TINE alarm system.
    /// </remarks>
    /// <param name="ads">is a pointer to an Alarm Definition Structure (ADS) containing the alarm table information which is to be appended to the alarm definition table.</param>
    /// <returns>0 upon success, otherwise a TINE error code.</returns>
    public int AppendAlarmInfoTable(ref ADS ads)
    {
      return native.AppendAlarmInfoTable(eqmName,ref ads);
    }
    /// <summary>
    /// Inserts a property to be monitored into the local alarm server's Watch Table.
    /// </summary>
    /// <remarks>
    /// Certain alarms are to be set whenever the value of a property exceeds a defineable threshold. 
    /// Such alarms can be managed automatically by the local alarm server if the alarm criteria are 
    /// entered into the alarm watch table. 
    /// This can be achieved by calling this routine (or supplying a startup configuration file 
    /// almwatch.csv).
    /// </remarks>
    /// <param name="prp">is the property which is to be called by the local alarm server.</param>
    /// <param name="dev">is the device name associated with the property to be called by the local alarm server.</param>
    /// <param name="siz">is the data array size to be called by the local alarm server.</param>
    /// <param name="fmt">is the TINE data format to be called by the local alarm server.</param>
    /// <param name="atyp">is the TINE data array type to be applied to the property called by the local alarm server.</param>
    /// <param name="sev">is the severity of the alarm issued when the data returned by the call exceed the given thresholds.</param>
    /// <param name="sys">is the alarm system identifier to be associated with the alarm.</param>
    /// <param name="thr">is an ALM_THRESHOLDS object specifying the threshold criteria for setting the alarm.</param>
    /// <returns>0 upon success, otherwise a TINE error code.</returns>
    public int AppendAlarmWatchTable(string prp, string dev, Int32 siz, Int32 fmt, Int32 atyp, Int32 sev, Int32 sys, ref ALM_THRESHOLDS thr)
    {
      return native.AppendAlarmWatchTable(eqmName, prp, dev, siz, fmt, atyp, sev, sys, ref thr);
    }
    /// <summary>
    /// Inserts a local history element into the local history server. 
    /// </summary>
    /// <remarks>
    /// A server can instruct the local history server to keep a history of the given property by utilizing this call. The local history server will periodically call the property as specified according to the following input parameters.
    ///</remarks> 
    /// <param name="prp">is the requested property for which a history is to be kept.</param>
    /// <param name="dev">is the device name to be associated with the property name supplied as the second parameter.</param>
    /// <param name="len">is the length of the local history call.</param>
    /// <param name="fmt">is the TINE format of the local history call.</param>
    /// <param name="idx">is the local history index to be identified with this local history element (Note: this must be unique with this server process and within the history data repository).</param>
    /// <param name="hspec">is a pointer to a HistorySpecification structure, where the relevant filtering criteria are given.</param>
    /// <returns>0 upon success, otherwise a TINE error code.</returns>
    /// \note In lieu of calling AppendHistoryInformation(), it is recommended to make use of the startup configuration file 'history.csv', in which all local history critieria can be introduced. This has the advantage that no information is hard-coded. Indeed by introducing a 'history.csv' file one can immediately instruct a server to maintain local history information (even if the server is developed and maintained by someone else!).
    /// 
    /// \b Example:
    /// \include eg_AppendHistory.cs
    public int AppendHistoryInformation(string prp, string dev, Int32 len, Int32 fmt, Int32 idx, ref HistorySpecification hspec)
    {
      return native.AppendHistoryInformation(eqmName, prp, dev, len, fmt, idx, ref hspec);
    }

    public int AppendHistoryData(string prp, string dev, TDataType d)
    {
      if (d != null)
      {
        DTYPE din = d.GetDTYPE();
        d.PutData(d.GetDataObject());
        return native.AppendHistoryData(eqmName, prp, dev, ref din);
      }
      return Errors.invalid_parameter;
    }
    
    public delegate int CallHandler(string dev, string prp, TDataType dout, TDataType din, UInt16 acc, TEquipmentModule eqm);
    public delegate void FcnDispatch();
    /// <summary>
    /// Returns the registered device number for the device name given.
    /// </summary>
    /// <param name="device">is the device name whose assigned device number is desired.
    /// If the device name entered does not correspond to a registered device the value
    /// '-1' is returned.
    /// </param>
    /// <returns>The corresponding device number or '-1' on error.</returns>
    public Int32 GetDeviceNumber(string device) { return native.GetDeviceNumber(eqmName, device); }
    /// <summary>
    /// Returns the registered device number for the device name and property given.
    /// </summary>
    /// <remarks>
    /// A pure device server will maintain a list of registered devices, each of which may support
    /// any particular given property.  In such cases the 'property' parameter is irrelevant, as any
    /// associated device number is independent of the property.
    /// 
    /// On the other hand, a 'property' server will treat properties as the more relevant identifier,
    /// whereby each property may support an independent set of 'device' names (or more probably 
    /// 'keyword' names).  In such cases the property name entered is of paramount importance in
    /// determining the registered device (or keyword) number.
    /// </remarks>
    /// <param name="device">is the device name whose assigned device number is desired.
    /// If the device name entered does not correspond to a registered device the value
    /// '-1' is returned.
    /// </param>
    /// <param name="property">is the property whose device set is being queried.</param>
    /// <returns>The corresponding device number or '-1' on error.</returns>
    /// \b Example:
    /// \include eg_GetPropId.cs
    public Int32 GetDeviceNumber(string device,string property) { return native.GetDeviceNumberEx(eqmName, device, property); }
    /// <summary>
    /// Returns the registered property identification number for the property entered.
    /// </summary>
    /// <param name="property">is the property whose assigned property identification is desired</param>
    /// <returns>The property identication number of -1 on error.</returns>
    public Int32 GetPropertyId(string property) { return native.GetPropertyId(eqmName, property); }
    public unsafe Int32 GetCallerInformation(ClnInfo[] clnInfo, ref int num) 
    { 
     // return native.GetCallerInformation(eqmName, clnInfo, ref num); 
      return 0;
    }
    public unsafe ClnInfo GetCallerInformation()
    {
      int num = 1;
      ClnInfo ci = new ClnInfo();
      int cc = native.GetCallerInformation(eqmName, ref ci, ref num);
      return ci;
    }
    /// <summary>
    /// Schedules the given property for immediate delivery to all attached clients according scope specified.
    /// </summary>
    /// <remarks>
    /// When clients are attached to a particular property, they have specified a polling rate, which 
    /// defines the maximum latency for receiving updates at the client side. This is usually fine. 
    /// However, if a server knows that important data have changed and which properties depend on these 
    /// data, it can signal the scheduler to call the given properties immediately and send the data to 
    /// all attached callers, regardless of their registered polling rates. In this way, a server can 
    /// signal an event to its attached clients. In this extended call, the caller can specify the 
    /// 'scope' of the call : CA_NETWORK (all listening clients), CA_HIST (the local history subsystem) 
    /// or CA_ALARM (the local alarm server).
    /// </remarks>
    /// <param name="property">is (comma separated) list of properties which are to be scheduled.</param>
    /// <param name="scope">is any of CA_NETWORK, CA_HIST, or CA_ALARM ORed together.</param>
    /// <returns>0 upon success or a TINE error code.</returns>
    public Int32 ScheduleProperty(string property,Int32 scope) { return native._SystemSchedulePropertyEx(eqmName, property, scope); }
    /// <summary>
    /// Schedules the given property for immediate delivery to all attached clients.
    /// </summary>
    /// <remarks>
    /// When clients are attached to a particular property, they have specified a polling rate, which 
    /// defines the maximum latency for receiving updates at the client side. This is usually fine. 
    /// However, if a server knows that important data have changed and which properties depend on these 
    /// data, it can signal the scheduler to call the given properties immediately and send the data to 
    /// all attached callers, regardless of their registered polling rates. In this way, a server can 
    /// signal an event to its attached clients. In this simple call, clients from all listening
    /// subsystems will be scheduled (i.e. the local history and alarm subsystem as well as any remote
    /// clients).
    /// </remarks>
    /// <param name="property">is (comma separated) list of properties which are to be scheduled.</param>
    /// <returns></returns>
    public Int32 ScheduleProperty(string property) 
    { 
      return native._SystemSchedulePropertyEx(eqmName, property, Access.CA_NETWORK | Access.CA_HIST | Access.CA_ALARM); 
    }
    /// <summary>
    /// Sets the error string to accompany the current server call
    /// </summary>
    /// <remarks>
    ///  * For user-defined errors codes (equal to or above 512) the accompanying error string
    /// should be defined by calling this routine prior to returning from the equipment module
    /// call.  When system error codes such as 'illegal_format' are used, a call to SetEqmCompletion() 
    /// is not necessary.
    /// </remarks>
    /// <param name="statusString">
    /// is the 96-character string which is to accompany the outgoing error code.
    /// </param>
    public void SetCompletion(string statusString)
    {
      native.SetEqmCompletion(eqmName, statusString);
    }
  }
}
